/*JS*********************************************************************
*
*    File    : GENMALLOC.H
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Generate allocation routines.
*
*              Copyright (C) 2004 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

#ifndef GENMALLOC_TYPE
# error "genmalloc.h": Macro "GENMALLOC_TYPE"  undefined!
#endif

#ifndef GENMALLOC_NAMEALLOC
# error "genmalloc.h": Macro "GENMALLOC_NAMEALLOC"  undefined!
#endif

#ifndef GENMALLOC_MEMSIZE
# define GENMALLOC_MEMSIZE     4096
#endif

/*************************************************************************/

#include <stdlib.h>

#define _intern_GENMALLOC_CONCAT(a,b)   _intern_GENMALLOC_CONCAT2(a,b)
#define _intern_GENMALLOC_CONCAT2(a,b)   a##b

/* ***  Define local type and variable names  *** */
#define _intern_GENMALLOC_UNION \
	_intern_GENMALLOC_CONCAT(GENMALLOC_NAMEALLOC,union_t)
#define _intern_GENMALLOC_MEMTYPE \
	_intern_GENMALLOC_CONCAT(GENMALLOC_NAMEALLOC,memtype_t)
#define _intern_GENMALLOC_VARMEM \
	_intern_GENMALLOC_CONCAT(_intern_VarMem,GENMALLOC_NAMEALLOC)
#define _intern_GENMALLOC_VARDEL \
	_intern_GENMALLOC_CONCAT(_intern_VarDel,GENMALLOC_NAMEALLOC)

/*************************************************************************/

/* ***  Define local types  *** */
union _intern_GENMALLOC_UNION {
  union _intern_GENMALLOC_UNION *_intern_GENMALLOC_NextPtr;
  GENMALLOC_TYPE                 _intern_GENMALLOC_Node;
};

struct _intern_GENMALLOC_MEMTYPE {
  struct _intern_GENMALLOC_MEMTYPE *_intern_GENMALLOC_Next;
  long                              _intern_GENMALLOC_Index;
  union _intern_GENMALLOC_UNION     _intern_GENMALLOC_Mem[GENMALLOC_MEMSIZE];
};

/* ***  Define static variables  *** */
static struct _intern_GENMALLOC_MEMTYPE *_intern_GENMALLOC_VARMEM;
static union _intern_GENMALLOC_UNION    *_intern_GENMALLOC_VARDEL;

/*************************************************************************/

/* ***  Allocation routine  *** */
#ifdef GENMALLOC_STATIC
static
#endif
GENMALLOC_TYPE *GENMALLOC_NAMEALLOC(void)
{
  struct _intern_GENMALLOC_MEMTYPE *cMem;
  union _intern_GENMALLOC_UNION *node;

  if((node = _intern_GENMALLOC_VARDEL)) {
    /*  Take from deleted ones  */
    _intern_GENMALLOC_VARDEL = node->_intern_GENMALLOC_NextPtr;
    return &(node->_intern_GENMALLOC_Node);
  }

  /*  Any nodes in global memory?  */
  for(cMem = _intern_GENMALLOC_VARMEM ; cMem ;
	 cMem = cMem->_intern_GENMALLOC_Next)
    if(cMem->_intern_GENMALLOC_Index < GENMALLOC_MEMSIZE) {
      node = &(cMem->_intern_GENMALLOC_Mem[cMem->_intern_GENMALLOC_Index++]);
      return &(node->_intern_GENMALLOC_Node);
    }

  /*  Get new chunk from system  */
  if((cMem = (struct _intern_GENMALLOC_MEMTYPE *)
	malloc(sizeof(*cMem))) == NULL)
    return NULL;

  node = &(cMem->_intern_GENMALLOC_Mem[0]);
  cMem->_intern_GENMALLOC_Index = 1;

  /*  Hang in front of global list  */
  cMem->_intern_GENMALLOC_Next = _intern_GENMALLOC_VARMEM;
  _intern_GENMALLOC_VARMEM = cMem;

  return &(node->_intern_GENMALLOC_Node);
}

/*************************************************************************/

/* ***  Free routine  *** */
#ifdef GENMALLOC_NAMEFREE
#ifdef GENMALLOC_STATIC
static
#endif
void GENMALLOC_NAMEFREE(GENMALLOC_TYPE *node)
{
  ((union _intern_GENMALLOC_UNION *)node)->_intern_GENMALLOC_NextPtr =
    _intern_GENMALLOC_VARDEL;

  _intern_GENMALLOC_VARDEL = (union _intern_GENMALLOC_UNION *)node;
}
#undef GENMALLOC_NAMEFREE
#endif

/*************************************************************************/

/* ***  Clean up memory  *** */
#ifdef GENMALLOC_NAMECLEANMEM
#ifdef GENMALLOC_STATIC
static
#endif
void GENMALLOC_NAMECLEANMEM(void)
{
  struct _intern_GENMALLOC_MEMTYPE *cMem;

  for(cMem = _intern_GENMALLOC_VARMEM ; cMem ;
      cMem = cMem->_intern_GENMALLOC_Next)
    cMem->_intern_GENMALLOC_Index = 0;

  _intern_GENMALLOC_VARDEL = NULL;
}

#undef GENMALLOC_NAMECLEANMEM
#endif

/*************************************************************************/

/* ***  Free up memory  *** */
#ifdef GENMALLOC_NAMEFREEMEM
#ifdef GENMALLOC_STATIC
static
#endif
void GENMALLOC_NAMEFREEMEM(void)
{
  struct _intern_GENMALLOC_MEMTYPE *cMem,*cNext;

  for(cMem = _intern_GENMALLOC_VARMEM ; cMem ; cMem = cNext) {
    cNext = cMem->_intern_GENMALLOC_Next;
    free(cMem);
  }

  _intern_GENMALLOC_VARMEM = NULL;
  _intern_GENMALLOC_VARDEL = NULL;
}

#undef GENMALLOC_NAMEFREEMEM
#endif

/*************************************************************************/

/* ***  Clean up macro definitions  *** */
#undef _intern_GENMALLOC_CONCAT
#undef _intern_GENMALLOC_CONCAT2
#undef _intern_GENMALLOC_MEMTYPE
#undef _intern_GENMALLOC_VARMEM
#undef _intern_GENMALLOC_VARDEL
#undef _intern_GENMALLOC_UNION
#undef _intern_GENMALLOC_MEMTYPE

#undef GENMALLOC_TYPE
#undef GENMALLOC_NAMEALLOC
#undef GENMALLOC_MEMSIZE
#undef GENMALLOC_STATIC
