#ifndef JEDLIB_H
/*JS*********************************************************************
*
*    Program : JEDLIB
*    Language: ANSI-C
*    Author  : Joerg Schoen
*
*    Part    : Header file for editing library
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

#define JEDLIB_H 1

#if defined(__alpha) || defined(__hpux)
#define _XOPEN_SOURCE_EXTENDED
#endif

#include <curses.h>

#include <cursesconf.h>

/* ***********	 Extracted definitions of library code  ********* */
/* ----------- FILE: jedapps.c            ------------ */
#include <stdio.h> /*  needed for "FILE" definition  */
#define FOFI_MPOINTFILES   (1<<4)

/* ----------- FILE: jedframe.c           ------------ */
/* ***  Definitions for text frames  *** */
#define TFRAME_INSMODE         (1<<0) /*  Insert characters, not overwrite  */
#define TFRAME_SCROLLBAR       (1<<1) /*  Display scroll bar  */
#define TFRAME_ALLOCATE        (1<<2) /*  Reallocate text region if growing  */
#define TFRAME_CHANGED         (1<<3) /*  Mark that something was changed  */
#define TFRAME_FILLBLANK       (1<<4) /*  Fill with blanks (instead of
				       *   'TFFillCharacter').
				       */
#define TFRAME_NONEWLINE       (1<<5) /*  Do not treat newline, return
				       *   'unknown'.
				       */

typedef struct {
  int TF1_LineNr;
  long TF1_LineOff;
} jedtextline_t;

typedef struct {
  int TF_Mode;

  /*  The text to display, its length, maximum length and number of lines  */
  char *TF_Text;
  long TF_Length, TF_MaxLen, TF_NLines;

  int TF_Line, TF_Column, TF_Height, TF_Width;
  int TF_TopL, TF_BottL;
  int TF_CuLine, TF_CuColumn; /*  cursor position  */
  long TF_CuOffset; /*  cursor position as offset  */
  long TF_CuMark; /*  current text mark  */

  int TF_ScrollBarS, TF_ScrollBarE; /*  start and end of scrollbar  */

  long TF_StartOff; /*  offset of start of first screen line  */
  jedtextline_t TF_Lines[1]; /*  line numbers/offset pairs  */
} jedtext_t;

#define jedtextdelete(tf)     free(tf)

/* ----------- FILE: jedhist.c            ------------ */
#define JEDLIB_HASHISTORY

/* ----------- FILE: jedlgraph.c          ------------ */
/* ***  Double line graphics. Only work on linux console  *** */
/*  single lines horizontally and single lines vertically  */
#define ACS_ULCORNER11     JEDACSMap[0x06]
#define ACS_LLCORNER11     JEDACSMap[0x03]
#define ACS_URCORNER11     JEDACSMap[0x0c]
#define ACS_LRCORNER11     JEDACSMap[0x09]
#define ACS_RTEE11         JEDACSMap[0x0d]
#define ACS_LTEE11         JEDACSMap[0x07]
#define ACS_BTEE11         JEDACSMap[0x0b]
#define ACS_TTEE11         JEDACSMap[0x0e]
#define ACS_HLINE11        JEDACSMap[0x0a]
#define ACS_VLINE11        JEDACSMap[0x05]
#define ACS_PLUS11         JEDACSMap[0x0f]

/*  double lines horizontally and single lines vertically  */
#define ACS_ULCORNER21     JEDACSMap[0x16]
#define ACS_LLCORNER21     JEDACSMap[0x13]
#define ACS_URCORNER21     JEDACSMap[0x1c]
#define ACS_LRCORNER21     JEDACSMap[0x19]
#define ACS_RTEE21         JEDACSMap[0x1d]
#define ACS_LTEE21         JEDACSMap[0x17]
#define ACS_BTEE21         JEDACSMap[0x1b]
#define ACS_TTEE21         JEDACSMap[0x1e]
#define ACS_HLINE21        JEDACSMap[0x1a]
#define ACS_VLINE21        JEDACSMap[0x15]
#define ACS_PLUS21         JEDACSMap[0x1f]

/*  single lines horizontally and double lines vertically  */
#define ACS_ULCORNER12     JEDACSMap[0x26]
#define ACS_LLCORNER12     JEDACSMap[0x23]
#define ACS_URCORNER12     JEDACSMap[0x2c]
#define ACS_LRCORNER12     JEDACSMap[0x29]
#define ACS_RTEE12         JEDACSMap[0x2d]
#define ACS_LTEE12         JEDACSMap[0x27]
#define ACS_BTEE12         JEDACSMap[0x2b]
#define ACS_TTEE12         JEDACSMap[0x2e]
#define ACS_HLINE12        JEDACSMap[0x2a]
#define ACS_VLINE12        JEDACSMap[0x25]
#define ACS_PLUS12         JEDACSMap[0x2f]

/*  double lines horizontally and double lines vertically  */
#define ACS_ULCORNER22     JEDACSMap[0x36]
#define ACS_LLCORNER22     JEDACSMap[0x33]
#define ACS_URCORNER22     JEDACSMap[0x3c]
#define ACS_LRCORNER22     JEDACSMap[0x39]
#define ACS_RTEE22         JEDACSMap[0x3d]
#define ACS_LTEE22         JEDACSMap[0x37]
#define ACS_BTEE22         JEDACSMap[0x3b]
#define ACS_TTEE22         JEDACSMap[0x3e]
#define ACS_HLINE22        JEDACSMap[0x3a]
#define ACS_VLINE22        JEDACSMap[0x35]
#define ACS_PLUS22         JEDACSMap[0x3f]

#define lgmvwaddch(win, y, x, ch) (wmove(win, y, x) == ERR ? ERR : lgwaddch(win, ch))

/* ----------- FILE: jedmenu.c            ------------ */
typedef struct jedmenufld_t jedmenufld_t;
struct jedmenufld_t {
  unsigned long MF_Type;
  jedmenufld_t    *MF_Next;

  char         *MF_Text;
  union {
    int         mf_Value;
    jedmenufld_t  *mf_SubMenu;
    long       *mf_Variable;
#define MF_Value      c.mf_Value
#define MF_SubMenu    c.mf_SubMenu
#define MF_Variable   c.mf_Variable
  } c;
  char         *MF_Help;
};

/*  For the type field  */
#define MENU_NMASK       ((1UL<<24)-1)
#define MENU_TMASK       (0xffUL<<24)
#define MENU_TEXT        (1UL<<24)
#define MENU_SUBMENU     (2UL<<24)
#define MENU_VARIABLE    (3UL<<24)

/*  For routine editMenu  */
#define MENUM_PRINTONLY  (1 << 1)
#define MENUM_VERTICAL   (1 << 2)
#define MENUM_NEWWIN     (1 << 3)
#define MENUM_FRAME      (1 << 4)
#define MENUM_MARKSUB    (1 << 5)

/* ----------- FILE: jedtable.c           ------------ */
/*  Lower byte reserved for return value  */
#define EDTAB_MASK         ((1UL << 8) - 1)
#define EDTAB_TOUCHED      (1UL << 8)
#define EDTAB_CHTITLE      (1UL << 9)
#define EDTAB_CHLINE       (1UL << 10)
#define EDTAB_CHTABLE      (1UL << 11)

/*  Modes for routine _tabformdokey  */
#define TABFORM_FNOREFRESH     (1<<1)
#define TABFORM_FREREADLINE    (1<<2)
#define TABFORM_FREREADTABLE   (1<<3)
#define TABFORM_FQUICKLOCATE   (1<<4)
#define TABFORM_FHIDECURRENT   (1<<5)

typedef struct {
  int TA_Mode;
  WINDOW *TA_Window;
  int TA_Line, TA_Column, TA_Height, TA_Width;

  int (*TA_GetLine)(void *uptr, long nr, char *dest, int m);
  long TA_Multi, TA_NLines;
  void *TA_UPtr;

  long TA_Start, TA_Current, TA_SeekIndex;
  int TA_ScrollBarS, TA_ScrollBarE;
  char TA_Lines[1];
} jedtable_t;

/* ----------- FILE: jedtree.c            ------------ */
#include <jsdllist.h>  /*  doubly linked lists  */

typedef struct jedtreenode_t jedtreenode_t;
struct jedtreenode_t {
  _jsdllnode_t JEDTN_Connect;

  long JEDTN_Flags;
  int JEDTN_SubLength;

  const char *JEDTN_Name;

  _jsdll_t JEDTN_Sub;

  void *JEDTN_User;
};

/* ***   For JEDTN_Flags  *** */
#define JEDT_TREE_FSUBNODES     (1UL << 0)
#define JEDT_TREE_FOPENED       (1UL << 1)
#define JEDT_TREE_FCALLBACK     (1UL << 2)
#define JEDT_TREE_FSELECTED     (1UL << 3)

#ifdef JEDLIB_INTERNAL
#define JEDT_TREE_FFREESTRING   (1UL << 8)
#define JEDT_TREE_FOPENING      (1UL << 9)
#endif

typedef struct jedtree_t jedtree_t;
struct jedtree_t {
  jedtreenode_t jedt_Content;
#define JEDT_Flags       jedt_Content.JEDTN_Flags
#define JEDT_NLines      jedt_Content.JEDTN_SubLength
#define JEDT_Title       jedt_Content.JEDTN_Name
#define JEDT_Nodes       jedt_Content.JEDTN_Sub
#define JEDT_User        jedt_Content.JEDTN_User

  WINDOW *JEDT_Window;
  int JEDT_Line, JEDT_Column, JEDT_Height, JEDT_Width;

  jedtreenode_t *JEDT_TopScreen;
  int JEDT_TopScreenLine, JEDT_TopScreenLvl;

  jedtreenode_t *JEDT_Current;
  int JEDT_CurrentLine;

  int JEDT_ColOff, JEDT_NCols;

  int (*JEDT_Setup)(void *user, jedtree_t *edtr, jedtreenode_t *node);
};

/*  For JEDT_Flags. Since the field might be misinterpreted for JEDTN_Flags,
 *   we use separate bits
 */
#define JEDT_TREEWIN_FDELWINDOW    (1UL << 16)
#define JEDT_TREEWIN_FFRAME        (1UL << 17)
#define JEDT_TREEWIN_FSCROLLBAR    (1UL << 18)
#define JEDT_TREEWIN_FLINENR       (1UL << 19)

/* **  Wrapper macros used internally to deal with doubly linked lists  *** */

#define _jedtdlln2node(nptr)   _jsdllnode2utype(nptr, jedtreenode_t, JEDTN_Connect)
#define _jedtdllh2node(hptr)   _jsdllnode2utype(hptr, jedtreenode_t, JEDTN_Sub)

#define jedtisempty(edtr)         _jsdllisempty(&(edtr->JEDT_Nodes))

#define jedtgetfirstnode(edtr)    _jedtdlln2node(_jsdllgetfirstnode(&(edtr->JEDT_Nodes)))
#define jedtgetlastnode(edtr)     _jedtdlln2node(_jsdllgetlastnode(&(edtr->JEDT_Nodes)))

#define	jedtgetnext(node)         _jedtdlln2node(_jsdllgetnext(&((node)->JEDTN_Connect)))
#define	jedtgetprev(node)         _jedtdlln2node(_jsdllgetprev(&((node)->JEDTN_Connect)))

#define jedtgetsubfirstnode(node) _jedtdlln2node(_jsdllgetfirstnode(&((node)->JEDTN_Sub)))
#define jedtgetsublastnode(node)  _jedtdlln2node(_jsdllgetlastnode(&((node)->JEDTN_Sub)))

#define jedtnodeistail(node)      _jsdllistail(&((node)->JEDTN_Connect))
#define jedtnodeishead(node)      _jsdllishead(&((node)->JEDTN_Connect))

#define jedtistreenode(node)      !_jsdllnodeinlist(&((node)->JEDTN_Connect))

#define jedthead2uppernode(node)  _jedtdllh2node(_jsdllhead2dll(&((node)->JEDTN_Connect)))
#define jedttail2uppernode(node)  _jedtdllh2node(_jsdlltail2dll(&((node)->JEDTN_Connect)))

/*  Modes for jedtreeaddnode etc.  */
#define JEDT_TREEADD_FADDHEAD      (1 << 0)
#define JEDT_TREEADD_FADDAFTER     (1 << 1)
#define JEDT_TREEADD_FADDBEFORE    (3 << 0)
#define JEDT_TREEADD_FDUPSTRING    (1 << 2)
#define JEDT_TREEADD_FKEEPDIRTY    (1 << 3)

#define JEDT_TREEADD_FKEEPSORT     (1 << 8)
#define JEDT_TREEADD_FCREATE       (1 << 9)
#define JEDT_TREEADD_FALLOWDUP     (1 << 10)


/* ----------- FILE: jedsubs.c            ------------ */
/*  For the mode variable  */
#define FOFI_MINITCURSES     (1<<0)
#define FOFI_MFRAME          (1<<1)
#define FOFI_MOWNWINDOW      (1<<2)
#define FOFI_NUMTABLE        (1<<3)  /* for edTable: display ordinals  */
#define FOFI_QUICKLOCATE     (1<<4)
#define FOFI_SCROLLBAR       (1<<5)

#ifdef JEDLIB_INTERNAL
#define _EDLIB_FCOLIND    0
#define _EDLIB_TCOLIND    4
#define _EDLIB_MCOLIND    7
#endif

/*  Macro to get number of color pair for a foreground/background combination */
#define GET_COLORPAIR(fg, bg)  (((fg) % 8) + 8 * ((bg) % 8) + 1)
#define GET_COLOR(fg, bg)      (((fg) & 8) ? \
			       A_BOLD | COLOR_PAIR(GET_COLORPAIR(fg, bg)) : \
			       COLOR_PAIR(GET_COLORPAIR(fg, bg)))

#define GET_COLORBG(color)    ((((color) - 1) / 8) % 8)
#define GET_COLORFG(color)    (((color) - 1) % 8)

#define CURSMODE_NOEXPLODE    (1 << 0)
#define CURSMODE_NOCOLORS     (1 << 1)
#define CURSMODE_NOMOUSE      (1 << 2)
#define CURSMODE_NOACSGRAPH   (1 << 3)

typedef struct {
  short KT_Old, KT_New;
} _jedkeytrans_t;

#ifdef JEDLIB_INTERNAL
/*  Standard editing keys as used by the routines  */
#define CNTRL(c)         ((c) & 0x1f)
#define EDKEY_REDRAW     CNTRL('L')

#define EDKEY_LEFT       KEY_LEFT
#define EDKEY_RIGHT      KEY_RIGHT
#define EDKEY_UP         KEY_UP
#define EDKEY_DOWN       KEY_DOWN
#define EDKEY_PUP        KEY_PPAGE
#define EDKEY_PDOWN      KEY_NPAGE

#define EDKEY_LSTART     KEY_HOME
#define EDKEY_LEND       KEY_END
#define EDKEY_START      CNTRL('A')
#define EDKEY_END        CNTRL('E')

#define EDKEY_DELETE     KEY_DC
#define EDKEY_BS         KEY_BACKSPACE

#define EDKEY_KILLEOL    CNTRL('K')
#define EDKEY_SETMARK    CNTRL(' ')
#define EDKEY_KILLREGION CNTRL('W')
#define EDKEY_YANK       CNTRL('Y')
#define EDKEY_ROTYANK    CNTRL('X')
#define EDKEY_TOGGLEINS  KEY_IC

#define EDKEY_PREV       KEY_F(7)
#define EDKEY_NEXT       KEY_F(8)

#define EDKEY_ENTER      KEY_F(1)

#define EDKEY_QUIT       '\033'

/*  Specifying line = col = height = width = 0
 *   gives the whole screen. line == -1 means
 *   centered vertically, col == -1 centered
 *   horizontally.
 */
#define _checkcoordinates(line, col, height, width) \
	do { \
	  if((line) == -1) (line) = (LINES - (height)) / 2; \
	  else if((line) < 0) (line) += LINES; \
	  if((height) <= 0) (height) += LINES - (line); \
          \
	  if((col) == -1) (col) = (COLS - (width)) / 2; \
	  else if((col) < 0) (col) += COLS; \
	  if((width) <= 0) (width) += COLS - (col); \
	} while(0)

#define _wcheckcoordinates(win, line, col, height, width) \
	do { \
	  int _h, _w; \
          \
	  getmaxyx((win), _h, _w); \
          \
	  if((line) == -1) (line) = (_h - (height)) / 2; \
	  else if((line) < 0) (line) += _h; \
	  if((height) <= 0) (height) += _h - (line); \
          \
	  if((col) == -1) (col) = (_w - (width)) / 2; \
	  else if((col) < 0) (col) += _w; \
	  if((width) <= 0) (width) += _w - (col); \
	} while(0)
#endif

/* ----------- FILE: jedform.c            ------------ */
typedef struct {
  unsigned long FF_Type;

  /*  Position on screen where to display  */
  int           FF_Line, FF_Column;

  /*  To restrict the field seen on the screen  */
  int           FF_ScreenHeight, FF_ScreenWidth;

  /*  The field contents  */
  union {
    char       *ff_Text;
    const char **ff_List;
  } c1;
#define FF_Text     c1.ff_Text
#define FF_List     c1.ff_List

  union {
    int         ff_Len;
    int         ff_Curr;
  } c2;
#define FF_Len   c2.ff_Len
#define FF_Curr  c2.ff_Curr

  union {
    int         ff_MaxLen;
    int         ff_NList;
  } c3;
#define FF_MaxLen   c3.ff_MaxLen
#define FF_NList    c3.ff_NList

  const char   *FF_Name;
  const char   *FF_Help;

  void         *FF_User;
} formfield_t;

/*  For the type field  */
#define FOFI_TMASK       0x000f0000UL
#define FOFI_LMASK       0x0000ffffUL

#define FOFI_TASCII      0x00010000UL /*  any character  */
#define FOFI_TINUM       0x00020000UL /*  solely digits  */
#define FOFI_TFNUM       0x00030000UL /*  floating point number  */
#define FOFI_TENUM       0x00040000UL
#define FOFI_TTEXT       0x00050000UL /*  text in multiple lines  */
#define FOFI_TBUTTON     0x00060000UL /*  lower bits give return value  */

#define FOFI_FNONSEL     0x00100000UL /*  field can not be selected  */
#define FOFI_FREFRESH    0x00200000UL /*  refresh always  */
#define FOFI_FEMPHASIZE  0x00400000UL /*  highlighted   */
#define FOFI_FSHOWSIGNS  0x00800000UL /*  Show '<' and '>' on long fields  */
#define FOFI_FALLOCATE   0x01000000UL /*  Reallocate field if necessary.  */
#define FOFI_FHIDDEN     0x02000000UL /*  Display field hidden  */
#define FOFI_FCHANGED    0x08000000UL /*  Set on exit if field changed */

typedef struct {
  int EF_LineOff, EF_ColOff;
  int EF_TopL, EF_BottL, EF_Width;
  int EF_ScrollBarS, EF_ScrollBarE;

  formfield_t *EF_Forms;
  int EF_FormHeight, EF_NForms;

  const _jedkeytrans_t *EF_KeyTrans;
  int EF_Mode, EF_CurrForm;
  WINDOW *EF_Window, *EF_InWin;

#ifdef NCURSES_MOUSE_VERSION
  int EF_PrevX, EF_PrevY;
#endif

  void *EF_FieldAux[1];
} jedform_t;


/* *******   Prototypes of library functions  ***** */
/*  For C++ compatibility   */
# ifdef __cplusplus
#  define Prototype extern "C"
# else
#  define Prototype extern
# endif
/* ***** File: jedapps.c             ***** */
Prototype const char    *jedshowdirectory(const char *directory, int line, int column,
				       int height, int width, int *pKey, int mode);
Prototype jedtree_t     *jeddirtreecreate(WINDOW *win, const char *directory,
					 int line, int column, int height, int width,
					 int mode);
Prototype int            jedtdirtreedelete(jedtree_t *edtr);
Prototype int            jedtextfileedit(const char *title, const char *name, int line,
				   int column, int height, int width, int mode);

/* ***** File: jedform.c             ***** */
Prototype int		 jedformdisplay(int line, int column, int height,
					int width, formfield_t *forms, int n,
					int mode);
Prototype jedform_t     *jedformcreate(WINDOW *win, int line, int column,
				       int height, int width,
				       formfield_t *forms, int n,
				       const _jedkeytrans_t *kt, int mode);
Prototype int            jedformdokey(jedform_t *ef, int key);
Prototype int            jedformdelete(jedform_t *ef);

/* ***** File: jedframe.c            ***** */
Prototype jedtext_t     *jedtextcreate(char *text, long maxlen, int line,
				       int column, int height, int width,
				       int mode);
Prototype int            jedtextdokey(WINDOW *win, jedtext_t *tf, int key,
				      int mode);

/* ***** File: jedhist.c             ***** */
Prototype char           JEDHistFile[];

/* ***** File: jedlgraph.c           ***** */
Prototype int            lgwaddch(WINDOW *win, chtype c);
Prototype chtype         JEDACSMap[];
Prototype chtype         JEDScrollChars[];

/* ***** File: jedmenu.c             ***** */
Prototype int            jedmenudokey(int line, int column, jedmenufld_t *menus,
				  int key, int mode);
Prototype jedmenufld_t  *jedmenufldadd(jedmenufld_t **pstart, const char **names,
				       int nnames, long type);
Prototype jedmenufld_t  *jedmenufldfind(jedmenufld_t *start, const char **names,
					int nnames);

/* ***** File: jedsubs.c             ***** */
Prototype int            jedinitialize(int mode);
Prototype int            jeddeinitialize(void);
Prototype int            jeddokeytrans(int key, const _jedkeytrans_t *table);
Prototype void           jedprogbarshow(WINDOW *win, int line, int column,
					int height, int width, double perc,
					int mode);
Prototype void           jedprogbarclear(WINDOW *win, int line, int column,
					 int height, int width);
Prototype int            jedpictoverlay(WINDOW *win, int voff, int hoff, const char **alines,
					int *attrs);
Prototype int            JEditCursesMode;
Prototype long           JEditColors[];
Prototype int            JEditHelpTextY;
Prototype int            JEditTFFillChars;

/* ***** File: jedtable.c            ***** */
Prototype int		 jedtabdisplay(char *title, int line, int column,
				       int height, int width, long multi, long nLines,
				       int (*getLine)(void *uptr, long nr,
						      char *dest, int m),
				       int (*handle)(void *uptr, int ch, long nr,
						     char *title, char *line,
						     long *pnLines),
				       void *uPtr, long *pStart, int mode);
Prototype jedtable_t    *jedtabcreate(WINDOW *win, int line, int column,
				       int height, int width,
				       int (*getLine)(void *, long, char *, int),
				       void *uptr, long multi, long nLines,
				       long *pInit, int mode);
Prototype int            jedtabdelete(jedtable_t *ta);
Prototype int            jedtabdokey(jedtable_t *ta, int key, int mode);

/* ***** File: jedtree.c             ***** */
Prototype jedtree_t     *jedtreecreate(WINDOW *win, const char *title,
				       int line, int column, int height,
				       int width, int flags);
Prototype void           jedtreedelete(jedtree_t *edtr);
Prototype int            jedtreesetup(jedtree_t *edtr);
Prototype int            jedtreesetname(jedtreenode_t *node, const char *name,
					int mode);
Prototype jedtreenode_t *jedtreeaddnode(jedtree_t *edtr, jedtreenode_t *parent,
					const char *name, int mode);
Prototype jedtreenode_t *jedtreeaddnodepath(jedtree_t *edtr,
					    jedtreenode_t *root,
					    const char *path, int mode);
Prototype jedtreenode_t *jedtreeaddnodepath2(jedtree_t *edtr,
					     jedtreenode_t *root,
					     const char *parts[], int mode);
Prototype void           jedtreedelnode(jedtreenode_t *node, int mode);
Prototype int            jedtreegetpath(jedtreenode_t *node, const char *sep,
					char *dest, int dlen);
Prototype int            jedtreedokey(jedtree_t *edtr, int key, int arg,
				      int mode);
Prototype int            jedtreesearch(jedtree_t *edtr,
				       int (*function)(void *, jedtreenode_t *, int, int),
				       void *uptr, int mode);

/* ***** File: jedwstack.c           ***** */
Prototype WINDOW        *jedwstacknew(int nlines, int ncols, int y, int x);
Prototype int            jedwstackdel(WINDOW *win);
Prototype int            jedwstackrefresh(WINDOW *win);
Prototype int            wnoutrefresh_s(WINDOW *win);
Prototype int            jedwstacktouch(WINDOW *win);
Prototype int            jedwstackdoupdate(int flag);
Prototype int            jedwstackgetch(WINDOW *win);


# undef Prototype   /* I like to be clean  */

# ifdef JEDLIB_INTERNAL
/*  For C++ compatibility   */
#  ifdef __cplusplus
#   define IPrototype extern "C"
#  else
#   define IPrototype extern
#  endif
/* ***** File: jedhist.c             ***** */
IPrototype int           _jedhistload(void);
IPrototype int           _jedhistdump(void);
IPrototype const char  **_jedhistgetvalues(const char *fldname,
					   const char *content,
					   int *pNEntries);
IPrototype int           _jedhistremember(const char *fldname,
					  const char *content,
					  int len);

/* ***** File: jedlgraph.c           ***** */
IPrototype void          _initmyacs(int mode);

/* ***** File: jedsubs.c             ***** */
IPrototype void          _jedprintframe(WINDOW *win, int offy, int offx,
				     int height, int width, int mode);
IPrototype int           _jedprinttitle(WINDOW *win, int left, int width,
				     const char *title);
IPrototype int           _jedflashwin(WINDOW *win, int mode);
IPrototype int           _jedsetscrollbar(long pos, long n, int height, int *pstart,
				       int *pend);
IPrototype int           _jedinskillring(const char *s, long len);
IPrototype const char   *_jedgetyank(void);
IPrototype void          _jedrotkillring(int d);
IPrototype const _jedkeytrans_t JEdLibStdKeyTab[];

/* ***** File: jedwstack.c           ***** */
IPrototype int           _jedwstackpush(WINDOW *win, int pos);
IPrototype int           _jedwstackpop(WINDOW *win);
IPrototype int           _jedwstackflag(WINDOW *win, int flag);


# undef IPrototype   /* I like to be clean  */
# endif

#endif /* JEDLIB_H  */
