/*JS*********************************************************************
*
*    Program : JSBITUTIL
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Define some simple macros to manipulate bits in arrays.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

#include <limits.h>  /* to know CHAR_BIT  */

#define BITS_PER_ENTRY(a)			\
  (sizeof(*(a)) * CHAR_BIT)

#define BITS_NRENTRIES(a, nBits)				\
  (((nBits) + (BITS_PER_ENTRY(a) - 1)) / BITS_PER_ENTRY(a))

#define BITS_NRBYTES(a, nBits)				\
  (BITS_NRENTRIES(a, nBits) * sizeof(*(a)))

#define BIT_SET(a, n)							\
  ((a)[(n) / (sizeof(*(a)) * CHAR_BIT)] |=				\
   (1UL << ((n) % (sizeof(*(a)) * CHAR_BIT))))

#define BIT_CLEAR(a, n)							\
  ((a)[(n) / (sizeof(*(a)) * CHAR_BIT)] &=				\
   ~(1UL << ((n) % (sizeof(*(a)) * CHAR_BIT))))

#define BIT_FLIP(a, n)				\
  ((a)[(n) / (sizeof(*(a)) * CHAR_BIT)] ^=	\
   (1UL << ((n) % (sizeof(*(a)) * CHAR_BIT))))

/*  The "!= 0" ensures that we have an integer  */
#define BIT_ISSET(a, n)							\
  (((a)[(n) / (sizeof(*(a)) * CHAR_BIT)] &				\
   (1UL << ((n) % (sizeof(*(a)) * CHAR_BIT)))) != 0)

#define BIT_SETRNG(a, n, l, v)					     \
  ((a)[(n) / (sizeof(*(a)) * CHAR_BIT)] =			     \
   ((a)[(n) / (sizeof(*(a)) * CHAR_BIT)] &			     \
    ~((((1UL << (l)) - 1)) <<					     \
      ((n) % (sizeof(*(a)) * CHAR_BIT)))) |			     \
   ((v) & ((1UL << (l)) - 1)) <<				     \
   ((n) % (sizeof(*(a)) * CHAR_BIT)))

#define BIT_GETRNG(a, n, l)					 \
  (((a)[(n) / (sizeof(*(a)) * CHAR_BIT)] >>			 \
    ((n) % (sizeof(*(a)) * CHAR_BIT))) &			 \
   ((1UL << (l)) - 1))
