/*JS*********************************************************************
*
*    Program : JSCONFIG
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Contains machine dependent settings for all programs.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/
#ifndef _jsconfig_h_
#define _jsconfig_h_

/* *************************** */
/* ***   General settings  *** */
/* *************************** */
/* **  Force system header files to disable POSIX extensions  ** */
/*#define _POSIX_SOURCE*/

/* **  Do we have 'strdup' (is neither POSIX nor ANSI conformant)?  ** */
/*#define CONFIG_NO_STRDUP*/

/* **  Do not use 'alloca' memory allocator  ** */
/*#define CONFIG_NO_ALLOCA*/

#if defined(__alpha) || defined(__x86_64__)
/* **  Set if your long ints have 64 bits  ** */
# define CONFIG_64BIT_LONG
/* **  'int' has a different size than 'long' and 'short'  ** */
# define CONFIG_SEP_INT_SIZE
#endif

#ifdef __GNUC__
/*  The integer type "long long" is available  */
# define CONFIG_HAS_LONGLONG
# define INLINE inline
#else
# define INLINE
#endif

/* ***  Z/OS (OS/390)  *** */
#ifdef __MVS__
#define _VARARG_EXT_       1 /*  make stdarg work on Z/OS  */
#define _XOPEN_SOURCE    500 /*  needed to make some prototypes visible  */
#define CONFIG_NO_STRDUP
#endif

/* **  System's delimiter for directories in a filename ** */
#if defined __MSDOS__ || defined _WIN32
#define CONFIG_DIR_DELIMITER  '\\'
#else
#define CONFIG_DIR_DELIMITER  '/'
#endif

/* **  Don't search directories starting with "."  ("." and ".." are
       never looked at)  ** */
#define CONFIG_IGNORE_POINTDIRS

/* **  Delimiter for PATH like variables  ** */
#define CONFIG_LIST_DELIMITER      ':'
#define CONFIG_LIST_DELIMITER_STR  ":"

/*  Actually, djgpp has most of POSIX functionality  */
#if defined(__MSDOS__)
/* **  Resort to the ANSI standard  ** */
# define CONFIG_NO_POSIX
# define CONFIG_NO_REGEX
# ifndef __DJGPP__ /*  alloca seems to work with djgpp  */
#  define CONFIG_NO_ALLOCA
# endif
#elif defined(_WIN32)
# define CONFIG_NO_REGEX
# define CONFIG_NO_ALLOCA
#endif

/* ***  File "readparam.c": Customizations  ****/
/* **  Allow arithmetric expressions like $[3+4*5]  ** */
/*# define CONFIG_READPARAM_EXPR*/

/*  Since CONFIG_READPARAM_SHELLCMD uses "popen", disable
 *   it without POSIX
 */
#if !defined(CONFIG_READPARAM_NO_SHELLCMD) && !defined(CONFIG_NO_POSIX)
 /* ** Allow shell command evaluation like $(..) ** */
# define CONFIG_READPARAM_SHELLCMD
#endif

/* ***  File "compexpr.c": Use "intpow" if exponent is an integer value  *** */
#define CONFIG_USE_INTPOW

#if defined(__MSDOS__) && !defined(__DJGPP__)
/* ***  File "bigsort.c": Don't use too much memory under this ancient OS *** */
# define BIGSORT_MAXMEM   50000

/* ***  File "fcache*.c": Restrict cache size under DOS  *** */
#define CONFIG_FCACHE_DEFNRPAGES      48
#define CONFIG_FCACHE_DEFPAGESIZE   1024
#endif

#ifdef __sgi
/* ***  File "bigsort.c": We need this for Silicon Graphics with -mips2  *** */
# define CONFIG_BIGSORT_ALIGN
#endif

#if 0
/*  Enable if you have the GNU readline package installed  */
# define CONFIG_HAS_READLINE
#endif

#ifdef __linux
# define CONFIG_HAS_PREADWRITE
# define CONFIG_HAS_ASYNCIO
#endif

#ifdef __sparc
/*  It's unfortunate, but this value on SunOS is terribly low (25)  */
# define CONFIG_FCACHE_NRLOCKSEMS  16
# define CONFIG_FCACHE_MAXSEMOPS   10
#endif

/* ***  Read additional configuration file if given  *** */
#ifdef JSCONFIG_LOCAL_FILE
# include JSCONFIG_LOCAL_FILE
#endif

/* ***  For JO-SQL data base library  *** */
#ifndef JOSQL_SYSDIR
/*  For include files: use JOSQL_SYSDIR/JNSQL_BASEDIR as
 *   well as HOME/JNSQL_BASEDIR
 */
# define JOSQL_SYSDIR          "/usr"
#endif

#ifndef CONFIG_JOSQLD_SOCKPATT
/*  Default pattern for UNIX domain sockets the server is
 *   listening to. Is used with printf and the port number
 */
# define CONFIG_JOSQLD_SOCKPATTERN  "/tmp/.josqlserv_%d"
#endif

#ifndef CONFIG_JOSQLD_PORT
/*  Default port the server is listening too  */
# define CONFIG_JOSQLD_PORT     19023 /*  0x4a4f "JO"  */
#endif

#endif /*  "#ifndef _jsconfig_h_"  */
