/*JS*********************************************************************
*
*    Program : JSLLIO
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Header file for low-level I/O, file and other functions
*              that are OS specific
*
*              Copyright (C) 2008-2010 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/
#ifndef _jsllio_h_
#define _jsllio_h_

#if defined(__unix__) || defined(__unix)
# if defined(__linux__) && !defined(_GNU_SOURCE)
#  define _GNU_SOURCE   /*  make O_DIRECT visible  */
# endif

# include <sys/types.h>
# include <unistd.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <utime.h>
# include <dirent.h>

# ifdef S_ISLNK
#  define CONFIG_HAS_SYMLINKS
# endif

#elif defined __BORLANDC__ || defined _MSC_VER

# include <sys/types.h>
# include <fcntl.h>
# include <sys/stat.h>
# include <io.h>
# include <utime.h>
# include <dir.h>
# include <dirent.h>
# include <process.h>

# define popen(n, m)      _popen(n, m)
# define pclose(f)        _pclose(f)
# define mkdir(path, mode) mkdir(path)

/* structure + function  */
# define lstat            _stat
# define ftruncate        chsize
# define utime            _utime

# define fsync(handle)    0

typedef signed long ssize_t;

# define getuid()         0
# define getgid()         0

# define S_IRGRP          S_IRUSR
# define S_IROTH          S_IRUSR
# define S_IWGRP          S_IWUSR
# define S_IWOTH          S_IWUSR
# define S_IXGRP          0
# define S_IXOTH          0

# define S_IRWXG          (S_IRGRP | S_IWGRP | S_IXGRP)
# define S_IRWXO          (S_IROTH | S_IWOTH | S_IXOTH)

#elif defined __LCC__ || defined __MINGW32__ 
# include <sys/types.h>
# include <fcntl.h>
# include <sys/stat.h>
# include <io.h>
# include <utime.h>
# include <dirent.h>
# include <process.h>

# define mkdir(path, mode) mkdir(path)

/* structure + function  */
# define lstat            _stat
# define ftruncate        chsize
# define utime            _utime

# define fsync(handle)    0

# define getuid()         0
# define getgid()         0

# define S_IRGRP          S_IRUSR
# define S_IROTH          S_IRUSR
# define S_IWGRP          S_IWUSR
# define S_IWOTH          S_IWUSR
# define S_IXGRP          0
# define S_IXOTH          0

# define S_IRWXG          (S_IRGRP | S_IWGRP | S_IXGRP)
# define S_IRWXO          (S_IROTH | S_IWOTH | S_IXOTH)

typedef long off_t;
#ifndef __MINGW32__
typedef signed long ssize_t;
#endif

#endif
#endif
