/*JS*********************************************************************
*
*    Program : JSSTDINT
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Header file to wrap "stdint.h" which is not always available
*
*              Copyright (C) 2009 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/
#ifndef _jsstdint_h_
#define _jsstdint_h_

#if defined __BORLANDC__ || defined _MSC_VER

# define jsint8_t         __int8
# define jsint16_t        __int16
# define jsint32_t        __int32
# define jsint64_t        __int64

# define jsuint8_t        unsigned __int8
# define jsuint16_t       unsigned __int16
# define jsuint32_t       unsigned __int32
# define jsuint64_t       unsigned __int64

#ifdef _MSC_VER
#define off_t long
#endif

#elif defined __sun || defined __hpux

# define jsint8_t         signed char
# define jsint16_t        short
# define jsint32_t        long
# define jsint64_t        long long

# define jsuint8_t        unsigned char
# define jsuint16_t       unsigned short
# define jsuint32_t       unsigned long
# define jsuint64_t       unsigned long long

#else
# include <stdint.h>

# define jsint8_t         int8_t
# define jsint16_t        int16_t
# define jsint32_t        int32_t
# define jsint64_t        int64_t

# define jsuint8_t        uint8_t
# define jsuint16_t       uint16_t
# define jsuint32_t       uint32_t
# define jsuint64_t       uint64_t
#endif

/*JS*********************************************************************
*   Store an integer efficiently, using 1-9 bytes. Returns the number of
*    bytes actually written.
*************************************************************************/

#define JSVARINT_MAXLEN       9

#define JSVARINTSTORE16(dest, value)			\
    ((value) < (1U << 7) ?				\
     *(dest)++ = ((value) & 0x7f) :			\
     							\
     (value) < (1U << 14) ?				\
     (*(dest)++ = (((value) >>  7) & 0x7f) | 0x80,	\
      *(dest)++ = ( (value)        & 0x7f)       ) : 	\
							\
     (*(dest)++ = (((value) >> 14) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >>  7) & 0x7f) | 0x80,	\
      *(dest)++ = ( (value)        & 0x7f)       )	\
     )

#define JSVARINTSTORE32(dest, value)			\
    ((value) < (1U << 21) ?				\
     JSVARINTSTORE16(dest, value) :			\
							\
     (value) < (1U << 28) ?				\
     (*(dest)++ = (((value) >> 21) & 0x7f) | 0x80,      \
      *(dest)++ = (((value) >> 14) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >>  7) & 0x7f) | 0x80,	\
      *(dest)++ = ( (value)        & 0x7f)       ) :	\
							\
     (*(dest)++ = (((value) >> 28) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 21) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 14) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >>  7) & 0x7f) | 0x80,	\
      *(dest)++ = ( (value)        & 0x7f)       )	\
     )

#define JSVARINTSTORE64(dest, value)			\
    ((value) < (1ULL << 35) ?				\
     JSVARINTSTORE32(dest, value) :			\
			  				\
     (value) < (1ULL << 42) ?				\
     (*(dest)++ = (((value) >> 35) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 28) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 21) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 14) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >>  7) & 0x7f) | 0x80,	\
      *(dest)++ = ( (value)        & 0x7f)       ) :	\
							\
     (value) < (1ULL << 49) ?				\
     (*(dest)++ = (((value) >> 42) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 35) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 28) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 21) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 14) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >>  7) & 0x7f) | 0x80,	\
      *(dest)++ = ( (value)        & 0x7f)       ) :	\
							\
     (value) < (1ULL << 56) ?				\
     (*(dest)++ = (((value) >> 49) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 42) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 35) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 28) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 21) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 14) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >>  7) & 0x7f) | 0x80,	\
      *(dest)++ = ( (value)        & 0x7f)       ) :	\
							\
     (*(dest)++ = (((value) >> 56) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 49) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 42) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 35) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 28) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 21) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >> 14) & 0x7f) | 0x80,	\
      *(dest)++ = (((value) >>  7) & 0x7f) | 0x80,	\
      *(dest)++ = ( (value)        & 0x7f)       )	\
     )

#define JSVARINTREAD16(src)			\
    ((src)[0] < (1U << 7) ?			\
    ((src) += 1,				\
     (jsuint32_t)(src)[-1]) :			\
    						\
    (src)[1] < (1U << 7) ?			\
    ((src) += 2,				\
     ((jsuint32_t)((src)[-2] & 0x7f) <<  7) |	\
     ((jsuint32_t)(src)[-1])) :			\
						\
    (src)[2] < (1U << 7) ?			\
    ((src) += 3,				\
     ((jsuint32_t)((src)[-3] & 0x7f) << 14) |	\
     ((jsuint32_t)((src)[-2] & 0x7f) <<  7) |	\
     (jsuint32_t)(src)[-1]) :			\
						\
     (jsuint64_t)-1)

#define JSVARINTREAD32(src)			\
    ((src)[0] < (1U << 7) ?			\
    ((src) += 1,				\
     (jsuint32_t)(src)[-1]) :			\
    						\
    (src)[1] < (1U << 7) ?			\
    ((src) += 2,				\
     ((jsuint32_t)((src)[-2] & 0x7f) <<  7) |	\
     ((jsuint32_t)(src)[-1])) :			\
						\
    (src)[2] < (1U << 7) ?			\
    ((src) += 3,				\
     ((jsuint32_t)((src)[-3] & 0x7f) << 14) |	\
     ((jsuint32_t)((src)[-2] & 0x7f) <<  7) |	\
     (jsuint32_t)(src)[-1]) :			\
    						\
     (src)[3] < (1U << 7) ?			\
     ((src) += 4,				\
      ((jsuint32_t)((src)[-4] & 0x7f) << 21) |	\
      ((jsuint32_t)((src)[-3] & 0x7f) << 14) |	\
      ((jsuint32_t)((src)[-2] & 0x7f) <<  7) |	\
      (jsuint32_t)(src)[-1]) :			\
     						\
     (src)[4] < (1U << 7) ?			\
     ((src) += 5,				\
      ((jsuint64_t)((src)[-5] & 0x7f) << 28) |	\
      ((jsuint64_t)((src)[-4] & 0x7f) << 21) |	\
      ((jsuint64_t)((src)[-3] & 0x7f) << 14) |	\
      ((jsuint64_t)((src)[-2] & 0x7f) <<  7) |	\
      (jsuint64_t)(src)[-1]) :			\
						\
    (jsuint64_t)-1)

#define JSVARINTREAD64(src)			\
    ((src)[0] < (1U << 7) ?			\
    ((src) += 1,				\
     (jsuint32_t)(src)[-1]) :			\
    						\
    (src)[1] < (1U << 7) ?			\
    ((src) += 2,				\
     ((jsuint32_t)((src)[-2] & 0x7f) <<  7) |	\
     ((jsuint32_t)(src)[-1])) :			\
						\
    (src)[2] < (1U << 7) ?			\
    ((src) += 3,				\
     ((jsuint32_t)((src)[-3] & 0x7f) << 14) |	\
     ((jsuint32_t)((src)[-2] & 0x7f) <<  7) |	\
     (jsuint32_t)(src)[-1]) :			\
    						\
     (src)[3] < (1U << 7) ?			\
     ((src) += 4,				\
      ((jsuint32_t)((src)[-4] & 0x7f) << 21) |	\
      ((jsuint32_t)((src)[-3] & 0x7f) << 14) |	\
      ((jsuint32_t)((src)[-2] & 0x7f) <<  7) |	\
      (jsuint32_t)(src)[-1]) :			\
     						\
     (src)[4] < (1U << 7) ?			\
     ((src) += 5,				\
      ((jsuint64_t)((src)[-5] & 0x7f) << 28) |	\
      ((jsuint64_t)((src)[-4] & 0x7f) << 21) |	\
      ((jsuint64_t)((src)[-3] & 0x7f) << 14) |	\
      ((jsuint64_t)((src)[-2] & 0x7f) <<  7) |	\
      (jsuint64_t)(src)[-1]) :			\
     						\
    (src)[5] < (1U << 7) ?			\
    ((src) += 6,				\
     ((jsuint64_t)((src)[-6] & 0x7f) << 35) |	\
     ((jsuint64_t)((src)[-5] & 0x7f) << 28) |	\
     ((jsuint64_t)((src)[-4] & 0x7f) << 21) |	\
     ((jsuint64_t)((src)[-3] & 0x7f) << 14) |	\
     ((jsuint64_t)((src)[-2] & 0x7f) <<  7) |	\
     (jsuint64_t)(src)[-1]) :			\
						\
    (src)[6] < (1U << 7) ?			\
    ((src) += 7,				\
     ((jsuint64_t)((src)[-7] & 0x7f) << 42) |	\
     ((jsuint64_t)((src)[-6] & 0x7f) << 35) |	\
     ((jsuint64_t)((src)[-5] & 0x7f) << 28) |	\
     ((jsuint64_t)((src)[-4] & 0x7f) << 21) |	\
     ((jsuint64_t)((src)[-3] & 0x7f) << 14) |	\
     ((jsuint64_t)((src)[-2] & 0x7f) <<  7) |	\
     (jsuint64_t)(src)[-1]) :			\
						\
    (src)[7] < (1U << 7) ?			\
    ((src) += 8,				\
     ((jsuint64_t)((src)[-8] & 0x7f) << 49) |	\
     ((jsuint64_t)((src)[-7] & 0x7f) << 42) |	\
     ((jsuint64_t)((src)[-6] & 0x7f) << 35) |	\
     ((jsuint64_t)((src)[-5] & 0x7f) << 28) |	\
     ((jsuint64_t)((src)[-4] & 0x7f) << 21) |	\
     ((jsuint64_t)((src)[-3] & 0x7f) << 14) |	\
     ((jsuint64_t)((src)[-2] & 0x7f) <<  7) |	\
     (jsuint64_t)(src)[-1]) :			\
						\
    (src)[8] < (1U << 7) ?			\
    ((src) += 9,				\
     ((jsuint64_t)((src)[-9] & 0x7f) << 56) |	\
     ((jsuint64_t)((src)[-8] & 0x7f) << 49) |	\
     ((jsuint64_t)((src)[-7] & 0x7f) << 42) |	\
     ((jsuint64_t)((src)[-6] & 0x7f) << 35) |	\
     ((jsuint64_t)((src)[-5] & 0x7f) << 28) |	\
     ((jsuint64_t)((src)[-4] & 0x7f) << 21) |	\
     ((jsuint64_t)((src)[-3] & 0x7f) << 14) |	\
     ((jsuint64_t)((src)[-2] & 0x7f) <<  7) |	\
     (jsuint64_t)(src)[-1]) :			\
						\
     (jsuint64_t)-1)

#endif
