/*JS*********************************************************************
*
*    Program : JEDAPPS
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application routines
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <jsconfig.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef CONFIG_NO_POSIX
# include <utime.h>
# include <sys/stat.h>
# include <dirent.h>
# include <sys/time.h>
# include <unistd.h>
#endif

#include <jssubs.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS  */
#include <stdio.h> /*  needed for "FILE" definition  */
#define FOFI_MPOINTFILES   (1<<4)
/*  END-DEFINITIONS  */

/*  Auxiliary structure for edit tree  */
typedef struct {
  int Mode;
  int OVisi;
  double SleepTime;
  char Directory[JS_FILENAME_MAX + 1];

  DIR *DirHandle;
  int Pos, Off;
} diredit_t;

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype const char    *jedshowdirectory(const char *directory, int line, int column,
				       int height, int width, int *pKey, int mode);

Prototype jedtree_t     *jeddirtreecreate(WINDOW *win, const char *directory,
					 int line, int column, int height, int width,
					 int mode);
Prototype int            jedtdirtreedelete(jedtree_t *edtr);
static int               setupdir(void *user, jedtree_t *edtr,
				  jedtreenode_t *node);

Prototype int            jedtextfileedit(const char *title, const char *name, int line,
				   int column, int height, int width, int mode);

/*JS*********************************************************************
*
*************************************************************************/

const char *jedshowdirectory(const char *directory, int line, int column,
			     int height, int width, int *pKey, int mode)

/************************************************************************/
{
#ifndef CONFIG_NO_POSIX
static char NameBuffer[JS_FILENAME_MAX + 1];
  jedtree_t *edtr;
  int key;

  if((edtr = jeddirtreecreate(NULL, directory, line, column, height, width,
			      mode)) == NULL)
    return NULL;

  jedtreedokey(edtr, 0, 0, 0); /*  redraw  */
  for(;;) {
    key = wgetch(edtr->JEDT_Window);
    if(key == '\n' || key == '\r' || key == KEY_ENTER) {
      int l;

      if(edtr->JEDT_Current == NULL) {
	beep();
	continue;
      }

      strcpy(NameBuffer, directory);
      l = strlen(NameBuffer);
      if(jedtreegetpath(edtr->JEDT_Current, "/", NameBuffer + l,
			sizeof(NameBuffer) - 1 - l) == -1)
	EGOTO(error);

#if 1
      jedtextfileedit(NameBuffer, NameBuffer, line, column, height - 4, width - 4,
		mode);
#else
      break;
#endif

    } else if(key == 'q' || key == '\033') {
      NameBuffer[0] = '\0';
      break;
    }

    if(jedtreedokey(edtr, key, 1, 1) == 0)
      beep();
  }

  *pKey = key;

  jedtdirtreedelete(edtr);

  return NameBuffer;
error:
#endif
  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

jedtree_t *jeddirtreecreate(WINDOW *win, const char *directory,
			   int line, int column, int height, int width,
			   int mode)

/************************************************************************/
{
  jedtree_t *edtr;
  diredit_t *pedit;

  if((pedit = (diredit_t *)malloc(sizeof(*pedit))) == NULL)
    return NULL;

  pedit->Mode = mode & ~256;
  pedit->SleepTime = (mode & 1) ? 0.02 : 0.0;
  strcpy(pedit->Directory, directory);

  pedit->OVisi = curs_set(0);
  pedit->DirHandle = NULL;
  pedit->Pos = pedit->Off = 0;

  if((edtr = jedtreecreate(win, directory, line, column, height, width,
			   JEDT_TREEWIN_FFRAME |
			   JEDT_TREEWIN_FSCROLLBAR)) == NULL) {
    free(pedit);
    return NULL;
  }

  /*  This will trigger the 'setupdir' function when the
   *   node is first printed
   */
  edtr->JEDT_Flags |= JEDT_TREE_FCALLBACK;

  edtr->JEDT_Setup = setupdir;
  edtr->JEDT_User = pedit;

  return edtr;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedtdirtreedelete(jedtree_t *edtr)

/************************************************************************/
{
  diredit_t *pedit;

  pedit = (diredit_t *)edtr->JEDT_User;

  if(pedit->OVisi != ERR) curs_set(pedit->OVisi);

#ifndef CONFIG_NO_POSIX
  if(pedit->DirHandle != NULL &&
     closedir(pedit->DirHandle) == -1)
    return -1;
#endif

  free(pedit);
  jedtreedelete(edtr);

  return 0;
}

#ifndef CONFIG_NO_POSIX
/*JS*********************************************************************
*
*************************************************************************/

static int setupdir(void *user, jedtree_t *edtr, jedtreenode_t *node)

/************************************************************************/
{
  struct dirent *dent;

#define DEDIT    ((diredit_t *)user)

  if((DEDIT->Mode & 256) || !(node->JEDTN_Flags & JEDT_TREE_FOPENED)) {
    if(DEDIT->SleepTime > 0) {
      jedtreenode_t *sub;

      sub = jedtgetsubfirstnode(node);
      if(!jedtnodeistail(sub)) {
	/*  Delete tail node  */
	jedtreedelnode(sub, 1);
	dsleep(DEDIT->SleepTime);

	/*  User should see how sub entries vanish  */
	node->JEDTN_Flags |= JEDT_TREE_FOPENED;

	/*  Prepare for another wrap-around  */
	DEDIT->Mode |= 256;
	return 1;
      }

      if(node->JEDTN_Flags & JEDT_TREE_FOPENED) {
	/*  OK, last round, so close now and one more time  */
	node->JEDTN_Flags &= ~JEDT_TREE_FOPENED;
	return 1;
      }
    }

    /*  Ok, deleting subnodes done, we are through  */
    DEDIT->Mode &= ~256;
    return 0;
  }

  if(DEDIT->DirHandle == NULL) {
    /*  First open the directory  */
    DEDIT->Pos = strlen(DEDIT->Directory);

    DEDIT->Directory[DEDIT->Pos] = '/';

    if((DEDIT->Off = jedtreegetpath(node, "/", DEDIT->Directory + DEDIT->Pos + 1,
				    sizeof(DEDIT->Directory) - 1 - DEDIT->Pos)) == -1)
      EGOTO(error);

    DEDIT->Off += DEDIT->Pos + 1;

    /*dbgprintf("OPEN '%s'\n", DEDIT->Directory);*/
    if((DEDIT->DirHandle = opendir(DEDIT->Directory)) == NULL)
      EGOTO(error);

    DEDIT->Directory[DEDIT->Off++] = '/';
    assert(DEDIT->Off < sizeof(DEDIT->Directory));
  }

  while((dent = readdir(DEDIT->DirHandle)) != NULL) {
    jedtreenode_t *entry;
    const char *parts[2];
    struct stat stbuf;

    if(dent->d_name[0] == '.' &&
       (!(DEDIT->Mode & 1) ||
	dent->d_name[1] == '\0' ||
	(dent->d_name[1] == '.' && dent->d_name[2] == '\0')))
      continue;

    parts[0] = dent->d_name;
    parts[1] = NULL;

    if((entry = jedtreeaddnodepath2(edtr, node, parts, JEDT_TREEADD_FDUPSTRING |
				    JEDT_TREEADD_FKEEPDIRTY)) == NULL)
      EGOTO(error);

    strcpy(DEDIT->Directory + DEDIT->Off, dent->d_name);
    if(stat(DEDIT->Directory, &stbuf) == -1)
      EGOTO(error);

    if(S_ISDIR(stbuf.st_mode)) {
      entry->JEDTN_Flags |= JEDT_TREE_FCALLBACK | JEDT_TREE_FSUBNODES;
    }

    /*dbgprintf("READ '%s'\n", dent->d_name);*/

    if(DEDIT->SleepTime > 0) {
      /*  Ok, only one more entry, than prepare for another round  */
      dsleep(DEDIT->SleepTime);

      return 1;
    }
  }

  if(closedir(DEDIT->DirHandle) == -1)
    EGOTO(error);

  /*  Clean up  */
  DEDIT->DirHandle = NULL;

  if(DEDIT->SleepTime == 0)
    /*  No need to call us over and over again  */
    node->JEDTN_Flags &= ~JEDT_TREE_FCALLBACK;

  DEDIT->Directory[DEDIT->Pos] = '\0';

  return 0;
error:
  DEDIT->Directory[DEDIT->Pos] = '\0';
  DEDIT->DirHandle = NULL;
  return -1;
}
#undef DEDIT
#endif

/*JS*********************************************************************
*
*************************************************************************/

int jedtextfileedit(const char *title, const char *name, int line,
	      int column, int height, int width, int mode)

/************************************************************************/
{
  char *text;
  long len;
  WINDOW *win;
  jedtext_t *tf;
  int save, key;

  save = JEditTFFillChars;

  if((text = file2buffer(name, &len)) == NULL) goto error;

  _checkcoordinates(line, column, height, width);

  if((win = newwin(height, width, line, column)) == NULL) goto error;

  keypad(win, TRUE);
  intrflush(win, FALSE);
  werase(win);

  if((tf = jedtextcreate(text, len, 1, 1, height - 2, width - 3, TFRAME_SCROLLBAR|
			TFRAME_ALLOCATE|TFRAME_INSMODE)) == NULL)
    goto error;

  _jedprintframe(win, 0, 0, height, width, 9);
  if((len = (width - strlen(title)) / 2) < 0) len = 0;
  mvwaddstr(win, 0, len, title);

#ifdef CONFIG_HAS_COLORS
  wbkgd(win, JEditColors[_EDLIB_FCOLIND + 0] | ' ');
#endif

  JEditTFFillChars = ' ';
  for(key = EDKEY_REDRAW ; ; ) {
    int ret;

    if((ret = jedtextdokey(win, tf, jeddokeytrans(key, NULL), 0)) == -1)
      goto error;

    if(ret == -2 || ret == -3) {
      beep();
      wrefresh(win);
    }

    if((key = wgetch(win)) == EDKEY_QUIT || key == KEY_F(1))
      break;
  }

  if(key == KEY_F(1)) {
    FILE *fp;

    if(mkBackupFile(name, 1) == NULL || (fp = fopen(name, "w")) == NULL ||
       fwrite(tf->TF_Text, 1, tf->TF_Length, fp) != tf->TF_Length ||
       fclose(fp) == EOF)
      goto error;
  }
  free(tf->TF_Text);

  if(delwin(win) == ERR) goto error;

  JEditTFFillChars = save;
  return(0);
error:
  JEditTFFillChars = save;
  return(-1);
}
