/*JS*********************************************************************
*
*    Program : JEDFORM
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Edit forms.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <jsconfig.h>
#include <jssubs.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

/*********     DEFINES                                          *********/
#ifndef FOFI_TMASK
/*  START-DEFINITIONS 2 */
typedef struct {
  unsigned long FF_Type;

  /*  Position on screen where to display  */
  int           FF_Line, FF_Column;

  /*  To restrict the field seen on the screen  */
  int           FF_ScreenHeight, FF_ScreenWidth;

  /*  The field contents  */
  union {
    char       *ff_Text;
    const char **ff_List;
  } c1;
#define FF_Text     c1.ff_Text
#define FF_List     c1.ff_List

  union {
    int         ff_Len;
    int         ff_Curr;
  } c2;
#define FF_Len   c2.ff_Len
#define FF_Curr  c2.ff_Curr

  union {
    int         ff_MaxLen;
    int         ff_NList;
  } c3;
#define FF_MaxLen   c3.ff_MaxLen
#define FF_NList    c3.ff_NList

  const char   *FF_Name;
  const char   *FF_Help;

  void         *FF_User;
} formfield_t;

/*  For the type field  */
#define FOFI_TMASK       0x000f0000UL
#define FOFI_LMASK       0x0000ffffUL

#define FOFI_TASCII      0x00010000UL /*  any character  */
#define FOFI_TINUM       0x00020000UL /*  solely digits  */
#define FOFI_TFNUM       0x00030000UL /*  floating point number  */
#define FOFI_TENUM       0x00040000UL
#define FOFI_TTEXT       0x00050000UL /*  text in multiple lines  */
#define FOFI_TBUTTON     0x00060000UL /*  lower bits give return value  */

#define FOFI_FNONSEL     0x00100000UL /*  field can not be selected  */
#define FOFI_FREFRESH    0x00200000UL /*  refresh always  */
#define FOFI_FEMPHASIZE  0x00400000UL /*  highlighted   */
#define FOFI_FSHOWSIGNS  0x00800000UL /*  Show '<' and '>' on long fields  */
#define FOFI_FALLOCATE   0x01000000UL /*  Reallocate field if necessary.  */
#define FOFI_FHIDDEN     0x02000000UL /*  Display field hidden  */
#define FOFI_FCHANGED    0x08000000UL /*  Set on exit if field changed */

typedef struct {
  int EF_LineOff, EF_ColOff;
  int EF_TopL, EF_BottL, EF_Width;
  int EF_ScrollBarS, EF_ScrollBarE;

  formfield_t *EF_Forms;
  int EF_FormHeight, EF_NForms;

  const _jedkeytrans_t *EF_KeyTrans;
  int EF_Mode, EF_CurrForm;
  WINDOW *EF_Window, *EF_InWin;

#ifdef NCURSES_MOUSE_VERSION
  int EF_PrevX, EF_PrevY;
#endif

  void *EF_FieldAux[1];
} jedform_t;
/*  END-DEFINITIONS  */
#endif

#ifdef CONFIG_NO_STRDUP
# define strdup  mystrdup  /*  strdup is not POSIX conformant  */
#endif

#ifdef CONFIG_USE_SHADOWS
# define CONFIG_FORM_SHADOWS
#endif

#ifdef CONFIG_HAS_COLORS
# define FORM_BKGRND        JEditColors[_EDLIB_FCOLIND + 0]
# define FORM_COLOR         JEditColors[_EDLIB_FCOLIND + 1]
# define FORM_ENUMCOLOR     JEditColors[_EDLIB_FCOLIND + 2]
# define FORM_HELPCOLOR     JEditColors[_EDLIB_FCOLIND + 3]
#endif

/*  Pop up the enumeration list automatically when entering the field  */
/*#define CONFIG_POPUPENUM*/

/*  Internal modes for routine editField  */
#define EFIELD_KEEPWINDOW   (1<<0)
#define EFIELD_FRAME        (1<<1)

#define EFIELD_PRINTONLY    (1<<8)
#define EFIELD_INSERT       (1<<9)
#define EFIELD_SCROLLBAR    (1<<10)

#define FORM_SCROLL_SKIP      1

typedef struct {
  int Start, Current;

  WINDOW *Window;
  int Height;

  const char **Values;
  int NValues;

  int WinOff, WinCurr;
} text1aux_t;

typedef struct {
  jedtable_t *EA_Table;
} enumaux_t;

#define edformsetscbar(ef) _jedsetscrollbar((ef)->EF_TopL - (ef)->EF_LineOff, \
                                         (ef)->EF_FormHeight, \
                                         (ef)->EF_BottL - (ef)->EF_TopL + 1, \
		                         &((ef)->EF_ScrollBarS), \
                                         &((ef)->EF_ScrollBarE))

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
Prototype int		 jedformdisplay(int line, int column, int height,
					int width, formfield_t *forms, int n,
					int mode);

Prototype jedform_t     *jedformcreate(WINDOW *win, int line, int column,
				       int height, int width,
				       formfield_t *forms, int n,
				       const _jedkeytrans_t *kt, int mode);
static int               mvwaddstr2(WINDOW *win, int line, int column,
				    const char *text);

Prototype int            jedformdokey(jedform_t *ef, int key);
Prototype int            jedformdelete(jedform_t *ef);

static int               doeditfield(jedform_t *ef, int key, int mode);
static int               doredrawform(jedform_t *ef);
static int               doredrawscbar(jedform_t *ef);

static int               dochgfield(jedform_t *ef, int newcurr, int d);
static int               doformalign(jedform_t *ef);
static int               doformscroll(jedform_t *ef, int l);
static void              dochkfield(jedform_t *ef, int d);

static int               getfldattrs(formfield_t *ff, int flag);

static text1aux_t       *mkauxtext1(formfield_t *ff);
static int               dotext1(jedform_t *ef, int line, int column,
				 formfield_t *ff, text1aux_t *paux, int key,
				 int mode);
static int               doinsert1(jedform_t *ef, formfield_t *ff, long off,
				   const char *s, int len, int mode);
static void              freeauxtext1(jedform_t *ef, formfield_t *ff,
				      text1aux_t *ea);

static int               dobutton(WINDOW *win, int line, int column,
				  formfield_t *ff, int key, int mode);

static enumaux_t        *mkauxenum(formfield_t *ff);
static int               doenum(jedform_t *ef, int line, int column,
				formfield_t *ff, enumaux_t *ea, int key,
				int mode);
static int               domkenumwin(jedform_t *ef, int line, int column,
				     formfield_t *ff, enumaux_t *ea);
static int               dofreeenumwin(jedform_t *ef, formfield_t *ff,
				       enumaux_t *ea);
static int               dogetenum(const char *const *list, long nr,
				   char *dest, int m);
static void              freeauxenum(jedform_t *ef, formfield_t *ff,
				     enumaux_t *ea);

/* ERROR-DEFINITIONS from editForm label _ERR_EDITFORM
   Screen too small for window
*/

/*JS*********************************************************************
*
*************************************************************************/

int jedformdisplay(int line, int column, int height, int width,
		   formfield_t *forms, int n, int mode)

/************************************************************************/
{
  jedform_t *ef;
  int ret, key;

#if 0
  height /= 2;
#endif

  if((ef = jedformcreate(NULL, line, column, height, width, forms, n,
			 JEdLibStdKeyTab, mode)) == NULL)
    EGOTO(error);

  /*  Do not refresh the window before _flashWindow, since it needs to
   *   see the underlying screen for flashing. If flashing is disabled,
   *   _flashWindow will refresh anyway. We need to refresh manually
   *   at least once, since the first field might be a popped up
   *   enumeration field. In that case the form window would never have
   *   been refreshed once.
   */
  _jedflashwin(ef->EF_Window, 3);

  if(ef->EF_CurrForm == -1) {
    doredrawform(ef);
    curs_set(0);
    ret = jedwstackgetch(ef->EF_InWin);
    curs_set(1);
  } else {
    for(key = EDKEY_REDRAW ;; key = jedwstackgetch(ef->EF_InWin)) {
      if((ret = jedformdokey(ef, key))) {
	if(ret == -2 || ret == -3) {
	  beep();
	  jedwstackrefresh(ef->EF_InWin);
	} else if(ret > 0) {
	  ret--;
	  break;
	}
      }
    }
  }
  _jedflashwin(ef->EF_Window, 0);

  if(jedformdelete(ef) == -1) EGOTO(error);

  return(ret);
error:
  return(-1);
}

/*JS*********************************************************************
*
*************************************************************************/

jedform_t *jedformcreate(WINDOW *win, int line, int column,
			 int height, int width,
			 formfield_t *forms, int n,
			 const _jedkeytrans_t *kt, int mode)

/************************************************************************/
{
  jedform_t *ef;
  int i, j, k, wheight, wwidth;

  if((ef = (jedform_t *)malloc(sizeof(*ef) + (n - 1) *
			      sizeof(*(ef->EF_FieldAux)))) == NULL)
    return NULL;

  ef->EF_Mode = EFIELD_INSERT;
  ef->EF_Forms  = forms;
  ef->EF_NForms = n;
  ef->EF_KeyTrans = kt;

  wheight = height;
  wwidth  = width;

  /*  Check for selectable fields and help texts  */
  j = k = 0;
  for(i = 0 ; i < n ; i++) {
    int bline;

    forms[i].FF_Type &= ~FOFI_FCHANGED;
    if(!(forms[i].FF_Type & FOFI_FNONSEL)) k |= 1;
    if(forms[i].FF_Help) k |= 2;

    if((bline = forms[i].FF_Line + forms[i].FF_ScreenHeight) > j) j = bline;

    forms[i].FF_Len = strlen(forms[i].FF_Text);
    if(forms[i].FF_ScreenWidth == -1 &&
       (forms[i].FF_ScreenWidth = forms[i].FF_MaxLen) == -1)
      forms[i].FF_ScreenWidth = forms[i].FF_Len;
  }

  ef->EF_FormHeight = j;
  if(k & 2) wheight++; /*  need additional line for help text  */

  if(mode & FOFI_MFRAME) {
    ef->EF_Mode |= EFIELD_FRAME;

#ifdef CONFIG_FORM_SHADOWS
    wheight += 3; wwidth += 4;
#else
    wheight += 2; wwidth += 2;
#endif
    ef->EF_LineOff = 1;
  } else
    ef->EF_LineOff = 0;

  /*  Center form and make sure that values are non-negative  */
  if(line < 0 && (line = (LINES - wheight) / 2) < 0) line = 0;
  if(column < 0 && (column = (COLS - wwidth) / 2) < 0) column = 0;

  if(win) {
    ef->EF_Mode |= EFIELD_KEEPWINDOW;
  } else {
    if(wheight > LINES) {
      i = wheight - LINES;
      wheight = LINES;
      if((height -= i) < 1) {
	JSSETERROR(_ERR_EDITFORM, 0);
	EGOTO(error);
      }
    }
    if(wwidth > COLS) wwidth = COLS;

    /*  Create a new window for the form and set it up  */
    if((win = jedwstacknew(wheight, wwidth, line, column)) == NULL) EGOTO(error);

#ifdef CONFIG_HAS_COLORS
    if(!(JEditCursesMode & CURSMODE_NOCOLORS) && wbkgd(win, FORM_BKGRND|' ') == ERR)
      EGOTO(error);
#endif

    keypad(win, TRUE); intrflush(win, FALSE); werase(win);
  }
  ef->EF_InWin = ef->EF_Window = win;

  ef->EF_TopL  = ef->EF_ColOff = ef->EF_LineOff;
  ef->EF_BottL = ef->EF_TopL + height - 1;

  ef->EF_Width = width;

#ifdef NCURSES_MOUSE_VERSION
  ef->EF_PrevX = ef->EF_PrevY = -1;
#endif

  if(ef->EF_FormHeight > height) {
    ef->EF_Mode |= EFIELD_SCROLLBAR;
    edformsetscbar(ef);
  }

  /*  Get work space for fields  */
  for(i = 0 ; i < n ; i++)
    switch(forms[i].FF_Type & FOFI_TMASK) {
    case FOFI_TASCII: case FOFI_TINUM: case FOFI_TFNUM:
      /*  Set up single line text and print them  */
      if((ef->EF_FieldAux[i] = mkauxtext1(&forms[i])) == NULL)
	EGOTO(error);
      break;
    case FOFI_TENUM:
      if((ef->EF_FieldAux[i] = mkauxenum(&forms[i])) == NULL)
	EGOTO(error);
      break;
    case FOFI_TTEXT:
      {
	jedtext_t *tf;
	int mmode;

	/*  Set up multiline fields properly  */
	mmode = TFRAME_INSMODE | TFRAME_NONEWLINE;
	if(forms[i].FF_Type & FOFI_FALLOCATE) mmode |= TFRAME_ALLOCATE;
	if(forms[i].FF_Type & FOFI_FSHOWSIGNS) mmode |= TFRAME_SCROLLBAR;
	if(forms[i].FF_Type & FOFI_FNONSEL) mmode |= TFRAME_FILLBLANK;

	if((tf = jedtextcreate(forms[i].FF_Text, forms[i].FF_MaxLen,
			      ef->EF_LineOff + forms[i].FF_Line,
			      ef->EF_ColOff  + forms[i].FF_Column,
			      forms[i].FF_ScreenHeight, forms[i].FF_ScreenWidth,
			      mmode)) == NULL)
	  EGOTO(error);
	ef->EF_FieldAux[i] = tf;

	/*  When scrolling, tell the frame which region to occupy  */
	tf->TF_TopL  = ef->EF_TopL;
	tf->TF_BottL = ef->EF_BottL;
      }
      break;
    case FOFI_TBUTTON: default:
      ef->EF_FieldAux[i] = NULL;
      break;
    }

  /*  Print all fields  */
  if(doredrawform(ef) == -1) EGOTO(error);

  if(k & 1) {
    /*  Set up current field  */
    ef->EF_CurrForm = 0;
    dochkfield(ef, 1);
  } else
    /*  nothing selectable  */
    ef->EF_CurrForm = -1;

  return(ef);
error:
  return(NULL);
}

/*JS*********************************************************************
*
*************************************************************************/

static int mvwaddstr2(WINDOW *win, int line, int column, const char *text)

/************************************************************************/
{
  wmove(win, line, column);
  for( ; *text ; text++) {
    int c;

    c = *(unsigned char *)text;
    if(c == '&' && text[1]) {
      c = *(unsigned char *)++text;

      if(JEditCursesMode & CURSMODE_NOACSGRAPH) {
	switch(c) {
	case 'g': case 'l': case 'n': case 'j':	case 'r': case 'm':
	case 'o': case 'G': case 'L': case 'N': case 'J': case 'R':
	case 'M': case 'O': case 'T': case 't':
	case '#': case '+':
	  c = '+';
	  break;

	case '=': case '-':
	  c = '-';
	  break;

	case '!': case '|':
	  c = '|';
	  break;
	}

      } else {
	switch(c) {
	case 'g': c = ACS_ULCORNER; break;
	case 'l': c = ACS_LLCORNER; break;
	case 'n': c = ACS_URCORNER; break;
	case 'j': c = ACS_LRCORNER; break;
	case 'r': c = ACS_LTEE; break;
	case 'm': c = ACS_RTEE; break;
	case 'o': c = ACS_BTEE; break;
	case 't': c = ACS_TTEE; break;
	case '-': c = ACS_HLINE; break;
	case '|': c = ACS_VLINE; break;
	case '+': c = ACS_PLUS; break;
	case 'G': c = ACS_ULCORNER22; break;
	case 'L': c = ACS_LLCORNER22; break;
	case 'N': c = ACS_URCORNER22; break;
	case 'J': c = ACS_LRCORNER22; break;
	case 'R': c = ACS_LTEE22; break;
	case 'M': c = ACS_RTEE22; break;
	case 'O': c = ACS_BTEE22; break;
	case 'T': c = ACS_TTEE22; break;
	case '=': c = ACS_HLINE22; break;
	case '!': c = ACS_VLINE22; break;
	case '#': c = ACS_PLUS22; break;
	}
      }
    }

    waddch(win, c);
  }

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedformdokey(jedform_t *ef, int key)

/************************************************************************/
{
  int ret;

  if(ef->EF_KeyTrans && key > 0)
    key = jeddokeytrans(key, ef->EF_KeyTrans);

  switch(key) {
  case EDKEY_PREV:
    ret = dochgfield(ef, ef->EF_CurrForm - 1, -1);
    break;
  case EDKEY_NEXT:
    ret = dochgfield(ef, ef->EF_CurrForm + 1, 1);
    break;
  case EDKEY_PUP:
    if((ret = doformscroll(ef, 1)) == -1) return -1;
    if(ret) {
      if(doformalign(ef) == -1) return -1;
      ret = doeditfield(ef, EDKEY_REDRAW, 0);
    } else
      ret = -3;
    break;
  case EDKEY_PDOWN:
    if((ret = doformscroll(ef, -1)) == -1) return -1;
    if(ret) {
      if(doformalign(ef) == -1) return -1;
      ret = doeditfield(ef, EDKEY_REDRAW, 0);
    } else
      ret = -3;
    break;
#ifdef NCURSES_MOUSE_VERSION
  case KEY_MOUSE:
    /*  Change current field and then handle mouse event in that  */
    {
      int wl, wc, ml, mc, i;
      MEVENT mev;

      /*  Get position relative to window  */
      if(getmouse(&mev) == ERR) EGOTO(error);
      getbegyx(ef->EF_Window, wl, wc);
      ml = mev.y - wl;
      mc = mev.x - wc;

      if((mev.bstate & BUTTON1_PRESSED) &&
	 ml == 0 && 1 <= mc && mc <= ef->EF_Width) {
	ef->EF_PrevX = mev.x;
	ef->EF_PrevY = mev.y;
	ret = 0;
	break;
      }

      if((mev.bstate & BUTTON1_RELEASED) && ef->EF_PrevX != -1 &&
	 (ef->EF_PrevX != mev.x || ef->EF_PrevY != mev.y)) {
	/*  Move window  */
	wl += mev.y - ef->EF_PrevY;
	wc += mev.x - ef->EF_PrevX;

	getmaxyx(ef->EF_Window, ml, mc);
	if(wl < 0) wl = 0;
	else if((wl + ml) > LINES) wl = LINES - ml;
	if(wc < 0) wc = 0;
	else if((wc + mc) > COLS) wc = COLS - mc;

	if(mvwin(ef->EF_Window, wl, wc) == ERR) ret = -3;
	jedwstacktouch(stdscr);
	jedwstackdoupdate(1);
	ret = 0;
	break;
      }

      ef->EF_PrevX = ef->EF_PrevY = -1;
      ml -= ef->EF_LineOff;
      mc -= ef->EF_ColOff;

      for(i = 0 ; i < ef->EF_NForms ; i++) {
	formfield_t *ff;
	int w, h;

	ff = &(ef->EF_Forms[i]);
	if(ff->FF_Type & FOFI_FNONSEL) continue;
	h = ff->FF_ScreenHeight;
	w = ff->FF_ScreenWidth;

	if(ff->FF_Line <= ml && ml < (ff->FF_Line + h) &&
	   ff->FF_Column <= mc && mc < ff->FF_Column + w)
	  break;
      }

      if(i == ef->EF_NForms ||
 	 /*  Prevent changing to non-selectable field  */
	 (ef->EF_Forms[i].FF_Type & FOFI_FNONSEL)) {
	ret = -3;
	break;
      }

      if(i != ef->EF_CurrForm && dochgfield(ef, i, 1) == -1) EGOTO(error);
      ret = 0;
    }
#endif
  default:
    ret = doeditfield(ef, key, 0);

    if(ret == -2) {
      if(key < 0) {
	key = -key;
	if(ef->EF_KeyTrans)
	  key = jeddokeytrans(key, ef->EF_KeyTrans);
      }

      switch(key) {
      case EDKEY_UP:
	ret = dochgfield(ef, ef->EF_CurrForm - 1, -1);
	break;
      case EDKEY_DOWN:
	ret = dochgfield(ef, ef->EF_CurrForm + 1, 1);
	break;
      case EDKEY_START:
	ret = dochgfield(ef, 0, 1);
	break;
      case EDKEY_END:
	ret = dochgfield(ef, ef->EF_NForms - 1, -1);
	break;
      case EDKEY_ENTER:
	ret = 1;
	break;
      case EDKEY_QUIT:
	ret = 2;
	break;
      }
    }
    break;
  }

  return ret;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedformdelete(jedform_t *ef)

/************************************************************************/
{
  WINDOW *win;
  int i;

  win = ef->EF_Window;

  for(i = 0 ; i < ef->EF_NForms ; i++)
    switch(ef->EF_Forms[i].FF_Type & FOFI_TMASK) {
    case FOFI_TASCII: case FOFI_TINUM: case FOFI_TFNUM:
      freeauxtext1(ef, &(ef->EF_Forms[i]), (text1aux_t *)(ef->EF_FieldAux[i]));
      break;
    case FOFI_TENUM:
      freeauxenum(ef, &(ef->EF_Forms[i]), (enumaux_t *)(ef->EF_FieldAux[i]));
      break;
    case FOFI_TTEXT:
      if(((jedtext_t *)ef->EF_FieldAux[i])->TF_Mode & TFRAME_CHANGED) {
	ef->EF_Forms[i].FF_Type |= FOFI_FCHANGED;

	/*  Because of reallocation, this pointer might have changed  */
	ef->EF_Forms[i].FF_Text = ((jedtext_t *)ef->EF_FieldAux[i])->TF_Text;
      }
      jedtextdelete((jedtext_t *)ef->EF_FieldAux[i]);
      break;
    case FOFI_TBUTTON:
      break;
    }

  i = ef->EF_Mode;
  free(ef);

  if(!(i & EFIELD_KEEPWINDOW) && jedwstackdel(win) == ERR) return(-1);

  return(0);
}

/*JS*********************************************************************
*
*************************************************************************/

static int doeditfield(jedform_t *ef, int key, int mode)

/************************************************************************/
{
  int curr, attrs, type, line, column, ret;

  curr = ef->EF_CurrForm;
  type = ef->EF_Forms[curr].FF_Type & FOFI_TMASK;

  ret = (mode & 4) ? 0 : 1;
  if(ef->EF_Mode & EFIELD_KEEPWINDOW) ret |= 2;
  attrs = getfldattrs(&ef->EF_Forms[curr], ret);
  wattron(ef->EF_Window, attrs);

  ret = 0;

  line   = ef->EF_LineOff + ef->EF_Forms[curr].FF_Line;
  column = ef->EF_ColOff  + ef->EF_Forms[curr].FF_Column;

#ifdef CONFIG_HAS_CURSSET
  if(mode == 0)
    curs_set(type == FOFI_TBUTTON || type == FOFI_TENUM ? 0 : 1);
#endif

  switch(type) {
  case FOFI_TASCII: case FOFI_TINUM: case FOFI_TFNUM:
    ret = dotext1(ef, line, column, &ef->EF_Forms[curr],
		  (text1aux_t *)(ef->EF_FieldAux[curr]), key, mode);
    break;
  case FOFI_TENUM:
    ret = doenum(ef, line, column, &ef->EF_Forms[curr],
		 (enumaux_t *)(ef->EF_FieldAux[curr]), key, mode);
    break;
  case FOFI_TTEXT:
    ret = jedtextdokey(ef->EF_Window, (jedtext_t *)(ef->EF_FieldAux[curr]), key,
		       mode);
    break;
  case FOFI_TBUTTON:
    ret = dobutton(ef->EF_Window, line, column, &(ef->EF_Forms[curr]), key, mode);
    break;
  }

  wattroff(ef->EF_Window, attrs);

  return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doredrawform(jedform_t *ef)

/************************************************************************/
{
  formfield_t *forms;
  int i, n;

  /*  Print a surrounding frame  */
  if(ef->EF_Mode & EFIELD_FRAME) {
#ifdef CONFIG_FORM_SHADOWS
    _jedprintframe(ef->EF_Window, 0, 0, 0, 0, 11);
#else
    _jedprintframe(ef->EF_Window, 0, 0, 0, 0, 9);
#endif
  }

  forms = ef->EF_Forms;
  n = ef->EF_NForms;
  for(i = 0 ; i < n ; i++) {
    int attrs, l, c;

    l = ef->EF_LineOff + forms[i].FF_Line;
    c = ef->EF_ColOff  + forms[i].FF_Column;
#if 1
    /*  If field is completely invisible, skip it  */
    if((l + forms[i].FF_ScreenHeight) <= ef->EF_TopL || l > ef->EF_BottL)
      continue;
#endif
    attrs = getfldattrs(&forms[i], (ef->EF_Mode & EFIELD_KEEPWINDOW) ? 2 : 0);
    wattron(ef->EF_Window, attrs);

    switch(forms[i].FF_Type & FOFI_TMASK) {
    case FOFI_TASCII: case FOFI_TINUM: case FOFI_TFNUM:
      if(dotext1(ef, l, c, &forms[i],
		 (text1aux_t *)(ef->EF_FieldAux[i]), EDKEY_REDRAW,
		 (forms[i].FF_Type & FOFI_FNONSEL) ? 15 : 7))
	EGOTO(error);
      break;
    case FOFI_TENUM:
      if(doenum(ef, l, c, &forms[i], (enumaux_t *)(ef->EF_FieldAux[i]), EDKEY_REDRAW,
		(forms[i].FF_Type & FOFI_FNONSEL) ? 15 : 7))
	EGOTO(error);
      break;
    case FOFI_TTEXT:
      if(jedtextdokey(ef->EF_Window, (jedtext_t *)(ef->EF_FieldAux[i]),
		      EDKEY_REDRAW, 3))
	EGOTO(error);
      break;
    case FOFI_TBUTTON:
      /*  Buttons are just displayed unpressed  */
      if(dobutton(ef->EF_Window, l, c, &forms[i], EDKEY_REDRAW, 7) == -1)
	EGOTO(error);
      break;
    default:
      /*  Just print non-selectable fields  */
      mvwaddstr2(ef->EF_Window, l, c, forms[i].FF_Text);
      break;
    }

    wattroff(ef->EF_Window, attrs);
  }

  if((ef->EF_Mode & EFIELD_SCROLLBAR) && doredrawscbar(ef) == -1) EGOTO(error);

  return(0);
error:
  return(-1);
}

/*JS*********************************************************************
*
*************************************************************************/

static int doredrawscbar(jedform_t *ef)

/************************************************************************/
{
  int i, n, col;

  col = ef->EF_ColOff + ef->EF_Width - 1;
  n = ef->EF_BottL - ef->EF_TopL + 1;
  for(i = 0 ; i < n ; i++) {
    int l = ef->EF_TopL + i;

    if(mvwaddch(ef->EF_Window, l, col,
		JEDScrollChars[ef->EF_ScrollBarS <= i && i < ef->EF_ScrollBarE ? 1 : 0]) == ERR)
      return -1;
  }

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int dochgfield(jedform_t *ef, int newcurr, int d)

/************************************************************************/
{
  /*  Redraw old current field not highlighted  */
  if(doeditfield(ef, EDKEY_REDRAW, 7) == -1)
    return -1;

  /*  Go to new field  */
  ef->EF_CurrForm = newcurr;

  dochkfield(ef, d);
  if(doformalign(ef) == -1)
    return -1;

  /*  Now redraw new field  */
  return doeditfield(ef, EDKEY_REDRAW, 0);
}

/*JS*********************************************************************
*
*************************************************************************/

static int doformalign(jedform_t *ef)

/************************************************************************/
{
  int s, e, d;

  s = ef->EF_LineOff + ef->EF_Forms[ef->EF_CurrForm].FF_Line;
  e = s + ef->EF_Forms[ef->EF_CurrForm].FF_ScreenHeight - 1;

  d = 0;
  if(e > ef->EF_BottL) d = ef->EF_BottL - e;
  if((s + d) < ef->EF_TopL) d = ef->EF_TopL - s;

  return(doformscroll(ef, d));
}

/*JS*********************************************************************
*
*************************************************************************/

static int doformscroll(jedform_t *ef, int l)

/************************************************************************/
{
  int i;

  i = ef->EF_LineOff;

  ef->EF_LineOff += l;

  if(ef->EF_LineOff > ef->EF_TopL) ef->EF_LineOff = ef->EF_TopL;
  l = ef->EF_LineOff + ef->EF_FormHeight - 1;
  if(l < ef->EF_BottL) ef->EF_LineOff = ef->EF_BottL - ef->EF_FormHeight + 1;

  if((l = ef->EF_LineOff - i) == 0) return(0);

  for(i = 0 ; i < ef->EF_NForms ; i++)
    if((ef->EF_Forms[i].FF_Type & FOFI_TMASK) == FOFI_TTEXT)
      ((jedtext_t *)(ef->EF_FieldAux[i]))->TF_Line += l;

  if(ef->EF_Mode & EFIELD_SCROLLBAR) edformsetscbar(ef);

  werase(ef->EF_Window);
  doredrawform(ef);

  if((ef->EF_Mode & EFIELD_SCROLLBAR) && doredrawscbar(ef) == -1) return(-1);

  return(1);
}

/*JS*********************************************************************
*
*************************************************************************/

static void dochkfield(jedform_t *ef, int d)

/************************************************************************/
{
  int i;

  if(ef->EF_CurrForm == -1) return;

  /*  Select next user editable field  */
  for(i = ef->EF_CurrForm ; ; i += d) {
    if(i == ef->EF_NForms) i = 0;
    else if(i < 0) i = ef->EF_NForms - 1;

    if(!(ef->EF_Forms[i].FF_Type & FOFI_FNONSEL)) break;
  }
  ef->EF_CurrForm = i;
}

/*JS*********************************************************************
*
*************************************************************************/

static int getfldattrs(formfield_t *ff, int flag)

/************************************************************************/
{
  int attrs;

  attrs = ((ff->FF_Type & FOFI_FEMPHASIZE) || (flag & 1)) ? A_BOLD : 0;

#ifdef CONFIG_HAS_COLORS
  if(!(JEditCursesMode & CURSMODE_NOCOLORS) && !(flag & 2) &&
     !(ff->FF_Type & FOFI_FNONSEL))
    attrs |= FORM_COLOR;
#endif

  return attrs;
}

/*JS*********************************************************************
*
*************************************************************************/

static text1aux_t *mkauxtext1(formfield_t *ff)

/************************************************************************/
{
  text1aux_t *paux;

  if((paux =(text1aux_t *)malloc(sizeof(*paux))) == NULL)
    return NULL;

  paux->Window = NULL;
  paux->Start = paux->Current = 0;

  return paux;
}

#define FLAG_REDRAW      (1<<0)
#define FLAG_MOVECURSOR  (1<<1)
#define FLAG_REFRESH     (1<<2)
#define FLAG_ILLEGAL     (1<<3)
#define FLAG_UNKNOWN     (1<<4)

#define FLAG_STORE       (1<<5)
#define FLAG_CHECKCURR   (1<<6)

#define FLAG_CHECKHELP   (1<<7)

/*JS*********************************************************************
*
*************************************************************************/

static int dotext1(jedform_t *ef, int line, int column,
		   formfield_t *ff, text1aux_t *paux, int key, int mode)

/************************************************************************/
{
  int flag, start, curr, type;

  start = paux->Start;
  curr = paux->Current;
  flag = 0;

  switch(key) {
  case EDKEY_REDRAW:
    flag |= FLAG_REDRAW;
    break;
  case EDKEY_LEFT:
    if(curr == 0) {
      flag |= FLAG_ILLEGAL;
      break;
    }
    curr--;
    flag |= FLAG_CHECKCURR|FLAG_REDRAW|FLAG_STORE;
    break;
  case EDKEY_RIGHT:
    if(curr == ff->FF_Len) {
      flag |= FLAG_ILLEGAL;
      break;
    }
    curr++;
    flag |= FLAG_CHECKCURR|FLAG_REDRAW|FLAG_STORE;
    break;
  case EDKEY_BS:
    if(curr == 0) break;
    curr--;
    flag |= FLAG_CHECKCURR|FLAG_MOVECURSOR|FLAG_STORE;
  case EDKEY_DELETE:
    {
      int ret;

      if((ret = doinsert1(ef, ff, curr, NULL, 1, 0)) == -1) EGOTO(error);
      if(ret) {
	flag |= FLAG_ILLEGAL;
	break;
      }
      flag |= FLAG_REDRAW | FLAG_CHECKHELP;
    }
    break;
  case EDKEY_LSTART:
    curr = 0;
    flag |= FLAG_CHECKCURR|FLAG_REDRAW|FLAG_STORE;
    break;
  case EDKEY_LEND:
    curr = ff->FF_Len;
    flag |= FLAG_CHECKCURR|FLAG_REDRAW|FLAG_STORE;
    break;
  case EDKEY_KILLEOL:
    {
      int ret;

      if((ret = doinsert1(ef, ff, curr, NULL, ff->FF_Len - curr, 1)) == -1)
	EGOTO(error);

      if(ret)
	flag |= FLAG_ILLEGAL;
      else
	flag |= FLAG_REDRAW;
    }
    break;
  case EDKEY_YANK:
    {
      const char *s;
      int ret;

      if((s = _jedgetyank())) {
	int l;

	l = strlen(s);
	if((ret = doinsert1(ef, ff, curr, s, l, 0)) == -1) EGOTO(error);
	if(ret == 0) {
	  curr += l;
	  flag |= FLAG_CHECKCURR|FLAG_REDRAW|FLAG_STORE;
	} else
	  flag |= FLAG_ILLEGAL;
      } else
	flag |= FLAG_ILLEGAL;
    }
    break;
  case EDKEY_ROTYANK:
    _jedrotkillring(-1);
    break;
  case EDKEY_TOGGLEINS:
    ef->EF_Mode ^= EFIELD_INSERT;
    break;
#ifdef NCURSES_MOUSE_VERSION
  case KEY_MOUSE:
    {
      MEVENT mev;
      int ml, mc;

      /*  Get position relative to window  */
      if(getmouse(&mev) == ERR) EGOTO(error);
      getbegyx(ef->EF_Window, ml, mc);
      if(mev.y != (ml + line)) {
	flag |= FLAG_ILLEGAL;
	break;
      }

      curr = mev.x - mc - column;
      flag |= FLAG_CHECKCURR|FLAG_REDRAW|FLAG_STORE;
    }
    break;
#endif
  case EDKEY_DOWN:
  case EDKEY_UP:
    if(paux->Window == NULL) {
      flag |= FLAG_UNKNOWN;
      break;
    }

    if(key == EDKEY_DOWN ? paux->WinCurr < paux->NValues - 1 : paux->WinCurr >= 0) {
      int ret;

      if(key == EDKEY_DOWN) {
	paux->WinCurr++;

	/*  maybe scroll?  */
	if((paux->WinCurr - paux->WinOff) >= paux->Height)
	  paux->WinOff++;
      } else if(paux->WinCurr > 0) {
	paux->WinCurr--;

	/*  maybe scroll?  */
	if(paux->WinCurr < paux->WinOff)
	  paux->WinOff--;
      } else {
	paux->WinCurr = -1;
      }

      if(paux->WinCurr >= 0) {
	const char *s = paux->Values[paux->WinCurr];

	if((ret = doinsert1(ef, ff, 0, s, strlen(s), 2)) == -1)
	  EGOTO(error);
      } else
	ret = 0;

      if(ret == 0) {
	start = 0;
	curr = ff->FF_Len;
	flag |= FLAG_CHECKCURR | FLAG_REDRAW | FLAG_STORE;
      } else
	flag |= FLAG_ILLEGAL;
    }

    break;

  default:
    /*  Literal key?  */
    if(key < 0) key = -key;
    if(key >= 256 || key == '\n' || !isprint(key)) {
      flag |= FLAG_UNKNOWN;
      break;
    }

    type = (ff->FF_Type & FOFI_TMASK);
    if((type == FOFI_TINUM && !isdigit(key)) ||
       (type == FOFI_TFNUM && !(isdigit(key) || strchr(".+-eE", key)))) {
      flag |= FLAG_ILLEGAL;
      break;
    }

    {
      char c;
      int ret;

      c = key;
      if((ret = doinsert1(ef, ff, curr, &c, 1, 0)) == -1)
	EGOTO(error);

      if(ret == 0) {
	curr++;
	flag |= FLAG_CHECKCURR | FLAG_REDRAW | FLAG_MOVECURSOR | FLAG_STORE | FLAG_CHECKHELP;

      } else
	flag |= FLAG_ILLEGAL;
    }
    break;
  }

  if(flag & FLAG_CHECKCURR) {
    int w;

    w = ff->FF_ScreenWidth;
    if(curr < 0) {
      curr = 0;
      flag |= FLAG_STORE;
    } else if(curr > ff->FF_Len) {
      curr = ff->FF_Len;
      flag |= FLAG_STORE;
    }

    if(curr < start) {
      /*  Move window to the right  */
      do
	start -= FORM_SCROLL_SKIP;
      while(curr < start);
      if(start < 0) start = 0;

      flag |= FLAG_REDRAW|FLAG_STORE;
    } else if(curr > (start + w)) {
      /*  Move window to the left  */
      do
	start += FORM_SCROLL_SKIP;
      while(curr > (start + w));
      if(start > ff->FF_Len) start = ff->FF_Len;

      flag |= FLAG_REDRAW|FLAG_STORE;
    }
  }

  if(mode & 4) {
    if(paux->Window != NULL) {
      jedwstackdel(paux->Window);
      paux->Window = NULL;
      flag |= FLAG_REDRAW;
    }

    /*  Remember this field value if it has a length  */
    if((ff->FF_Type & FOFI_FCHANGED) && !(ff->FF_Type & (FOFI_FNONSEL|FOFI_FHIDDEN)) &&
       ff->FF_Len > 0 && ff->FF_Name && ff->FF_Name[0] &&
       _jedhistremember(ff->FF_Name, ff->FF_Text, ff->FF_Len) == -1)
      EGOTO(error);

  } else if(flag & FLAG_CHECKHELP) {
    if(ff->FF_Type & (FOFI_FNONSEL | FOFI_FHIDDEN)) {
      paux->Values = NULL;
      paux->NValues = 0;
    } else if((paux->Values = _jedhistgetvalues(ff->FF_Name, ff->FF_Text,
						&(paux->NValues))) == NULL)
      EGOTO(error);

    if(paux->Window != NULL) {
      int nlines, ncols;

      /*  Check if the window size changed  */
      getmaxyx(paux->Window, nlines, ncols);
      if(nlines != paux->NValues) {
	jedwstackdel(paux->Window);
	paux->Window = NULL;
	flag |= FLAG_REDRAW;
      }
    }

    if(paux->NValues > 0) {
      if(paux->Window == NULL) {
	int y, x, w;

	getbegyx(ef->EF_Window, y, x);
	y += line + 1;
	x += column;

 	paux->Height = MIN(MIN(6, paux->NValues), ef->EF_FormHeight - line - 1);
	w = ff->FF_ScreenWidth;

#ifdef CONFIG_USE_SHADOWSx
	/*  FIXME: Doesn't work - refresh problem  */
	if((paux->Window = newwin(paux->Height + 1, w + 2, y, x)) == NULL)
	  EGOTO(error);

	_jedprintframe(paux->Window, 0, 0, paux->Height + 1, w + 2, 19);
#else
	if((paux->Window = newwin(paux->Height, w, y, x)) == NULL)
	  EGOTO(error);
#endif

#ifdef CONFIG_HAS_COLORS
	if(!(JEditCursesMode & CURSMODE_NOCOLORS))
	  wattron(paux->Window, FORM_HELPCOLOR);
#endif

	paux->WinOff = 0;
	paux->WinCurr = -1;

	flag |= FLAG_REDRAW | FLAG_MOVECURSOR;
      }
    }
  }

  if(flag & FLAG_REDRAW) {
    WINDOW *win;
    const unsigned char *text;
    int slen, len2, c;

    /*  Show helper window  */
    if((win = paux->Window) != NULL) {
      int sbstart, sbend;

      slen = ff->FF_ScreenWidth;
      if(paux->Height < paux->NValues) {
	--slen;
	_jedsetscrollbar(paux->WinOff, paux->NValues, paux->Height,
			 &sbstart, &sbend);
      } else
	sbstart = sbend = -1;

      for(c = 0 ; c < paux->Height ; ++c) {
	const char *string;
	int i;

	i = paux->WinOff + c;
	string = paux->Values[i];

	if(i == paux->WinCurr) wattron(win, A_REVERSE);
	mvwaddnstr(win, c, 0, string, slen);
	len2 = slen - strlen(string);
	while(len2-- > 0) waddch(win, ' ');
	if(i == paux->WinCurr) wattroff(win, A_REVERSE);

	if(sbstart >= 0)
	  waddch(win, JEDScrollChars[sbstart <= c && c < sbend ? 1 : 0]);
      }
    }

    slen = ff->FF_ScreenWidth;

    len2 = ff->FF_Len - start;
    if(len2 > slen) len2 = slen;

    win = ef->EF_Window;

    if(ff->FF_Type & FOFI_FSHOWSIGNS)
      mvwaddch(win, line, column - 1, start <= 0 ? ' ' :
	       (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '<' :
	       ACS_LARROW);
    else
      wmove(win, line, column);

    text = (const unsigned char *)ff->FF_Text + start;
    for(c = 0 ; c < len2 ; ++c) {
      int c2 = (ff->FF_Type & FOFI_FHIDDEN) ? '*' : text[c];

#if 0
      if(c2 == '\n')
	c2 = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '<' : ACS_LARROW;
      else if(c2 == '\t')
	c2 = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '>' : ACS_RARROW;
#else
      if(iscntrl(c2))
	c2 = A_REVERSE | ('@' + (c2 & 0x1f));
#endif
      waddch(win, c2);
    }

    c = (mode & 8) ? ' ' : JEditTFFillChars;
    while(len2++ < slen)
      waddch(win, c);

    if(ff->FF_Type & FOFI_FSHOWSIGNS)
      waddch(win, ff->FF_Len < slen || text[slen] == '\0' ? ' ' :
	     (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '>' :
	     ACS_RARROW);

    flag |= FLAG_REFRESH | FLAG_MOVECURSOR;
  }

  if((flag & FLAG_MOVECURSOR) && !(mode & 1)) {
    wmove(ef->EF_Window, line, column + curr - start);
  }

  if((flag & FLAG_REFRESH) && !(mode & 2)) {
    jedwstackrefresh(ef->EF_Window);

    if(paux->Window != NULL)
      wrefresh(paux->Window);
  }

  if(flag & FLAG_STORE) {
    paux->Start = start;
    paux->Current = curr;
  }

  if(flag & FLAG_UNKNOWN) return -2;
  if(flag & FLAG_ILLEGAL) return -3;

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*   mode bit 0: insert into kill ring.
*   bit 1: overwrite content starting from offset.
*************************************************************************/

static int doinsert1(jedform_t *ef, formfield_t *ff, long off,
		     const char *s, int len, int mode)

/************************************************************************/
{
  char *text;

  text = ff->FF_Text;
  if(s) {
    long e;

    e = !(mode & 2) && (ef->EF_Mode & EFIELD_INSERT) ? ff->FF_Len + len : off + len;

    if(e > ff->FF_MaxLen) {
      if(!(ff->FF_Type & FOFI_FALLOCATE)) return 1;

      /*  Increase field size  */
      ff->FF_MaxLen = MAX(e, ff->FF_MaxLen + 50);
      if((ff->FF_Text = text = (char *)
	  realloc(text, ff->FF_MaxLen + 1)) == NULL)
	EGOTO(error);
    }

    if(mode & 2) {
      ff->FF_Len = e;
      text[e] = '\0';
    } else if(ef->EF_Mode & EFIELD_INSERT) {
      memmove(text + off + len, text + off, ff->FF_Len - off + 1);
      ff->FF_Len += len;
    } else if(e > ff->FF_Len) {
      ff->FF_Len = e;
      text[e] = '\0';
    }
    memcpy(text + off, s, len);
  } else {
    if(len > (ff->FF_Len - off)) len = ff->FF_Len - off;

    if(len == 0) return 1;

    if((mode & 1) && _jedinskillring(text + off, len) == -1)
      EGOTO(error);

    memmove(text + off, text + off + len, ff->FF_Len - off - len + 1);
    ff->FF_Len -= len;
  }

  ff->FF_Type |= FOFI_FCHANGED;

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static void freeauxtext1(jedform_t *ef, formfield_t *ff, text1aux_t *ea)

/************************************************************************/
{
  if(ea->Window != NULL)
    jedwstackdel(ea->Window);

  free(ea);
}

/*JS*********************************************************************
*
*************************************************************************/

static int dobutton(WINDOW *win, int line, int column, formfield_t *ff,
		    int key, int mode)

/************************************************************************/
{
  int len;

  len = strlen(ff->FF_Text);
  if(mvwaddnstr(win, line, column, ff->FF_Text, len) == ERR)
    return(-1);

  _jedprintframe(win, line - 1, column - 1, 3, len + 2, (mode & 4) ? 1 : 5);

  if(key == EDKEY_ENTER
#ifdef NCURSES_MOUSE_VERSION
     || key == KEY_MOUSE
#endif
     )
    return(1 + (ff->FF_Type & FOFI_LMASK));

  if(key != EDKEY_REDRAW) return(-2);

  return(0);
}

/*JS*********************************************************************
*
*************************************************************************/

static enumaux_t *mkauxenum(formfield_t *ff)

/************************************************************************/
{
  enumaux_t *ea;
  int i, l;

  l = 0;
  for(i = 0 ; i < ff->FF_NList ; i++) {
    int l2 = strlen(ff->FF_List[i]);
    if(l2 > l) l = l2;
  }
  ff->FF_ScreenWidth = l;

  if((ea = (enumaux_t *)malloc(sizeof(ea)))) ea->EA_Table = NULL;
  ff->FF_Curr = 0;

  return(ea);
}

/*JS*********************************************************************
*
*************************************************************************/

static int doenum(jedform_t *ef, int line, int column, formfield_t *ff,
		  enumaux_t *ea, int key, int mode)

/************************************************************************/
{
  int flag;

restart:

  flag = 0;
  if(mode & 4) {
    if(ea->EA_Table && dofreeenumwin(ef, ff, ea) == -1)
      EGOTO(error);
  }
#ifdef CONFIG_POPUPENUM
  /*  We enter this field, so pop up list  */
  else if(ea->EA_Table == NULL && domkenumwin(ef, line, column, ff, ea) == -1)
    EGOTO(error);
#endif

  /*  If the table is already popped up, handle key in it  */
  if(ea->EA_Table) {
    int ret;

    mode = (mode & 2) ? TABFORM_FNOREFRESH|TABFORM_FQUICKLOCATE :
      TABFORM_FQUICKLOCATE;
    ret = jedtabdokey(ea->EA_Table, key, mode);
    if(key != EDKEY_ENTER) return ret;

    if(dofreeenumwin(ef, ff, ea) == -1) EGOTO(error);
    key = EDKEY_REDRAW;
  }

  switch(key) {
  case EDKEY_REDRAW:
    flag |= FLAG_REDRAW;
    break;
  case EDKEY_LEFT:
    if(ff->FF_Curr <= 0) {
      flag |= FLAG_ILLEGAL;
      break;
    }
    ff->FF_Curr--;
    flag |= FLAG_REDRAW;
    break;
  case EDKEY_RIGHT:
    if(ff->FF_Curr >= (ff->FF_NList - 1)) {
      flag |= FLAG_ILLEGAL;
      break;
    }
    ff->FF_Curr++;
    flag |= FLAG_REDRAW;
    break;
  default:
    if(!isprint(key)) {
      flag |= FLAG_UNKNOWN;
      break;
    }
  case EDKEY_ENTER:
    if(ea->EA_Table == NULL && domkenumwin(ef, line, column, ff, ea) == -1)
      EGOTO(error);

    if(!(mode & 2) && jedtabdokey(ea->EA_Table, EDKEY_REDRAW, 0) == -1)
      EGOTO(error);

    /*  Handle this key in the popped up window  */
    if(key != EDKEY_ENTER) goto restart;
    break;
  }

  if(flag & FLAG_REDRAW) {
#ifdef SHOW_ENUM_CENTERED
    int l;

    l = (ff->FF_ScreenWidth - strlen(ff->FF_List[ff->FF_Curr])) / 2;
    if(l < 0) l = 0;
#endif

    wattron(ef->EF_Window, A_REVERSE);
#ifdef SHOW_ENUM_CENTERED
    mvwprintw(ef->EF_Window, line, column, "%-*s%-*s",
	      l, "", ff->FF_ScreenWidth - l, ff->FF_List[ff->FF_Curr]);
#else
    mvwprintw(ef->EF_Window, line, column, "%-*s",
	      ff->FF_ScreenWidth, ff->FF_List[ff->FF_Curr]);
#endif
    mvwaddch(ef->EF_Window, line, column + ff->FF_ScreenWidth,
	     (JEditCursesMode & CURSMODE_NOACSGRAPH) ? A_REVERSE | 'V' :
	     A_REVERSE | ACS_DARROW);

    wattroff(ef->EF_Window, A_REVERSE);
    flag |= FLAG_REFRESH;
  }

  if((flag & FLAG_REFRESH) && !(mode & 2))
    jedwstackrefresh(ef->EF_Window);

  if(flag & FLAG_UNKNOWN) return -2;
  if(flag & FLAG_ILLEGAL) return -3;

  return 0;
error:
  return -1;
}

#undef FLAG_REDRAW
#undef FLAG_MOVECURSOR
#undef FLAG_REFRESH
#undef FLAG_ILLEGAL
#undef FLAG_UNKNOWN

#undef FLAG_STORE
#undef FLAG_CHECKCURR

#undef FLAG_CHECKHELP

/*JS*********************************************************************
*
*************************************************************************/

static int domkenumwin(jedform_t *ef, int line, int column,
		       formfield_t *ff, enumaux_t *ea)

/************************************************************************/
{
  WINDOW *win;
  long start[2];
  int l, c, h, w, l1, l2, mode;

  l = line - ff->FF_Curr;
  l1 = MAX(l, ef->EF_TopL);

  c = ef->EF_BottL;

  l2 = MIN(l + ff->FF_NList - 1, c);

  h = l2 - l1 + 1;
  w = ff->FF_ScreenWidth;

  if(h < ff->FF_NList) {
    mode = FOFI_SCROLLBAR;
    w++;
  } else
    mode = 0;

  getbegyx(ef->EF_Window, l, c);
  l += l1;
  c += column;

#ifdef CONFIG_USE_SHADOWS
  win = jedwstacknew(h + 1, w + 2, l, c);
#else
  win = jedwstacknew(h, w, l, c);
#endif
  if(win == NULL) EGOTO(error);

#ifdef CONFIG_HAS_COLORS
  if(!(JEditCursesMode & CURSMODE_NOCOLORS) &&
     wbkgd(win, FORM_ENUMCOLOR | ' ') == ERR)
    EGOTO(error);
#endif

#ifdef CONFIG_USE_SHADOWS
  _jedprintframe(win, 0, 0, h + 1, w + 2, 19);
#endif

  keypad(win, TRUE);
  intrflush(win, FALSE);

  start[0] = ff->FF_Curr - line + l1;
  start[1] = ff->FF_Curr;

  ef->EF_InWin = win;

  if((ea->EA_Table = (jedtable_t *)
      jedtabcreate(win, 0, 0, h, w,
		    (int (*)(void *, long, char *, int))dogetenum,
		    ff->FF_List, 1, ff->FF_NList, start, mode)) == NULL)
    EGOTO(error);

  return(0);
error:
  return(-1);
}

/*JS*********************************************************************
*
*************************************************************************/

static int dofreeenumwin(jedform_t *ef, formfield_t *ff, enumaux_t *ea)

/************************************************************************/
{
  ef->EF_InWin = ef->EF_Window;
  ff->FF_Curr = ea->EA_Table->TA_Current;

  jedwstackdel(ea->EA_Table->TA_Window);
  jedtabdelete(ea->EA_Table);
  ea->EA_Table = NULL;
  return(0);
}

/*JS*********************************************************************
*
*************************************************************************/

static int dogetenum(const char *const *list, long nr, char *dest, int m)

/************************************************************************/
{
  strncpy(dest, list[nr], m);
  dest[m] = '\0';
  return(0);
}

/*JS*********************************************************************
*
*************************************************************************/

static void freeauxenum(jedform_t *ef, formfield_t *ff, enumaux_t *ea)

/************************************************************************/
{
  if(ea->EA_Table) dofreeenumwin(ef, ff, ea);
  free(ea);
}
