/*JS*********************************************************************
*
*    Program : JEDLGRAPH
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Routines for line graphics.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <string.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS  */
/* ***  Double line graphics. Only work on linux console  *** */
/*  single lines horizontally and single lines vertically  */
#define ACS_ULCORNER11     JEDACSMap[0x06]
#define ACS_LLCORNER11     JEDACSMap[0x03]
#define ACS_URCORNER11     JEDACSMap[0x0c]
#define ACS_LRCORNER11     JEDACSMap[0x09]
#define ACS_RTEE11         JEDACSMap[0x0d]
#define ACS_LTEE11         JEDACSMap[0x07]
#define ACS_BTEE11         JEDACSMap[0x0b]
#define ACS_TTEE11         JEDACSMap[0x0e]
#define ACS_HLINE11        JEDACSMap[0x0a]
#define ACS_VLINE11        JEDACSMap[0x05]
#define ACS_PLUS11         JEDACSMap[0x0f]

/*  double lines horizontally and single lines vertically  */
#define ACS_ULCORNER21     JEDACSMap[0x16]
#define ACS_LLCORNER21     JEDACSMap[0x13]
#define ACS_URCORNER21     JEDACSMap[0x1c]
#define ACS_LRCORNER21     JEDACSMap[0x19]
#define ACS_RTEE21         JEDACSMap[0x1d]
#define ACS_LTEE21         JEDACSMap[0x17]
#define ACS_BTEE21         JEDACSMap[0x1b]
#define ACS_TTEE21         JEDACSMap[0x1e]
#define ACS_HLINE21        JEDACSMap[0x1a]
#define ACS_VLINE21        JEDACSMap[0x15]
#define ACS_PLUS21         JEDACSMap[0x1f]

/*  single lines horizontally and double lines vertically  */
#define ACS_ULCORNER12     JEDACSMap[0x26]
#define ACS_LLCORNER12     JEDACSMap[0x23]
#define ACS_URCORNER12     JEDACSMap[0x2c]
#define ACS_LRCORNER12     JEDACSMap[0x29]
#define ACS_RTEE12         JEDACSMap[0x2d]
#define ACS_LTEE12         JEDACSMap[0x27]
#define ACS_BTEE12         JEDACSMap[0x2b]
#define ACS_TTEE12         JEDACSMap[0x2e]
#define ACS_HLINE12        JEDACSMap[0x2a]
#define ACS_VLINE12        JEDACSMap[0x25]
#define ACS_PLUS12         JEDACSMap[0x2f]

/*  double lines horizontally and double lines vertically  */
#define ACS_ULCORNER22     JEDACSMap[0x36]
#define ACS_LLCORNER22     JEDACSMap[0x33]
#define ACS_URCORNER22     JEDACSMap[0x3c]
#define ACS_LRCORNER22     JEDACSMap[0x39]
#define ACS_RTEE22         JEDACSMap[0x3d]
#define ACS_LTEE22         JEDACSMap[0x37]
#define ACS_BTEE22         JEDACSMap[0x3b]
#define ACS_TTEE22         JEDACSMap[0x3e]
#define ACS_HLINE22        JEDACSMap[0x3a]
#define ACS_VLINE22        JEDACSMap[0x35]
#define ACS_PLUS22         JEDACSMap[0x3f]

#define lgmvwaddch(win, y, x, ch) (wmove(win, y, x) == ERR ? ERR : lgwaddch(win, ch))
/*  END-DEFINITIONS  */

#define USE_VT100_COL

#define IPrototype extern
#define Prototype extern
/*********     PROTOTYPES                                       *********/
IPrototype void          _initmyacs(int mode);
Prototype int            lgwaddch(WINDOW *win, chtype c);

Prototype chtype         JEDACSMap[];
Prototype chtype         JEDScrollChars[];

/*********     GLOBAL VARIABLES                                 *********/
/*  Special ACS characters (only for LINUX console)  */
chtype JEDACSMap[4 * 16]
#ifdef A_ALTCHARSET
 = {
#define A(x)     (A_ALTCHARSET | x)

  /*  single lines horizontally and vertically  */
  ' ',       0,       0,     A(0xc0),   0,     A(0xb3), A(0xda), A(0xc3),
    0,     A(0xd9), A(0xc4), A(0xc1), A(0xbf), A(0xb4), A(0xc2), A(0xc5),

  /*  double lines horizontally and single lines vertically  */
  ' ',       0,       0,     A(0xd4),   0,     A(0xb3), A(0xd5), A(0xc6),
    0,     A(0xbe), A(0xcd), A(0xcf), A(0xb8), A(0xb5), A(0xd1), A(0xd8),

  /*  single lines horizontally and double lines vertically  */
  ' ',       0,       0,     A(0xd3),   0,     A(0xba), A(0xd6), A(0xc7),
    0,     A(0xbd), A(0xc4), A(0xd0), A(0xb7), A(0xb6), A(0xd2), A(0xd7),

  /*  double lines horizontally and double lines vertically  */
  ' ',       0,       0,     A(0xc8),   0,     A(0xba), A(0xc9), A(0xcc),
    0,     A(0xbc), A(0xcd), A(0xca), A(0xbb), A(0xb9), A(0xcb), A(0xce)
}
#endif
;

#ifdef USE_VT100_COL
static chtype Mapvt100_col[4 * 16]
#ifdef A_ALTCHARSET
 = {
  /*  single lines horizontally and vertically  */
  ' ',       0,       0,     A(0xed),   0,     A(0xf8), A(0xec), A(0xf4),
    0,     A(0xea), A(0xf2), A(0xf6), A(0xeb), A(0xf5), A(0xf7), A(0xee),

  /*  double lines horizontally and single lines vertically  */
  ' ',       0,       0,     A(0xd3),   0,     A(0xf8), A(0xd2), A(0xd4),
    0,     A(0xd0), A(0xaf), A(0xd6), A(0xd1), A(0xd5), A(0xd7), A(0xfc),

  /*  single lines horizontally and double lines vertically  */
  ' ',       0,       0,     A(0xcb),   0,     A(0xb1), A(0xca), A(0xcc),
    0,     A(0xc8), A(0xf2), A(0xce), A(0xc9), A(0xcd), A(0xcf), A(0xa8),

  /*  double lines horizontally and double lines vertically  */
  ' ',       0,       0,     A(0xdb),   0,     A(0xb1), A(0xda), A(0xdc),
    0,     A(0xd8), A(0xaf), A(0xde), A(0xd9), A(0xdd), A(0xdf), A(0xe9),
}

#undef A
#endif
;
#endif

/*  Characters for the scrollbar  */
chtype JEDScrollChars[] = { ' ', '#' };

/*JS*********************************************************************
*
*************************************************************************/

void _initmyacs(int mode)

/************************************************************************/
{
  if(!(mode & 1)) {
#ifdef A_ALTCHARSET
    const char *t;

    t = getenv("TERM");

#ifdef USE_VT100_COL
    if(t && strcmp(t, "vt100-col") == 0)
      memcpy(JEDACSMap, Mapvt100_col, sizeof(JEDACSMap));
    else
#endif
      if(t != NULL && strncmp(t, "linux", 5) == 0) {
	JEDScrollChars[1] = A_BOLD|ACS_BLOCK;
	return;
      }
#endif
  }

  if(!(mode & 2)) {
    int i;

    /*  Not linux console, so fall back on one-line ACS characters  */
    for(i = 0 ; i < sizeof(JEDACSMap) / sizeof(JEDACSMap[0]) ; i += 16) {
      JEDACSMap[i + 0x6] = ACS_ULCORNER;
      JEDACSMap[i + 0x3] = ACS_LLCORNER;
      JEDACSMap[i + 0xc] = ACS_URCORNER;
      JEDACSMap[i + 0x9] = ACS_LRCORNER;
      JEDACSMap[i + 0xd] = ACS_RTEE;
      JEDACSMap[i + 0x7] = ACS_LTEE;
      JEDACSMap[i + 0xb] = ACS_BTEE;
      JEDACSMap[i + 0xe] = ACS_TTEE;
      JEDACSMap[i + 0xa] = ACS_HLINE;
      JEDACSMap[i + 0x5] = ACS_VLINE;
      JEDACSMap[i + 0xf] = ACS_PLUS;
    }

    JEDScrollChars[1] = A_BOLD|ACS_BLOCK;

  } else {
    int i;

    /*  Be on the safe side and 'emulate' the characters  */
    for(i = 0 ; i < sizeof(JEDACSMap) / sizeof(JEDACSMap[0]) ; i += 16) {
      JEDACSMap[i + 0x6] =
	JEDACSMap[i + 0x3] =
	JEDACSMap[i + 0xc] =
	JEDACSMap[i + 0x9] =
	JEDACSMap[i + 0xd] =
	JEDACSMap[i + 0x7] =
	JEDACSMap[i + 0xb] =
	JEDACSMap[i + 0xe] = '+';
      JEDACSMap[i + 0xa] = '-';
      JEDACSMap[i + 0x5] = '|';
      JEDACSMap[i + 0xf] = '+';
    }

    JEDScrollChars[1] = '#';
  }
}

/*JS*********************************************************************
*
*************************************************************************/

int lgwaddch(WINDOW *win, chtype c)

/************************************************************************/
{
#ifdef A_ALTCHARSET
  chtype o, n;
  int i, j;

  o = winch(win) & (A_ALTCHARSET|A_CHARTEXT);
  for(i = 0 ; i < sizeof(JEDACSMap) / sizeof(JEDACSMap[0]) ; i++)
    if(JEDACSMap[i] == o)
      break;
  if(i == sizeof(JEDACSMap) / sizeof(JEDACSMap[0]))
    i = 0;

  n = c & (A_ALTCHARSET|A_CHARTEXT);
  for(j = 0 ; j < sizeof(JEDACSMap) / sizeof(JEDACSMap[0]) ; j++)
    if(JEDACSMap[j] == n)
      break;
  if(j == sizeof(JEDACSMap) / sizeof(JEDACSMap[0]))
    j = 0;

  c = (c & ~(A_ALTCHARSET | A_CHARTEXT)) | JEDACSMap[i | j];
#endif

  return waddch(win, c);
}

#ifdef TEST

#define MKROUTINE(prframe) \
static void prframe(WINDOW *win, int offy, int offx, int height, int width) \
{ \
  int c, clb, clw, endx, endy, i; \
 \
  clb = clw = 0; \
  endx = offx + width - 1; \
  endy = offy + height - 1; \
 \
  lgmvwaddch(win, offy, offx, clb | C(ACS_ULCORNER)); \
  lgmvwaddch(win, offy, endx, clw | C(ACS_URCORNER)); \
  c = clb | C(ACS_HLINE); \
  for(i = offx + 1 ; i < endx ; i++) { \
    lgmvwaddch(win, offy, i, c); \
    lgmvwaddch(win, endy, i, clw | C(ACS_HLINE)); \
  } \
  lgmvwaddch(win, endy, offx, clb | C(ACS_LLCORNER)); \
  lgmvwaddch(win, endy, endx, clw | C(ACS_LRCORNER)); \
 \
  for(i = offy + 1 ; i < endy ; i++) { \
    lgmvwaddch(win, i, offx, clb | C(ACS_VLINE)); \
    lgmvwaddch(win, i, endx, clw | C(ACS_VLINE)); \
  } \
}

#define C(a)   a##11
MKROUTINE(prframe11)
#undef C

#define C(a)   a##12
MKROUTINE(prframe12)
#undef C

#define C(a)   a##21
MKROUTINE(prframe21)
#undef C

#define C(a)   a##22
MKROUTINE(prframe22)
#undef C

int main(int argc, char *argv[])
{
  int i;

  initscr();
  _initmyacs(argc > 1 ? atoi(argv[1]) : 0);

  printw("              --> Pattern to test line graphics <--\n\n\n\
      0123456789abcdef\n");
  for(i = 0 ; i < 64 ; i++) {
    if((i & 15) == 0) {
      if(i != 0) printw("'\n");
      printw("%#04x '", i);
    }
    addch(JEDACSMap[i] ? JEDACSMap[i] : ' ');
  }

#define PRPATTERN(prframe, x, y) \
  do { \
    prframe(stdscr, y  , x   , 5, 12); \
    prframe(stdscr, y  , x+11, 5, 12); \
    prframe(stdscr, y+4, x   , 5, 12); \
    prframe(stdscr, y+4, x+11, 5, 12); \
    prframe(stdscr, y+2, x+ 5, 5, 13); \
  } while(0)

  PRPATTERN(prframe11, 28, 1);
  PRPATTERN(prframe12, 52, 1);
  PRPATTERN(prframe21, 28, 10);
  PRPATTERN(prframe22, 52, 10);

  move(0, 0);

  getch();
  endwin();

  return 0;
}
#endif
