/*JS*********************************************************************
*
*    Program : JEDMENU
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Show a menu.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <jssubs.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

/*********     DEFINES                                          *********/
#ifdef CONFIG_NO_STRDUP
# define strdup  mystrdup  /*  strdup is not POSIX conformant  */
#endif

#ifdef CONFIG_USE_SHADOWS
#define CONFIG_MENU_SHADOWS
#endif

#ifndef MF_Value
/*  START-DEFINITIONS  */
typedef struct jedmenufld_t jedmenufld_t;
struct jedmenufld_t {
  unsigned long MF_Type;
  jedmenufld_t    *MF_Next;

  char         *MF_Text;
  union {
    int         mf_Value;
    jedmenufld_t  *mf_SubMenu;
    long       *mf_Variable;
#define MF_Value      c.mf_Value
#define MF_SubMenu    c.mf_SubMenu
#define MF_Variable   c.mf_Variable
  } c;
  char         *MF_Help;
};

/*  For the type field  */
#define MENU_NMASK       ((1UL<<24)-1)
#define MENU_TMASK       (0xffUL<<24)
#define MENU_TEXT        (1UL<<24)
#define MENU_SUBMENU     (2UL<<24)
#define MENU_VARIABLE    (3UL<<24)

/*  For routine editMenu  */
#define MENUM_PRINTONLY  (1 << 1)
#define MENUM_VERTICAL   (1 << 2)
#define MENUM_NEWWIN     (1 << 3)
#define MENUM_FRAME      (1 << 4)
#define MENUM_MARKSUB    (1 << 5)
/*  END-DEFINITIONS  */
#endif

#ifdef CONFIG_HAS_COLORS
# define MENU_BKGRND        JEditColors[_EDLIB_MCOLIND + 0]
# define MENU_COLOR         JEditColors[_EDLIB_MCOLIND + 1]
#endif

struct MStack {
  WINDOW    *S_Window;
  jedmenufld_t *S_Menu;
  int        S_Mode, S_N, S_Pos;
  int        S_Column, S_Line, S_Width, S_Height, S_Width2, S_Height2;
  int        S_CurrX, S_CurrY;
};

/*  How to mark toggle entries  */
#define CHECK_MARK_ON   '*'
#define CHECK_MARK_OFF  '-'

/*  Shall we wrap around when going down or up in a menu?  */
#define CONFIG_EDLIB_WRAPMENU

#define ESC   '\033'

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            jedmenudokey(int line, int column, jedmenufld_t *menus,
				  int key, int mode);
Prototype jedmenufld_t  *jedmenufldadd(jedmenufld_t **pstart, const char **names,
				       int nnames, long type);
Prototype jedmenufld_t  *jedmenufldfind(jedmenufld_t *start, const char **names,
					int nnames);

/*********     GLOBAL VARIABLES                                 *********/

#define FL_SETUP          (1 << 0)
#define FL_REFRESH        (1 << 1)
#define FL_DRAWMENU       (1 << 4)
#define FL_GOUP           (1 << 5)
#define FL_GODOWN         (1 << 6)
#define FL_KEEPINCHAR     (1 << 7)
#define FL_WRAPANDDOWN    (1 << 8)

/*JS*********************************************************************
*   Edit a menu, displaying it at position line,column. Assumes that
*    already 'key' has been pressed (if != -1).
*************************************************************************/

int jedmenudokey(int line, int column, jedmenufld_t *menus, int key, int mode)

/************************************************************************/
{
  struct MStack *stack;
  int ret, depth, mDepth, oldVisi, flag;
  int currx, curry, width, width2, height, height2, pos, n, c;
  WINDOW *win;
#ifdef NCURSES_MOUSE_VERSION
  MEVENT mev;
#endif

  stack = NULL;
  mDepth = depth = 0;
  ret = -1;

  flag = FL_SETUP;
  if(key != -1) {
    flag |= FL_KEEPINCHAR;

    c = key;
#ifdef NCURSES_MOUSE_VERSION
    if(c == KEY_MOUSE && getmouse(&mev) == ERR) goto ende;
#endif
  }

#ifdef CONFIG_HAS_CURSSET
  oldVisi = curs_set(0);
#endif

  while(ret == -1) {
    jedmenufld_t *ptr;
    int i;

    /* ***  Go one level up  *** */
    if(flag & FL_GOUP) {
      if(depth == 0) {
	ret = -2;
	break;
      }

      if(jedwstackdel(win) == ERR) break;

      depth--;
      win    = stack[depth].S_Window;
      menus  = stack[depth].S_Menu;
      mode   = stack[depth].S_Mode;
      win    = stack[depth].S_Window;
      n      = stack[depth].S_N;
      pos    = stack[depth].S_Pos;
      column = stack[depth].S_Column;
      line   = stack[depth].S_Line;
      width  = stack[depth].S_Width;
      height = stack[depth].S_Height;
      width2 = stack[depth].S_Width2;
      height2= stack[depth].S_Height2;
      currx  = stack[depth].S_CurrX;
      curry  = stack[depth].S_CurrY;

      flag |= FL_DRAWMENU;
    }

    if(flag & (FL_GODOWN|FL_SETUP)) {
      if(depth >= mDepth) {
	mDepth += 5;
	if((stack = (struct MStack *)realloc(stack, mDepth * sizeof(*stack)))
	   == NULL) goto ende;
      }
    }

    /* ***  Go one level down  *** */
    if(flag & FL_GODOWN) {
      for(i = 0, ptr = menus ; i < pos ; ++i) ptr = ptr->MF_Next;

      if((ptr->MF_Type & MENU_TMASK) == MENU_SUBMENU && ptr->MF_SubMenu) {
	stack[depth].S_Window = win;
	stack[depth].S_Menu   = menus;
	stack[depth].S_Mode   = mode;
	stack[depth].S_N      = n;
	stack[depth].S_Pos    = pos;
	stack[depth].S_Column = column;
	stack[depth].S_Line   = line;
	stack[depth].S_Width  = width;
	stack[depth].S_Height = height;
	stack[depth].S_Width2 = width2;
	stack[depth].S_Height2= height2;
	stack[depth].S_CurrX  = currx;
	stack[depth].S_CurrY  = curry;
	depth++;

	menus  = ptr->MF_SubMenu;
	column = currx;
	line   = curry;
	mode   = MENUM_VERTICAL | MENUM_NEWWIN | MENUM_FRAME | MENUM_MARKSUB;
	flag |= FL_SETUP; /* will finally setup currx and curry  */
      } else {
	beep();  flag |= FL_REFRESH;
      }
    }

    /* ***  Set up a menu  *** */
    if(flag & FL_SETUP) {
      for(n = 0, ptr = menus ; ptr ; ptr = ptr->MF_Next) ++n;

      /*  Calculate needed size  */
      if(mode & MENUM_VERTICAL) {
	height = 0;
	width = 0;
	for(ptr = menus ; ptr ; ptr = ptr->MF_Next, height++) {
	  int l;

          l = strlen(ptr->MF_Text);
	  if((ptr->MF_Type & MENU_TMASK) == MENU_VARIABLE ||
	     ((mode & MENUM_MARKSUB) &&
	      (ptr->MF_Type & MENU_TMASK) == MENU_SUBMENU))
	    l++;
	  if(l > width) width = l;
	}
      } else {
	height = 1;
	width = 0;
	for(ptr = menus ; ptr ; ptr = ptr->MF_Next) {
	  int l;

          l = ((mode & MENUM_MARKSUB) &&
	       (ptr->MF_Type & MENU_TMASK) == MENU_SUBMENU) ?
	    strlen(ptr->MF_Text) + 2 : strlen(ptr->MF_Text) + 1;
	  width += l;
	}
      }

      height2 = height;
      width2 = width;
      if(mode & MENUM_FRAME) { /*  Space for frame + shadow */
#ifdef CONFIG_MENU_SHADOWS
	height += 3; width += 4;
#else
	height += 2; width += 2;
#endif
      }

      /*  We check that the new menu will fit entirely on the
       *   screen, otherwise shift the line and column pair
       */
      if((line + height) > LINES) line = MAX(LINES - height, 0);
      if((column + width) > COLS) column = MAX(COLS - width, 0);

      if(mode & MENUM_NEWWIN) {
	if((win = jedwstacknew(height, width, line, column)) == NULL) goto ende;
#ifdef MENU_COLOR
	if(!(JEditCursesMode & CURSMODE_NOCOLORS) &&
	   wbkgd(win, (mode & MENUM_FRAME) ? (MENU_COLOR | ' ') :
		 (MENU_BKGRND | ' ')) == ERR)
	  goto ende;
#endif
	intrflush(win, FALSE);
	keypad(win, TRUE);
      } else
	win = stdscr;
      pos = 0;

      if(mode & MENUM_FRAME)
#ifdef CONFIG_MENU_SHADOWS
	_jedprintframe(win, 0, 0, height, width, 3);
#else
	_jedprintframe(win, 0, 0, height, width, 1);
#endif

      flag |= FL_DRAWMENU;
    }

    /* ***  Draw a menu  *** */
    if(flag & FL_DRAWMENU) {
      int x, y;

      if(mode & MENUM_NEWWIN) {
	x = y = (mode & MENUM_FRAME) ? 1 : 0;
      } else {
	x = column; y = line;
      }

#ifdef MENU_COLOR
      if(!(JEditCursesMode & CURSMODE_NOCOLORS))
	wattron(win, (mode & MENUM_FRAME) ? MENU_COLOR : MENU_BKGRND);
#endif

      for(ptr = menus, i = 0 ; ptr ; ptr = ptr->MF_Next, ++i) {
	int l = strlen(ptr->MF_Text);
	char cc;

	if(i == pos && !(mode & MENUM_PRINTONLY)) {
	  wattron(win, A_REVERSE);
	  currx = x;
	  if(mode & MENUM_VERTICAL) currx += width2 / 2;
	  if(mode & MENUM_FRAME) currx++;
	  curry = y + 1;
	  if(mode & MENUM_NEWWIN) {
	    currx += column;
	    curry += line;
	  }
	}

	wmove(win, y, x);
	if(mode & MENUM_VERTICAL) {
	  char *s;
	  int ll;

	  for(s = ptr->MF_Text, ll = l ; *s ; s++)
	    if(*s == '\t') {
	      do
		waddch(win, ' ');
	      while(ll++ < width2);
	    } else
	      waddch(win, *s);
	  while(ll++ < width2) waddch(win, ' ');
	} else
	  wprintw(win, "%s", ptr->MF_Text);

	if((mode & MENUM_MARKSUB) &&
	   (ptr->MF_Type & MENU_TMASK) == MENU_SUBMENU)
	  cc = '>';
	else if((ptr->MF_Type & MENU_TMASK) == MENU_VARIABLE)
	  cc = (*(ptr->MF_Variable) &
		(1UL << (ptr->MF_Type & MENU_NMASK))) ?
	    CHECK_MARK_ON : CHECK_MARK_OFF;
	else
	  cc = 0;

	if(cc) {
	  mvwaddch(win, y, x + ((mode & MENUM_VERTICAL) ? (width2 - 1): l), cc);
	  l++;
	}

	if(i == pos && !(mode & MENUM_PRINTONLY)) {
	  wattroff(win, A_REVERSE);
	  if(JEditHelpTextY) {
	    if(ptr->MF_Help) {
	      int ind;

	      ind = (COLS - strlen(ptr->MF_Help)) / 2;

	      mvprintw(JEditHelpTextY > 0 ? JEditHelpTextY : LINES + JEditHelpTextY,
		       0, "%*s[%s]", ind, "", ptr->MF_Help);
	    } else
	      move(JEditHelpTextY > 0 ? JEditHelpTextY : LINES + JEditHelpTextY, 0);
	    clrtoeol();
	    jedwstacktouch(stdscr);
	  }
	}

	if(mode & MENUM_VERTICAL) y++;
	else x += l + 1;
      }
#ifdef MENU_COLOR
      if(!(JEditCursesMode & CURSMODE_NOCOLORS))
	wattroff(win, (mode & MENUM_FRAME) ? MENU_COLOR : MENU_BKGRND);
#endif

      flag |= FL_REFRESH;

      if(mode & MENUM_PRINTONLY) {
	ret = 0;
	goto ende;
      }
    }

    /* ***  Continue loop and go down on next loop  *** */
    if(flag & FL_WRAPANDDOWN) {
      flag = FL_GODOWN;

      /*  We have to restore the underlying windows first  */
      wnoutrefresh_s(win);
      jedwstackdoupdate(0);
      continue;
    }

    /*  If routine was started with key != -1, don't do
     *   refresh in first loop, since appearance might
     *   change due to key
     */
    if(key != -1) {
      key = -1;
    } else if(flag & FL_REFRESH) {
      if(jedwstackrefresh(win) == ERR) goto ende;
    }

    if(!(flag & FL_KEEPINCHAR)) {
      /*  Get next character  */
      c = jedwstackgetch(win);

#ifdef NCURSES_MOUSE_VERSION
      if(c == KEY_MOUSE && getmouse(&mev) == ERR) {
	ret = -1;
	goto ende;
      }
#endif
    }

    /*  find current entry  */
    for(i = 0, ptr = menus ; i < pos ; ++i) ptr = ptr->MF_Next;

    flag = 0;
    switch(c) {
    case KEY_LEFT:
      if(mode & MENUM_VERTICAL) {
	if(depth == 0) break;
	if(!(stack[depth - 1].S_Mode & MENUM_VERTICAL)) {
	  if(--(stack[depth - 1].S_Pos) < 0) {
#ifdef CONFIG_EDLIB_WRAPMENU
	    stack[depth - 1].S_Pos = stack[depth - 1].S_N - 1;
#else
	    stack[depth - 1].S_Pos = 0;
	    beep();  flag |= FL_REFRESH;
#endif
	  }

	  flag = FL_GOUP | FL_WRAPANDDOWN;
	} else
	  flag = FL_GOUP | FL_KEEPINCHAR;
	break;
      }
    case KEY_UP:
      if(c == KEY_UP && !(mode & MENUM_VERTICAL)) {
	flag = FL_GOUP;
	break;
      }

      if(--pos < 0) {
#ifdef CONFIG_EDLIB_WRAPMENU
	pos = n - 1;
#else
	pos = 0;
	beep();  flag |= FL_REFRESH;
#endif
      }
      flag = FL_DRAWMENU;
      break;
    case KEY_RIGHT:
      if(mode & MENUM_VERTICAL) {
	if(depth == 0) break;
	if(!(stack[depth - 1].S_Mode & MENUM_VERTICAL)) {
	  if(++(stack[depth - 1].S_Pos) == stack[depth - 1].S_N) {
#ifdef CONFIG_EDLIB_WRAPMENU
	    stack[depth - 1].S_Pos = 0;
#else
	    stack[depth - 1].S_Pos = stack[depth - 1].S_N - 1;
	    beep();  flag |= FL_REFRESH;
#endif
	  }

	  flag = FL_GOUP | FL_WRAPANDDOWN;
	} else
	  flag = FL_GOUP | FL_KEEPINCHAR;
	break;
      }
    case KEY_DOWN:
      if(c == KEY_DOWN && !(mode & MENUM_VERTICAL)) {
	flag = FL_GODOWN;
	break;
      }
      if(++pos >= n) {
#ifdef CONFIG_EDLIB_WRAPMENU
	pos = 0;
#else
	pos = n - 1;
	beep();  flag |= FL_REFRESH;
#endif
      }
      flag = FL_DRAWMENU;
      break;
#ifdef KEY_ENTER
    case KEY_ENTER:
#endif
    case '\r': case '\n':
      if((ptr->MF_Type & MENU_TMASK) == MENU_SUBMENU) {
	flag |= FL_GODOWN;
      } else if((ptr->MF_Type & MENU_TMASK) == MENU_VARIABLE) {
	*(ptr->MF_Variable) ^= 1UL << (ptr->MF_Type & MENU_NMASK);
	ret = -2;
      } else {
	ret = ptr->MF_Value;
      }
      break;
    case ESC: case KEY_F(3): case CNTRL('C'):
      ret = -2;
      break;
    case KEY_BACKSPACE: case '\010':
      flag |= FL_GOUP;
      break;
#ifdef NCURSES_MOUSE_VERSION
    case KEY_MOUSE:
      {
	int ml, mc;

	if(!(mev.bstate & (BUTTON1_PRESSED|BUTTON1_RELEASED|BUTTON1_CLICKED)))
	  break;

	ml = mev.y; mc = mev.x;
	ml -= (mode & MENUM_FRAME) ? (line + 1) : line;
	mc -= (mode & MENUM_FRAME) ? (column + 1) : column;

	/*  Mouse points to any menu entry?  */
	if(0 <= ml && ml < height2 && 0 <= mc && mc < width2) {
	  if(mode & MENUM_VERTICAL) {
	    pos = ml;
	  } else {
	    for(pos = 0, ptr = menus ; ptr ; ++pos, ptr = ptr->MF_Next) {
	      int l = strlen(ptr->MF_Text) + 1;
	      if(mc < l) break;
	      mc -= l;
	    }
	    if(ptr == NULL) {
	      ret = -2;
	      break;
	    }
	  }

	  if((ptr->MF_Type & MENU_TMASK) == MENU_SUBMENU) {
	    flag |= FL_DRAWMENU | FL_WRAPANDDOWN;
	  } else {
	    ret = ptr->MF_Value;
	  }
	} else if(depth > 0) {
	  flag |= FL_KEEPINCHAR | FL_GOUP;
	} else {
	  /*  If mouse clicked outside menu, break  */
	  ret = -2;
	}
      }
      break;
#endif
    case ' ':
      if((ptr->MF_Type & MENU_TMASK) == MENU_VARIABLE) {
	*(ptr->MF_Variable) ^= 1UL << (ptr->MF_Type & MENU_NMASK);
	flag = FL_DRAWMENU;
	break;
      }
    default:
      beep();
      flag |= FL_REFRESH;
      break;
    }
  }

ende:
#ifdef CONFIG_HAS_CURSSET
  curs_set(oldVisi);
#endif

  for(;;) {
    if(win != stdscr && jedwstackdel(win) == ERR) {
      ret = -1;
      break;
    }
    if(--depth < 0) break;
    win = stack[depth].S_Window;
  }

  if(JEditHelpTextY) {
    move(JEditHelpTextY > 0 ? JEditHelpTextY : LINES + JEditHelpTextY, 0);
    clrtoeol();
  }

  /*  Restore old screen contents  */
  if(jedwstackdoupdate(1) == ERR) ret = -1;
  if(stack) free(stack);

  return(ret);
}

/* ERROR-DEFINITIONS from addmenu label _ERR_ADDMENU
   Cannot add to non-sub menu
*/

/*JS*********************************************************************
*
*************************************************************************/

jedmenufld_t *jedmenufldadd(jedmenufld_t **pstart, const char **names, int nnames,
		   long type)

/************************************************************************/
{
  jedmenufld_t *newentry, *start;

  if((newentry = (jedmenufld_t *)malloc(sizeof(*newentry))) == NULL)
    EGOTO(error);

  newentry->MF_Type = type;
  newentry->MF_Next = NULL;
  newentry->MF_Help = NULL;

  if((newentry->MF_Text = strdup(names[nnames - 1])) == NULL) EGOTO(error);
  newentry->MF_SubMenu = NULL;

  if(nnames > 1) {
    if((start = jedmenufldfind(*pstart, names, nnames - 1)) == NULL) EGOTO(error);

    if((start->MF_Type & MENU_TMASK) != MENU_SUBMENU) {
      JSSETERROR2(_ERR_ADDMENU, 0, start->MF_Text);
      EGOTO(error);
    }

    if(start->MF_SubMenu) {
      /*  Add at the end of the list  */
      start = start->MF_SubMenu;
      while(start->MF_Next) start = start->MF_Next;
      start->MF_Next = newentry;
    } else
      start->MF_SubMenu = newentry;
  } else if((start = *pstart)) {
    /*  Add at the end of the list  */
    while(start->MF_Next) start = start->MF_Next;
    start->MF_Next = newentry;
  } else
    *pstart = newentry;

  return newentry;
error:
  return NULL;
}

/* ERROR-DEFINITIONS from findmenu label _ERR_FINDMENU
   Non existant parent menu
   Cannot proceed into non-sub menu
*/

/*JS*********************************************************************
*
*************************************************************************/

jedmenufld_t *jedmenufldfind(jedmenufld_t *start, const char **names, int nnames)

/************************************************************************/
{
  for(;;) {
    for( ; start ; start = start->MF_Next)
      if(strcmp(start->MF_Text, *names) == 0) break;

    if(start == NULL) {
      JSSETERROR(_ERR_FINDMENU, 0);
      return(NULL);
    }

    if(--nnames == 0) break;

    names++;

    if((start->MF_Type & MENU_TMASK) != MENU_SUBMENU) {
      JSSETERROR2(_ERR_FINDMENU, 1, *names);
      return(NULL);
    }

    start = start->MF_SubMenu;
  }

  return(start);
}
