/*JS*********************************************************************
*
*    Program : JEDSUBS
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Auxiliary routines for edit package.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <string.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
#define CONFIG_EXPLDELAY   0.01

#ifndef FOFI_MFRAME
/*  START-DEFINITIONS 1 */
/*  For the mode variable  */
#define FOFI_MINITCURSES     (1<<0)
#define FOFI_MFRAME          (1<<1)
#define FOFI_MOWNWINDOW      (1<<2)
#define FOFI_NUMTABLE        (1<<3)  /* for edTable: display ordinals  */
#define FOFI_QUICKLOCATE     (1<<4)
#define FOFI_SCROLLBAR       (1<<5)

#ifdef JEDLIB_INTERNAL
#define _EDLIB_FCOLIND    0
#define _EDLIB_TCOLIND    4
#define _EDLIB_MCOLIND    7
#endif

/*  Macro to get number of color pair for a foreground/background combination */
#define GET_COLORPAIR(fg, bg)  (((fg) % 8) + 8 * ((bg) % 8) + 1)
#define GET_COLOR(fg, bg)      (((fg) & 8) ? \
			       A_BOLD | COLOR_PAIR(GET_COLORPAIR(fg, bg)) : \
			       COLOR_PAIR(GET_COLORPAIR(fg, bg)))

#define GET_COLORBG(color)    ((((color) - 1) / 8) % 8)
#define GET_COLORFG(color)    (((color) - 1) % 8)

#define CURSMODE_NOEXPLODE    (1 << 0)
#define CURSMODE_NOCOLORS     (1 << 1)
#define CURSMODE_NOMOUSE      (1 << 2)
#define CURSMODE_NOACSGRAPH   (1 << 3)

typedef struct {
  short KT_Old, KT_New;
} _jedkeytrans_t;

#ifdef JEDLIB_INTERNAL
/*  Standard editing keys as used by the routines  */
#define CNTRL(c)         ((c) & 0x1f)
#define EDKEY_REDRAW     CNTRL('L')

#define EDKEY_LEFT       KEY_LEFT
#define EDKEY_RIGHT      KEY_RIGHT
#define EDKEY_UP         KEY_UP
#define EDKEY_DOWN       KEY_DOWN
#define EDKEY_PUP        KEY_PPAGE
#define EDKEY_PDOWN      KEY_NPAGE

#define EDKEY_LSTART     KEY_HOME
#define EDKEY_LEND       KEY_END
#define EDKEY_START      CNTRL('A')
#define EDKEY_END        CNTRL('E')

#define EDKEY_DELETE     KEY_DC
#define EDKEY_BS         KEY_BACKSPACE

#define EDKEY_KILLEOL    CNTRL('K')
#define EDKEY_SETMARK    CNTRL(' ')
#define EDKEY_KILLREGION CNTRL('W')
#define EDKEY_YANK       CNTRL('Y')
#define EDKEY_ROTYANK    CNTRL('X')
#define EDKEY_TOGGLEINS  KEY_IC

#define EDKEY_PREV       KEY_F(7)
#define EDKEY_NEXT       KEY_F(8)

#define EDKEY_ENTER      KEY_F(1)

#define EDKEY_QUIT       '\033'

/*  Specifying line = col = height = width = 0
 *   gives the whole screen. line == -1 means
 *   centered vertically, col == -1 centered
 *   horizontally.
 */
#define _checkcoordinates(line, col, height, width) \
	do { \
	  if((line) == -1) (line) = (LINES - (height)) / 2; \
	  else if((line) < 0) (line) += LINES; \
	  if((height) <= 0) (height) += LINES - (line); \
          \
	  if((col) == -1) (col) = (COLS - (width)) / 2; \
	  else if((col) < 0) (col) += COLS; \
	  if((width) <= 0) (width) += COLS - (col); \
	} while(0)

#define _wcheckcoordinates(win, line, col, height, width) \
	do { \
	  int _h, _w; \
          \
	  getmaxyx((win), _h, _w); \
          \
	  if((line) == -1) (line) = (_h - (height)) / 2; \
	  else if((line) < 0) (line) += _h; \
	  if((height) <= 0) (height) += _h - (line); \
          \
	  if((col) == -1) (col) = (_w - (width)) / 2; \
	  else if((col) < 0) (col) += _w; \
	  if((width) <= 0) (width) += _w - (col); \
	} while(0)
#endif
/*  END-DEFINITIONS  */
#endif

/*  Character displayed in unused part of a field  */
#define FIELD_FILLUP_CHAR      '_'

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            jedinitialize(int mode);
Prototype int            jeddeinitialize(void);

IPrototype void          _jedprintframe(WINDOW *win, int offy, int offx,
				     int height, int width, int mode);
IPrototype int           _jedprinttitle(WINDOW *win, int left, int width,
				     const char *title);

IPrototype int           _jedflashwin(WINDOW *win, int mode);

Prototype int            jeddokeytrans(int key, const _jedkeytrans_t *table);

IPrototype int           _jedsetscrollbar(long pos, long n, int height, int *pstart,
				       int *pend);

IPrototype int           _jedinskillring(const char *s, long len);
IPrototype const char   *_jedgetyank(void);
IPrototype void          _jedrotkillring(int d);
static void              mirrorarray(char **a, int n);

Prototype void           jedprogbarshow(WINDOW *win, int line, int column,
					int height, int width, double perc,
					int mode);
Prototype void           jedprogbarclear(WINDOW *win, int line, int column,
					 int height, int width);
Prototype int            jedpictoverlay(WINDOW *win, int voff, int hoff, const char **alines,
					int *attrs);

IPrototype const _jedkeytrans_t JEdLibStdKeyTab[];
Prototype int            JEditCursesMode;
Prototype long           JEditColors[];
Prototype int            JEditHelpTextY;
Prototype int            JEditTFFillChars;

/*********     GLOBAL/STATIC VARIABLES                          *********/
int JEditCursesMode;
long JEditColors[] = {
#ifdef CONFIG_HAS_COLORS
  /* **  Colors for forms  ** */
  GET_COLOR(2, 4), /*  background  */
  GET_COLOR(7, 4), /*  form fields  */
  GET_COLOR(7, 3), /*  enumeration fields  */
  GET_COLOR(5, 6), /*  for help fields  */

  /* **  Colors for tables  ** */
  GET_COLOR(0, 6), /*  background  */
  GET_COLOR(0, 6), /*  table  */
  GET_COLOR(2, 0), /*  scroll bar  */

  /* **  Colors for menus  ** */
  GET_COLOR(5, 0), /*  background  */
  GET_COLOR(4, 5), /*  menus  */
#endif
  0
};

int JEditHelpTextY = -1;

const _jedkeytrans_t JEdLibStdKeyTab[] = {
  { CNTRL('D'), EDKEY_DELETE },
  { 0x7f,       EDKEY_BS     },

  { CNTRL('B'), EDKEY_PUP    },
  { CNTRL('F'), EDKEY_PDOWN  },

  { CNTRL('O'), EDKEY_TOGGLEINS },

#ifdef KEY_BTAB
  { KEY_BTAB,   EDKEY_PREV   },
#endif
  { '\n',       EDKEY_ENTER  },
  { '\t',       EDKEY_NEXT   },

  { CNTRL('C'), EDKEY_QUIT   },
  /*{ CNTRL(''), EDKEY_ },*/
  { -1, -1 }
};

int JEditTFFillChars = FIELD_FILLUP_CHAR;

static char **KillRing;
static int NKillRing, MKillRing;

/*JS*********************************************************************
*   Initialize the curses routines.
*    If Bit 0 in mode is set, use raw mode instead of cbreak mode.
*    If Bit 1 is set, do not use colors if available.
*    If Bit 2 is set, do not use mouse if available.
*    If Bit 3 is set, do not set the background color.
*    If bit 4 is set, do not use special line characters.
*    If bit 5 is set, do not use any line characters.
*************************************************************************/

int jedinitialize(int mode)

/************************************************************************/
{
#ifdef JEDLIB_HASHISTORY
  if(_jedhistload() == -1)
    return -1;
#endif

  /*  Initialize and get character at a time  */
  initscr();
  _initmyacs((mode & 32) ? 3 : (mode & 16) ? 1 : 0);

  if(mode & (16|32))
    JEditCursesMode |= CURSMODE_NOACSGRAPH;

#ifdef CONFIG_HAS_COLORS
  if(!(JEditCursesMode & CURSMODE_NOCOLORS) && !(mode & 2)) {
    start_color();

    if(has_colors()) {
      int i;

      for(i = 1 ; i <= COLOR_PAIRS ; i++) {
	short f, b;

	f = (i - 1) % COLORS;
	b = ((i - 1) / COLORS) % 8;
	init_pair(i, f, b);
      }
    } else
      JEditCursesMode |= CURSMODE_NOCOLORS;

  } else
      JEditCursesMode |= CURSMODE_NOCOLORS;
#endif

  if(mode & 1)
    raw();
  else
    cbreak();

  nonl(); noecho();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);
#if defined(__sgi) && !defined(NCURSES_VERSION)
  /*  work around a problem with keypad under SGI  */
  notimeout(stdscr, TRUE);
#endif

#ifdef CONFIG_HAS_CURSSET
  curs_set(1);
#endif

  if(!(JEditCursesMode & CURSMODE_NOMOUSE)) {
#ifdef NCURSES_MOUSE_VERSION
    /*  Enable mouse commands  */
    (void)mousemask(BUTTON1_PRESSED | BUTTON1_RELEASED | BUTTON1_CLICKED |
		    BUTTON1_DOUBLE_CLICKED
		    /* | REPORT_MOUSE_POSITION */,
		    NULL);
#elif defined(__sgi)
    (void)mouse_set(ALL_MOUSE_EVENTS);
#endif
  }

  if(!(mode & 8) && bkgd(COLOR_PAIR(8) | ' ') == ERR)
    return -1;

  if(_jedwstackpush(stdscr, -1) == -1)
    return -1;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

int jeddeinitialize(void)

/************************************************************************/
{
  int ret = 0;

  if(endwin() == ERR)
    ret = -1;

#ifdef JEDLIB_HASHISTORY
  if(_jedhistdump() == -1)
    ret = -1;
#endif

  return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jedprinttitle(WINDOW *win, int left, int width, const char *title)

/************************************************************************/
{
  int n2, len;
  chtype c;

  if((len = strlen(title)) > width) len = width;
  n2 = width - len; /*  remaining for '-'  */

  width = (n2 - 1) / 2; /*  # of left '-'  */
  n2 -= width; /*  # of right '-'  */

  wmove(win, 0, left);

  c = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '-' : ACS_HLINE;
  while(width-- > 0)
    waddch(win, c);
  waddnstr(win, title, len);
  while(n2-- > 0)
    waddch(win, c);

  return 0;
}

/*JS*********************************************************************
*   Print a frame at position offy,offx of the given height and width in
*    the window.
*    Parameter mode:
*    bit 0: make a 3D effect.
*    bit 1: print a shadow.
*    bit 2: the 3D effect is reversed.
*    bit 3: the headline gets a double line instead.
*    bit 4: only the shadow is redrawn.
*    bit 5: double lines are used for the whole frame.
*    bit 6: make the frame bold.
*    bit 8: the frame is cleared.
*************************************************************************/

void _jedprintframe(WINDOW *win, int offy, int offx, int height, int width,
		    int mode)

/************************************************************************/
{
  chtype clb, clw, c;
  int i, endx, endy;

  _wcheckcoordinates(win, offy, offx, height, width);

  if((mode & 3) == 3) {
    /*  with shadow  */
    endx = offx + width - 3;
    endy = offy + height - 2;
  } else {
    /*  without shadow  */
    endx = offx + width - 1;
    endy = offy + height - 1;
  }

  if(!(mode & 16)) {
    int c2;

#ifdef CONFIG_USE_SHADOWS
    /*  Make 3D effect by making white and black line  */
    if(!(JEditCursesMode & CURSMODE_NOCOLORS) && (mode & 1) && !(mode & 256)) {
      short bg, fg;

      pair_content(PAIR_NUMBER(getbkgd(win)), &fg, &bg);
      if(mode & 4) {
	clw = GET_COLOR(7, bg);
	clb = GET_COLOR(0, bg);
      } else {
	clb = GET_COLOR(7, bg);
	clw = GET_COLOR(0, bg);
      }
    } else
#endif
      clb = clw = 0;

    if(mode & 64) {
      clb |= A_BOLD;
      clw |= A_BOLD;
    }

    /*  Print the frame  */
    c = (mode & 256) ? ' ' :
      (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '+' :
      (mode & 8) ? ACS_ULCORNER21 :
      (mode & 32) ? ACS_ULCORNER22 :
      ACS_ULCORNER;
    lgmvwaddch(win, offy, offx, clb | c);

    c = (mode & 256) ? ' ' :
      (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '+' :
      (mode & 8) ? ACS_URCORNER21 :
      (mode & 32) ? ACS_URCORNER22 :
      ACS_URCORNER;
    lgmvwaddch(win, offy, endx, clw | c);

    c = (mode & 256) ? ' ' :
      (JEditCursesMode & CURSMODE_NOACSGRAPH) ? (clb | '-') :
      (mode & 40) ? (clb | ACS_HLINE22) :
      (clb | ACS_HLINE);
    c2 = (mode & 256) ? ' ' :
      (JEditCursesMode & CURSMODE_NOACSGRAPH) ? (clw | '-') :
      (mode & 32) ? (clw | ACS_HLINE22) :
      (clw | ACS_HLINE);
    for(i = offx + 1 ; i < endx ; i++) {
      lgmvwaddch(win, offy, i, c);
      lgmvwaddch(win, endy, i, c2);
    }

    c = (mode & 256) ? ' ' :
      (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '+' :
      (mode & 32) ? ACS_LLCORNER22 :
      ACS_LLCORNER;
    lgmvwaddch(win, endy, offx, clb | c);

    c = (mode & 256) ? ' ' :
      (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '+' :
      (mode & 32) ? ACS_LRCORNER22 :
      ACS_LRCORNER;
    lgmvwaddch(win, endy, endx, clw | c);

    c = (mode & 256) ? ' ' :
      (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '|' :
      (mode & 32) ? ACS_VLINE22 :
      ACS_VLINE;
    for(i = offy + 1 ; i < endy ; i++) {
      lgmvwaddch(win, i, offx, clb | c);
      lgmvwaddch(win, i, endx, clw | c);
    }
  }

  /*  Print shadow  */
#if defined(CONFIG_USE_SHADOWS) && defined(CONFIG_HAS_COLORS)
  if(!(JEditCursesMode & CURSMODE_NOCOLORS) && (mode & 3) == 3) {
    int begx, begy;

    getbegyx(win, begy, begx);
    clb = COLOR_PAIR(GET_COLORPAIR(4, 0));

    for(i = offx ; i <= (endx + 2) ; i++) {
      c = (mode & 256) ? ' ' : mvwinch(newscr, begy + endy + 1, begx + i);
      mvwaddch(win, endy + 1, i, i > offx ? (clb | (c & ~A_COLOR)) : c);
    }
    for(i = offy ; i <= endy ; i++) {
      c = (mode & 256) ? ' ' : mvwinch(newscr, begy + i, begx + endx + 1);
      mvwaddch(win, i, endx + 1, i > offy ? (clb | (c & ~A_COLOR)) : c);
      c = (mode & 256) ? ' ' : mvwinch(newscr, begy + i, begx + endx + 2);
      mvwaddch(win, i, endx + 2, i > offy ? (clb | (c & ~A_COLOR)) : c);
    }
  }
#endif
}

/*JS*********************************************************************
*   Make a window explode.
*************************************************************************/

int _jedflashwin(WINDOW *win, int mode)

/************************************************************************/
{
#ifdef NCURSES_VERSION
  if(!(JEditCursesMode & CURSMODE_NOEXPLODE)) {
    int begx, begy, lenx, leny, startx, starty, endx, endy, dx, dy, d, x, y;
    WINDOW *clone, *src, *dest;
    chtype c;

    /*  Clone window and set up  */
    getbegyx(win, begy, begx);
    getmaxyx(win, leny, lenx);
    if((clone = newwin(leny, lenx, begy, begx)) == NULL)
      goto error;

    if(mode & 1) {
      src = curscr;

    } else {
      jedwstacktouch(stdscr);

      /*  Disable this window temporarily, so we can see what is behind  */
      if(_jedwstackflag(win, 4) == -1)
	goto error;
      jedwstackdoupdate(0);

      /*  To make sure the next refresh will still show the old content  */
      jedwstacktouch(win);
      src = newscr;
    }

    for(x = 0 ; x < lenx ; x++)
      for(y = 0 ; y < leny ; y++) {
	c = mvwinch(src, begy + y, begx + x);
	mvwaddch(clone, y, x, c);
      }

    dx = lenx / 2;
    dy = leny / 2;
    if(mode & 1) {
      startx = dx + 1;
      starty = dy + 1;

    } else {
      startx = starty = -1;
    }
    endx  = lenx - startx - 1;
    endy  = leny - starty - 1;

    if(mode & 1) {
      src  = win;
      dest = clone;

    } else {
      dest = win;
      src  = clone;
    }

    d = dx > dy ? 0 : -1;
    while((mode & 1) ? (startx > 0 || starty > 0) :
	  (startx < endx || starty < endy)) {
      jedwstackrefresh(dest);
#ifdef CONFIG_EXPLDELAY
      if((startx + starty) & 1)
	dsleep(CONFIG_EXPLDELAY);
#endif

      if(d >= 0) {
	if(mode & 1) { startx--; endx++; } else { startx++; endx--; }
	d -= dy;

	for(y = starty ; y <= endy ; y++) {
	  c = mvwinch(src, y, startx);
	  mvwaddch(dest, y, startx, c);
	  c = mvwinch(src, y, endx);
	  mvwaddch(dest, y, endx, c);
	}

      } else {
	if(mode & 1) { starty--; endy++; } else { starty++; endy--; }
	d += dx;

	for(x = startx ; x <= endx ; x++) {
	  c = mvwinch(src, starty, x);
	  mvwaddch(dest, starty, x, c);
	  c = mvwinch(src, endy, x);
	  mvwaddch(dest, endy, x, c);
	}
      }
    }

    delwin(clone);
  }

  if((mode & 1) && ((mode & 2) ? wnoutrefresh(win) : wrefresh(win)) == ERR)
    goto error;

  return 0;
error:
  return -1;
#else
  return 0;
#endif
}

/*JS*********************************************************************
*
*************************************************************************/

int jeddokeytrans(int key, const _jedkeytrans_t *table)

/************************************************************************/
{
  if(table == NULL)
    table = JEdLibStdKeyTab;

  for( ; table->KT_Old != -1 ; table++)
    if(key == table->KT_Old)
      return table->KT_New;

  return key;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jedsetscrollbar(long pos, long n, int height, int *pstart, int *pend)

/************************************************************************/
{
  int start, end;

  /*  Prevent division by zero  */
  if(n == 0)
    n = 1;

  if((end = (height * height + n / 2) / n) == 0)
    end = 1;

  if((start = (pos * height + n / 2) / n) > (height - end))
    start = height - end;

  end += start;

  *pstart = start;
  *pend   = end;

  return start > 0 || end < height;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jedinskillring(const char *s, long len)

/************************************************************************/
{
  char *dest;

  /*  Increase kill ring  */
  if(NKillRing >= MKillRing) {
    MKillRing += 10;
    if((KillRing = (char **)
	realloc(KillRing, MKillRing * sizeof(*KillRing))) == NULL)
      return -1;
  }

  if((dest = (char *)malloc(len + 1)) == NULL)
    return -1;
  memcpy(dest, s, len);
  dest[len] = '\0';

  KillRing[NKillRing++] = dest;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

const char *_jedgetyank(void)

/************************************************************************/
{
  if(NKillRing == 0)
    return NULL;

  return KillRing[NKillRing - 1];
}

/*JS*********************************************************************
*
*************************************************************************/

void _jedrotkillring(int d)

/************************************************************************/
{
  if(NKillRing > 1) {
    if((d %= NKillRing) < 0)
      d += NKillRing;

    mirrorarray(KillRing, d);
    mirrorarray(KillRing + d, NKillRing - d);
    mirrorarray(KillRing, NKillRing);
  }
}

/*JS*********************************************************************
*
*************************************************************************/

static void mirrorarray(char **a, int n)

/************************************************************************/
{
  char **e, *tmp;

  for(e = &a[n - 1] ; a < e ; a++, e--) {
    tmp = *a; *a = *e; *e = tmp;
  }
}

/*JS*********************************************************************
*   mode:
*    Bit 0-7: Color code to use.
*    Bit   8: Show frame
*    Bit   9: Show numerical representation of percentage.
*    Bit  10: Use color code from bit 0-7.
*    Bit  11: Vertical bar.
*************************************************************************/

void jedprogbarshow(WINDOW *win, int line, int column, int height, int width,
		    double perc, int mode)

/************************************************************************/
{
  char buff[5];
  chtype bcol, cc;
  int l, c, limit, left, right, cline;

  bcol = (mode & 1024) ? COLOR_PAIR(mode & 255) : 0;

  /*  Show frame?  */
  if(mode & 256) {
    cc = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? bcol | '+' :
      bcol | ACS_ULCORNER;
    mvwaddch(win, line, column, cc);

    cc = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? bcol | '-' :
      bcol | ACS_HLINE;
    for(c = 2 ; c < width ; ++c)
      waddch(win, cc);

    cc = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? bcol | '+' :
      bcol | ACS_URCORNER;
    waddch(win, cc);

    cc = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? bcol | '|' :
      bcol | ACS_VLINE;
    for(l = 1 ; l < height ; ++l) {
      mvwaddch(win, line + l, column, cc);
      mvwaddch(win, line + l, column + width - 1, cc);
    }

    cc = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? bcol | '+' :
      bcol | ACS_LLCORNER;
    mvwaddch(win, line + height - 1, column, cc);

    cc = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? bcol | '-' :
      bcol | ACS_HLINE;
    for(c = 2 ; c < width ; ++c)
      waddch(win, cc);

    cc = (JEditCursesMode & CURSMODE_NOACSGRAPH) ? bcol | '+' :
      bcol | ACS_LRCORNER;
    waddch(win, cc);

    ++line;
    ++column;
    height -= 2;
    width -= 2;
  }

  limit = (mode & 2048) ? perc * height : perc * width;

  sprintf(buff, "%3d%%", (int)(100 * perc + 0.5));

  /*  Show number?  */
  if((mode & 512) && width >= 4) {
    cline = (height + 1) / 2 - 1;
    left  = (width / 2 - 2);
    right = (width / 2 + 2);

  } else
    cline = -1;

  line += height - 1;
  for(l = 0 ; l < height ; ++l) {
    wmove(win, line - l, column);
    for(c = 0 ; c < width ; ++c) {
      cc = l == cline && left <= c && c < right ? buff[c - left] : ' ';
      if((mode & 2048) ? l < limit : c < limit)
	cc |= /*A_BOLD |*/ A_REVERSE;
      waddch(win, bcol | cc);
    }
  }
}

/*JS*********************************************************************
*
*************************************************************************/

void jedprogbarclear(WINDOW *win, int line, int column, int height, int width)

/************************************************************************/
{
  int l, c;

  for(l = 0 ; l < height ; ++l) {
    wmove(win, line + l, column);
    for(c = 0 ; c < width ; ++c)
      waddch(win, ' ');
  }
}

/*JS*********************************************************************
*
*************************************************************************/

int jedpictoverlay(WINDOW *win, int voff, int hoff, const char **alines,
		   int *attrs)

/************************************************************************/
{
  const char *line;
  int i, j;

  if(voff == -1 || hoff == -1) {
    int w, h;

    w = 0;
    for(h = 0 ; alines[h] ; ++h) {
      int l;
      l = strlen(alines[h]);
      if(l > w) w = l;
    }

    _wcheckcoordinates(win, voff, hoff, h, w);
  }

  for(i = 0 ; (line = alines[i]) != NULL ; ++i) {
    for(j = 0 ; line[j] ; ++j) {
      chtype attr, c;

      if(line[j] == ' ') {
	attr = 0;
      } else {
	attr = A_STANDOUT;
	if(attrs != NULL) {
	  int k;

	  for(k = 0 ; attrs[k] ; k += 2)
	    if(line[j] == attrs[k]) {
	      attr = attrs[k + 1] & (A_ATTRIBUTES | A_COLOR);
	      break;
	    }
	}
      }

      /*  Use the original character, but the attributes from the lines  */
      c = attr | (mvwinch(win, voff + i, hoff + j) & A_CHARTEXT);

      if(mvwaddch(win, voff + i, hoff + j, c) == ERR)
	return -1;
    }
  }

  return 0;
}

#ifdef TEST
/* ***  Test the progress bar  *** */
#include <stdlib.h>
#include <math.h>

static void dothenecessary(double p, int hoff, int voff)
{
  /*  horizontal  */
  jedprogbarshow(stdscr, voff +  1, hoff +  2, 1, 78, 0.01 * p, 0x402);
  jedprogbarshow(stdscr, voff +  3, hoff +  2, 1, 78, 0.01 * p, 0x603);
  jedprogbarshow(stdscr, voff +  5, hoff +  1, 3, 80, 0.01 * p, 0x504);
  jedprogbarshow(stdscr, voff +  8, hoff +  1, 3, 80, 0.01 * p, 0x705);

  /*  vertical  */
  jedprogbarshow(stdscr, voff + 12, hoff +  1, 12, 2, 0.01 * p, 0xc06);
  jedprogbarshow(stdscr, voff + 12, hoff +  5, 12, 4, 0.01 * p, 0xe07);
  jedprogbarshow(stdscr, voff + 11, hoff + 10, 14, 6, 0.01 * p, 0xd08);
  jedprogbarshow(stdscr, voff + 11, hoff + 16, 14, 6, 0.01 * p, 0xf09);

  jedprogbarshow(stdscr, voff + 11, hoff + 23, 5, 58, 0.01 * p, 0x703);

  refresh();
}

int main(int argc, char *argv[])
{
  double p, delay, incr;
  int i, hoff, voff;

  delay = argc > 1 ? atof(argv[1]) : 0.01;
  incr  = argc > 2 ? atof(argv[2]) : 0.2;

  jedinitialize(0);
  curs_set(0);

  mvprintw(2, (COLS - 40) / 2, "Various supported types of progress bars");
  voff = 4;
  hoff = 0;
  for(i = 0 ; i < 4 ; ++i) {
    for(p = 0.0 ; p <= 100.1 ; p += incr) {
      dothenecessary(p, hoff, voff);
      dsleep(delay);
    }
    for(p = 100. ; p >= -0.04 ; p -= incr) {
      dothenecessary(p, hoff, voff);
      dsleep(delay);
    }
  }

  jeddeinitialize();

  return(0);
}
#endif
