/*JS*********************************************************************
*
*    Program : JEDTABLE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Display tables.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <jssubs.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

/*********     DEFINES                                          *********/
#ifndef EDTAB_TOUCHED
/*  START-DEFINITIONS  */
/*  Lower byte reserved for return value  */
#define EDTAB_MASK         ((1UL << 8) - 1)
#define EDTAB_TOUCHED      (1UL << 8)
#define EDTAB_CHTITLE      (1UL << 9)
#define EDTAB_CHLINE       (1UL << 10)
#define EDTAB_CHTABLE      (1UL << 11)

/*  Modes for routine _tabformdokey  */
#define TABFORM_FNOREFRESH     (1<<1)
#define TABFORM_FREREADLINE    (1<<2)
#define TABFORM_FREREADTABLE   (1<<3)
#define TABFORM_FQUICKLOCATE   (1<<4)
#define TABFORM_FHIDECURRENT   (1<<5)

typedef struct {
  int TA_Mode;
  WINDOW *TA_Window;
  int TA_Line, TA_Column, TA_Height, TA_Width;

  int (*TA_GetLine)(void *uptr, long nr, char *dest, int m);
  long TA_Multi, TA_NLines;
  void *TA_UPtr;

  long TA_Start, TA_Current, TA_SeekIndex;
  int TA_ScrollBarS, TA_ScrollBarE;
  char TA_Lines[1];
} jedtable_t;
/*  END-DEFINITIONS  */
#endif

#define CHECK

#ifdef CONFIG_HAS_COLORS
# define TABLE_FRAME        JEditColors[_EDLIB_TCOLIND + 0]
# define TABLE_COLOR        JEditColors[_EDLIB_TCOLIND + 1]
#endif

#define TAB_PREFIX          "%4ld:"

struct GLineWrap {
  jedtable_t *ta;
  int (*getline)(void *uptr, long nr, char *dest, int m);
  void *uptr;
};

/* ***  For tabform routines  *** */
#define LINEMEMLEN(ta)        ((ta)->TA_Width + 1)
#define LINEMEM(ta, nr)        ((ta)->TA_Lines + (nr) * LINEMEMLEN(ta))

#define TABLE_SCROLL_SKIP           1

#define tabformsetscbar(ta)    _jedsetscrollbar((ta)->TA_Start, \
                                            (ta)->TA_NLines, (ta)->TA_Height, \
                                            &((ta)->TA_ScrollBarS), \
                                            &((ta)->TA_ScrollBarE))

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int		 jedtabdisplay(char *title, int line, int column,
				       int height, int width, long multi, long nLines,
				       int (*getLine)(void *uptr, long nr,
						      char *dest, int m),
				       int (*handle)(void *uptr, int ch, long nr,
						     char *title, char *line,
						     long *pnLines),
				       void *uPtr, long *pStart, int mode);
static int               glinewrap(struct GLineWrap *glw, long nr, char *dest,
				   int m);

Prototype jedtable_t    *jedtabcreate(WINDOW *win, int line, int column,
				       int height, int width,
				       int (*getLine)(void *, long, char *, int),
				       void *uptr, long multi, long nLines,
				       long *pInit, int mode);
static int               tabformreadline(jedtable_t *ta, int nr);

Prototype int            jedtabdelete(jedtable_t *ta);
Prototype int            jedtabdokey(jedtable_t *ta, int key, int mode);
static int               tabformseekkey(jedtable_t *ta, int key);

static int               tabformprintline(jedtable_t *ta, int nr);
static int               tabformscroll(jedtable_t *ta, int d);
static int               tabformalign(jedtable_t *ta);

/*JS*********************************************************************
*
*************************************************************************/

int jedtabdisplay(char *title, int line, int column, int height, int width,
		  long multi, long nLines,
		  int (*getLine)(void *uptr, long nr, char *dest, int m),
		  int (*handle)(void *uptr, int ch, long nr,
				char *title, char *line,
				long *pnLines),
		  void *uPtr, long *pStart, int mode)

/************************************************************************/
{
  jedtable_t *ta;
  int ret, key, ovisi;
  struct GLineWrap glw;

  if((mode & FOFI_MINITCURSES) && jedinitialize(0))
    EGOTO(error);

  glw.uptr = uPtr;
  if(mode & FOFI_NUMTABLE) {
    glw.getline = getLine;
    getLine = (int (*)(void *, long, char *, int))glinewrap;
    uPtr = &glw;
  }

  /*  One more column for scrollbar  */
  if((ta = jedtabcreate(NULL, line, column, height, width, getLine, uPtr,
			multi, nLines, pStart, FOFI_SCROLLBAR | mode)) == NULL)
    EGOTO(error);

  if(_jedprinttitle(ta->TA_Window, ta->TA_Column, ta->TA_Width, title) == -1)
    EGOTO(errorf);

  if(mode & FOFI_NUMTABLE) glw.ta = ta;

#ifdef CONFIG_HAS_CURSSET
  /*  Turn cursor off, but preserve old setting  */
  ovisi = curs_set(0);
#endif

  ret = 0;
  for(key = EDKEY_REDRAW ; ; ) {
    int a;

    if((a = jedtabdokey(ta, key, (mode & FOFI_QUICKLOCATE) ?
			TABFORM_FQUICKLOCATE : 0)) == -1) EGOTO(errorf);
    if(a == -3) {
      beep();
      jedwstackdoupdate(1);
    } else if(a == -2) {
      int flag;

      if(handle == NULL) {
	ret = key;
	break;
      }

#ifdef CONFIG_HAS_CURSSET
      curs_set(ovisi);
#endif

      if(nLines > 0)
	a = (*handle)(glw.uptr, key, ta->TA_Current, title,
		      LINEMEM(ta, ta->TA_Current - ta->TA_Start), &nLines);
      else
	a = (*handle)(glw.uptr, key, -1, title, NULL, &nLines);

      if((a & EDTAB_MASK) != 0) {
	ret = (a & EDTAB_MASK);
	break;
      }

      flag = 0;
      if(nLines != ta->TA_NLines) {
	ta->TA_NLines = nLines;
	flag |= TABFORM_FREREADTABLE;
      } else if(a & EDTAB_CHTABLE)
	flag |= TABFORM_FREREADTABLE;
      else if(a & EDTAB_CHLINE)
	flag |= TABFORM_FREREADLINE;

      if(a & (EDTAB_CHTITLE|EDTAB_TOUCHED)) {
	if((a & EDTAB_CHTITLE) &&
	   _jedprinttitle(ta->TA_Window, ta->TA_Column, ta->TA_Width, title) == -1)
	  EGOTO(errorf);
	if(flag == 0) jedwstackrefresh(ta->TA_Window);
      }

      if(flag && jedtabdokey(ta, EDKEY_REDRAW, flag) == -1) EGOTO(errorf);

#ifdef CONFIG_HAS_CURSSET
      ovisi = curs_set(0);
#endif
    }

    key = jedwstackgetch(ta->TA_Window);
  }

  if(pStart) {
    pStart[0] = ta->TA_Start;
    pStart[1] = ta->TA_Current;
  }

  jedtabdelete(ta);

  return ret;
errorf:
  if(ta) jedtabdelete(ta);
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int glinewrap(struct GLineWrap *glw, long nr, char *dest, int m)

/************************************************************************/
{
  int l;

  l = sprintf(dest, TAB_PREFIX, 1 + nr);
  dest += l;
  m -= l;

  if(glw->getline == NULL) {
    strncpy(dest, ((const char **)(glw->uptr))[nr], m);
    return(0);
  }

  return((*(glw->getline))(glw->uptr, nr, dest, m));
}

/*JS*********************************************************************
*
*************************************************************************/

jedtable_t *jedtabcreate(WINDOW *win, int line, int column, int height, int width,
			 int (*getLine)(void *, long, char *, int),
			 void *uptr, long multi, long nLines,
			 long *pInit, int mode)

/************************************************************************/
{
  jedtable_t *ta;
  int i;

  if((ta = (jedtable_t *)malloc(sizeof(*ta) - 1 + (height + 1) * (width + 1)))
     == NULL) return(NULL);

  ta->TA_Mode = mode;

  if(height <= 0) height += LINES;
  if(width <= 0) width += COLS;

  /*  Center table  */
  if(line == -1 && (line = (LINES - height) / 2) < 0) line = 0;
  if(column == -1 && (column = (COLS - width) / 2) < 0) column = 0;

  if(win == NULL) {
    /*  Create window for table  */
    if((win = jedwstacknew(height, width, line, column)) == NULL)
      EGOTO(error);

#ifdef CONFIG_HAS_COLORS
    if(!(JEditCursesMode & CURSMODE_NOCOLORS) &&
       wbkgd(win, TABLE_COLOR | ' ') == ERR)
      EGOTO(errorw);
#endif

    keypad(win, TRUE); intrflush(win, FALSE);
    idlok(win, TRUE); /* ??? */
    werase(win);

    line = 0; column = 0; /*  now coordinates relative to new window  */

    ta->TA_Mode |= FOFI_MOWNWINDOW;
  }

  if(mode & FOFI_MFRAME) {
#ifdef CONFIG_HAS_COLORS
    if(!(JEditCursesMode & CURSMODE_NOCOLORS) &&
       wattron(win, TABLE_FRAME) == ERR) EGOTO(errorw);
#endif
    _jedprintframe(win, 0, 0, height, width, 5);
    line++; column++; height -= 2; width -= 2;
  }

  ta->TA_Window  = win;
  ta->TA_Line    = line;    ta->TA_Column = column;
  ta->TA_Height  = height;  ta->TA_Width  = width;
  ta->TA_GetLine = getLine; ta->TA_UPtr   = uptr;
  if((ta->TA_Multi = multi) == 0) ta->TA_Multi = 1;
  ta->TA_NLines  = nLines;

  /*  Initialize the table form  */
  if(pInit) {
    if((ta->TA_Start = pInit[0]) < 0 && (ta->TA_Start += nLines) < 0)
      ta->TA_Start = 0;

    if((ta->TA_Current = pInit[1]) < 0 && (ta->TA_Current += nLines) < 0)
      ta->TA_Current = 0;

    if(ta->TA_Current > 0 && ta->TA_Current >= nLines)
      ta->TA_Current = MAX(nLines - 1, 0);
  } else
    ta->TA_Start = ta->TA_Current = 0;

  ta->TA_SeekIndex = 0;

  for(i = 0 ; i < height ; i++)
    if(tabformreadline(ta, i) == -1) EGOTO(error);

  if(ta->TA_Mode & FOFI_SCROLLBAR) tabformsetscbar(ta);

  return(ta);
errorw:
  jedwstackdel(win);
error:
  free(ta);
  return(NULL);
}

/*JS*********************************************************************
*
*************************************************************************/

static int tabformreadline(jedtable_t *ta, int nr)

/************************************************************************/
{
  char *s;
  long nr2;

  nr2 = ta->TA_Start + nr;
  s = LINEMEM(ta, nr);
  if(nr2 >= ta->TA_NLines) {
    *s = '\0';
    return(0);
  }

  if(ta->TA_GetLine == NULL) {
    strncpy(s, ((const char **)(ta->TA_UPtr))[nr2], ta->TA_Width);
    return(0);
  }

  return((*(ta->TA_GetLine))(ta->TA_UPtr, nr2, s, ta->TA_Width));
}

/*JS*********************************************************************
*
*************************************************************************/

int jedtabdelete(jedtable_t *ta)

/************************************************************************/
{
  if((ta->TA_Mode & FOFI_MOWNWINDOW) && jedwstackdel(ta->TA_Window) == ERR)
    return -1;

  free(ta);
  return 0;
}

/*  Action flags  */
#define FLAG_REDRAW      (1<<0)
#define FLAG_NEWCURR     (1<<1)
#define FLAG_REFRESH     (1<<2)
#define FLAG_CHECKCURR   (1<<3)
#define FLAG_ILLEGAL     (1<<4)
#define FLAG_UNKNOWN     (1<<5)
#define FLAG_RESETSEEK   (1<<6)

/*JS*********************************************************************
*   Handle the key for the table window.
*************************************************************************/

int jedtabdokey(jedtable_t *ta, int key, int mode)

/************************************************************************/
{
  int flag, d, ocurr;

  flag = FLAG_RESETSEEK;
  ocurr = ta->TA_Current;
  switch(key) {
  case EDKEY_REDRAW:
    flag |= FLAG_REDRAW;
    mode |= TABFORM_FREREADTABLE;
    break;
  case EDKEY_UP:
    if(ta->TA_Current >= ta->TA_Multi) {
      ta->TA_Current -= ta->TA_Multi;
      flag |= tabformalign(ta) ? FLAG_REDRAW : FLAG_NEWCURR;
    } else
      flag |= FLAG_ILLEGAL;
    break;
  case EDKEY_DOWN:
    if(ta->TA_Current < (ta->TA_NLines - ta->TA_Multi)) {
      ta->TA_Current += ta->TA_Multi;
      flag |= tabformalign(ta) ? FLAG_REDRAW : FLAG_NEWCURR;
    } else
      flag |= FLAG_ILLEGAL;
    break;
  case EDKEY_PUP:
    /*  We only scroll in multiples of TA_Multi  */
    d = ta->TA_Height - ta->TA_Height % ta->TA_Multi;
    if((d = tabformscroll(ta, -d)) > 0) {
      ta->TA_Current -= d;
      flag |= FLAG_REDRAW|FLAG_CHECKCURR;
    } else
      flag |= FLAG_ILLEGAL;
    break;
  case EDKEY_PDOWN:
    /*  We only scroll in multiples of TA_Multi  */
    d = ta->TA_Height - ta->TA_Height % ta->TA_Multi;
    if((ta->TA_Start + d) < ta->TA_NLines && (d = tabformscroll(ta, d)) > 0) {
      ta->TA_Current += d;
      flag |= FLAG_REDRAW|FLAG_CHECKCURR;
    } else
      flag |= FLAG_ILLEGAL;
    break;
  case EDKEY_START:
    if((d = tabformscroll(ta, -ta->TA_Start)) > 0) {
      ta->TA_Current -= d;
      flag |= FLAG_CHECKCURR|FLAG_REDRAW;
    } else
      flag |= FLAG_ILLEGAL;
    break;
  case EDKEY_END:
    d = (ta->TA_NLines - ta->TA_Start - 1) / ta->TA_Height;
    d *= ta->TA_Height;
    if((d = tabformscroll(ta, d)) > 0) {
      ta->TA_Current += d;
      flag |= FLAG_CHECKCURR|FLAG_REDRAW;
    } else
      flag |= FLAG_ILLEGAL;
    break;
#ifdef NCURSES_MOUSE_VERSION
  case KEY_MOUSE:
    {
      int ml, mc;
      MEVENT mev;

      if(getmouse(&mev) == ERR) EGOTO(error);

      /*  Ignore press events, we will react on the release one  */
      if(mev.bstate & BUTTON1_PRESSED) break;

      getbegyx(ta->TA_Window, ml, mc);
      ml = mev.y - ta->TA_Line - ml;
      mc = mev.x - ta->TA_Column - mc;

      if(ml < 0 || ml >= ta->TA_Height) {
	flag |= FLAG_UNKNOWN;
	break;
      }

      if((ta->TA_Mode & FOFI_SCROLLBAR) && mc == (ta->TA_Width - 1)) {
	ta->TA_Current = (ml * ta->TA_NLines + ta->TA_Height / 2) /
	  (ta->TA_Height - 1);
	flag |= tabformalign(ta) ? (FLAG_REDRAW|FLAG_CHECKCURR) :
	  (FLAG_NEWCURR|FLAG_CHECKCURR);
      } else if(0 <= mc && mc < (ta->TA_Width - 1)) {
	ta->TA_Current = ta->TA_Start + ml;
	flag |= FLAG_CHECKCURR | FLAG_NEWCURR;
      } else
	flag |= FLAG_UNKNOWN;

      /*  Other buttons bring us to the current line, but return
       *   with "unknown", so the next layer up can decide what
       *   to do.
       */
      if(!(mev.bstate & (BUTTON1_PRESSED|BUTTON1_CLICKED))) {
	flag |= FLAG_UNKNOWN;
	break;
      }
    }
    break;
#endif
  case EDKEY_BS:
    if(!(mode & TABFORM_FQUICKLOCATE)) {
      flag |= FLAG_UNKNOWN;
      break;
    }

    flag &= ~FLAG_RESETSEEK;
    if(tabformseekkey(ta, -1) == -1) EGOTO(error);
    flag |= tabformalign(ta) ? FLAG_REDRAW : FLAG_REDRAW;/* FLAG_NEWCURR;*/
    break;
  default:
    if(key >= 256 || !(mode & TABFORM_FQUICKLOCATE) || !isprint(key)) {
      flag |= FLAG_UNKNOWN;
      break;
    }

    flag &= ~FLAG_RESETSEEK;
    if((d = tabformseekkey(ta, key)) == -1) EGOTO(error);

    if(d == -3) {
      flag |= FLAG_ILLEGAL;
      break;
    }
    flag |= tabformalign(ta) ? FLAG_REDRAW : FLAG_NEWCURR;
    break;
  }

  if(mode & TABFORM_FREREADTABLE) {
    /*  Reread the whole table  */
    for(d = 0 ; d < ta->TA_Height ; d++)
      if(tabformreadline(ta, d) == -1) EGOTO(error);
    flag |= FLAG_REDRAW;
  } else if(mode & TABFORM_FREREADLINE) {
    int e;

    /*  Reread current line  */
    d = ta->TA_Current - ta->TA_Start;
    e = MIN(d + ta->TA_Multi, ta->TA_Height);
    for( ; d < e ; d++)
      if(tabformreadline(ta, d) == -1) EGOTO(error);
    flag |= FLAG_NEWCURR;
  }

  if(flag & FLAG_CHECKCURR) {
    ta->TA_Current -= ta->TA_Current % ta->TA_Multi;
    if(ta->TA_Current < 0)
      ta->TA_Current = 0;
    else if(ta->TA_Current >= ta->TA_NLines)
      ta->TA_Current = ta->TA_NLines - ta->TA_Multi;
  }

  if(flag & FLAG_RESETSEEK) {
    if(ta->TA_SeekIndex > 0) flag |= FLAG_NEWCURR;
    ta->TA_SeekIndex = 0;
  }

  if(flag & FLAG_NEWCURR) {
    if(ocurr != ta->TA_Current) {
      ocurr -= ta->TA_Start;
      d = MIN(ocurr + ta->TA_Multi, ta->TA_Height);
      for( ; ocurr < d ; ocurr++)
	tabformprintline(ta, ocurr);
    }

    /*  Print current line  */
    ocurr = ta->TA_Current - ta->TA_Start;
    d = MIN(ocurr + ta->TA_Multi, ta->TA_Height);
    for( ; ocurr < d ; ocurr++)
      tabformprintline(ta, ocurr);

    flag |= FLAG_REFRESH;
  } else if(flag & FLAG_REDRAW) {
    ocurr = ta->TA_Current;
    if(mode & TABFORM_FHIDECURRENT) ta->TA_Current = -1;
    for(d = 0 ; d < ta->TA_Height ; d++) {
      tabformprintline(ta, d);

      if(ta->TA_Mode & FOFI_SCROLLBAR)
	waddch(ta->TA_Window, JEDScrollChars[ta->TA_ScrollBarS <= d &&
					     d < ta->TA_ScrollBarE ? 1 : 0]);
    }
    ta->TA_Current = ocurr;
    flag |= FLAG_REFRESH;
  }

  if((flag & FLAG_REFRESH) && !(mode & TABFORM_FNOREFRESH))
    jedwstackrefresh(ta->TA_Window);

  /*  Having done all this, let's tell the caller the news  */
  if(flag & FLAG_UNKNOWN) return(-2);
  if(flag & FLAG_ILLEGAL) return(-3);

  return(0);
error:
  return(-1);
}

#undef FLAG_REDRAW
#undef FLAG_NEWCURR
#undef FLAG_REFRESH
#undef FLAG_CHECKCURR
#undef FLAG_ILLEGAL
#undef FLAG_UNKNOWN
#undef FLAG_RESETSEEK

/*JS*********************************************************************
*
*************************************************************************/

static int tabformseekkey(jedtable_t *ta, int key)

/************************************************************************/
{
  int i;
  const char *os;

  os = LINEMEM(ta, ta->TA_Current - ta->TA_Start);

  if(ta->TA_SeekIndex == 0)
    i = 0;
  else if(key == -1) {
    ta->TA_SeekIndex--;
    i = 0;
  } else
    i = ta->TA_Current;

  for( ; i < ta->TA_NLines ; i += ta->TA_Multi) {
    char *s;

    if(ta->TA_Start <= i && i < (ta->TA_Start + ta->TA_Height))
      s = LINEMEM(ta, i - ta->TA_Start);
    else {
      s = LINEMEM(ta, ta->TA_Height);
      if((*(ta->TA_GetLine))(ta->TA_UPtr, i, s, ta->TA_Width) == -1)
	return(-1);
    }

    if(ta->TA_SeekIndex && jsstrincmp(os, s, ta->TA_SeekIndex) != 0) continue;

    if(key == -1) {
      ta->TA_Current = i;
      return(0);
    }

    if(tolower((unsigned char)s[ta->TA_SeekIndex]) == tolower(key)) {
      ta->TA_SeekIndex++;
      ta->TA_Current = i;
      return(0);
    }
  }

  return(-3);
}

/*JS*********************************************************************
*
*************************************************************************/

static int tabformprintline(jedtable_t *ta, int nr)

/************************************************************************/
{
  const char *s;
  chtype attr;
  int d, w, i;

  wmove(ta->TA_Window, ta->TA_Line + nr, ta->TA_Column);

  s = LINEMEM(ta, nr);
  d = ta->TA_Start + nr - ta->TA_Current;
  i = ta->TA_SeekIndex;
  if(0 <= d && d < ta->TA_Multi)
    attr = i > 0 ? (A_BOLD | A_REVERSE) : A_REVERSE;
  else
    attr = 0;

  w = (ta->TA_Mode & FOFI_SCROLLBAR) ? ta->TA_Width - 1 : ta->TA_Width;
  while(w-- > 0) {
    int c;

    if((c = *(unsigned char *)s)) {
      s++;
      if(strchr("\n\v\f\r", c)) c = ' ';
    } else
      c = ' ';

    waddch(ta->TA_Window, attr | c);
    if(--i == 0) attr &= ~A_BOLD;
  }

  return(0);
}

/*JS*********************************************************************
*
*************************************************************************/

static int tabformscroll(jedtable_t *ta, int d)

/************************************************************************/
{
  int c;

  if(d < 0) { /* ***  scroll window down  *** */
    d = -d;
    if(d > ta->TA_Start) d = ta->TA_Start;

    if(d < ta->TA_Height)
      memmove(LINEMEM(ta, d), LINEMEM(ta, 0), (ta->TA_Height - d) *
	      LINEMEMLEN(ta));

    ta->TA_Start -= d;

    for(c = MIN(d, ta->TA_Height) - 1 ; c >= 0 ; c--)
      if(tabformreadline(ta, c) == -1) EGOTO(error);
  } else { /* ***  scroll window up  *** */
    c = ta->TA_NLines - ta->TA_Start - ta->TA_Multi;
    if(d > c) d = c;

    if(d < ta->TA_Height)
      memmove(LINEMEM(ta, 0), LINEMEM(ta, d), (ta->TA_Height - d) *
	      LINEMEMLEN(ta));

    ta->TA_Start += d;
    for(c = MAX(ta->TA_Height - d, 0) ; c < ta->TA_Height ; c++)
      if(tabformreadline(ta, c) == -1) EGOTO(error);
  }

  if(ta->TA_Mode & FOFI_SCROLLBAR) tabformsetscbar(ta);

  return(d);
error:
  return(-1);
}

/*JS*********************************************************************
*
*************************************************************************/

static int tabformalign(jedtable_t *ta)

/************************************************************************/
{
  int ret, d;

  ret = 0;
  if(ta->TA_Current < 0)
    ta->TA_Current = 0;
  else if(ta->TA_Current >= ta->TA_NLines)
    ta->TA_Current = ta->TA_NLines - ta->TA_Multi;

  if(ta->TA_Current < ta->TA_Start) {
    d = MIN(ta->TA_Current - ta->TA_Start, -TABLE_SCROLL_SKIP * ta->TA_Multi);
    if(tabformscroll(ta, d) == 0)
      EGOTO(error);

    ret = 1;

  } else if((ta->TA_Current + ta->TA_Multi) > (ta->TA_Start + ta->TA_Height)) {
    d = MAX(ta->TA_Current + ta->TA_Multi - ta->TA_Start - ta->TA_Height,
	    TABLE_SCROLL_SKIP * ta->TA_Multi);
    if(tabformscroll(ta, d) == 0)
      EGOTO(error);

    ret = 1;
  }

  if(ta->TA_Mode & FOFI_SCROLLBAR) tabformsetscbar(ta);

  return(ret);
error:
  return(-1);
}

#ifdef TEST
#define LINE    1
#define COLUMN  1
#define HEIGHT  25
#define WIDTH   75

#ifndef TABARRAY
static int getline(void *user, long nr, char *dest, int m)
{
  sprintf(dest, "This is line number %ld", 1 + nr);
  return(0);
}
#else
static const char *Lines[] = {
  "One", "Two", "Three", "Four", "Five", "Six", "Seven", "Eight", "Nine",
  "Ten", "Eleven", "Twelve"
};
#endif

int main(int argc, char *argv[])
{
  jedtable_t *ta;
  int key;

  jedinitialize(0);
  curs_set(0);

#ifndef TABARRAY
  ta = jedtabcreate(stdscr, LINE, COLUMN, HEIGHT, WIDTH, getline, NULL, 1, 121, NULL,
		     FOFI_SCROLLBAR);
#else
  ta = jedtabcreate(stdscr, LINE, COLUMN, HEIGHT, WIDTH, NULL, Lines, 1,
		     sizeof(Lines) / sizeof(Lines[0]), NULL,
		     FOFI_SCROLLBAR);
#endif

  _jedprintframe(stdscr, LINE - 1, COLUMN - 1, HEIGHT + 2, WIDTH + 3, 0);

  for(key = EDKEY_REDRAW ; ; ) {
    jedtabdokey(ta, key, 0);
    key = getch();
  }

  jedtabdelete(ta);

  jeddeinitialize();

  return(0);
}
#endif
