/*JS*********************************************************************
*
*    Program : JEDTREE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Tree display.
*
*              Copyright (C) 2006-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*
 * TODO:
 *  - implement selections.
 *  - JEDT_TREE_FHIDDEN attribute: skip in jedtmove
 *  - global variable for type of frame; subtract right amount if
 *    frame has shadows etc.
 */
/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

/*#include <dbgalloc.h>*/

#include <jsconfig.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <jssubs.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

#define DEBUG

/*********     DEFINES                                          *********/
#ifndef JEDT_TREE_FSUBNODES
/*  START-DEFINITIONS  */
#include <jsdllist.h>  /*  doubly linked lists  */

typedef struct jedtreenode_t jedtreenode_t;
struct jedtreenode_t {
  _jsdllnode_t JEDTN_Connect;

  long JEDTN_Flags;
  int JEDTN_SubLength;

  const char *JEDTN_Name;

  _jsdll_t JEDTN_Sub;

  void *JEDTN_User;
};

/* ***   For JEDTN_Flags  *** */
#define JEDT_TREE_FSUBNODES     (1UL << 0)
#define JEDT_TREE_FOPENED       (1UL << 1)
#define JEDT_TREE_FCALLBACK     (1UL << 2)
#define JEDT_TREE_FSELECTED     (1UL << 3)

#ifdef JEDLIB_INTERNAL
#define JEDT_TREE_FFREESTRING   (1UL << 8)
#define JEDT_TREE_FOPENING      (1UL << 9)
#endif

typedef struct jedtree_t jedtree_t;
struct jedtree_t {
  jedtreenode_t jedt_Content;
#define JEDT_Flags       jedt_Content.JEDTN_Flags
#define JEDT_NLines      jedt_Content.JEDTN_SubLength
#define JEDT_Title       jedt_Content.JEDTN_Name
#define JEDT_Nodes       jedt_Content.JEDTN_Sub
#define JEDT_User        jedt_Content.JEDTN_User

  WINDOW *JEDT_Window;
  int JEDT_Line, JEDT_Column, JEDT_Height, JEDT_Width;

  jedtreenode_t *JEDT_TopScreen;
  int JEDT_TopScreenLine, JEDT_TopScreenLvl;

  jedtreenode_t *JEDT_Current;
  int JEDT_CurrentLine;

  int JEDT_ColOff, JEDT_NCols;

  int (*JEDT_Setup)(void *user, jedtree_t *edtr, jedtreenode_t *node);
};

/*  For JEDT_Flags. Since the field might be misinterpreted for JEDTN_Flags,
 *   we use separate bits
 */
#define JEDT_TREEWIN_FDELWINDOW    (1UL << 16)
#define JEDT_TREEWIN_FFRAME        (1UL << 17)
#define JEDT_TREEWIN_FSCROLLBAR    (1UL << 18)
#define JEDT_TREEWIN_FLINENR       (1UL << 19)

/* **  Wrapper macros used internally to deal with doubly linked lists  *** */

#define _jedtdlln2node(nptr)   _jsdllnode2utype(nptr, jedtreenode_t, JEDTN_Connect)
#define _jedtdllh2node(hptr)   _jsdllnode2utype(hptr, jedtreenode_t, JEDTN_Sub)

#define jedtisempty(edtr)         _jsdllisempty(&(edtr->JEDT_Nodes))

#define jedtgetfirstnode(edtr)    _jedtdlln2node(_jsdllgetfirstnode(&(edtr->JEDT_Nodes)))
#define jedtgetlastnode(edtr)     _jedtdlln2node(_jsdllgetlastnode(&(edtr->JEDT_Nodes)))

#define	jedtgetnext(node)         _jedtdlln2node(_jsdllgetnext(&((node)->JEDTN_Connect)))
#define	jedtgetprev(node)         _jedtdlln2node(_jsdllgetprev(&((node)->JEDTN_Connect)))

#define jedtgetsubfirstnode(node) _jedtdlln2node(_jsdllgetfirstnode(&((node)->JEDTN_Sub)))
#define jedtgetsublastnode(node)  _jedtdlln2node(_jsdllgetlastnode(&((node)->JEDTN_Sub)))

#define jedtnodeistail(node)      _jsdllistail(&((node)->JEDTN_Connect))
#define jedtnodeishead(node)      _jsdllishead(&((node)->JEDTN_Connect))

#define jedtistreenode(node)      !_jsdllnodeinlist(&((node)->JEDTN_Connect))

#define jedthead2uppernode(node)  _jedtdllh2node(_jsdllhead2dll(&((node)->JEDTN_Connect)))
#define jedttail2uppernode(node)  _jedtdllh2node(_jsdlltail2dll(&((node)->JEDTN_Connect)))

/*  Modes for jedtreeaddnode etc.  */
#define JEDT_TREEADD_FADDHEAD      (1 << 0)
#define JEDT_TREEADD_FADDAFTER     (1 << 1)
#define JEDT_TREEADD_FADDBEFORE    (3 << 0)
#define JEDT_TREEADD_FDUPSTRING    (1 << 2)
#define JEDT_TREEADD_FKEEPDIRTY    (1 << 3)

#define JEDT_TREEADD_FKEEPSORT     (1 << 8)
#define JEDT_TREEADD_FCREATE       (1 << 9)
#define JEDT_TREEADD_FALLOWDUP     (1 << 10)

/*  END-DEFINITIONS  */
#endif

#ifdef CONFIG_HAS_COLORS
# define TREE_FRAME        JEditColors[_EDLIB_TCOLIND + 0]
# define TREE_COLOR        JEditColors[_EDLIB_TCOLIND + 1]
#endif

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype jedtree_t     *jedtreecreate(WINDOW *win, const char *title,
				       int line, int column, int height,
				       int width, int flags);
Prototype void           jedtreedelete(jedtree_t *edtr);
Prototype int            jedtreesetup(jedtree_t *edtr);

Prototype int            jedtreesetname(jedtreenode_t *node, const char *name,
					int mode);

Prototype jedtreenode_t *jedtreeaddnode(jedtree_t *edtr, jedtreenode_t *parent,
					const char *name, int mode);
Prototype jedtreenode_t *jedtreeaddnodepath(jedtree_t *edtr,
					    jedtreenode_t *root,
					    const char *path, int mode);
Prototype jedtreenode_t *jedtreeaddnodepath2(jedtree_t *edtr,
					     jedtreenode_t *root,
					     const char *parts[], int mode);

static jedtreenode_t    *_jedtreenodealloc(void);
static void              _jedtreenodefree(jedtreenode_t *node);
/*
static void              _jedtreememclean(void);
static void              _jedtreememfree(void);
*/

Prototype void           jedtreedelnode(jedtreenode_t *node, int mode);

Prototype int            jedtreegetpath(jedtreenode_t *node, const char *sep,
					char *dest, int dlen);

Prototype int            jedtreedokey(jedtree_t *edtr, int key, int arg,
				      int mode);
static int               jedtreeopennode(jedtree_t *edtr, jedtreenode_t *node,
					 int mode);
static int               jedtreeclosenode(jedtree_t *edtr, jedtreenode_t *node);

static int               jedtreeopenall(jedtree_t *edtr);
static int               jedtreecloseall(jedtree_t *edtr);

#ifdef DEBUG
static void              jedtreecheck(jedtree_t *edtr);
#endif
static int               jedtreeprint(jedtree_t *edtr, int start, int end,
				      int mode);
static int               jedtmove(jedtreenode_t **pNode, int *pLevel,
				  int amount);

Prototype int            jedtreesearch(jedtree_t *edtr,
				       int (*function)(void *, jedtreenode_t *, int, int),
				       void *uptr, int mode);

/*  0: Scroll one screen size.
 *  > 0: scroll so many lines.
 *  < 0: Scroll screen size / -ScreenScroll lines
 */
int EdLibPageScroll = 1;

#define GETPAGESCROLL(pagesize)  (EdLibPageScroll == 0 ? (pagesize) : \
				  EdLibPageScroll > 0 ? EdLibPageScroll : \
				  (pagesize) / -EdLibPageScroll)

/* ERROR-DEFINITIONS from editTree label _ERR_EDITTREE
   Intermediate nodes does not exist
   Node already exists
*/

/*JS*********************************************************************
*
*************************************************************************/

jedtree_t *jedtreecreate(WINDOW *win, const char *title, int line, int column,
			 int height, int width, int flags)

/************************************************************************/
{
  jedtree_t *edtr;

  if((edtr = (jedtree_t *)malloc(sizeof(*edtr))) == NULL)
    return NULL;

  edtr->JEDT_Flags = flags;

  edtr->JEDT_Title = title;

  /*  Center tree  */
  if(height <= 0) height += LINES;
  if(width <= 0) width += COLS;
  if(line == -1 && (line = (LINES - height) / 2) < 0) line = 0;
  if(column == -1 && (column = (COLS - width) / 2) < 0) column = 0;

  if(win == NULL) {
    /*  Create window for the tree  */
    if((win = jedwstacknew(height, width, line, column)) == NULL)
      EGOTO(error);

#ifdef CONFIG_HAS_COLORS
    if(!(JEditCursesMode & CURSMODE_NOCOLORS) &&
       wbkgd(win, TREE_COLOR | ' ') == ERR) {
      jedwstackdel(win);
      EGOTO(error);
    }
#endif

    keypad(win, TRUE); intrflush(win, FALSE);
    idlok(win, TRUE); /* ??? */
    werase(win);

    line = 0; column = 0; /*  now coordinates relative to new window  */

    edtr->JEDT_Flags |= JEDT_TREEWIN_FDELWINDOW;
  }

  edtr->JEDT_Window = win;

  if(flags & JEDT_TREEWIN_FFRAME) {
    ++line; ++column;
    height -= 2; width -= 2;
  }

  edtr->JEDT_Line = line;
  edtr->JEDT_Column = column;
  edtr->JEDT_Height = height;
  edtr->JEDT_Width = width;

  edtr->JEDT_ColOff = edtr->JEDT_NCols = 0;

  /*  Initialize linked lists  */
  edtr->jedt_Content.JEDTN_Flags |= JEDT_TREE_FSUBNODES;
  edtr->JEDT_NLines = 0;
  _jsdllcleannode(&(edtr->jedt_Content.JEDTN_Connect));
  _jsdllinit(&(edtr->JEDT_Nodes));

  edtr->JEDT_TopScreen = edtr->JEDT_Current = NULL;
  edtr->JEDT_TopScreenLine = edtr->JEDT_TopScreenLvl =
    edtr->JEDT_CurrentLine = 0;

  edtr->JEDT_Setup = NULL;
  edtr->JEDT_User = NULL;

  return edtr;
error:
  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

void jedtreedelete(jedtree_t *edtr)

/************************************************************************/
{
  /*  Free content  */
  jedtreedelnode(&(edtr->jedt_Content), 0);

  if(edtr->JEDT_Flags & JEDT_TREEWIN_FDELWINDOW)
    jedwstackdel(edtr->JEDT_Window);

  free(edtr);
}

/*JS*********************************************************************
*   Recalculates and sets lengths of all subnodes. Used after adding
*    nodes with 'JEDT_TREEADD_FKEEPDIRTY' set.
*************************************************************************/

int jedtreesetup(jedtree_t *edtr)

/************************************************************************/
{
  jedtreenode_t *node;
  int line;

  node = jedtgetsubfirstnode(&(edtr->jedt_Content));
  line = 0;
  for(;;) {
    if(jedtnodeistail(node)) {
      node = jedttail2uppernode(node);

      if(jedtistreenode(node))
	/*  We have reached the end of our walk  */
	break;

      if(node->JEDTN_Flags & JEDT_TREE_FOPENED) {
	node->JEDTN_SubLength = line - node->JEDTN_SubLength - 1;
      } else {
	line = node->JEDTN_SubLength + 1;
	node->JEDTN_SubLength = 0;
      }

      node = jedtgetnext(node);

    } else if(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) {
      /*  Memorize current line so we can later put the length there
       *   or restore the current line in case the node is closed
       */
      node->JEDTN_SubLength = line;

      node = jedtgetsubfirstnode(node);
      ++line;

    } else {
      node = jedtgetnext(node);
      ++line;
    }
  }

  edtr->JEDT_NLines = line;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedtreesetname(jedtreenode_t *node, const char *name, int mode)

/************************************************************************/
{
  if(node->JEDTN_Flags & JEDT_TREE_FFREESTRING)
    free((void *)node->JEDTN_Name);

  if(mode & JEDT_TREEADD_FDUPSTRING) {
    if((name = strdup(name)) == NULL)
      return -1;

    node->JEDTN_Flags |= JEDT_TREE_FFREESTRING;
  } else
    node->JEDTN_Flags &= ~JEDT_TREE_FFREESTRING;

  node->JEDTN_Name = name;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

jedtreenode_t *jedtreeaddnode(jedtree_t *edtr, jedtreenode_t *parent,
			      const char *name, int mode)

/************************************************************************/
{
  jedtreenode_t *node;
  int line;

  if((node = _jedtreenodealloc()) == NULL)
    EGOTO(error);

  node->JEDTN_Flags = 0;
  if(mode & JEDT_TREEADD_FDUPSTRING) {
    if((name = strdup(name)) == NULL)
      EGOTO(error);

    node->JEDTN_Flags |= JEDT_TREE_FFREESTRING;
  }
  node->JEDTN_Name = name;

  node->JEDTN_SubLength = 0;

  /*  Initialize linked list of sub nodes  */
  _jsdllinit(&(node->JEDTN_Sub));
  node->JEDTN_User = NULL;

  if(parent == NULL) {
    assert(!(mode & JEDT_TREEADD_FADDAFTER));

    parent = &(edtr->jedt_Content);
  }

#ifdef DEBUG
  if(Debug & 4096)
    jedtreecheck(edtr);
#endif

  /*  Determine the line number of added node to
   *   be able to correct screen stuff below
   *   We walk later up to the top level node
   *   which has line number -1.
   */
  line = -1;

  if(mode & JEDT_TREEADD_FADDAFTER) {
    /* ***  Hang in on same level  *** */
    if(mode & JEDT_TREEADD_FADDHEAD) {
      _jsdlladdbefore(&(parent->JEDTN_Connect), &(node->JEDTN_Connect));
      parent = node;
    } else {
      _jsdlladdafter(&(parent->JEDTN_Connect), &(node->JEDTN_Connect));

      if(parent->JEDTN_Flags & JEDT_TREE_FOPENED)
	line += parent->JEDTN_SubLength;

      ++line;
    }

    if(!(mode & JEDT_TREEADD_FKEEPDIRTY)) {
      /*  Walk up to *real* parent for the length increase below  */
      do {
	assert(jedtistreenode(parent));
	parent = jedtgetprev(parent);
	++line;
      } while(!jedtnodeishead(parent));

      parent = jedthead2uppernode(parent);
    }

  } else if(mode & JEDT_TREEADD_FKEEPSORT) {
    jedtreenode_t *sub;

    /* ***  Find the right place where to insert  *** */
    for(sub = jedtgetsubfirstnode(parent), ++line ;
	!jedtnodeistail(sub) ; sub = jedtgetnext(sub), ++line)
      if(strcmp(sub->JEDTN_Name, name) > 0)
	break;

    /*  We are at the end or 'parent' points to a node which is greater  */
    _jsdlladdbefore(&(sub->JEDTN_Connect), &(node->JEDTN_Connect));

    parent->JEDTN_Flags |= JEDT_TREE_FSUBNODES;

  } else {
    /* ***  Hang in as a child node  *** */
    if(mode & JEDT_TREEADD_FADDHEAD) {
      _jsdlladdhead(&(parent->JEDTN_Sub), &(node->JEDTN_Connect));
      ++line;
    } else {
      _jsdlladdtail(&(parent->JEDTN_Sub), &(node->JEDTN_Connect));
      line += parent->JEDTN_SubLength + 1;
    }

    parent->JEDTN_Flags |= JEDT_TREE_FSUBNODES;
  }

  if(!(mode & JEDT_TREEADD_FKEEPDIRTY)) {
    /*  Correct length of parent nodes and determine current line of parent  */
    while(parent->JEDTN_Flags & JEDT_TREE_FOPENED) {
      /*  This will also affect JEDT_NLines in case parent == treenode  */
      parent->JEDTN_SubLength++;

      if(jedtistreenode(parent)) {
	/*  The added node is visible, so correct the screen lines  */
	if(edtr->JEDT_TopScreen != NULL && edtr->JEDT_TopScreenLine >= line)
	  edtr->JEDT_TopScreenLine++;

        if(edtr->JEDT_Current != NULL && edtr->JEDT_CurrentLine >= line) {
	  edtr->JEDT_CurrentLine++;

          /*  The current line might have moved out of the screen  */
          if((line = edtr->JEDT_CurrentLine - (edtr->JEDT_TopScreenLine +
					       edtr->JEDT_Height - 1)) > 0) {
            line = jedtmove(&(edtr->JEDT_TopScreen), &(edtr->JEDT_TopScreenLvl),
		            line);
            edtr->JEDT_TopScreenLine += line;
          }
        }

	break;
      }

      /*  Walk up to the end of the sequence  */
      do {
	parent = jedtgetprev(parent);

	if(parent->JEDTN_Flags & JEDT_TREE_FOPENED)
	  line += parent->JEDTN_SubLength;

	++line;
      } while(!jedtnodeishead(parent));

      parent = jedthead2uppernode(parent);
    }

#ifdef DEBUG
    if(Debug & 2048)
      jedtreecheck(edtr);
#endif
  }

  return node;
error:
  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

jedtreenode_t *jedtreeaddnodepath(jedtree_t *edtr, jedtreenode_t *root,
				  const char *path, int mode)

/************************************************************************/
{
static const char **Parts;
static int MParts;
  char *buffer;
  int mbuffer, n;

  if(*path == '/') {
    /*  An absolute path  */
    root = &(edtr->jedt_Content);
    ++path;
  }

  buffer = NULL;
  mbuffer = 0;
  for(n = 0 ; *path ; ++n) {
    int i;

    /*  Locate next path element. '//' is an escaped slash. Disadvantage
     *   is that this does not allow empty parts. Use jedtreeaddnodep2 for this.
     */
    for(i = 0 ; *path ; ++path) {
      if(*path == '/') {
	if(path[1] != '/') break;
	++path;
      }

      if((i + 2) >= mbuffer) {
	REALLOC_RESIZE(mbuffer, 80);
	if((buffer = (char *)realloc(buffer, mbuffer)) == NULL)
	  EGOTO(error);
      }

      buffer[i++] = *path;
    }
    buffer[i] = '\0';

    if(n >= MParts) {
      REALLOC_RESIZE(MParts, 10);
      if((Parts = (const char **)realloc(Parts, MParts * sizeof(*Parts))) == NULL)
	EGOTO(error);
    }

    Parts[n] = buffer;
    buffer = NULL; mbuffer = 0;

    if(*path == '/') ++path;
  }

  Parts[n] = NULL;

#ifdef XDEBUG
  for(n = 0 ; Parts[n] ; ++n)
    printf("Parts[%d]='%s'\n", n, Parts[n]);
#endif

  return jedtreeaddnodepath2(edtr, root, Parts, mode & ~4);
error:
  return NULL;
}

/*JS*********************************************************************
*   mode bit 3: when inserting, sort entries alphabetically.
*        bit 4: If set, create intermediate nodes automatically.
*        bit 5: allow duplicate entries.
*************************************************************************/

jedtreenode_t *jedtreeaddnodepath2(jedtree_t *edtr, jedtreenode_t *root,
				   const char *parts[], int mode)

/************************************************************************/
{
  int i;

  if(root == NULL) root = &(edtr->jedt_Content);

  for(i = 0 ; parts[i] ; ++i) {
    jedtreenode_t *sub, *sEntry;
    const char *part;

    /*  If duplicates are allowed, just scan through to find the tail
     *   during the last iteration
     */
    part = (mode & JEDT_TREEADD_FALLOWDUP) && parts[i + 1] == NULL ?
      NULL : parts[i];

#ifdef XDEBUG
    printf("SEARCH %d %s\n", root, part ? part : "<NULL>");
#endif

    /*  FIXME: This code is probably not working correctly,
     *   but I am too tired now to fix it....
     */

    sEntry = NULL;
    for(sub = jedtgetsubfirstnode(root) ; !jedtnodeistail(sub) ;
	sub = jedtgetnext(sub)) {
      if((mode & JEDT_TREEADD_FKEEPSORT) &&
	 strcmp(sub->JEDTN_Name, parts[i]) < 0)
	sEntry = sub;

      if(part && strcmp(sub->JEDTN_Name, part) == 0) {
	if(!(mode & JEDT_TREEADD_FALLOWDUP) &&
	   parts[i + 1] == NULL) {
	  JSSETERROR2(_ERR_EDITTREE, 1, part);
	  EGOTO(error);
	}

	break;
      }
    }

    if(jedtnodeistail(sub)) {
      int mode2;

      /*  Node not found  */
      if(parts[i + 1] != NULL && !(mode & JEDT_TREEADD_FCREATE)) {
	JSSETERROR2(_ERR_EDITTREE, 0, part);
	EGOTO(error);
      }

      mode2 = (mode & JEDT_TREEADD_FDUPSTRING);
      if(mode & JEDT_TREEADD_FKEEPSORT)
	mode2 |= sEntry == NULL ? JEDT_TREEADD_FADDHEAD : JEDT_TREEADD_FADDAFTER;

      if((sub = jedtreeaddnode(edtr, root, parts[i], mode2)) == NULL)
	EGOTO(error);
    }

    root = sub;
  }

  return root;
error:
  return NULL;
}

/*JS*********************************************************************
*   Generate allocation routine for nodes
*************************************************************************/

#define GENMALLOC_TYPE          jedtreenode_t
#define GENMALLOC_NAMEALLOC     _jedtreenodealloc
#define GENMALLOC_NAMEFREE      _jedtreenodefree

/*
#define GENMALLOC_NAMECLEANMEM  _jedtreememclean
#define GENMALLOC_NAMEFREEMEM   _jedtreememfree
*/

#define GENMALLOC_MEMSIZE       5000
#define GENMALLOC_STATIC

#include "genmalloc.h"

/*JS*********************************************************************
*
*************************************************************************/

void jedtreedelnode(jedtreenode_t *node, int mode)

/************************************************************************/
{
  if((node->JEDTN_Flags & JEDT_TREE_FSUBNODES) &&
     !_jsdllisempty(&(node->JEDTN_Sub))) {
    jedtreenode_t *sub, *next;

    sub = node;
    do {
      if(sub->JEDTN_Flags & JEDT_TREE_FSUBNODES) {
	/*  Clear bit because we meet this node again and want to
	 *   avoid endless recursion. Then step down to sub nodes.
	 */
	sub->JEDTN_Flags &= ~JEDT_TREE_FSUBNODES;
	next = jedtgetsubfirstnode(sub);

	/*  Deal with empty subnodes  */
	if(!jedtnodeistail(next)) {
	  sub = next;
	  continue;
	}
      }

      next = jedtgetnext(sub);

      if(sub->JEDTN_Flags & JEDT_TREE_FFREESTRING)
	free((void *)sub->JEDTN_Name);
      _jedtreenodefree(sub);

      if(jedtnodeistail(next))
	/*  End of list reached, walk up  */
	next = jedttail2uppernode(next);

      sub = next;
    } while(sub != node);

    /*  After deleting all subnodes, clear linked list  */
    _jsdllinit(&(node->JEDTN_Sub));
  }

  if(mode & 1) {
    _jsdllremove(&(node->JEDTN_Connect));

    if(node->JEDTN_Flags & JEDT_TREE_FFREESTRING)
      free((void *)node->JEDTN_Name);

    free(node);
  }

  /*  FIXME: Need to correct screen lines!  */
}

/*JS*********************************************************************
*
*************************************************************************/

int jedtreegetpath(jedtreenode_t *node, const char *sep,
		   char *dest, int dlen)

/************************************************************************/
{
  int offset, len;

  if(!jedtistreenode(node)) {
    jedtreenode_t *parent;

    /*  Find parent node  */
    for(parent = node ; !jedtnodeistail(parent) ; )
      parent = jedtgetnext(parent);

    parent = jedttail2uppernode(parent);

    /*  Recursive call: put parent path in front of us  */
    if((offset = jedtreegetpath(parent, sep, dest, dlen)) == -1)
      EGOTO(error);

    len = strlen(sep);
    if(offset + len >= dlen)
      EGOTO(error);

    strcpy(dest + offset, sep);
    offset += len;
  } else
    offset = 0;

  if(!jedtistreenode(node)) {
    /*  Print our name  */
    len = strlen(node->JEDTN_Name);
    if(offset + len >= dlen)
      EGOTO(error);

    strcpy(dest + offset, node->JEDTN_Name);
    offset += len;
  }
  dest[offset] = '\0';

  return offset;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedtreedokey(jedtree_t *edtr, int key, int arg, int mode)

/************************************************************************/
{
  jedtreenode_t *node;
  int ret, start, end, pmode;

  /*  Refresh defaults to whole screen  */
  start = 0;
  end = edtr->JEDT_Height - 1;
  pmode = 1;

  /* ****************************************** */
  if(key == KEY_UP) { /* ***  MOVE DOWN ONE LINE *** */
    if((ret = jedtmove(&(edtr->JEDT_Current), NULL, -1)) == 0)
      goto notpossible;

    if((edtr->JEDT_CurrentLine += ret) < edtr->JEDT_TopScreenLine) {
      ret = jedtmove(&(edtr->JEDT_TopScreen), &(edtr->JEDT_TopScreenLvl),
		     -GETPAGESCROLL(edtr->JEDT_Height));
      edtr->JEDT_TopScreenLine += ret;
    } else {
      start = edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine;
      end = start + 1;
      pmode = 0;
    }

  /* ****************************************** */
  } else if(key == KEY_DOWN) { /* ***  MOVE UP ONE LINE  *** */
    if((ret = jedtmove(&(edtr->JEDT_Current), NULL, 1)) == 0)
      goto notpossible;

    if((edtr->JEDT_CurrentLine += ret) >=
       (edtr->JEDT_TopScreenLine + edtr->JEDT_Height)) {
      ret = jedtmove(&(edtr->JEDT_TopScreen), &(edtr->JEDT_TopScreenLvl),
		     GETPAGESCROLL(edtr->JEDT_Height));
      edtr->JEDT_TopScreenLine += ret;
    } else {
      end = edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine;
      start = end - 1;
      pmode = 0;
    }

  /* ****************************************** */
  } else if(key == 0 || key == EDKEY_REDRAW) { /* *** REDRAW  *** */
    start = end = -1;

  /* ****************************************** */
  } else if(key == 'I') { /* ***  SHOW INACTIVE  *** */
    start = end = -1;
    pmode |= 2;

  /* ****************************************** */
  } else if(key == KEY_LEFT) { /* ***  SCROLL LEFT  *** */
    if(edtr->JEDT_ColOff <= 0)
      goto notpossible;

    edtr->JEDT_ColOff--;
    pmode = 0;

  /* ****************************************** */
  } else if(key == KEY_RIGHT) { /* ***  SCROLL RIGHT  *** */
    if(edtr->JEDT_NCols != 0 && edtr->JEDT_ColOff >= (edtr->JEDT_NCols - 1))
      goto notpossible;

    edtr->JEDT_ColOff++;
    pmode = 0;

  /* ****************************************** */
  } else if(key == KEY_PPAGE) { /* ***  MOVE DOWN ONE PAGE  *** */
    if((ret = jedtmove(&(edtr->JEDT_TopScreen), &(edtr->JEDT_TopScreenLvl),
		       -edtr->JEDT_Height)) == 0)
      goto notpossible;

    edtr->JEDT_TopScreenLine += ret;
    edtr->JEDT_CurrentLine += jedtmove(&(edtr->JEDT_Current), NULL, ret);

  } else if(key == KEY_NPAGE) { /* ***  MOVE UP ONE PAGE  *** */
    if((ret = jedtmove(&(edtr->JEDT_TopScreen), &(edtr->JEDT_TopScreenLvl),
		       edtr->JEDT_Height)) == 0)
      goto notpossible;

    edtr->JEDT_TopScreenLine += ret;
    edtr->JEDT_CurrentLine += jedtmove(&(edtr->JEDT_Current), NULL, ret);

  /* ****************************************** */
  } else if(key == '+') { /* ***  OPEN CURRENT NODE  *** */
    if((node = edtr->JEDT_Current) == NULL)
      goto notpossible;

    if((ret = jedtreeopennode(edtr, node, 0)) == -1)
      EGOTO(error);

    if(ret == 0)
      goto notpossible;

    start = edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine;

  /* ****************************************** */
  } else if(key == '-') { /* ***  CLOSE CURRENT NODE  *** */
    if((node = edtr->JEDT_Current) == NULL)
      goto notpossible;

    if(jedtreeclosenode(edtr, node) == 0)
      goto notpossible;

    start = edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine;

  /* ****************************************** */
  } else if(key == ' ') { /* ***  TOGGLE CURRENT NODE  *** */
    if((node = edtr->JEDT_Current) == NULL)
      goto notpossible;

    if((ret = (node->JEDTN_Flags & JEDT_TREE_FOPENED) ?
	jedtreeclosenode(edtr, node) :
	jedtreeopennode(edtr, node, 0)) == -1)
      EGOTO(error);

    if(ret == 0)
      goto notpossible;

    start = edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine;

  /* ****************************************** */
  } else if(key == 'C') { /* ***  CLOSE ALL NODES  *** */
    if(jedtreecloseall(edtr) == 0)
      goto notpossible;

    start = end = -1;

  /* ****************************************** */
  } else if(key == 'O') { /* ***  OPEN ALL NODES  *** */
    if(jedtreeopenall(edtr) == 0)
      goto notpossible;

    /*  The current line might have moved out of the screen  */
    ret = edtr->JEDT_CurrentLine - (edtr->JEDT_TopScreenLine +
				    edtr->JEDT_Height);
    if(ret > 0) {
      ret = jedtmove(&(edtr->JEDT_TopScreen), &(edtr->JEDT_TopScreenLvl),
                     ret);
      edtr->JEDT_TopScreenLine += ret;
    }

    start = end = -1;

  } else {
    /*  Unknown key  */
    goto notpossible;
  }

#ifdef DEBUG
  if(Debug & 2048)
    jedtreecheck(edtr);
#endif

  /*wclear(edtr->JEDT_Window);*/

  if(jedtreeprint(edtr, start, end, pmode) == -1)
    return -1;

  wrefresh(edtr->JEDT_Window);

  return 1;
notpossible:
  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int jedtreeopennode(jedtree_t *edtr, jedtreenode_t *node, int mode)

/************************************************************************/
{
  jedtreenode_t *sub;
  int length;

  if(!(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) ||
     (node->JEDTN_Flags & JEDT_TREE_FOPENED))
    return 0;

  node->JEDTN_Flags |= JEDT_TREE_FOPENED;

  if(edtr->JEDT_Setup != NULL) {
    while(node->JEDTN_Flags & JEDT_TREE_FCALLBACK) {
      int ret;

      if((ret = edtr->JEDT_Setup(edtr->JEDT_User, edtr, node)) == -1)
	EGOTO(error);

      if(ret == 0)
	break;

      if(mode & 1)
	/*  We are setting up the top level node, so make sure first line
	 *   is correct
	 */
	edtr->JEDT_TopScreen = NULL;

      jedtreeprint(edtr, -1, -1, 0);
      wrefresh(edtr->JEDT_Window);
    }
  }

  length = 0;
  for(sub = jedtgetsubfirstnode(node) ; !jedtnodeistail(sub) ;
      sub = jedtgetnext(sub))
    length += 1 + sub->JEDTN_SubLength;

  node->JEDTN_SubLength = length;

  /*  Also correct length of all our parent nodes  */
  while(!jedtistreenode(node)) {
    /*  Walk up to the end of the sequence  */
    do
      node = jedtgetnext(node);
    while(!jedtnodeistail(node));

    node = jedttail2uppernode(node);

    if(!(node->JEDTN_Flags & JEDT_TREE_FOPENED))
      break;

    node->JEDTN_SubLength += length;
  }

  /*  FIXME: In case the opened node is the current one, none
   *   of the screen related lines changes. Otherwise, we have
   *   to correct them.
   */

  return 1;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int jedtreeclosenode(jedtree_t *edtr, jedtreenode_t *node)

/************************************************************************/
{
  int length;

  if(!(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) ||
     !(node->JEDTN_Flags & JEDT_TREE_FOPENED))
    return 0;

  node->JEDTN_Flags &= ~JEDT_TREE_FOPENED;
  if(edtr->JEDT_Setup != NULL) {
    while(node->JEDTN_Flags & JEDT_TREE_FCALLBACK) {
      int ret;

      if((ret = edtr->JEDT_Setup(edtr->JEDT_User, edtr, node)) == -1)
	return -1;

      if(ret == 0)
	break;

      jedtreeprint(edtr, -1, -1, 0);
      wrefresh(edtr->JEDT_Window);
    }
  }

  length = node->JEDTN_SubLength;
  node->JEDTN_SubLength = 0;

  do {
    while(!jedtnodeistail(node))
      node = jedtgetnext(node);

    node = jedttail2uppernode(node);
    node->JEDTN_SubLength -= length;
  } while(!jedtistreenode(node));

  /*  FIXME: In case the closed node is the current one, none
   *   of the screen related lines changes. Otherwise, we have
   *   to correct them.
   */

  return 1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int jedtreeopenall(jedtree_t *edtr)

/************************************************************************/
{
  jedtreenode_t *node;
  int line, ret;

  ret = 0;

  node = jedtgetsubfirstnode(&(edtr->jedt_Content));
  line = 0;
  for(;;) {
    /*  Adapt screen related lines  */
    if(node == edtr->JEDT_Current)
      edtr->JEDT_CurrentLine = line;

    if(node == edtr->JEDT_TopScreen)
      edtr->JEDT_TopScreenLine = line;

    if(jedtnodeistail(node)) {
      node = jedttail2uppernode(node);

      if(jedtistreenode(node))
	/*  We have reached the end of our walk  */
	break;

      node->JEDTN_SubLength = line - node->JEDTN_SubLength - 1;
      if(!(node->JEDTN_Flags & JEDT_TREE_FOPENED)) {
	node->JEDTN_Flags |= JEDT_TREE_FOPENED;

	if(edtr->JEDT_Setup != NULL &&
	   (node->JEDTN_Flags & JEDT_TREE_FCALLBACK) &&
	   edtr->JEDT_Setup(edtr->JEDT_User, edtr, node) == -1)
	  EGOTO(error);

	ret = 1; /*  we did something  */
      }

      node = jedtgetnext(node);

    } else if(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) {
      /*  Memorize current line so we can later put the length there  */
      node->JEDTN_SubLength = line;

      node = jedtgetsubfirstnode(node);
      ++line;

    } else {
      node = jedtgetnext(node);
      ++line;
    }
  }

  edtr->JEDT_NLines = line;

  return ret;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int jedtreecloseall(jedtree_t *edtr)

/************************************************************************/
{
  jedtreenode_t *node, *visinode;
  int line, level, ret;

  ret = 0;

  node = jedtgetsubfirstnode(&(edtr->jedt_Content));
  line = 0;
  level = 0; /*  only nodes on level 0 are visible  */
  for(;;) {
    if(level == 0)
      /*  Track visbible nodes (only on level 0)  */
      visinode = node;

    /*  Adapt screen related stuff  */
    if(node == edtr->JEDT_Current) {
      edtr->JEDT_Current = visinode;
      edtr->JEDT_CurrentLine = line;
    }

    if(node == edtr->JEDT_TopScreen) {
      edtr->JEDT_TopScreen = visinode;
      edtr->JEDT_TopScreenLine = line;
      edtr->JEDT_TopScreenLvl = 0;
    }

    if(jedtnodeistail(node)) {
      node = jedttail2uppernode(node);
      --level;

      if(jedtistreenode(node))
	/*  We have reached the end of our walk  */
	break;

      if(node->JEDTN_Flags & JEDT_TREE_FOPENED) {
	node->JEDTN_Flags &= ~JEDT_TREE_FOPENED;
	node->JEDTN_SubLength = 0;

	ret = 1; /*  we did something  */
      }

      node = jedtgetnext(node);
      if(level == 0)
	++line;

    } else {
      if(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) {
	node = jedtgetsubfirstnode(node);
	++level;

      } else {
	node = jedtgetnext(node);

	if(level == 0)
	  ++line;
      }
    }
  }

  edtr->JEDT_NLines = line;

  return ret;
}

#ifdef DEBUG
/*JS*********************************************************************
*  Check the consistency of the tree.
*************************************************************************/

static void jedtreecheck(jedtree_t *edtr)

/************************************************************************/
{
  jedtreenode_t *node;
  long alines[1000];
  int line, level;

  node = jedtgetsubfirstnode(&(edtr->jedt_Content));
  level = 1;
  alines[0] = -1;
  line = 0;
  for(;;) {
    assert(node != edtr->JEDT_TopScreen ||
	   edtr->JEDT_TopScreenLine == line);

    assert(node != edtr->JEDT_Current ||
	   edtr->JEDT_CurrentLine == line);

    if(jedtnodeistail(node)) {
      node = jedttail2uppernode(node);
      --level;

      if(node->JEDTN_Flags & JEDT_TREE_FOPENED) {
	assert(node->JEDTN_SubLength == (line - alines[level] - 1));
      } else {
	assert(node->JEDTN_SubLength == 0);

	line = alines[level] + 1;
      }

      if(jedtistreenode(node))
	/*  We have reached the end of our walk  */
	break;

      node = jedtgetnext(node);

    } else if(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) {
      alines[level] = line;

      node = jedtgetsubfirstnode(node);
      ++line;
      ++level;

    } else {
      node = jedtgetnext(node);
      ++line;
    }
  }

  assert(level == 0);
  assert(edtr->JEDT_NLines == line);
}
#endif

/*JS*********************************************************************
*   mode bit 0: update scrollbar.
*   mode bit 1: show as deselected
*************************************************************************/

static int jedtreeprint(jedtree_t *edtr, int start, int end, int mode)

/************************************************************************/
{
  WINDOW *win;
  jedtreenode_t *node, *current;
  int width, line, level, maxCols;

  win = edtr->JEDT_Window;

  /*  Maybe this is the first time we are called and we should initialize
   *   the root node on the fly?
   */
  if(!(edtr->JEDT_Flags & JEDT_TREE_FOPENED) &&
     !(edtr->JEDT_Flags & JEDT_TREE_FOPENING)) {
    edtr->JEDT_Flags |= JEDT_TREE_FSUBNODES | JEDT_TREE_FOPENING;
    level = jedtreeopennode(edtr, &(edtr->jedt_Content), 1);
    edtr->JEDT_Flags &= ~JEDT_TREE_FOPENING;

    if(level == -1) EGOTO(error);
  }

  if(edtr->JEDT_TopScreen == NULL && !jedtisempty(edtr)) {
    edtr->JEDT_TopScreen = jedtgetfirstnode(edtr);
    edtr->JEDT_TopScreenLine = edtr->JEDT_TopScreenLvl = 0;
  }

  if(!(edtr->JEDT_Flags & JEDT_TREE_FOPENING) &&
     edtr->JEDT_Current == NULL && !jedtisempty(edtr)) {
    edtr->JEDT_Current = edtr->JEDT_TopScreen;
    edtr->JEDT_CurrentLine = edtr->JEDT_TopScreenLine;
  }

  current = !(mode & 2) ? edtr->JEDT_Current : NULL;

  /*  We print frame and title only on the very first call  */
  if(start == -1 && end == -1) {
#ifdef CONFIG_HAS_COLORS
    if(edtr->JEDT_Flags & JEDT_TREEWIN_FFRAME) {
      wattron(win, TREE_FRAME);
      _jedprintframe(win, edtr->JEDT_Line - 1, edtr->JEDT_Column - 1,
		     edtr->JEDT_Height + 2, edtr->JEDT_Width + 2,
		     (mode & 2) ? 1 | 4 : 64 | 4 | 1);
      wattroff(win, TREE_FRAME);
    }
#endif

    if(edtr->JEDT_Title != NULL)
      _jedprinttitle(win, edtr->JEDT_Column, edtr->JEDT_Width, edtr->JEDT_Title);
  }

  /*  'start' and 'end' 'line' are screen lines  */
  if(start < 0) start = 0;
  if(end < 0) end += edtr->JEDT_Height;

  node = edtr->JEDT_TopScreen;
  level = edtr->JEDT_TopScreenLvl;
  maxCols = edtr->JEDT_NCols - edtr->JEDT_ColOff;
  width = edtr->JEDT_Width;
  if(edtr->JEDT_Flags & JEDT_TREEWIN_FSCROLLBAR) --width;
  if(edtr->JEDT_Flags & JEDT_TREEWIN_FLINENR)
    width -= 6;

  line = 0;
  if(node != NULL) {
    while(!jedtnodeistail(node) && line <= end) {
      int i;

      /*  Display current node  */
      if(line >= start) {
	int col;

	wmove(win, edtr->JEDT_Line + line, edtr->JEDT_Column);

	if(edtr->JEDT_Flags & JEDT_TREEWIN_FLINENR)
	  wprintw(win, "%6d", edtr->JEDT_TopScreenLine + line);

	col = -edtr->JEDT_ColOff;
	for(i = 0 ; i < level ; ++i) {
	  if(col++ >= 0) waddch(win, '|');
	  if(col++ >= 0) waddch(win, ' ');
	}

	if(col++ >= 0)
	  waddch(win, !(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) ? '|' :
		 (node->JEDTN_Flags & JEDT_TREE_FOPENED) ? '-' : '+');
	if(col++ >= 0) waddch(win, ' ');

#ifdef DEBUG
	if(Debug & 1024)
	  col += wprintw(win, "#%d", node->JEDTN_SubLength);
#endif

	i = strlen(node->JEDTN_Name);
	if(i >= -col) {
	  int nc;

	  nc = width - 1 - MAX(col, 0);
	  if(nc > i) nc = i;

#if 0
	  if(current != NULL &&
	     (node == current ?
	      line != (edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine) :
	      line == (edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine))) {
	    dbgprintf("GOT %d: %d %d %d\n", node == current,
		      line, edtr->JEDT_CurrentLine, edtr->JEDT_TopScreenLine);
	  }
#endif

	  assert(current == NULL ||
		 (node == current ?
		  line == (edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine) :
		  line != (edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine)));

	  if(node == current) wattron(win, A_STANDOUT);
	  waddnstr(win, node->JEDTN_Name + MAX(-col, 0), nc);
	  col += nc;

	  if(node == current) wattroff(win, A_STANDOUT);
	}

	if(col > maxCols) maxCols = col;

	if(col < 0) col = 0;
	while(++col < width)
	  waddch(win, ' ');
      }
      ++line;

      /*  Move down one line  */
      if(jedtmove(&node, &level, 1) != 1)
	break;
    }
  }

  edtr->JEDT_NCols = maxCols + edtr->JEDT_ColOff;

  if(line < start)
    line = start;

  for( ; line <= end ; ++line) {
    int col;

    wmove(win, edtr->JEDT_Line + line, edtr->JEDT_Column);
    for(col = 0 ; ++col < width ; )
      waddch(win, ' ');
  }

  /*  Update whole scrollbar  */
  if((mode & 1) && (edtr->JEDT_Flags & JEDT_TREEWIN_FSCROLLBAR)) {
    int sbStart, sbEnd;

    _jedsetscrollbar(edtr->JEDT_TopScreenLine, edtr->JEDT_NLines,
		  edtr->JEDT_Height, &sbStart, &sbEnd);

    for(line = 0 ; line < edtr->JEDT_Height ; ++line)
      mvwaddch(win, edtr->JEDT_Line + line,
	       edtr->JEDT_Column + edtr->JEDT_Width - 1,
	       JEDScrollChars[sbStart <= line && line < sbEnd ? 1 : 0]);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int jedtmove(jedtreenode_t **pNode, int *pLevel, int amount)

/************************************************************************/
{
  jedtreenode_t *node, *next;
  int count;

  if((node = *pNode) == NULL)
    return 0;

  if(amount > 0) { /* ****  MOVE DOWN  **** */
    for(count = 0 ; count < amount ; ) {
      int ldiff, add;

      add = 1;
      ldiff = 0;
      if(!(node->JEDTN_Flags & JEDT_TREE_FOPENED)) {
	next = jedtgetnext(node);

      } else if((count + node->JEDTN_SubLength) < amount) {
	/*  Fast track: Just skip the nodes as one  */
	add += node->JEDTN_SubLength;

	next = jedtgetnext(node);

      } else {
	next = jedtgetsubfirstnode(node);

	/*  The case that the node is open but has no children
	 *   is handled below when checking for "istail"
	 */

	++ldiff;
      }

      /*  Check if we have to move up the tree  */
      while(jedtnodeistail(next)) {
	next = jedttail2uppernode(next);
	--ldiff;

	if(jedtistreenode(next))
	  /*  We reached the top level node, so stop  */
	  goto ende;

	next = jedtgetnext(next);
      }

      if(pLevel) *pLevel += ldiff;

      node = next;

      count += add;
    }

  } else { /* ****  MOVE UP  **** */
    for(count = 0 ; count > amount ; --count) {
      next = jedtgetprev(node);

      if(jedtnodeishead(next)) {
	next = jedthead2uppernode(next);

	if(jedtistreenode(next))
	  /*  We reached the top level node, so stop  */
	  break;

	if(pLevel) --*pLevel;
      } else {
	while(next->JEDTN_Flags & JEDT_TREE_FOPENED) {
	  jedtreenode_t *child;

	  if((count - next->JEDTN_SubLength) > amount) {
	    /*  Fast track: Just skip the subnodes as one  */
	    count -= next->JEDTN_SubLength;
	    break;
	  }

	  child = jedtgetsublastnode(next);
	  if(jedtnodeishead(child))
	    /*  node is open, but has no children  */
	    break;

	  next = child;

	  if(pLevel) ++*pLevel;
	}
      }

      node = next;
    }
  }

ende:
  *pNode = node;
  return count;
}

/*JS*********************************************************************
*   mode bit 0: If set, start from the beginning of the tree instead of
*               the current node.
*        bit 1: If set, search backwards.
*************************************************************************/

int jedtreesearch(jedtree_t *edtr, int (*function)(void *, jedtreenode_t *, int, int),
		  void *uptr, int mode)

/************************************************************************/
{
  jedtreenode_t **stack, *node;
  int level, mlevel, linenr, ret;

  stack = NULL;
  mlevel = 0;

  if((mode & 1) || edtr->JEDT_Current == NULL) {
    node = jedtgetfirstnode(edtr);
    level = 0;
    linenr = 0;

  } else {
    int cnt;

    /*  We don't know the current level, so navigate to this  */
    node = edtr->JEDT_TopScreen;
    level = edtr->JEDT_TopScreenLvl;

    for(cnt = edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine ; cnt > 0 ; --cnt)
      if(jedtmove(&node, &level, 1) != 1)
	break;

    linenr = edtr->JEDT_CurrentLine;

    if(level > 0) {
      mlevel += level + 10;
      if((stack = (jedtreenode_t **)malloc(mlevel * sizeof(*stack))) == NULL)
	EGOTO(error);
    }

    for(cnt = 0 ; cnt < level ; ++cnt)
      stack[cnt] = NULL;
  }

  ret = 0;
  for(;;) {
    /* ***  Searching must start with next line after current, so first advance  *** */
    if(mode & 2) { /* ***  Walk backwards  *** */
      node = jedtgetprev(node);

      if(jedtnodeishead(node)) {
	/*  We cannot use 'stack[--level]' here since this might be NULL  */
	node = jedthead2uppernode(node);
	--level;

	if(jedtistreenode(node)) {
	  assert(level < 0);
	  goto ende;
	}

      } else {
	while(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) { /*  Enter node  */
	  if(level >= mlevel) {
	    mlevel += MAX(level, mlevel) + 10;
	    if((stack = (jedtreenode_t **)realloc(stack, mlevel * sizeof(*stack))) == NULL)
	      EGOTO(error);
	  }

	  stack[level++] = node;
	  node = jedtgetsublastnode(node);

	  if(jedtnodeishead(node)) {
	    /*  node is open, but has no children  */
	    node = stack[--level];
	    break;
	  }
	}
      }

      --linenr;

    } else { /* ***  Walk forward  *** */
      if(node->JEDTN_Flags & JEDT_TREE_FSUBNODES) { /*  Enter node  */
	if(level >= mlevel) {
	  mlevel += MAX(level, mlevel) + 10;
	  if((stack = (jedtreenode_t **)realloc(stack, mlevel * sizeof(*stack))) == NULL)
	    EGOTO(error);
	}

	stack[level++] = node;
	node = jedtgetsubfirstnode(node);

      } else
	node = jedtgetnext(node);

      while(jedtnodeistail(node)) {
	/*  We cannot use 'stack[--level]' here since this might be NULL  */
	node = jedttail2uppernode(node);
	--level;

	if(jedtistreenode(node)) {
	  assert(level < 0);
	  goto ende;
	}

	node = jedtgetnext(node);
      }

      ++linenr;
    }

    if((ret = (*function)(uptr, node, linenr, level)) == -1)
      EGOTO(error);

    if(ret) {
      int flag;

      flag = FALSE;
      if(ret & 2) {
	int l;

	for(l = level - 1 ; l >= 0 ; --l) {
	  if(stack[l] && jedtreeopennode(edtr, stack[l], 0) == -1)
	    EGOTO(error);

	  flag = TRUE;
	}
      }

      if(ret & 4) {
	if(!flag && edtr->JEDT_TopScreenLine <= linenr && linenr < edtr->JEDT_TopScreenLine + edtr->JEDT_Height) {
	  edtr->JEDT_Current = node;
	  edtr->JEDT_CurrentLine = linenr;

	} else {
	  int line;

	  /*  Remember last screen line...  */
	  line = edtr->JEDT_CurrentLine - edtr->JEDT_TopScreenLine;

	  edtr->JEDT_TopScreen = node;
	  edtr->JEDT_TopScreenLvl = level;
	  edtr->JEDT_TopScreenLine = linenr;
	  edtr->JEDT_Current = node;
	  edtr->JEDT_CurrentLine = linenr;

	  /*  ... and try to stay on screen where we were  */
	  if(line > 0)
	    edtr->JEDT_TopScreenLine += jedtmove(&(edtr->JEDT_TopScreen), &(edtr->JEDT_TopScreenLvl),
						 -line);
	}
      }

      break;
    }
  }
ende:

  free(stack);

  return ret;
error:
  return -1;
}

/************************************************************************/
#ifdef TEST
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>

static double SleepTime = 0.0;

static int Mode = 1;

static DIR *Dir;
static int Pos, Off;
static int DelMode = 0;

static int setupdir(void *user, jedtree_t *edtr, jedtreenode_t *node)
{
  struct dirent *dent;

  if(DelMode || !(node->JEDTN_Flags & JEDT_TREE_FOPENED)) {
    if(SleepTime > 0) {
      jedtreenode_t *sub;

      sub = jedtgetsubfirstnode(node);
      if(!jedtnodeistail(sub)) {
	/*  Delete tail node  */
	jedtreedelnode(sub, 1);
	dsleep(SleepTime);

	/*  User should see how sub entries vanish  */
	node->JEDTN_Flags |= JEDT_TREE_FOPENED;

	/*  Prepare for another wrap-around  */
	DelMode = 1;
	return 1;
      }

      if(node->JEDTN_Flags & JEDT_TREE_FOPENED) {
	/*  OK, last round, so close now and one more time  */
	node->JEDTN_Flags &= ~JEDT_TREE_FOPENED;
	return 1;
      }
    }

    /*  Ok, deleting subnodes done, we are through  */
    DelMode = 0;
    return 0;
  }

  if(Dir == NULL) {
    /*  First open the directory  */
    Pos = strlen((char *)user);

    ((char *)user)[Pos] = '/';

    if((Off = jedtreegetpath(node, "/", (char *)user + Pos + 1,
			     JS_FILENAME_MAX - Pos - 1)) == -1)
      EGOTO(error);

    Off += Pos + 1;

    if((Dir = opendir((char *)user)) == NULL)
      EGOTO(error);

    ((char *)user)[Off++] = '/';
  }

  while((dent = readdir(Dir)) != NULL) {
    jedtreenode_t *entry;
    const char *parts[2];
    struct stat stbuf;

    if(dent->d_name[0] == '.' &&
       (!(Mode & 2) ||
	dent->d_name[1] == '\0' ||
	(dent->d_name[1] == '.' && dent->d_name[2] == '\0')))
      continue;

    parts[0] = dent->d_name;
    parts[1] = NULL;

    if((entry = jedtreeaddnodep2(NULL, node, parts, 4 | 8)) == NULL)
      EGOTO(error);

    strcpy((char *)user + Off, dent->d_name);
    if(stat((char *)user, &stbuf) == -1)
      EGOTO(error);

    if(S_ISDIR(stbuf.st_mode)) {
      entry->JEDTN_Flags |= JEDT_TREE_FCALLBACK | JEDT_TREE_FSUBNODES;
    }

    if(SleepTime > 0) {
      /*  Ok, only one more entry, than prepare for another round  */
      dsleep(SleepTime);

      return 1;
    }
  }

  if(closedir(Dir) == -1)
    EGOTO(error);

  /*  Clean up  */
  Dir = NULL;

  if(SleepTime == 0)
    /*  No need to call us over and over again  */
    node->JEDTN_Flags &= ~JEDT_TREE_FCALLBACK;

  ((char *)user)[Pos] = '\0';

  return 0;
error:
  ((char *)user)[Pos] = '\0';
  Dir = NULL;
  return -1;
}

static int showNode(jedtree_t *edtr)
{
  jedtreenode_t *node;
  int level;

  node = jedtgetfirstnode(edtr);
  for(level = 0 ;; ) {
    printf("%*s>'%s'\n", level, "", node->JEDTN_Name);

    /*  Move down  */
    if(jedtmove(&node, &level, 1) != 1)
      break;
  }

  return 0;
}

int main(int argc, char *argv[])
{
  jedtree_t *edtr;
  char path[JS_FILENAME_MAX];

  while(argc > 1 && argv[1][0] == '-') {
    if(argv[1][1] == 'a') { /*  animate  */
      SleepTime = 0.05;
    } else if(argv[1][1] == 's') { /*  only show  */
      Mode &= ~1;
    } else {
      fprintf(stderr, "ERROR: Invalid option '%s'!\n", argv[1]);
      return 5;
    }

    --argc; ++argv;
  }

  if(argc != 2) goto usage;

  strcpy(path, argv[1]);

  if(Mode & 1) {
    jedinitialize(0);
    curs_set(0);

    if((edtr = jedtreecreate(NULL, path, 4, 4, -20, -8,
			     JEDT_TREEWIN_FFRAME | JEDT_TREEWIN_FSCROLLBAR)) == NULL)
      EGOTO(error);
  } else {
    if((edtr = jedtreecreate(NULL, path, 2, 2, 20, 80, 0)) == NULL)
      EGOTO(error);
  }

  edtr->JEDT_Setup = setupdir;
  edtr->JEDT_User = path;

  /*  This will trigger the 'setupdir' function when the
   *   node is first printed
   */
  edtr->JEDT_Flags |= JEDT_TREE_FCALLBACK;

  if(Mode & 1) {
    int i, line, selFlag;

    line = edtr->JEDT_Line + edtr->JEDT_Height + 2;

    jedtreedokey(edtr, 0, 0, 0); /*  redraw  */
    selFlag = TRUE;
    while((i = wgetch(edtr->JEDT_Window)) != 'q') {
      if(i == '\t') {
	if(jedtreedokey(edtr, selFlag ? 'I' : 0, 1, 1) == 0)
	  beep();

	selFlag = !selFlag;
	continue;
      }

      if(i == '\n' || i == '\r' || i == KEY_ENTER) {
	if(edtr->JEDT_Current != NULL) {
	  char cpath[200];

	  if(jedtreegetpath(edtr->JEDT_Current, "/", cpath, sizeof(cpath)) == -1)
	    EGOTO(error);

	  mvprintw(line++, 1, "SIZE %d/%d %d CURR %d %d; Path '%s'\n",
		   edtr->JEDT_TopScreenLine, edtr->JEDT_NLines,
		   edtr->JEDT_Height, edtr->JEDT_CurrentLine,
		   edtr->JEDT_Current, cpath);

	  if(line == LINES - 1) line = edtr->JEDT_Line + edtr->JEDT_Height + 2;

	  refresh();
	} else
	  beep();
	continue;
      }

      if(jedtreedokey(edtr, i, 1, 1) == 0)
	beep();
    }

    jeddeinitialize();
  } else {
    showNode(edtr);
  }

  jedtreedelete(edtr);

  return 0;
usage:
  fprintf(stderr, "Usage: testedtree [<opts>] <directory\n");
  return 5;
error:
  jsperror("testedtree");
  return 30;
}
#endif
