/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Preprocessor and parser
*
*    Part    : Header file.
*
*              Copyright (C) 2003-2005 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

#ifndef JSCPP_H
#define JSCPP_H  1

/*********     INCLUDES 					*********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <jscpplib.h>

#include <jsparse.h>

#ifdef CONFIG_NO_STRDUP
# define strdup  mystrdup  /*  strdup is not POSIX conformant  */
#endif

/*********     DEFINES						*********/
/* ----------- FILE: jscpp.c              ------------ */
/* ***  Close streams  *** */
#define closePreProcStream(jps, flag)     jsppclose(jps, flag)
#define closeParserStream(jpa, flag)      jsparsclose(jpa, flag)

#define EXTFLAG_DOUBLEASSIGN              (JPAM_FUSERFLAGS << 0)
#define EXTFLAG_INSERTCAST                (JPAM_FUSERFLAGS << 1)
#define EXTFLAG_INSERTCAST2               (JPAM_FUSERFLAGS << 2)


/* ----------- FILE: jsfctsep.c           ------------ */
#define JSCPP_MODE_FCTSEP      "fctsep"

/* ----------- FILE: jsfcttoc.c           ------------ */
#define JSCPP_MODE_PROTO      "proto"
#define JSCPP_MODE_FCTTOC     "fcttoc"

/* ----------- FILE: jsfdeps.c            ------------ */
#define JSCPP_NAME_FDEPS        "jsfdeps"
#define JSCPP_MODE_FDEPS        "fdeps"

/* ***  Various modes for doFDeps function  *** */
/*  Use jsmakedeps output style  */
#define JSCPP_FDEPS_STYLEDEPS     (1 << 0)
/*  Use topological sort output style  */
#define JSCPP_FDEPS_STYLETSORT    (1 << 1)
/*  Hide definitions from include file */
#define JSCPP_FDEPS_HIDEINC       (1 << 2)

/*  Also include dependencies on global/static variables  */
#define JSCPP_FDEPS_VARDEPS       (1 << 3)

/*  Include static functions and variables as well  */
#define JSCPP_FDEPS_STATICDEF     (1 << 4)

/*  Prefix static variables/functions with the function name to
 *   make them globally unique
 */
#define JSCPP_FDEPS_STATICPREFIX  (1 << 5)

/*  Global Variable names are prefixed with ''  */
#define JSCPP_FDEPS_VARPREFIX     (1 << 6)

/*  When showing called functions, do not avoid duplicates and do not sort  */
#define JSCPP_FDEPS_DUPLICATES    (1 << 7)

/*  Also show functions that start with __ or _[A-Z]  */
#define JSCPP_FDEPS_SHOWALL       (1 << 8)

/*  Exclude ANSI symbols from dependencies  */
#define JSCPP_FDEPS_ANSISYMS      (1 << 9)

/* ----------- FILE: jsftrace.c           ------------ */
#define JSCPP_FTRACE_SYMBOL     "__JSFTRACE__"

/*  When called like this, all functions are traced as default  */
#define JSCPP_NAME_FTRACE       "jsftrace"
#define JSCPP_MODE_FTRACE       "ftrace"

/*  When called like this, nothing is traced/changed as default  */
#define JSCPP_NAME_TRACE        "jstrace"
#define JSCPP_MODE_TRACE        "trace"

/*  When called like this, all functions are traced by wrapping them  */
#define JSCPP_NAME_FCTWRAP      "jsfctwrap"
#define JSCPP_MODE_FCTWRAP      "fctwrap"

/* ***  Various modes for doFTrace function  *** */
/*  Emit trace call on start of function  */
#define JSCPP_TRACE_FCTCALL           (1 <<  0)
/*  Start trace calls contain functions arguments  */
#define JSCPP_TRACE_FCTCALLARGS       (1 <<  1)
/*  Function arguments are labeled as default  */
#define JSCPP_TRACE_FCTCALLLBLD       (1 <<  2)

/*  Emit trace call on return statements  */
#define JSCPP_TRACE_FCTRET            (1 <<  4)
/*  The "return" trace calls contain return value  */
#define JSCPP_TRACE_FCTRETARG         (1 <<  5)
/*  Label return values as well  */
#define JSCPP_TRACE_FCTRETLBLD        (1 <<  6)
/*  The "return" trace calls contain the line  */
#define JSCPP_TRACE_FCTRETLINE        (1 <<  7)

/*  Produce trace calls for each statement  */
#define JSCPP_TRACE_STMTTRACE         (1 << 12)
/*  Produce trace calls only for statements where flow branches  */
#define JSCPP_TRACE_STMTTRACE_BRANCH  (1 << 13)

/*  Produce code coverage  */
#define JSCPP_TRACE_CODECOVERAGE      (1 << 14)

/*  Generate wrapper functions for each function call and redirect
 *   them
 */
#define JSCPP_TRACE_FCTWRAP           (1 << 16)

/*  To recognize that wrapper generator has to be called  */
#define JSCPP_TRACE_NEEDWRAPPING      (1 << 17)

/*  Use internal mapping of types to trace format  */
#define JSCPP_TRACE_INTTYPEMAPPING    (1 << 20)

/*  Print prototype of trace function at the start  */
#define JSCPP_TRACE_PROTOTYPEFCT      (1 << 21)
/*  Print prototype for variablel trace function  */
#define JSCPP_TRACE_PROTOTYPEVAR      (1 << 22)
/*  Print prototype for statement trace function  */
#define JSCPP_TRACE_PROTOTYPESTMT     (1 << 23)
/*  Print prototype for code coverage function + array  */
#define JSCPP_TRACE_PROTOTYPECOV      (1 << 24)

/*  Yield preprocessed tokens instead of original ones  */
#define JSCPP_TRACE_OUTPREPROC        (1 << 25)
/*  Expand system include files  */
#define JSCPP_TRACE_EXPUSERINCS       (1 << 26)
/*  Expand macros defined in user include files  */
#define JSCPP_TRACE_EXPUSERMACS       (1 << 27)

/*  outName points to a directory  */
#define JSCPP_TRACE_DIRECTORY         (1 << 29)

/*  Generate 'line' directive at the start to identify sources
 *   correctly
 */
#define JSCPP_TRACE_LINEDIRECTIVE     (1 << 30)

/* ********************************************************** */
/*  If this is defined and we have a trace pragma with explicit expressions
 *   and no label is given, the expression is used as a label.
 *   E. g. "call(arg * 5)" will use the label "arg * 5".
 *   If you want to suppress a label, specify an empty one like ""
 */
#define CONFIG_JSPARSE_ARGLABELS

typedef struct {
  const char *Label;
  const char *Expression;

  /*  Contains the tokens to read in the expression  */
  jsqueuestream_t *QStream;
} argdesc_t;

typedef struct {
  int Type;

  const char **Patterns;
  int NPatterns;

  argdesc_t *CallDescriptions;
  int NCallDescriptions; /*  -1: patch in the standard way  */

  argdesc_t *ReturnDescriptions;
  int NReturnDescriptions; /*  -1: patch in the standard way  */
} patchdesc_t;

typedef struct {
  const char *Pattern;
  int Level;
} tracelevel_t;

/*  Maps one expression type to an expression  */
typedef struct {
  jsparsexprtype_t ExprType;

  argdesc_t *Descriptions;
  int NDescriptions;
} typemapping_t;

typedef struct {
  const char *TypeName;
  const char *SourceName, *TargetName;

} typerename_t;

/*  Maps a given type to a character for the format string of the
 *   trace function
 */
typedef struct {
  /*
   *  Bit 0: ExprType not used, this format is for pointers
   *  Bit 1: Never use a label.
   *  Bit 2: Always use a label.
   */
  int FFlag;

  jsparsexprtype_t ExprType;

  const char *Format[2];
} typeformat_t;

typedef struct {
  const char *FuncPattern;
  const char *VarPattern;

  long Flag; /*  bit 0: trace read access.
	      *  bit 1: trace write access.
	      *  bit 8: only global variables matching the pattern.
	      *  bit 9: only local variables matching the pattern.
	      */
} vartrace_t;

typedef struct {
  const char *Name;
  int Flag;
} wrapfunc_t;

#define VARTRACE_FREAD      1
#define VARTRACE_FWRITE     2
#define VARTRACE_FGLOBAL  256
#define VARTRACE_FLOCAL   512

#define PDESC_TYPE_TRACE              1
#define PDESC_TYPE_MASK            0xff

#define PDESC_TYPE_FCALL          (1 <<  8)
#define PDESC_TYPE_FCALLLBL       (1 <<  9)

#define PDESC_TYPE_FRETURN        (1 << 12)
#define PDESC_TYPE_FRETURNLINE    (1 << 13)
#define PDESC_TYPE_FRETURNLBL     (1 << 14)

#define PDESC_TYPE_FALLSTMT       (1 << 16) /*  trace all statements  */
#define PDESC_TYPE_FBRANCHSTMT    (1 << 17) /*  trace only after branches  */

#define PDESC_TYPE_FCODECOVERAGE  (1 << 18) /*  trace only after branches  */

#define PDESC_TYPE_FFCTWRAP       (1 << 20)

typedef struct {
  int Mode;
  int PrototypeFlag; /*  marks which prototypes were generated  */

  const char *CurrentFile;

  /*  If non-zero, use proper indentation for trace statements,
   *   possibly inserting newlines
   */
  int Indentation;

  int MaxMappingLevel;
  int MaxRecursiveLevel;

  jsparscnode_t **NodeStack;
  int INodeStack, MNodeStack;

  int *ExpandMapStack;
  int IExpandMapStack, MExpandMapStack;

  patchdesc_t *Descriptions;
  int NDescriptions, MDescriptions;

  wrapfunc_t *WrappedFunctions;
  int NWrappedFunctions, MWrappedFunctions;

  tracelevel_t *TraceLevels;
  int NTraceLevels, MTraceLevels;

  typemapping_t *TypeMappings;
  int NTypeMappings, MTypeMappings;

  typerename_t *TypeRenames;
  int NTypeRenames, MTypeRenames;

  typeformat_t *TypeFormats;
  int NTypeFormats, MTypeFormats;

  const char *TraceFormatCall;
  const char *TraceFormatReturn;
  const char *TraceFormatStmt;

  const char *NameRetVariable;
  const char *TraceVarFunction;

  const char *LabelFormat;

  vartrace_t *VarTraces;
  int NVarTraces, MVarTraces;

  const char *PrefixWrapFunction;
  const char *PrefixWrapFunction2;

  int CurrentVarNumber;

  /*  For code coverage  */
  int *CovRange;
  int NCovRange, MCovRange;

} traceaux_t;

#define CALLTRACE_FUNCTION   "_jstracefunc"


/* ----------- FILE: jsftrace2.c          ------------ */
typedef struct {
    int Start, Curr, End;
} tokenseq_t;

#define tokseqrewind(tokseq)    ((tokseq)->Curr = (tokseq)->Start)

/* ----------- FILE: jsglobvars.c         ------------ */
#define JSCPP_MODE_GLOBVARS     "globvars"

/* ----------- FILE: jsptrdeps.c          ------------ */
#define JSCPP_MODE_PTRDEPS     "ptrdeps"

/*  Do also try to add a const for the parameter itself (not yet
 *   working)
 */
#define JSCPP_PTRDEP_ARGCONST         (1 << 0)

/*  Only try to add const to the specifier  */
#define JSCPP_PTRDEP_SPECONLY         (1 << 1)

/*  Print changed routines on stdout  */
#define JSCPP_PTRDEP_VERBOSE          (1 << 2)

#define JSCPP_PTRDEP_NUMFORMAT        (1 << 3)

/* ----------- FILE: jsstmtstep.c         ------------ */
#define JSCPP_MODE_STMTSTEP        "stmtstep"
#define JSCPP_MODE_STMTSTEP0       "stmtstep0"
#define JSCPP_MODE_CODECOVER       "codecov"

#define STMTTRACE_FUNCTION         "_jstracestmt"
#define STMT_FILENAME_VAR          "_jstracefname"

#define CODECOVER_REGFUNCTION      "_jscovregister"
#define CODECOVER_COUNTARRAY       "_jscovcounts"
#define CODECOVER_DESCRIPTARRAY    "_jscovdescript"

/* ----------- FILE: jsunused.c           ------------ */
#define JSCPP_MODE_PPUNUSED     "ppunused"


/* *******   Prototypes of library functions  ***** */
/*  For C++ compatibility   */
# ifdef __cplusplus
#  define Prototype extern "C"
# else
#  define Prototype extern
# endif
/* ***** File: jscpp.c               ***** */
Prototype FILE          *openFile(const char *directory, const char *name);
Prototype jstrbuff_t     TempBuffer;

/* ***** File: jsfctsep.c            ***** */
Prototype int            doFctSep(jsparsenviron_t *ppenv,
				  const char *names[],
				  int nnames, FILE *fpOut, int mode);
Prototype const char    *getFunctionName(jsparscnode_t *node,
					 jctokfifo_t *queue);

/* ***** File: jsfcttoc.c            ***** */
Prototype int            doFctToc(jsparsenviron_t *ppenv,
				  const char *names[],
				  int nnames, FILE *fpOut, int mode);

/* ***** File: jsfdeps.c             ***** */
Prototype int            doFDeps(jsparsenviron_t *ppenv, const char *names[],
				 int nnames, FILE *fpOut, int depsMode);

/* ***** File: jsftrace.c            ***** */
Prototype int            doFTrace(jsparsenviron_t *ppenv, const char *names[],
				  int nnames, const char *outName, int mode);
Prototype void           freepatchdesc(patchdesc_t *pdesc);
Prototype jstrbuff_t      SBuffTypes, SBuffArgs;
Prototype int             DefaultIndentation;

/* ***** File: jsftrace1.c           ***** */
Prototype int            doProcFTrace(traceaux_t *paux, jsparsstream_t *jpa,
                                      jsparscnode_t *node, jctokfifo_t *queue);
Prototype int            doProcTypeRename(traceaux_t *paux, jsparsstream_t *jpa,
                                          jsparscnode_t *node, jctokfifo_t *queue);
Prototype int            doGetName(traceaux_t *paux, jsparscnode_t *node,
                                   const char **pName, jctokfifo_t *queue,
                                   int mode);
Prototype int            checkPattern(const char *pattern,
                                      const char *fileName,
                                      const char *funcName);
Prototype char           *genTraceFctCall(traceaux_t *paux,
                                          const patchdesc_t *pdesc,
                                          jctokfifo_t *queue, jsparscnode_t *node,
                                          const char *format, const char *marker,
                                          const char *funcname,
                                          const char *types, const char *args,
                                          const char *prefix[]);

/* ***** File: jsftrace2.c           ***** */
Prototype int            doPragma(jsparsstream_t *jpa, traceaux_t *paux,
				  jsparscnode_t *node, jctokfifo_t *queue,
				  int mode);
Prototype int            findTraceLevel(traceaux_t *paux,
					const char *funcname);
Prototype int            findTypeRename(traceaux_t *paux, const char *typeName,
                                        const char *sourceName);

/* ***** File: jsftrace3.c           ***** */
Prototype int            doPatchStmt(traceaux_t *paux,
				     jsparsstream_t *jpa,
				     jsparsvardef_t *fctDef,
				     jsparscnode_t *node,
				     const patchdesc_t *pdesc,
				     int mode);

/* ***** File: jsftrace4.c           ***** */
Prototype int            genTraceStmt(traceaux_t *paux,
				      jsparsstream_t *jpa,
				      jsparsvartab_t *vTab,
				      jstrbuff_t *sbuffTypes,
				      jstrbuff_t *sbuffArgs,
				      argdesc_t *desc, int n,
				      const char *name,
				      const char *defLabel,
				      int mode);
Prototype int            genTracePart(traceaux_t *paux, jsparsstream_t *jpa,
				      jsparsvartab_t *vTab,
				      jstrbuff_t *sbuffTypes,
				      jstrbuff_t *sbuffArgs,
				      const char *label,
				      const char *expression,
				      jsparsexprtype_t *pType,
				      int *pUseLabel, int mode);
Prototype const char    *getTraceFormat(traceaux_t *paux,
					jsparsstream_t *jpa,
					jsparsvartab_t *vTab,
					jsparscnode_t *decl,
					jsparscnode_t *spec,
					jctokfifo_t *queue,
					int *pUseLabel, int mode);

/* ***** File: jsftrace5.c           ***** */
Prototype int            readPatterns(patchdesc_t *pdesc, int start, int end,
				      jctokfifo_t *queue);
Prototype int            readArgs(traceaux_t *paux, jsparsstream_t *jpa,
				  argdesc_t **pList, int *pNList,
				  int start, int end, jctokfifo_t *queue,
				  int mode);
Prototype jsparsvartab_t*readExprType(jsparsstream_t *jpa, jctokfifo_t *queue,
				      int start, int end,
				      jsparsexprtype_t *eType);
Prototype jctokfifo_t   *copyToks(jsparsstream_t *jpa, jctokfifo_t *queue,
				  int *pStart, int *pEnd);

/* ***** File: jsglobvars.c          ***** */
Prototype int            doGlobVars(jsparsenviron_t *ppenv,
				    const char *names[],
				    int nnames, FILE *fpOut, int mode);

/* ***** File: jsideps.c             ***** */
Prototype int            doIDeps(jsppenviron_t *ppenv, const char *names[],
				 int nnames, FILE *fpOut);

/* ***** File: jspreproc.c           ***** */
Prototype jsppstream_t  *getPreProcStream(jsppenviron_t *jppenv,
                                          const char *name, int mode);
Prototype jsparsstream_t *getParserStream(jsparsenviron_t *ppenv,
                                          const char *name, int omode, int mode);
Prototype int            doPreProcess(jsppenviron_t *ppenv, const char *names[],
                                      int nnames, const char *outName, int mode);

/* ***** File: jsptrdeps.c           ***** */
Prototype int            doPtrDeps(jsparsenviron_t *ppenv,
				   const char *names[],
				   int nnames, FILE *fpOut, int mode);

/* ***** File: jsstmtstep.c          ***** */
Prototype int            doStmtTracing(traceaux_t *paux,
				       const patchdesc_t *pdesc,
				       jsparsstream_t *jpa,
				       jsparscnode_t *tree,
				       jctokfifo_t *queue,
				       const char *funcName);
Prototype int            checkAssignment(jsparsstream_t *jpa,
					 jctokfifo_t *queue,
					 jsparscnode_t *node);
Prototype int            tokInsertString(jsparsstream_t *jpa,
					 jsparscnode_t *tree,
					 jctokfifo_t *queue, int pos,
					 const char *string,
					 int mode);

/* ***** File: jsunused.c            ***** */
Prototype int            doPPUnUsed(jsppenviron_t *ppenv, const char *names[],
				    int nnames, FILE *fpOut, int mode);

#endif
