/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor.
 *    Part    : File subroutines.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include <string.h>
#include <stdlib.h>

#include <errno.h>

#define JSCPP_INTERNAL
#include <jscpplib.h>

/*********     DEFINES                                          *********/
#ifndef IF_FLAG_MACRO
/*  START-DEFINITIONS 1 */
/* ***  Structure to store open file  *** */
#define IF_FLAG_EOF            (1 <<  0)
#define IF_FLAG_MACRO          (1 <<  1)
#define IF_FLAG_PPSTREAM       (1 <<  2)
#define IF_FLAG_NOEXP          (1 <<  3)
#define IF_FLAG_ORIGTOK        (1 <<  4)
#define IF_FLAG_COPYTOKS       (1 <<  5)
#define IF_FLAG_EXPANDORIG     (1 <<  6)
#define IF_FLAG_EXPANDORIGFULL (1 <<  7)

#define IF_FLAG_EXPANDINGARG   (1 <<  8)
#define IF_FLAG_DISABLEEXPAND  (1 <<  9)
#define IF_FLAG_LINETOKENS     (1 << 10)

#ifdef JSCPP_INTERNAL
#ifdef CONFIG_JSCPP_OPTINC
/*  Description:
 *   NONSPC1:    Used internally to set NONSPC *after* a non-space
 *               token has been read.
 *   NONSPC:     Marks if a non-space token has been read in the file.
 *   INCONSTRUCT: Within '#ifndef' .. 'endif' construct.
 *   CONSTRUCT:  '#ifndef SYMBOL' has been seen in file with solely
 *               blanks prior to '#ifndef'.
 */
# define IF_FLAG_NONSPC1      (1 << 12)
# define IF_FLAG_NONSPC       (1 << 13)
# define IF_FLAG_INCONSTRUCT  (1 << 14)
# define IF_FLAG_CONSTRUCT    (1 << 15)
#endif
#endif

#define IF_SYSINCLUDE         (1 << 16)

#ifdef JSCPP_INTERNAL
#define IF_ISMACRO(iFile)     ((iFile)->IF_Flag & IF_FLAG_MACRO)
#define IF_ISFILE(iFile)      (!IF_ISMACRO(iFile))
#define IF_ISATEOF(iFile)     ((iFile) == NULL ||			\
                               ((iFile)->IF_Flag & IF_FLAG_EOF))
#endif

typedef struct {
    jsctoken_t *MA_Token;
    int         MA_UseCount;
} _jsppmacarg_t;

/* **  For an input channel  ** */
typedef struct jsppfile_t jsppfile_t;
struct jsppfile_t {
    jsppfile_t *IF_Next;
    int         IF_Flag;
    const char *IF_Name;   /*  The file name or the macro identifier  */
    union {
	extfile_t       *if_FFPtr;
	jsctoken_t      *if_MTokens; /*  Tokens for current macro  */
	jsqueuestream_t *if_QStream;
#define IF_FFPtr    if_c1.if_FFPtr
#define IF_MTokens  if_c1.if_MTokens
#define IF_QStream  if_c1.if_QStream
    } if_c1;
    union {
	int         if_LineNr; /*  Contains corrected line number  */
	_jsppmacarg_t *if_MArgs;    /*  Arguments of current macro  */
#define IF_LineNr   if_c2.if_LineNr
#define IF_MArgs    if_c2.if_MArgs
    } if_c2;

    int IF_CondLevel;

#ifdef CONFIG_JSCPP_OPTINC
    const char *IF_Symbol;
#endif
};

/*  For include files  */
typedef struct {
    const char *IN_Directory;
    int         IN_Flags;
} jsppincdir_t;

#define JPP_FILE_FSYSINC           (1 << 0)
#define JPP_FILE_FNODIRSEP         (1 << 1)
#define JPP_FILE_FADDTOTOP         (1 << 2) /*  is cleared after adding at the front  */
#define JPP_FILE_FALLOCATED        (1 << 3)

/* **  For routine jspppushfile  ** */
#define JPP_PUSHFILE_FNOCURRDIR    (1 << 0)
#define JPP_PUSHFILE_FINCNEXT      (1 << 1)
#define JPP_PUSHFILE_FSINCLUDE     (1 << 2)
#define JPP_PUSHFILE_FNOLINEDIR    (1 << 3)

#define jspppushtokens(jps, tokens)    _jspppushmacro(jps, NULL, tokens, NULL, 0)
/*  END-DEFINITIONS  */
#endif

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            _jspppushmacro(jsppstream_t *jps, const char *name,
                                        jsctoken_t *pTokens,
                                        _jsppmacarg_t *args, int flags);

Prototype int            jspppushfile(jsppstream_t *jps, const char *name,
                                      int mode);
Prototype int            jspppushppstream(jsppstream_t *jps,
                                          jsqueuestream_t *jsqs, int mode);

Prototype int            jspppushinplevel(jsppstream_t *jps);
Prototype int            jspppopinplevel(jsppstream_t *jps);

IPrototype int           _jspppopistack(jsppstream_t *jps);
static int               dopopfile(jsppstream_t *jps, jsppfile_t *ifile);
static int               dopopmacro(jsppstream_t *jps, jsppfile_t *ifile);

static int               pushLineDir(jsppstream_t *jps, int lineNr,
                                     const char *fileName);

Prototype int            jsppaddincdir(jsppstream_t *jps, const char *name,
                                       int flags);
Prototype int            jsppaddincpath(jsppstream_t *jps,
                                        const char *includepath, int flags);
Prototype void           jsppshowinc(jsppstream_t *jps, FILE *fp, int mode);

Prototype int            jsppgetfilepos(jsppstream_t *jps,
                                        const char **pFileName, int mode);

#ifdef CONFIG_JSCPP_OPTINC
static int               saveOnceFile(jsppstream_t *jps, const char *name,
                                      const char *symbol);
static int               checkOnceFile(jsppstream_t *jps, const char *name);
Prototype void           jsppfreeoncefl(jsppstream_t *jps);
#endif

/*JS*********************************************************************
 *   Push the expansion of macro 'name' on the input stack. pTokens
 *    contains a list of tokens ended with a CTK_EOF, args is a list of
 *    current arguments, starting with args[1] .. args[n], args[0] may
 *    contain a pointer that is freed when the macro expansion is popped.
 *    Every argument is terminated in the same fashion. The name is stored
 *    to prevent recursive expansion of this macro during reading of the
 *    token sequence. If name = "", then macro expansion will be disabled
 *    completely.
 *************************************************************************/

int _jspppushmacro(jsppstream_t *jps, const char *name, jsctoken_t *pTokens,
                   _jsppmacarg_t *args, int flags)

/************************************************************************/
{
    jsppfile_t *newDef;
    int size, count;

#if 0
    {
	int i;

	printf("PTOKENS %d flags %d\n", pTokens, flags);
	for(i = 0 ; pTokens[i].TK_Value != CTK_EOF ; ++i)
	    printf(" %d: %d '%s'\n", i,
		   pTokens[i].TK_Value, pTokens[i].TK_String ?
		   pTokens[i].TK_String : "<NULL>");
    }
#endif

    size = sizeof(*newDef);
    count = 0;
    if(flags & IF_FLAG_COPYTOKS) {
	while(pTokens[count].TK_Value != CTK_EOF)
	    ++count;
	++count;

	size += count * sizeof(*pTokens);
    }

    if((newDef = (jsppfile_t *)malloc(size)) == NULL)
	return -1;

    if(count > 0) {
	jsctoken_t *pTokensNew;

	pTokensNew = (jsctoken_t *)&newDef[1];
	while(--count >= 0) pTokensNew[count] = pTokens[count];

	pTokens = pTokensNew;
    }

    /*  Preserve this flag even in macro expansions  */
    if(jps->JSPP_File != NULL && (jps->JSPP_File->IF_Flag & IF_SYSINCLUDE))
	flags |= IF_SYSINCLUDE;

    /*  Preset structure  */
    newDef->IF_Flag    = IF_FLAG_MACRO | flags;
    newDef->IF_Name    = name;
#ifdef CONFIG_JSCPP_OPTINC
    newDef->IF_Symbol  = NULL;
#endif
    newDef->IF_MTokens = pTokens;
    newDef->IF_MArgs   = args;

    /*  We have to keep this flag if previous levels have set it as well  */
    if(jps->JSPP_File) {
	if(jps->JSPP_File->IF_Flag & IF_FLAG_EXPANDORIGFULL)
	    newDef->IF_Flag |= IF_FLAG_EXPANDORIGFULL;

	if((flags & IF_FLAG_EXPANDINGARG) &&
	   (jps->JSPP_File->IF_Flag & IF_FLAG_EXPANDORIG))
	    newDef->IF_Flag |= IF_FLAG_EXPANDORIG;
    }

    /*  Push on stack  */
    newDef->IF_Next = jps->JSPP_File;
    jps->JSPP_File = newDef;

    return 0;
}
/*JS*********************************************************************
 *   Push the input file on stack of inclusion.
 *    Parameter mode:
 *      JPP_PUSHFILE_NOCURRDIR:
 *        Look only in directory list, not in directory of previous included
 *        file.
 *
 *      JPP_PUSHFILE_INCNEXT:
 *        Do command "include_next": Keep scanning in directory list, if a file
 *        is already on the input stack.
 *
 *      JPP_PUSHFILE_SINCLUDE:
 *        Silent include. Do not report any errors when attempting to open
 *        file.
 *
 *      JPP_PUSHFILE_NOLINEDIR:
 *        Do not print a line directive at the start of the file.
 *
 *************************************************************************/

int jspppushfile(jsppstream_t *jps, const char *name, int mode)

/************************************************************************/
{
    jsppfile_t *newDef;
    char fName[JS_FILENAME_MAX + 1];
    extfile_t *eFp;
    int len, flag;

    newDef = NULL;
    flag = 0;

    if(name) {
	int i, ilast, saveErrno, prevFlag;
	jsppfile_t *iPtr;

	eFp = NULL;

	prevFlag = 0;
	/*  Get name of current include file  */
	for(iPtr = jps->JSPP_File ; !IF_ISATEOF(iPtr) ; iPtr = iPtr->IF_Next)
	    if(IF_ISFILE(iPtr)) {
		prevFlag = iPtr->IF_Flag;
		break;
	    }

	/*  Scan all directories  */
	saveErrno = errno;
	ilast = jps->JSPP_NIncDir;

	/*  We break the strong rule for system includes. That allows
	 *   us to overwrite system includes if we specify
	 *       -I -i/mydir/patchincludes -i/usr/include
	 *   on the command line.
	 */
	if(prevFlag & IF_SYSINCLUDE)
	    ilast++;

	for(i = (mode & JPP_PUSHFILE_FNOCURRDIR) ? 0 : -1 ;
	    i < ilast ; i++) {
	    if(name[0] == CONFIG_DIR_DELIMITER) {
		/*  absolute filename  */
		strcpy(fName, name);

	    } else if(i == -1 || i == jps->JSPP_NIncDir) {
		const char *lastFile, *string;

		if(iPtr && (string = strrchr(lastFile = iPtr->IF_Name,
					     CONFIG_DIR_DELIMITER))) {
		    /*  open in directory of previous include file  */
		    sprintf(fName, "%.*s%c%s", (int)(string - lastFile), lastFile,
			    CONFIG_DIR_DELIMITER, name);

		} else {
		    /*  open in current directory  */
		    strcpy(fName, name);
		}

	    } else {
		unsigned int l = strlen(jps->JSPP_IncDir[i].IN_Directory);

		/*  Avoid another '/' if already at the end of the include dir  */
		if(jps->JSPP_IncDir[i].IN_Directory[l - 1] == CONFIG_DIR_DELIMITER ||
		   (jps->JSPP_IncDir[i].IN_Flags & JPP_FILE_FNODIRSEP))
		    sprintf(fName, "%s%s", jps->JSPP_IncDir[i].IN_Directory, name);
		else
		    sprintf(fName, "%s%c%s", jps->JSPP_IncDir[i].IN_Directory,
			    CONFIG_DIR_DELIMITER, name);
	    }

	    /*  Strip any '\\\n' in file before opening  */
	    _jspprmlinecont(fName);

	    if((eFp = efopen(fName, "r")) != NULL) {
#ifdef CONFIG_JSCPP_INCNEXT
		if(mode & JPP_PUSHFILE_FINCNEXT) {
		    /*  Look in current file stack, compare with fName
		     *   and *not* with FOpenName.
		     */

		    for(newDef = jps->JSPP_File ; !IF_ISATEOF(newDef) ;
			newDef = newDef->IF_Next)
			if(newDef->IF_Name != NULL && tkStrcmp(newDef->IF_Name, fName) == 0)
			    break;

		    if(newDef) {
			/*  This file already opened  */
			newDef = NULL;
			if(efclose(eFp) == -1)
			    EGOTO(error);

			eFp = NULL;
			continue; /*  continue looking for file  */
		    }
		}
#endif

		name = fName;

		/*  Only treat as system include file if included
		 *   via #include <file> or included from a system
		 *   include file itself.
		 */
		if((i >= 0 && (jps->JSPP_IncDir[i].IN_Flags & JPP_FILE_FSYSINC) &&
		    (mode & JPP_PUSHFILE_FNOCURRDIR)) ||
		   (prevFlag & IF_SYSINCLUDE))
		    flag |= IF_SYSINCLUDE;

		break;

	    }

	    if(name[0] == CONFIG_DIR_DELIMITER)
		/*  Break if opening of absolute pathname failed  */
		break;

	    JSCLEARERROR();
	}

	if(eFp == NULL) {
	    if(mode & JPP_PUSHFILE_FSINCLUDE) {
		errno = saveErrno;
		return 0;
	    }

	    jsppprinterr(jps, "Cannot open file \"%s\" %d", 1, name, mode);
	    EGOTO(error);
	}
#ifdef CONFIG_JSCPP_OPTINC
	/*  We must not check once-include files if we are doing strange stuff
	 *   with original token in included files.
	 */
	else if(!(jps->JSPP_Mode & JPPM_EXPUSERFILES) &&
		checkOnceFile(jps, name)) {
	    /*  Don't include file multiple times  */
	    efclose(eFp);
	    return 0;
	}
#endif

	errno = saveErrno;
	len = strlen(name) + 1;
    } else {
	/*  Read from stdin  */
	len = 0;
	if((eFp = efopen(NULL, "r")) == NULL)
	    EGOTO(error);
    }

    if((newDef = (jsppfile_t *)malloc(sizeof(*newDef) + len)) == NULL)
	EGOTO(error);

    newDef->IF_FFPtr = eFp;
    if(name) {
	newDef->IF_Name = (char *)&newDef[1];
	strcpy((char *)&newDef[1], name);
    } else {
	newDef->IF_Name = "<stdin>";
    }
#ifdef CONFIG_JSCPP_OPTINC
    newDef->IF_Symbol  = NULL;
#endif

    /*  Preset structure  */
    newDef->IF_Flag = flag;

    /*  To detect stack underflow in conditional processing  */
    newDef->IF_CondLevel = ifstacklevel(&(jps->JSPP_Conds));

    /*  Hang in global list  */
    newDef->IF_Next = jps->JSPP_File;
    jps->JSPP_File = newDef;
    jps->JSPP_NFiles++;

    /*  Generate '#line' directive  */
    if((jps->JSPP_Mode & JPPM_LINEDIR) && !(mode & JPP_PUSHFILE_FNOLINEDIR) &&
       pushLineDir(jps, 1, newDef->IF_Name) == -1)
	EGOTO(error);

    /*  Call user routine  */
    if(jps->JSPP_OpenFile &&
       (*(jps->JSPP_OpenFile))(jps, newDef->IF_Name, jps->JSPP_NFiles - 1,
			       flag) == -1)
	EGOTO(error);

    return 0;
 error:
    if(newDef) free(newDef);
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jspppushppstream(jsppstream_t *jps, jsqueuestream_t *jsqs, int mode)

/************************************************************************/
{
    jsppfile_t *newDef;

    if((newDef = (jsppfile_t *)malloc(sizeof(*newDef))) == NULL)
	return -1;

    newDef->IF_Flag = IF_FLAG_PPSTREAM;
    newDef->IF_Name = NULL;
#ifdef CONFIG_JSCPP_OPTINC
    newDef->IF_Symbol  = NULL;
#endif
    newDef->IF_QStream = jsqs;

    /*  Push on stack  */
    newDef->IF_Next = jps->JSPP_File;
    jps->JSPP_File = newDef;

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jspppushinplevel(jsppstream_t *jps)

/************************************************************************/
{
    if(jps->JSPP_File)
	jps->JSPP_File->IF_Flag |= IF_FLAG_EOF;

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jspppopinplevel(jsppstream_t *jps)

/************************************************************************/
{
    if(jps->JSPP_File)
	jps->JSPP_File->IF_Flag &= ~IF_FLAG_EOF;

    return 0;
}

/*JS*********************************************************************
 *   Pops the input file from the stack. Returns > 0 if last file, 0 if ok
 *    and < 0 on error (No file open).
 *************************************************************************/

int _jspppopistack(jsppstream_t *jps)

/************************************************************************/
{
    jsppfile_t *ifile;
    int oiflag;

    ifile = jps->JSPP_File;
    if(IF_ISATEOF(ifile))
	EGOTO(error2);

    oiflag = ifile->IF_Flag;
    jps->JSPP_File = ifile->IF_Next;

    if(oiflag & IF_FLAG_MACRO) {
	if(dopopmacro(jps, ifile) == -1)
	    EGOTO(error);

    } else if(!(oiflag & IF_FLAG_PPSTREAM)) {
	if(dopopfile(jps, ifile) == -1)
	    EGOTO(error);
    }

    free(ifile);

    /*  Last file ?  */
    ifile = jps->JSPP_File;
    if(IF_ISATEOF(ifile))
	return 1;

    if((jps->JSPP_Mode & JPPM_LINEDIR) &&
       !(oiflag & (IF_FLAG_MACRO|IF_FLAG_PPSTREAM))) {
	/*  Use "EF_LineNr" since we need the line number after the
	 *   newline token at the end of the "#include" line, not before.
	 */
	if(pushLineDir(jps, ifile->IF_FFPtr->EF_LineNr, ifile->IF_Name))
	    EGOTO(error2);
    }

    return 0;
 error:
    if(ifile)
	free(ifile);
 error2:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int dopopfile(jsppstream_t *jps, jsppfile_t *ifile)

/************************************************************************/
{
    /*  We are at the end of a file. Ensure that we believe that a line
     *   was ended (file might not end with a newline)
     */
    jps->JSPP_LastToken2 = CTK_BLANKSEOL;

    jps->JSPP_NFiles--;

#ifdef CONFIG_JSCPP_OPTINC
    /*  Check if no non-space came after '#endif' and
     *   construct was successfully recognized.
     */
    if((ifile->IF_Flag & (IF_FLAG_CONSTRUCT | IF_FLAG_NONSPC)) ==
       IF_FLAG_CONSTRUCT) {
	/*  Register file  */
	if(saveOnceFile(jps, ifile->IF_Name, ifile->IF_Symbol) == -1)
	    EGOTO(error);
    }

    if(ifile->IF_Symbol)
	free((void *)ifile->IF_Symbol);
#endif

    /*  Close the file and free memory  */
    if(IF_ISFILE(ifile) && ifile->IF_FFPtr && efclose(ifile->IF_FFPtr))
	EGOTO(error);

    /*  Check if a ifdef construct was not finished  */
    if(ifile->IF_CondLevel != ifstacklevel(&(jps->JSPP_Conds))) {
	jsppprinterr(jps, "Unfinished '#if' construct in file", 1);
	EGOTO(error);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Pops the macro expansion from stack, frees the list of this macros
 *    parameters.
 *************************************************************************/

static int dopopmacro(jsppstream_t *jps, jsppfile_t *ifile)

/************************************************************************/
{
    if(ifile->IF_MArgs) {
	jsctoken_t *ptr;
	int i, count;

	/*  Count number of missing newlines  */
	count = ifile->IF_MArgs[0].MA_UseCount;

	/*  Free list of macro's argument  */
	/*  First entry contains only pointer for freeing  */
	for(i = 1 ; (ptr = ifile->IF_MArgs[i].MA_Token) ; i++) {
	    if(ifile->IF_MArgs[i].MA_UseCount == 0) {
		for( ; ptr->TK_Value != CTK_EOF ; ptr++) {
		    if(ptr->TK_String)
			count += countnlCToken(ptr->TK_String);

		    freeCToken(*ptr);
		}
	    } else {
		for( ; ptr->TK_Value != CTK_EOF ; ptr++)
		    freeCToken(*ptr);
	    }
	}

	if(ifile->IF_MArgs[0].MA_Token)
	    free(ifile->IF_MArgs[0].MA_Token);
	free(ifile->IF_MArgs);

	/*  Push missing newlines as a macro itself  */
	if(count > 0) {
	    static jsctoken_t Trailer[2];

	    Trailer[0] = _jsppgetnnewlines(count);
	    Trailer[1].TK_Value = CTK_EOF;
	    Trailer[1].TK_String = NULL;

	    if(_jspppushmacro(jps, NULL, Trailer, NULL, (ifile->IF_Flag & IF_SYSINCLUDE) |
			      IF_FLAG_DISABLEEXPAND) == -1)
		EGOTO(error);
	}
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Generate a '#line <nr> "file"' line and push it as a macro expansion.
 *************************************************************************/

static int pushLineDir(jsppstream_t *jps, int lineNr, const char *fileName)

/************************************************************************/
{
    static char NumBuff[20];
    static jsctoken_t LineTokens[] = {
	CTOKEN_GENSTATIC(PTK_STARTIGN,     NULL   ),
	CTOKEN_GENSTATIC(PTK_LINEDIRSTART, NULL   ),
	CTOKEN_GENSTATIC(CTK_NUMBER,       NumBuff),
	CTOKEN_GENSTATIC(CTK_BLANKS,       " "    ),
#define INDEX    4
	CTOKEN_GENSTATIC(CTK_STRING,       NULL   ), /*  Will be filled in  */
	CTOKEN_GENSTATIC(CTK_BLANKSEOL,    "\n"   ),
	CTOKEN_GENSTATIC(PTK_ENDIGN,       NULL   ),
	CTOKEN_GENSTATIC(CTK_EOF,          NULL   )
    };

    /*  Generate '#line' directive  */
    sprintf(NumBuff, "%d", lineNr);
    LineTokens[INDEX].TK_String = fileName;

    return _jspppushmacro(jps, NULL, LineTokens, NULL,
			  IF_FLAG_LINETOKENS | IF_FLAG_DISABLEEXPAND);
}
#undef INDEX

/*JS*********************************************************************
 *   Add include directory to file inclusion list.
 *************************************************************************/

int jsppaddincdir(jsppstream_t *jps, const char *name, int flags)

/************************************************************************/
{
    if(name == NULL) {
	int i;

	/*  Clear list of include directories  */
	for(i = 0 ; i < jps->JSPP_NIncDir ; ++i)
	    free((void *)jps->JSPP_IncDir[i].IN_Directory);

	free(jps->JSPP_IncDir);
	jps->JSPP_IncDir = NULL;
	jps->JSPP_NIncDir = 0;
    } else {
	int pos;

	if((jps->JSPP_IncDir = (jsppincdir_t *)
	    realloc(jps->JSPP_IncDir, (jps->JSPP_NIncDir + 1) *
		    sizeof(*(jps->JSPP_IncDir)))) == NULL)
	    return -1;

	if(flags & JPP_FILE_FADDTOTOP) {
	    flags &= ~JPP_FILE_FADDTOTOP;

	    if(jps->JSPP_NIncDir > 0)
		memmove(&(jps->JSPP_IncDir[1]), &(jps->JSPP_IncDir[0]),
			jps->JSPP_NIncDir * sizeof(jps->JSPP_IncDir[0]));

	    pos = 0;
	} else
	    pos = jps->JSPP_NIncDir++;

	if(!(flags & JPP_FILE_FALLOCATED) &&
	   (name = strdup(name)) == NULL)
	    return -1;

	jps->JSPP_IncDir[pos].IN_Directory = name;
	jps->JSPP_IncDir[pos].IN_Flags = flags;
    }

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsppaddincpath(jsppstream_t *jps, const char *includepath, int flags)

/************************************************************************/
{
    for(;;) {
	char *string;
	int l;

	if((string = (char *)strchr(includepath, CONFIG_LIST_DELIMITER)))
	    l = string - includepath;
	else
	    l = strlen(includepath);

	if((string = (char *)malloc(l + 1)) == NULL)
	    EGOTO(error);

	strncpy(string, includepath, l);
	string[l] = '\0';

	if(jsppaddincdir(jps, string, JPP_FILE_FALLOCATED | flags) == -1)
	    EGOTO(error);

	includepath += l;
	if(*includepath == '\0')
	    break;

	++includepath;
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsppshowinc(jsppstream_t *jps, FILE *fp, int mode)

/************************************************************************/
{
    int i;

    for(i = 0 ; i < jps->JSPP_NIncDir ; ++i)
	fprintf(fp, "INCDIR %c %s\n",
		(jps->JSPP_IncDir[i].IN_Flags & JPP_FILE_FSYSINC) ? 'S' : 'U',
		jps->JSPP_IncDir[i].IN_Directory);
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsppgetfilepos(jsppstream_t *jps, const char **pFileName, int mode)

/************************************************************************/
{
    jsppfile_t *iPtr;

    /*  Scan for current file  */
    for(iPtr = jps->JSPP_File ; !IF_ISATEOF(iPtr) ; iPtr = iPtr->IF_Next)
	if(IF_ISFILE(iPtr)) {
	    if(pFileName) {
		const char *name;

		name = iPtr->IF_Name;
		if((mode & 1) &&
		   (name = sharestrdup(&(jps->JSPP_FileNames), name)) == NULL)
		    return -1;

		*pFileName = name;
	    }

	    return iPtr->IF_LineNr;
	}

    /*  not found  */
    if(pFileName)
	*pFileName = NULL;

    return -1;
}

#ifdef CONFIG_JSCPP_OPTINC
/*JS*********************************************************************
 *   Save the file as one that need not be included multiple times if the
 *    symbol is defined. symbol is *not* copied.
 *************************************************************************/

static int saveOnceFile(jsppstream_t *jps, const char *name,
                        const char *symbol)

/************************************************************************/
{
    _jspponcefile_t *oFile;

    /*  Prevent multiple registering of file  */
    for(oFile = jps->JSPP_OnceFiles ; oFile ; oFile = oFile->OF_Next)
	if(strcmp(oFile->OF_Name, name) == 0)
	    return 0;

    if((oFile = (_jspponcefile_t *)malloc(sizeof(*oFile))) == NULL)
	EGOTO(error);

    if((oFile->OF_Name = strdup(name)) == NULL) {
	free(oFile);
	EGOTO(error);
    }

    if((oFile->OF_Symbol = strdup(symbol)) == NULL) {
	free((void*)oFile->OF_Name);
	free(oFile);
	EGOTO(error);
    }

    oFile->OF_Next = jps->JSPP_OnceFiles;
    jps->JSPP_OnceFiles = oFile;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Check if file has to be included.
 *************************************************************************/

static int checkOnceFile(jsppstream_t *jps, const char *name)

/************************************************************************/
{
    _jspponcefile_t *oFile;

    for(oFile = jps->JSPP_OnceFiles ; oFile ; oFile = oFile->OF_Next)
	if(strcmp(oFile->OF_Name, name) == 0) {
	    _jsppmacentry_t *entry;

	    /*  File inclusion can be skipped if symbol is defined  */
	    if((entry = _jsppfindmacentry(jps, oFile->OF_Symbol)) != NULL &&
	       !(entry->ME_Flags & MAFLAG_UNDEFINED))
		return TRUE;
	}

    return FALSE;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsppfreeoncefl(jsppstream_t *jps)

/************************************************************************/
{
    _jspponcefile_t *oFile, *nFile;

    for(oFile = jps->JSPP_OnceFiles ; oFile ; oFile = nFile) {
	nFile = oFile->OF_Next;
	free((void *)oFile->OF_Name);
	free((void *)oFile->OF_Symbol);
	free(oFile);
    }
}
#endif
