/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Initialization routines
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/
#ifndef CONFIG_NO_POSIX
# include <sys/types.h>
# include <sys/stat.h>
#endif
#include <stdlib.h>
#include <time.h>

#define JSCPP_INTERNAL
#include <jscpplib.h>

/*********     DEFINES						*********/
#ifndef JPPM_LINEDIR
/*  START-DEFINITIONS 2 */
#ifdef CONFIG_JSCPP_OPTINC
typedef struct _jspponcefile_t _jspponcefile_t;
struct _jspponcefile_t {
    _jspponcefile_t *OF_Next;
    const char      *OF_Name;
    const char      *OF_Symbol;
};
#endif

typedef struct {
    int JSOQ_Mode;
    jctokfifo_t JSOQ_OQueue;
    jctokfifo_t JSOQ_MacOQueue;
} _jsppoqueue_t;

typedef struct jsppstream_t jsppstream_t;
struct jsppstream_t {
    int JSPP_Mode;

    int JSPP_OLineNr;

    int JSPP_CTokMode;
    int JSPP_LastToken, JSPP_LastToken2;

    const jsctokops_t *JSPP_Ops;

    jsppincdir_t *JSPP_IncDir;
    int JSPP_NIncDir;

    /*  The current stream (file or macro expansion) we are reading from  */
    jsppfile_t *JSPP_File;
    int JSPP_NFiles;

    ifstack_t JSPP_Conds;

    /*  For all macro definitions  */
    _jsppmacentry_t *JSPP_Macros[JSPP_MACTABSIZE];

    jctokfifo_t JSPP_Queue;

    _jsppoqueue_t *JSPP_OQueue;
    int JSPP_QueueLevel, JSPP_MaxQueueLevel;

#ifdef CONFIG_JSCPP_ASSERTION
    _jspppredicate_t *JSPP_Predicates;
#endif

    /*  Called to print error messages  */
    void (*JSPP_PrintError)(void *user, const char *message);

    /*  If set, is called if pragma line encountered  */
    int (*JSPP_DoPragma)(jsppstream_t *jps, jsctoken_t *pToken, int n,
			 void *user);

    void *JSPP_User;

    jsctoken_t (*JSPP_SpecToken)(jsppstream_t *jps, int value);

    /*  When non-zero, is called every time a new file is opened  */
    int (*JSPP_OpenFile)(jsppstream_t *jps, const char *name, int incLevel,
			 int flag);

    int JSPP_NLLeftOver, JSPP_NLLeftOver2;
    int JSPP_Ignored;

    jssharestr_t     JSPP_FileNames;

#ifdef CONFIG_JSCPP_OPTINC
    _jspponcefile_t *JSPP_OnceFiles;
#endif
};

/* **  For JSPP_Mode  ** */
#define JPPM_LINEDIR               (1 <<  0) /*  Print '#line' directives in output  */
#define JPPM_PRESCOMM              (1 <<  1) /*  Preserve comments in output  */
#define JPPM_PRESSPACE             (1 <<  2) /*  Preserve spaces in output  */
#ifdef CONFIG_JSCPP_OPTLINES
# define JPPM_SAVENEWLN            (1 <<  3) /*  Save newlines by using '#line' directives */
#endif
#define JPPM_PPMACARGS             (1 <<  4) /*  Do preprocessing within macro args - not well tested  */
#define JPPM_PRWARNING             (1 <<  5)
/*  Do not accept redefinition of macros to a different token sequence  */
#define JPPM_STRICT                (1 <<  6)

/* Accept several errors in preprocessor lines and generate solely
 * warnings instead of errors (i. e. for unknown commands, garbage
 * at line end, illegal '#' and '##' tokens in macro definitions,
 * empty arguments to macro, illegal usages of '##' operators).
 * This is especially useful if the preprocessor parses things other
 * than valid C-code.
 */
#define JPPM_LAZYSTYLE             (1 <<  8)

#define JPPM_MACVARARG             (1 <<  9)

/*  Return original tokens in non-system includes  */
#define JPPM_EXPUSERFILES          (1 << 10)

/*  Expand macros in original stream in non-system includes  */
#define JPPM_EXPUSERMACROS         (1 << 11)

#define JPPM_FLAG_SEENDEFEREDMAC   (1 << 16)

#define jsppopen(jps, filename, ops, omode, mode)		\
    jsppopen2(jps, NULL, filename, NULL, ops, omode, mode)

/* ***  To simplify writing applications that parse cmdline options  *** */
typedef struct {
    const char *JSPES_String;
    int JSPES_Type;
} _jsppenvstype_t;

#define JSPE_INDEX_TINCS       0
#define JSPE_INDEX_TDEFS       1
#define JSPE_INDEX_TKEY        2
#define JSPE_INDEX_TASS        3
#define JSPE_INDEX_TSTARTFILE  4

#define JSPE_INDEX_MAX         5

#define jsppenvincdir(jppenv, dir, flag)  _jsppenvadd(jppenv, JSPE_INDEX_TINCS, dir, flag)
#define jsppenvdefine(jppenv, string)     _jsppenvadd(jppenv, JSPE_INDEX_TDEFS, string, 0)

typedef struct {
    long JSPE_Mode, JSPE_IMode;
    long JSPE_CTokMode;

    _jsppenvstype_t *JSPE_STypes[JSPE_INDEX_MAX];
    int JSPE_NSTypes[JSPE_INDEX_MAX];
} jsppenviron_t;

#ifdef CONFIG_JSCPP_ASSERTION
#define JSCPP_OPTSTRING_ADDASS  "A:"
#else
#define JSCPP_OPTSTRING_ADDASS   ""
#endif

/*  These are the standard options understood by jsppenvhandle  */
#define JSCPP_OPTSTRING       "I;i:D:U;x:PCWS:e;K:V"  JSCPP_OPTSTRING_ADDASS

#define jsppsetprinterr(jps, printError)   ((jps)->JSPP_PrintError = (printError))
/*  END-DEFINITIONS  */
#endif

/*  TRACE-DEFINITION symbol TRACETYPE_JSCPP name JSCPP */

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
Prototype int            jsppinit(int mode);

Prototype jsppstream_t  *jsppopen2(jsppstream_t *jps, jsppenviron_t *jppenv,
				   const char *filename, const char *incpath,
				   const jsctokops_t *ops, int omode, int mode);
static int               addstdinc(jsppstream_t *jps, const char *incpath);
static int               dopredefine(jsppstream_t *jps,
				     const char *filename);

Prototype int            jsppclose(jsppstream_t *jps, int flag);
Prototype void           jsppdelmacdef(jsppstream_t *jps, int flag);

Prototype void           jsppenvclear(jsppenviron_t *jppenv);
Prototype int            jsppenvhandle(jsppenviron_t *jppenv, int opt, const char *string);
Prototype int            _jsppenvadd(jsppenviron_t *jppenv, int nr, const char *string,
				     int type);
static int               _jsppenvapply(jsppenviron_t *jppenv, jsppstream_t *jps);
#ifdef CONFIG_JSCPP_ASSERTION
static int               defAssertion(jsppstream_t *jps, char *string);
#endif
Prototype void           jsppenvfree(jsppenviron_t * jppenv);


Prototype const char     JSPPStdOptsHelp[];
Prototype const char     JSPPStdOptsHelp2[];
Prototype int            JSPPLanguageLevel;
Prototype const char   **JSPPPragmaIntroducer;

/*********     GLOBAL /STATIC DEFINITIONS			*********/
const char JSPPStdOptsHelp[] =
    "    -I<dir>   Add <dir> to list of include directories.\n"
    "    -i<dir>   Add <dir> to list of include directories as a system directory.\n"
    "    -I        Clear internal list of standard include directories. To see\n"
    "              the default, use option '-V'.\n"
    "    -D<symb>  or  -D<symb>=<value>  or  -D<symb>(arg,..)=<value>\n"
    "              Define a symbol to the preprocessor. Is equivalent to\n"
    "              the '#define' command. If no <value> is given, <symb> is\n"
    "              defined to '1'.\n"
    "    -U<symb>  Undefine the specified symbol. Can be used to undefine one\n"
    "              of the predefined symbols that are not protected.\n"
    "    -U        Undefine user defined symbols from the command line.\n"
    "    -x<level> Specify the language level. Determines which macros are.\n"
    "              predefined according to the platform it was compiled on.\n"
    "                Level 0: No symbols are predefined.\n"
    "                Level 1: Internally defined like '__FILE__' without\n"
    "                         platform specific macros.\n"
    "                Level 2: Pure ANSI level with platform specific additions.\n"
    "                Level 3: Relaxed ANSI level (default).\n"
    "                Level 4: Extended level, i. e. macros that the C compiler\n"
    "                         usually defines are defined.\n"
    "              Use option '-V' to show predefined macros.\n"
    "    -V        Show predefined macros and include directories and then\n"
    "              exit program.\n"
#ifdef CONFIG_JSCPP_ASSERTION
    "    -Apredicate(answer) Define a new assertion.\n"
    "    -A-       Undefine all assertions.\n"
#endif
    "    -C	      Preserve comments in preprocessed file. Giving this option\n"
    "              twice preserves also spaces in the output.\n"
    "    -P        Suppress '#line' directives in the output.\n"
    "    -S<startfile> Behaves as if the first line of the input file was\n"
    "                  #include \"startfile\"\n"
    "              This option can be given multiple times.\n"
    "    -K<newkey>=<oldkey>\n"
    "              Introduce to the parser the keyword <newkey> with the same\n"
    "              meaning as <oldkey>.\n"
    "    -e[<extensions>] Switch on extension.\n"
    "              Show a list by using '-ehelp'.\n"
    ;

const char JSPPStdOptsHelp2[] =
    "    -W        Print warnings. Giving this option twice produces more\n"
    "              warnings.\n"
    ;

int JSPPLanguageLevel = CONFIG_JSCPPLIB_LANGUAGELEVEL;

static const char *DefaultPragmaIntroducer[] = {
    JSCPP_MACRO, NULL
};

const char **JSPPPragmaIntroducer = DefaultPragmaIntroducer;

static const dbgopts_t JSCPPExtensionOptions[] = {
    { "cppcomm",      1, NULL,
      "Allow C++ style comments."  },
    { "dollarident",  2, NULL,
      "Allow '$' in identifiers."  },
    { "macvararg",    4, NULL,
      "Allow C99 variadic macro arguments."  },
    { "strictdef",    8, NULL,
      "Do not allow redefinition of macros to a different token sequence."  },
#if 0
    { "ppmacarg",    16, NULL,
      "Preprocessing on during macro arguments (not working well)."  },
#endif
    { "lazy",        32, NULL,
      "Be lazy about unknown preprocessor commands, garbage in preprocessor lines,\n\
empty macro arguments."  },
    { NULL,           0, NULL, NULL }
};

/*JS*********************************************************************
 *
 *************************************************************************/

int jsppinit(int mode)

/************************************************************************/
{
#ifdef CONFIG_JSCPP_PIGGYBACKINCLUDES
    /*  Make standard includes compiled into the program available to
     *   extfile_t opener
     */
    if(epushopener(CONFIG_JSCPP_PIGGYBACKINCLUDES, piggybackopener,
		   (void *)JSPPStubIncludes) == -1)
	return -1;
#endif

    return 0;
}

/*JS*********************************************************************
 *   Creates a preprocessing stream, either in the pregiven jps or it is
 *    allocated. 'filename' is used for the __FILE__ macro.
 *    If omode != 1, it is opened and any start files from jppenv are
 *    opened as well.
 *************************************************************************/

jsppstream_t *jsppopen2(jsppstream_t *jps, jsppenviron_t *jppenv,
			const char *filename, const char *incpath,
			const jsctokops_t *ops, int omode, int mode)

/************************************************************************/
{
    int i, allocated;

    if(jps != NULL) {
	allocated = FALSE;
    } else {
	allocated = TRUE;
	if((jps = (jsppstream_t *)malloc(sizeof(*jps))) == NULL)
	    return NULL;
    }

    /*  Default is to generate line directives  */
    jps->JSPP_Mode = JPPM_LINEDIR | mode;

#ifdef CONFIG_JSCPP_OPTLINES
    jps->JSPP_Mode |= JPPM_SAVENEWLN;
#endif

    jps->JSPP_CTokMode = 0;

    jps->JSPP_LastToken = jps->JSPP_LastToken2 = CTK_BLANKSEOL;

    if(ifstackcreate(&(jps->JSPP_Conds)) == NULL ||
       jctokcreate(&(jps->JSPP_Queue), 0) == NULL)
	EGOTO(error);

    /*  This is only needed for mode JPPM_PPMACARGS, so let's be conservative  */
    jps->JSPP_MaxQueueLevel = 1;
    if((jps->JSPP_OQueue = (_jsppoqueue_t *)
	malloc(jps->JSPP_MaxQueueLevel * sizeof(*(jps->JSPP_OQueue)))) == NULL)
	EGOTO(error);

    for(i = 0 ; i < jps->JSPP_MaxQueueLevel ; ++i)
	if(jctokcreate(&(jps->JSPP_OQueue[i].JSOQ_OQueue), 0) == NULL ||
	   jctokcreate(&(jps->JSPP_OQueue[i].JSOQ_MacOQueue), 0) == NULL)
	    EGOTO(error);
    jps->JSPP_QueueLevel = 0;
    jps->JSPP_OQueue[0].JSOQ_Mode = 0;

    /*  Clear macro definitions  */
    for(i = 0 ; i < JSPP_MACTABSIZE ; ++i)
	jps->JSPP_Macros[i] = NULL;

#ifdef CONFIG_JSCPP_ASSERTION
    jps->JSPP_Predicates = NULL;
#endif

    /*  FIXME: This is less than perfect, but otherwise expressions
     *   in preprocessor lines don't work if they come through
     *   expansion of a macro. This is because during macro definition
     *   operators like '<<' are not recognized and later appear as
     *   '<' '<' giving syntax errors.
     *   We actually don't need more than this in the way we use
     *   it currently.
     */
    jps->JSPP_Ops = ops ? ops : _jsppOperators;

    jps->JSPP_IncDir = NULL;
    jps->JSPP_NIncDir = 0;

    /*  Clear input stacks  */
    jps->JSPP_File = NULL;
    jps->JSPP_NFiles = 0;

    jps->JSPP_PrintError = NULL;
    jps->JSPP_DoPragma   = NULL;
    jps->JSPP_User       = NULL;

    jps->JSPP_SpecToken  = NULL;
    jps->JSPP_OpenFile   = NULL;

    jps->JSPP_NLLeftOver = jps->JSPP_NLLeftOver2 = 0;
    jps->JSPP_Ignored = -1;

    sharestrinit(&(jps->JSPP_FileNames));

#ifdef CONFIG_JSCPP_OPTINC
    jps->JSPP_OnceFiles = NULL;
#endif

    if(JSPPLanguageLevel > 0 && dopredefine(jps, filename) == -1)
	EGOTO(error);

    if(addstdinc(jps, incpath) == -1)
	EGOTO(error);

    /*  First apply the environment settings to make sure JSPP_Mode
     *  is affected before we attempt anything else
     */
    if(jppenv != NULL && _jsppenvapply(jppenv, jps) == -1)
	EGOTO(error);

    /*  Open the input file and push it on the stream  */
    if(omode != -1) {
	int nstartfiles;

	nstartfiles = jppenv != NULL ? jppenv->JSPE_NSTypes[JSPE_INDEX_TSTARTFILE] : 0;

	if(nstartfiles > 0) omode |= JPP_PUSHFILE_FNOLINEDIR;

	if(jspppushfile(jps, filename, omode) == -1)
	    EGOTO(error);

	for(i = 0 ; i < nstartfiles ; ++i)
	    if(jspppushfile(jps, jppenv->JSPE_STypes[JSPE_INDEX_TSTARTFILE][i].JSPES_String, 0) == -1)
		EGOTO(error);
    }

    return jps;
 error:
    if(allocated)
	free(jps);

    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addstdinc(jsppstream_t *jps, const char *incpath)

/************************************************************************/
{
    char *string;

    /*  Environment variable for include directories overwrites
     *   internal settings
     */
    if((string = getenv(CONFIG_JSCPP_ENVINCPATH)) != NULL)
	return jsppaddincpath(jps, string, JPP_FILE_FSYSINC);

    if(incpath != NULL &&
       /*  Add path supplied by higher layer before our own  */
       jsppaddincpath(jps, incpath, JPP_FILE_FSYSINC) == -1)
	return -1;

    /*  Use compiled in default  */
    return jsppaddincpath(jps, CONFIG_JSCPP_INCLUDEDIR, JPP_FILE_FSYSINC);
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int dopredefine(jsppstream_t *jps, const char *filename)

/************************************************************************/
{
    time_t tim;
    struct tm *tp;
    char buffer[30];
    jsctoken_t token;

    /* ********************************** */
    /* **  Define recognition macros   ** */
    /* ********************************** */
    token.TK_Value = CTK_NUMBER;
    token.TK_String = buffer;
    sprintf(buffer, "%ld", JSCPPNVersion);
    if(jsppdefundefmac(jps, JSCPP_MACRO, &token, 1, MAFLAG_INTERNAL) == -1)
	EGOTO(error);

    strcpy(buffer, JSPPLanguageLevel <= 2 ? "1" : "0");
    if(jsppdefundefmac(jps, "__STDC__", &token, 1, MAFLAG_INTERNAL) == -1)
	EGOTO(error);

    /* ********************************** */
    /* **  Define time and date macro  ** */
    /* ********************************** */
    token.TK_Value = CTK_STRING;
    token.TK_String = buffer;

    /*  Get local time and date  */
    if((tim = time(NULL)) == -1)
	EGOTO(error);
    tp = localtime(&tim);

    strftime(buffer, sizeof(buffer), "%H:%M:%S", tp); /*  Format time  */
    if(jsppdefundefmac(jps, "__TIME__", &token, 1, MAFLAG_PROTECTED|
		       MAFLAG_INTERNAL) == -1)
	EGOTO(error);

    strftime(buffer, sizeof(buffer), "%b %d %Y", tp); /*  Format date  */
    if(jsppdefundefmac(jps, "__DATE__", &token, 1, MAFLAG_PROTECTED|
		       MAFLAG_INTERNAL) == -1)
	EGOTO(error);

#ifdef CONFIG_JSCPP_TIMESTAMP
# ifndef CONFIG_NO_POSIX
    /*  Get time of last modification, otherwise use current time  */
    if(filename && strcmp(filename, "-") != 0) {
	struct stat stat_buf;

	if(stat(filename, &stat_buf) == -1)
	    EGOTO(error);

	tp = localtime(&stat_buf.st_mtime);
    }
# endif

    strftime(buffer, sizeof(buffer), "%a %b %d %H:%M:%S %Y", tp);
    if(jsppdefundefmac(jps, "__TIMESTAMP__", &token, 1, MAFLAG_PROTECTED|
		       MAFLAG_INTERNAL) == -1)
	EGOTO(error);
#endif

    /* ********************************** */
    /* **  Define line and file macro  ** */
    /* ********************************** */
    /*  These macros are protected and cannot be undefined  */
    token.TK_String = NULL;
    token.TK_Value = PTK___LINE;
    if(jsppdefundefmac(jps, "__LINE__", &token, 1, MAFLAG_PROTECTED|
		       MAFLAG_INTERNAL) == -1)
	EGOTO(error);
    token.TK_Value = PTK___FILE;
    if(jsppdefundefmac(jps, "__FILE__", &token, 1, MAFLAG_PROTECTED|
		       MAFLAG_INTERNAL) == -1)
	EGOTO(error);

    /* ******************************* */
    /* **  Some common extensions:  ** */
    /* ******************************* */
#ifdef CONFIG_JSCPP_INCLEVEL
    /*  Note: They are not protected and hence may be undefined  */
    token.TK_Value = PTK___INCLEVEL;
    if(jsppdefundefmac(jps, "__INCLUDE_LEVEL__", &token, 1, MAFLAG_PROTECTED|
		       MAFLAG_INTERNAL) == -1)
	EGOTO(error);
#endif
#ifdef CONFIG_JSCPP_BASEFILE
    token.TK_Value = PTK___BASEFILE;
    if(jsppdefundefmac(jps, "__BASE_FILE__", &token, 1, MAFLAG_PROTECTED|
		       MAFLAG_INTERNAL) == -1)
	EGOTO(error);
#endif

#ifdef JSCPP_USE_PREDEFINES
    if(JSPPLanguageLevel >= 2 &&
       _jspppredefine(jps, JSPPLanguageLevel - 2) == -1)
	EGOTO(error);
#endif

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   If bit 0 of flag is set, do not free jps. If bit 1 is set, do not
 *    check for errors.
 *************************************************************************/

int jsppclose(jsppstream_t *jps, int flag)

/************************************************************************/
{
    int i, ret = 0;

    if(ifstackfree(&(jps->JSPP_Conds), flag & 2) == -1)
	ret = -1;

    /*  Free all queues  */
    jctokfree(&(jps->JSPP_Queue), 0);
    for(i = 0 ; i < jps->JSPP_MaxQueueLevel ; ++i) {
	jctokfree(&(jps->JSPP_OQueue[i].JSOQ_OQueue), 0);
	jctokfree(&(jps->JSPP_OQueue[i].JSOQ_MacOQueue), 0);
    }
    free(jps->JSPP_OQueue);

    jsppdelmacdef(jps, 3);

#ifdef CONFIG_JSCPP_ASSERTION
    jsppdelpred(jps, NULL, NULL);
#endif

    for(i = 0 ; i < jps->JSPP_NIncDir ; ++i)
	free((void *)jps->JSPP_IncDir[i].IN_Directory);

    free(jps->JSPP_IncDir);

    sharestrdeinit(&(jps->JSPP_FileNames), 0);

#ifdef CONFIG_JSCPP_OPTINC
    jsppfreeoncefl(jps);
#endif

    if(!(flag & 1)) free(jps);

    return ret;
}

/*JS*********************************************************************
 *   If bit 1 of flag is set, deletes also the protected macros, if bit 0
 *    is set, delete also the internal ones.
 *************************************************************************/

void jsppdelmacdef(jsppstream_t *jps, int flag)

/************************************************************************/
{
    int i;

    /*  Free all macro definitions  */
    for(i = 0 ; i < JSPP_MACTABSIZE ; i++) {
	_jsppmacentry_t *entry, *oEntry, *nEntry;

	oEntry = NULL;
	for(entry = jps->JSPP_Macros[i] ; entry ; entry = nEntry) {
	    nEntry = entry->ME_Next;
	    if((entry->ME_Flags & MAFLAG_PROTECTED) ? (flag & 2) :
	       (!(entry->ME_Flags & MAFLAG_INTERNAL) || (flag & 1))) {
		/*  Free the macro definition  */
		_jsppfreemacdef(jps, entry);

		if(oEntry) oEntry->ME_Next = nEntry;
		else jps->JSPP_Macros[i] = nEntry;
	    } else
		oEntry = entry;
	}
    }
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsppenvclear(jsppenviron_t *jppenv)

/************************************************************************/
{
    int i;

    jppenv->JSPE_Mode = 0;
    jppenv->JSPE_IMode = 0;

    jppenv->JSPE_CTokMode = 0;

    for(i = 0 ; i < JSPE_INDEX_MAX ; ++i) {
	jppenv->JSPE_STypes[i] = NULL;
	jppenv->JSPE_NSTypes[i] = 0;
    }
}

/* ERROR-DEFINITIONS from jsppenvhandle label _ERR_JSPPENVHANDLE
   Illegal '-U' option
*/

/*JS*********************************************************************
 *
 *************************************************************************/

int jsppenvhandle(jsppenviron_t *jppenv, int opt, const char *string)

/************************************************************************/
{
    switch(opt) {
    case 'I':
	if(_jsppenvadd(jppenv, JSPE_INDEX_TINCS, string, 0) == -1)
	    EGOTO(error);
	break;

    case 'i':
	if(_jsppenvadd(jppenv, JSPE_INDEX_TINCS, string, JPP_FILE_FSYSINC) == -1)
	    EGOTO(error);
	break;

    case 'D':
	if(_jsppenvadd(jppenv, JSPE_INDEX_TDEFS, string, 0) == -1)
	    EGOTO(error);
	break;

    case 'U':
	if(string == NULL) {
	    if(_jsppenvadd(jppenv, JSPE_INDEX_TDEFS, NULL, 0) == -1)
		EGOTO(error);

	} else if(isalpha(*(unsigned char *)string) || *string == '_') {
	    /*  Undefine specified macro  */
	    if(_jsppenvadd(jppenv, JSPE_INDEX_TDEFS, string, -1) == -1)
		EGOTO(error);

	} else {
	    JSSETERROR2(_ERR_JSPPENVHANDLE, 0, string);
	    EGOTO(error);
	}
	break;

    case 'x':
	JSPPLanguageLevel = atoi(string);
	break;

#ifdef CONFIG_JSCPP_ASSERTION
    case 'A':
	if(_jsppenvadd(jppenv, JSPE_INDEX_TASS, string, 0) == -1)
	    EGOTO(error);
	break;
#endif

    case 'P': /*  No '#line' directives  */
	jppenv->JSPE_IMode |= JPPM_LINEDIR;
	break;

    case 'C': /*  Preserve comments  */
	if(jppenv->JSPE_Mode & JPPM_PRESCOMM)
	    jppenv->JSPE_Mode |= JPPM_PRESSPACE;

	jppenv->JSPE_Mode |= JPPM_PRESCOMM;
	break;

    case 'W': /*  Print warnings  */
	jppenv->JSPE_Mode |= JPPM_PRWARNING;
	break;

    case 'S': /*  File included at the start  */
	if(_jsppenvadd(jppenv, JSPE_INDEX_TSTARTFILE, string, 0) == -1)
	    EGOTO(error);
	break;

    case 'K':
	if(_jsppenvadd(jppenv, JSPE_INDEX_TKEY, string, 0) == -1)
	    EGOTO(error);
	break;

    case 'e': /*  Switch on extensions  */
	{
	    long val;

	    if(string == NULL)
		val =  1 | 2 | 4 | 32;
	    else if(parsedbgopts(JSCPPExtensionOptions, string,
				 &val, 3) == -1)
		EGOTO(error);

	    if(val &  1) jppenv->JSPE_CTokMode |= GCTK_CPPCOMMENT;
	    if(val &  2) jppenv->JSPE_CTokMode |= GCTK_EXTENSIONS;

	    if(val &  4) jppenv->JSPE_Mode |= JPPM_MACVARARG;
	    if(val &  8) jppenv->JSPE_Mode |= JPPM_STRICT;
	    if(val & 16) jppenv->JSPE_Mode |= JPPM_PPMACARGS;
	    if(val & 32) jppenv->JSPE_Mode |= JPPM_LAZYSTYLE;
	}
	break;

    case 'V': /*  Show predefined options and exit program  */
	{
	    jsppstream_t jps;

	    if(jsppopen2(&jps, jppenv, NULL, NULL, NULL, 0, 0) == NULL)
		EGOTO(error);

	    jsppshowmac(&jps, stdout, -1);

	    jsppshowinc(&jps, stdout, 0);

	    jsppclose(&jps, 1);
	}
	exit(0);

    default:
	/*  This option is not for us  */
	return -2;
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int _jsppenvadd(jsppenviron_t *jppenv, int nr, const char *string,
		int type)

/************************************************************************/
{
    int pos;

    pos = jppenv->JSPE_NSTypes[nr];

    if((jppenv->JSPE_STypes[nr] = (_jsppenvstype_t *)
	realloc(jppenv->JSPE_STypes[nr], (pos + 1) *
		sizeof(*(jppenv->JSPE_STypes[nr])))) == NULL)
	return -1;

    jppenv->JSPE_STypes[nr][pos].JSPES_String = string;
    jppenv->JSPE_STypes[nr][pos].JSPES_Type = type;

    jppenv->JSPE_NSTypes[nr]++;

    return pos;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int _jsppenvapply(jsppenviron_t *jppenv, jsppstream_t *jps)

/************************************************************************/
{
    const char *string;
    int i, type;

    jps->JSPP_Mode &= ~(jppenv->JSPE_IMode);
    jps->JSPP_Mode |= jppenv->JSPE_Mode;

    jps->JSPP_CTokMode |= jppenv->JSPE_CTokMode;

    for(i = 0 ; i < jppenv->JSPE_NSTypes[JSPE_INDEX_TINCS] ; ++i) {
	string = jppenv->JSPE_STypes[JSPE_INDEX_TINCS][i].JSPES_String;
	type = jppenv->JSPE_STypes[JSPE_INDEX_TINCS][i].JSPES_Type;

	if(jsppaddincdir(jps, string, type) == -1)
	    EGOTO(error);
    }

    for(i = 0 ; i < jppenv->JSPE_NSTypes[JSPE_INDEX_TDEFS] ; ++i) {
	string = jppenv->JSPE_STypes[JSPE_INDEX_TDEFS][i].JSPES_String;
	type = jppenv->JSPE_STypes[JSPE_INDEX_TDEFS][i].JSPES_Type;

	if(string == NULL) {
	    jsppdelmacdef(jps, type);
	} else if(type == 0) {
	    if(jsppdefinemacro(jps, string, 0) == -1)
		EGOTO(error);
	} else {
	    if(jsppdefundefmac(jps, string, NULL, -1, 0) == -1)
		EGOTO(error);
	}
    }

#ifdef CONFIG_JSCPP_ASSERTION
    for(i = 0 ; i < jppenv->JSPE_NSTypes[JSPE_INDEX_TASS] ; ++i) {
	string = jppenv->JSPE_STypes[JSPE_INDEX_TASS][i].JSPES_String;
	type = jppenv->JSPE_STypes[JSPE_INDEX_TASS][i].JSPES_Type;

	if(strcmp(string, "-") == 0) {
	    /*  Delete all assertions  */
	    if(jsppdelpred(jps, NULL, NULL) == -1)
		EGOTO(error);
	} else if(defAssertion(jps, (char *)string) == -1)
	    EGOTO(error);
    }
#endif

    return 0;
 error:
    return -1;
}

#ifdef CONFIG_JSCPP_ASSERTION
/*JS*********************************************************************
 *
 *************************************************************************/

static int defAssertion(jsppstream_t *jps, char *string)

/************************************************************************/
{
    char *pred, *ans, *string2;
    int len;

    /*  Skip initial spaces  */
    while(isspace(*(unsigned char *)string))
	string++;

    /*  Get predicate  */
    pred = string;
    if((len = jstrtoid(string)) == 0)
	EGOTO(error);
    string2 = string += len;

    /*  Skip heading space in answer  */
    while(isspace(*(unsigned char *)string))
	string++;

    if(*string != '(')
	EGOTO(error);

    *string2 = '\0'; /*  Delimit predicate  */
    ans = string2 = ++string; /*  skip '('  */

    /*  Copy answer, changing sequences of spaces to single spaces  */
    while(*string && *string != ')') {
	*string2++ = *string;

	/*  Skip sequences of spaces  */
	if(isspace(*(unsigned char *)string)) {
	    do
		string++;
	    while(isspace(*(unsigned char *)string));
	} else
	    string++;
    }
    if(*string != ')')
	EGOTO(error);

    /*  Remove heading and trailing spaces  */
    while(isspace(*(unsigned char *)ans) && ans < string2)
	++ans;

    string2--;
    while(isspace(*(unsigned char *)string2) && ans < string2)
	string2--;
    string2[1] = '\0';

    if(*ans == '\0')
	EGOTO(error);

    if(jsppnewpred(jps, pred, ans) == -1)
	return -1;

    return 0;
 error:
    fprintf(stderr, "ERROR: Illegal assertion in command line -- Break.\n");
    return -1;
}
#endif

/*JS*********************************************************************
 *
 *************************************************************************/

void jsppenvfree(jsppenviron_t * jppenv)

/************************************************************************/
{
    unsigned int i;

    for(i = 0 ; i < JSPE_INDEX_MAX ; ++i)
	free(jppenv->JSPE_STypes[i]);
}
