/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Macro expansion code.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/

#define JSCPP_INTERNAL
#include <jscpplib.h>

/*********     DEFINES                                          *********/

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
IPrototype jsctoken_t    _jspptkbasic(jsppstream_t *jps,
                                      jsctoken_t *pOToken, int mode);
IPrototype jsctoken_t    _jspptkbasic2(jsppstream_t *jps,
                                       jsctoken_t *pOToken, int mode);
static jsctoken_t        _jspptkfile(jsppstream_t *jps,
                                     jsctoken_t *pOToken, int mode);
static jsctoken_t        getSpecial(jsppstream_t *jps, int value);

static int               doConcat(jsppstream_t *jps, jsctoken_t *dest,
                                  jsctoken_t tokens[], int n);
static int               isValidSuffix(const char *string);
static int               isOperator(jsppstream_t *jps, jsctoken_t tk);

/*JS*********************************************************************
 *   Wrapper for getTkBasic2 to prevent PTK_IDENTNOEXP to be seen outside.
 *************************************************************************/

jsctoken_t _jspptkbasic(jsppstream_t *jps, jsctoken_t *pOToken, int mode)

/************************************************************************/
{
    jsctoken_t token;

    token = _jspptkbasic2(jps, pOToken, mode);

#ifdef TRACETYPE_JSCPP
    if(jstracelevel(TRACETYPE_JSCPP) >= 6) {
	const char *tokString;

	if((tokString = jspptok2string(token)) != NULL)
	    jstrace(TRACETYPE_JSCPP, 6,
		    "PPTOKEN: %s ", tokString);
	else
	    jstrace(TRACETYPE_JSCPP, 6,
		    "PPTOKEN: %d ", (int)token.TK_Value);

	if(token.TK_String)
	    jstrace(TRACETYPE_JSCPP, 6,
		    " \"%s\"\n", token.TK_String);
	else
	    jstrace(TRACETYPE_JSCPP, 6,
		    "\n");
    }
#endif

    if(token.TK_Value == PTK_IDENTNOEXP)
	token.TK_Value = CTK_IDENT;

    if(pOToken && pOToken->TK_Value == PTK_IDENTNOEXP)
	pOToken->TK_Value = CTK_IDENT;

    return token;
}

/*JS*********************************************************************
 *   Get next token with routine getCToken or from macro expansion. Handle
 *    macro arguments and special preprocessor tokens '#', '##', PTK___LINE,
 *    PTK___FILE etc. Macro expansion is done only when inserting a macros
 *    parameter as required by the ANSI standard unless the flag GTK_NO_EXP
 *    in mode is set.
 *************************************************************************/

jsctoken_t _jspptkbasic2(jsppstream_t *jps, jsctoken_t *pOToken, int mode)

/************************************************************************/
{
    jsctoken_t token;

    for(;;) {
	jsppfile_t *iPtr;

	/*  No file open or simulated EOF?  */
	if(IF_ISATEOF(jps->JSPP_File)) {
	    token.TK_Flags = 0; token.TK_Value = CTK_EOF; token.TK_String = NULL;

	    /*  We read from a file, so original token is identical  */
	    if(pOToken) *pOToken = token;

	    break; /*  We got it  */
	}

	if(IF_ISFILE(jps->JSPP_File)) {
	    token = _jspptkfile(jps, pOToken, mode);

	    break; /*  We got it  */
	}

	/*  Get next token from macro expansion  */
	token = *((jps->JSPP_File->IF_MTokens)++);

	if((mode & GTK_PREPROCOPS) && token.TK_Value >= CTK_USERTOKEN &&
	   token.TK_String) {
	    int i;

	    /*  During macro definition, another operator token list was active, so
	     *   other tokens were generated. But now we need them for our
	     *   preprocessor line, so check.
	     *   FIXME: This won't work if preprocessor relevant operators like
	     *   '<<' are not in 'jps->JSPP_Ops'.
	     */
	    for(i = 0 ; _jsppOperators[i].KD_String ; ++i)
		if(strcmp(_jsppOperators[i].KD_String, token.TK_String) == 0) {
		    token.TK_Flags = 0;
		    token.TK_Value = _jsppOperators[i].KD_Token;
		    break;
		}
	}

	if(token.TK_Value == CTK_EOF) {
	    /*  Pop this macro expansion  */
	    if(_jspppopistack(jps) == -1)
		EGOTO(error);

	    continue;  /*  Redo looping  */
	}

	if(pOToken) {
	    if(jps->JSPP_File->IF_Flag & (IF_FLAG_ORIGTOK |
					  IF_FLAG_EXPANDORIG |
					  IF_FLAG_EXPANDORIGFULL)) {
		*pOToken = token;
	    } else {
		/*  Corresponding original token is empty  */
		pOToken->TK_Flags  = 0;
		pOToken->TK_Value  = PTK_EMPTY;
		pOToken->TK_String = NULL;
	    }
	}

	/*  Before popping EOF's, push if macro's argument  */
	if(PTK_ISMACROARG(token.TK_Value)) {
            _jsppmacarg_t *pmacArg;
	    int nr;

	    /*  If no arguments present, keep it in output  */
	    if(jps->JSPP_File->IF_MArgs == NULL)
		break;

	    /*  This is a macro's parameter  */
	    nr = 1 + PTK_GETMACROARG(token.TK_Value);
            pmacArg = &(jps->JSPP_File->IF_MArgs[nr]);

            if((pmacArg->MA_UseCount)++ == 1) {
		jsctoken_t *ptr;

		/*  Argument used the second time, clear all newlines  */
                for(ptr = pmacArg->MA_Token ; ptr->TK_Value != CTK_EOF ; ptr++)
		    if(ptr->TK_String)
                        tkDelNewLines((char *)ptr->TK_String);
	    }

            if(_jspppushmacro(jps, NULL, pmacArg->MA_Token, NULL,
                              IF_FLAG_EXPANDINGARG) == -1)
		EGOTO(error);

	    continue; /*  Redo looping  */
	}

	/*  Look for subtokens  */
	if(token.TK_Value == PTK_SUBTOKENS) {
	    if(_jspppushmacro(jps, NULL, (jsctoken_t *)token.TK_Ptr, NULL,
			      IF_FLAG_EXPANDINGARG) == -1)
		EGOTO(error);

	    continue; /*  Redo looping  */
	}

	if(token.TK_Value == PTK_STRINGIZE) {
            _jsppmacarg_t *pmacArg;
            int nr;

	    /*  If no arguments present, keep it in output  */
	    if(jps->JSPP_File->IF_MArgs == NULL)
		break;

	    /*  Skip intermediate non macro arguments  */
	    do {
		token = *((jps->JSPP_File->IF_MTokens)++);

	    } while(!PTK_ISMACROARG(token.TK_Value));

            nr = 1 + PTK_GETMACROARG(token.TK_Value);
            pmacArg = &(jps->JSPP_File->IF_MArgs[nr]);

            token = _jsppstringize(pmacArg->MA_Token, 7);
	    break; /*  We got it  */
	}

    redo:
	/*  Get next lookahead token and check for '##'  */
	if(!(mode & GTK_NO_CONCAT) && !(jps->JSPP_File->IF_Flag & IF_FLAG_NOEXP)) {
	    iPtr = jps->JSPP_File;
	    do {
		jsctoken_t *ptr;

		ptr = iPtr->IF_MTokens;

		/*  Skip blank tokens  */
		while(CTK_BLANK_START <= ptr->TK_Value &&
		      ptr->TK_Value <= CTK_BLANK_END)
		    ptr++;

		/*  Next lookahead token is '##'?  */
		if(ptr->TK_Value == PTK_CONCAT) {
		    /*  We need to look ahead for 2 more tokens. This is for cases like
		     *   "two ## 4tea": The latter tokens are read as *TWO* tokens, but
		     *   the overall result should be *ONE*, the identifier "two4tea".
		     */
#define NR_TOKENS      3  /*  1 before the "##" and two after  */
		    static jsctoken_t freeToken = CTOKEN_GENSTATIC(CTK_EOF, NULL);
		    jsctoken_t saveTokens[NR_TOKENS + 1];
                    unsigned int flags[NR_TOKENS], i, j, n;

		    /*
		     *   Since 'token' might be a result of 'a ## b',
		     *    we make a copy before calling getTkBasic
		     *    recursively.
		     */
		    saveTokens[0] = copyCToken(token, 0);

		    /*  n == 0: look for '##'
		     *  n == 1: look for token after the '##'.
		     *  We ignore space around the '##' token.
		     */
		    for(n = 0 ; n < NR_TOKENS ; ++n) {
			do {
			    /*  Call recursive. GTK_NO_CONCAT yields left
			     *   associativity of '##' operator.
			     */
			    token = _jspptkbasic2(jps, NULL, mode |
						  (GTK_NO_EXP | GTK_NO_CONCAT));
			} while(n <= 1 && CTK_BLANK_START <= token.TK_Value &&
				token.TK_Value <= CTK_BLANK_END);

			/*  I set this afterwards. The current input stack still
			 *   reflects the status for this token.
			 */
			if(IF_ISFILE(jps->JSPP_File))
			    flags[n] = IF_FLAG_ORIGTOK;
			else
			    flags[n] = (jps->JSPP_File->IF_Flag &
					(IF_FLAG_ORIGTOK | IF_FLAG_EXPANDORIG |
					 IF_FLAG_EXPANDORIGFULL)) ? IF_FLAG_ORIGTOK : 0;

			if(token.TK_Value == CTK_EOF)
			    break;

			if(n > 0)
			    saveTokens[n] = copyCToken(token, 0);
		    }

		    /*  Try to concatenate two tokens  */
		    if((i = doConcat(jps, &token, saveTokens, n)) == -1)
			EGOTO(error);

		    /*  i is at least 1  */
		    if(i == 1) {
			/*  Nothing was concatenated, so token hasn't changed.
			 *   We save it to be freed next time we come around
			 */
			freeCToken(freeToken);
			freeToken = token;

		    } else {
			/*  If there was any concatenation, we can free the
			 *   concatenated ones
			 */
			for(j = 0 ; j < i ; ++j)
			    freeCToken(saveTokens[j]);
		    }

		    if(i < n) {
			jsctoken_t *rem;
			int flagrem;

			/*  Some tokens which could not be concatenated remain;
			 *   return them individually
			 */

			/*  FIXME: We should not have to copy it, because it
			 *   is from a macro expansion etc., so it still
			 *   exists somewhere.
			 */
			rem = NULL;
			for(j = 0 ; i < n ; ++i, ++j) {
			    if(j > 0 && flags[j - 1] != flags[i]) {
				/*  We have to split it into two parts, one
				 *   which contains original tokens, one
				 *   which does not.
				 */
				saveTokens[j].TK_Flags = 0;
				saveTokens[j].TK_Value = CTK_EOF;
				saveTokens[j].TK_String = NULL;
				++j;
				rem = &saveTokens[j];
				flagrem = flags[i];
			    }

			    saveTokens[j] = saveTokens[i];
			    flags[j] = flags[i];
			}
			saveTokens[j].TK_Flags = 0;
			saveTokens[j].TK_Value = CTK_EOF;
			saveTokens[j].TK_String = NULL;

			/*  We push it as expansion of " " to make sure further
			 *   macro expansion takes place
			 */
			if(rem && _jspppushmacro(jps, NULL, rem, NULL,
						 IF_FLAG_COPYTOKS | flagrem) == -1)
			    EGOTO(error);

			if(_jspppushmacro(jps, NULL, saveTokens, NULL,
					  IF_FLAG_COPYTOKS | flags[0]) == -1)
			    EGOTO(error);

			/*  We need this dummy expansion to represent the correct
			 *   status for the current token we have: It should *not*
			 *   look as if it stems from the macro expansion of saveTokens
			 *   above which could have (flags[0] & IF_FLAG_ORIGTOK) == TRUE
			 *   This would incorrectly treat this token as an *original* one
			 *   and produce it in the output stream for original tokens
			 *   despite the fact that it was generated by concatentation!
			 *
			 *  Here is an example where it went wrong for the original
			 *   tokens:
			 *
			 *     #define good         append(too)
			 *     #define append(name) name##bad
			 *     #define toobad       worse
			 *     int good;
			 *
			 */
			if(_jspppushmacro(jps, NULL, &saveTokens[j], NULL,
					  IF_FLAG_COPYTOKS) == -1)
			    EGOTO(error);
		    }

		    goto redo; /*  We got it  */

		} else if(ptr->TK_Value != CTK_EOF)
		    break;

		iPtr = iPtr->IF_Next;

	    } while(iPtr && IF_ISMACRO(iPtr));
	}

	if(token.TK_Value == PTK___LINE ||
#ifdef CONFIG_JSCPP_INCLEVEL
	   token.TK_Value == PTK___INCLEVEL ||
#endif
#ifdef CONFIG_JSCPP_BASEFILE
	   token.TK_Value == PTK___BASEFILE ||
#endif
	   token.TK_Value == PTK___FILE) {
	    token = getSpecial(jps, token.TK_Value);
	    break; /*  We got it  */
	}

	if(jps->JSPP_SpecToken) {
	    jsctoken_t token2;

	    /*  Try user supplied routine  */
	    token2 = jps->JSPP_SpecToken(jps, token.TK_Value);
	    if(token2.TK_Value == CTK_ERROR)
		EGOTO(error);

	    if(token2.TK_Value != CTK_EOF) {
		token = token2;
		break;
	    }
	}

	if(token.TK_Value == CTK_IDENT) {
	    /*  Are we within a macro's parameter expansion?  */
	    for(iPtr = jps->JSPP_File ; iPtr && IF_ISMACRO(iPtr) ;
		iPtr = iPtr->IF_Next)
		if(iPtr->IF_Flag & IF_FLAG_EXPANDINGARG)
		    /*  We have hit a macro's parameter expansion, so don't look
		     *   any further. This is to enable macros that appear in
		     *   arguments of themselves like in MACRO(MACRO(x))
		     */
		    break;

	    if(iPtr && IF_ISMACRO(iPtr) && !(iPtr->IF_Flag & IF_FLAG_NOEXP)) {
		jsctoken_t *ptr;

		/*  Look if this macro argument is followed by '##' and prevent
		 *   macro expansion
		 */
		ptr = jps->JSPP_File->IF_MTokens;
		while(CTK_BLANK_START <= ptr->TK_Value &&
		      ptr->TK_Value <= CTK_BLANK_END)
		    ptr++;

		if(ptr->TK_Value == CTK_EOF) {
		    ptr = jps->JSPP_File->IF_Next->IF_MTokens;
		    while(CTK_BLANK_START <= ptr->TK_Value &&
			  ptr->TK_Value <= CTK_BLANK_END)
			ptr++;
		}

		if(ptr->TK_Value != PTK_CONCAT) {
		    jsppfile_t *iPtr2;
		    int ret;

		    /*  We are expanding a macro's argument. ANSI requires
		     *   that recursive expansion of macros is done first in
		     *   a macro's argument.
		     *  Mark that we are already expanding this macro argument
		     *   and stop expansion while collecting macro arguments
		     */
		    iPtr->IF_Flag |= IF_FLAG_NOEXP;

		    ret = _jsppexpand(jps, token.TK_String, mode);

		    /*  Check if iPtr is still valid  */
		    for(iPtr2 = jps->JSPP_File ; iPtr2 && IF_ISMACRO(iPtr2) ;
			iPtr2 = iPtr2->IF_Next)
			if(iPtr2 == iPtr) {
			    iPtr2->IF_Flag &= ~IF_FLAG_NOEXP;
			    break;
			}

		    if(ret == 0)
			continue;

		    if(ret < 0)
			EGOTO(error);

		    /*  Mark to prohibit reexpansion of macro  */
		    if(ret == 2) {
			token.TK_Flags = 0;
			token.TK_Value = PTK_IDENTNOEXP;
		    }
		}
	    }
	}

	break; /*  We got it  */
    }

    return token;
 error:
    token.TK_Flags = 0; token.TK_Value = CTK_ERROR; token.TK_String = NULL;
    return token;
}

/*JS*********************************************************************
 *   Read a token from a file.
 *************************************************************************/

static jsctoken_t _jspptkfile(jsppstream_t *jps, jsctoken_t *pOToken,
                              int mode)

/************************************************************************/
{
    jsctoken_t token;
    jsppfile_t *jfile;

    jfile = jps->JSPP_File;
    if(!(jfile->IF_Flag & IF_FLAG_PPSTREAM)) {
	const jsctokops_t *ops;

	/*  Save line number now  */
	jfile->IF_LineNr = jfile->IF_FFPtr->EF_LineNr;

	if(mode & GTK_PREPROCOPS)
	    /*  Regardless of which operators the caller sees, within preprocessor
	     *   lines we have our own set of known and allowed operators
	     */
	    ops = _jsppOperators;
	else
	    ops = jps->JSPP_Ops;

	token = getCToken(jfile->IF_FFPtr, ops, jps->JSPP_CTokMode |
			  (mode & (GCTK_FLAG_USER - 1)));

    } else {
	jsqueuestream_t *jsqs;
	jctokfifo_t *queue;
	int indx;

	jsqs = jfile->IF_QStream;
	queue = jsqsgetqueue(jsqs);

	if((indx = jsqsget(jsqs, 0)) == -1)
	    EGOTO(error);

	if(indx == -2) {
	    token.TK_Flags = 0;
	    token.TK_Value = CTK_EOF;
	    token.TK_String = NULL;

	} else
	    token = queue->TKFF_Token[indx];
    }

#ifdef CONFIG_JSCPP_OPTINC
    if(!(jfile->IF_Flag & IF_FLAG_INCONSTRUCT)) {
	/* **  Set flag that we have seen a non-space token  ** */
	if(jfile->IF_Flag & IF_FLAG_NONSPC1) {
	    /*  Delete NONSPC1 flag and set NONSPC flag  */
	    jfile->IF_Flag = (jfile->IF_Flag & ~IF_FLAG_NONSPC1) | IF_FLAG_NONSPC;
	}

	if(!((CTK_BLANK_START <= token.TK_Value &&
	      token.TK_Value <= CTK_BLANK_END) || token.TK_Value == CTK_EOF))
	    jfile->IF_Flag |= IF_FLAG_NONSPC1;
    }
#endif

    /*  We read from a file, so original token is identical  */
    if(pOToken)
	*pOToken = token;

    if(token.TK_Value != CTK_EOF) {
	int tokValue;

	tokValue = token.TK_Value;

	/*  If the first line in a file contains spaces before
	 *   the preprocesor line '#', we have to maintain
	 *   CTK_BLANKSEOL as the last token.
	 */
	if(jps->JSPP_LastToken2 == CTK_BLANKSEOL &&
	   CTK_BLANK_START <= tokValue && tokValue <= CTK_BLANK_END)
	    tokValue = CTK_BLANKSEOL;

	jps->JSPP_LastToken = jps->JSPP_LastToken2;
	jps->JSPP_LastToken2 = tokValue;
    }

    return token;
 error:
    token.TK_Flags = 0;
    token.TK_Value = CTK_ERROR;
    token.TK_String = NULL;
    return token;
}

/*JS*********************************************************************
 *   Get special token as a string.
 *************************************************************************/

static jsctoken_t getSpecial(jsppstream_t *jps, int value)

/************************************************************************/
{
    jsppfile_t *iPtr;
    jsctoken_t token;
    static char Buffer[10];

    /*  Scan for current file  */
    iPtr = jps->JSPP_File;
    while(!IF_ISFILE(iPtr))
	iPtr = iPtr->IF_Next;

    token.TK_Flags = 0;
    switch(value) {
    case PTK___LINE: /* **  LINE  ** */
	/*  Use "IF_LineNr" since we wanna have the line number when
	 *   the "__LINE__" token started, not when it ended.
	 */
	sprintf(Buffer, "%d", iPtr->IF_LineNr);

	token.TK_Value = CTK_NUMBER;
	token.TK_String = Buffer;
	break;

    case PTK___FILE: /* **  FILE  ** */
	token.TK_Value = CTK_STRING;
	token.TK_String = iPtr->IF_Name;
	break;

#ifdef CONFIG_JSCPP_INCLEVEL
    case PTK___INCLEVEL: /* **  INCLUDE LEVEL  ** */
	sprintf(Buffer, "%d", jps->JSPP_NFiles - 1);

	token.TK_Value = CTK_NUMBER;
	token.TK_String = Buffer;
	break;
#endif

#ifdef CONFIG_JSCPP_BASEFILE
    case PTK___BASEFILE: /* **  BASE FILE ** */
	token.TK_Value = CTK_STRING;
	token.TK_String = iPtr->IF_Name;

	while((iPtr = iPtr->IF_Next))
	    if(IF_ISFILE(iPtr))
		token.TK_String = iPtr->IF_Name;

	break;
#endif

    default:
	token.TK_Value = CTK_ERROR;
	token.TK_String = NULL;
	break;
    }

    return token;
}

/*JS*********************************************************************
 *   The tokens are concatenated together to yield a new token if possible
 *    (ANSI's '##' operator). The concatenated tokens are stored in a
 *    static buffer that is reused. This might crash if tk2 points to this
 *    buffer, too.
 *    Returns the number of tokens that were successfully concatenated,
 *    at least 1. It returns -1 on error.
 *************************************************************************/

static int doConcat(jsppstream_t *jps, jsctoken_t *dest,
                    jsctoken_t tokens[], int n)

/************************************************************************/
{
    static char *Buffer;
    static int MaxB;
    jsctoken_t curr, res;
    int count;

    for(curr = tokens[0], count = 1 ; count < n ; curr = res, ++count) {
	jsctoken_t next;
	int len, flag;

	flag = 3;

	next = tokens[count];

        /*printf("CONCAT %d %d : '%s' '%s'\n", curr.TK_Value, next.TK_Value, curr.TK_String, next.TK_String);*/

	/*  Check all possibilities that are allowed        */
	if(curr.TK_Value == CTK_STRING && next.TK_Value == CTK_STRING) {
	    /*  String concatenation  */
	    res.TK_Value = CTK_STRING;

	} else if(curr.TK_Value == CTK_WSTRING &&
		  (next.TK_Value == CTK_STRING || next.TK_Value == CTK_WSTRING)) {
	    /*  String concatenation  */
	    res.TK_Value = CTK_WSTRING;

	} else if(curr.TK_Value == CTK_CHARCONST &&
		  (next.TK_Value == CTK_CHARCONST ||
		   next.TK_Value == CTK_WCHARCONST)) {
	    /*  Character constants concatenation  */
	    res.TK_Value = CTK_CHARCONST;

	} else if(curr.TK_Value == CTK_WCHARCONST &&
		  (next.TK_Value == CTK_CHARCONST ||
		   next.TK_Value == CTK_WCHARCONST)) {
	    /*  Character constants concatenation  */
	    res.TK_Value = CTK_WCHARCONST;

	} else if(curr.TK_Value == CTK_IDENT &&
		  (next.TK_Value == CTK_IDENT || next.TK_Value == CTK_NUMBER ||
                   next.TK_Value == CTK_INUMBER || next.TK_Value == CTK_DNUMBER)) {
	    /*  ident + some digits yields again an ident  */
	    res.TK_Value = CTK_IDENT;

	} else if((curr.TK_Value == CTK_NUMBER || curr.TK_Value == CTK_INUMBER) &&
		  (next.TK_Value == CTK_NUMBER || next.TK_Value == CTK_INUMBER)) {
	    /*  Two numbers will again be a number (hopefully :)  */
	    res.TK_Value = CTK_NUMBER;

        } else if(curr.TK_Value == CTK_DNUMBER &&
                  (next.TK_Value == CTK_NUMBER || next.TK_Value == CTK_INUMBER || next.TK_Value == CTK_DNUMBER)) {
            /*  Two numbers will again be a number (hopefully :)  */
            res.TK_Value = CTK_DNUMBER;

	} else if((curr.TK_Value == CTK_NUMBER || curr.TK_Value == CTK_INUMBER) &&
                  next.TK_String && isValidSuffix(next.TK_String)) {
	    res.TK_Value = CTK_INUMBER;

	} else if(curr.TK_Value == CTK_IDENT && jsstricmp(curr.TK_String, "L") == 0 &&
		  (next.TK_Value == CTK_STRING || next.TK_Value ==CTK_CHARCONST)){
	    res.TK_Value = next.TK_Value == CTK_STRING ? CTK_WSTRING :
		CTK_WCHARCONST;
	    flag = 2; /*  only take second string  */

	} else if(curr.TK_Value == CTK_IDENT &&
		  jsstricmp(curr.TK_String, "u") == 0 &&
		  (next.TK_Value == CTK_STRING || next.TK_Value ==CTK_CHARCONST)) {
	    res.TK_Value = next.TK_Value == CTK_STRING ? CTK_USTRING :
		CTK_UCHARCONST;
	    flag = 2; /*  only take second string  */

	} else if(jps->JSPP_Ops && isOperator(jps, curr) && isOperator(jps, next)) {
	    static char SmallBuffer[10];

	    /*  Operator concatenation        */
	    if(curr.TK_String == NULL) {
		SmallBuffer[0] = curr.TK_Value;
		len = 1;
	    } else {
		strcpy(SmallBuffer, curr.TK_String);
		len = strlen(SmallBuffer);
	    }

	    if(next.TK_String == NULL) {
		SmallBuffer[len] = next.TK_Value;
		SmallBuffer[len + 1] = '\0';
	    } else
		strcpy(SmallBuffer + len, next.TK_String);

	    /*  Search in operator table for valid operator   */
	    for(len = 0 ; jps->JSPP_Ops[len].KD_String ; len++)
		if(tkStrcmp(jps->JSPP_Ops[len].KD_String, SmallBuffer) == 0)
		    break;

	    if(jps->JSPP_Ops[len].KD_String == NULL)
		/*  No concatenation possible  */
		break;

	    /*  Found token  */
	    res.TK_Flags  = 0;
	    res.TK_Value  = jps->JSPP_Ops[len].KD_Token;
	    res.TK_String = jps->JSPP_Ops[len].KD_String;

	    continue;

	} else
	    /*  This concatenation doesn't make sense  */
	    break;

	/*  Generate new long token in static buffer        */
	len = 1;
	if(flag & 1) len += strlen(curr.TK_String);
	if(flag & 2) len += strlen(next.TK_String);

	if(len >= MaxB) {
	    MaxB = len;
	    if((Buffer = (char*)realloc(Buffer, MaxB)) == NULL) {
		MaxB = 0;
		EGOTO(error);
	    }
	}

	if((flag & 1) && curr.TK_String != Buffer) {
	    strcpy(Buffer, curr.TK_String);
	    len = strlen(curr.TK_String);
	} else
	    len = 0;

	if(flag & 2)
	    strcpy(Buffer + len, next.TK_String);

	res.TK_String = Buffer;
    }

    *dest = curr;

    return count;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int isValidSuffix(const char *string)

/************************************************************************/
{
    for( ; *string ; ++string) {
        int c;

        c = tolower(*string);
        if(c != 'l' && c != 'u')
            return FALSE;
    }

    return TRUE;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int isOperator(jsppstream_t *jps, jsctoken_t tk)

/************************************************************************/
{
    if(tk.TK_Value >= CTK_EOF && tk.TK_Value < 256 && ispunct(tk.TK_Value))
	return TRUE;

    /* Search in operator table for valid operator  */
    if(jps->JSPP_Ops != NULL && tk.TK_String != NULL) {
	int i;

	for(i = 0 ; jps->JSPP_Ops[i].KD_String ; ++i)
	    if(tkStrcmp(jps->JSPP_Ops[i].KD_String, tk.TK_String) == 0)
		return TRUE;
    }

    return FALSE;
}
