/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Macro expansion code.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include <assert.h>

#define JSCPP_INTERNAL

#include <jscpplib.h>

/*********     DEFINES                                          *********/
#ifndef GTK_NO_EXP
/*  START-DEFINITIONS  */
#ifdef JSCPP_INTERNAL
/*  Flags for routine jspptkmacro and _jspptkbasic  */
#define GTK_NO_EXP           (GCTK_FLAG_USER)
#define GTK_NO_CONCAT        (GCTK_FLAG_USER << 1)
#define GTK_PREPROCOPS       (GCTK_FLAG_USER << 2)

/*  Flags for routine _jsppexpand  */
#define EXPM_DEFINED_MACRO   (GCTK_FLAG_USER << 4)
#endif
#define EXPM_DEFEREXPANSION  (GCTK_FLAG_USER << 5)
#ifdef JSCPP_INTERNAL
#define EXPM_SAVE_ORIG       (GCTK_FLAG_USER << 6)
#define EXPM_SAVE_ORIG2      (GCTK_FLAG_USER << 7)
#endif
/*  END-DEFINITIONS  */
#endif

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
Prototype jsctoken_t     jspptkmacro(jsppstream_t *jps,
                                     jsctoken_t *pOToken, int mode);
static jsctoken_t        _jspptkmacro2(jsppstream_t *jps,
                                       jsctoken_t *pOToken, int mode);
IPrototype int           _jsppexpand(jsppstream_t *jps,
                                     const char *name, int mode);

/*JS*********************************************************************
 *   Wrapper for getTkMacro2 to prevent PTK_IDENTNOEXP to be seen outside.
 *************************************************************************/

jsctoken_t jspptkmacro(jsppstream_t *jps, jsctoken_t *pOToken, int mode)

/************************************************************************/
{
    jsctoken_t token;

    token = _jspptkmacro2(jps, pOToken, mode);

    if(token.TK_Value == PTK_IDENTNOEXP)
	token.TK_Value = CTK_IDENT;

    if(pOToken && pOToken->TK_Value == PTK_IDENTNOEXP)
	pOToken->TK_Value = CTK_IDENT;

    return token;
}

/*JS*********************************************************************
 *   Get next token with routine getTkBasic and check for macro expansion.
 *************************************************************************/

static jsctoken_t _jspptkmacro2(jsppstream_t *jps, jsctoken_t *pOToken,
                                int mode)

/************************************************************************/
{
    jsctoken_t token;

    for(;;) {
	if(pOToken) {
	    /*  Read from macro's queue for original tokens  */
	    *pOToken = jctokget(&(jps->JSPP_OQueue[jps->JSPP_QueueLevel].
				  JSOQ_MacOQueue));

	    /*  Queue empty?  */
	    if(pOToken->TK_Value != CTK_EOF) {
		/*  Return only original token, no preprocessed one  */
		token.TK_Flags = 0;
		token.TK_Value = PTK_EMPTY;
		token.TK_String = NULL;
		break;
	    }

	    /*  Clear "#define"-line-mode after exhausting orig queue  */
	    jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_Mode = 0;
	}

	/*  Read next token, may be CTK_EOF too  */
	token = _jspptkbasic2(jps, pOToken, mode);

	/*  Check for macro expansion  */
	if(token.TK_Value == CTK_IDENT && !(mode & GTK_NO_EXP)) {
	    int ret, mode2;

	    /*  Do not expand while collecting macro's arguments  */
	    mode2 = mode | GTK_NO_EXP;
	    if(pOToken) {
		/*  We are interested in original tokens, so if we tap into
		 *   original ones while e. g. looking for macro arguments,
		 *   reenable saving them
		 */
		mode2 |= EXPM_SAVE_ORIG2;

		if(IF_ISFILE(jps->JSPP_File) ||
		   (jps->JSPP_File->IF_Flag & (IF_FLAG_ORIGTOK | IF_FLAG_EXPANDORIG)))
		    /*  We are already within orginal ones, so save everything,
		     *   including the name of this macro.
		     */
		    mode2 |= EXPM_SAVE_ORIG;
	    }

	    if((ret = _jsppexpand(jps, token.TK_String, mode2)) == -1)
		EGOTO(error);

	    /*  Return first the sequence of original tokens  */
	    if(ret == 0)
		continue;

	    /*  Mark to prohibit reexpansion of macro  */
	    if(ret == 2) {
		token.TK_Flags = 0;
		token.TK_Value = PTK_IDENTNOEXP;
	    }
	}
#ifdef CONFIG_JSCPP_ASSERTION
	else if((mode & EXPM_DEFINED_MACRO) && token.TK_Value == '#') {
	    int mode2;

	    mode2 = mode;
	    if(pOToken) {
		mode2 |= EXPM_SAVE_ORIG2;
		if(IF_ISFILE(jps->JSPP_File) ||
		   ((jps->JSPP_Mode & JPPM_EXPUSERMACROS) && jsppisusrinclude(jps, 0)))
		    mode2 |= EXPM_SAVE_ORIG;
	    }

	    if(_jsppppdefass(jps, mode2, TRUE,
			     &(jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_MacOQueue)) == -1)
		EGOTO(error);

	    continue;
	}
#endif
	break;
    }

    return token;
 error:
    token.TK_Flags  = 0;
    token.TK_Value  = CTK_ERROR;
    token.TK_String = NULL;
    return token;
}

/*JS*********************************************************************
 *   The macro name has already been read, now expand it. Read parameters
 *    if this is a macro with parameters. Returns 0 if the macro has been
 *    successfully expanded, 1 otherwise. If the macro is tried to be
 *    expanded recursively, it returns 2. A negative value signals an error.
 *************************************************************************/

int _jsppexpand(jsppstream_t *jps, const char *name, int mode)

/************************************************************************/
{
    _jsppmacentry_t *entry;
    int ptrIndex;
    jsctoken_t token, *ptr;
    _jsppmacarg_t *mArgs;
    long macFlags;
    unsigned int origIndexName = -1;

    /*  Initialize pointers  */
    ptr = NULL;
    mArgs = NULL;

    /*  Recognition of special preprocessor macro "defined(name)"?  */
    if((mode & EXPM_DEFINED_MACRO) && tkStrcmp("defined", name) == 0)
	return _jsppppdefass(jps, mode, FALSE,
			     &(jps->JSPP_OQueue[jps->JSPP_QueueLevel].
			       JSOQ_MacOQueue));

    /*  Any macro definition for identifier name?  */
    if((entry = _jsppfindmacentry(jps, name)) == NULL)
	return 1;

    if((macFlags = entry->ME_Flags) & MAFLAG_UNDEFINED)
	return 1;

    if(mode & EXPM_DEFEREXPANSION) {
	/*  Do not expand macros that have already been expanded  */
	if(IF_ISATEOF(jps->JSPP_File->IF_Next) &&
	   !(macFlags & MAFLAG_DEFEREXPANSION))
	    return 1;

    } else {
	/*  Defer expansion of marked macros  */
	if(macFlags & MAFLAG_DEFEREXPANSION) {
	    jps->JSPP_Mode |= JPPM_FLAG_SEENDEFEREDMAC;
	    return 1;
	}
    }

    /*  Check if expansion is disabled  */
    if((macFlags & MAFLAG_DISABLED) ||
       ((macFlags & MAFLAG_DISABLEDINFILE) && IF_ISFILE(jps->JSPP_File)))
	return 1;

    if((jps->JSPP_Mode & JPPM_EXPUSERMACROS) && !(macFlags & MAFLAG_SYSTEMHEADER))
	/*  We basically treat this as if EXPANDORIG was requested for
	 *   this macro.
	 */
	macFlags |= MAFLAG_EXPANDORIG;

    /*  Check if this macro is recursively expanded  */
    {
	jsppfile_t *iPtr;

	for(iPtr = jps->JSPP_File ; iPtr && IF_ISMACRO(iPtr) ;
	    iPtr = iPtr->IF_Next) {
	    /*   We cannot expand it if this is within a macro expansion itself  */
	    macFlags &= ~(MAFLAG_EXPANDORIG | MAFLAG_EXPANDORIGFULL);

	    if(iPtr->IF_Flag & IF_FLAG_EXPANDINGARG) {
		/*  Within expansion of a macros parameter, don't check the parent
		 *   macro of this parameter. Otherwise, in a case like "MAC(MAC(x))",
		 *   the inner macro call will not be expanded because we think it is
		 *   a recursive expansion.
		 */
		iPtr = iPtr->IF_Next;

	    } else if((iPtr->IF_Flag & IF_FLAG_DISABLEEXPAND) ||
		      (iPtr->IF_Name &&
		       tkStrcmp(iPtr->IF_Name, entry->ME_Ident) == 0)) {
		return 2; /* Don't expand  */
	    }
	}
    }

    if(macFlags & (MAFLAG_EXPANDORIG | MAFLAG_EXPANDORIGFULL))
	mode &= ~(EXPM_SAVE_ORIG | EXPM_SAVE_ORIG2);

    /*  FIXME: This seems to work experimentally  */
    {
	jsppfile_t *iPtr;

	for(iPtr = jps->JSPP_File ; iPtr ; iPtr = iPtr->IF_Next) {
	    if(iPtr->IF_Flag & IF_FLAG_EXPANDORIG) {
		mode |= EXPM_SAVE_ORIG;
		break;
	    }
	    break; /*  FIXME: this is rubbish  */
	}
    }

    /*  Put name into original queue  */
    if(mode & EXPM_SAVE_ORIG) {
	/*  Preset token with macro's name  */
	token.TK_Flags = 0; token.TK_Value = CTK_IDENT; token.TK_String = entry->ME_Ident;

	if((origIndexName = jctokinsert(&(jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_MacOQueue),
				    token, 0)) == -1)
	    EGOTO(error);
    }

    if(entry->ME_NArgs >= 0) {
	unsigned int ptrMax, iArg;
	int brackets;

	/*  Allocate a list for the arguments  */
	if((mArgs = (_jsppmacarg_t *)
	    malloc((entry->ME_NArgs + 2) * sizeof(*mArgs))) == NULL)
	    EGOTO(error);

	ptrMax = 0;
	ptrIndex = 0;

	/*  First entry of parameters contains pointer to free  */
	mArgs[0].MA_Token = ptr; /* == NULL  */
	mArgs[0].MA_UseCount = countnlCToken(name);
	iArg = 1; /*  start with first argument  */

	/*  Read parameters  */
	for(brackets = -1 ; brackets >= -1 ; ) {
            unsigned int origIndex, flag;

	    /*  Save original value of token if wanted  */
	    origIndex = -1;

	    if(mode & (EXPM_SAVE_ORIG | EXPM_SAVE_ORIG2)) {
		jsctoken_t oToken;

		if((jps->JSPP_Mode & JPPM_PPMACARGS) && brackets >= 0) {
		    jps->JSPP_QueueLevel++;
		    if(jps->JSPP_QueueLevel >= jps->JSPP_MaxQueueLevel) {
			unsigned int maxQueue = jps->JSPP_MaxQueueLevel;

			/*  Increase the number of queues by one  */
			if((jps->JSPP_OQueue = (_jsppoqueue_t *)
			    realloc(jps->JSPP_OQueue, (maxQueue + 1) *
				    sizeof(*(jps->JSPP_OQueue)))) == NULL)
			    EGOTO(error);

			if(jctokcreate(&(jps->JSPP_OQueue[maxQueue].JSOQ_OQueue), 0) == NULL ||
			   jctokcreate(&(jps->JSPP_OQueue[maxQueue].JSOQ_MacOQueue), 0) == NULL)
			    EGOTO(error);

			jps->JSPP_MaxQueueLevel++;
		    }
		    jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_Mode = 0;

		    /*  jsppgettoken might call us recursively if during the
		     *   argument collection we again see conditionals which
		     *   again have macros etc. We need to make sure it is
		     *   using different queues. Otherwise, a construct like
		     *
		     *     #define A    B(a)
		     *     #define B(x) 1
		     *     #if A == A
		     *     #endif
		     *     #if A == B(x)
		     *     #endif
		     *
		     *   gives the wrong result for the original tokens in the
		     *   '#if' lines.
		     *
		     *  The whole stuff is only relevant for the strange
		     *   non-ANSI mode "JPPM_PPMACARGS".
		     */
		    token = jsppgettoken(jps, &oToken, mode);
		    jps->JSPP_QueueLevel--;

		} else
		    token = _jspptkbasic2(jps, &oToken, mode);

		if(IF_ISFILE(jps->JSPP_File))
		    /*  We tapped into original tokens, so start saving them.
		     *   This applies in cases like the following:
		     *
		     *    #define MACRO(arg)   This is "MACRO" with ARG arg
		     *    #define Macro        MACRO
		     *    Macro(one)
		     *
		     *   We expand "Macro" to "MACRO". While we re-scan "MACRO",
		     *   we disable looking for originals, since we don't want to
		     *   see "MACRO" in the original stream. But "MACRO" looks for
		     *   its arguments which stem from the original stream. So we
		     *   have to reenable scanning for originals.
		     */
		    mode |= EXPM_SAVE_ORIG;

		if(oToken.TK_Value != PTK_EMPTY && (mode & EXPM_SAVE_ORIG)) {
		    if((origIndex = jctokinsert(&(jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_MacOQueue),
						oToken, 0)) == (unsigned int)-1)
			EGOTO(error);

                    if(oToken.TK_Value != token.TK_Value)
			origIndex = -1;
		}

	    } else {
		token = (jps->JSPP_Mode & JPPM_PPMACARGS) && brackets >= 0 ?
		    jsppgettoken(jps, NULL, mode) :
		    _jspptkbasic2(jps, NULL, mode);
	    }

	    /*  Enough memory?  */
	    if((ptrIndex + 1) >= ptrMax) {
		jsctoken_t *oPtr = ptr;
		unsigned int i;

		/*  Increase memory  */
		ptrMax += 10;
		if((ptr = (jsctoken_t *)realloc(ptr, ptrMax * sizeof(*ptr))) == NULL)
		    EGOTO(error);

		/*  Change mArgs array  */
		for(i = 0 ; i < iArg ; i++)
		    /*  The expression
		     *
		     *      mArgs[i].MA_Token += ptr - oPtr;
		     *
		     *   looks syntactically equivalent to the one below,
		     *   but produces wrong results if the alignment
		     *   of *ptr is lower than sizeof(*ptr).
		     *
		     *  Example: (sizeof(*ptr) == 16)
		     *
		     *    mArgs[i].MA_Token = oPtr = NULL;
		     *    ptr = 24;
		     *
		     *   Then (ptr - oPtr) == 1, but after the above statement,
		     *
		     *    mArgs[i].MA_Token == 16 !!!
		     */
		    mArgs[i].MA_Token = &ptr[mArgs[i].MA_Token - oPtr];
	    }

	    if(token.TK_Value == CTK_EOF) {
		if(brackets >= 0) {
		    jsppprinterr(jps, "EOF while scanning macro \"%s\"", 1,
				 entry->ME_Ident);
		    EGOTO(error);
		}
		goto macNoParam;
	    }

	    if(brackets < 0) {
		/*  Look for opening bracket  */
		if(token.TK_Value == '(') {
		    if(entry->ME_NArgs > 0) {
			flag = 0;

			/*  Start of first argument  */
			mArgs[iArg].MA_Token = &ptr[ptrIndex + 1];
			mArgs[iArg].MA_UseCount = 0;
			iArg++;
		    }

		    brackets = 0;

		    /*  Generate end marker for space token between ident and '('  */
		    token.TK_Value = CTK_EOF;
		    origIndex = -1;

		} else if(!(CTK_BLANK_START <= token.TK_Value &&
			    token.TK_Value <= CTK_BLANK_END))
		    goto macNoParam;

	    } else if(brackets == 0) {
		if(token.TK_Value == ',' &&
		   !(macFlags & MAFLAG_ONEARGUMENT) &&
		   !((macFlags & MAFLAG_VARARG) && (iArg - 1) == entry->ME_NArgs)) {
		    token.TK_Value = CTK_EOF; /*  Generate end marker  */
		    origIndex = -1;

		    /*  Set next argument pointer  */
		    if(iArg > entry->ME_NArgs) {
			jsppprinterr(jps, "Too many arguments for macro \"%s\"", 1, entry->ME_Ident);
			EGOTO(error);
		    }

		    mArgs[iArg].MA_Token = &ptr[ptrIndex + 1];
		    mArgs[iArg].MA_UseCount = 0;
		    iArg++;

		} else if(token.TK_Value == ')') {
		    /*  Mark last argument  */
		    mArgs[iArg].MA_Token = NULL;
		    mArgs[iArg].MA_UseCount = 0;
		    iArg++;

		    token.TK_Value = CTK_EOF; /*  Generate end marker  */
		    origIndex = -1;

		    brackets = -2; /*  last argument found  */
		}
	    }

	    if(token.TK_Value == '(')
		++brackets;
	    else if(token.TK_Value == ')')
		--brackets;

	    /*  Check if macro arguments are empty  */
	    if(token.TK_Value != PTK_EMPTY) {
		/*  Any non blanks in last macro argument?  */
		if(token.TK_Value == CTK_EOF) {
		    if(entry->ME_NArgs > 0 && iArg > 2 && flag == 0) {
			if(!(jps->JSPP_Mode & JPPM_LAZYSTYLE) ||
			   (jps->JSPP_Mode & JPPM_PRWARNING))
			    jsppprinterr(jps, "Argument %d to macro \"%s\" is empty",
					 (jps->JSPP_Mode & JPPM_LAZYSTYLE) ? 0 : 1,
					 iArg - 2, entry->ME_Ident);

			if(!(jps->JSPP_Mode & JPPM_LAZYSTYLE))
			    EGOTO(error);
		    }

		    flag = 0;

                } else if(token.TK_Value < CTK_BLANK_START || CTK_BLANK_END < token.TK_Value) {
		    /*  Mark if macro argument contains any non space  */
		    flag = 1;
		}

		if(origIndex != (unsigned int)-1) {
		    jsctoken_t otoken;

		    otoken = jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_MacOQueue.TKFF_Token[origIndex];

                    /*printf("HAVE %d %d '%s' '%s'\n", otoken.TK_Value, token.TK_Value,
                      otoken.TK_String, token.TK_String);*/

		    assert(otoken.TK_Value == token.TK_Value &&
			   (otoken.TK_String == NULL ? token.TK_String == NULL :
			    token.TK_String != NULL &&
			    strcmp(token.TK_String, otoken.TK_String) == 0));

		    ptr[ptrIndex] = copyCToken(otoken, 1);
                    ptr[ptrIndex].TK_Flags |= CTOKEN_FORIGTOKEN;

                } else {
                    int cmode;

                    cmode = token.TK_Flags & CTOKEN_FORIGTOKEN ? 1 : 0;
                    ptr[ptrIndex] = copyCToken(token, cmode);
		}

		ptrIndex++;
	    }
	}

	/*  Check correct number of arguments  */
	if((macFlags & MAFLAG_VARARG) ? (iArg - 2) < entry->ME_NArgs :
	   (iArg - 2) != entry->ME_NArgs) {
	    jsppprinterr(jps, "To few parameters %d for macro \"%s\", expected %d", 1, iArg - 2,
			 entry->ME_Ident, entry->ME_NArgs);
	    EGOTO(error);
	}

#ifdef XDEBUG
	{
	    unsigned int i, j;

	    printf("\n***Collected arguments \"%s\" (%d %d %d):\n", entry->ME_Ident,
		   iArg, ptrIndex, ptrMax);

	    for(i = 0 ; mArgs[i].MA_Token ; i++) {
                printf("Arg[%d]: <", i);

		for(j = 0 ; mArgs[i].MA_Token[j].TK_Value != CTK_EOF ; j++) {
		    putchar('!');
		    printCToken(stdout, mArgs[i].MA_Token[j]);
		    putchar('@');
		}

		printf(">\n");
	    }
#if 0
	    for(i = 0 ; entry->ME_Token[i].TK_Value != CTK_EOF ; i++)
		printf("  %d: %d\n", i, entry->ME_Token[i].TK_Value);
#endif
	}
#endif
    } else {
	unsigned int count;

	if((count = countnlCToken(name)) > 0) {
	    /*  Need at least one entry to save number of newlines in macro name
	     *   so that line number synchronization works correctly.
	     */
	    if((mArgs = (_jsppmacarg_t *)malloc(2 * sizeof(*mArgs))) == NULL) EGOTO(error);
	    mArgs[0].MA_Token = mArgs[1].MA_Token = NULL;
	    mArgs[0].MA_UseCount = count;
	    mArgs[1].MA_UseCount = 0;
	}
    }

    {
	long pushFlags = (macFlags & MAFLAG_SYSTEMHEADER) ? IF_SYSINCLUDE : 0;

	/*  Any space token between identifier and '('?  */
	/*  We push this *before* the macro expansion, so it actually comes
	 *   *after* it. This ensures that we don't have some spaces in front
	 *   of a macro expansion.
	 *   When the parser creates nodes, it looks if there are only empty
	 *   tokens in front of a token. The node interval then contains the
	 *   original tokens as well. This does not work if we have first the
	 *   original tokens, then space tokens and then the preprocessed
	 *   tokens.
	 */
	if(mArgs && mArgs[0].MA_Token && mArgs[0].MA_Token[0].TK_Value != CTK_EOF &&
	   _jspppushmacro(jps, NULL, mArgs[0].MA_Token, NULL, pushFlags |
			  IF_FLAG_EXPANDINGARG | IF_FLAG_COPYTOKS) == -1)
	    EGOTO(error);

	/*  Push macro expansion  */
	if(macFlags & MAFLAG_EXPANDORIGFULL)
	    pushFlags |= IF_FLAG_EXPANDORIGFULL;
	else if(macFlags & MAFLAG_EXPANDORIG)
	    pushFlags |= IF_FLAG_EXPANDORIG;

	if(_jspppushmacro(jps, entry->ME_Ident, entry->ME_Token, mArgs, pushFlags) == -1)
	    EGOTO(error);
    }

    entry->ME_Flags |= MAFLAG_USED;

    return 0;
 error:
    if(ptr) {
	unsigned int i;

	for(i = 0 ; i < ptrIndex ; i++)
	    freeCToken(ptr[i]);
	free(ptr);
    }

    if(mArgs) free(mArgs);
    return -1;
 macNoParam: /* ***  Macro with parameters called without  *** */
    {
	static jsctoken_t TName[2] = { CTOKEN_GENSTATIC(CTK_IDENT, NULL),
				       CTOKEN_GENSTATIC(CTK_EOF,   NULL) };
	unsigned int pMode;

	/*  Delete collected original tokens  */
	if(origIndexName != (unsigned int)-1)
	    jctokdelete(&(jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_MacOQueue),
			origIndexName, -1, 0);

	/*  Push read tokens as pseudo macro expansion  */
	ptr[ptrIndex++] = copyCToken(token, 0); /*  copy this one  */

	/*  Generate finisher  */
	token.TK_Flags = 0;
	token.TK_Value = CTK_EOF;
	token.TK_String = NULL;
	ptr[ptrIndex] = copyCToken(token, 0);

	mArgs[1].MA_Token = NULL; /*  No parameters  */
	mArgs[1].MA_UseCount = 0;

	pMode = (mode & EXPM_SAVE_ORIG) ? IF_FLAG_ORIGTOK : 0;

	/*  Push lookahead tokens as macro expansion.
	 *   Use an invalid name (e.g. a space) so
	 *   that the read token will be expanded
	 *   if applicable.
	 */
	if(_jspppushmacro(jps, NULL, ptr, mArgs, pMode) == -1)
	    EGOTO(error);

	/*  But before that, put missing macro name
	 *   as expansion of itself to prevent
	 *   infinite recursion of this routine.
	 */
	TName[0].TK_String = entry->ME_Ident;
	if(_jspppushmacro(jps, entry->ME_Ident, TName, NULL, pMode) == -1)
	    EGOTO(error);

#ifdef XDEBUG
	{
	    unsigned int i;

	    printf("\n***Collected tokens : <");
	    for(i = 0 ; ptr[i].TK_Value != CTK_EOF ; i++)
		printCToken(stdout, ptr[i]);
	    printf(">\n");
	}
#endif
    }
    return 0; /*  signal macro expansion was successful  */
}
