/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Queue reader
 *
 *              Copyright (C) 2004-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/
#include <assert.h>

/*#define JSCPP_INTERNAL*/ /*  only public things are allowed  */
#include <jscpplib.h>

/*********     DEFINES						*********/
#ifndef jsqsgetqueue
/*  START-DEFINITIONS  */
typedef struct jsqueuestream_t jsqueuestream_t;
/*  END-DEFINITIONS c */

/*  START-DEFINITIONS 3 */
#define JSQS_TYPE_PPSTREAM    1
#define JSQS_TYPE_USER        2
#define JSQS_TYPE_FMASK       0xff

#define JSQS_TYPE_FFREEQUEUE  (1<<8)

struct jsqueuestream_t {
    int JSQS_Type;
    jctokfifo_t *JSQS_Queue;

    union {
	jsppstream_t *jsqs_PPStream;
	int (*jsqs_GetIndex)(void *user, jctokfifo_t *queue);
    } c;
#define JSQS_PPStream         c.jsqs_PPStream
#define JSQS_GetIndex         c.jsqs_GetIndex

    void *JSQS_User;
};

#define jsqsgetqueue(jsqs)     ((jsqs)->JSQS_Queue)

#define jsqsgetppstream(jsqs)  (((jsqs)->JSQS_Type & JSQS_TYPE_FMASK) == \
				JSQS_TYPE_PPSTREAM ? (jsqs)->JSQS_PPStream : \
				NULL)
/*  END-DEFINITIONS  */
#endif

#define Prototype extern
/*********     PROTOTYPES					*********/
Prototype jsqueuestream_t *jsqscreateppstream(jsppstream_t *jps,
					      jctokfifo_t *queue, int qmode);
Prototype jsqueuestream_t *jsqscreatequeuestream(jctokfifo_t *queue,
						 int (*routine)(void *,
								jctokfifo_t *),
						 void *user);
Prototype void           jsqsfree(jsqueuestream_t *jsqs);
Prototype int            jsqsget(jsqueuestream_t *jsqs, int mode);

/*JS*********************************************************************
 *
 *************************************************************************/

jsqueuestream_t *jsqscreateppstream(jsppstream_t *jps, jctokfifo_t *queue,
				    int qmode)

/************************************************************************/
{
    jsqueuestream_t *jsqs;

    if((jsqs = (jsqueuestream_t *)malloc(sizeof(*jsqs))) == NULL)
	return NULL;

    jsqs->JSQS_Type = JSQS_TYPE_PPSTREAM;

    if(queue == NULL) {
	if((queue = jctokcreate(NULL, qmode)) == NULL) {
	    free(jsqs);
	    return NULL;
	}

	jsqs->JSQS_Type |= JSQS_TYPE_FFREEQUEUE;
    }

    jsqs->JSQS_Queue     = queue;
    jsqs->JSQS_PPStream  = jps;
    jsqs->JSQS_User      = NULL;

    return jsqs;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsqueuestream_t *jsqscreatequeuestream(jctokfifo_t *queue,
				       int (*routine)(void *, jctokfifo_t *),
				       void *user)

/************************************************************************/
{
    jsqueuestream_t *jsqs;

    if((jsqs = (jsqueuestream_t *)malloc(sizeof(*jsqs))) == NULL)
	return NULL;

    jsqs->JSQS_Type = JSQS_TYPE_USER;

    if(queue == NULL) {
	if((queue = jctokcreate(NULL, 0)) == NULL) {
	    free(jsqs);
	    return NULL;
	}

	jsqs->JSQS_Type |= JSQS_TYPE_FFREEQUEUE;
    }

    jsqs->JSQS_Queue = queue;
    jsqs->JSQS_GetIndex = routine;
    jsqs->JSQS_User = user;

    return jsqs;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsqsfree(jsqueuestream_t *jsqs)

/************************************************************************/
{
    if(jsqs->JSQS_Type & JSQS_TYPE_FFREEQUEUE)
	jctokfree(jsqs->JSQS_Queue, 1);

    free(jsqs);
}

/*JS*********************************************************************
 *   If bit 0 of mode is set, return the preprocessed tokens as original
 *    ones outside of system includes and macro expansion of macros
 *    defined in system includes.
 *************************************************************************/

int jsqsget(jsqueuestream_t *jsqs, int mode)

/************************************************************************/
{
    if((jsqs->JSQS_Type & JSQS_TYPE_FMASK) == JSQS_TYPE_PPSTREAM) {
	jsppstream_t *jps;
	jctokfifo_t  *queue;
	jsctoken_t    tlist[JCTOK_MAXSTEP];
	unsigned int  ind, wmode;

	queue = jsqsgetqueue(jsqs);
	jps = jsqs->JSQS_PPStream;
	wmode = JCTOK_MODE_INSERT;

	if((ind = jctokindorig(queue)) == -1U) {
	    tlist[0] = jsppgettoken(jps, NULL, 0);

	} else {
	    /*  Outside system includes and expansions of macros defined
	     *   in system includes, we return the preprocessed tokens
	     *   as the original stream.
	     */
	    tlist[0] = jsppgettoken(jps, &tlist[ind], 0);

	    /*  Sequence matters! *After* getting the token we check if we are
	     *   within a system include file
	     */
	    if((mode & 1) && jsppisusrinclude(jps, 0)) {
#if 0 /*  FIXME: Expanding the user defined includes first, but then still
       *   keeping them in the output is a bad idea.
       *   See the docu for an extensive discussion.
       */
		/*  Check if this is a "#define" line we should preserve  */
		if(!(tlist[0].TK_Value == PTK_EMPTY &&
		     (jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_Mode & 1)))
#endif
		    tlist[ind] = tlist[0];
	    }

	    /*  We want identical tokens to have identical pointers in TK_String,
	     *   so if one side is empty, we enabled checking the pointers!
	     */
	    if(tlist[0].TK_Value == PTK_EMPTY || tlist[ind].TK_Value == PTK_EMPTY)
		wmode |= JCTOK_MODE_CHKPTR;
	    else
		/*  otherwise we leave a region of preprocessed stuff and can
		 *   safely forget our pointer memory.
		 */
		jctokclearptrmap(queue);
	}

	if(tlist[0].TK_Value == CTK_EOF)
	    return -2;

	if((ind = jctokindfpos(queue)) != -1U) {
	    /*  Store file position in last token  */
	    tlist[ind].TK_Flags = 0;
	    tlist[ind].TK_Value = jsppgetfilepos(jsqs->JSQS_PPStream,
						 (const char **)&(tlist[ind].TK_String),
						 1);
	}

	/*  Save tokens and return position  */
	return jctokwriten(jsqs->JSQS_Queue, tlist, jctokgetstep(queue), -1, wmode);
    }

    assert((jsqs->JSQS_Type & JSQS_TYPE_FMASK) == JSQS_TYPE_USER);

    if(jsqs->JSQS_GetIndex)
	return jsqs->JSQS_GetIndex(jsqs->JSQS_User, jsqs->JSQS_Queue);

    return jctokgetindex(jsqs->JSQS_Queue);
}
