/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Predefined symbols.
 *              Compile this source with "-DSHOW_SYMBOLS" to get a
 *              list of symbols visible.
 *
 *              Copyright (C) 2004-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*  These come *before* any include so we get symbols that stem from
 *   the preprocessor itself!
 */

/*  IGNORE-DEFINE-SECTION  */

/*  This macro will produce the value of the macro as a string  */
#define STRINGIZE(x)   STRINGIZEX(x)
#define STRINGIZEX(x)  #x

static const struct {
    const char *Name;
    const char *Value;
    const int   Mask;
} PreDefSymbols[] = {

    /* *************************** Linux (i386) ***************************** */

    /*  The following four are GNU macros which we normally don't define  */
#ifdef __GNUC__
    { "__GNUC__",               STRINGIZE(__GNUC__),               0x0c },
#endif
#ifdef __GNUC_MINOR__
    { "__GNUC_MINOR__",         STRINGIZE(__GNUC_MINOR__),         0x0c },
#endif
#ifdef __GNUC_PATCHLEVEL__
    { "__GNUC_PATCHLEVEL__",    STRINGIZE(__GNUC_PATCHLEVEL__),    0x0c },
#endif
#ifdef __GXX_ABI_VERSION
    { "__GXX_ABI_VERSION",      STRINGIZE(__GXX_ABI_VERSION),      0x0c },
#endif

#ifdef __CHAR_UNSIGNED__
    { "__CHAR_UNSIGNED__",      STRINGIZE(__CHAR_UNSIGNED__),      0x0f },
#endif
#ifdef __WCHAR_UNSIGNED__
    { "__WCHAR_UNSIGNED__",     STRINGIZE(__WCHAR_UNSIGNED__),     0x0f },
#endif
#ifdef __CHAR_BIT__
    { "__CHAR_BIT__",           STRINGIZE(__CHAR_BIT__),           0x0f },
#endif
#ifdef __SCHAR_MAX__
    { "__SCHAR_MAX__",          STRINGIZE(__SCHAR_MAX__),          0x0f },
#endif
#ifdef __WCHAR_MAX__
    { "__WCHAR_MAX__",          STRINGIZE(__WCHAR_MAX__),          0x0f },
#endif
#ifdef __SHRT_MAX__
    { "__SHRT_MAX__",           STRINGIZE(__SHRT_MAX__),           0x0f },
#endif
#ifdef __INT_MAX__
    { "__INT_MAX__",            STRINGIZE(__INT_MAX__),            0x0f },
#endif
#ifdef __LONG_MAX__
    { "__LONG_MAX__",           STRINGIZE(__LONG_MAX__),           0x0f },
#endif
#ifdef __LONG_LONG_MAX__
    { "__LONG_LONG_MAX__",      STRINGIZE(__LONG_LONG_MAX__),      0x0f },
#endif
#ifdef __FLT_RADIX__
    { "__FLT_RADIX__",          STRINGIZE(__FLT_RADIX__),          0x0f },
#endif
#ifdef __FLT_MANT_DIG__
    { "__FLT_MANT_DIG__",       STRINGIZE(__FLT_MANT_DIG__),       0x0f },
#endif
#ifdef __DBL_MANT_DIG__
    { "__DBL_MANT_DIG__",       STRINGIZE(__DBL_MANT_DIG__),       0x0f },
#endif
#ifdef __LDBL_MANT_DIG__
    { "__LDBL_MANT_DIG__",      STRINGIZE(__LDBL_MANT_DIG__),      0x0f },
#endif
#ifdef __FLT_DIG__
    { "__FLT_DIG__",            STRINGIZE(__FLT_DIG__),            0x0f },
#endif
#ifdef __DBL_DIG__
    { "__DBL_DIG__",            STRINGIZE(__DBL_DIG__),            0x0f },
#endif
#ifdef __LDBL_DIG__
    { "__LDBL_DIG__",           STRINGIZE(__LDBL_DIG__),           0x0f },
#endif
#ifdef __FLT_MIN_EXP__
    { "__FLT_MIN_EXP__",        STRINGIZE(__FLT_MIN_EXP__),        0x0f },
#endif
#ifdef __DBL_MIN_EXP__
    { "__DBL_MIN_EXP__",        STRINGIZE(__DBL_MIN_EXP__),        0x0f },
#endif
#ifdef __LDBL_MIN_EXP__
    { "__LDBL_MIN_EXP__",       STRINGIZE(__LDBL_MIN_EXP__),       0x0f },
#endif
#ifdef __FLT_MIN_10_EXP__
    { "__FLT_MIN_10_EXP__",     STRINGIZE(__FLT_MIN_10_EXP__),     0x0f },
#endif
#ifdef __DBL_MIN_10_EXP__
    { "__DBL_MIN_10_EXP__",     STRINGIZE(__DBL_MIN_10_EXP__),     0x0f },
#endif
#ifdef __LDBL_MIN_10_EXP__
    { "__LDBL_MIN_10_EXP__",    STRINGIZE(__LDBL_MIN_10_EXP__),    0x0f },
#endif
#ifdef __FLT_MAX_EXP__
    { "__FLT_MAX_EXP__",        STRINGIZE(__FLT_MAX_EXP__),        0x0f },
#endif
#ifdef __DBL_MAX_EXP__
    { "__DBL_MAX_EXP__",        STRINGIZE(__DBL_MAX_EXP__),        0x0f },
#endif
#ifdef __LDBL_MAX_EXP__
    { "__LDBL_MAX_EXP__",       STRINGIZE(__LDBL_MAX_EXP__),       0x0f },
#endif
#ifdef __FLT_MAX_10_EXP__
    { "__FLT_MAX_10_EXP__",     STRINGIZE(__FLT_MAX_10_EXP__),     0x0f },
#endif
#ifdef __DBL_MAX_10_EXP__
    { "__DBL_MAX_10_EXP__",     STRINGIZE(__DBL_MAX_10_EXP__),     0x0f },
#endif
#ifdef __LDBL_MAX_10_EXP__
    { "__LDBL_MAX_10_EXP__",    STRINGIZE(__LDBL_MAX_10_EXP__),    0x0f },
#endif
#ifdef __FLT_MAX__
    { "__FLT_MAX__",            STRINGIZE(__FLT_MAX__),            0x0f },
#endif
#ifdef __DBL_MAX__
    { "__DBL_MAX__",            STRINGIZE(__DBL_MAX__),            0x0f },
#endif
#ifdef __LDBL_MAX__
    { "__LDBL_MAX__",           STRINGIZE(__LDBL_MAX__),           0x0f },
#endif
#ifdef __FLT_EPSILON__
    { "__FLT_EPSILON__",        STRINGIZE(__FLT_EPSILON__),        0x0f },
#endif
#ifdef __DBL_EPSILON__
    { "__DBL_EPSILON__",        STRINGIZE(__DBL_EPSILON__),        0x0f },
#endif
#ifdef __LDBL_EPSILON__
    { "__LDBL_EPSILON__",       STRINGIZE(__LDBL_EPSILON__),       0x0f },
#endif
#ifdef __FLT_MIN__
    { "__FLT_MIN__",            STRINGIZE(__FLT_MIN__),            0x0f },
#endif
#ifdef __DBL_MIN__
    { "__DBL_MIN__",            STRINGIZE(__DBL_MIN__),            0x0f },
#endif
#ifdef __LDBL_MIN__
    { "__LDBL_MIN__",           STRINGIZE(__LDBL_MIN__),           0x0f },
#endif
#ifdef __FLT_EVAL_METHOD__
    { "__FLT_EVAL_METHOD__",    STRINGIZE(__FLT_EVAL_METHOD__),    0x0f },
#endif
#ifdef __DECIMAL_DIG__
    { "__DECIMAL_DIG__",        STRINGIZE(__DECIMAL_DIG__),        0x0f },
#endif
#ifdef __LP64__
    { "__LP64__",               STRINGIZE(__LP64__),               0x0f },
#endif
#ifdef _LP64
    { "_LP64",                  STRINGIZE(_LP64),                  0x0f },
#endif
#ifdef __ELF__
    { "__ELF__",                STRINGIZE(__ELF__),                0x0f },
#endif
#ifdef __gnu_linux__
    { "__gnu_linux__",          STRINGIZE(__gnu_linux__),          0x0f },
#endif
#ifdef __unix__
    { "__unix__",               STRINGIZE(__unix__),               0x0f },
#ifdef CONFIG_JSCPP_ASSERTION
    { "system",                 "unix",                         0x1000f },
    { "system",                 "posix",                        0x1000f },
#endif
#endif
#ifdef __linux__
    { "__linux__",              STRINGIZE(__linux__),              0x0f },
#endif
#ifdef __unix
    { "__unix",                 STRINGIZE(__unix),                 0x0f },
    { "unix",                   STRINGIZE(unix),                   0x0c },
#endif
#ifdef __linux
    { "__linux",                STRINGIZE(__linux),                0x0f },
    { "linux",                  STRINGIZE(linux),                  0x0c },
#endif
#ifdef __NO_INLINE__
    { "__NO_INLINE__",          STRINGIZE(__NO_INLINE__),          0x0f },
#endif
#ifdef __STDC_HOSTED__
    { "__STDC_HOSTED__",        STRINGIZE(__STDC_HOSTED__),        0x0f },
#endif
#ifdef __i386
    { "__i386",                 STRINGIZE(__i386),                 0x0f },
    { "i386",                   STRINGIZE(i386),                   0x0c },
#endif
#ifdef __i386__
    { "__i386__",               STRINGIZE(__i386__),               0x0f },
#ifdef CONFIG_JSCPP_ASSERTION
    { "cpu",                    "i386",                         0x1000f },
    { "machine",                "i386",                         0x1000f },
#endif
#endif
#ifdef __tune_i486__
    { "__tune_i486__",          STRINGIZE(__tune_i486__),          0x0f },
#endif

#ifdef __VERSION
    { "__VERSION",              STRINGIZE(__VERSION),              0x0f },
#endif

#ifdef __VERSION__
    { "__VERSION__",            STRINGIZE(__VERSION__),            0x0f },
#endif

#ifdef __SIZE_TYPE
    { "__SIZE_TYPE=" STRINGIZE(__SIZE_TYPE),       (const char *)0L, 0x0f },
#endif

#ifdef __PTRDIFF_TYPE__
    { "__PTRDIFF_TYPE__=" STRINGIZE(__PTRDIFF_TYPE__), (const char *)0L,  0x0f },
#endif

#ifdef __WCHAR_TYPE__
    { "__WCHAR_TYPE__=" STRINGIZE(__WCHAR_TYPE__), (const char *)0L, 0x0f },
#endif

#ifdef __USER_LABEL_PREFIX__
    { "__USER_LABEL_PREFIX__",  STRINGIZE(__USER_LABEL_PREFIX__),  0x0f },
#endif

#ifdef __REGISTER_PREFIX__
    { "__REGISTER_PREFIX__",    STRINGIZE(__REGISTER_PREFIX__),    0x0f },
#endif

#ifdef __IMMEDIATE_PREFIX__
    { "__IMMEDIATE_PREFIX__",   STRINGIZE(__IMMEDIATE_PREFIX__),   0x0f },
#endif

    /* *************************** Linux (amd) ***************************** */
#ifdef __bounded
    { "__bounded",              STRINGIZE(__bounded),              0x07 },
#endif
#ifdef __ptrvalue
    { "__ptrvalue",             STRINGIZE(__ptrvalue),             0x07 },
#endif
#ifdef __unbounded
    { "__unbounded",            STRINGIZE(__unbounded),            0x07 },
#endif
#ifdef __WINT_TYPE__
    { "__WINT_TYPE__=" STRINGIZE(__WINT_TYPE__), (const char *)0L, 0x07 },
#endif
#ifdef __x86_64
    { "__x86_64",               STRINGIZE(__x86_64),               0x0f },
#endif
#ifdef __x86_64__
    { "__x86_64__",             STRINGIZE(__x86_64__),             0x0f },
#endif
#ifdef __amd64
    { "__amd64",                STRINGIZE(__amd64),                0x0f },
#endif
#ifdef __amd64__
    { "__amd64__",              STRINGIZE(__amd64__),              0x0f },
#endif
#ifdef __tune_k8__
    { "__tune_k8__",            STRINGIZE(__tune_k8__),            0x07 },
#endif
#ifdef __MMX__
    { "__MMX__",                STRINGIZE(__MMX__),                0x07 },
#endif
#ifdef __SSE__
    { "__SSE__",                STRINGIZE(__SSE__),                0x07 },
#endif
#ifdef __SSE2__
    { "__SSE2__",               STRINGIZE(__SSE2__),               0x07 },
#endif
#ifdef __SSE_MATH__
    { "__SSE_MATH__",           STRINGIZE(__SSE_MATH__),           0x07 },
#endif
#ifdef __SSE2_MATH__
    { "__SSE2_MATH__",          STRINGIZE(__SSE2_MATH__),          0x07 },
#endif
#ifdef __athlon
    { "__athlon",               STRINGIZE(__athlon),               0x07 },
#endif
#ifdef __athlon__
    { "__athlon__",             STRINGIZE(__athlon__),             0x07 },
#endif
#ifdef __athlon_sse__
    { "__athlon_sse__",         STRINGIZE(__athlon_sse__),         0x07 },
#endif

    /* *************************** Linux (ia64) ***************************** */
#ifdef __ia64__
    { "__ia64__",               STRINGIZE(__ia64),                 0x0f },
    { "__ia64",                 STRINGIZE(__ia64__),               0x0c },
#endif

    /* ************************** Linux (others) **************************** */
#ifdef  __ppc_64__
    { "__ppc_64__",             STRINGIZE(__ppc_64__),             0x0f },
#endif
#ifdef __PPC64__
    { "__PPC64__",              STRINGIZE(__PPC64__),              0x0f },
#endif
#ifdef __s390__
    { "__s390__",               STRINGIZE(__s390__),               0x0f },
#endif
#ifdef __s390x__
    { "__s390x__",              STRINGIZE(__s390x__),              0x0f },
#endif

    /* ***************************** OSF1 (Dec) ***************************** */

#ifdef __DECC
    { "__DECC",                 STRINGIZE(__DECC),                 0x0f },
#endif
#ifdef __osf__
    { "__osf__",                STRINGIZE(__osf__),                0x0f },
#endif
#ifdef __arch64__
    { "__arch64__",             STRINGIZE(__arch64__),             0x0f },
#endif
#ifdef __PRAGMA_ENVIRONMENT
    { "__PRAGMA_ENVIRONMENT",   STRINGIZE(__PRAGMA_ENVIRONMENT),   0x0f },
#endif
#ifdef _LONGLONG
    { "_LONGLONG",              STRINGIZE(_LONGLONG),              0x0f },
#endif
#ifdef __digital__
    { "__digital__",            STRINGIZE(__digital__),            0x0f },
#endif
#ifdef __X_FLOAT
    { "__X_FLOAT",              STRINGIZE(__X_FLOAT),              0x0f },
#endif
#ifdef __STDC_VERSION__
    { "__STDC_VERSION__",       STRINGIZE(__STDC_VERSION__),       0x06 },
#endif
#ifdef __DECC_MODE_RELAXED
    { "__DECC_MODE_RELAXED",    STRINGIZE(__DECC_MODE_RELAXED),    0x0f },
#endif
#ifdef __DECC_MODE_STRICT
    { "__DECC_MODE_STRICT",     STRINGIZE(__DECC_MODE_STRICT),     0x0f },
#endif
#ifdef __DECC_VER
    { "__DECC_VER",             STRINGIZE(__DECC_VER),             0x0f },
#endif
#ifdef _SYSTYPE_BSD
    { "_SYSTYPE_BSD",           STRINGIZE(_SYSTYPE_BSD),           0x0f },
#endif
#ifdef __ALPHA
    { "__ALPHA",                STRINGIZE(__ALPHA),                0x0f },
#endif
#ifdef __IEEE_FLOAT
    { "__IEEE_FLOAT",           STRINGIZE(__IEEE_FLOAT),           0x0f },
#endif
#ifdef __Alpha_AXP
    { "__Alpha_AXP",            STRINGIZE(__Alpha_AXP),            0x0f },
#endif
#ifdef __INITIAL_POINTER_SIZE
    { "__INITIAL_POINTER_SIZE", STRINGIZE(__INITIAL_POINTER_SIZE), 0x0f },
#endif
#ifdef __LANGUAGE_C__
    { "__LANGUAGE_C__",         STRINGIZE(__LANGUAGE_C__),         0x0f },
#endif
#ifdef __alpha
    { "__alpha",                STRINGIZE(__alpha),                0x0f },
#endif

    /* ***************************** SunOS ***************************** */

#ifdef __SunOS_5_8
    { "__SunOS_5_8",            STRINGIZE(__SunOS_5_8),            0x0f },
#endif
#ifdef __SUNPRO_C
    { "__SUNPRO_C",             STRINGIZE(__SUNPRO_C),             0x0f },
#endif
#ifdef __SVR4
    { "__SVR4",                 STRINGIZE(__SVR4),                 0x0f },
#endif
#ifdef __sun
    { "__sun",                  STRINGIZE(__sun),                  0x0f },
#endif
#ifdef __sparc
    { "__sparc",                STRINGIZE(__sparc),                0x0f },
#endif
#ifdef __BUILTIN_VA_ARG_INCR
    { "__BUILTIN_VA_ARG_INCR",  STRINGIZE(__BUILTIN_VA_ARG_INCR),  0x0f },
#endif
#ifdef __PRAGMA_REDEFINE_EXTNAME
    { "__PRAGMA_REDEFINE_EXTNAME", STRINGIZE(__PRAGMA_REDEFINE_EXTNAME), 0x0f },
#endif
#ifdef sun
    { "sun",                    STRINGIZE(sun),                    0x0c },
#endif
#ifdef sparc
    { "sparc",                  STRINGIZE(sparc),                  0x0c },
#endif
#ifdef __RESTRICT
    { "__RESTRICT",             STRINGIZE(__RESTRICT),             0x0c },
#endif

    /* ***************************** HP-UX  ***************************** */

#ifdef __hp9000s700
    { "__hp9000s700",           STRINGIZE(__hp9000s700),           0x0f },
#endif
#ifdef __hp9000s800
    { "__hp9000s800",           STRINGIZE(__hp9000s800),           0x0f },
#endif
#ifdef __hppa
    { "__hppa",                 STRINGIZE(__hppa),                 0x0f },
#endif
#ifdef __hpux
    { "__hpux",                 STRINGIZE(__hpux),                 0x0f },
#endif
#ifdef _ILP32
    { "_ILP32",                 STRINGIZE(_ILP32),                 0x08 },
#endif
#ifdef _ILP64
    { "_ILP64",                 STRINGIZE(_ILP64),                 0x08 },
#endif
#ifdef _PA_RISC2_0
    { "_PA_RISC2_0",            STRINGIZE(_PA_RISC2_0),            0x0f },
#endif
#ifdef _HPUX_SOURCE
    { "_HPUX_SOURCE",           STRINGIZE(_HPUX_SOURCE),           0x0f },
#endif
#ifdef __STDC_EXT__
    { "__STDC_EXT__",           STRINGIZE(__STDC_EXT__),           0x0f },
#endif
    /*
      #ifdef __HP_cc
      { "__HP_cc",                STRINGIZE(__HP_cc),                0x0f },
      #endif
    */
#ifdef __CLASSIC_C__
    { "__CLASSIC_C__",          STRINGIZE(__CLASSIC_C__),          0x0f },
#endif
#ifdef __ia64
    { "__ia64",                 STRINGIZE(__ia64),                 0x0f },
#endif
#ifdef _BIG_ENDIAN
    { "_BIG_ENDIAN",            STRINGIZE(_BIG_ENDIAN),            0x0f },
#endif

    /* *****************************  AIX  ***************************** */

#ifdef _AIX
    { "_AIX",                   STRINGIZE(_AIX),                   0x0f },
#endif
#ifdef _AIX32
    { "_AIX32",                 STRINGIZE(_AIX32),                 0x0f },
#endif
#ifdef _AIX41
    { "_AIX41",                 STRINGIZE(_AIX41),                 0x0f },
#endif
#ifdef _AIX43
    { "_AIX43",                 STRINGIZE(_AIX43),                 0x0f },
#endif
#ifdef _IBMR2
    { "_IBMR2",                 STRINGIZE(_IBMR2),                 0x0f },
#endif
#ifdef _POWER
    { "_POWER",                 STRINGIZE(_POWER),                 0x0f },
#endif
#ifdef __xlc__
    { "__xlc__",                STRINGIZE(__xlc),                  0x0f },
#endif
#ifdef _LONG_LONG
    { "_LONG_LONG",             STRINGIZE(_LONG_LONG),             0x0f },
#endif
#ifdef __64BIT__
    { "__64BIT__",              STRINGIZE(__64BIT__),              0x0f },
#endif

    /* *****************************  Windows  ***************************** */
#ifdef _WIN32
    { "_WIN32",                 STRINGIZE(_WIN32),                 0x0f },
#endif
#ifdef _CHAR_UNSIGNED
    { "_CHAR_UNSIGNED",         STRINGIZE(_CHAR_UNSIGNED),         0x0f },
#endif
#ifdef _DLL
    { "_DLL",                   STRINGIZE(_DLL),                   0x0f },
#endif
#ifdef _M_ALPHA
    { "_M_ALPHA",               STRINGIZE(_M_ALPHA),               0x0f },
#endif
#ifdef _M_IX86
    { "_M_IX86",                STRINGIZE(_M_IX86),                0x0f },
#endif
#ifdef _M_MPPC
    { "_M_MPPC",                STRINGIZE(_M_MPPC),                0x0f },
#endif
#ifdef _M_MRX000
    { "_M_MRX000",              STRINGIZE(_M_MRX000),              0x0f },
#endif
#ifdef _M_PPC
    { "_M_PPC",                 STRINGIZE(_M_PPC),                 0x0f },
#endif
#ifdef _MFC_VER
    { "_MFC_VER",               STRINGIZE(_MFC_VER),               0x0f },
#endif
#ifdef _MSC_EXTENSIONS
    { "_MSC_EXTENSIONS",        STRINGIZE(_MSC_EXTENSIONS),        0x0f },
#endif
#ifdef _MSC_VER
    { "_MSC_VER",               STRINGIZE(_MSC_VER),               0x0f },
#endif
#ifdef _MT
    { "_MT",                    STRINGIZE(_MT),                    0x0f },
#endif
#ifdef __32BIT__
    { "__32BIT__",              STRINGIZE(__32BIT__),              0x0f },
#endif

    /* *****************************  OS/390  ***************************** */
#ifdef _EXT
    { "_EXT",                   STRINGIZE(_EXT),                   0x0f },
#endif
#ifdef _MI_BUILTIN
    { "_MI_BUILTIN",            STRINGIZE(_MI_BUILTIN),            0x0f },
#endif
#ifdef _NO_PROTO
    { "_NO_PROTO",              STRINGIZE(_NO_PROTO),              0x0f },
#endif
#ifdef _OPEN_DEFAULT
    { "_OPEN_DEFAULT",          STRINGIZE(_OPEN_DEFAULT),          0x0f },
#endif
#ifdef __ARCH__
    { "__ARCH__",               STRINGIZE(__ARCH__),               0x0f },
#endif
#ifdef __COMPILER_VER__
    { "__COMPILER_VER__",       STRINGIZE(__COMPILER_VER__),       0x0f },
#endif
#ifdef __DIGRAPHS__
    { "__DIGRAPHS__",           STRINGIZE(__DIGRAPHS__),           0x0f },
#endif
#ifdef __EXTENDED__
    { "__EXTENDED__",           STRINGIZE(__EXTENDED__),           0x0f },
#endif
#ifdef __IBMC__
    { "__IBMC__",               STRINGIZE(__IBMC__),               0x0f },
#endif
#ifdef __LONGNAME__
    { "__LONGNAME__",           STRINGIZE(__LONGNAME__),           0x0f },
#endif
#ifdef __MVS__
    { "__MVS__",                STRINGIZE(__MVS__),                0x0f },
#endif
#ifdef __PTR32
    { "__PTR32",                STRINGIZE(__PTR32),                0x0f },
#endif
#ifdef __TARGET_LIB__
    { "__TARGET_LIB__",         STRINGIZE(__TARGET_LIB__),         0x0f },
#endif

    { (const char *)0L, (const char *)0L, 0 }
};

/*********     INCLUDES 					*********/
#include <stddef.h>

#ifndef SHOW_SYMBOLS
#define JSCPP_INTERNAL
#include <jscpplib.h>
#endif

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSCPP_USE_PREDEFINES
/*  END-DEFINITIONS  */

#define IPrototype extern
/*********     PROTOTYPES					*********/
#ifndef SHOW_SYMBOLS
IPrototype int           _jspppredefine(jsppstream_t *jps, int level);
#endif

/*********     STATIC VARIABLES					*********/

/*  Language Level:
 *    Only macros with (Mask & (1 << level)) != 0 are enabled.
 *
 *  -1: all predefined platform dependent macros are off.
 *   0: pure ANSI, __STDC__ = 1
 *   1: ANSI, __STDC__ = 0
 *   2: (DEFAULT) extended level
 *
 *  Bit 16 is used for assertions
 */

#ifndef SHOW_SYMBOLS
/*JS*********************************************************************
 *
 *************************************************************************/

int _jspppredefine(jsppstream_t *jps, int level)

/************************************************************************/
{
    jsctoken_t token;
    int i, bmask;

    if(level < 0)
	return 0;

    bmask = 1 << level;

    token.TK_Flags = 0;
    token.TK_Value = CTK_NUMBER;
    for(i = 0 ; PreDefSymbols[i].Name ; ++i) {
	if((PreDefSymbols[i].Mask & bmask) == 0)
	    continue;

	if(PreDefSymbols[i].Mask & (1 << 16)) {
#ifdef CONFIG_JSCPP_ASSERTION
	    if(jsppnewpred(jps, PreDefSymbols[i].Name, PreDefSymbols[i].Value) == -1)
		EGOTO(error);
#endif
	} else if((token.TK_String = PreDefSymbols[i].Value) != NULL) {
	    if(jsppdefundefmac(jps, PreDefSymbols[i].Name,
			       &token, 1, MAFLAG_INTERNAL) == -1)
		EGOTO(error);
	} else {
	    /*  More complex definition  */
	    if(jsppdefinemacro(jps, PreDefSymbols[i].Name, 0) == -1)
		EGOTO(error);
	}
    }

    return 0;
 error:
    return -1;
}
#endif

#ifdef SHOW_SYMBOLS
#include <stdio.h>

int main(void)
{
    int i;

    for(i = 0 ; PreDefSymbols[i].Name ; ++i)
	if(PreDefSymbols[i].Value) {
	    printf("%02x: %-25s   %-40s\n",
		   PreDefSymbols[i].Mask, PreDefSymbols[i].Name,
		   PreDefSymbols[i].Value);
	} else {
	    printf("%02x: %-60s\n", PreDefSymbols[i].Mask,
		   PreDefSymbols[i].Name);
	}

    return 0;
}
#endif
