#ifndef __SIGNAL_H_
#define __SIGNAL_H_

typedef void (*sighandler_t)(int);
extern sighandler_t signal(int signum, sighandler_t handler);

#define SIG_DFL   -3
#define SIG_IGN   -2
#define SIG_ERR   -1

#define SIGHUP        1
#define SIGINT        2
#define SIGQUIT       3
#define SIGILL        4
#define SIGTRAP       5
#define SIGABRT       6
#define SIGIOT  SIGABRT
#define SIGEMT        7
#define SIGFPE        8
#define SIGKILL       9
#define SIGBUS       10
#define SIGSEGV      11
#define SIGSYS       12
#define SIGPIPE      13
#define SIGALRM      14
#define SIGTERM      15
#define SIGUSR1      16
#define SIGUSR2      17
#define SIGCHLD      18
#define SIGPWR       19
#define SIGWINCH     20
#define SIGURG       21
#define SIGIO        22
#define SIGPOLL      SIGIO
#define SIGSTOP      23
#define SIGTSTP      24
#define SIGCONT      25
#define SIGTTIN      26
#define SIGTTOU      27
#define SIGVTALRM    28
#define SIGPROF      29
#define SIGXCPU      30
#define SIGXFSZ      31

#define SIG_BLOCK    1
#define SIG_UNBLOCK  2
#define SIG_SETMASK  3

#ifndef _WIN32  /*  That's POSIX stuff!  */
#include <time.h>
#include <sys/types.h>

typedef long sigset_t;
typedef long sigval_t;
typedef struct {
  int      si_signo;
  int      si_errno;
  int      si_code;
  pid_t    si_pid;
  uid_t    si_uid;
  int      si_status;
  clock_t  si_utime;
  clock_t  si_stime;
  sigval_t si_value;
  int      si_int;
  void    *si_ptr;
  void    *si_addr;
  int      si_band;
  int      si_fd;
} siginfo_t;

struct sigaction {
  void (*sa_handler)(int);
  void (*sa_sigaction)(int, siginfo_t *, void *);
  sigset_t sa_mask;
  int sa_flags;
  void (*sa_restorer)(void);
};

#define SA_NOCLDSTOP    1
#define SA_ONESHOT      2
#define SA_RESETHAND    SA_ONESHOT
#define SA_ONSTACK      4
#define SA_RESTART      8
#define SA_NOMASK       16
#define SA_NODEFER      SA_NOMASK
#define SA_SIGINFO      32

extern int raise(int sig);

extern int sigaction(int signum, const struct sigaction *act,
		     struct sigaction *oldact);

extern int sigprocmask(int how, const sigset_t *set, sigset_t *oldset);
extern int sigpending(sigset_t *set);
extern int sigsuspend(const sigset_t *mask);

extern int sigemptyset(sigset_t *set);
extern int sigfillset(sigset_t *set);
extern int sigaddset(sigset_t *set, int signum);
extern int sigdelset(sigset_t *set, int signum);
extern int sigismember(const sigset_t *set, int signum);
#endif

#endif
