#ifndef __STDIO_H_
#define __STDIO_H_
#include <stddef.h>
#include <stdarg.h>

typedef void FILE;
typedef long fpos_t;

#define EOF              -1

#define _IONBF            1
#define _IOLBF            2
#define _IOFBF            3

#define L_tmpnam         20

#define FILENAME_MAX   1024

extern FILE *fopen(const char *name,const char *mode);
extern FILE *freopen(const char *path, const char *mode, FILE *stream);
extern int fflush(FILE *stream);
extern int fclose(FILE *fp);

extern int remove(const char *pathname);
extern int rename(const char *oldpath, const char *newpath);

extern FILE *tmpfile (void);
extern char *tmpnam(char *s);

extern void setbuf(FILE *stream, char *buf);
extern void setbuffer(FILE *stream, char *buf, size_t size);
extern void setlinebuf(FILE *stream);
extern int setvbuf(FILE *stream,char *buf,int mode,size_t size);

extern int fgetc(FILE *stream);
extern char *fgets(char *s,int size,FILE *stream);
extern int getc(FILE *stream);
extern int getchar(void);
extern char *gets(char *s);
extern int ungetc(int c, FILE *stream);

extern int printf(const char *format, ...);
extern int fprintf(FILE *stream, const char *format, ...);
extern int sprintf(char *str, const char *format, ...);
extern int vprintf(const char *format, va_list ap);
extern int vfprintf(FILE *stream, const char *format, va_list ap);
extern int vsprintf(char *str, const char *format, va_list ap);

extern int fputc(int c, FILE *stream);
extern int fputs(const char *s, FILE *stream);
extern int putc(int c, FILE *stream);
extern int putchar(int c);
extern int puts(const char *s);

extern void perror(const char *s);

extern int scanf(const char *format, ...);
extern int fscanf(FILE *stream, const char *format, ...);
extern int sscanf(const char *str, const char *format, ...);
extern int vscanf(const char *format, va_list ap);
extern int vsscanf(const char *str, const char *format, va_list ap);
extern int vfscanf(FILE *stream, const char *format, va_list ap);

extern size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream);
extern size_t fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream);
extern int fseek(FILE *stream, long offset, int whence);

extern long ftell(FILE *stream);
extern void rewind(FILE *stream);
extern int fgetpos(FILE *stream, fpos_t *pos);
extern int fsetpos(FILE *stream, fpos_t *pos);

extern void clearerr(FILE *stream);
extern int feof(FILE *stream);
extern int ferror(FILE *stream);
extern int fileno(FILE *stream);

extern FILE *stdin, *stdout, *stderr;

extern FILE *popen(const char *command, const char *type);
extern int pclose(FILE *stream);
extern FILE *fdopen(int fildes, const char *mode);

#ifdef JSFTRACE_TRACE_STDLIB
#define fopen _jswrap_fopen
#define freopen _jswrap_freopen
#define fflush _jswrap_fflush
#define fclose _jswrap_fclose
#define remove _jswrap_remove
#define rename _jswrap_rename
#define tmpfile _jswrap_tmpfile
#define tmpnam _jswrap_tmpnam
#define setbuf _jswrap_setbuf
#define setbuffer _jswrap_setbuffer
#define setlinebuf _jswrap_setlinebuf
#define setvbuf _jswrap_setvbuf
#define fgetc _jswrap_fgetc
#define fgets _jswrap_fgets
#define getc _jswrap_getc
#define getchar _jswrap_getchar
#define gets _jswrap_gets
#define ungetc _jswrap_ungetc
#define printf _jswrap_printf
#define fprintf _jswrap_fprintf
#define sprintf _jswrap_sprintf
#define vprintf _jswrap_vprintf
#define vfprintf _jswrap_vfprintf
#define vsprintf _jswrap_vsprintf
#define fputc _jswrap_fputc
#define fputs _jswrap_fputs
#define putc _jswrap_putc
#define putchar _jswrap_putchar
#define puts _jswrap_puts
#define scanf _jswrap_scanf
#define fscanf _jswrap_fscanf
#define sscanf _jswrap_sscanf
#define vscanf _jswrap_vscanf
#define vsscanf _jswrap_vsscanf
#define vfscanf _jswrap_vfscanf
#define fread _jswrap_fread
#define fwrite _jswrap_fwrite
#define fseek _jswrap_fseek
#define ftell _jswrap_ftell
#define rewind _jswrap_rewind
#define fgetpos _jswrap_fgetpos
#define fsetpos _jswrap_fsetpos
#define clearerr _jswrap_clearerr
#define feof _jswrap_feof
#define ferror _jswrap_ferror
#define fileno _jswrap_fileno

#pragma __JSCPP__ expandorig fopen, freopen, fflush, fclose, remove, \
	rename, tmpfile, tmpnam, setbuf, setbuffer, setlinebuf, setvbuf, \
	fgetc, fgets, getc, getchar, gets, ungetc, printf, fprintf, \
	sprintf, vprintf, vfprintf, vsprintf, fputc, fputs, putc, putchar, \
	puts, scanf, fscanf, sscanf, vscanf, vsscanf, vfscanf, fread, \
	fwrite, fseek, ftell, rewind, fgetpos, fsetpos, clearerr, feof, \
	ferror, fileno
#endif
#endif
