#ifndef __STDLIB_H_
#define __STDLIB_H_

#include <stddef.h>
#include <time.h>
#include <wchar.h>

typedef struct {
  long quot,rem;
} ldiv_t;
typedef struct {
  int quot,rem;
} div_t;

extern div_t div(int numerator, int denominator);
extern ldiv_t ldiv(long numerator, long denominator);

extern int abs(int j);
extern long int labs(long int j);

extern int putenv(char *string);
extern char *getenv(const char *name);

extern void *calloc(size_t nmemb, size_t size);
extern void *malloc(size_t size);
extern void *alloca(size_t size);
extern void free(void *ptr);
extern void *realloc(void *ptr, size_t size);

extern void qsort(void *base, size_t nmemb, size_t size,
                  int (*compar)(const void *, const void *));

extern void *bsearch(const void *key, const void *base, size_t nmemb,
		     size_t size, int (*compar)(const void *, const void *));

extern unsigned long int strtol(const char *nptr, char **endptr, int base);
extern unsigned long int strtoul(const char *nptr, char **endptr, int base);

extern int atoi(const char *);
extern long atol(const char *nptr);
extern double atof(const char *);

extern double strtod(const char *, char **);

extern void exit(int);
extern void abort(void);
extern int atexit(void (*function)(void));

extern int mblen(const char *s, size_t n);
extern size_t mbstowcs(wchar_t *dest, const char *src, size_t n);
extern int mbtowc(wchar_t *pwc, const char *s, size_t n);
extern size_t wcstombs(char *dest, const wchar_t *src, size_t n);
extern int wctomb(char *s, wchar_t wc);

extern int rand(void);
extern void srand(unsigned int seed);

extern int system(const char *command);

#ifdef JSFTRACE_TRACE_STDLIB
#define putenv _jswrap_putenv
#define getenv _jswrap_getenv
#define calloc _jswrap_calloc
#define malloc _jswrap_malloc
#define free _jswrap_free
#define realloc _jswrap_realloc
#define strtol _jswrap_strtol
#define strtoul _jswrap_strtoul
#define atoi _jswrap_atoi
#define atof _jswrap_atof
#define strtod _jswrap_strtod

#define mblen _jswrap_mblen
#define mbstowcs _jswrap_mbstowcs
#define mbtowc _jswrap_mbtowc
#define wcstombs _jswrap_wcstombs
#define wctomb _jswrap_wctomb

#pragma __JSCPP__ expandorig putenv, getenv, calloc, malloc, free, \
	realloc, strtol, strtoul, atoi, atof, strtod, mblen, mbstowcs, \
	mbtowc, wcstombs, wctomb
#endif
#endif
