#ifndef __SYS_SOCKET_H_
#define __SYS_SOCKET_H_

#include <stddef.h>
#include <sys/uio.h>

typedef long socklen_t;

struct cmsghdr {
  size_t cmsg_len;
  int cmsg_level;
  int cmsg_type;
};

#define SCM_CREDENTIALS    1

struct msghdr {
  void        *msg_name;
  socklen_t    msg_namelen;
  struct iovec *msg_iov;
  size_t       msg_iovlen;
  void         *msg_control;
  socklen_t    msg_controllen;
  int          msg_flags;   
};

#define PF_UNIX            1
#define PF_LOCAL           PF_UNIX
#define PF_INET            2

#define AF_UNIX            1
#define AF_INET            2

#define SOCK_STREAM        1
#define SOL_SOCKET         1

#define SO_REUSEADDR       1
#ifdef __linux__
#define SO_PASSCRED        2
#endif

/*  Avoid bogus "function calls"  */
#define CMSG_FIRSTHDR(a)   0
#define CMSG_NXTHDR(a,b)   0
#define CMSG_ALIGN(a)      0
#define CMSG_SPACE(a)      0
#define CMSG_LEN(a)        0
#define CMSG_DATA(a)       0

extern int socket(int domain, int type, int protocol);
extern int getsockopt(int s, int level, int optname, void *optval, socklen_t *optlen);
extern int setsockopt(int s, int level, int optname,  const  void  *optval,
		      socklen_t optlen);
extern int bind(int sockfd, const struct sockaddr *my_addr, socklen_t addrlen);
extern int listen(int s, int backlog);
extern int accept(int s, struct sockaddr *addr, socklen_t *addrlen);

extern ssize_t send(int s, const void *buf, size_t len, int flags);
extern ssize_t  sendto(int  s,  const  void  *buf, size_t len, int flags,
                       const struct sockaddr *to, socklen_t tolen);
extern ssize_t sendmsg(int s, const struct msghdr *msg, int flags);

#endif
