#ifndef __SYS_STAT_H_
#define __SYS_STAT_H_

#include <sys/types.h>
#include <time.h>

struct stat {
  dev_t         st_dev;
  ino_t         st_ino;
  mode_t        st_mode;
  nlink_t       st_nlink;
  uid_t         st_uid;
  gid_t         st_gid;
  dev_t         st_rdev;
  off_t         st_size;
  blksize_t     st_blksize;
  blkcnt_t      st_blocks;
  time_t        st_atime;
  time_t        st_mtime;
  time_t        st_ctime;
};

/*  Avoid bogus "function calls"  */
#define S_ISLNK(m)  m
#define S_ISREG(m)  m
#define S_ISDIR(m)  m
#define S_ISCHR(m)  m
#define S_ISBLK(m)  m
#define S_ISFIFO(m) m
#define S_ISSOCK(m) m

#define S_IRUSR  0001
#define S_IWUSR  0002
#define S_IXUSR  0004

#ifndef _WIN32
#define S_IRGRP  0010
#define S_IWGRP  0020
#define S_IXGRP  0040

#define S_IROTH  0100
#define S_IWOTH  0200
#define S_IXOTH  0400

#define	S_IRWXU	 (S_IRUSR | S_IWUSR | S_IXUSR)
#define	S_IRWXG	 (S_IRGRP | S_IWGRP | S_IXGRP)
#define	S_IRWXO	 (S_IROTH | S_IWOTH | S_IXOTH)
#endif

extern int stat(const char *file_name, struct stat *buf);
extern int fstat(int filedes, struct stat *buf);
extern int lstat(const char *file_name, struct stat *buf);

#ifndef _WIN32
extern mode_t umask(mode_t mask);
extern int chmod(const char *path, mode_t mode);
extern int fchmod(int fildes, mode_t mode);
extern int mkdir(const char *pathname, mode_t mode);
#endif

#endif
