#ifndef __TERMIOS_H_
#define __TERMIOS_H_

/*  c_iflag  */
#define IGNBRK      1
#define BRKINT      2
#define IGNPAR      4
#define PARMRK      8
#define INPCK      16
#define ISTRIP     32
#define INLCR      64
#define IGNCR     128
#define ICRNL     256
#ifdef __linux__
#define IUCLC     512
#endif
#define IXON     1024
#define IXANY    2048
#define IXOFF    4096
#ifdef __linux__
#define IMAXBEL  8192
#endif

/*  c_oflag  */
#define OPOST       1
#ifdef __linux__
#define OLCUC       2
#endif
#define ONLCR       4
#define OCRNL       8
#define ONOCR      16
#define ONLRET     32
#define OFILL      64
#ifdef __linux__
#define OFDEL     128
#endif
#define NLDLY     256
#define CRDLY     512
#define TABDLY   1024
#define BSDLY    2048
#define VTDLY    4096
#define FFDLY    8192

/*  c_cflag  */
#ifdef __linux__
#define CBAUD       1
#endif
#define CSIZE       4
#define CSTOPB      8
#define CREAD      16
#define PARENB     32
#define PARODD     64
#define HUPCL     128
#define CLOCAL    256
#ifdef __linux__
#define LOBLK     512
#define CIBAUD   1024
#define CRTSCTS  2048
#endif

/*  c_lflag  */
#define ISIG        1
#define ICANON      2
#define ECHO        8
#define ECHOE      16
#define ECHOK      32
#define ECHONL     64
#ifdef __linux__
#define ECHOCTL   128
#endif
#define ECHOPRT   256
#ifdef __linux__
#define ECHOKE    512
#define DEFECHO  1024
#define FLUSHO   2048
#endif
#define NOFLSH   8192
#define TOSTOP  16384
#define IEXTEN  32768

/*  c_cc  */
#define VINTR       1
#define VQUIT       2
#define VERASE      3
#define VKILL       4
#define VEOF        5
#define VMIN        6
#define VEOL        7
#define VTIME       8
#define VEOL2       9
#define VSWTCH     10
#define VSTART     11
#define VSTOP      12
#define VSUSP      13
#ifdef __linux__
#define VDSUSP     14
#define VLNEXT     15
#define VWERASE    16
#define VREPRINT   17
#define VDISCARD   18
#define VSTATUS    19
#endif

#define NCCS       20

typedef unsigned long tcflag_t;
typedef unsigned long cc_t;
typedef unsigned long speed_t;

struct termios {
  tcflag_t c_iflag;
  tcflag_t c_oflag;
  tcflag_t c_cflag;
  tcflag_t c_lflag;
  cc_t c_cc[NCCS];
};

#define TCSANOW     1
#define TCSADRAIN   2
#define TCSAFLUSH   4

#define TCIFLUSH    1
#define TCOFLUSH    2
#define TCIOFLUSH   3

#define TCOOFF      1
#define TCOON       2
#define TCIOFF      3
#define TCION       4

#define B0      0x00001
#define B50     0x00002
#define B75     0x00004
#define B110    0x00008
#define B134    0x00010
#define B150    0x00020
#define B200    0x00040
#define B300    0x00080
#define B600    0x00100
#define B1200   0x00200
#define B1800   0x00400
#define B2400   0x00800
#define B4800   0x01000
#define B9600   0x02000
#define B19200  0x04000
#define B38400  0x08000
#define CBAUDEX 0x0ffff

#define B57600  0x10000
#define B115200 0x20000
#define B230400 0x40000

extern int tcgetattr(int fd, struct termios *termios_p);
extern int tcsetattr(int fd, int optional_actions, const struct termios *termios_p);
extern int tcsendbreak(int fd, int duration);
extern int tcdrain(int fd);
extern int tcflush(int fd, int queue_selector);
extern int tcflow(int fd, int action);
extern void cfmakeraw(struct termios *termios_p);
extern speed_t cfgetispeed(const struct termios *termios_p);
extern speed_t cfgetospeed(const struct termios *termios_p);
extern int cfsetispeed(struct termios *termios_p, speed_t speed);
extern int cfsetospeed(struct termios *termios_p, speed_t speed);

#endif
