#ifndef __UNISTD_H_
#define __UNISTD_H_

#include <stddef.h>
#include <signal.h>
#include <sys/types.h>
#define _POSIX_VERSION 199506L
#define _POSIX_THREADS

extern int truncate(const char *path, off_t length);
extern int ftruncate(int fd, off_t length);
extern int close(int fd);

extern off_t lseek(int fildes, off_t offset, int whence);
extern ssize_t write(int fd, const void *buf, size_t count);
extern ssize_t read(int fd, void *buf, size_t count);

extern int pipe(int filedes[2]);
extern pid_t fork(void);
extern pid_t wait(int *status);
extern pid_t waitpid(pid_t pid, int *status, int options);

extern pid_t setsid(void);

extern int symlink(const char *oldpath, const char *newpath);
extern int readlink(const char *path, char *buf, size_t bufsiz);

extern unsigned int sleep(unsigned int seconds);

extern uid_t getuid(void);
extern uid_t geteuid(void);
extern gid_t getgid(void);
extern gid_t getegid(void);

extern pid_t getpid(void);
extern pid_t getppid(void);

extern int rmdir(const char *pathname);
extern int unlink(const char *pathname);

extern char *getcwd(char *buf, size_t size);
extern char *getwd(char *buf);

extern int chdir(const char *path);
extern int fchdir(int fd);

extern int select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
		  struct timeval *timeout);
extern int pselect(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
		   const struct timespec *timeout, const sigset_t *sigmask);

#define _SC_ARG_MAX        1
#define _SC_CHILD_MAX      2
#define _SC_HOST_NAME_MAX  3
#define _SC_LOGIN_NAME_MAX 4
#define _SC_CLK_TCK        5
#define _SC_OPEN_MAX       6
#define _SC_PAGESIZE       7
#define _SC_RE_DUP_MAX     8
#define _SC_STREAM_MAX     9
#define _SC_TTY_NAME_MAX  10
#define _SC_TZNAME_MAX    11
#define _SC_VERSION       12

#define ARG_MAX         4096
#define CHILD_MAX         25
#define HOST_NAME_MAX    255
#define LOGIN_NAME_MAX     9
#define OPEN_MAX          20
#define PAGESIZE        1024
#define PAGE_SIZE   PAGESIZE
#define RE_DUP_MAX       255
#define STREAM_MAX         8
#define SYMLOOP_MAX        8
#define TTY_NAME_MAX       9
#define TZNAME_MAX         6

long sysconf(int name);

#endif
