/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Application for preprocessor library.
 *
 *    Part    : Add tracing calls to functions.
 *
 *              Copyright (C) 2003-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/
#ifndef tokseqrewind
/*  START-DEFINITIONS  */
typedef struct {
    int Start, Curr, End;
} tokenseq_t;

#define tokseqrewind(tokseq)    ((tokseq)->Curr = (tokseq)->Start)
/*  END-DEFINITIONS  */
#endif

#define initPatchDescription(pdesc, type)	\
    ((pdesc)->Type = (type),			\
     (pdesc)->Patterns = NULL,			\
     (pdesc)->NPatterns = 0,			\
     (pdesc)->CallDescriptions = NULL,		\
     (pdesc)->NCallDescriptions = -1,		\
     (pdesc)->ReturnDescriptions = NULL,	\
     (pdesc)->NReturnDescriptions = -1)

#define NAME_CLAUSE_TRACE               "trace"
#define NAME_CLAUSE_TRACELEVEL          "tracelevel"
#define NAME_CLAUSE_OPTIONS             "options"
#define NAME_CLAUSE_TYPEMAP             "typemap"
#define NAME_CLAUSE_TYPERENAME          "typerename"
#define NAME_CLAUSE_TYPEFORMAT          "typeformat"
#define NAME_CLAUSE_VARTRACE            "vartrace"

#define NAME_CLAUSE_STATEMENT           "statement"
#define NAME_CLAUSE_BRANCH              "branch"

#define NAME_CLAUSE_TRACEFORMATCALL     "traceformatcall"
#define NAME_CLAUSE_TRACEFORMATRETURN   "traceformatreturn"
#define NAME_CLAUSE_TRACEFORMATSTMT     "traceformatstatement"
#define NAME_CLAUSE_RETURNVARIABLE      "returnvariable"
#define NAME_CLAUSE_LABELFORMAT         "labelformat"

#define NAME_CLAUSE_CALL                "call"
#define NAME_CLAUSE_RETURN              "return"
#define NAME_CLAUSE_FCTWRAP             "wrap"

#define NAME_CLAUSE_OFF                 "off"
#define NAME_CLAUSE_READ                "read"
#define NAME_CLAUSE_WRITE               "write"
#define NAME_CLAUSE_GLOBAL              "global"
#define NAME_CLAUSE_LOCAL               "local"

#define NAME_CLAUSE_LABEL               "label"
#define NAME_CLAUSE_NOLABEL             "nolabel"
#define NAME_CLAUSE_NOLINE              "noline"
#define NAME_CLAUSE_RETURNLABEL         "retlabel"

#define NAME_CLAUSE_INDENT              "indent"
#define NAME_CLAUSE_MAPLEVEL            "maplevel"
#define NAME_CLAUSE_RECURLEVEL          "recurlevel"

#define NAME_POINTER_TYPE               "__POINTER__"

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doPragma(jsparsstream_t *jpa, traceaux_t *paux,
				  jsparscnode_t *node, jctokfifo_t *queue,
				  int mode);

static int               parseTrace(jsparsstream_t *jpa, traceaux_t *paux,
				    int start, int end,
				    jctokfifo_t *queue);
static int               addDescription(traceaux_t *paux, patchdesc_t *pdesc);

static int               parseVarTrace(jsparsstream_t *jpa, traceaux_t *paux,
				       int start, int end,
				       jctokfifo_t *queue);
static int               addVarTrace(traceaux_t *paux, char *pattern,
				     long flag);

static int               parseTraceLevel(jsparsstream_t *jpa, traceaux_t *paux,
					 int start, int end,
					 jctokfifo_t *queue);
static int               addTraceLevel(traceaux_t *paux, const char *pattern,
				       int level);
Prototype int            findTraceLevel(traceaux_t *paux,
					const char *funcname);

static int               parseOptions(jsparsstream_t *jpa, traceaux_t *paux,
				      int start, int end, jctokfifo_t *queue);

static int               parseTypeMap(jsparsstream_t *jpa, traceaux_t *paux,
				      int start, int end, jctokfifo_t *queue);
static int               addTypeMap(jsparsstream_t *jpa, traceaux_t *paux,
				    jsparsvartab_t *vTab,
				    typemapping_t *pMap);

static int               parseTypeRename(jsparsstream_t *jpa, traceaux_t *paux,
                                         int start, int end, jctokfifo_t *queue);
static int               addTypeRename(jsparsstream_t *jpa, traceaux_t *paux,
                                       const char *typeName,
                                       const char *sourceName, const char *targetName);
static int               locateTypeRename(traceaux_t *paux, const char *typeName,
                                          const char *sourceName);
Prototype int            findTypeRename(traceaux_t *paux, const char *typeName,
                                        const char *sourceName);

static int               parseTypeFormat(jsparsstream_t *jpa, traceaux_t *paux,
					 int start, int end,
					 jctokfifo_t *queue);
static int               addTypeFormat(jsparsstream_t *jpa, traceaux_t *paux,
				       jsparsvartab_t *vTab,
				       typeformat_t *lMap,
				       jctokfifo_t *queue);

/*JS*********************************************************************
 *   mode bit 0: If set, we have a pragma within a function body.
 *************************************************************************/

int doPragma(jsparsstream_t *jpa, traceaux_t *paux,
	     jsparscnode_t *node, jctokfifo_t *queue, int mode)

/************************************************************************/
{
    const char *name;
    int start, end, step;

    step = jctokgetstep(queue);

    /*  The actual content is in the child node of the pragma node  */
    assert(node->JPACN_Sub && node->JPACN_Sub->JPACN_Type == JPACT_PRAGMACONT);
    start = node->JPACN_Sub->JPACN_Start;
    end = node->JPACN_Sub->JPACN_End;

    if(queue->TKFF_Token[start].TK_Value != CTK_IDENT) {
	name = "Syntax error in pragma line";
	goto invalid;
    }

    name = queue->TKFF_Token[start].TK_String;
    start += step;

    /* *********************************************************************** */
    if(strcmp(name, NAME_CLAUSE_TRACE) == 0) {
	if((start = parseTrace(jpa, paux, start, end, queue)) == -1)
	    EGOTO(error);

	if(start == -2) {
	    name = "Invalid patterns in " NAME_CLAUSE_TRACE " line";
	    goto invalid;
	}

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_TRACELEVEL) == 0) {
	if((start = parseTraceLevel(jpa, paux, start, end, queue)) == -1)
	    EGOTO(error);

	if(start == -2) {
	    name = "Invalid expression in " NAME_CLAUSE_TRACELEVEL " line";
	    goto invalid;
	}

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_OPTIONS) == 0) {
	if((start = parseOptions(jpa, paux, start, end, queue)) == -1)
	    EGOTO(error);

	if(start == -2) {
	    name = "Invalid expression in " NAME_CLAUSE_OPTIONS " line";
	    goto invalid;
	}

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_TYPEMAP) == 0) {
	if((start = parseTypeMap(jpa, paux, start, end, queue)) == -1)
	    EGOTO(error);

	if(start == -2) {
	    name = "Invalid expression in " NAME_CLAUSE_TYPEMAP " line";
	    goto invalid;
	}

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_TYPERENAME) == 0) {
	if((start = parseTypeRename(jpa, paux, start, end, queue)) == -1)
	    EGOTO(error);

	if(start == -2) {
	    name = "Invalid expression in " NAME_CLAUSE_TYPERENAME " line";
	    goto invalid;
	}

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_TYPEFORMAT) == 0) {
	while(start <= end &&
	      CTK_BLANK_START <= queue->TKFF_Token[start].TK_Value &&
	      queue->TKFF_Token[start].TK_Value <= CTK_BLANK_END)
	    start += step;

	if(start > end) {
	    name = "Premature ended " NAME_CLAUSE_TYPEFORMAT " line";
	    EGOTO(error);
	}

	if((start = parseTypeFormat(jpa, paux, start, end, queue)) == -1)
	    EGOTO(error);

	if(start == -2) {
	    name = "Invalid syntax in " NAME_CLAUSE_TYPEFORMAT " line";
	    goto invalid;
	}

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_VARTRACE) == 0) {
	if((start = parseVarTrace(jpa, paux, start, end, queue)) == -1)
	    EGOTO(error);

	if(start == -2) {
	    name = "Invalid patterns in " NAME_CLAUSE_VARTRACE " line";
	    goto invalid;
	}

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_TRACEFORMATCALL) == 0 ||
	      strcmp(name, NAME_CLAUSE_TRACEFORMATRETURN) == 0 ||
	      strcmp(name, NAME_CLAUSE_TRACEFORMATSTMT) == 0) {
	char *string;

	while(start <= end && CTK_BLANK_START <=
	      queue->TKFF_Token[start].TK_Value &&
	      queue->TKFF_Token[start].TK_Value <= CTK_BLANK_END)
	    start += step;

	if((string = jctoktostring(queue, start, end)) == NULL)
	    EGOTO(error);

	if(strcmp(name, NAME_CLAUSE_TRACEFORMATCALL) == 0)
	    paux->TraceFormatCall = string;
	else if(strcmp(name, NAME_CLAUSE_TRACEFORMATRETURN) == 0)
	    paux->TraceFormatReturn = string;
	else
	    paux->TraceFormatStmt = string;

	/*  We inhibit generating prototype if it is not the standard trace
	 *   function
	 */
	paux->Mode &= ~JSCPP_TRACE_PROTOTYPEFCT;

	start = end + step;

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_RETURNVARIABLE) == 0) {
	while(start <= end &&
	      CTK_BLANK_START <= queue->TKFF_Token[start].TK_Value &&
	      queue->TKFF_Token[start].TK_Value <= CTK_BLANK_END)
	    start += step;

	if(queue->TKFF_Token[start].TK_Value != CTK_IDENT) {
	    name = "Expected identifier in " NAME_CLAUSE_RETURNVARIABLE
		" pragma line";
	    goto invalid;
	}

	if((paux->NameRetVariable = strdup(queue->TKFF_Token[start].TK_String))
	   == NULL)
	    EGOTO(error);

	/* *********************************************************************** */
    } else if(strcmp(name, NAME_CLAUSE_LABELFORMAT) == 0) {
	while(start <= end &&
	      CTK_BLANK_START <= queue->TKFF_Token[start].TK_Value &&
	      queue->TKFF_Token[start].TK_Value <= CTK_BLANK_END)
	    start += step;

	if((paux->LabelFormat = jctoktostring(queue, start, end)) == NULL)
	    EGOTO(error);

	start = end + step;

	/* *********************************************************************** */
    } else {
	name = "Unknown pragma line";
	goto invalid;
    }

    /*  Check remainder  */
    for( ; start <= end ; start += step) {
	int tokval;

	tokval = queue->TKFF_Token[start].TK_Value;
	if(!(CTK_BLANK_START <= tokval && tokval <= CTK_BLANK_END)) {
	    name = "Trailing garbage in pragma line";
	    /*printCToken(stdout, queue->TKFF_Token[start]);*/
	    goto invalid;
	}
    }

    /* *********************************************************************** */

    /*  Delete pragma line, since we have handled it  */
    start = node->JPACN_Start;
    end   = node->JPACN_End;
    if(jpa->JSPA_Mode & JPAM_FGETORIG) { ++start; ++end; }

    /*  Substitute it by the newlines   */
    if(jctoksubstitute(queue, start, end, "%n") == -1)
	EGOTO(error);

    return 0;
 invalid:
    if(node->JPACN_Type == JPACT_PRAGMALINE2 && jpa->JSPA_WarnFlags != 0)
	/*  At least emit a warning if the pragma line was decorated with
	 *   our own introducer. Through that, our user gets at least some
	 *   feedback that he/she did something wrong.
	 */
	jsparsprinterr(jpa, queue, node, name, 0);

    /*  As required by ANSI, invalid pragma lines are no error  */
    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int parseTrace(jsparsstream_t *jpa, traceaux_t *paux,
		      int start, int end, jctokfifo_t *queue)

/************************************************************************/
{
    patchdesc_t desc;
    int count, step, isOff;

    step = jctokgetstep(queue);

    initPatchDescription(&desc, PDESC_TYPE_TRACE);

    if((start = readPatterns(&desc, start, end, queue)) == -1)
	EGOTO(error);

    if(start == -2)
	EGOTO(invalid);

    count = 0;
    isOff = FALSE;
    while(start <= end) {
	const char *name;
	int tokval;

	tokval = queue->TKFF_Token[start].TK_Value;
	if(CTK_BLANK_START <= tokval && tokval <= CTK_BLANK_END) {
	    start += step;
	    continue;
	}

	if(tokval != CTK_IDENT)
	    EGOTO(invalid);

	name = queue->TKFF_Token[start].TK_String;
	start += step;
	++count;

	/*  We recognize "off" only as the very first word  */
	if(count == 1 && strcmp(name, NAME_CLAUSE_OFF) == 0) {
	    isOff = TRUE;
	    continue;
	}

	/*  Nothing more is allowed after "off"  */
	if(isOff) EGOTO(invalid);

	if(!(desc.Type & PDESC_TYPE_FCALL) &&
	   strcmp(name, NAME_CLAUSE_CALL) == 0) {
	    desc.Type |= PDESC_TYPE_FCALL;
	    if((start = readArgs(paux, jpa, &(desc.CallDescriptions),
				 &(desc.NCallDescriptions), start, end,
				 queue, 0)) == -1 || start == -2)
		EGOTO(error);

	} else if(!(desc.Type & PDESC_TYPE_FRETURN) &&
		  strcmp(name, NAME_CLAUSE_RETURN) == 0) {
	    desc.Type |= PDESC_TYPE_FRETURN | PDESC_TYPE_FRETURNLINE;
	    if((start = readArgs(paux, jpa, &(desc.ReturnDescriptions),
				 &(desc.NReturnDescriptions), start, end,
				 queue, 1)) == -1)
		EGOTO(invalid);

	} else if(!(desc.Type & PDESC_TYPE_FCALLLBL) &&
		  strcmp(name, NAME_CLAUSE_LABEL) == 0) {
	    desc.Type |= PDESC_TYPE_FCALLLBL;

	} else if((desc.Type & PDESC_TYPE_FRETURNLINE) &&
		  strcmp(name, NAME_CLAUSE_NOLINE) == 0) {
	    desc.Type &= ~PDESC_TYPE_FRETURNLINE;

	} else if(!(desc.Type & PDESC_TYPE_FRETURNLBL) &&
		  strcmp(name, NAME_CLAUSE_RETURNLABEL) == 0) {
	    desc.Type |= PDESC_TYPE_FRETURNLBL;

	} else if(!(desc.Type & PDESC_TYPE_FALLSTMT) &&
		  strcmp(name, NAME_CLAUSE_STATEMENT) == 0) {
	    desc.Type |= PDESC_TYPE_FALLSTMT;

	} else if(!(desc.Type & PDESC_TYPE_FFCTWRAP) &&
		  strcmp(name, NAME_CLAUSE_FCTWRAP) == 0) {
	    desc.Type |= PDESC_TYPE_FFCTWRAP;
	    paux->Mode |= JSCPP_TRACE_NEEDWRAPPING;

	} else if(!(desc.Type & PDESC_TYPE_FALLSTMT) &&
		  strcmp(name, NAME_CLAUSE_BRANCH) == 0) {
	    desc.Type |= PDESC_TYPE_FBRANCHSTMT;

	} else {
	    EGOTO(invalid);
	}
    }

    if((desc.Type & ~PDESC_TYPE_FFCTWRAP) == PDESC_TYPE_TRACE && !isOff)
	/*  Nothing specified, defaults to "call return line"  */
	desc.Type |= PDESC_TYPE_FCALL | PDESC_TYPE_FRETURN |
	    PDESC_TYPE_FRETURNLINE;

    /*  Add description to structure  */
    if(addDescription(paux, &desc) == -1)
	EGOTO(error);

    return start;
 error:
    return -1;
 invalid:
    return -2;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addDescription(traceaux_t *paux, patchdesc_t *pdesc)

/************************************************************************/
{
    int i, j;

    for(j = i = 0 ; i < paux->NDescriptions ; ++i) {
	patchdesc_t *odesc;
	int k, l, diddelete;

	odesc = &(paux->Descriptions[i]);

	/*  Go to all the old patterns and check if the new one
	 *   makes them obsolete
	 */
	diddelete = FALSE;
	for(k = 0 ; k < odesc->NPatterns ; ++k) {
	    const char *opattern = odesc->Patterns[k];

	    for(l = 0 ; l < pdesc->NPatterns ; ++l) {
		const char *pattern = pdesc->Patterns[l];

		if(jiswildcard(pattern) ?
		   /*  We want to handle situations like opattern = "*",
		    *   pattern = "?" properly:
		    *   opattern is matched by pattern, but not obsolete!
		    */
		   (!jiswildcard(opattern) && wildcmp(opattern, pattern, 0)) :
		   strcmp(opattern, pattern) == 0)
		    break;
	    }

	    if(l < pdesc->NPatterns) {
		/*  This pattern is obsolete  */
		free((void *)odesc->Patterns[k]);
		odesc->Patterns[k] = NULL;
		diddelete = TRUE;
	    }
	}

	if(diddelete) {
	    /*  Remove empty patterns  */
	    for(l = k = 0 ; k < odesc->NPatterns ; ++k)
		if(odesc->Patterns[k] != NULL) {
		    if(l != k)
			odesc->Patterns[l] = odesc->Patterns[k];

		    ++l;
		}

	    odesc->NPatterns = l;
	    if(l == 0) {
		/*  This one is obsolete  */
		freepatchdesc(odesc);
		continue;
	    }
	}

	/*  Copy this description  */
	if(j != i)
	    paux->Descriptions[j] = *odesc;
	++j;
    }
    paux->NDescriptions = j;

    /*  Now add the new patch description  */
    if(paux->NDescriptions >= paux->MDescriptions) {
	REALLOC_RESIZE(paux->MDescriptions, 50);
	if((paux->Descriptions = (patchdesc_t *)
	    realloc(paux->Descriptions, paux->MDescriptions *
		    sizeof(*(paux->Descriptions)))) == NULL)
	    EGOTO(error);
    }

    paux->Descriptions[paux->NDescriptions++] = *pdesc;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int parseVarTrace(jsparsstream_t *jpa, traceaux_t *paux,
			 int start, int end, jctokfifo_t *queue)

/************************************************************************/
{
    patchdesc_t desc;
    long flag;
    int i, step;

    step = jctokgetstep(queue);

    initPatchDescription(&desc, 0);

    if((start = readPatterns(&desc, start, end, queue)) == -1)
	EGOTO(error);

    if(start == -2)
	EGOTO(invalid);

    flag = -1;
    while(start <= end) {
	const char *name;
	int tokval;

	tokval = queue->TKFF_Token[start].TK_Value;
	if(CTK_BLANK_START <= tokval && tokval <= CTK_BLANK_END) {
	    start += step;
	    continue;
	}

	if(tokval != CTK_IDENT)
	    EGOTO(invalid);

	name = queue->TKFF_Token[start].TK_String;
	start += step;

	if(flag == -1) flag = 0;
	if(strcmp(name, NAME_CLAUSE_OFF) == 0) {
	    flag = 0;
	} else if(strcmp(name, NAME_CLAUSE_READ) == 0) {
	    flag |= VARTRACE_FREAD;
	} else if(strcmp(name, NAME_CLAUSE_WRITE) == 0) {
	    flag |= VARTRACE_FWRITE;
	} else if(strcmp(name, NAME_CLAUSE_GLOBAL) == 0) {
	    flag |= VARTRACE_FGLOBAL;
	} else if(strcmp(name, NAME_CLAUSE_LOCAL) == 0) {
	    flag |= VARTRACE_FLOCAL;
	} else
	    EGOTO(invalid);
    }

    /*  Default: only writing to the variable is traced  */
    if(flag == -1)
	flag = VARTRACE_FWRITE;
    else if(flag != 0 && !(flag & (VARTRACE_FREAD | VARTRACE_FWRITE)))
	flag |= VARTRACE_FWRITE;

    for(i = 0 ; i < desc.NPatterns ; ++i)
	if(addVarTrace(paux, (char *)desc.Patterns[i], flag) == -1)
	    EGOTO(error);

    free((void *)desc.Patterns);

    return start;
 error:
    return -1;
 invalid:
    return -2;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addVarTrace(traceaux_t *paux, char *pattern, long flag)

/************************************************************************/
{
    char *sep;

    /*  Now add the new patch description  */
    if(paux->NVarTraces >= paux->MVarTraces) {
	REALLOC_RESIZE(paux->MVarTraces, 50);
	if((paux->VarTraces = (vartrace_t *)
	    realloc(paux->VarTraces, paux->MVarTraces *
		    sizeof(*(paux->VarTraces)))) == NULL)
	    EGOTO(error);
    }

    if((sep = strchr(pattern, ':')) == NULL) {
	/*  "VARPATTERN"  */
	paux->VarTraces[paux->NVarTraces].FuncPattern = NULL;
	paux->VarTraces[paux->NVarTraces].VarPattern  = pattern;

    } else {
	/*  Separate it into two parts  */
	*sep++ = '\0';

	if(*pattern != '\0') {
	    /*  "FUNCPATTERN:VARPATTERN"  */
	    paux->VarTraces[paux->NVarTraces].FuncPattern = pattern;
	    paux->VarTraces[paux->NVarTraces].VarPattern = *sep ? sep : NULL;

	} else {
	    paux->VarTraces[paux->NVarTraces].FuncPattern = NULL;

	    if(*sep != '\0') {
		/*  ":VARPATTERN"  */

		/*  This is to ensure we can free it later correctly  */
		memmove(pattern, sep, strlen(sep) + 1);
		paux->VarTraces[paux->NVarTraces].VarPattern = pattern;

	    } else {
		/*  ":"  */
		free(pattern);
		paux->VarTraces[paux->NVarTraces].VarPattern = NULL;
	    }
	}
    }

    paux->VarTraces[paux->NVarTraces].Flag = flag;
    paux->NVarTraces++;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int parseTraceLevel(jsparsstream_t *jpa, traceaux_t *paux,
			   int start, int end, jctokfifo_t *queue)

/************************************************************************/
{
    patchdesc_t desc;
    int i, level, step;

    step = jctokgetstep(queue);

    initPatchDescription(&desc, 0);

    if((start = readPatterns(&desc, start, end, queue)) == -1)
	EGOTO(error);

    for( ; start <= end ; start += step) {
	int tokVal = queue->TKFF_Token[start].TK_Value;
	if(!(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END))
	    break;
    }

    if(start > end || queue->TKFF_Token[start].TK_Value != CTK_NUMBER)
	EGOTO(invalid);

    level = strtol(queue->TKFF_Token[start].TK_String, NULL, 0);

    for(i = 0 ; i < desc.NPatterns ; ++i)
	if(addTraceLevel(paux, desc.Patterns[i], level) == -1)
	    EGOTO(error);

    free((void *)desc.Patterns);

    return start + step;
 error:
    return -1;
 invalid:
    return -2;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addTraceLevel(traceaux_t *paux, const char *pattern, int level)

/************************************************************************/
{
    int i;

    for(i = 0 ; i < paux->NTraceLevels ; ++i) {
	const char *opattern = paux->TraceLevels[i].Pattern;

	if(jiswildcard(pattern) ?
	   /*  We want to handle situations like opattern = "*",
	    *   pattern = "?" properly:
	    *   opattern is matched by pattern, but not obsolete!
	    */
	   (!jiswildcard(opattern) && wildcmp(opattern, pattern, 0)) :
	   strcmp(opattern, pattern) == 0)
	    break;
    }

    if(paux->NTraceLevels >= paux->MTraceLevels) {
	REALLOC_RESIZE(paux->MTraceLevels, 20);
	if((paux->TraceLevels = (tracelevel_t *)
	    realloc(paux->TraceLevels, paux->MTraceLevels *
		    sizeof(*(paux->TraceLevels)))) == NULL)
	    EGOTO0(error);
    }

    if(i < paux->NTraceLevels)
	memmove(&(paux->TraceLevels[i + 1]), &(paux->TraceLevels[i]),
		(paux->NTraceLevels - i) * sizeof(*(paux->TraceLevels)));

    paux->NTraceLevels++;

    if((paux->TraceLevels[i].Pattern = strdup(pattern)) == NULL)
	EGOTO0(error);

    paux->TraceLevels[i].Level = level;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int findTraceLevel(traceaux_t *paux, const char *funcname)

/************************************************************************/
{
    int i;

    for(i = 0 ; i < paux->NTraceLevels ; ++i) {
	int ret;

	if((ret = checkPattern(paux->TraceLevels[i].Pattern,
			       paux->CurrentFile, funcname)) == -1)
	    return -1;

	if(ret != 0)
	    return paux->TraceLevels[i].Level;
    }

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int parseOptions(jsparsstream_t *jpa, traceaux_t *paux,
			int start, int end, jctokfifo_t *queue)

/************************************************************************/
{
    int step, state;

    step = jctokgetstep(queue);

    for(state = 0 ; start <= end ; start += step) {
	const char *string;
	int tokVal;

	tokVal = queue->TKFF_Token[start].TK_Value;
	string = queue->TKFF_Token[start].TK_String;

	if(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END)
	    continue;

	switch(state) {
	case 0:
	    if(tokVal != CTK_IDENT)
		goto invalid;

	    if(strcmp(string, NAME_CLAUSE_INDENT) == 0)
		state = 10;
	    else if(strcmp(string, NAME_CLAUSE_MAPLEVEL) == 0)
		state = 20;
	    else if(strcmp(string, NAME_CLAUSE_RECURLEVEL) == 0)
		state = 30;
	    else
		goto invalid;

	    break;
	case 10: case 20: case 30:
	    if(tokVal != '=')
		goto invalid;
	    state += 2;
	    break;
	case 12: /*  "indent = <number>"  */
	case 22: /*  "maplevel = <number>"  */
	case 32: /*  "recurlevel = <number>"  */
	    if(tokVal == '-') {
		++state;
		break;
	    }
	case 13: /*  "indent = <number>"  */
	case 23: /*  "maplevel = <number>"  */
	case 33: /*  "recurlevel = <number>"  */
	    {
		int value;

		if(tokVal != CTK_NUMBER)
		    goto invalid;

		value = strtol(string, NULL, 0);
		if(state & 1)
		    value = -value;

		switch(state & ~1) {
		case 12:
		    paux->Indentation = value;
		    break;
		case 22:
		    paux->MaxMappingLevel = value;
		    break;
		case 32:
		    paux->MaxRecursiveLevel = value;
		    break;
		}
		state = 0;
	    }
	    break;
	default:
	    return -1;
	}
    }

    return start;
 invalid:
    return -2;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int parseTypeMap(jsparsstream_t *jpa, traceaux_t *paux,
			int start, int end, jctokfifo_t *queue)

/************************************************************************/
{
    jsparsvartab_t *vTab;
    typemapping_t tMap;
    int iColon, step;

    step = jctokgetstep(queue);

    /*  Find the separating colon  */
    for(iColon = start ; iColon <= end ; iColon += step)
	if(queue->TKFF_Token[iColon].TK_Value == ':')
	    break;

    /*  There must be something after the colon  */
    if(iColon >= end)
	goto invalid;

    /* ***   Read what follows the colon  *** */
    tMap.Descriptions = NULL;
    tMap.NDescriptions = 0;
    if((end = readArgs(paux, jpa, &(tMap.Descriptions), &(tMap.NDescriptions),
		       iColon + step, end, queue, 2 | 4 | 8)) == -1)
	EGOTO(error);

    /* ***  Read the type until the ':'  *** */
    if(queue != &(jpa->JSPA_GlobalQueue) &&
       (queue = copyToks(jpa, queue, &start, &iColon)) == NULL)
	EGOTO(error);

    if((vTab = readExprType(jpa, queue, start, iColon - step,
			    &(tMap.ExprType))) == NULL)
	EGOTO(error);

    if(addTypeMap(jpa, paux, vTab, &tMap) == -1)
	EGOTO(error);

    return end;
 error:
    return -1;
 invalid:
    return -2;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addTypeMap(jsparsstream_t *jpa, traceaux_t *paux,
		      jsparsvartab_t *vTab, typemapping_t *pMap)

/************************************************************************/
{
    int i;

    if(paux->NTypeMappings >= paux->MTypeMappings) {
	REALLOC_RESIZE(paux->MTypeMappings, 20);

	if((paux->TypeMappings = (typemapping_t *)
	    realloc(paux->TypeMappings, paux->MTypeMappings *
		    sizeof(*(paux->TypeMappings)))) == NULL)
	    EGOTO0(error);
    }

    /*  Make sure that types get sorted properly:
     *    typedef char MYCHAR;
     *   1. addTypeMap("char")
     *   2. addTypeMap("MYCHAR")
     *   We have to insert "MYCHAR" before "char", otherwise
     *   it will never be used.
     */
    for(i = 0 ; i < paux->NTypeMappings ; ++i) {
	int ret;

	if((ret = jsparsisvartype(jpa, vTab, &(pMap->ExprType),
				  &(paux->TypeMappings[i].ExprType), 0)) == -1)
	    EGOTO(error);

	if(ret != 0)
	    break;
    }

    if(i < paux->NTypeMappings)
	memmove(&(paux->TypeMappings[i + 1]), &(paux->TypeMappings[i]),
		(paux->NTypeMappings - i) * sizeof(paux->TypeMappings[i]));

    paux->TypeMappings[i] = *pMap;
    paux->NTypeMappings++;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int parseTypeRename(jsparsstream_t *jpa, traceaux_t *paux,
                           int start, int end, jctokfifo_t *queue)

/************************************************************************/
{
    const char *typeName, *sourceName, *targetName;
    int iColon, step;

    step = jctokgetstep(queue);

    typeName = sourceName = targetName = NULL;
    for(iColon = start ; iColon <= end ; iColon += step) {
	const char *string;
	int tokVal;

	tokVal = queue->TKFF_Token[iColon].TK_Value;

	if(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END)
	    continue;

	if(tokVal != CTK_IDENT)
	    goto invalid;

	string = queue->TKFF_Token[iColon].TK_String;
	if(typeName == NULL)
	    typeName = string;
	else if(sourceName == NULL)
	    sourceName = string;
	else if(targetName == NULL)
	    targetName = string;
	else
	    goto invalid;
    }

    if(targetName == NULL)
	goto invalid;

    if(addTypeRename(jpa, paux, typeName, sourceName, targetName) == -1)
	EGOTO(error);

    return end + step;
 error:
    return -1;
 invalid:
    return -2;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addTypeRename(jsparsstream_t *jpa, traceaux_t *paux, const char *typeName,
                         const char *sourceName, const char *targetName)

/************************************************************************/
{
    unsigned int lower;

    if(paux->NTypeRenames >= paux->MTypeRenames) {
	REALLOC_RESIZE(paux->MTypeRenames, 20);

	if((paux->TypeRenames = (typerename_t *)
	    realloc(paux->TypeRenames, paux->MTypeRenames *
		    sizeof(*(paux->TypeRenames)))) == NULL)
	    EGOTO0(error);
    }

    /*  Make sure that renames are sorted properly  */
    lower = locateTypeRename(paux, typeName, sourceName);

    if(lower < paux->NTypeRenames &&
       strcmp(paux->TypeRenames[lower].TypeName, typeName) == 0 &&
       strcmp(paux->TypeRenames[lower].SourceName, sourceName) == 0) {
	jsparsprinterr(jpa, NULL, NULL, "Duplicate type rename for %s.%s", 0, typeName, sourceName);
	EGOTO(error);
    }

    if(lower < paux->NTypeRenames)
	memmove(&(paux->TypeRenames[lower + 1]), &(paux->TypeRenames[lower]),
		(paux->NTypeRenames - lower) * sizeof(paux->TypeRenames[lower]));

    if((paux->TypeRenames[lower].TypeName   = strdup(typeName)) == NULL ||
       (paux->TypeRenames[lower].SourceName = strdup(sourceName)) == NULL ||
       (paux->TypeRenames[lower].TargetName = strdup(targetName)) == NULL)
	EGOTO(error);

    paux->NTypeRenames++;

#ifdef TRACETYPE_JSCPP
    jstrace(TRACETYPE_JSCPP, 3,
	    "Added type rename '%s': '%s' -> '%s'\n",
	    paux->TypeRenames[lower].TypeName, paux->TypeRenames[lower].SourceName,
	    paux->TypeRenames[lower].TargetName);
#endif

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int locateTypeRename(traceaux_t *paux, const char *typeName,
                     const char *sourceName)

/************************************************************************/
{
    unsigned int lower, upper;

    /*  Make sure that renames are sorted properly  */
    lower = 0;   upper = paux->NTypeRenames;
    while(lower < upper) {
	unsigned int m = (lower + upper) >> 1;
	int cmp;

	cmp = strcmp(paux->TypeRenames[m].TypeName, typeName);
	if(cmp == 0 && sourceName != NULL)
	    cmp = strcmp(paux->TypeRenames[m].SourceName, sourceName);

	if(cmp < 0)
	    lower = m + 1;
	else
	    upper = m;
    }

    return lower;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int findTypeRename(traceaux_t *paux, const char *typeName, const char *sourceName)

/************************************************************************/
{
    unsigned int lower;

    lower = locateTypeRename(paux, typeName, sourceName);
    if(lower < paux->NTypeRenames &&
       strcmp(paux->TypeRenames[lower].TypeName, typeName) == 0 &&
       (sourceName == NULL ||
	strcmp(paux->TypeRenames[lower].SourceName, sourceName) == 0))
        return lower;

    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int parseTypeFormat(jsparsstream_t *jpa, traceaux_t *paux,
			   int start, int end, jctokfifo_t *queue)

/************************************************************************/
{
    jsparsvartab_t *vTab;
    typeformat_t lMap;
    char formatMem[2][2];
    int iColon, indx, step, tokVal, endType, nFormat;

    step = jctokgetstep(queue);

    /*  Find the separating colon  */
    for(iColon = start ; iColon <= end ; iColon += step)
	if(queue->TKFF_Token[iColon].TK_Value == ':')
	    break;

    /*  There must be something after the colon  */
    if(iColon >= end) {
	int flag;

	/*  Alternative syntax: single word "off" to switch off
	 *   internal type mappings.
	 */
	for(flag = FALSE ; start <= end ; start += step) {
	    tokVal = queue->TKFF_Token[start].TK_Value;

	    if(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END)
		continue;

	    if(tokVal != CTK_IDENT)
		EGOTO(invalid);

	    if(strcmp(queue->TKFF_Token[start].TK_String, NAME_CLAUSE_OFF) == 0)
		paux->Mode &= ~JSCPP_TRACE_INTTYPEMAPPING;
	    else
		EGOTO(invalid);

	    flag = TRUE;
	}

	if(!flag) /*  empty line - not allowed  */
	    EGOTO(invalid);

	return end + queue->TKFF_Step;
    }

    /*  Look for the letter string(s)
     *   We only allow four different formats:
     *   1. "nolabelstring" "labelstring"
     *   2. "nolabelstring"
     *   3. "nolabelstring" nolabel
     *   4. "labelstring" label
     */
    lMap.FFlag = 0;
    lMap.Format[0] = lMap.Format[1] = NULL;
    for(nFormat = 0, indx = iColon + step ;
	nFormat < 2 && indx <= end ; indx += step) {
	tokVal = queue->TKFF_Token[indx].TK_Value;
	if(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END)
	    continue;

	/*  Clause at second position  */
	if(nFormat == 1 && tokVal == CTK_IDENT) {
	    char *s = queue->TKFF_Token[indx].TK_String;

	    if(strcmp(s, NAME_CLAUSE_NOLABEL) == 0) {
		lMap.FFlag |= 2;
	    } else if(strcmp(s, NAME_CLAUSE_LABEL) == 0) {
		lMap.Format[1] = lMap.Format[0];
		lMap.Format[0] = NULL;
		lMap.FFlag |= 4;
	    } else
		EGOTO(invalid);
	} else {
	    /*  The letter string will be duplicated by addTypeFormat  */
	    if(tokVal == CTK_STRING) {
		lMap.Format[nFormat] = queue->TKFF_Token[indx].TK_String;
	    } else if(tokVal < 256) {
		formatMem[nFormat][0] = tokVal;
		formatMem[nFormat][1] = '\0';
		lMap.Format[nFormat] = formatMem[nFormat];
	    } else
		EGOTO(invalid);
	}

	++nFormat;
    }

    if(nFormat == 0)
	/*  No string after the colon  */
	EGOTO(invalid);

    if(queue != &(jpa->JSPA_GlobalQueue) &&
       (queue = copyToks(jpa, queue, &start, &iColon)) == NULL)
	EGOTO(error);

    endType = iColon - step;
    for( ; start <= endType ; start += step) {
	tokVal = queue->TKFF_Token[start].TK_Value;
	if(!(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END))
	    break;
    }

    /*  Check for special syntax __POINTER__  */
    if(start <= endType && strcmp(queue->TKFF_Token[start].TK_String,
				  NAME_POINTER_TYPE) == 0) {
	lMap.FFlag |= 1; /*  just mark "any pointer"  */
	vTab = NULL;
    } else {
	/*  The remainder must be a type, convert it to a tree  */
	if((vTab = readExprType(jpa, queue, start, endType,
				&(lMap.ExprType))) == NULL)
	    EGOTO(error);
    }

    if(addTypeFormat(jpa, paux, vTab, &lMap, queue) == -1)
	EGOTO(error);

    return end + queue->TKFF_Step;
 error:
    return -1;
 invalid:
    return -2;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addTypeFormat(jsparsstream_t *jpa, traceaux_t *paux,
			 jsparsvartab_t *vTab,
			 typeformat_t *lMap, jctokfifo_t *queue)

/************************************************************************/
{
    int i;

    if(paux->NTypeFormats >= paux->MTypeFormats) {
	REALLOC_RESIZE(paux->MTypeFormats, 20);

	if((paux->TypeFormats = (typeformat_t *)
	    realloc(paux->TypeFormats, paux->MTypeFormats *
		    sizeof(*(paux->TypeFormats)))) == NULL)
	    EGOTO0(error);
    }

    /*  Copy the strings if they exist  */
    for(i = 0 ; i < 2 ; ++i)
	if(lMap->Format[i] && (lMap->Format[i] = strdup(lMap->Format[i])) == NULL)
	    EGOTO0(error);

    /*  Make sure that types get sorted properly:
     *    typedef char MYCHAR;
     *   1. addTypeFormat("char")
     *   2. addTypeFormat("MYCHAR")
     *   We have to insert "MYCHAR" before "char", otherwise
     *   it will never be used.
     */
    for(i = 0 ; i < paux->NTypeFormats ; ++i) {
	if(lMap->FFlag & 1) {
	    /*  Generic pointer type format gets substituted or
	     *   inserted at the end
	     */
	    if(paux->TypeFormats[i].FFlag & 1)
		break;
	} else {
	    int ret;

	    /*  Specific format has to be inserted BEFORE generic pointer type  */
	    if(paux->TypeFormats[i].FFlag & 1)
		break;

	    if((ret = jsparsisvartype(jpa, vTab, &(lMap->ExprType),
				      &(paux->TypeFormats[i].ExprType), 0)) == -1)
		EGOTO(error);

	    if(ret != 0)
		break;
	}
    }

    if(i < paux->NTypeFormats && (lMap->FFlag & 1)) {
	/*  Replace it  */
	free((void *)paux->TypeFormats[i].Format[0]);
	free((void *)paux->TypeFormats[i].Format[1]);

	paux->TypeFormats[i] = *lMap;
    } else {
	/*  Insert it  */
	if(i < paux->NTypeFormats)
	    memmove(&(paux->TypeFormats[i + 1]), &(paux->TypeFormats[i]),
		    (paux->NTypeFormats - i) * sizeof(paux->TypeFormats[i]));

	paux->TypeFormats[i] = *lMap;
	paux->NTypeFormats++;
    }

    return 0;
 error:
    return -1;
}
