 /*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application for preprocessor library.
*
*    Part    : Include file dependency
*
*              Copyright (C) 2003-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include "jscpp.h"

/*********     DEFINES                                          *********/

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doIDeps(jsppenviron_t *ppenv, const char *names[],
				 int nnames, FILE *fpOut);
static int               prDeps(jsppstream_t *jsp, const char *name,
				int incLevel, int flag);
static int               addIName(const char *name);
static void              sortFileNames(char **array, unsigned long n);
static void              freeNames(void);

static FILE *FpOut;
static char **FileNames;
static int NFileNames, MFileNames;

/*JS*********************************************************************
*
*************************************************************************/

int doIDeps(jsppenviron_t *ppenv, const char *names[], int nnames,
	    FILE *fpOut)

/************************************************************************/
{
  int i;

  FpOut = fpOut;

  for(i = 0 ; i < nnames ; ++i) {
    jsppstream_t *jps;
    jsctoken_t token;
    int pos;

    if((jps = getPreProcStream(ppenv, names[i], 0)) == NULL)
      EGOTO(error);

    /*  Make preprocessor call user routine  */
    jps->JSPP_OpenFile = prDeps;

    freeNames();
    if(addIName(names[i]) == -1)
      EGOTO(error);

    /*  Read and discard tokens  */
    do {
      token = jsppgettoken(jps, NULL, 0);

      if(token.TK_Value == CTK_ERROR)
	EGOTO(error);
    } while(token.TK_Value != CTK_EOF);

    pos = fprintf(fpOut, "%s:", FileNames[0]);
    if(NFileNames > 1) {
      int j;

      sortFileNames(FileNames + 1, NFileNames - 1);
      for(j = 1 ; j < NFileNames ; ++j)
	if(strcmp(FileNames[j - 1], FileNames[j]) != 0) {
	  /*  Print next dependency  */
	  if((pos + strlen(FileNames[j]) + 1) > 78) {
	    /*  Start new line  */
	    fprintf(FpOut, " \\\n\t");
	    pos = 8 + fprintf(FpOut, "%s", FileNames[j]);
	  } else
	    pos += fprintf(FpOut, " %s", FileNames[j]);
	}
    }
    putc('\n', fpOut);

    if(closePreProcStream(jps, 0) == -1)
      EGOTO(error);
  }

  freeNames();

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int prDeps(jsppstream_t *jps, const char *name, int incLevel, int flag)

/************************************************************************/
{
  if(incLevel != 0 && !(flag & IF_SYSINCLUDE)) {
    const char *s;

    s = (name[0] == '.' && name[1] == '/') ? name + 2 : name;
    if(addIName(s) == -1)
      return -1;
  }

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int addIName(const char *name)

/************************************************************************/
{
  if(NFileNames >= MFileNames) {
    REALLOC_RESIZE(MFileNames,10);
    if((FileNames = (char **)
	realloc(FileNames, MFileNames * sizeof(*FileNames))) == NULL)
      return -1;
  }

  if((FileNames[NFileNames] = strdup(name)) == NULL)
    return -1;

  NFileNames++;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static void freeNames(void)

/************************************************************************/
{
  if(FileNames) {
    int j;

    for(j = 0 ; j < NFileNames ; ++j)
      free(FileNames[j]);
    free(FileNames);
    FileNames = NULL;
  }

  NFileNames = MFileNames = 0;
}

/*JS*********************************************************************
*
*************************************************************************/

/*  Generated automatically with "gensort.h"  */
#define GENSORT_NAME		     sortFileNames
#define GENSORT_TYPE		     char *
#define GENSORT_KEYTYPE 	     char *
#define GENSORT_GETKEY(a)            a
#define GENSORT_COMPAREKEYS(k1, k2)   strcmp(k1, k2) < 0
#define GENSORT_STATIC
#include <gensort.h>
