/*JS*********************************************************************
 *
 *    Program : JSPARSE
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : ANSI-C parser.
 *    Part    : Routines to scan
 *
 *              Copyright (C) 2004-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/
/*#define NDEBUG*/
#include <assert.h>

#include <stdlib.h>

#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES						*********/
#ifndef JPA_VARTYPE_FALLOCDECL
/*  START-DEFINITIONS 2 */
typedef struct {
    unsigned long JSPE_Flags;

    /*  Only valid if flag JPA_VARTYPE_FLONGCONST or JPA_VARTYPE_FARRAYSIZE
     *  set
     */
    long JSPE_LongConst;

    jsparscnode_t *JSPE_Declarator; /*  without top node JPACT_DECLARATOR */
    jsparscnode_t *JSPE_Specifier;
    jsparsvartab_t *JSPE_SubTable;

    jctokfifo_t   *JSPE_Queue;
} jsparsexprtype_t;

/*  For JSPE_Flags field  */
#define JPA_VARTYPE_FALLOCDECL     1
#define JPA_VARTYPE_FCOPYDECL      2
#define JPA_VARTYPE_FFREESUBTAB    4
#define JPA_VARTYPE_FISINTNULL     8

#define JPA_VARTYPE_FLONGCONST  1024
#define JPA_VARTYPE_FARRAYSIZE  2048
/*  END-DEFINITIONS */
#endif

#define Prototype extern
/*********     PROTOTYPES					*********/
Prototype int            jsparsgenvardef2(jsparsstream_t *jpa,
					  jstrbuff_t *sbuffer,
					  jsparsexprtype_t *pType,
					  const char *varname);
Prototype int            jsparsgenvardef(jsparsstream_t *jpa,
					 jstrbuff_t *sbuffer,
					 jsparsvardef_t *vDef,
					 const char *varname);
Prototype int            jsparsgenvardef3(jsparsstream_t *jpa,
					  jstrbuff_t *sbuffer,
					  jsparscnode_t *decl,
					  jsparscnode_t *spec,
					  jctokfifo_t *queue,
					  const char *varname);
static int               dogenvardef(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
				     unsigned int dStarts[2], unsigned int dEnds[2],
				     jsparscnode_t *spec, jctokfifo_t *queue,
				     const char *varname);
static jsctoken_t        hidespecifiers(void *user, jsctoken_t token);

Prototype int            jsparsgenwrapper(jsparsstream_t *jpa,
					  jstrbuff_t *sbuffer,
					  jsparscnode_t *node,
					  const char *newname,
					  const char *retvar,
					  const char *strings[3],
					  jctokfifo_t *queue,
					  jstrbuff_t *auxBuffer,
					  const char *(*getArgName)(void *user),
					  void *uPtr, int mode);
Prototype int            jsparsgenwrapper2(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
					   jsparscnode_t *decl, jsparscnode_t *spec,
					   const char *newname, const char *retvar,
					   const char *strings[2],
					   jctokfifo_t *queue, jstrbuff_t *auxBuffer,
					   const char *(*getArgName)(void *user),
					   void *uPtr, int mode);

/* ERROR-DEFINITIONS from jsparsgenvardef2 label _ERR_JSPARSGENVARDEF
   Illegal declarator
*/

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsgenvardef2(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
		     jsparsexprtype_t *pType, const char *varname)

/************************************************************************/
{
    jsparscnode_t *decl;
    jctokfifo_t *queue;
    unsigned int starts[2], ends[2], step, n;

    queue = pType->JSPE_Queue;
    step = jctokgetstep(queue);

    n = 0;
    for(decl = pType->JSPE_Declarator ; decl ; decl = decl->JPACN_Next) {
	unsigned int i;

	/*
	 *           Case I                  Case II
	 *         |<-decl->|  |<-area->|  |<-decl->|
	 *
	 *       START     END START   END START   END
	 *         |        |  |        |  |        |
	 *         v        V  V        V  V        V
	 *  ----------------------------------------------------------------
	 *  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
	 *  ----------------------------------------------------------------
	 */
	for(i = 0 ; i < n ; ++i) {
	    /*  Case I: Check if current declarator is adjacent from the left  */
	    if(decl->JPACN_End + step == starts[i]) {
		starts[i] = decl->JPACN_Start;
		break;
	    }

	    /*  Case II: ... or from the right  */
	    if(ends[i] + step == decl->JPACN_Start) {
		ends[i] = decl->JPACN_End;
		break;
	    }
	}

	if(i == n) {
	    /*  New area  */
	    if(n == 2) {
#if 0
		fprintf(stderr, "GOT [%d,%d] [%d,%d], merge [%d,%d]\n",
			starts[0], ends[0], starts[1], ends[1],
			decl->JPACN_Start, decl->JPACN_End);
#endif

		jsparsseterrcode(_ERR_JSPARSGENVARDEF, 0, queue, decl);
		EGOTO(error);
	    }
	    starts[n] = decl->JPACN_Start;
	    ends[n]   = decl->JPACN_End;
	    ++n;
	}
    }

    while(n < 2)
	starts[n++] = -1;

#if 0
    fprintf(stderr, "GOT [%d,%d] [%d,%d]\n",
	    starts[0], ends[0], starts[1], ends[1]);
#endif

    return dogenvardef(jpa, sbuffer, starts, ends,
		       pType->JSPE_Specifier, queue, varname);
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsgenvardef(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
		    jsparsvardef_t *vDef, const char *varname)

/************************************************************************/
{
    return jsparsgenvardef3(jpa, sbuffer, vDef->JPAVD_Declarator,
			    vDef->JPAVD_Specifier,
			    vDef->JPAVD_Queue, varname);
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsgenvardef3(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
		     jsparscnode_t *decl, jsparscnode_t *spec,
		     jctokfifo_t *queue, const char *varname)

/************************************************************************/
{
    jsparscnode_t *node;
    unsigned int starts[2], ends[2];

    node = decl->JPACN_Sub;
    assert(node->JPACN_Type == JPACT_IDENT);

    starts[0] = decl->JPACN_Start;
    ends[0]   = node->JPACN_Start - queue->TKFF_Step;
    if(ends[0] < starts[0])
	starts[0] = -1;

    starts[1] = node->JPACN_End   + queue->TKFF_Step;
    ends[1]   = decl->JPACN_End;

    /*  Exclude function arguments if this is a function declaration  */
    node = node->JPACN_Next;
    if(node != NULL && node->JPACN_Type == JPACT_FUNCCALL)
	starts[1] = node->JPACN_End + queue->TKFF_Step;

    if(ends[1] < starts[1])
	starts[1] = -1;

    return dogenvardef(jpa, sbuffer, starts, ends, spec, queue, varname);
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int dogenvardef(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
		       unsigned int dStarts[2], unsigned int dEnds[2],
		       jsparscnode_t *spec, jctokfifo_t *queue,
		       const char *varname)

/************************************************************************/
{
    /*  "(" in case of cast  */
    if(varname == NULL && jstrbuffprint(sbuffer, "(") == -1)
	EGOTO(error);

    if(spec != NULL) {
	jsparscnode_t *node;

	node = spec;
	while(node->JPACN_Next)
	    node = node->JPACN_Next;

	if(jctoktostrbuff(sbuffer, queue, node->JPACN_Start,
			  spec->JPACN_End, hidespecifiers, NULL) == -1)
	    EGOTO(error);

	if(jstrbuffprint(sbuffer, " ") == -1)
	    EGOTO(error);
    } else {
	/*  The default is type int  */
	if(jstrbuffprint(sbuffer, "int ") == -1)
	    EGOTO(error);
    }

    if(dStarts[0] != (unsigned int)-1 &&
       jctoktostrbuff(sbuffer, queue, dStarts[0], dEnds[0],
		      NULL, NULL) == -1)
	EGOTO(error);

    /*  varname may be NULL. We then basically generate a cast  */
    if(varname != NULL && jstrbuffprint(sbuffer, varname) == -1)
	EGOTO(error);

    if(dStarts[1] != (unsigned int)-1 &&
       jctoktostrbuff(sbuffer, queue, dStarts[1], dEnds[1],
		      NULL, NULL) == -1)
	EGOTO(error);

    /*  ")" in case of cast  */
    if(varname == NULL && jstrbuffprint(sbuffer, ")") == -1)
	EGOTO(error);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static jsctoken_t hidespecifiers(void *user, jsctoken_t token)

/************************************************************************/
{
    if(token.TK_Value == CTK_IDENT) {
	if(tkStrcmp(token.TK_String, "static") == 0 ||
	   tkStrcmp(token.TK_String, "extern") == 0) {
	    token.TK_Value = PTK_EMPTY;
	    token.TK_String = NULL;
	}
    }

    return token;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsgenwrapper(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
		     jsparscnode_t *node,
		     const char *newname, const char *retvar,
		     const char *strings[2],
                     jctokfifo_t *queue, jstrbuff_t *auxBuffer,
		     const char *(*getArgName)(void *user),
		     void *uPtr, int mode)

/************************************************************************/
{
    jsparscnode_t *sub, *spec;

    if(node->JPACN_Type == JPACT_FUNCDEF)
	node = node->JPACN_Sub;

    if(node->JPACN_Type != JPACT_DECLARATION)
	EGOTO(error);

    node = node->JPACN_Sub;
    sub = (node->JPACN_Type == JPACT_LIST) ? node->JPACN_Sub : node;
    spec = node->JPACN_Next;

    /*  Walk through all declarations  */
    for( ; sub ; sub = sub->JPACN_Next) {
	int ret;

	if((ret = jsparsgenwrapper2(jpa,sbuffer, sub, spec, newname, retvar, strings, queue,
				    auxBuffer, getArgName, uPtr, mode)) == -1)
	    EGOTO(error);

	/*  This doesn't work, print at least a comment  */
	if(ret == 0 && (mode & 1) &&
	   jstrbuffprintf(sbuffer,
			  "/*  Function is variadic - no wrapper generated  */\n") == -1)
	    EGOTO(error);

	if(node->JPACN_Type != JPACT_LIST)
	    break;
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Generate a wrapper function from the prototype or from the ANSI
 *    type function definition. The wrapper routine is named 'newname'
 *    which may contain a '%' which is substituted by the function name.
 *    strings[0] is put before the actual function call.
 *    strings[1] or strings[2] is put after the actual function call.
 *               The first one is used for non-void functions, the second
 *               one for void functions
 *************************************************************************/

int jsparsgenwrapper2(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
		      jsparscnode_t *decl, jsparscnode_t *spec,
		      const char *newname, const char *retvar,
		      const char *strings[3],
		      jctokfifo_t *queue, jstrbuff_t *auxBuffer,
		      const char *(*getArgName)(void *user),
		      void *uPtr, int mode)

/************************************************************************/
{
    jsparscnode_t *node, *node2;
    unsigned int step, isVoid, specStart, specEnd;
    struct { unsigned int Start, End; } ranges[3];
    const char *funcName;
    unsigned int thisIsVoid, nr, j;

#if 0
    printf("----------\n");
    jsparsprctree(stdout, queue, decl, 0, 2);
    printf("----------\n");
    jsparsprctree(stdout, queue, spec, 0, 4);
    printf("----------\n");
#endif

    step = jctokgetstep(queue);

    /*  Find the begin and end of the specifiers  */
    specStart = specEnd = spec->JPACN_End;
    isVoid = FALSE;
    for(node = spec ; node ; node = node->JPACN_Next) {
	if(node->JPACN_Type == JPACT_VOID) isVoid = TRUE;

	if(node->JPACN_End < specStart) specStart = node->JPACN_End;
    }

    /*  This is where we generate the body of the wrapper function  */
    jstrbuffempty(auxBuffer);

    /*
     * ===========DECLARATOR=============....
     * ......IDENTIFIER......(ARGS)......
     * <---->          <---->      <---->
     *  r[0]            r[1]        r[2]
     */

    assert(decl->JPACN_Type == JPACT_DECLARATOR);
    ranges[0].Start = decl->JPACN_Start;
    ranges[2].End   = decl->JPACN_End;

    node = decl->JPACN_Sub;
    assert(node->JPACN_Type == JPACT_IDENT);
    funcName = queue->TKFF_Token[node->JPACN_IndexPP].TK_String;
    ranges[0].End   = node->JPACN_Start - step;
    ranges[1].Start = node->JPACN_End + step;

    node = node->JPACN_Next;
    if(node == NULL || node->JPACN_Type != JPACT_FUNCCALL)
	/*  A declaration for a variable, not a function prototype  */
	return 0;

    ranges[1].End   = node->JPACN_Start;
    ranges[2].Start = node->JPACN_End;

    if((node2 = node->JPACN_Next))
	/*  maybe "void *" or similar  */
	thisIsVoid = FALSE;
    else
	thisIsVoid = isVoid;

    node = node->JPACN_Sub;

    /*  Check for variadic arguments  */
    for(node2 = node ; node2 ; node2 = node2->JPACN_Next)
	if(node2->JPACN_Sub->JPACN_Type == JPACT_ELLIPSIS)
	    return 0;

    if(jstrbuffprint(auxBuffer, (mode & 2) ? "\n{\n" : "{") == -1)
	EGOTO(error);

    /*  Generate a temporary variable for the return value  */
    if(!thisIsVoid) {
	if(((mode & 2) && jstrbuffprint(auxBuffer, "  ") == -1) ||
	   jsparsgenvardef3(jpa, auxBuffer, decl, spec,
			    queue, retvar) == -1)
	    EGOTO(error);

	if(jstrbuffprint(auxBuffer, (mode & 2) ? ";\n" : ";") == -1)
	    EGOTO(error);
    }

    if(strings && strings[0] &&
       jstrbuffprint(auxBuffer, strings[0]) == -1)
	EGOTO(error);

    if(thisIsVoid) {
	if(jstrbuffprintf(auxBuffer, (mode & 2) ? "  %s(" : "%s(",
			  funcName) == -1)
	    EGOTO(error);
    } else {
	if(jstrbuffprintf(auxBuffer, (mode & 2) ? "  %s = %s(" : "%s = %s(",
			  retvar, funcName) == -1)
	    EGOTO(error);
    }

    /*  Print specifier to destination  */
    if(jctoktostrbuff(sbuffer, queue, specStart, specEnd, hidespecifiers,
		      NULL) == -1 ||
       jstrbuffprint(sbuffer, " ") == -1)
	EGOTO(error);

    /*  Generate first part of declarator up to function name  */
    if(jctoktostrbuff(sbuffer, queue, ranges[0].Start, ranges[0].End,
		      NULL, NULL) == -1)
	EGOTO(error);

    /*  Substitute function name with pregiven wrapped name  */
    if(jstrbuffstore(sbuffer, NULL, strlen(newname) + strlen(funcName)) == -1)
	EGOTO(error);

    for(j = 0 ; newname[j] ; ++j)
	if(newname[j] == '%') {
	    if(jstrbuffprint(sbuffer, funcName) == -1)
		EGOTO(error);
	} else
	    jstrbuffputc(sbuffer, newname[j]);

    /*  Generate next part up to function arguments  */
    if(jctoktostrbuff(sbuffer, queue, ranges[1].Start, ranges[1].End,
		      NULL, NULL) == -1)
	EGOTO(error);

    /*  put in function args  */
    for(nr = 0 ; node ; node = node->JPACN_Next, ++nr) {
	const char *parName;
	unsigned int pos, spaceFlag;

	pos = node->JPACN_End;
	spaceFlag = FALSE;

	assert(node->JPACN_Type == JPACT_PARDECL);
	node2 = node->JPACN_Sub;
	if(node2->JPACN_Type == JPACT_DECLARATOR) {
	    node2 = node2->JPACN_Sub;
	    if(node2 && node2->JPACN_Type == JPACT_IDENT) {
		parName = queue->TKFF_Token[node2->JPACN_IndexPP].TK_String;
		pos = -1;
	    } else {
		if((parName = getArgName(uPtr)) == NULL)
		    EGOTO(error);

		/*  this is where the separator is supposed to go  */
		pos = node2->JPACN_End;
	    }
	} else if(node2->JPACN_Type == JPACT_VOID) {
	    /*  no function arguments at all  */
	    parName = NULL;
	    pos = -1;
	} else {
	    if((parName = getArgName(uPtr)) == NULL)
		EGOTO(error);

	    /*  Include a space to make sure specifier and declarator
	     *   are well separated.
	     */
	    spaceFlag = TRUE;
	}

	if(nr > 0 && jstrbuffprint(sbuffer, ", ") == -1)
	    EGOTO(error);

	for(j = node->JPACN_Start ; j <= node->JPACN_End ; j += step) {
	    if(jctoktostrbuff1(sbuffer, queue->TKFF_Token[j]) == -1)
		EGOTO(error);

	    if(j == pos) {
		if(spaceFlag)
		    jstrbuffputc(sbuffer, ' ');

		if(jstrbuffprint(sbuffer, parName) == -1)
		    EGOTO(error);
	    }
	}

	if(parName != NULL) {
	    if(nr > 0 && jstrbuffprint(auxBuffer, ", ") == -1)
		EGOTO(error);

	    if(jstrbuffprint(auxBuffer, parName) == -1)
		EGOTO(error);
	}
    }

    if(jstrbuffprint(auxBuffer, (mode & 2) ? ");\n" : ");") == -1)
	EGOTO(error);

    /*  Generate remainder of function body  */
    if(strings && strings[thisIsVoid ? 2 : 1] &&
       jstrbuffprint(auxBuffer, strings[thisIsVoid ? 2 : 1]) == -1)
	EGOTO(error);

    if(!thisIsVoid && jstrbuffprintf(auxBuffer, (mode & 2) ?
				     "  return %s;\n" : "return %s;",
				     retvar) == -1)
	EGOTO(error);

    if(jstrbuffprint(auxBuffer, (mode & 2) ? "}\n" : " } ") == -1)
	EGOTO(error);

    /*  Generate remainder of declarator  */
    if(jctoktostrbuff(sbuffer, queue, ranges[2].Start, ranges[2].End,
		      NULL, NULL) == -1)
	EGOTO(error);

    /*  Print the function body to the destination  */
    if(jstrbuffprint(sbuffer, jstrbuffer(auxBuffer)) == -1)
	EGOTO(error);

    return 1;
 error:
    return -1;
}
