/*JS*********************************************************************
*
*    Program : JSPARSE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : ANSI-C parser.
*    Part    : Routines to scan
*
*              Copyright (C) 2004-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES 					*********/
/*#define NDEBUG*/
#include <assert.h>

#include <stdlib.h>

#include <jsalloca.h>

#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS */
/*  Returned values from jsparsgetvartype* routines  */
#define JPA_VARTYPE_VOID     (1 << 0)
#define JPA_VARTYPE_CHAR     (1 << 1)
#define JPA_VARTYPE_SHORT    (1 << 2)
#define JPA_VARTYPE_INT      (1 << 3)
#define JPA_VARTYPE_LONG     (1 << 4)
#define JPA_VARTYPE_LONGLONG (1 << 5)

#define JPA_VARTYPE_FLOAT    (1 << 8)
#define JPA_VARTYPE_DOUBLE   (1 << 9)

#define JPA_VARTYPE_WCHAR    (1 << 12)

#define JPA_VARTYPE_STRUCT   (1 << 16)
#define JPA_VARTYPE_UNION    (1 << 17)
#define JPA_VARTYPE_ENUM     (1 << 18)
/*  END-DEFINITIONS */

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
Prototype jsparsexprtype_t *jsparsmakeetype(jsparsexprtype_t *pType,
					    jsparscnode_t *decl,
					    jsparscnode_t *spec,
					    jctokfifo_t *queue);

Prototype jsparsexprtype_t *jsparsnode2type(jsparsexprtype_t *pType,
					    jsparscnode_t *node,
					    jctokfifo_t *queue);

Prototype jsparsexprtype_t *jsparsvdef2type(jsparsstream_t *jpa,
					    jsparsvartab_t *vTab,
					    jsparsexprtype_t *pType,
					    const jsparsvardef_t *vDef,
					    int mode);

Prototype long           jsparsgetvartype3(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab,
					   jsparsvardef_t *vDef,
					   int *pPtrCnt, int mode);

Prototype long           jsparsgetvartype2(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab,
					   jsparsexprtype_t *pType,
					   int *pPtrCnt, int mode);

Prototype long           jsparsgetvartype(jsparsstream_t *jpa,
					  jsparsvartab_t *vTab,
					  jsparscnode_t *decl,
					  jsparscnode_t *spec,
					  jctokfifo_t *queue, int *pPtrCnt,
					  int mode);

Prototype int            jsparsisvartype(jsparsstream_t *jpa,
					 jsparsvartab_t *vTab,
					 jsparsexprtype_t *type,
					 jsparsexprtype_t *refType,
					 int mode);
static int               compareDeclarators(jsparscnode_t *decl,
					    jsparscnode_t *refDecl,
					    int mode);
static int               compareSpecifiers(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab,
					   jsparscnode_t *spec,
					   jctokfifo_t *queue,
					   jsparscnode_t *refSpec,
					   jctokfifo_t *refQueue, int mode);

IPrototype int           _jsparsres1type(jsparsstream_t *jpa,
					 jsparsvartab_t *vTab,
					 jsparsexprtype_t *pType);
IPrototype int           _jsparsrestype(jsparsstream_t *jpa,
					jsparsvartab_t *vTab,
					jsparscnode_t *node,
					jctokfifo_t *queue,
					jsparsexprtype_t *pType);

IPrototype void          _jsparssetuptype(jsparsexprtype_t *pType,
					  const jsparsvardef_t *vDef);

/*JS*********************************************************************
*
*************************************************************************/

jsparsexprtype_t *jsparsmakeetype(jsparsexprtype_t *pType,
				  jsparscnode_t *decl, jsparscnode_t *spec,
				  jctokfifo_t *queue)

/************************************************************************/
{
  if(pType == NULL)
    if((pType = (jsparsexprtype_t *)malloc(sizeof(*pType))) == NULL)
      EGOTO0(error);

  if(decl != NULL) {
    if(decl->JPACN_Type == JPACT_DECLARATOR)
      decl = decl->JPACN_Sub;

    if(decl != NULL && decl->JPACN_Type == JPACT_IDENT)
      decl = decl->JPACN_Next;
  }

  pType->JSPE_Flags = 0;
  pType->JSPE_Declarator = decl;
  pType->JSPE_Specifier = spec;
  pType->JSPE_SubTable = NULL;
  pType->JSPE_Queue = queue;

  return pType;
error:
  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

jsparsexprtype_t *jsparsnode2type(jsparsexprtype_t *pType,
				  jsparscnode_t *node,
				  jctokfifo_t *queue)

/************************************************************************/
{
  jsparscnode_t *decl, *spec;

  spec = node->JPACN_Sub;
  if(spec->JPACN_Type == JPACT_DECLARATOR) {
    decl = spec;
    spec = spec->JPACN_Next;
  } else
    decl = NULL;

  return jsparsmakeetype(pType, decl, spec, queue);
}

/*JS*********************************************************************
*
*************************************************************************/

jsparsexprtype_t *jsparsvdef2type(jsparsstream_t *jpa,
				  jsparsvartab_t *vTab,
				  jsparsexprtype_t *pType,
				  const jsparsvardef_t *vDef, int mode)

/************************************************************************/
{
  if(pType == NULL)
    if((pType = (jsparsexprtype_t *)malloc(sizeof(*pType))) == NULL)
      EGOTO0(error);

  _jsparssetuptype(pType, vDef);

  if(mode & 1) {
    jsparscnode_t *node;

    node = pType->JSPE_Declarator;

    if(node && node->JPACN_Type == JPACT_FUNCCALL)
      pType->JSPE_Declarator = node->JPACN_Next;
  }

  if(mode & 2) {
    int ret;

    do {
      if((ret = _jsparsres1type(jpa, vTab, pType)) == -1)
	EGOTO(error);
    } while(ret != 0); /*  repeat it until we reach a basic type  */
  }

  return pType;
error:
  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

long jsparsgetvartype3(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		       jsparsvardef_t *vDef, int *pPtrCnt, int mode)

/************************************************************************/
{
  jsparscnode_t *decl;

  decl = vDef->JPAVD_Declarator->JPACN_Sub;

  return jsparsgetvartype(jpa, vTab, decl, vDef->JPAVD_Specifier,
			  vDef->JPAVD_Queue, pPtrCnt, mode);
}

/*JS*********************************************************************
*
*************************************************************************/

long jsparsgetvartype2(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		       jsparsexprtype_t *pType, int *pPtrCnt, int mode)

/************************************************************************/
{
  return jsparsgetvartype(jpa, vTab, pType->JSPE_Declarator,
			  pType->JSPE_Specifier, pType->JSPE_Queue,
			  pPtrCnt, mode);
}

/* ERROR-DEFINITIONS from jsparsgetvartype label _ERR_JSPARSGETVARTYPE
   Illegal node encountered
*/

/*JS*********************************************************************
*  mode bit 0: If set, function calls is not considered as indirection
*  mode bit 1: If set, recognize "wchar_t" as special type.
*************************************************************************/

long jsparsgetvartype(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		      jsparscnode_t *decl, jsparscnode_t *spec,
		      jctokfifo_t *queue, int *pPtrCnt, int mode)

/************************************************************************/
{
  int ptrCnt, flag;

  /*  Count level of indirection of this type  */
  for(ptrCnt = 0 ; decl ; decl = decl->JPACN_Next)
    if(decl->JPACN_Type == JPACT_VECTOR ||
       decl->JPACN_Type == JPACT_POINTER ||
       (!(mode & 1) && decl->JPACN_Type == JPACT_FUNCCALL))
      ++ptrCnt;

  for(flag = 0 ; spec ; spec = spec->JPACN_Next) {
    switch(spec->JPACN_Type) {
    case JPACT_VOID:   flag |= JPA_VARTYPE_VOID;  break;
    case JPACT_CHAR:   flag |= JPA_VARTYPE_CHAR;  break;
    case JPACT_SHORT:  flag |= JPA_VARTYPE_SHORT; break;
    case JPACT_INT:    flag |= JPA_VARTYPE_INT;   break;
    case JPACT_LONG:
      if(flag & JPA_VARTYPE_LONG)
	flag = (flag & ~JPA_VARTYPE_LONG) | JPA_VARTYPE_LONGLONG;
      else
	flag |= JPA_VARTYPE_LONG;
      break;

    case JPACT_FLOAT:  flag |= JPA_VARTYPE_FLOAT;  break;
    case JPACT_DOUBLE: flag |= JPA_VARTYPE_DOUBLE; break;

    case JPACT_STRUCT: flag |= JPA_VARTYPE_STRUCT; break;
    case JPACT_UNION:  flag |= JPA_VARTYPE_UNION;  break;
    case JPACT_ENUM:   flag |= JPA_VARTYPE_ENUM;   break;

    case JPACT_AUTO:     case JPACT_REGISTER:
    case JPACT_STATIC:   case JPACT_EXTERN:
    case JPACT_CONST:    case JPACT_VOLATILE:
    case JPACT_SIGNED:   case JPACT_UNSIGNED:
      /*  ignore them  */
      break;

    case JPACT_TYPEDEFNAME:
      if((mode & 2) && strcmp(queue->TKFF_Token[spec->JPACN_IndexPP].TK_String,
			      "wchar_t") == 0) {
	flag |= JPA_VARTYPE_WCHAR;
      } else {
	jsparsexprtype_t newType;
	int ptrCnt2, flag2;

	if(_jsparsrestype(jpa, vTab, spec, queue, &newType) == -1)
	  EGOTO(error);

	/*  Recursive call, so we handle typedefs that are defined
	 *   through typedefs correctly
	 */
	if((flag2 = jsparsgetvartype2(jpa, vTab, &newType,
				      &ptrCnt2, mode)) == -1)
	  EGOTO(error);

	/*  This is a bit useless: actually, we should substitute flag
	 *   with flag2
	 */
	flag   |= flag2;
	ptrCnt += ptrCnt2;
      }
      break;

    default:
      jsparsseterrcode(_ERR_JSPARSGETVARTYPE, 0, queue, spec);
      EGOTO0(error);
    }
  }

  if(flag == 0)
    /*  Bad style - no declarator which defaults to int  */
    flag = JPA_VARTYPE_INT;

  if(flag == (JPA_VARTYPE_LONGLONG|JPA_VARTYPE_INT))
    flag = JPA_VARTYPE_LONGLONG; /*  "long long int" == "long long"  */
  else if(flag == (JPA_VARTYPE_LONG|JPA_VARTYPE_INT))
    flag = JPA_VARTYPE_LONG; /*  "long int" == "long"  */
  else if(flag == (JPA_VARTYPE_SHORT|JPA_VARTYPE_INT))
    flag = JPA_VARTYPE_SHORT; /*  "short int" == "short"  */
  else if(flag == (JPA_VARTYPE_LONG|JPA_VARTYPE_FLOAT))
    flag = JPA_VARTYPE_DOUBLE; /*  "long float" == "double"  */

  if(pPtrCnt)
    *pPtrCnt = ptrCnt;

  return flag;
error:
  return -1;
}

/*JS*********************************************************************
*   Check if expression type 'type' is defined through 'refType'.
*    mode bit 0: Only exact matches are tried, no typedefs are expanded
*                in 'type'.
*    Example:
*     typedef char MYCHAR, *MYSTRING;
*     typedef MYCHAR *MYCHARSTRING;
*     MYCHAR ch1, *str1;
*     MYSTRING str2;
*     MYCHARSTRING str3;
*     char ch2, *str4;
*
*     isvartype(ch1,   ch2)  -> TRUE
*     isvartype(ch2,   ch1)  -> FALSE
*     isvartype(str2,  str1) -> FALSE
*     isvartype(str2,  str4) -> TRUE
*     isvartype(str3,  str1) -> TRUE
*     isvartype(str3,  str2) -> FALSE
*     isvartype(str4,  str1) -> FALSE
*     isvartype(*str2, ch1)  -> TRUE
*
*************************************************************************/

int jsparsisvartype(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		    jsparsexprtype_t *type, jsparsexprtype_t *refType,
		    int mode)

/************************************************************************/
{
  int ret;
  jsparsexprtype_t locType;

  /*  Make a copy of the type we are going to modify  */
  locType = *type;
  locType.JSPE_Flags = 0; /*  to avoid double free  */
  do {
#ifdef TRACETYPE_JSCPP
    if(jstracelevel(TRACETYPE_JSCPP) >= 6) {
      FILE *tfp = jstracegetfp(TRACETYPE_JSCPP, 6);

      jstrace(TRACETYPE_JSCPP, 6,
	      "COMPARE DECL = %d:\n",
	      compareDeclarators(locType.JSPE_Declarator,
				 refType->JSPE_Declarator, mode));
      jsparsprctree(tfp, locType.JSPE_Queue, locType.JSPE_Declarator, 0, 0);
      jstrace(TRACETYPE_JSCPP, 6, "with\n");
      jsparsprctree(tfp, refType->JSPE_Queue, refType->JSPE_Declarator,
		    0, 0);
    }
#endif

    /*  Check for equality of declarators and specifiers  */
    if(compareDeclarators(locType.JSPE_Declarator, refType->JSPE_Declarator,
			  mode)) {
      if((ret = compareSpecifiers(jpa, vTab,
				  locType.JSPE_Specifier, locType.JSPE_Queue,
				  refType->JSPE_Specifier, refType->JSPE_Queue,
				  mode)) == -1)
	EGOTO(error);


#ifdef TRACETYPE_JSCPP
      if(jstracelevel(TRACETYPE_JSCPP) >= 6) {
	FILE *tfp = jstracegetfp(TRACETYPE_JSCPP, 6);

	jstrace(TRACETYPE_JSCPP, 6,
		"COMPARE SPEC = %d:\n", ret);
	jsparsprctree(tfp, locType.JSPE_Queue, locType.JSPE_Specifier,
		      0, 0);
	jstrace(TRACETYPE_JSCPP, 6, "with\n");
	jsparsprctree(tfp, refType->JSPE_Queue, refType->JSPE_Specifier,
		      0, 0);
      }
#endif

      if(ret)
	return TRUE;
    }

    if(mode & 1)
      /*  do not resolve typedefs  */
      break;

    if((ret = _jsparsres1type(jpa, vTab, &locType)) == -1)
      EGOTO(error);
  } while(ret != 0);

  return FALSE;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int compareDeclarators(jsparscnode_t *decl, jsparscnode_t *refDecl,
			      int mode)

/************************************************************************/
{
  if(decl && decl->JPACN_Type == JPACT_IDENT)
    decl = decl->JPACN_Next;

  if(refDecl && refDecl->JPACN_Type == JPACT_IDENT)
    refDecl = refDecl->JPACN_Next;

  while(decl && refDecl) {
    int dType, rType;

    /*  Skip qualifiers  */
    if((dType = decl->JPACN_Type) == JPACT_CONST || dType == JPACT_VOLATILE) {
      decl = decl->JPACN_Next;
      continue;
    }

    if((rType = refDecl->JPACN_Type) == JPACT_CONST ||
       rType == JPACT_VOLATILE) {
      refDecl = refDecl->JPACN_Next;
      continue;
    }

    /*  Pointer and array are equivalent  */
    if(dType == JPACT_VECTOR) dType = JPACT_POINTER;
    if(rType == JPACT_VECTOR) rType = JPACT_POINTER;

    if(rType != dType)
      return FALSE;

    /*  Advance both  */
    decl = decl->JPACN_Next;
    refDecl = refDecl->JPACN_Next;
  }

  /*  Must have the same length  */
  return decl == NULL && refDecl == NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static int compareSpecifiers(jsparsstream_t *jpa, jsparsvartab_t *vTab,
			     jsparscnode_t *spec, jctokfifo_t *queue,
			     jsparscnode_t *refSpec, jctokfifo_t *refQueue,
			     int mode)

/************************************************************************/
{
  jsparscnode_t *node, **refList;
  int n, count;

  /*  Convert nodes to a list  */
  n = 0;
  for(node = refSpec ; node ; node = node->JPACN_Next)
    ++n;

  if((refList = (jsparscnode_t **)
#ifdef CONFIG_USE_ALLOCA
      alloca((n + 1) * sizeof(*refList))
#else
      malloc((n + 1) * sizeof(*refList))
#endif
       ) == NULL)
    EGOTO0(error);

  n = 0;
  for(node = refSpec ; node ; node = node->JPACN_Next)
    /*  Ignore qualifier  */
    if(node->JPACN_Type != JPACT_CONST &&
       node->JPACN_Type != JPACT_VOLATILE)
      refList[n++] = node;

  /*  Now check if all specifiers appear  */
  for(count = n ; spec ; spec = spec->JPACN_Next) {
    unsigned int sType;
    int i;

    sType = spec->JPACN_Type;

    /*  Ignore qualifier  */
    if(sType == JPACT_CONST || sType == JPACT_VOLATILE)
      continue;

    for(i = 0 ; i < n ; ++i) {
      if((node = refList[i]) == NULL)
	continue;

      if(sType == node->JPACN_Type) {
	if(sType == JPACT_TYPEDEFNAME) {
	  if(tkStrcmp(queue->TKFF_Token[spec->JPACN_IndexPP].TK_String,
		      refQueue->TKFF_Token[node->JPACN_IndexPP].TK_String)
	     == 0)
	    break;
	} else if(sType == JPACT_STRUCT || sType == JPACT_UNION ||
		  sType == JPACT_ENUM) {
	  jsparscnode_t *ssub, *nsub;
	  const char *sname, *nname;

	  ssub = spec->JPACN_Sub;
	  nsub = node->JPACN_Sub;

	  sname = ssub && ssub->JPACN_Type == JPACT_IDENT ?
	    queue->TKFF_Token[ssub->JPACN_IndexPP].TK_String : NULL;
	  nname = nsub && nsub->JPACN_Type == JPACT_IDENT ?
	    refQueue->TKFF_Token[nsub->JPACN_IndexPP].TK_String : NULL;

	  /*  Either both structs have a name or both not  */
	  if(sname && nname) {
	    if(tkStrcmp(sname, nname) == 0)
	      break;
	  } else if(ssub && ssub->JPACN_Type == JPACT_BLOCK &&
		    nsub && nsub->JPACN_Type == JPACT_BLOCK) {
	    /*  Compare an unnamed struct  */
	    for(ssub = ssub->JPACN_Sub, nsub = nsub->JPACN_Sub ;
		ssub && nsub ;
		ssub = ssub->JPACN_Next, nsub = nsub->JPACN_Next) {
	      jsparscnode_t *sidn, *nidn;
	      const char *sIdent, *nIdent;
	      jsparsexprtype_t seType, neType;
	      int cmp;

	      sidn = ssub->JPACN_Sub;
	      sIdent = NULL;
	      if(sidn->JPACN_Type == JPACT_DECLARATOR) {
		sidn = sidn->JPACN_Sub;
		if(sidn->JPACN_Type == JPACT_IDENT)
		  sIdent = queue->TKFF_Token[sidn->JPACN_IndexPP].TK_String;
	      }

	      nidn = nsub->JPACN_Sub;
	      nIdent = NULL;
	      if(nidn->JPACN_Type == JPACT_DECLARATOR) {
		nidn = nidn->JPACN_Sub;
		if(nidn->JPACN_Type == JPACT_IDENT)
		  nIdent = refQueue->TKFF_Token[nidn->JPACN_IndexPP].TK_String;
	      }

	      if(sIdent == NULL || nIdent == NULL ||
		 tkStrcmp(sIdent, nIdent) != 0)
		break;

	      if(jsparsnode2type(&seType, ssub, queue) == NULL ||
		 jsparsnode2type(&neType, nsub, refQueue) == NULL)
		EGOTO(error);

	      if((cmp = jsparsisvartype(jpa, vTab, &seType, &neType, 1)) == -1)
		EGOTO(error);

	      if(!cmp)
		break;
	    }

	    if(ssub == NULL && nsub == NULL)
	      /*  All match and length is the same  */
	      break;
	  }
	} else
	  /*  yes, they match  */
	  break;

      }
    }

    if(i == n)
      break; /*  not found  */

    refList[i] = NULL;
    --count;
  }

#ifndef CONFIG_USE_ALLOCA
  free(refList);
#endif

  /*  Check that all specifiers were found and that the
   *   reference type doesn't have any specifiers left over
   */
  return spec == NULL && count == 0;
error:
  return -1;
}

/*JS*********************************************************************
*   Resolves *one* typedef. Returns 1 if something was done, 0 if no
*    typedef was found.
*************************************************************************/

int _jsparsres1type(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		    jsparsexprtype_t *pType)

/************************************************************************/
{
  jsparscnode_t *node;

  for(node = pType->JSPE_Specifier ; node ; node = node->JPACN_Next)
    if(node->JPACN_Type == JPACT_TYPEDEFNAME) {
      jsparsexprtype_t newType;

      if(_jsparsrestype(jpa, vTab, node, pType->JSPE_Queue, &newType) == -1)
	EGOTO(error);

      if(pType->JSPE_Declarator == NULL) {
	/*  New type might have a declarator with a dimension, so take
	 *   it from there.
	 */
	pType->JSPE_Flags = newType.JSPE_Flags;
	pType->JSPE_LongConst = newType.JSPE_LongConst;
      } else {
	/*  Set flags from new type, but preserve constant and possible
	 *   array size from old declarator.
	 */
	pType->JSPE_Flags = newType.JSPE_Flags |
	  (pType->JSPE_Flags & (JPA_VARTYPE_FARRAYSIZE | JPA_VARTYPE_FLONGCONST));
      }

      pType->JSPE_SubTable = newType.JSPE_SubTable;

      if((node = pType->JSPE_Declarator) != NULL
	 /*  Always copy it otherwise we destroy the original  */
	 /*&& pType->JSPE_Queue != newType.JSPE_Queue*/
	) {
	/*  We need to duplicate the declarator, so we can append
	 *   something to it.
	 */
	if((node = jsparscopyncnode(node, pType->JSPE_Queue,
				    newType.JSPE_Queue)) == NULL)
	  EGOTO(error);

	pType->JSPE_Declarator = node;
	pType->JSPE_Flags |= JPA_VARTYPE_FCOPYDECL;
      }

      if(newType.JSPE_Declarator != NULL) {
	if(node != NULL) {
	  jsparscnode_t *newDecl;

	  /*
	   *  Append the new declarator at the end
	   */
	  while(node->JPACN_Next)
	    node = node->JPACN_Next;

	  newDecl = newType.JSPE_Declarator;
	  if(!(newType.JSPE_Flags & JPA_VARTYPE_FCOPYDECL)) {
	    if((newDecl = jsparscopyncnode(newDecl, newType.JSPE_Queue,
					   newType.JSPE_Queue)) == NULL)
	      EGOTO(error);
	  }

	  node->JPACN_Next = newDecl;
	} else
	  pType->JSPE_Declarator = newType.JSPE_Declarator;
      }

      pType->JSPE_Queue = newType.JSPE_Queue;

      pType->JSPE_Specifier = newType.JSPE_Specifier;

      return 1; /*  we did something!  */
    }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jsparsrestype(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		   jsparscnode_t *node, jctokfifo_t *queue,
		   jsparsexprtype_t *pType)

/************************************************************************/
{
  const char *name;
  jsparsvardef_t *vDef;

  name = queue->TKFF_Token[node->JPACN_IndexPP].TK_String;

  if((vDef = jsparsgetvardef(jpa, vTab, name, JVDF_TTYPEDEF, 1)) == NULL)
    EGOTO(error);

  _jsparssetuptype(pType, vDef);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

void _jsparssetuptype(jsparsexprtype_t *pType, const jsparsvardef_t *vDef)

/************************************************************************/
{
  jsparscnode_t *node;

  if((node = vDef->JPAVD_Declarator) != NULL) {
    node = node->JPACN_Sub;
    if(node && node->JPACN_Type == JPACT_IDENT)
      node = node->JPACN_Next;
  }

  pType->JSPE_Flags      = 0;
  pType->JSPE_Declarator = node;
  pType->JSPE_Specifier  = vDef->JPAVD_Specifier;
  pType->JSPE_SubTable   = vDef->JPAVD_SubTable;
  pType->JSPE_Queue      = vDef->JPAVD_Queue;

  if((vDef->JPAVD_Flags & JVDF_TMASK) == JVDF_TENUMIDENT) {
    pType->JSPE_Flags |= JPA_VARTYPE_FLONGCONST;
    pType->JSPE_LongConst = vDef->JPAVD_LongValue;
  }

  if(vDef->JPAVD_Flags & JVDF_FARRAYSIZE) {
    pType->JSPE_Flags |= JPA_VARTYPE_FARRAYSIZE;
    pType->JSPE_LongConst = vDef->JPAVD_LongValue;
  }
}
