/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Subroutines for the grammar
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/
#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define jsparsopen(jpa, jps, filename, omode, mode)	\
    jsparsopen2(jpa, jps, NULL, filename, omode, mode)
/*  END-DEFINITIONS  */

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
Prototype int            jsparsinit(int mode);

Prototype jsparsstream_t *jsparsopen2(jsparsstream_t *jpa, jsppstream_t *jps,
				      jsparsenviron_t *jpaenv, const char *filename,
				      int omode, int mode);
static int               _jsparsenvapply(jsparsenviron_t *jpaenv,
					 jsparsstream_t *jpa);
Prototype int            jsparsclose(jsparsstream_t *jpa, int flag);

Prototype void           jsparssetprinterr(jsparsstream_t *jpa,
					   void (*printError)(void *user,
							      const char *message));
Prototype int            jsparsaddkeyword(jsparsstream_t *jpa,
					  const char *name,
					  const char *oldName);

Prototype int            jsparsget(jsparsstream_t *jpa, jsparscnode_t **pNode,
				   int mode);

Prototype int            jsparsflush(jsparsstream_t *jpa, jsparscnode_t *node,
				     FILE *fp, int mode);

IPrototype _jsparsident_t *_jsparsgetident(const char *name, _jsparsident_t *
					   hTab[JPA_KEYWORD_TABSIZE]);

IPrototype int           _jsparspushbtype(jsparsstream_t *jpa,
					  _jsparstypedef_t *def, int flags);
IPrototype int           _jsparspopbtype(jsparsstream_t *jpa);
IPrototype void          _jsparsfreebtype(_jsparstypedef_t *def);
IPrototype int           _jsparschkstrunion(jsparsstream_t *jpa,
					    jctokfifo_t *queue,
					    jsparscnode_t *node);


IPrototype jsparscnode_t *_jsparsmkunary(jsparscnode_t *node,
					 jsparscnode_t *sub,
					 jctokfifo_t *queue);

IPrototype _jsparsident_t *_jsparsputident(const char *name, int value,
					   _jsparsident_t *hTab[JPA_KEYWORD_TABSIZE]);

IPrototype jsparscnode_t  *_jsparsmkpragmaline(jctokfifo_t *queue,
					       int start, int end);

Prototype const dbgopts_t JSParseDebugOptions[];

/*********     GLOBAL VARIABLES                                 *********/
#ifdef CONFIG_JSCPPLIB_DEBUG
const dbgopts_t JSParseDebugOptions[] = {
    { "prtree",     32 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Print parser tree before processing" },
    { "prtreeq",     (32 + 64) << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Print parser tree before processing including queue" },
    { "prtok",      128 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Show tokens as processed by lexical analyzer" },
    { "vars",     256 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Show generated variable tables. Also, do not exit on parser error" },
    { "varsdecl",     512 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Show declarator and specifier for each variable" },
    { "varssub",     1024 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Show also sub variable tables of struct and unions" },
    { "yaccdbg",     2048 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Switch on yacc debugging if compiled with debugging" },

    { "flow",     4096 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Generate and show statement flow" },
    { "flowstmt",     8192 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Generate and show statement flow along with statements" },
    { "flownode",     (4096+8192) << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Generate and show statement flow along with statement nodes" },

    { "cmpdecl",     16384 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Show how variable types are compared" },
    { "boolass",     32768 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Report assignments where right hand side is a boolean expression" },
    { "varsize",     65536 << CONFIG_JSCPPLIB_DEBUG, NULL,
      "Report variable size determination" },

    { NULL,          0, NULL, NULL }
};
#endif

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsinit(int mode)

/************************************************************************/
{
#ifdef CONFIG_JSPARSE_PIGGYBACKINCLUDES
    /*  Make standard includes compiled into the program available  */
    if(epushopener(CONFIG_JSPARSE_PIGGYBACKINCLUDES, piggybackopener,
		   (void *)JSParseWrapIncludes) == -1)
	EGOTO(error);
#endif

    if(jsppinit(mode) == -1)
	EGOTO(error);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsstream_t *jsparsopen2(jsparsstream_t *jpa, jsppstream_t *jps,
			    jsparsenviron_t *jpaenv, const char *filename,
			    int omode, int mode)

/************************************************************************/
{
    int allocated, i;

    if(jpa != NULL) {
	allocated = FALSE;
    } else {
	if((jpa = (jsparsstream_t *)malloc(sizeof(*jpa))) == NULL)
	    return NULL;

	allocated = TRUE;
    }

    if(jpaenv != NULL) mode |= jpaenv->JPAE_PAMode;

    /*  Get underlying preprocessor stream  */
    if(jps != NULL) {
	omode = jps->JSPP_Mode;

	/*  Let the preprocessor understand our operators  */
	jps->JSPP_Ops = JSParserCOperators;
    } else {
	const char *incpath;

	if(mode & JPAM_FEXPANDUSERFILES) omode |= JPPM_EXPUSERFILES;

#ifdef CONFIG_JSPARSE_INCLUDEDIR
	incpath = CONFIG_JSPARSE_INCLUDEDIR;
#else
	incpath = NULL;
#endif

	/*  We opened it, so we have to close it as well  */
	mode |= JPAM_FCLOSESTREAM;
	if((jps = jsppopen2(NULL, jpaenv ? &(jpaenv->JPAE_PPEnviron) : NULL,
			    filename, incpath, JSParserCOperators, omode, 0)) == NULL)
	    EGOTO(error);
    }

    jpa->JSPA_WarnFlags = jpaenv != NULL ? jpaenv->JPAE_WarnOpts : 0;

    if(mode & JPAM_FPRINTORIG)
	jps->JSPP_Mode |= JPPM_PRESSPACE;

#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & (3 << (12 + CONFIG_JSCPPLIB_DEBUG)))
	mode |= JPAM_FSTMTFLOW;
#endif

    jpa->JSPA_Mode = mode;

    /* ***  Generate queues  *** */
    i = 0;
    if(mode & JPAM_FGETFPOS) i |= JCTOK_TYPE_FFILEPOS;
    if(mode & JPAM_FGETORIG) i |= JCTOK_TYPE_FORIGTOK;

    /*  FIXME: Global queue could go without above flags?!  */
    jctokcreate(&(jpa->JSPA_GlobalQueue), i);

    if((jpa->JSPA_Stream = jsqscreateppstream(jps, NULL, i)) == NULL)
	EGOTO(error);

    /*  Clear internal structure  */
    for(i = 0 ; i < JPA_KEYWORD_TABSIZE ; ++i)
	jpa->JSPA_KeyWordTable[i] = NULL;
    jpa->JSPA_TypeDefs = jpa->JSPA_DeferBlockDef =
	jpa->JSPA_SavedBlockDef = NULL;

    /*  We also need variable tables in case of generating statement
     *   flow (to have goto labels scanned for and to evaluate
     *   constant expression).
     */
    if(mode & (JPAM_FVARTABLES | JPAM_FSTMTFLOW)) {
	jpa->JSPA_MVarTables = 50;
	if((jpa->JSPA_VarTables = (jsparsvartab_t **)
	    malloc(jpa->JSPA_MVarTables * sizeof(*(jpa->JSPA_VarTables)))) == NULL)
	    EGOTO(error);

	/*  Preset global variable table  */
	jpa->JSPA_NVarTables = 1;
	if((jpa->JSPA_VarTables[0] = _jsparsgetvartab(NULL, 0, 0, 0)) == NULL)
	    EGOTO(error);
    } else {
	/*  We don't need this  */
	jpa->JSPA_VarTables = NULL;
	jpa->JSPA_MVarTables = jpa->JSPA_NVarTables = 0;
    }

    jpa->JSPA_BoundChkPragma = NULL;
    jpa->JSPA_NBoundChkPragma = jpa->JSPA_MBoundChkPragma = 0;

    jpa->JSPA_StmtFlow = NULL;

    jpa->JSPA_PrintError = NULL;
    jpa->JSPA_User     = NULL;

    for(i = 0 ; JSParserCKeyWords[i].KD_String ; i++)
	if(_jsparsputident(JSParserCKeyWords[i].KD_String,
			   JSParserCKeyWords[i].KD_Token,
			   jpa->JSPA_KeyWordTable) == NULL)
	    EGOTO(error);

    /*  Predefined symbols + macros  */
    if(_jsparsdopredefine(jpa) == -1)
	EGOTO(error);

    /*  Make it recognize our own pragmas  */
    jps->JSPP_DoPragma = _jsparsdopragma;
    jps->JSPP_User = jpa;

    /*  Generate global block  */
    if(_jsparspushbtype(jpa, NULL, BTD_CMPSTMT) == -1)
	EGOTO(error);

#ifdef XDEBUG
    /*  Shows entries in hash table with a chain longer than 1  */
    for(i = 0 ; i < JPA_KEYWORD_TABSIZE ; i++) {
	_jsparsident_t *ident;

	if((ident = jpa->JSPA_KeyWordTable[i]) && ident->JPAID_Next) {
	    printf("------ %d -------\n", i);

	    for( ; ident ; ident = ident->JPAID_Next)
		printf("%s: %d\n", ident->JPAID_Name, ident->JPAID_Value);
	}
    }
#endif

    if(jpaenv != NULL && _jsparsenvapply(jpaenv, jpa) == -1)
	EGOTO(error);

    return jpa;
 error:
    if(allocated)
	free(jpa);
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int _jsparsenvapply(jsparsenviron_t *jpaenv, jsparsstream_t *jpa)

/************************************************************************/
{
    jsppenviron_t *jppenv;
    int i;

    jppenv = &(jpaenv->JPAE_PPEnviron);

    if(jpa->JSPA_WarnFlags & JPAM_WARNPREPROC)
	jppenv->JSPE_Mode |= JPPM_PRWARNING;

    if(jppenv->JSPE_Mode & JPPM_LAZYSTYLE)
	jpa->JSPA_Mode |= JPAM_FALLOWREDECL;

    for(i = 0 ; i < jppenv->JSPE_NSTypes[JSPE_INDEX_TKEY] ; ++i) {
	const char *string;
	char *string2;

	string = jppenv->JSPE_STypes[JSPE_INDEX_TKEY][i].JSPES_String;
	if((string2 = strchr(string, '=')) == NULL)
	    EGOTO(error);

	*string2++ = '\0';
	if(jsparsaddkeyword(jpa, string, string2) == -1)
	    EGOTO(error);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Deinitializes the parser. If bit 0 of flag is set, do not free jps.
 *    If bit 1 is set, do not check for errors.
 *************************************************************************/

int jsparsclose(jsparsstream_t *jpa, int flag)

/************************************************************************/
{
    jsppstream_t *jps;
    int i, ret = 0;
    _jsparstypedef_t *tdef, *tnext;

    if(!(jpa->JSPA_Mode & JPAM_SEENNODE) &&
       (jpa->JSPA_WarnFlags & JPAM_WARNSTYLE))
	jsparsprinterr(jpa, NULL, NULL, "ANSI C forbids empty source file", 0);

    jps = jsqsgetppstream(jpa->JSPA_Stream);

    if(jps != NULL && jsppdefundefmac(jps, "__ignore__", NULL, -1, 0) == -1)
	ret = -1;

    if(jpa->JSPA_DeferBlockDef)
	_jsparsfreebtype(jpa->JSPA_DeferBlockDef);

    for(tdef = jpa->JSPA_TypeDefs ; tdef ; tdef = tnext) {
	tnext = tdef->JPATD_Next;
	_jsparsfreebtype(tdef);
    }

    for(i = 0 ; i < JPA_KEYWORD_TABSIZE ; ++i) {
	_jsparsident_t *id, *nid;

	for(id = jpa->JSPA_KeyWordTable[i] ; id ; id = nid) {
	    nid = id->JPAID_Next;
	    free(id);
	}
    }

    jctokfree(&(jpa->JSPA_GlobalQueue), 0);
    jsqsfree(jpa->JSPA_Stream);

    if(jpa->JSPA_VarTables) {
	for(i = 0 ; i < jpa->JSPA_NVarTables ; ++i)
	    jsparsfreevartab(jpa->JSPA_VarTables[i]);

	free(jpa->JSPA_VarTables);
    }

    if(jpa->JSPA_BoundChkPragma != NULL) {
	for(i = 0 ; i < jpa->JSPA_NBoundChkPragma ; ++i)
	    free((void *)jpa->JSPA_BoundChkPragma[i].JPABC_Name);
	free(jpa->JSPA_BoundChkPragma);
    }

    if(jpa->JSPA_StmtFlow)
	_jsparsfreeflow(jpa->JSPA_StmtFlow);

    if((jpa->JSPA_Mode & JPAM_FCLOSESTREAM) && jps != NULL &&
       jsppclose(jps, flag & 2) == -1)
	ret = -1;

    if(!(flag & 1))
	free(jpa);

    return ret;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsparssetprinterr(jsparsstream_t *jpa, void (*printError)(void *user,
							       const char *message))

/************************************************************************/
{
    jsppstream_t *jps;

    jpa->JSPA_PrintError = printError;

    if((jps = jsqsgetppstream(jpa->JSPA_Stream)) != NULL)
	jsppsetprinterr(jps, printError);
}

/*JS*********************************************************************
 *   Adds identifier name to the global keyword table having the same
 *    meaning as oldName.
 *************************************************************************/

int jsparsaddkeyword(jsparsstream_t *jpa, const char *name,
		     const char *oldName)

/************************************************************************/
{
    _jsparsident_t *pIdent;

    /*  Look for name and store it with same meaning  */
    if((pIdent = _jsparsgetident(oldName, jpa->JSPA_KeyWordTable)) == NULL)
	return -1;

    if(_jsparsputident(name, pIdent->JPAID_Value,
		       jpa->JSPA_KeyWordTable) == NULL)
	return -1;

    return 0;
}

/*JS*********************************************************************
 *   Gets the token value of name from the hash table. Returns NULL if not
 *    found.
 *************************************************************************/

_jsparsident_t *_jsparsgetident(const char *name, _jsparsident_t *
				hTab[JPA_KEYWORD_TABSIZE])

/************************************************************************/
{
    _jsparsident_t *ident;

    for(ident = hTab[jspptkhash(name, JPA_KEYWORD_TABSIZE)] ; ident ;
	ident = ident->JPAID_Next)
	if(tkStrcmp(ident->JPAID_Name, name) == 0)
	    return ident;

    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsget(jsparsstream_t *jpa, jsparscnode_t **pNode, int mode)

/************************************************************************/
{
    jctokfifo_t *queue;

    if(jpa->JSPA_StmtFlow) {
	_jsparsfreeflow(jpa->JSPA_StmtFlow);
	jpa->JSPA_StmtFlow = NULL;
    }

    if(_jsparsgetnode(jpa, jpa->JSPA_Stream, pNode, -1) == -1)
	EGOTO(error);

    queue = jsqsgetqueue(jpa->JSPA_Stream);

    if(*pNode != NULL)
	jpa->JSPA_Mode |= JPAM_SEENNODE;

#ifdef CONFIG_JSCPPLIB_DEBUG
    /*  Print tree before processing  */
    if(Debug & ((32 + 64) << CONFIG_JSCPPLIB_DEBUG)) {
	int pmode = (Debug >> (CONFIG_JSCPPLIB_DEBUG + 5)) & 3;

	if(pmode == 3)
	    jctokshowqueue(stdout, queue);

	if(*pNode) jsparsprctree(stdout, queue, *pNode, 0, pmode);
    }
#endif

#ifdef TRACETYPE_JSCPP
    /*  Trace tree before processing  */
    if(jstracelevel(TRACETYPE_JSCPP) >= 6) {
	FILE *tfp = jstracegetfp(TRACETYPE_JSCPP, 6);

	jctokshowqueue(tfp, queue);
	if(*pNode) jsparsprctree(tfp, queue, *pNode, 0, 3);
    }
#endif

    if(*pNode) {
	/*  Does our caller want to see variable definitions?  */
	if(jpa->JSPA_VarTables != NULL &&
	   _jsparsfindvariables(jpa, *pNode) == -1)
	    EGOTO(error);

	/*  Does our caller want to see a statement flow chart?  */
	if(((jpa->JSPA_Mode & JPAM_FSTMTFLOW)
#ifdef CONFIG_JSCPPLIB_DEBUG
	    || (Debug & (4096 << CONFIG_JSCPPLIB_DEBUG))
#endif
	    ) && (*pNode)->JPACN_Type == JPACT_FUNCDEF &&
	   (jpa->JSPA_StmtFlow = _jsparsscanflow(jpa, *pNode, queue)) == NULL)
	    EGOTO(error);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsflush(jsparsstream_t *jpa, jsparscnode_t *node, FILE *fp, int mode)

/************************************************************************/
{
    /*
     *  Tokens may be flushed after one
     *   translation unit has been processed.
     */
    if(fp != NULL) {
	jctokfifo_t *queue = jsqsgetqueue(jpa->JSPA_Stream);

	if(jsparsflushtokens(jpa, fp) == -1)
	    EGOTO(error);

	/*  Clear the queue  */
	jctokclear(queue);
    }

    if(node != NULL)
        jsparsfreecnode(node);

    if(jpa->JSPA_VarTables != NULL &&
       _jsparsclearvariables(jpa) == -1)
        EGOTO(error);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Push a new block on the block stack. If def is non-zero, reuse it,
 *    otherwise create empty block definition.
 *************************************************************************/

int _jsparspushbtype(jsparsstream_t *jpa, _jsparstypedef_t *def, int flags)

/************************************************************************/
{
    _jsparstypedef_t *newDef;

    /*  Anything to delete from last time?  */
    if(jpa->JSPA_DeferBlockDef) {
	_jsparsfreebtype(jpa->JSPA_DeferBlockDef);

	jpa->JSPA_DeferBlockDef = NULL;
    }

    if(def) {
	newDef = def;
    } else {
	if((newDef = (_jsparstypedef_t *)calloc(1, sizeof(*newDef))) == NULL)
	    return -1;
    }

    newDef->JPATD_Flags = flags;
    newDef->JPATD_Next = jpa->JSPA_TypeDefs;
    jpa->JSPA_TypeDefs = newDef;

    return 0;
}

/*JS*********************************************************************
 *   Pops a block from the block stack and stores it in jpa->JSPA_DeferBlockDef.
 *************************************************************************/

int _jsparspopbtype(jsparsstream_t *jpa)

/************************************************************************/
{
    _jsparstypedef_t *oldDef;

    if((oldDef = jpa->JSPA_TypeDefs) == NULL)
	return -1;

    jpa->JSPA_TypeDefs = oldDef->JPATD_Next;

    /*  Anything to delete from last time?  */
    if(jpa->JSPA_DeferBlockDef)
	_jsparsfreebtype(jpa->JSPA_DeferBlockDef);

    /*  Save it in static variable for later deletion  */
    jpa->JSPA_DeferBlockDef = oldDef;

    return 0;
}

/*JS*********************************************************************
 *   Free a block definition.
 *************************************************************************/

void _jsparsfreebtype(_jsparstypedef_t *def)

/************************************************************************/
{
    int i;

    for(i = 0 ; i < JPA_KEYWORD_TABSIZE ; i++) {
	_jsparsident_t *pIdent, *next;

	for(pIdent = def->JPATD_HashTable[i] ; pIdent ; pIdent = next) {
	    next = pIdent->JPAID_Next;
	    free(pIdent);
	}
    }
    free(def);
}

/*JS*********************************************************************
 *
 *************************************************************************/

int _jsparschkstrunion(jsparsstream_t *jpa, jctokfifo_t *queue,
		       jsparscnode_t *node)

/************************************************************************/
{
    /*  Might be struct, union or enum specification, otherwise invalid  */
    while(node) {
	if(node->JPACN_Type == JPACT_STRUCT ||
	   node->JPACN_Type == JPACT_UNION ||
	   node->JPACN_Type == JPACT_ENUM)
	    return 0;

	node = node->JPACN_Next;
    }

    jsparsprinterr(jpa, queue, node, "Illegal declaration", 1);
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparscnode_t *_jsparsmkunary(jsparscnode_t *node, jsparscnode_t *sub,
			      jctokfifo_t *queue)

/************************************************************************/
{
    if(node->JPACN_Type != JPACT_UNARYEXPR) {
	/*  Wrap it  */
	if((node = jsparsmkcnode(NULL, JPACT_UNARYEXPR, node,
				 node->JPACN_Start, node->JPACN_End, queue)) == NULL)
	    return NULL;
    }

    node->JPACN_Sub = _jsparslinktail(node->JPACN_Sub, sub);

    return node;
}

/* ERROR-DEFINITIONS from putIdentifier label _ERR_JSPARSPUTIDENT
   Keyword redefined
*/

/*JS*********************************************************************
 *   Puts an identifier in a hash table.
 *************************************************************************/

_jsparsident_t *_jsparsputident(const char *name, int value,
				_jsparsident_t *hTab[JPA_KEYWORD_TABSIZE])

/************************************************************************/
{
    _jsparsident_t *pIdent, *newIdent;
    unsigned int hash;

    hash = jspptkhash(name, JPA_KEYWORD_TABSIZE);

    for(pIdent = hTab[hash] ; pIdent ; pIdent = pIdent->JPAID_Next)
	/*  Redefinition of identifier in same block?  */
	if(tkStrcmp(pIdent->JPAID_Name, name) == 0) {
	    /*  Accept redefinition of identifiers, since we may have
	     *   multiple declarations of a function or variable.
	     */
	    return pIdent;
	}

    if((newIdent = (_jsparsident_t *)malloc(sizeof(*newIdent) +
					    strlen(name))) == NULL)
	return NULL;

    /*  Preset structure  */
    strcpy(newIdent->JPAID_Name, name);
    newIdent->JPAID_Value = value;

    newIdent->JPAID_File = NULL;
    newIdent->JPAID_Line = 0;

    /*  Hang in list  */
    newIdent->JPAID_Next = hTab[hash];
    hTab[hash] = newIdent;

    return newIdent;
}

/*JS*********************************************************************
 *   A pragma line has its own syntax rules, so all we do is collect the
 *    tokens and generate a JPACT_PRAGMALINE/JPACT_PRAGMALINE2 node.
 *    The second one is used if we have the syntax "#pragma XXXX ...."
 *    where "XXXX" is one of the JSPPPragmaIntroducer strings. In that
 *    case we *know* that the line was meant for us and that we should
 *    bark in case we find a syntax error.
 *************************************************************************/

jsparscnode_t *_jsparsmkpragmaline(jctokfifo_t *queue, int start, int end)

/************************************************************************/
{
    jsparscnode_t *node;
    int step, curr, type, start2, end2;

    step = jctokgetstep(queue);

    /*  Find first useful token and change type accordingly  */
    type = JPACT_PRAGMALINE;
    for(curr = 1, start2 = start ; start2 < end ; start2 += step) {
	jsctoken_t token = queue->TKFF_Token[start2];

	if(token.TK_Value == PTK_EMPTY || token.TK_Value == PTK_STARTSPC ||
	   (CTK_BLANK_START <= token.TK_Value && token.TK_Value <= CTK_BLANK_END))
	    continue;

	if(curr == 1) { /* "#" */
	    if(token.TK_Value != '#')
		break;

	    ++curr;
	} else if(curr == 2) { /* "pragma" */
	    if(token.TK_String == NULL || strcmp(token.TK_String, "pragma") != 0)
		break;

	    ++curr;
	} else if(curr == 3) { /*  look for special pragma introducer  */
	    int i;

	    if(token.TK_String == NULL)
		break;

	    for(i = 0 ; JSPPPragmaIntroducer[i] ; ++i)
		if(tkStrcmp(token.TK_String, JSPPPragmaIntroducer[i]) == 0)
		    break;

	    /*  None matched, so stop  */
	    if(JSPPPragmaIntroducer[i] == NULL)
		break;

	    type = JPACT_PRAGMALINE2;
	    ++curr;
	} else if(curr == 4) /*  first useful token  */
	    break;
    }

    /*  Delete trailing useless tokens  */
    for(end2 = end ; start2 < end2 ; end2 -= step) {
	int tokValue = queue->TKFF_Token[end2].TK_Value;

	if(tokValue != PTK_EMPTY && tokValue != PTK_ENDSPC &&
	   !(CTK_BLANK_START <= tokValue && tokValue <= CTK_BLANK_END))
	    break;
    }

    if((node = jsparsmkcnode(NULL, JPACT_PRAGMACONT, NULL,
			     start2, end2, queue)) == NULL)
	EGOTO(error);

    return jsparsmkcnode(NULL, type, node, start, end, queue);
 error:
    return NULL;
}
