/*JS*********************************************************************
 *
 *    Program : JSPARSE
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : ANSI-C parser.
 *    Part    : Routines to scan for variables
 *
 *              Copyright (C) 2003-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/

/*#define NDEBUG*/
#include <assert.h>

#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES						*********/
#ifndef JVDF_TMASK
/*  START-DEFINITIONS */
typedef struct jsparsvartab_t jsparsvartab_t;
typedef struct jsparsvardef_t jsparsvardef_t;
/*  END-DEFINITIONS c */
#endif

#ifndef JVDF_TMASK
/*  START-DEFINITIONS 2 */
/* ***  Variable definitions  *** */
struct jsparsvardef_t {
    jsparsvardef_t       *JPAVD_Next; /*  To link within hash table  */

    const char           *JPAVD_Name;
    unsigned long         JPAVD_Flags;
    jsparscnode_t        *JPAVD_Declarator;
    jsparscnode_t        *JPAVD_Specifier;

    /*  queue where tokens of definition are stored  */
    jctokfifo_t          *JPAVD_Queue;
    jsparsvartab_t       *JPAVD_SubTable; /* In case of struct/union
					   *  or a typedef with
					   *  anonymous struct
					   */

    unsigned long         JPAVD_LongValue; /*  if this is an enumeration,
					    *   the value.
					    *  If it is an automatically
					    *   sized array (flag
					    */

    /*  To let the user store some other relevant information  */
    union {
	void               *jpavd_UserPtr;
	long               *jpavd_UserLong;
    } u;
#define JPAVD_UserPtr    u.jpavd_UserPtr
#define JPAVD_UserLong   u.jpavd_UserLong
};

/*  For the flags field  */
#define JVDF_TMASK       0x1fUL  /*  Mask for type field  */

/*  Submask to distinguish different name spaces:
 *   Reject multiple definition within one namespace
 */
#define JVDF_TGMASK      0x18UL

/*  Name space 0 for struct/union/enums  */
#define JVDF_TSTRUCT     ((0UL << 3) | 1UL)
#define JVDF_TUNION      ((0UL << 3) | 2UL)
#define JVDF_TENUM       ((0UL << 3) | 3UL)

/*  Name space 1 for variable type identifiers  */
#define JVDF_TVARIABLE   ((1UL << 3) | 1UL)
#define JVDF_TTYPEDEF    ((1UL << 3) | 2UL)
#define JVDF_TENUMIDENT  ((1UL << 3) | 3UL)
#define JVDF_TFUNCTION   ((1UL << 3) | 4UL)

/*  Name space 2 for statement labels  */
#define JVDF_TLABEL      ((2UL << 3) | 1UL)

#define JVDF_FDECLARED   (1UL <<  8) /*  declared  */
#define JVDF_FDEFINED    (1UL <<  9) /*  defined  */
#define JVDF_FFREESUBTAB (1UL << 10) /*  Free subtable when freeing resources  */
#define JVDF_FFREENODES  (1UL << 11) /*  free declarator/specifier  */

#define JVDF_FGLOBAL     (1UL << 12) /*  variable in global variable table  */

/*  Some information flags about this variable  */
#define JVDF_IEXTERN     (1UL << 13) /*  external variable  */
#define JVDF_ISTATIC     (1UL << 14) /*  static variable  */
#define JVDF_IFUNCARG    (1UL << 15) /*  function argument  */
#define JVDF_ICONST      (1UL << 16) /*  constant  */

#define JVDF_FREAD       (1UL << 17) /*  has been read  */
#define JVDF_FWRITE      (1UL << 18) /*  has been written to  */
#define JVDF_FADDR       (1UL << 19) /*  address has been taken of  */

#define JVDF_FARRAYSIZE  (1UL << 20) /*  JPAVD_LongValue contains
				      *   automatically determined
				      *   array size
				      */

#define JVDF_FUSER       (1UL << 24) /*  user flags  */

/* **  Size of variable hash table  ** */
#ifndef JPA_VARIABLE_TABSIZE
# define JPA_VARIABLE_TABSIZE   251  /*  prime number  */
#endif

    struct jsparsvartab_t {
	jsparsvartab_t *JPAVT_Next;

	long            JPAVT_Flags;

	/*  Indexes into a token queue for which range this
	 *   variable table is valid
	 */
	int             JPAVT_StartIndex, JPAVT_EndIndex;

	jsparsvardef_t *JPAVT_Vars[JPA_VARIABLE_TABSIZE];
    };

#define JVDT_FSTRUCTUNION                 (1UL << 0)

#define jsparsgetglobalvartab(jpa)        ((jpa)->JSPA_VarTables[0])
#define jsparsisglobalvartab(jpa, vTab)   ((vTab) == (jpa)->JSPA_VarTables[0])
/*  END-DEFINITIONS  */
#endif

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
IPrototype jsparsvartab_t *_jsparsgetvartab(jsparsvartab_t *next, long flags,
					    int start, int end);
Prototype int             jsparsinsertvtab(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab);
static int                jsparslocatevartab(jsparsstream_t *jpa, int ind);

Prototype jsparsvardef_t *jsparsfindvariable(jsparsstream_t *jpa,
					     jctokfifo_t *queue,
					     jsparscnode_t *node,
					     int type, int mode);

Prototype jsparsvartab_t *jsparsfindvartab(jsparsstream_t *jpa,
					   jsparscnode_t *node);
Prototype jsparsvartab_t *jsparsfindvartab2(jsparsstream_t *jpa, int start,
					    int end);

Prototype jsparsvardef_t *jsparsdefinevar2(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab, int type,
					   jctokfifo_t *queue,
					   const char *name,
					   jsparscnode_t *declarator,
					   jsparscnode_t *specifier,
					   int mode);
Prototype jsparsvardef_t *jsparsgetvardef(jsparsstream_t *jpa,
					  jsparsvartab_t *vTab,
					  const char *name, int type,
					  int mode);
Prototype jsparsvardef_t *jsparsgetvardef2(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab,
					   const char *name, int type,
					   jsparsvartab_t **pTab,
					   int mode);
static int               findSubTable(jsparsstream_t *jpa,
				      jsparsvartab_t *vTab,
				      jsparsvardef_t *vDef);

Prototype int            jsparsdelvardef(jsparsstream_t *jpa,
					 jsparsvartab_t *vTab,
					 const char *name, int type,
					 int mode);

Prototype void           jsparsfreevartab(jsparsvartab_t *vTab);

static jsparsvardef_t   *jsparsfreevardef(jsparsvardef_t *vDef,
					  int mode);

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsvartab_t *_jsparsgetvartab(jsparsvartab_t *next, long flags,
				 int start, int end)

/************************************************************************/
{
    jsparsvartab_t *vTab;
    int i;

    if((vTab = (jsparsvartab_t *)malloc(sizeof(*vTab))) == NULL)
	EGOTO0(error);

    vTab->JPAVT_Next = next;

    vTab->JPAVT_Flags      = flags;
    vTab->JPAVT_StartIndex = start;
    vTab->JPAVT_EndIndex   = end;

    for(i = 0 ; i < JPA_VARIABLE_TABSIZE ; ++i)
	vTab->JPAVT_Vars[i] = NULL;

    return vTab;
 error:
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsinsertvtab(jsparsstream_t *jpa, jsparsvartab_t *vTab)

/************************************************************************/
{
    int ind;

    if(jpa->JSPA_NVarTables >= jpa->JSPA_MVarTables) {
	jpa->JSPA_MVarTables += 50;

	if((jpa->JSPA_VarTables = (jsparsvartab_t **)
	    realloc(jpa->JSPA_VarTables, jpa->JSPA_MVarTables *
		    sizeof(*(jpa->JSPA_VarTables)))) == NULL)
	    EGOTO(error);
    }

    ind = jsparslocatevartab(jpa, vTab->JPAVT_StartIndex);

    assert(ind > 0);

    if(ind < jpa->JSPA_NVarTables)
	memmove(&(jpa->JSPA_VarTables[ind + 1]),
		&(jpa->JSPA_VarTables[ind]),
		(jpa->JSPA_NVarTables - ind) *
		sizeof(*(jpa->JSPA_VarTables)));

    jpa->JSPA_VarTables[ind] = vTab;

    jpa->JSPA_NVarTables++;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int jsparslocatevartab(jsparsstream_t *jpa, int ind)

/************************************************************************/
{
    jsparsvartab_t **tables;
    int lower, upper;

    tables = jpa->JSPA_VarTables;

    lower = 0;
    upper = jpa->JSPA_NVarTables;
    while(lower < upper) {
	int m = (lower + upper) >> 1;

	if(tables[m]->JPAVT_StartIndex <= ind)
	    lower = m + 1;
	else
	    upper = m;
    }

    return lower;
}

/* ERROR-DEFINITIONS from jsparsfindvariable label _ERR_JSPARSFINDVARIABLE
   Variable not found
*/

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsvardef_t *jsparsfindvariable(jsparsstream_t *jpa, jctokfifo_t *queue,
				   jsparscnode_t *node, int type, int mode)

/************************************************************************/
{
    jsparsvartab_t *vTab;
    jsparsvardef_t *vDef;
    const char *name;
    int ind;

    if((vTab = jsparsfindvartab(jpa, node)) == NULL)
	EGOTO(error);

    ind = node->JPACN_IndexPP;

    name = queue->TKFF_Token[ind].TK_String;

    if((vDef = jsparsgetvardef(jpa, vTab, name, type, mode)) == NULL) {
	jsparsseterrcode2(_ERR_JSPARSFINDVARIABLE, 0, name, queue, node);
	EGOTO(error);
    }

    return vDef;
 error:
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsvartab_t *jsparsfindvartab(jsparsstream_t *jpa, jsparscnode_t *node)

/************************************************************************/
{
    return jsparsfindvartab2(jpa, node->JPACN_Start, node->JPACN_End);
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsvartab_t *jsparsfindvartab2(jsparsstream_t *jpa, int start, int end)

/************************************************************************/
{
    int ind;

    ind = jsparslocatevartab(jpa, start);
    --ind;

    while(ind > 0 &&
	  jpa->JSPA_VarTables[ind]->JPAVT_EndIndex < end)
	--ind;

    return jpa->JSPA_VarTables[ind];
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsvardef_t *jsparsdefinevar2(jsparsstream_t *jpa,
				 jsparsvartab_t *vTab, int type,
				 jctokfifo_t *queue,
				 const char *name,
				 jsparscnode_t *declarator,
				 jsparscnode_t *specifier,
				 int mode)

/************************************************************************/
{
    jsparsvardef_t *vDef;
    jsparscnode_t *node;

    if((vDef = jsparsgetvardef(jpa, vTab, name, type, 0)) == NULL)
	EGOTO(error);

    /*  Free previous definitions if they are on global queue  */
    if(vDef->JPAVD_Flags & JVDF_FFREENODES) {
	if(vDef->JPAVD_Declarator)
	    jsparsfreecnode(vDef->JPAVD_Declarator);

	for(node = vDef->JPAVD_Specifier ; node ; )
	    node = jsparsfreecnode(node);
    }

    if(jsparsisglobalvartab(jpa, vTab))
	vDef->JPAVD_Flags |= JVDF_FGLOBAL;

    /*  Check for global variables if we have to copy it  */
    if((mode & 2) || jsparsisglobalvartab(jpa, vTab)) {
	jctokfifo_t *dqueue;

	/*  Definition should go to global variable table, so
	 *   make it permanent
	 */
	dqueue = &(jpa->JSPA_GlobalQueue);

	if((declarator != NULL &&
	    (declarator = jsparscopycnode(declarator, queue, dqueue)) == NULL) ||
	   (specifier != NULL &&
	    (specifier = jsparscopyncnode(specifier, queue, dqueue)) == NULL))
	    EGOTO(error);

	vDef->JPAVD_Flags |= JVDF_FFREENODES;

	queue = dqueue;
    }

    /*  Set new definition  */
    vDef->JPAVD_Queue = queue;
    vDef->JPAVD_Declarator = declarator;
    vDef->JPAVD_Specifier = specifier;

    /*  Check for specific flags  */
    for(node = specifier ; node ; node = node->JPACN_Next)
	if(node->JPACN_Type == JPACT_STATIC)
	    vDef->JPAVD_Flags |= JVDF_ISTATIC;
	else if(node->JPACN_Type == JPACT_EXTERN)
	    vDef->JPAVD_Flags |= JVDF_IEXTERN;
	else if(node->JPACN_Type == JPACT_CONST)
	    vDef->JPAVD_Flags |= JVDF_ICONST;

    return vDef;
 error:
    return NULL;
}

/* ERROR-DEFINITIONS from jsparsgetvardef label _ERR_JSPARSGETVARDEF
   Variable defined twice
   Struct/union not found
   Typedef not found
   Enumeration not found
   Variable not found
*/

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsvardef_t *jsparsgetvardef(jsparsstream_t *jpa, jsparsvartab_t *vTab,
				const char *name, int type, int mode)

/************************************************************************/
{
    return jsparsgetvardef2(jpa, vTab, name, type, NULL, mode);
}

/*JS*********************************************************************
 *   Gets the definition with the required type. If VDF_FDEFINED is set in
 *    type, create the variable in the current block. If the variable type
 *    is JVDF_TENUMIDENT, it is defined in the block that is not type
 *    JVDT_FSTRUCTUNION.
 *   mode bit 0: If set, fix JPAVD_SubTable field before returning.
 *   mode bit 1: only find variables defined in global block.
 *   mode bit 2: only find variables defined in local blocks.
 *   mode bit 3: only look in current vTab, don't go to next higher level
 *************************************************************************/

jsparsvardef_t *jsparsgetvardef2(jsparsstream_t *jpa, jsparsvartab_t *vTab,
				 const char *name, int type,
				 jsparsvartab_t **pTab, int mode)

/************************************************************************/
{
    jsparsvartab_t *vTab2;
    jsparsvardef_t *vDef;
    unsigned int hashValue;
    int createFlag;
    unsigned int type2;

    createFlag = (type & (JVDF_FDEFINED | JVDF_FDECLARED));

    /*  Hash value for this variable  */
    hashValue = jspptkhash(name, JPA_VARIABLE_TABSIZE);

    type2 = type & JVDF_TMASK;
    for(vTab2 = vTab ; vTab2 ; vTab2 = vTab2->JPAVT_Next) {
	if((mode & 6) == 2 && vTab2->JPAVT_Next != NULL)
	    continue; /*  skip local blocks  */

	if((mode & 6) == 4 && vTab2->JPAVT_Next == NULL)
	    break; /*  skip global block  */

	for(vDef = vTab2->JPAVT_Vars[hashValue] ; vDef ; vDef = vDef->JPAVD_Next)
	    if((vDef->JPAVD_Flags & JVDF_TMASK) == type2 &&
	       tkStrcmp(vDef->JPAVD_Name, name) == 0) {
#ifdef XDEBUG
		printf("getVarDef: \"%s\" found: \"%s\", flags %#x\n", name,
		       vDef->JPAVD_Name, vDef->JPAVD_Flags);
#endif

		if(vDef->JPAVD_SubTable == NULL && (mode & 1))
		    findSubTable(jpa, vTab, vDef);

		/*  There must be only one definition  */
		if((createFlag & JVDF_FDEFINED) &&
		   (vDef->JPAVD_Flags & JVDF_FDEFINED)) {
		    /* ***  We have a redefinition  *** */
		    if((jpa->JSPA_Mode & JPAM_FALLOWREDECL) &&
		       (type2 == JVDF_TSTRUCT || type2 == JVDF_TUNION ||
			type2 == JVDF_TTYPEDEF || type2 == JVDF_TENUM)) {
			/*  Make redeclaration of typedef etc only a warning  */
			if(jpa->JSPA_WarnFlags & JPAM_WARNVARS)
			    jsparsprinterr(jpa, vDef->JPAVD_Queue, vDef->JPAVD_Declarator,
					   "Redefinition of %s %s", 0,
					   type2 == JVDF_TSTRUCT ? "struct" :
					   type2 == JVDF_TUNION  ? "union" :
					   type2 == JVDF_TTYPEDEF  ? "typedef" :
					   "enumeration", name);
		    } else if(!(jsparsisglobalvartab(jpa, vTab2) &&
				type2 == JVDF_TVARIABLE)) {
			/*  Redefinition of variables is allowed in global space, but
			 *   not in local space
			 */
			jsparsseterrcode2(_ERR_JSPARSGETVARDEF, 0, name,
					  vDef->JPAVD_Queue, vDef->JPAVD_Declarator);
			EGOTO(error);
		    }
		}

		if(pTab) *pTab = vTab2;

		return vDef;
	    }

	/*  Stop searching if creation / declaration forced  */
	if(createFlag || (mode & 8))
	    break;
    }

    /*  Is creation required?  */
    if(createFlag == 0) {
	int errNr = type2 == JVDF_TSTRUCT || type2 == JVDF_TUNION ? 1 :
	    type2 == JVDF_TTYPEDEF  ? 2 : type2 == JVDF_TENUM ? 3 : 4;
	jsparsseterrcode2(_ERR_JSPARSGETVARDEF, errNr, name, NULL, NULL);
	return NULL;
    }

    /* **  Generate new definition in this block  ** */
    if((vDef = (jsparsvardef_t *)malloc(sizeof(*vDef) +
					strlen(name) + 1)) == NULL)
	EGOTO(error);

    vDef->JPAVD_Name = (const char *)&vDef[1];
    strcpy((char *)vDef->JPAVD_Name, name);

    vDef->JPAVD_Flags = type;

    /*  Preset fields  */
    vDef->JPAVD_Declarator = vDef->JPAVD_Specifier = NULL;
    vDef->JPAVD_Queue = NULL;
    vDef->JPAVD_SubTable = NULL;

    vDef->JPAVD_UserPtr = NULL;

    if(type2 == JVDF_TENUMIDENT)
	/*  Enumeration constants are defined within surrounding block  */
	while(vTab && (vTab->JPAVT_Flags & JVDT_FSTRUCTUNION))
	    vTab = vTab->JPAVT_Next;

    vDef->JPAVD_Next = vTab->JPAVT_Vars[hashValue];
    vTab->JPAVT_Vars[hashValue] = vDef;

#ifdef XDEBUG
    printf("getVarDef: \"%s\" generated\n", name);
#endif

    if(pTab) *pTab = vTab;

    return vDef;
 error:
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int findSubTable(jsparsstream_t *jpa, jsparsvartab_t *vTab,
			jsparsvardef_t *vDef)

/************************************************************************/
{
    jsparscnode_t *node;

    /*  Try to fix this field by looking for the struct/union definition  */
    for(node = vDef->JPAVD_Specifier ; node ; node = node->JPACN_Next) {
	if(node->JPACN_Type == JPACT_STRUCT || node->JPACN_Type == JPACT_UNION) {
	    int type;

	    type = node->JPACN_Type;
	    node = node->JPACN_Sub;

	    if(node->JPACN_Type == JPACT_IDENT) {
		const char *name;
		jsparsvardef_t *subDef;

		name = vDef->JPAVD_Queue->TKFF_Token[node->JPACN_IndexPP].TK_String;
		if((subDef = jsparsgetvardef(jpa, vTab, name, type == JPACT_STRUCT ?
					     JVDF_TSTRUCT : JVDF_TUNION, 0)) != NULL) {
		    vDef->JPAVD_SubTable = subDef->JPAVD_SubTable;
		}
	    }
	    break;
	}
    }
    return 0;
}


/*JS*********************************************************************
 *   mode bit 1: only find variables defined in global block.
 *   mode bit 2: only find variables defined in local blocks.
 *   mode bit 3: only look in current vTab, don't go to next higher level
 *************************************************************************/

int jsparsdelvardef(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		    const char *name, int type, int mode)

/************************************************************************/
{
    unsigned int hashValue;

    hashValue = jspptkhash(name, JPA_VARIABLE_TABSIZE);

    type &= JVDF_TMASK;
    for( ; vTab ; vTab = vTab->JPAVT_Next) {
	jsparsvardef_t *vDef, *prev;

	if((mode & 6) == 2 && vTab->JPAVT_Next != NULL)
	    continue; /*  skip local blocks  */

	if((mode & 6) == 4 && vTab->JPAVT_Next == NULL)
	    break; /*  skip global block  */

	for(prev = NULL, vDef = vTab->JPAVT_Vars[hashValue] ; vDef ;
	    prev = vDef, vDef = vDef->JPAVD_Next)
	    if((int)(vDef->JPAVD_Flags & JVDF_TMASK) == type &&
	       tkStrcmp(vDef->JPAVD_Name, name) == 0) {
		if(prev == NULL)
		    vTab->JPAVT_Vars[hashValue] = jsparsfreevardef(vDef, 0);
		else
		    prev->JPAVD_Next = jsparsfreevardef(vDef, 0);

		/*  Report success  */
		return 1;
	    }

	if(mode & 8)
	    break; /*  Only look in current block  */
    }

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsparsfreevartab(jsparsvartab_t *vTab)

/************************************************************************/
{
    int i;

    for(i = 0 ; i < JPA_VARIABLE_TABSIZE; i++) {
	jsparsvardef_t *vDef;

	for(vDef = vTab->JPAVT_Vars[i] ; vDef ; )
	    vDef = jsparsfreevardef(vDef, 0);
    }

    free(vTab);
}

/*JS*********************************************************************
 *
 *************************************************************************/

static jsparsvardef_t *jsparsfreevardef(jsparsvardef_t *vDef, int mode)

/************************************************************************/
{
    jsparsvardef_t *next;

    next = vDef->JPAVD_Next;

    if(vDef->JPAVD_SubTable && (vDef->JPAVD_Flags & JVDF_FFREESUBTAB))
	jsparsfreevartab(vDef->JPAVD_SubTable);

    if(vDef->JPAVD_Flags & JVDF_FFREENODES) {
	jsparscnode_t *node;

	if(vDef->JPAVD_Declarator)
	    jsparsfreecnode(vDef->JPAVD_Declarator);

	for(node = vDef->JPAVD_Specifier ; node ; )
	    node = jsparsfreecnode(node);
    }

    free(vDef);

    return next;
}
