/*JS*********************************************************************
*
*    Program : JSPARSE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : ANSI-C parser.
*    Part    : Routines to scan for variables
*
*              Copyright (C) 2003-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES 					*********/

/*#define NDEBUG*/
#include <assert.h>

#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES						*********/
typedef struct {
  long Start, End;
} usageinterval_t;

typedef struct {
  int Mode;
  long Flags;
} auxprocvar_t;

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
IPrototype int           _jsparsfindvariables(jsparsstream_t *jpa,
					      jsparscnode_t *node);
IPrototype int           _jsparsscanvars(jsparsstream_t *jpa,
					 jsparscnode_t *node,
					 jctokfifo_t *queue,
					 jsparsvartab_t *vTab, int mode);
static jsparsvardef_t   *jsparsdefinevar(jsparsstream_t *jpa,
					 jsparsvartab_t *vTab, int type,
					 jctokfifo_t *queue, int ind,
					 jsparscnode_t *declarator,
					 jsparscnode_t *specifier,
					 int mode);
static int               doDefinitions(jsparsstream_t *jpa,
				       jsparscnode_t *node, jctokfifo_t *queue,
				       jsparsvartab_t *vTab, int mode);
static int               scanEnumeration(jsparsstream_t *jpa,
					 jsparscnode_t *node,
					 jctokfifo_t *queue,
					 jsparsvartab_t *vTab);
static int               checkArrayBounds(jsparsstream_t *jpa,
					  jsparscnode_t *expr,
					  jctokfifo_t *queue,
					  jsparsvartab_t *vTab);
static int               checkArrayBoundary(jsparsstream_t *jpa, jsparscnode_t *arrArg,
					    jsparscnode_t *indArg, jctokfifo_t *queue,
					    jsparsvartab_t *vTab,
					    const char *name, int level,
					    _jsparsboundchk_t *pboundchk);

IPrototype int           _jsparsclearvariables(jsparsstream_t *jpa);
Prototype void           jsparsclearvarflags(jsparsvartab_t *vTab, long flags);

Prototype int            jsparsprocessvars(jsparsstream_t *jpa,
					   jsparscnode_t *node,
					   jctokfifo_t *queue,
					   int (*doVariable)(jsparsstream_t *,
							     jsparscnode_t *,
							     jctokfifo_t *,
							     jsparsvartab_t *,
							     jsparsvardef_t *,
							     long, void *),
					   void *user, int mode);

static int               processFlags(void *user, jsparscnode_t *node);
static int               doProcVar(jsparsstream_t *jpa, jsparscnode_t *node,
				   jctokfifo_t *queue,
				   int (*doVariable)(jsparsstream_t *,
						     jsparscnode_t *,
						     jctokfifo_t *,
						     jsparsvartab_t *,
						     jsparsvardef_t *,
						     long, void *),
				   long flags, void *user, int mode);

#ifdef TRACETYPE_JSCPP
static void              prVarTab(jctokfifo_t *queue, int ident,
				  jsparsvartab_t *vTab);
static void              prVarDef(jctokfifo_t *queue, int ident,
				  jsparsvardef_t *vDef);

static int               scanBadCast(jsparsstream_t *jpa, jctokfifo_t *queue,
				     jsparscnode_t *node);
#endif

static int               warnVars(jsparsstream_t *jpa, jsparscnode_t *node);
static int               doVarCheck(jsparsstream_t *jpa, jsparscnode_t *node,
				    jctokfifo_t *queue, jsparsvartab_t *vTab,
				    jsparsvardef_t *vDef,
				    long flags, void *user);
static usageinterval_t  *allocusaint(void);
/*
static void              freeusaint(usageinterval_t *ui);
*/
static void              cleanusaint(void);

static int               warnStackUse(jsparsstream_t *jpa, jctokfifo_t *queue,
				      jsparscnode_t *node);

Prototype int            jsparsgetvarname(jsparscnode_t *node,
					  jsparscnode_t **pVarNode,
					  int (*handleOps)(void *u, jsparscnode_t *op),
					  void *user);

/*********     GLOBAL VARIABLES                                 *********/
unsigned long JSParseMinStackSizeWarn = 100000;

/*JS*********************************************************************
*
*************************************************************************/

int _jsparsfindvariables(jsparsstream_t *jpa, jsparscnode_t *node)

/************************************************************************/
{
  jctokfifo_t *queue;

  queue = jsqsgetqueue(jpa->JSPA_Stream);

  /*  Scan for variables  */
  if(_jsparsscanvars(jpa, node, queue, jpa->JSPA_VarTables[0], 0) == -1)
    EGOTO(error);

#ifdef TRACETYPE_JSCPP
  if(jstracelevel(TRACETYPE_JSCPP) >= 6) {
    unsigned int i;

    for(i = 0 ; i < jpa->JSPA_NVarTables ; ++i)
      prVarTab(queue, 0, jpa->JSPA_VarTables[i]);
  }

  if(jstracelevel(TRACETYPE_JSCPP) >= 6 &&
     scanBadCast(jpa, queue, node) == -1)
    EGOTO(error);
#endif

  /*  Look for unused variables or variables that could be localized  */
  if(node->JPACN_Type == JPACT_FUNCDEF) {
    if((jpa->JSPA_WarnFlags & JPAM_WARNVARS) &&
       warnVars(jpa, node) == -1)
      EGOTO(error);

    if((jpa->JSPA_WarnFlags & JPAM_WARNSTACKUSE) &&
       warnStackUse(jpa, queue, node) == -1)
      EGOTO(error);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*   mode bit 0: If set, also treat parameter declarations.
*   mode bit 1: If set, copy the decl/spec to the global queue.
*   mode bit 2: If set, we are the block of a 'switch' statement.
*************************************************************************/

int _jsparsscanvars(jsparsstream_t *jpa, jsparscnode_t *node,
		    jctokfifo_t *queue, jsparsvartab_t *vTab, int mode)

/************************************************************************/
{
  jsparsvartab_t *subTab;
  jsparscnode_t *sub;

  subTab = vTab;

  /*  Check first for labels to avoid too much recursion  */
  if(node->JPACN_Type == JPACT_LABELEDSTMT) {
    jsparsvartab_t *gTab, *nTab;

    /*  Define the label in the topmost variable list, i. e. it
     *   is visible in the whole function.
     *   Store the node in the Declarator field.
     */
    for(gTab = vTab ; ; gTab = nTab) {
      nTab = gTab->JPAVT_Next;
      assert(nTab != NULL);
      if(jsparsisglobalvartab(jpa, nTab))
	break;
    }

    assert(!jsparsisglobalvartab(jpa, gTab));

    node = node->JPACN_Sub; /*  the labels  */

    for(sub = node->JPACN_Type == JPACT_LIST ? node->JPACN_Sub : node ;
	sub ; sub = sub->JPACN_Next) {
      assert(sub->JPACN_Type == JPACT_LABEL ||
	     sub->JPACN_Type == JPACT_CASE ||
	     sub->JPACN_Type == JPACT_DEFAULT ||
	     sub->JPACN_Type == JPACT_PRAGMALINE ||
	     sub->JPACN_Type == JPACT_PRAGMALINE2);

      if(sub->JPACN_Type == JPACT_LABEL) {
	if(jsparsdefinevar(jpa, gTab, JVDF_FDEFINED | JVDF_TLABEL,
			   queue, sub->JPACN_IndexPP,
			   node->JPACN_Next, NULL, 0) == NULL)
	  EGOTO(error);
      }

      if(node->JPACN_Type != JPACT_LIST)
	break;
    }

    /*  Scan recursively with the actual statement. There might be none if
     *   we have trailing labels
     */
    if((node = node->JPACN_Next) == NULL)
      return 0;
  }

  if(node->JPACN_Type == JPACT_DECLARATION ||
     node->JPACN_Type == JPACT_TYPEDEF ||
     /*  We make it dependent on the mode, so we can choose
      *   to ignore parameter declaration, e. g. in
      *   "x = (int (*)(int a,int b))y;" we don't want to
      *   define "a" and "b".
      */
     ((mode & 1) && node->JPACN_Type == JPACT_PARDECL)) {
    return doDefinitions(jpa, node, queue, vTab, mode);
  }

  if(node->JPACN_Type == JPACT_CAST) {
    /*  The cast might look like "(struct Bad1 { int Worse; } *) vptr" amd
     *   we don't want to define "Worse", so simply ignore it.
     */
    return 0;
  }

  if(node->JPACN_Type == JPACT_UNARYEXPR) {
    if((jpa->JSPA_WarnFlags & JPAM_WARNBOUNDS) &&
       checkArrayBounds(jpa, node, queue, vTab) == -1)
      EGOTO(error);

  } else if(node->JPACN_Type == JPACT_FUNCDEF) {
    jsparsvardef_t *vDef;
    const char *funcName;
    jsparscnode_t *decl, *pars, *pars2;
    int start, end;

    /*  The variable table wraps the whole function definition  */
    start = node->JPACN_Start;
    end = node->JPACN_End;

    node = node->JPACN_Sub;
    assert(node->JPACN_Type == JPACT_DECLARATION);

    decl = node->JPACN_Sub;
    assert(decl->JPACN_Type == JPACT_DECLARATOR);

    sub = decl->JPACN_Sub;
    assert(sub->JPACN_Type == JPACT_IDENT);

    node = node->JPACN_Next;
    if(node->JPACN_Type == JPACT_LIST ||
       node->JPACN_Type == JPACT_DECLARATION) {
      /*  K&R-type function definition  */
      pars = node;

      node = node->JPACN_Next;

      /*  Memorize the function brackets  */
      pars2 = sub->JPACN_Next;
      assert(pars2->JPACN_Type == JPACT_FUNCCALL);
      pars2 = pars2->JPACN_Sub;
    } else {
      pars = sub->JPACN_Next;
      assert(pars->JPACN_Type == JPACT_FUNCCALL);
      pars = pars->JPACN_Sub;

      if(pars && pars->JPACN_Type == JPACT_IDENT)
	/*  We have something like
	 *    kr_all_ints(one, two)
	 *    {
	 *    }
	 */
	pars2 = pars;
      else
	pars2 = NULL;
    }

    assert(node->JPACN_Type == JPACT_COMPSTMT);

    funcName = queue->TKFF_Token[sub->JPACN_IndexPP].TK_String;
    if((vDef = jsparsgetvardef(jpa, vTab, funcName,
			       JVDF_TFUNCTION, 0)) != NULL) {
      if((vDef->JPAVD_Flags & JVDF_ISTATIC) &&
	 (jpa->JSPA_WarnFlags & JPAM_WARNPROTO)) {
	jsparscnode_t *n;

	for(n = decl->JPACN_Next ; n ; n = n->JPACN_Next)
	  if(n->JPACN_Type == JPACT_STATIC)
	    break;

	if(n == NULL)
	  jsparsprinterr(jpa, queue, sub,
			 "Function '%s' becomes static due to prototype.", 0,
			 funcName);
      }
    } else {
#if 0 /*  This produces to many spurious warnings  */
      if(jpa->JSPA_WarnFlags & JPAM_WARNPROTO)
	jsparsprinterr(jpa, queue, sub,
		       "Function '%s' defined without previous prototype.", 0,
		       funcName);
#endif

      /*  Define the function in the global scope  */
      if(jsparsdefinevar(jpa, vTab, JVDF_FDEFINED | JVDF_TFUNCTION,
			 queue, sub->JPACN_IndexPP,
			 decl, decl->JPACN_Next, mode) == NULL)
	EGOTO(error);
    }

#if 1 /*  FIXME: Not needed  */
    sub = sub->JPACN_Next;
    assert(sub->JPACN_Type == JPACT_FUNCCALL);
#endif

    /* We need a topmost variable table in any case -
     *  we might have labels
     */
    if((subTab = _jsparsgetvartab(vTab, 0, start, end)) == NULL)
      EGOTO(error);

    /*  Define the parameters as variables in the function body */
    for( ; pars ; pars = pars->JPACN_Next)
      if(_jsparsscanvars(jpa, pars, queue, subTab, 1 | mode) == -1)
	EGOTO(error);

    /*  K&R: Check if all variables are already defined and define
     *   remaining ones as 'int'
     */
    for( ; pars2 ; pars2 = pars2->JPACN_Next) {
      const char *name;

      assert(pars2->JPACN_Type == JPACT_IDENT);

      name = queue->TKFF_Token[pars2->JPACN_IndexPP].TK_String;

      if(jsparsgetvardef(jpa, subTab, name, JVDF_TVARIABLE, 0) == NULL) {
static const jsparscnode_t TokenInt = { NULL, JPACT_INT,     NULL, -1, -1 };

        /*  Define this variable as of type integer  */
	if(jsparsdefinevar2(jpa, subTab, JVDF_FDEFINED | JVDF_TVARIABLE,
			    queue, name, NULL, (jsparscnode_t *)&TokenInt,
			    0) == NULL)
	  EGOTO(error);
      }
    }

    /*  Now scan the body  */
  }

  /* ******************************************************************** */

  if(node->JPACN_Type == JPACT_COMPSTMT && subTab == vTab) {
    /*  Check if we need our own variable table for this block  */

    /*  If we want to find variables that could be moved to inner blocks,
     *   we even generate empty variable tables. The only exception is
     *   if we have a block of a 'switch' statement which usually doesn't
     *   have variables. In this case, we avoid telling the user to move
     *   a variable to the 'switch' block.
     *   However, if the 'switch' block has variable definitions, we can
     *   handle it.
     */
    if((!(mode & 4) && (jpa->JSPA_WarnFlags & JPAM_WARNVARS)) ||
       /*  Check if this block has local variable definitions  */
       ((sub = node->JPACN_Sub) != NULL &&
	(sub->JPACN_Type == JPACT_DECLARATION ||
	 sub->JPACN_Type == JPACT_TYPEDEF))) {
      if((subTab = _jsparsgetvartab(vTab, 0, node->JPACN_Start,
				   node->JPACN_End)) == NULL)
	EGOTO(error);
    }
  }

  if(node->JPACN_Type == JPACT_SWITCH)
    mode |= 4;
  else
    /*  We have used it above, so disable for follow up calls.  */
    mode &= ~4;

  /* ***  Scan recursively  *** */
  for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next)
    if(_jsparsscanvars(jpa, sub, queue, subTab, mode) == -1)
      EGOTO(error);

  if(subTab != vTab && jsparsinsertvtab(jpa, subTab) == -1)
    EGOTO(error);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static jsparsvardef_t *jsparsdefinevar(jsparsstream_t *jpa,
				       jsparsvartab_t *vTab, int type,
				       jctokfifo_t *queue, int ind,
				       jsparscnode_t *declarator,
				       jsparscnode_t *specifier,
				       int mode)

/************************************************************************/
{
  const char *name;

  name = queue->TKFF_Token[ind].TK_String;

  return jsparsdefinevar2(jpa, vTab, type, queue, name, declarator,
			  specifier, mode);
}

/*JS*********************************************************************
*
*************************************************************************/

static int doDefinitions(jsparsstream_t *jpa,
			 jsparscnode_t *node, jctokfifo_t *queue,
			 jsparsvartab_t *vTab, int mode)

/************************************************************************/
{
  jsparscnode_t *decl, *n, *n2, *subdef, *spec;
  jsparsvartab_t *subTable;
  jsparsvardef_t *vDef;
  int type, stype, flags;

  if(node->JPACN_Type == JPACT_TYPEDEF)
    type = JVDF_TTYPEDEF;
  else if(node->JPACN_Type == JPACT_PARDECL)
    /*  Function parameters are always initialized  */
    type = JVDF_TVARIABLE | JVDF_IFUNCARG | JVDF_FWRITE;
  else
    type = JVDF_TVARIABLE;

  decl = node->JPACN_Sub;
  if(decl->JPACN_Type == JPACT_LIST || decl->JPACN_Type == JPACT_DECLARATOR ||
     decl->JPACN_Type == JPACT_ASS) {
    spec = decl->JPACN_Next;
  } else {
    /*  There is no declarator (i. e. a pure struct or union
     *   declaration or definition)
     */
    spec = decl;
    decl = NULL;
  }

  /* ***  Process specifier  *** */
  /*  Check for definition or declaration  */
  flags = JVDF_FDEFINED;
  stype = -1;
  subdef = NULL;
  for(n = spec ; n ; n = n->JPACN_Next) {
    if(n->JPACN_Type == JPACT_EXTERN)
      flags = JVDF_FDECLARED;
    else if(n->JPACN_Type == JPACT_STRUCT) {
      stype = JVDF_TSTRUCT;
      subdef = n;
    } else if(n->JPACN_Type == JPACT_UNION) {
      stype = JVDF_TUNION;
      subdef = n;
    } else if(n->JPACN_Type == JPACT_ENUM) {
      stype = JVDF_TENUM;
      subdef = n;
    }
  }
  type |= flags;

  subTable = NULL;
  flags = 0;
  if(stype != -1) {
    n = subdef->JPACN_Sub;
    vDef = NULL;
    if(n->JPACN_Type == JPACT_IDENT) {
      n2 = n->JPACN_Next;

      /*  A single line like "struct bingo;" or
       *   "struct bingo i;" will not trigger
       *   a declaration of struct "bingo".
       *   No one needs this.
       */
      if(n2 && n2->JPACN_Type == JPACT_BLOCK) {
	stype |= JVDF_FDEFINED;

	if((vDef = jsparsdefinevar(jpa, vTab, stype, queue, n->JPACN_IndexPP,
				   NULL, spec, mode)) == NULL)
	  EGOTO(error);
      }

      n = n2;
    }

    if(n && n->JPACN_Type == JPACT_BLOCK) {
      if((stype & JVDF_TMASK) == JVDF_TENUM) {
	if(scanEnumeration(jpa, n, queue, vTab) == -1)
	  EGOTO(error);
      } else {
	int smode, ret;

	/*  struct/union with block  */
	if((subTable = _jsparsgetvartab(vTab, JVDT_FSTRUCTUNION,
					n->JPACN_Start,
					n->JPACN_End)) == NULL)
	  EGOTO(error);

	smode = mode;
	if(jsparsisglobalvartab(jpa, vTab))
	  /*  The struct/union was defined globally, so we
	   *   have to copy the declarator/specifier to the global
	   *   definition as well.
	   */
	  smode |= 2;

	ret = _jsparsscanvars(jpa, n, queue, subTable, smode);

	/*  We'll keep subTable->JPAVT_Next because it defines the next
	 *   layer for the stucture where e. g. global definitions are found
	 */

	if(ret == -1)
	  EGOTO(error);

	if(vDef) {
	  /*  A named structure, so the structure definition owns
	   *   the table
	   */
	  vDef->JPAVD_SubTable = subTable;
	  vDef->JPAVD_Flags |= JVDF_FFREESUBTAB;
	} else
	  /*  An anonymous structure, so make sure that someone
	   *   becomes the owner of the subtable
	   */
	  flags = JVDF_FFREESUBTAB;
      }
    }
  }

  /* ***  Process declarators  *** */
  if(decl != NULL) {
    n = (decl->JPACN_Type == JPACT_LIST) ? decl->JPACN_Sub : decl;

    for( ; n ; n = n->JPACN_Next) {
      jsparscnode_t *currdecl, *ass;

      currdecl = n;
      if(currdecl->JPACN_Type == JPACT_ASS) {
	currdecl = currdecl->JPACN_Sub;
	ass = currdecl->JPACN_Next;
	if(ass->JPACN_Type == JPACT_UNARYEXPR)
	  ass = ass->JPACN_Sub;
      } else
	ass = NULL;

      n2 = currdecl->JPACN_Sub;
      if(n2->JPACN_Type == JPACT_IDENT) {
	int currType;

	/*  A line like "typedef int IFUNC(...);" is *not* a function
	 *   declaration, it is a typedef
	 */
	if((type & JVDF_TMASK) != JVDF_TTYPEDEF &&
	   n2->JPACN_Next && n2->JPACN_Next->JPACN_Type == JPACT_FUNCCALL)
	  currType = JVDF_FDECLARED | JVDF_TFUNCTION;
	else
	  currType = type;

	if((vDef = jsparsdefinevar(jpa, vTab, currType,
				   queue, n2->JPACN_IndexPP,
				   currdecl, spec, mode)) == NULL)
	  EGOTO(error);

	vDef->JPAVD_SubTable = subTable;
	vDef->JPAVD_Flags |= flags;
	flags = 0;  /*  signal that it is used  */

	n2 = n2->JPACN_Next;
	if(n2 && n2->JPACN_Type == JPACT_VECTOR && n2->JPACN_Sub == NULL && ass) {
	  int arraySize;

	  /*  There is a vector that should be automatically sized  */
	  if(ass->JPACN_Type == JPACT_BLOCK) {
	    arraySize = 0;
	    for(ass = ass->JPACN_Sub ; ass ; ass = ass->JPACN_Next)
	      ++arraySize;
	  } else if(ass->JPACN_Type == JPACT_STRING ||
		    ass->JPACN_Type == JPACT_WSTRING ||
		    ass->JPACN_Type == JPACT_USTRING) {
	    int i, step;

	    step = jctokgetstep(queue);
	    arraySize = 1; /*  for the '\0'  */
	    for(i = ass->JPACN_IndexPP ; i <= ass->JPACN_End ; i += step)
	      arraySize += strlen(queue->TKFF_Token[i].TK_String);
	  } else
	    arraySize = -1;

	  if(arraySize != -1) {
	    /*  Memorize determined array size  */
	    vDef->JPAVD_LongValue = arraySize;
	    vDef->JPAVD_Flags |= JVDF_FARRAYSIZE;
	  }
	}
      }

      if(decl->JPACN_Type != JPACT_LIST)
	break;
    }
  }

  if(flags && subTable) {
    /*  That's strange - something like "struct { int x; };", so free
     *   the table we generated.
     */
    jsparsfreevartab(subTable);
  }

  return 0;
error:
  return -1;
}

/* ERROR-DEFINITIONS from scanEnumeration label _ERR_JSPARSSCANENUM
   Invalid value for enumeration constant
*/

/*JS*********************************************************************
*
*************************************************************************/

static int scanEnumeration(jsparsstream_t *jpa,
			   jsparscnode_t *node, jctokfifo_t *queue,
			   jsparsvartab_t *vTab)

/************************************************************************/
{
  unsigned long value;

  for(value = 0, node = node->JPACN_Sub ; node ;
      ++value, node = node->JPACN_Next) {
    jsparsvardef_t *vDef;
    jsparscnode_t *sub, *valNode;
    const char *name;

    sub = node;
    if(sub->JPACN_Type == JPACT_ASS) {
      sub = sub->JPACN_Sub;
      valNode = sub->JPACN_Next;
    } else
      valNode = NULL;

    assert(sub->JPACN_Type == JPACT_IDENT);
    name = queue->TKFF_Token[sub->JPACN_IndexPP].TK_String;

    if(valNode != NULL) {
      jsparsexprtype_t eType;

      /*  Enumeration constants must not have any side effects  */
      if(jsparsgetexprtype(&eType, jpa, vTab, valNode, queue, 1) == NULL)
	EGOTO(error);

      if(!(eType.JSPE_Flags & JPA_VARTYPE_FLONGCONST)) {
	jsparsseterrcode2(_ERR_JSPARSSCANENUM, 0, name, queue, sub);
	EGOTO(error);
      }

      value = eType.JSPE_LongConst;

      jsparsfreeexprtype(jpa, &eType, 1);
    }

    /*  Define this enumeration constant  */
    if((vDef = jsparsgetvardef(jpa, vTab, name, JVDF_FDEFINED | JVDF_TENUMIDENT,
			       0)) == NULL)
      EGOTO(error);

    vDef->JPAVD_LongValue = value;
  }

  return 0;
error:
  return -1;
}

/* ERROR-DEFINITIONS from jsparscheckarrayaccess label _ERR_JSPARSCHECKARRBOUND
   Array access to non-array type
*/

/*JS*********************************************************************
*
*************************************************************************/

static int checkArrayBounds(jsparsstream_t *jpa, jsparscnode_t *expr,
			    jctokfifo_t *queue, jsparsvartab_t *vTab)

/************************************************************************/
{
  jsparscnode_t *prev, *node;
  const char *name;
  unsigned int level;

  name = NULL;
  level = 0;
  for(prev = NULL, node = expr->JPACN_Sub ; node ;
      prev = node, node = node->JPACN_Next) {
    if(node->JPACN_Type == JPACT_IDENT) {
      jsparscnode_t *next;

      name = queue->TKFF_Token[node->JPACN_IndexPP].TK_String;

      if(jpa->JSPA_NBoundChkPragma > 0 &&
	 (next = node->JPACN_Next) != NULL &&
	 next->JPACN_Type == JPACT_FUNCCALL) {
	unsigned int i;

	for(i = 0 ; i < jpa->JSPA_NBoundChkPragma ; ++i)
	  if(strcmp(jpa->JSPA_BoundChkPragma[i].JPABC_Name, name) == 0) {
	    _jsparsboundchk_t *pboundchk;
	    jsparscnode_t *sub, *arrayArg, *indexArg;
	    int nr;

	    pboundchk = &(jpa->JSPA_BoundChkPragma[i]);

	    /*  Find the required arguments  */
	    arrayArg = indexArg = NULL;
	    for(nr = 1, sub = next->JPACN_Sub ; sub ;
		++nr, sub = sub->JPACN_Next)
	      if(nr == pboundchk->JPABC_ArrayArg)
		arrayArg = sub;
	      else if(nr == pboundchk->JPABC_IndexArg)
		indexArg = sub;

	    if(arrayArg != NULL && indexArg != NULL &&
	       checkArrayBoundary(jpa, arrayArg, indexArg,
				  queue, vTab, name,
				  -1, pboundchk) == -1)
	      EGOTO(error);
	  }
      }
    } else if(node->JPACN_Type == JPACT_VECTOR) {
      int ret;

      ++level;

      /*  Modify the expression temporarily to find the array dimension  */
      assert(prev != NULL);
      prev->JPACN_Next = NULL;
      ret = checkArrayBoundary(jpa, expr, node->JPACN_Sub,
			       queue, vTab, name, level, NULL);
      prev->JPACN_Next = node;
      if(ret == -1)
	EGOTO(error);
    }
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int checkArrayBoundary(jsparsstream_t *jpa, jsparscnode_t *arrArg,
			      jsparscnode_t *indArg, jctokfifo_t *queue,
			      jsparsvartab_t *vTab,
			      const char *name, int level,
			      _jsparsboundchk_t *pboundchk)

/************************************************************************/
{
  jsparsexprtype_t aType, dimType;
  jsparscnode_t *node;
  long arrayIndex, arrayDimension;

  /*  Check if array index is an integer constant  */
  if(jsparsgetexprtype(&aType, jpa, vTab, indArg, queue, 2) == NULL)
    EGOTO(error2);

  /*  If the array index is not a constant (most common case), we are done  */
  if(!(aType.JSPE_Flags & JPA_VARTYPE_FLONGCONST)) {
    jsparsfreeexprtype(jpa, &aType, 1);
    return 0;
  }

  arrayIndex = aType.JSPE_LongConst;

  jsparsfreeexprtype(jpa, &aType, 1);

  if(jsparsgetexprtype(&aType, jpa, vTab, arrArg, queue, 2) == NULL)
    EGOTO(error2);

  if((node = aType.JSPE_Declarator) == NULL) {
    jsparsseterrcode(_ERR_JSPARSCHECKARRBOUND, 0, queue, arrArg);
    EGOTO(error);
  }

  /*  Check if this is defined as an array or maybe only a pointer  */
  if(node->JPACN_Type != JPACT_VECTOR || node->JPACN_Sub == NULL) {
    jsparsfreeexprtype(jpa, &aType, 1);
    return 0;
  }

  if(jsparsgetexprtype(&dimType, jpa, vTab, node->JPACN_Sub,
		       aType.JSPE_Queue, 2) == NULL)
    EGOTO(error);

  if(!(dimType.JSPE_Flags & (JPA_VARTYPE_FARRAYSIZE | JPA_VARTYPE_FLONGCONST))) {
    /*  Looks like we don't know the dimension of this array  */
    jsparsfreeexprtype(jpa, &dimType, 1);
    jsparsfreeexprtype(jpa, &aType, 1);
    return 0;
  }

  arrayDimension = dimType.JSPE_LongConst;
  jsparsfreeexprtype(jpa, &dimType, 1);
  jsparsfreeexprtype(jpa, &aType, 1);

  if(pboundchk) {
    if(pboundchk->JPABC_Flag & 1)
      --arrayDimension;

    /*  The "index" may be arrayDimension  */
    if(arrayIndex <= arrayDimension)
      return 0;
  } else {
    if(arrayIndex < arrayDimension)
      return 0;
  }

  if(pboundchk == NULL)
    jsparsprinterr(jpa, queue, arrArg,
		   "Array access on '%s' (%d. level): index %ld exceeds dimension %ld!",
		   0, name ? name : "????", level, arrayIndex, arrayDimension);
  else
    jsparsprinterr(jpa, queue, arrArg,
		   "Function '%s': %d. arg %ld exceeds dimension %ld of %d. arg!",
		   0, name ? name : "?????", pboundchk->JPABC_IndexArg, arrayIndex,
		   arrayDimension, pboundchk->JPABC_ArrayArg);

  return 0;
error:
  jsparsfreeexprtype(jpa, &aType, 1);
error2:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jsparsclearvariables(jsparsstream_t *jpa)

/************************************************************************/
{
  unsigned int i;

  /*  Delete old stuff from previous run EXCEPT global variables  */
  for(i = 1 ; i < jpa->JSPA_NVarTables ; ++i)
    jsparsfreevartab(jpa->JSPA_VarTables[i]);

  jpa->JSPA_NVarTables = 1;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

void jsparsclearvarflags(jsparsvartab_t *vTab, long flags)

/************************************************************************/
{
  unsigned int i;

  flags = ~flags;

  for(i = 0 ; i < JPA_VARIABLE_TABSIZE ; ++i) {
    jsparsvardef_t *vDef;

    for(vDef = vTab->JPAVT_Vars[i] ; vDef ; vDef = vDef->JPAVD_Next)
      vDef->JPAVD_Flags &= flags;
  }
}

/*JS*********************************************************************
*
*************************************************************************/

int jsparsprocessvars(jsparsstream_t *jpa, jsparscnode_t *node,
		      jctokfifo_t *queue,
		      int (*doVariable)(jsparsstream_t *, jsparscnode_t *,
					jctokfifo_t *, jsparsvartab_t *,
					jsparsvardef_t *,
					long, void *),
		      void *user, int mode)

/************************************************************************/
{
restart:
  switch(node->JPACN_Type) {
  case JPACT_TYPEDEF: case JPACT_PARDECL:
    /*  Don't look at them again - we already did it when creating
     *   the variable definitions
     */
    break;

  case JPACT_STRUCT: case JPACT_UNION: case JPACT_ENUM:
    /*  We are not looking at struct/union/enum names  */
    break;

  case JPACT_DECLARATION:
    node = node->JPACN_Sub;

    /*  Deal with initialization in the definition  */
    if(node->JPACN_Type == JPACT_ASS) {
      if(jsparsprocessvars(jpa, node, queue, doVariable, user, mode) == -1)
	EGOTO(error);
    } else if(node->JPACN_Type == JPACT_LIST) {
      for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next)
	if(node->JPACN_Type == JPACT_ASS)
	  if(jsparsprocessvars(jpa, node, queue, doVariable, user, mode) == -1)
	    EGOTO(error);
    }

    break;

  case JPACT_ASS:       case JPACT_MULASS:    case JPACT_DIVASS:
  case JPACT_MODASS:    case JPACT_ADDASS:    case JPACT_SUBASS:
  case JPACT_LSHIFTASS: case JPACT_RSHIFTASS:
  case JPACT_ANDASS:    case JPACT_EXORASS:   case JPACT_ORASS:
    {
      jsparscnode_t *n;
      long flags;

      /*  Treat assignments  */
      flags = JVDF_FWRITE;
      if(node->JPACN_Type != JPACT_ASS)
	/*  Assignments where the previous values are used as well  */
	flags |= JVDF_FREAD;

      /*  Treat also expressions like "a[1]", "*a" etc.  */
      node = node->JPACN_Sub;

      assert(node->JPACN_Type == JPACT_UNARYEXPR ||
	     node->JPACN_Type == JPACT_DECLARATOR);

      if(doProcVar(jpa, node, queue, doVariable, flags, user, mode) == -1)
	EGOTO(error);

      /*  There might be other variables like in a[i] or p->a[i], so call
       *   recursively on what follows.
       */
      for(n = node->JPACN_Sub ; n ; n = n->JPACN_Next)
	if(jsparsprocessvars(jpa, n, queue, doVariable, user, mode) == -1)
	  EGOTO(error);

      node = node->JPACN_Next;
    }

    /*  Now handle the right hand side of the assignment  */
    goto restart; /*  avoid unnecessary recursion  */

  case JPACT_MEMBER: case JPACT_POINTTO:
    /*  Don't look at the member names  */
    break;

  case JPACT_LABEL:
  case JPACT_GOTO:
    break;

  case JPACT_UNARYEXPR:
    /*  A variable might be used  */
    if(doProcVar(jpa, node, queue, doVariable, JVDF_FREAD,
		 user, mode) == -1)
      EGOTO(error);

    /*  There could be unary expressions in there, so scan recursively  */

  default:
    /*  Scan recursively  */
    for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next)
      if(jsparsprocessvars(jpa, node, queue, doVariable, user, mode) == -1)
	EGOTO(error);

    break;
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int processFlags(void *user, jsparscnode_t *node)

/************************************************************************/
{
  int type;

#define paux    ((auxprocvar_t *)user)

  if(paux->Mode & 1) {
    type = node->JPACN_Type;
    if(type == JPACT_ADDRESS)
      paux->Flags |= JVDF_FADDR;
    else if(type == JPACT_POSTINC || type == JPACT_POSTDEC ||
	    type == JPACT_PREINC  || type == JPACT_PREDEC)
      paux->Flags |= JVDF_FREAD | JVDF_FWRITE;
    else if(type == JPACT_POINTTO || type == JPACT_POINTER)
      /*  The pointer value is used to access the underlying thingadoo  */
      paux->Flags |= JVDF_FREAD;
    else if(type == JPACT_FUNCCALL)
      paux->Mode |= 2;

    paux->Mode &= ~1; /*  bit 0: If set, this is the first operator  */
  }

#undef paux

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doProcVar(jsparsstream_t *jpa, jsparscnode_t *node,
		     jctokfifo_t *queue,
		     int (*doVariable)(jsparsstream_t *, jsparscnode_t *,
				       jctokfifo_t *, jsparsvartab_t *,
				       jsparsvardef_t *,
				       long, void *),
		     long flags, void *user, int mode)

/************************************************************************/
{
  jsparsvartab_t *vTab, *vTab2;
  jsparsvardef_t *vDef;
  jsparscnode_t *varNode;
  const char *name;
  auxprocvar_t aux;

  aux.Mode = 1;
  aux.Flags = flags;
  if(jsparsgetvarname(node, &varNode, processFlags, (void *)&aux) == 0)
    /*  nothing to do  */
    return 0;

  flags = aux.Flags;

  /*  Find the variable table responsible for this node  */
  if((vTab = jsparsfindvartab(jpa, node)) == NULL)
    EGOTO(error);

  name = queue->TKFF_Token[varNode->JPACN_IndexPP].TK_String;

  /*  ANSI feature: We can call a function via a function pointer just
   *   like that, so we first check if the identifier is known as a
   *   variable.
   */
  if((vDef = jsparsgetvardef2(jpa, vTab, name, JVDF_TVARIABLE,
			      &vTab2, 0)) == NULL) {
    /*  Maybe we refer to a global function with this name?  */
    if((vDef = jsparsgetvardef2(jpa, vTab, name, JVDF_TFUNCTION,
				&vTab2, 0)) == NULL) {
      if(!(aux.Mode & 2)) {
	/*  An unknown variable - the coding is wrong  */
	jsparsprinterr(jpa, queue, node,
		       "Variable '%s' not found.", (mode & 1) ? 0 : 1, name);

	if(mode & 1)
	  /*  Ignore unknown variables  */
	  return 0;

	EGOTO(error);
      }

      /*  Function call without prototype  */
      if(jpa->JSPA_WarnFlags & JPAM_WARNPROTO)
	jsparsprinterr(jpa, queue, node,
		       "Function '%s' is called without prototype", 0,
		       name);
    }
  }

  if(vDef != NULL) {
    jsparscnode_t *decl;

    /*  If this is an array and is accessed directly, this is
     *   equivalent to taken the 'address'
     */
    if((aux.Mode & 1) && (decl = vDef->JPAVD_Declarator) != NULL) {
      if(decl->JPACN_Type == JPACT_DECLARATOR)
	decl = decl->JPACN_Sub;

      if(decl->JPACN_Type == JPACT_IDENT)
	decl = decl->JPACN_Next;

      if(decl != NULL && decl->JPACN_Type == JPACT_VECTOR)
	flags |= JVDF_FADDR;
    }

    vDef->JPAVD_Flags |= flags;

    if(doVariable(jpa, varNode, queue, vTab2, vDef, flags, user) == -1)
      EGOTO(error);
  }

  return 0;
error:
  return -1;
}

#ifdef TRACETYPE_JSCPP
/*JS*********************************************************************
*
*************************************************************************/

static void prVarTab(jctokfifo_t *queue, int ident,
		     jsparsvartab_t *vTab)

/************************************************************************/
{
  unsigned int i, step;

  step = queue ? jctokgetstep(queue) : 1;

  jstrace(TRACETYPE_JSCPP, 6,
	  "%*sVariable list {%p,%p} [%d,%d]:\n", ident, "",
	  (void *)vTab, (void *)vTab->JPAVT_Next,
	  vTab->JPAVT_StartIndex / step, vTab->JPAVT_EndIndex / step);

  for(i = 0 ; i < JPA_VARIABLE_TABSIZE ; i++) {
    jsparsvardef_t *vDef;

    for(vDef = vTab->JPAVT_Vars[i] ; vDef ; vDef = vDef->JPAVD_Next)
      prVarDef(queue, ident + 2, vDef);
  }
}

/*JS*********************************************************************
*
*************************************************************************/

static void prVarDef(jctokfifo_t *queue, int ident,
		     jsparsvardef_t *vDef)

/************************************************************************/
{
  int fpos;

  jstrace(TRACETYPE_JSCPP, 6,
	  "%*s%-15s: %c%c %c%c%c %c%c%c%c ", ident, "", vDef->JPAVD_Name,
	  (vDef->JPAVD_Flags & JVDF_FDECLARED) ? 'd' : '-',
	  (vDef->JPAVD_Flags & JVDF_FDEFINED)  ? 'D' : '-',

	  (vDef->JPAVD_Flags & JVDF_IEXTERN)   ? 'E' : '-',
	  (vDef->JPAVD_Flags & JVDF_ISTATIC)   ? 'S' : '-',
	  (vDef->JPAVD_Flags & JVDF_IFUNCARG)  ? 'F' : '-',

	  (vDef->JPAVD_Flags & JVDF_ICONST)    ? 'C' : '-',
	  (vDef->JPAVD_Flags & JVDF_FREAD)     ? 'r' : '-',
	  (vDef->JPAVD_Flags & JVDF_FWRITE)    ? 'w' : '-',
	  (vDef->JPAVD_Flags & JVDF_FADDR)     ? 'a' : '-');

  switch(vDef->JPAVD_Flags & JVDF_TMASK) {
  case JVDF_TSTRUCT:    jstrace(TRACETYPE_JSCPP, 6, "Struct");    break;
  case JVDF_TUNION:     jstrace(TRACETYPE_JSCPP, 6, "Union");     break;
  case JVDF_TENUM:      jstrace(TRACETYPE_JSCPP, 6, "Enum");      break;
  case JVDF_TVARIABLE:
    if(vDef->JPAVD_Flags & JVDF_FARRAYSIZE)
      jstrace(TRACETYPE_JSCPP, 6, "ArrayVariable[%ld]", vDef->JPAVD_LongValue);
    else
      jstrace(TRACETYPE_JSCPP, 6, "Variable");
    break;
  case JVDF_TTYPEDEF:   jstrace(TRACETYPE_JSCPP, 6, "Typedef");   break;
  case JVDF_TENUMIDENT:
    jstrace(TRACETYPE_JSCPP, 6, "EnumIdent (%ld)",  vDef->JPAVD_LongValue);
    break;
  case JVDF_TFUNCTION:  jstrace(TRACETYPE_JSCPP, 6, "Function");  break;
  case JVDF_TLABEL:     jstrace(TRACETYPE_JSCPP, 6, "Label");     break;
  default:              jstrace(TRACETYPE_JSCPP, 6, "???");       break;
  }

  if((fpos = jctokindfpos(queue)) != -1) {
    jsparscnode_t *node;

    if((node = vDef->JPAVD_Declarator) == NULL)
      node = vDef->JPAVD_Specifier;

    if(node != NULL) {
      jsctoken_t postok;

      postok = vDef->JPAVD_Queue->TKFF_Token[node->JPACN_Start + fpos];

      jstrace(TRACETYPE_JSCPP, 6,
	      "\tl. %3d '%s'", postok.TK_Value, postok.TK_String);
    }
  }
  jstrace(TRACETYPE_JSCPP, 6, "\n");

  if(jstracelevel(TRACETYPE_JSCPP) >= 6) {
    if(vDef->JPAVD_Declarator) {
      FILE *tfp = jstracegetfp(TRACETYPE_JSCPP, 6);

      jstrace(TRACETYPE_JSCPP, 6,
	      "%*sDeclarator:\n", ident + 2, "");
      jsparsprctree(tfp, vDef->JPAVD_Queue, vDef->JPAVD_Declarator,
		    ident + 4, 1);
    }

    if(vDef->JPAVD_Specifier) {
      FILE *tfp = jstracegetfp(TRACETYPE_JSCPP, 6);
      jsparscnode_t *node;

      jstrace(TRACETYPE_JSCPP, 6,
	      "%*sSpecifier:\n", ident + 2, "");
      for(node = vDef->JPAVD_Specifier ; node ; node = node->JPACN_Next)
	jsparsprctree(tfp, vDef->JPAVD_Queue, node, ident + 4, 1);
    }
  }

  if(vDef->JPAVD_SubTable) {
    jstrace(TRACETYPE_JSCPP, 6,
	    "%*sSUBLIST%s:\n", ident + 4, "",
	    (vDef->JPAVD_Flags & JVDF_FFREESUBTAB) ? " (to free)" : "");
    prVarTab(queue, ident + 6, vDef->JPAVD_SubTable);
  }
}

/*JS*********************************************************************
*
*************************************************************************/

static int scanBadCast(jsparsstream_t *jpa, jctokfifo_t *queue,
		       jsparscnode_t *node)

/************************************************************************/
{
  jsparscnode_t *sub;
  int flag;

  flag = 0;
  for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next) {
    if(sub->JPACN_Type == JPACT_CAST) flag = 1;

    if(scanBadCast(jpa, queue, sub) == -1)
      return -1;
  }

  if(flag) { /*  expression has a cast, so check it  */
    jsparsexprtype_t eType;
    jsparsvartab_t *vTab;

    if((vTab = jsparsfindvartab(jpa, node)) == NULL)
      return -1;

    if(jsparsgetexprtype(&eType, jpa, vTab, node, queue, 0) == NULL) {
#ifdef DEBUG
      fprintf(stderr, "Failed getexprtypr for:\n");
      jsparsprctree(stderr, queue, node, 0, 0);
#endif
    }
  }

  return 0;
}
#endif

/*JS*********************************************************************
*
*************************************************************************/

static int warnVars(jsparsstream_t *jpa, jsparscnode_t *node)

/************************************************************************/
{
  jsparscnode_t *n;
  unsigned int k;
  jctokfifo_t *queue;

  queue = jsqsgetqueue(jpa->JSPA_Stream);

  /*  First mark all variable usages  */
  cleanusaint();
  if(jsparsprocessvars(jpa, node, queue, doVarCheck, NULL, 0) == -1)
    EGOTO(error);

  n = node->JPACN_Sub; /*  DECLARATION  */
  n = n->JPACN_Sub; /*  DECLARATOR  */
  n = n->JPACN_Sub; /*  IDENTIFIER  */

  for(k = 1 ; k < jpa->JSPA_NVarTables ; ++k) {
    jsparsvartab_t *vTab;
    unsigned int i;

    vTab = jpa->JSPA_VarTables[k];

    for(i = 0 ; i < JPA_VARIABLE_TABSIZE ; ++i) {
      jsparsvardef_t *vDef;

      for(vDef = vTab->JPAVT_Vars[i] ; vDef ; vDef = vDef->JPAVD_Next) {
	jsparscnode_t *decl;
	usageinterval_t *ui;
	long flags;

	decl = vDef->JPAVD_Declarator;

	flags = vDef->JPAVD_Flags;

	if((ui = (usageinterval_t *)(vDef->JPAVD_UserPtr)) == NULL) {
	  /*  This variable hasn't even been used at all  */
	  if((flags & JVDF_TMASK) == JVDF_TVARIABLE && (flags & JVDF_FDEFINED))
	    jsparsprinterr(jpa, queue, decl,
			   "Unused %s '%s'", 0, (flags & JVDF_IFUNCARG) ?
			   "parameter" : "variable", vDef->JPAVD_Name);
	} else {
	  jsparsvartab_t *vTab2;

	  /*  Non-pointer function parameters that are only set are odd, too  */
	  if((!(flags & JVDF_IFUNCARG) || decl->JPACN_Next == NULL) &&
	     (flags & JVDF_FWRITE) && !(flags & (JVDF_FREAD|JVDF_FADDR))) {
	    /*  Variable was used, but only set and value was never read  */
	    jsparsprinterr(jpa, queue, vDef->JPAVD_Declarator,
			   "%s '%s' was modified, but its value never used",
			   0, (flags & JVDF_IFUNCARG) ? "Parameter" : "Variable",
			   vDef->JPAVD_Name);
	  }

	  /*  Find the variable table that would be enough for the range
	   *   where the variable was used.
	   */
	  if((vTab2 = jsparsfindvartab2(jpa, ui->Start, ui->End)) == NULL)
	    EGOTO(error);

	  /*  Check if we could move the variable to an inner block.
	   *   Don't suggest it if the address of the variable was taken.
	   *   This is to handle situations like this correctly:
	   *   {
	   *      int x, *p;
	   *
	   *      if(...) {
	   *        p = &x;
	   *        x = 1;
	   *      }
	   *
	   *      ...
	   *
	   *      return *p;
	   *   }
	   *
	   *   By moving "x" to the 'if' block it will no longer be
	   *   valid at the return statement, thus generating invalid
	   *   code.
	   */
	  if(!(flags & (JVDF_IFUNCARG|JVDF_FADDR)) && vTab2 != vTab) {
	    unsigned int indfpos;

	    if((indfpos = jctokindfpos(queue)) != (unsigned int)-1)
	      jsparsprinterr(jpa, queue, vDef->JPAVD_Declarator,
			     "Variable '%s' could be moved to block starting at line %d",
			     0, vDef->JPAVD_Name,
			     queue->TKFF_Token[vTab2->JPAVT_StartIndex + indfpos].TK_Value);
	    else
	      jsparsprinterr(jpa, queue, vDef->JPAVD_Declarator,
			     "Definition of '%s' could be moved to an inner block",
			     0, vDef->JPAVD_Name);
	  }

	  /*  Clean up user pointer (is expected by callers)  */
	  vDef->JPAVD_UserPtr = NULL;
	}
      }
    }
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doVarCheck(jsparsstream_t *jpa, jsparscnode_t *node,
		      jctokfifo_t *queue, jsparsvartab_t *vTab,
		      jsparsvardef_t *vDef,
		      long flags, void *user)

/************************************************************************/
{
  usageinterval_t *ui;
  unsigned int ind;

  /*  Only do this for local variables  */
  if((vDef->JPAVD_Flags & JVDF_TMASK) != JVDF_TVARIABLE ||
     jsparsisglobalvartab(jpa, vTab))
    return 0;

  ind = node->JPACN_IndexPP;

  if((ui = vDef->JPAVD_UserPtr) == NULL) {
    if((ui = allocusaint()) == NULL)
      EGOTO(error);

    vDef->JPAVD_UserPtr = ui;
    ui->Start = ind;
    ui->End   = ind;
  } else {
    if(ind < ui->Start)
      ui->Start = ind;
    else if(ind > ui->End)
      ui->End = ind;
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*   Generate allocation routine
*************************************************************************/

#define GENMALLOC_TYPE          usageinterval_t
#define GENMALLOC_NAMEALLOC     allocusaint
/*#define GENMALLOC_NAMEFREE      freeusaint*/
#define GENMALLOC_NAMECLEANMEM  cleanusaint
#define GENMALLOC_MEMSIZE       100000
#define GENMALLOC_STATIC

#include "genmalloc.h"

/*JS*********************************************************************
*
*************************************************************************/

static int warnStackUse(jsparsstream_t *jpa, jctokfifo_t *queue,
			jsparscnode_t *node)

/************************************************************************/
{
  unsigned long *sizes, maxsize;
  const char *funcName;
  unsigned int k;

  if((sizes = (unsigned long *)
      malloc(jpa->JSPA_NVarTables * sizeof(*sizes))) == NULL)
    EGOTO(error);

  for(k = 1 ; k < jpa->JSPA_NVarTables ; ++k) {
    jsparsvartab_t *vTab;
    unsigned int i;

    vTab = jpa->JSPA_VarTables[k];

    sizes[k] = 0;
    for(i = 0 ; i < JPA_VARIABLE_TABSIZE ; ++i) {
      jsparsvardef_t *vDef;

      for(vDef = vTab->JPAVT_Vars[i] ; vDef ; vDef = vDef->JPAVD_Next) {
	unsigned long vsize;
	unsigned int type;

	/*  Only look at local variables, skip function arguments  */
	type = vDef->JPAVD_Flags & JVDF_TMASK;
	if(type != JVDF_TVARIABLE ||
	   (vDef->JPAVD_Flags & (JVDF_IFUNCARG|JVDF_IEXTERN|JVDF_ISTATIC)))
	  continue;

        if((vsize = jsparsgetvarsize2(jpa, vTab, vDef)) == (unsigned long)-1)
	  EGOTO(error);

#ifdef TRACETYPE_JSCPP
	jstrace(TRACETYPE_JSCPP, 6,
		"  VARIABLE %-10s SIZE %ld\n", vDef->JPAVD_Name, vsize);
#endif

	sizes[k] += vsize;
      }
    }
  }

  /*  Handle inner variable blocks: Walk up to highes block, but skip
   *   global and add up sizes
   */
  for(k = 1 ; k < jpa->JSPA_NVarTables ; ++k) {
    jsparsvartab_t *vTab;
    unsigned int pos;

    /*  Parent variable table  */
    vTab = jpa->JSPA_VarTables[k]->JPAVT_Next;
    if(vTab->JPAVT_Next == NULL) /*  skip global list  */
      continue;

    for(pos = 1 ; pos < jpa->JSPA_NVarTables ; ++pos)
      if(vTab == jpa->JSPA_VarTables[pos])
	break;
    assert(pos < jpa->JSPA_NVarTables);

    sizes[k] += sizes[pos];
  }

  /*  Determine maximum  */
  maxsize = sizes[1];
  for(k = 2 ; k < jpa->JSPA_NVarTables ; ++k) {
#ifdef TRACETYPE_JSCPP
    jstrace(TRACETYPE_JSCPP, 6,
            "  VARTABLE #%d SIZE %ld\n", k, sizes[k]);
#endif

    if(sizes[k] > maxsize)
      maxsize = sizes[k];
  }

  free(sizes);
  /*sizes=NULL;*/

  /*  Get name of function  */
  node = node->JPACN_Sub->JPACN_Sub;
  assert(node->JPACN_Type == JPACT_DECLARATOR);
  node = node->JPACN_Sub;
  funcName = queue->TKFF_Token[node->JPACN_IndexPP].TK_String;

#ifdef TRACETYPE_JSCPP
  jstrace(TRACETYPE_JSCPP, 1,
	  "Stack usage of %8ld bytes in function '%s'.\n",
	  maxsize, funcName);
#endif

  if(maxsize >= JSParseMinStackSizeWarn) {
    jsparsprinterr(jpa, queue, node,
		   "Stack usage of %ld bytes in function '%s' might cause stack overflow.", 0,
		   maxsize, funcName);
  }

  return 0;
error:
  free(sizes);
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int jsparsgetvarname(jsparscnode_t *node,
		     jsparscnode_t **pVarNode,
		     int (*handleOps)(void *u, jsparscnode_t *op),
		     void *user)

/************************************************************************/
{
  int type, ret;

  type = node->JPACN_Type;
  if(type != JPACT_UNARYEXPR && type != JPACT_DECLARATOR)
    return 0;

  if((node = node->JPACN_Sub) == NULL)
    return 0;

  if(node->JPACN_Type == JPACT_IDENT) {
    *pVarNode = node;

    ret = 1;
  } else if(node->JPACN_Type == JPACT_ADD || node->JPACN_Type == JPACT_SUB) {
    jsparscnode_t *sub;

    /*  Treat cases like "*(a + 3)" etc.  */
    ret = 0;
    for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next) {
      if((ret = jsparsgetvarname(sub, pVarNode, handleOps, user)) == -1)
	EGOTO(error);

      if(ret)
	break;
    }
  } else if((ret = jsparsgetvarname(node, pVarNode, handleOps, user)) == -1)
    EGOTO(error);

  /*  If we found a variable, process the operators, but not in case like
   *      char *s = "hello";
   *   where we should *not* treat the
   *   "*" as an actual deferencing operator
   */
  if(ret && type != JPACT_DECLARATOR && handleOps) {
    while((node = node->JPACN_Next) != NULL)
      if(handleOps(user, node) == -1)
	EGOTO(error);
  }

  return ret;
error:
  return -1;
}
