/*JS*********************************************************************
*
*    Program : CRITICAL
*    Language: ANSI-C with UNIX SVR4 signal functions
*    Author  : Joerg Schoen
*    Purpose : To surround critical regions of code where a CONTROL-C
*              should not occur. May be nested!
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>

#include <jsconfig.h>

#ifdef _WIN32
#define CONFIG_NO_POSIX
#endif

#ifndef CONFIG_NO_POSIX
# include <unistd.h>
#endif

#include <signal.h>

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype void           criticalStart(void);
Prototype void           criticalEnd(void);

/*********     GLOBAL / STATIC VARIABLES                        *********/
#ifndef CONFIG_NO_POSIX
static sigset_t OldSet;
static int Counter = 0;

/*JS*********************************************************************
*   Blocks CNTRL-C and kill pid.
*************************************************************************/

void criticalStart(void)

/************************************************************************/
{
  /*  Allow multiple callings  */
  if(Counter++ == 0) {
    sigset_t block;

    /*  Create signal mask with SIGINT and SIGTERM  */
    sigemptyset(&block);
    sigaddset(&block, SIGINT);
    sigaddset(&block, SIGTERM);

    /*  Now block these signals and save old signal mask  */
    sigprocmask(SIG_BLOCK, &block, &OldSet);
  }
}

/*JS*********************************************************************
*   Reset to old signal mask.
*************************************************************************/

void criticalEnd(void)

/************************************************************************/
{
  /*  Prevent Counter become negative  */
  if(Counter <= 0) return;

  /*  Allow multiple callings  */
  if(--Counter == 0) {
    /*  Reset old signal mask  */
    sigprocmask(SIG_SETMASK, &OldSet, NULL);
  }
}
#else
/* ***  Dumb functions when not POSIX compliant  *** */
void criticalStart(void) {  /*  empty  */ }

void criticalEnd(void) {  /*  empty  */  }
#endif

#ifdef TEST
main()
{
  printf("Uncritical region ...\n");
  sleep(15);

  criticalStart();
  printf("Critical region ...\n");
  sleep(15);

  criticalEnd();

  printf("Uncritical region ...\n");
  sleep(15);
}
#endif
