/*JS*********************************************************************
*
*    Program : CSTRCOMPILE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Compile a string with C escape sequences.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype long           cStrCompile(const char *src, char *dest, int mode);

/*JS*********************************************************************
*   Understands C like escape sequences, including escaped quotes. Returns
*    the length of the compiled string in dest ('strlen' may produce wrong
*    results if the string contains '\0'). The line continuation sequence
*    "\\\n" is removed. src and dest may be identical. Unknown escape
*    sequences are kept.
*************************************************************************/

long cStrCompile(const char *src, char *dest, int mode)

/************************************************************************/
{
static const char InEsc[]  = "\"\'ntrfvbae\\";
static const char OutEsc[] = "\"\'\n\t\r\f\v\b\a\033\\";
  unsigned char *d;

  for(d = (unsigned char *)dest ; *src ; src++) {
    if((mode & 1) && *src == '\'' && src[1] == '\'') {
      /*  Substitute '' -> ' */
      *d++ = *src++;
      continue;
    }

    if((mode & 2) && *src == '"' && src[1] == '"') {
      /*  Substitute "" -> " */
      *d++ = *src++;
      continue;
    }

    if(*src == '\\') {
      char *string;

      src++; /*  Skip '\\'  */

      /*  Common escape sequence?  */
      if((string = strchr(InEsc, *src))) {
	*d = OutEsc[string - InEsc];

      } else if(*src == '\n') {
	continue; /*  Skip "\\\n" completely  */

      } else if('0' <= *src && *src <= '7') { /*  Octal number  */
	int i;

	i = 1;
	*d = *src - '0';
	while('0' <= src[1] && src[1] <= '7' && i++ < 3)
	  *d = (*d << 3) | (*++src - '0');

      } else if(*src == 'x') { /*  hexadecimal number  */
	int i;

	*d = 0;
	for(i = 0 ; isxdigit(*(const unsigned char *)src) && i < 2 ; i++) {
	  int s;

	  s = ('a' <= *src && *src <= 'f') ? 'a' :
	    (('A' <= *src && *src <= 'F') ? 'A' : '0');
	  *d = (*d << 4) | (*++src - s);
	}

      } else {
	/*  Unknown escape sequence, keep it in output  */
	*d++ = '\\';
	*d = *src;
      }

    } else
      *d = *src;

    d++;
  }

  return (char *)d - dest;
}

#ifdef TEST
int main(int argc, char *argv[])
{
  char buffer[1000];
  long len, i;

  while(--argc > 0) {
    len = cStrCompile(*++argv, buffer);

    for(i = 0 ; i < len ; i++)
      printf("%#x ", ((unsigned char *)buffer)[i]);

    printf("\n");
    printf("<%.*s>\n", len, buffer);
  }

  return 0;
}
#endif
