/*JS*********************************************************************
*
*    Program : CSTRINGPRINT
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Print a character array as a C-like string into a string.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            cStrSPrint(char *dest, const char *src, long len);

#ifndef CSTRPRINT
#define CSTRPRINT        cStrSPrint
#define CSTRARGS         char *s
#define PUTC(c)          *s++ = (c)
#undef CHECKERROR
#endif

/*JS*********************************************************************
*
*************************************************************************/

int CSTRPRINT(CSTRARGS, const char *_src, long _len)

/************************************************************************/
{
  int _outLen;

#ifdef CHECKERROR
#define DOPUTC(c)      if(DOPUTC2(c) == EOF) goto _error
#else
#define DOPUTC(c)      DOPUTC2(c)
#endif

#ifdef CSTRSTORECHAR
  char _c2[2];

  _c2[1] = '\0';
#define DOPUTC2(c)       (_c2[0] = (c), PUTC(_c2[0]))
#else
#define DOPUTC2(c)       PUTC(c)
#endif

  for(_outLen = 0 ; _len-- > 0 ; ++_src) {
static const char _InEsc[]  = "\\\"\n\t\r\f\v\b\a";
static const char _OutEsc[] = "\\\"ntrfvba";
    const char *_s;
    char _c;

    _c = *(unsigned char *)_src;
    if(_c && (_s = strchr(_InEsc, _c))) {
      /*  Print "\n" in case of newline char etc.  */
      DOPUTC('\\');
      _outLen++;
      _c = _OutEsc[_s - _InEsc];
    } else if(!isprint(_c)) { /*  Print octal representation  */
      int val, flag;

      DOPUTC('\\');
      ++_outLen;

      /*  If a digit follows or this is the last char, make it a full
       *   length octal representation
       */
      flag = isdigit(_src[1]) || _len == 0;
      val = ((int)_c & 0xc0) >> 6;
      if(val || flag) {
        DOPUTC('0' + val); ++_outLen;
	flag = 1;
      }
      val = ((int)_c & 0x38) >> 3;
      if(val || flag) {
        DOPUTC('0' + val); ++_outLen;
	flag = 1;
      }

      _c = '0' + ((int)_c & 0x07);
    }

    DOPUTC(_c); _outLen++;
  }
#undef DOPUTC
#undef DOPUTC2

  return _outLen;
#ifdef CHECKERROR
_error:
  return -1;
#endif
}
#undef CSTRPRINT
#undef CSTRARGS
#undef PUTC

#ifdef TEST
#define N    4096

char In[N], Out[4 * N + 1];

int main(int argc, char *argv[])
{
  int len, i;

  while((len = fread(In, 1, sizeof(In), stdin)) > 0) {
    i = cStrSPrint(Out, In, len);
    Out[i] = '\0';
    fputs(Out, stdout);
  }
  putchar('\n');

  return(0);
}
#endif
