/*JS*********************************************************************
*
*    Program : DBGOPTS
*    Language: ANSI - C
*    Author  : Joerg Schoen
*    Purpose : Parse debugging options
*
*              Copyright (C) 2005 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <ctype.h>
#include <string.h>

#include <jssubs.h>

/*********     DEFINES						*********/
#if 0
/*  START-DEFINITIONS  */
typedef struct {
  const char    *DBOP_Name;
  unsigned long  DBOP_Flag;
  void          *DBOP_Pointer;
  const char    *DBOP_Help;
} dbgopts_t;
/*  END-DEFINITIONS  */
#endif

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            parsedbgopts(const dbgopts_t *optdef,
				      const char *options, long *pvalue,
				      int mode);

/*JS*********************************************************************
*
*************************************************************************/

int parsedbgopts(const dbgopts_t *optdef, const char *options, long *pvalue,
		 int mode)

/************************************************************************/
{
  int count = 0;

  if(options == NULL)
    return 0;

  if(isdigit(*options)) {
    *pvalue = jstrtol2(options, NULL, 0);
    return 1;
  }

  *pvalue = 0;
  while(*options != '\0') {
    const dbgopts_t *currDef;
    const char *next;
    int len;

    while(isspace(*options)) ++options;

    if((len = jstrtoid(options)) == 0)
      EGOTO(error);

    if((next = jstrstartswith(options, "help")) != NULL &&
       (next - options) == len) {
      for(currDef = optdef ; currDef ;
	  currDef = (dbgopts_t *)currDef->DBOP_Pointer) {
	for( ; currDef->DBOP_Name != NULL ; ++currDef) {
	  fprintf(stderr, "'%s':\n", currDef->DBOP_Name);

	  if((next = currDef->DBOP_Help)) {
	    while(*next) {
	      const char *t;
	      int l;

	      if((t = strchr(next, '\n')) == NULL)
		l = strlen(next);
	      else
		l = t - next;

	      fprintf(stderr, "\t%.*s\n", l, next);
	      next += l;
	      if(*next == '\n')
		++next;
	    }
	  }

	  ++count;
	}

	if((mode & 7) == 1)
	  fprintf(stderr, "'all':\n\tAll of the above flags are set.\n");
      }

      if((mode & 7) == 3)
	fprintf(stderr, "'all':\n\tAll of the above flags are set.\n");
    } else if((mode & 1) &&
	      (next = jstrstartswith(options, "all")) != NULL &&
	      (next - options) == len) {
      for(currDef = optdef ; currDef ;
	  currDef = (dbgopts_t *)currDef->DBOP_Pointer) {
	for( ; currDef->DBOP_Name != NULL ; ++currDef) {
	  long flag, *ptr;

	  ptr = currDef->DBOP_Pointer ? (long *)currDef->DBOP_Pointer : pvalue;

	  if((flag = currDef->DBOP_Flag) >= 0)
	    *ptr |= flag; /*  set bits  */
	  else
	    *ptr &= flag; /*  clear bits  */

	}

	if(!(mode & 2))
	  break;
      }
    } else {
      for(currDef = optdef ; currDef ;
	  currDef = (dbgopts_t *)currDef->DBOP_Pointer) {
	for( ; currDef->DBOP_Name != NULL ; ++currDef) {
	  if((next = jstrstartswith(options, currDef->DBOP_Name)) != NULL &&
	     (next - options) == len)
	    break;
	}

	if(currDef->DBOP_Name) {
	  long flag, *ptr;

	  ptr = currDef->DBOP_Pointer ? (long *)currDef->DBOP_Pointer : pvalue;

	  /*  By this, we cannot use the highes bit (31)  */
	  if((flag = currDef->DBOP_Flag) >= 0)
	    *ptr |= flag; /*  set bits  */
	  else
	    *ptr &= flag; /*  clear bits  */

	  ++count;

	  break;
	}
      }
    }

    options += len;

    /*  valid delimiters  */
    while(*options && strchr(",; ", *options))
      ++options;
  }

  return count;
error:
  return -1;
}
