/*JS*********************************************************************
*
*    Program : DSLEEP
*    Language: ANSI-C using 'select'
*    Author  : Joerg Schoen
*    Purpose : Sleep function with double argument to wait for fractions
*              of seconds as well.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>

#ifndef _WIN32
# include <sys/types.h>
# include <sys/time.h>
# include <unistd.h>
#else
# include <windows.h>
#endif

/*********     DEFINES                                          *********/

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype double         dsleep(double secs);

/*JS*********************************************************************
*
*************************************************************************/

double dsleep(double secs)

/************************************************************************/
{
#ifndef _WIN32
  struct timeval tVal;

  tVal.tv_sec  = (long)secs;
  tVal.tv_usec = (time_t) (1000000 * (secs - (double)tVal.tv_sec));

  select(1, NULL, NULL, NULL, &tVal);
#else
  Sleep((DWORD) (1000.0 * secs));
#endif

  return 0.0;
}

#ifdef TEST
#include <stdlib.h>

#ifndef DEFTIME
# define DEFTIME   0.05
#endif

int main(int argc, char *argv[])
{
  double secs;
  const char *s = "-\\|/";
  int i;

  if(argc >= 2)
    secs = strtod(argv[1], NULL);
  else
    secs = DEFTIME;

  i = 0;
  while(1) {
    putchar('\r');
    putchar((unsigned char)s[i]);
    fflush(stdout);
    dsleep(secs);
    i = (i+1) % 4;
  }

  return(0);
}
#endif
