/*JS*********************************************************************
*
*    Program : FEXISTS
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Kind of alternative to 'access'
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <jsconfig.h>

#include <jssubs.h>

#include <jsllio.h>

#include <errno.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS  */
#define FILEEXISTS_TFILE           1
#define FILEEXISTS_TDIRECTORY      2
#define FILEEXISTS_TSYMLINK        3
#define FILEEXISTS_TFIFO           4

#define fileexists(name)          fileexists2(name, 0)
/*  END-DEFINITIONS  */

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            fileexists2(const char *name, int mode);

/*JS*********************************************************************
*   Check if a file exists. Returns '1' if it is a file, '2' in case of
*    a directory and '0' if it does not exist, -1 on error.
*   mode bit 0: If set, do not check for symlinks
*************************************************************************/

int fileexists2(const char *name, int mode)

/************************************************************************/
{
  struct stat stbuf;
  int rc;

#ifdef S_ISLNK
  if(!(mode & 1))
    rc = lstat(name, &stbuf);
  else
#endif
    rc = stat(name, &stbuf);

  if(rc == -1) {
    if(errno == ENOENT) {
      JSCLEARSYSERROR();
      return 0;
    }

    EGOTO(error);
  }

  if(S_ISDIR(stbuf.st_mode))
    return FILEEXISTS_TDIRECTORY;

#ifdef S_ISLNK
  if(S_ISLNK(stbuf.st_mode))
    return FILEEXISTS_TSYMLINK;
#endif

#ifdef S_ISFIFO
  if(S_ISFIFO(stbuf.st_mode))
    return FILEEXISTS_TFIFO;
#endif

  return FILEEXISTS_TFILE;
error:
  return -1;
}
