/*JS*********************************************************************
*
*    Program : FILE2BUFFER
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Read a file into a string.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
/*  Read chunks of this length  */
#define NEXTCHUNK     4000

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype char          *file2buffer(const char *name, long *pLen);
Prototype char          *ffile2buffer(FILE *fp, long *pLen);
Prototype char          *efile2buffer(extfile_t *efp, long *pLen);

/*JS*********************************************************************
 *
*************************************************************************/

char *file2buffer(const char *name, long *pLen)

/************************************************************************/
{
  FILE *fp;
  char *res;

  if((fp = fopen(name, "r")) == NULL)
    return NULL;

  res = ffile2buffer(fp, pLen);

  if(fclose(fp) == -1) {
    if(res) free(res);
    return NULL;
  }

  return res;
}

/*JS*********************************************************************
*   Read the file completely into an allocated buffer, which is returned.
*    *pLen contains the length of the data.
*************************************************************************/

char *ffile2buffer(FILE *fp, long *pLen)

/************************************************************************/
{
  char *buffer;
  long i, len, offset;

  buffer = NULL;
  len = offset = 0;
  do {
    len += NEXTCHUNK;
    if((buffer = (char *)realloc(buffer, len)) == NULL)
      EGOTO(error);

    if((i = fread(buffer + offset, sizeof(*buffer), NEXTCHUNK, fp)) < 0)
      EGOTO(error);

    offset += i;

  } while(i == NEXTCHUNK);

  /*  Reduce memory usage  */
  buffer = (char *)realloc(buffer, offset + 1);
  buffer[offset] = '\0';
  *pLen = offset;

  return buffer;
 error:
  if(buffer) free(buffer);
  return NULL;
}

/*JS*********************************************************************
*   Read the file completely into an allocated buffer, which is returned.
*    *pLen contains the length of the data.
*************************************************************************/

char *efile2buffer(extfile_t *efp, long *pLen)

/************************************************************************/
{
  char *buffer;
  long i, len, offset;

  buffer = NULL;
  len = offset = 0;
  do {
    len += NEXTCHUNK;
    if((buffer = (char *)realloc(buffer, len)) == NULL)
      EGOTO(error);

    if((i = efread(buffer + offset, NEXTCHUNK, efp)) < 0)
      EGOTO(error);

    offset += i;

  } while(i == NEXTCHUNK);

  /*  Reduce memory usage  */
  buffer = (char *)realloc(buffer, offset + 1);
  buffer[offset] = '\0';
  *pLen = offset;

  return buffer;
error:
  if(buffer)
    free(buffer);
  return NULL;
}
