/*JS*********************************************************************
*
*    Program : JSERRNO
*    Language: ANSI - C
*    Author  : Joerg Schoen
*    Purpose : Is linked for error setting.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

#include <jssubs.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS 1 */
#include <errno.h>

#define JSCLEARSYSERROR()       (errno = 0)

#define JSERROR_MAXLENGTH       1000

#ifndef _ERRMODE_STRING

# ifdef DEBUG_EGOTO_VERBOSE
#  define JSCLEARERROR()     (JSErrorStackLevel = 0,	\
			      JSErrNo = 0,		\
			      JSErrString[0] = '\0')
# else
#  define JSCLEARERROR()     (JSErrNo = 0,		\
			      JSErrString[0] = '\0')
# endif

# define JSSETERROR(code, off)						\
  (JSCLEARSYSERROR(),							\
   JSErrNo = (code) + (off), JSErrLine = 0,				\
   JSErrString[0] = '\0')

# define JSSETERROR2X(code, off, string)				\
  (JSCLEARSYSERROR(),							\
   JSErrNo = (code) + (off), JSErrLine = 0,				\
   strncpy(JSErrString, string, sizeof(JSErrString) - 1))
#else
# ifdef DEBUG_EGOTO_VERBOSE
#  define JSCLEARERROR()						\
  (JSErrorStackLevel = 0,						\
   JSErrNo = JSErrLine = 0, JSErrLabel = NULL,				\
   JSErrString[0] = '\0')
# else
#  define JSCLEARERROR()					\
  (JSErrNo = JSErrLine = 0, JSErrLabel = NULL,			\
   JSErrString[0] = '\0')
# endif

# define JSSETERROR(code, off)						\
  (JSCLEARSYSERROR(),							\
   JSErrLabel = #code, JSErrNo = 1 + (off),				\
   JSErrLine = 0, JSErrString[0] = '\0')

# define JSSETERROR2X(code, off, string)				\
  (JSCLEARSYSERROR(),							\
   JSErrLabel = #code, JSErrNo = 1 + (off),				\
   strncpy(JSErrString, string, sizeof(JSErrString) - 1))
#endif

# define JSSETERRSTRING(string)					\
  strncpy(JSErrString, string, sizeof(JSErrString) - 1)

# define JSSETSYSERRSTRING(string)			\
  (JSCLEARERROR(), JSSETERRSTRING(string))

#define JSSETERROR2(code, off, string)				\
  JSSETERROR2X(code, off, (string) ? (string) : "<NULL>")

#define JSSETERRFILE(file, line)		\
  JSSETERRFILE2(file, line, 1)

#define JSSETERRFILE2(file, line, column)				\
  (strncpy(JSErrFile, (file) ? (file) : "<NULL>", sizeof(JSErrFile) - 1), \
   JSErrLine = (line), JSErrColumn = (column))

/*  These are internal error codes for debugging  */
#define JSMINIERROR             10000
#define JSSETIERROR(code, off)   JSSETERROR(JSMINIERROR + (code), off)
#define JSSETIERROR2(code, off, s) JSSETERROR2(JSMINIERROR + (code), off, s)
/*  END-DEFINITIONS  */

#define Prototype extern
/*********     PROTOTYPES					*********/
Prototype int            JSErrNo;
Prototype char           JSErrString[JSERROR_MAXLENGTH + 1];

Prototype const char    *JSErrLabel;

Prototype int            JSErrLine, JSErrColumn;
Prototype char           JSErrFile[JS_FILENAME_MAX + 1];

Prototype int            JSErrorStackLevel;
Prototype char           JSErrorStack[100][JSERROR_MAXLENGTH];

/*********     GLOBAL VARIABLES                                 *********/
int JSErrNo;
char JSErrString[JSERROR_MAXLENGTH + 1];

#ifdef _ERRMODE_STRING
const char *JSErrLabel;
#endif

int JSErrLine, JSErrColumn;
char JSErrFile[JS_FILENAME_MAX + 1];

/*  We define this always - only some code might be compiled with 'DEBUG_EGOTO_VERBOSE'  */
int JSErrorStackLevel;
char JSErrorStack[100][JSERROR_MAXLENGTH];
