/*JS*********************************************************************
*
*    Program : JSPERROR
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Similar to "perror", but recognizes library errors as well
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <jsconfig.h>

#include <jssubs.h>

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype void           jsperror(const char *progname);

/*JS*********************************************************************
*   The difficult question is: which error code has precedence? Currently
*    this is the library.
*************************************************************************/

void jsperror(const char *progname)

/************************************************************************/
{
#ifdef DEBUG_EGOTO_VERBOSE
  int i;

  for(i = 0 ; i < JSErrorStackLevel ; ++i)
    fprintf(stderr, "ERROR %s\n", JSErrorStack[i]);
#endif

  /*  System errors have precedence  */
  if(JSErrNo != 0 || errno != 0) {
    fprintf(stderr, "%s:", progname);
    if(JSErrLine > 0)
      fprintf(stderr, " l. %d '%s'", JSErrLine, JSErrFile);

    fprintf(stderr, " %s", JSErrNo != 0 ?
	    jsStrError(JSErrNo) : strerror(errno));

    if(JSErrString[0])
      fprintf(stderr, " \"%s\"", JSErrString);

    putc('\n', stderr);

  } else
    /*  Leave handling of errno == 0 to perror :)  */
    perror(progname);

  return;
}
