/*JS*********************************************************************
*
*    Program : JSSTRERROR
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Get error string for internal error code.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <string.h>

#include <jsconfig.h>

#include <jssubs.h>

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype const char    *jsStrError(int err);

Prototype int            JSErrorMode;

/*********     GLOBAL VARIABLES                                 *********/
/*  Default is to include routine name in error string  */
int JSErrorMode = 1;

/*JS*********************************************************************
*
*************************************************************************/

const char *jsStrError(int err)

/************************************************************************/
{
  const char *string;
static char Buffer[500]; /*  should be enough for error messages  */

  if(err == 0)
    return "No error";

#ifdef _ERRMODE_STRING
  if(err == JSErrNo) {
    int i;

    for(i = 0 ; JSErrorsS[i] ; ++i)
      if(strcmp(JSErrorsS[i], JSErrLabel) == 0)
	break;

    err += JSErrorsI[i] - 1;
  }
#endif

#ifdef _ERR_BLOCKSIZE
  /*  Errors are allocated in blocks. Use indirection table to get the
   *   real offset
   */
  err = JSErrorsI[(err / _ERR_BLOCKSIZE) - 1] + (err % _ERR_BLOCKSIZE);
#endif

  if(JSErrors[0]) { /* ***  Read errors from file  *** */
    FILE *fp;

    if((fp = fopen(JSErrors[0], "r")) == NULL)
      goto nomsg;

    do {
      if(fgets(Buffer, sizeof(Buffer) - 1, fp) == NULL)
	goto nomsg;

    } while(err-- > 1);

    fclose(fp);

    /*  Delete trailing '\n'  */
    if(Buffer[0])
      Buffer[strlen(Buffer) - 1] = '\0';
    string = Buffer;

  } else if(err < _ERR_MAXERROR) {
    string = JSErrors[err];

  } else if(err >= JSMINIERROR) {
    sprintf(Buffer, "INTERNAL ERROR #%d", err - JSMINIERROR);
    string = Buffer;

  } else
    goto nomsg;

  if(JSErrorMode == 0) {
    const char *string2;

    /*  Strip routine name  */
    if((string2 = strchr(string, ':')))
      string = string2 + 1;
  }

  return string;
nomsg:
  return "FATAL: Error message not available";
}
