/*JS*********************************************************************
*
*    Program : MILLITIME
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Get absolute time with milliseconds resolution
*
*              Copyright (C) 2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#ifndef _WIN32
#include <sys/time.h>
#else
#include <windows.h>
#endif

#include <jssubs.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS  */
#include <time.h>
/*  END-DEFINITIONS  */

#ifdef _WIN32
#if defined(_MSC_VER) || defined(_MSC_EXTENSIONS)
  #define DELTA_EPOCH_IN_USEC  11644473600000000Ui64
#else
  #define DELTA_EPOCH_IN_USEC  11644473600000000ULL
#endif
#endif

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            getMicroTime(time_t *pTime, long *pMicroTime);

/*JS*********************************************************************
*
*************************************************************************/

int getMicroTime(time_t *pTime, long *pMicroTime)

/************************************************************************/
{
#ifndef _WIN32
    struct timeval tStamp;

    if(gettimeofday(&tStamp, NULL) == -1)
      return -1;

    *pTime = tStamp.tv_sec;
    *pMicroTime = tStamp.tv_usec;

    return 0;
#else
    FILETIME ftime;
    __int64 timeval;

    GetSystemTimeAsFileTime(&ftime);

    timeval = ftime.dwHighDateTime << 32;
    timeval |= ftime.dwLowDateTime;
    timeval /= 10; /*  from nanosec to usec  */
    timeval -= DELTA_EPOCH_IN_USEC;

    *pTime      = timeval / 1000000;
    *pMicroTime = timeval % 1000000;

    return 0;
#endif
}
