/*JS*********************************************************************
*
*    Program : MKBACKUP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Makes a backup file.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <string.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS  */
#define mkBackupFile(fileName, mode)   mkBackupFile2(fileName, NULL, mode)
/*  END-DEFINITIONS  */

#define MAX_BACKUPS      100

#define BACKUP_PATTERN   ".~%d~"

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype const char    *mkBackupFile2(const char *fileName, char *oName,
				       int mode);
Prototype int            mkBackupName(char *name, int mlen, const char *ext);

/*JS*********************************************************************
*   Move fileName to backup file in emacs style ("~<num>~"). First loop
*    to find unused number. If oName is non-NULL, the backup name is
*    stored there.
*************************************************************************/

const char *mkBackupFile2(const char *fileName, char *oName, int mode)

/************************************************************************/
{
  int count;
static char Buffer[JS_FILENAME_MAX];

  if(oName == NULL) oName = Buffer;

  strcpy(oName, fileName);

  if((count = mkBackupName(oName, MAX_BACKUPS, NULL)) == -1)
    goto error;

  if(count > 0 && mode && rename(fileName, oName) == -1) goto error;

  return oName;
error:
  return NULL;
}

/* ERROR-DEFINITIONS from mkBackupName label _ERR_MKBACKUPNAME
   Too many backup files
*/

/*JS*********************************************************************
*
*************************************************************************/

int mkBackupName(char *name, int mlen, const char *ext)

/************************************************************************/
{
  int i, l, count;

  l = strlen(name);

  /*  Find a file that does not exist  */
  for(count = 0 ;; ++count) {
    if(count > 0) {
      if(ext) {
	for(i = 0 ; ext[i] ; ++i) {
	  if(ext[i] == '\b') {
	    if(l > 0) --l;
	    continue;
	  }
	  name[l++] = ext[i];
	}
	name[l] = '\0';
      } else {
	/*  emacs style  */
	sprintf(name + l, BACKUP_PATTERN, count);
      }
    }

    if((i = fileexists(name)) == -1)
      return -1;

    if(i == 0)
      break;

    if(ext ? l >= (mlen - 1) : count >= mlen) {
      JSSETERROR(_ERR_MKBACKUPNAME, 0);
      return -1;
    }
  }

  return count;
}

#ifdef TEST
#include <stdlib.h>

int main(int argc, char *argv[])
{
  char name[JS_FILENAME_MAX];
  const char *ext;
  int ret, mlen;

  strcpy(name, argv[1]);

  if(argc == 4) {
    ext = argv[2];
    mlen = atoi(argv[3]);
  } else {
    ext = NULL;
    mlen = 5;
  }

  ret = mkBackupName(name, mlen, ext);

  printf("'%s' -> '%s' %d\n", argv[1], name, ret);

  return 0;
}
#endif
