/*JS*********************************************************************
*
*    Program : SHARESTRDUP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Similar to 'strdup', but allows sharing of identical
*              strings.
*
*              Copyright (C) 2003 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <string.h>

/*********     DEFINES						*********/
#ifndef JSHST_TABSIZE
/*  START-DEFINITIONS  */
#define JSHST_TABSIZE  251   /*  prime number  */

typedef struct _jsshareident_t _jsshareident_t;
struct _jsshareident_t {
  _jsshareident_t *JSHID_Next;
  char             JSHID_String[1];
};

typedef struct {
  _jsshareident_t *JSHST_HashTable[JSHST_TABSIZE];
} jssharestr_t;
/*  END-DEFINITIONS  */
#endif

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype jssharestr_t  *sharestrinit(jssharestr_t *ss);
Prototype void           sharestrdeinit(jssharestr_t *ss, int flag);

Prototype const char    *sharestrdup(jssharestr_t *ss, const char *string);
static unsigned int      stringhash(const char *string,
				    unsigned int tabSize);

/*JS*********************************************************************
*
*************************************************************************/

jssharestr_t *sharestrinit(jssharestr_t *ss)

/************************************************************************/
{
  int i;

  if(ss == NULL && (ss = malloc(sizeof(*ss))) == NULL)
    return NULL;

  for(i = 0 ; i < JSHST_TABSIZE ; ++i)
    ss->JSHST_HashTable[i] = NULL;

  return ss;
}

/*JS*********************************************************************
*
*************************************************************************/

void sharestrdeinit(jssharestr_t *ss, int flag)

/************************************************************************/
{
  int i;

  for(i = 0 ; i < JSHST_TABSIZE ; ++i) {
    _jsshareident_t *id, *next;

    for(id = ss->JSHST_HashTable[i] ; id ; id = next) {
      next = id->JSHID_Next;
      free(id);
    }
  }

  if(flag) free(ss);
}

/*JS*********************************************************************
*
*************************************************************************/

const char *sharestrdup(jssharestr_t *ss, const char *string)

/************************************************************************/
{
  _jsshareident_t *id;
  int hash;

  hash = stringhash(string, JSHST_TABSIZE);
  for(id = ss->JSHST_HashTable[hash] ; id ; id = id->JSHID_Next)
    if(strcmp(id->JSHID_String, string) == 0)
      return id->JSHID_String;

  if((id = (_jsshareident_t *)malloc(sizeof(*id) + strlen(string)))
     == NULL)
    return NULL;

  strcpy(id->JSHID_String, string);
  id->JSHID_Next = ss->JSHST_HashTable[hash];
  ss->JSHST_HashTable[hash] = id;

  return id->JSHID_String;
}

/*JS*********************************************************************
*
*************************************************************************/

static unsigned int stringhash(const char *string, unsigned int tabSize)

/************************************************************************/
{
  unsigned int hash, count;

#define HASH_BITS   32
#define HASH_SHIFT  3

  /*  Generate polynomial   */
  for(hash = count = 0 ; *string ; ++string) {
    hash = (hash << HASH_SHIFT) | (hash >> (HASH_BITS - HASH_SHIFT));
    hash += *(const unsigned char *)string;
    count++;
  }

  return (hash ^ count) % tabSize;
}
#undef HASH_BITS
#undef HASH_SHIFT
