/*JS*********************************************************************
*
*    Program : STRCOMPARE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Useful string compare function.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <ctype.h>
#include <string.h>
#include <stddef.h>

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype const char    *jstrstartswith(const char *input, const char *string);
Prototype const char    *jstristartswith(const char *input, const char *string);
Prototype int            jstrendswith(const char *input, const char *string);

/*JS*********************************************************************
*   Compares 'string' totally with the beginning of 'input'. Returns NULL
*    when the comparison fails, otherwise returns the first position after
*    the string in 'input'.
*************************************************************************/

const char *jstrstartswith(const char *input, const char *string)

/************************************************************************/
{
  while(*string) {
    if(*string++ !=  *input++)
      return NULL;
  }

  return input;
}

/*JS*********************************************************************
*   Compares 'string' with beginning of 'input' ignoring upper/lower case.
*    Upper case characters in 'string' mark a valid abbreviation (up to
*    the character before the upper case one) of the word. On failure
*    returns NULL, otherwise first position after 'string' in the input.
*************************************************************************/

const char *jstristartswith(const char *input, const char *string)

/************************************************************************/
{
  while(*string) {
    if(tolower(*(unsigned char *)string) != tolower(*(unsigned char *)input))
      return isupper(*(unsigned char *)string) ? input : (const char *)NULL;

    input++;
    string++;
  }

  return input;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstrendswith(const char *input, const char *string)

/************************************************************************/
{
  int off;

  off = strlen(input) - strlen(string);

  return off >= 0 && strcmp(input + off, string) == 0;
}
