/*JS*********************************************************************
*
*    Program : STRTOL2
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Enhancement to "strtol".
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype long           jstrtol2(const char *input, char **endp, int base);
Prototype unsigned long  jstrtoul2(const char *input, char **endp, int base);

/*JS*********************************************************************
*   Same as strtol, but understands additionally binary values starting
*    with '%'.
*************************************************************************/

long jstrtol2(const char *input, char **endp, int base)

/************************************************************************/
{
  if(base == 0) {
    const char *string;

    /*   Drop spaces     */
    string = input;
    while(isspace(*(unsigned char *)string)) string++;

    /*   Binary value?    */
    if(*string++ == '%') {
      long value;

      value = strtol(string, endp, 2);

      /*  Failure?  */
      if(endp && ((const char *)*endp) == string) *endp = (char *)input;

      return value;
    }
  }

  return strtol(input, endp, base);
}

/*JS*********************************************************************
*   Same as strtoul, but understands additionally binary values starting
*    with '%'.
*************************************************************************/

unsigned long jstrtoul2(const char *input, char **endp, int base)

/************************************************************************/
{
  if(base == 0) {
    const char *string;

    /*   Drop spaces     */
    string = input;
    while(isspace(*(unsigned char *)string)) string++;

    /*   Binary value?    */
    if(*string++ == '%') {
      unsigned long value;

      value = strtoul(string, endp, 2);

      /*  Failure?  */
      if(endp && ((const char *)*endp) == string) *endp = (char *)input;

      return value;
    }
  }

  return strtoul(input, endp, base);
}
