/*JS*********************************************************************
*
*    Program : TOKPATMAT
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Pattern matcher for C-tokens
*
*              Copyright (C) 2005 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
#ifndef TOMASD_TYPE_TOKENS
/*  START-DEFINITIONS 2 */
/* ***  A single state description of the finite state machine  *** */
typedef struct tomastatedescr_t tomastatedescr_t;
struct tomastatedescr_t {
    unsigned int TOMASD_Type;
#define TOMASD_TYPE_TOKENS           1
#define TOMASD_TYPE_CHOICE           2
#define TOMASD_TYPE_REPEAT           3

#define TOMASD_TYPE_EXPR             4
#define TOMASD_TYPE_EXPRSTART        5
#define TOMASD_TYPE_EXPREND          6

#define TOMASD_TYPE_STMT             7
#define TOMASD_TYPE_STMTSTART        8
#define TOMASD_TYPE_STMTEND          9

#define TOMASD_TYPE_IDENT           10
#define TOMASD_TYPE_NUMBER          11
#define TOMASD_TYPE_STRING          12

#define TOMASD_TYPE_LINENR          13
#define TOMASD_TYPE_COUNTER         14

    tomastatedescr_t *TOMASD_Next;
    union {
	/*  For TOMASD_TYPE_TOKENS. The indexes refer to TOPA_StateQueue  */
	struct {
	    unsigned int tomasd_TokStart, tomasd_TokEnd;

	} tomasd_us1;
#define TOMASD_TokStart     tomasd_u.tomasd_us1.tomasd_TokStart
#define TOMASD_TokEnd       tomasd_u.tomasd_us1.tomasd_TokEnd

	/*  For TOMASD_TYPE_CHOICE and TYPE_REPEAT  */
	struct {
	    tomastatedescr_t *tomasd_Sub;
	    unsigned int tomasd_ILoopVar;
	    unsigned int tomasd_MinLoop, tomasd_MaxLoop;
	} tomasd_us2;
#define TOMASD_Sub          tomasd_u.tomasd_us2.tomasd_Sub
#define TOMASD_ILoopVar     tomasd_u.tomasd_us2.tomasd_ILoopVar
#define TOMASD_MinLoop      tomasd_u.tomasd_us2.tomasd_MinLoop
#define TOMASD_MaxLoop      tomasd_u.tomasd_us2.tomasd_MaxLoop

	/*  For TOMASD_TYPE_EXPR + TOMASD_TYPE_STMT + TOMASD_TYPE_IDENT +
	 *   TOMASD_TYPE_NUMBER + TOMASD_TYPE_STRING
	 */
	unsigned int tomasd_ExprIndex;
#define TOMASD_ExprIndex     tomasd_u.tomasd_ExprIndex

    } tomasd_u;

    /*  Used for subordinates of TYPE_CHOICE to connect them  */
    tomastatedescr_t *TOMASD_Connect;
};

/* ***  Runtime structure to describe a state where we are  *** */
/*  To keep what a single variable matches  */
typedef struct {
    unsigned int MASTV_Start, MASTV_End;

} topavarmatch_t;

typedef struct matcherstate_t matcherstate_t;
struct matcherstate_t {
    matcherstate_t *MAST_Next;

    unsigned int MAST_TokenIndex; /*  refers to TOPA_InputQueue  */
    tomastatedescr_t *MAST_StateDescription;

    unsigned int *MAST_LoopCounter;
    topavarmatch_t MAST_VariableMatch[1];
};

/* ***  The overall structure  *** */
typedef struct {
    const char *TOPD_Name;

    tomastatedescr_t *TOPD_StartState;

    /*  Indexes into TOPA_StateQueue that contain the replacement or -1 if none  */
    unsigned int TOPD_Start, TOPD_End;

} tokpattdef_t;

#define TOPA_MEMORY_SIZE      200
typedef struct topamemory_t topamemory_t;
struct topamemory_t {
    topamemory_t *TOPM_Next;

    unsigned int TOPM_Index;
    tomastatedescr_t TOPM_Memory[TOPA_MEMORY_SIZE];
};

typedef struct {
    extfile_t *TOPA_Input;
    jctokfifo_t TOPA_InputQueue;

    unsigned int TOPA_NVariables;
    unsigned int TOPA_NCounters;
    unsigned int TOPA_MaxLoopCounter;

    tokpattdef_t *TOPA_Definitions;
    unsigned int TOPA_NDefinitions;

    jctokfifo_t TOPA_StateQueue; /*  for TOMASD_TYPE_TOKENS  */

    /*  Linked list to recycle matcher states  */
    matcherstate_t *TOPA_MatcherStates;

    /*  Automatically incremented counters  */
    unsigned long *TOPA_Counters;

    /*  Memory chunks for states  */
    topamemory_t *TOPA_StateMemory;

} tokpattmatch_t;
/*  END-DEFINITIONS  */
#endif

#define GETCTOKEN(efp) getCToken(efp, NULL, GCTK_CPPCOMMENT | GCTK_EXTENSIONS)

/*  Special tokens for substitute  */
#define TOPA_TOKEN_FMASK           ((1L << 16) - 1)
#define TOPA_TOKEN_TMASK           (0xff << 16)

#define TOPA_TOKEN_EXPRESSION      (1L << 16)
#define TOPA_TOKEN_TRIMEXPRESSION  (2L << 16)
#define TOPA_TOKEN_LINENR          (5L << 16)
#define TOPA_TOKEN_COUNTER         (6L << 16)

typedef struct {
    unsigned int Flags;
    tomastatedescr_t *State;
    tomastatedescr_t *Head;
    tomastatedescr_t *LNext;

} topaparserstack_t;

typedef struct {
    unsigned int Type;
    const char *Name;

} topaexprname_t;

typedef struct {
    tokpattmatch_t *Topa;

    topaexprname_t *VarNames;
    unsigned int NVarNames, MVarNames;

    const char **CounterNames;
    unsigned int NCounterNames, MCounterNames;

    topaparserstack_t *Stack;
    unsigned int NStack, MStack;

} topacraux_t;

/*  Top level tags  */
#define TAGNAME_COMMANDS     "commands"
#define TAGNAME_MATCHER      "matcher"
#define TAGNAME_PATTERN      "pattern"
#define TAGNAME_SUBSTITUTE   "substitute"

/*  Regular expression stuff  */
#define TAGNAME_REPEAT       "repeat"
#define ATTRNAME_MIN         "min"
#define ATTRNAME_MAX         "max"
#define ATTRNAME_COUNT       "count"

#define TAGNAME_OPTIONAL     "optional"
#define TAGNAME_CHOICE       "choice"
#define TAGNAME_BRANCH       "branch"

/*  terminal nodes  */
#define TAGNAME_IDENTIFIER   "identifier"
#define TAGNAME_NUMBER       "number"
#define TAGNAME_STRING       "string"
#define TAGNAME_EXPRESSION   "expression"
#define TAGNAME_STATEMENT    "statement"

#define ATTRNAME_NAME        "name"
#define ATTRNAME_TRIM        "trim"

/*  special nodes for substitute  */
#define TAGNAME_LINENR       "line"
#define TAGNAME_COUNTER      "counter"

/*  Whenever the input queue has this many unused tokens, we delete them  */
#define GC_INTERVAL          1000

/*  TRACE-DEFINITION symbol TRACETYPE_TOKPATMAT name TOPA */

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype tokpattmatch_t *topacreate(const char *fileName);
static int               analyzeVariableTag(topacraux_t *paux, saxparse_t *sapa,
					     const char **pName, int mode);
static tomastatedescr_t *getExpressionState(topacraux_t *paux, unsigned int type,
					  saxparse_t *sapa);
static int               getCounterIndex(topacraux_t *paux, const char *name);
static int               findExpressionIndex(topacraux_t *paux,
					     unsigned int type,
					     const char *name);

static tomastatedescr_t *getRepeatState(topacraux_t *paux, unsigned int type,
					unsigned int loopVar, saxparse_t *sapa);

static tomastatedescr_t *getTokenState(topacraux_t *paux,
				       saxparse_t *sapa);
static int               addCTokens(topacraux_t *paux, const char *string);
static void              connectNext(tomastatedescr_t *list,
				     tomastatedescr_t *next);
static tomastatedescr_t *linkTail(tomastatedescr_t *list,
				  tomastatedescr_t *node);
static int               addDefinition(tokpattmatch_t *topa, const char *name,
				       tomastatedescr_t *startState,
				       unsigned int start, unsigned int end);
static void              cleanaux(topacraux_t *aux, int flag);

Prototype void           topadelete(tokpattmatch_t *topa);

Prototype int            topaprocess(tokpattmatch_t *topa, extfile_t *efp,
				     const char *fileName, FILE *fpOut,
				     const char **matchers, int mode);
static int               findMatch(tokpattmatch_t *topa, jctokfifo_t *iQueue,
				   tomastatedescr_t *start,
				   unsigned int startIndex, matcherstate_t **pState,
				   int mode);
#ifdef TRACETYPE_TOKPATMAT
static void              showStates(const tokpattmatch_t *topa);
static void              printTokens(const jctokfifo_t *queue, int start, int end);
static int               findStateNr(const tokpattmatch_t *topa,
				     const tomastatedescr_t *state);
static void              showMatcherStates(const tokpattmatch_t *topa,
					   const matcherstate_t *head,
					   const jctokfifo_t *iQueue);
#endif
static int               putState(tokpattmatch_t *topa, matcherstate_t **pHead,
				  matcherstate_t *state);

static int               checkTokens(tokpattmatch_t *topa, jctokfifo_t *iQueue,
				     unsigned int position,
				     unsigned int fifoStart, unsigned int fifoEnd,
				     int mode);
static int               getExprMatcher(tokpattmatch_t *topa, jctokfifo_t *iQueue,
					matcherstate_t *current,
					unsigned int varIndex,
					matcherstate_t **pHead, int mode);

static int               getTokenMatcher(tokpattmatch_t *topa, jctokfifo_t *iQueue,
					 matcherstate_t *current, int varIndex,
					 matcherstate_t **pHead, int mode);

static tomastatedescr_t *getStateDescription(tokpattmatch_t *topa,
					     unsigned int type);

static matcherstate_t   *getMatcherState(tokpattmatch_t *topa,
					 tomastatedescr_t *stateDescr,
					 int tokenIndex, matcherstate_t *old);
static matcherstate_t   *freeMatcherState(tokpattmatch_t *topa,
					  matcherstate_t *state);

static const jsctoken_t *indexedToken(tokpattmatch_t *topa,
				      jctokfifo_t *iQueue, int tokIndex);

/*********     GLOBAL VARIABLES                                 *********/

/* ERROR-DEFINITIONS from topacreate label _ERR_TOPACREATE
   Syntax error
   Missing XML version
   Illegal XML element
   Invalid nested XML element
   Missing variable name in substitution
   Unknown variable name in substitution
   Unknown attribute for expression tag
   Non numeric attribute
   Illegal nesting of expression tags
   Empty repeat body
   Unknown attribute for matcher tag
   Attributes 'count' and 'min', 'max' are mutually exclusive
*/

/*JS*********************************************************************
*
*************************************************************************/

tokpattmatch_t *topacreate(const char *fileName)

/************************************************************************/
{
    FILE *fpXml;
    tomastatedescr_t *head, *lNext;
    tokpattmatch_t *topa;
    const char *string, *matcherName;
    saxparse_t *sapa;
    topacraux_t aux;
    unsigned int type, parserState, linePos[2];
    unsigned int startIndex, endIndex, loopVar, sapamode;

    CASSERT(TOPA_TOKEN_EXPRESSION >= CTK_USERTOKEN);

    if((fpXml = fopen(fileName, "r")) == NULL)
	EGOTO(error0);

    /*  Allocate main structure and preset  */
    if((topa = (tokpattmatch_t *)malloc(sizeof(*topa))) == NULL ||
       jctokcreate(&(topa->TOPA_StateQueue), 0) == NULL ||
       jctokcreate(&(topa->TOPA_InputQueue), 0) == NULL)
	EGOTO(error);

    topa->TOPA_Definitions = NULL;
    topa->TOPA_NDefinitions = 0;
    topa->TOPA_MaxLoopCounter = 0;
    topa->TOPA_NVariables = 0;
    topa->TOPA_NCounters = 0;
    topa->TOPA_MatcherStates = NULL;
    topa->TOPA_StateMemory = NULL;
    topa->TOPA_Counters = NULL;

    sapamode = /*SAXTYPE_FTRIMSPACES |*/ SAXTYPE_FSKIPSPACES | SAXTYPE_FUPDCOLUMN;
    if((sapa = saxparsecreate(sapamode)) == NULL)
	EGOTO(error);

    linePos[0] = linePos[1] = 1;

    /*  Read prologue. We don't care about the XML version and other stuff  */
    if((type = saxparse(sapa, fpXml, linePos)) != SAXTYPE_XMLVERSION) {
	JSSETERROR(_ERR_TOPACREATE, 1);
	EGOTO(error);
    }

    aux.Topa = topa;
    aux.VarNames = NULL;
    aux.NVarNames = aux.MVarNames = 0;
    aux.CounterNames = NULL;
    aux.NCounterNames = aux.MCounterNames = 0;
    aux.Stack = NULL;
    aux.NStack = aux.MStack = 0;

#define stackisempty()  (aux.NStack == 0)
#define tos()           (&(aux.Stack[aux.NStack - 1]))

    matcherName = NULL;
    head = lNext = NULL;
    loopVar = 0;
    startIndex = endIndex = -1;
    string = NULL; /*  For an empty file, we end up at syntaxError where string is needed */
    parserState = 0;
    while((type = saxparse(sapa, fpXml, linePos)) != SAXTYPE_EOF) {
	tomastatedescr_t *current;
	const char **attrs;
	unsigned int i, nattrs, flag;

	if(type == SAXTYPE_COMMENT || type == SAXTYPE_PROCINSTR || type == SAXTYPE_DOCTYPE)
	    continue;

#ifdef TRACETYPE_TOKPATMAT
	if(jstracelevel(TRACETYPE_TOKPATMAT) > 4) {
	    jstrace(TRACETYPE_TOKPATMAT, 4,
		    "  PARSE %2d INPUT %d '%s'; HEAD #%d; LNEXT",
		    parserState, type, saxparsegetname(sapa), findStateNr(topa, head));

	    for(current = lNext ; current ; current = current->TOMASD_Next)
		jstrace(TRACETYPE_TOKPATMAT, 4,
			" (#%d: T%ld)", findStateNr(topa, current), current->TOMASD_Type);
	    jstrace(TRACETYPE_TOKPATMAT, 4,
		    "\n");

	    for(i = 0 ; i < aux.NStack ; ++i) {
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"    STACK %d: FL %d, ST %d HD %d LN %d\n", i, aux.Stack[i].Flags,
			findStateNr(topa, aux.Stack[i].State),
			findStateNr(topa, aux.Stack[i].Head),
			findStateNr(topa, aux.Stack[i].LNext));
	    }
	}
#endif

	string = saxparsegetname(sapa);

#define FLAG_NO_LNEXT    1
#define FLAG_PUSH        2
#define FLAG_PUSHHDLN    4
#define FLAG_POP         8
#define FLAG_USER      256
	flag = 0;

	current = NULL;

	switch(parserState) {
	case 0: /* ***  looking for <commands>  *** */
	    if(type == SAXTYPE_STARTELEMENT) {
		if(strcmp(string, TAGNAME_COMMANDS) == 0) {
		    parserState = 5;
		    break;
		}
	    }

	    JSSETERROR2(_ERR_TOPACREATE, 2, string);
	    EGOTO(error);

	case 5: /* ***  looking for <matcher> *** */
	    if(type == SAXTYPE_STARTELEMENT) {
		if(strcmp(string, TAGNAME_MATCHER) == 0) {
		    attrs = saxparsegetattrs(sapa);
		    nattrs = saxparsegetnattrs(sapa);
		    matcherName = NULL;
		    for(i = 0 ; i < nattrs ; i += 2) {
			if(strcmp(attrs[i], ATTRNAME_NAME) == 0) {
			    if((matcherName = strdup(attrs[i + 1])) == NULL)
				EGOTO(error);
			} else {
			    JSSETERROR2(_ERR_TOPACREATE, 10, attrs[i]);
			    EGOTO(error);
			}
		    }

		    /*  Preset for next round  */
		    cleanaux(&aux, 0);
		    head = lNext = NULL;
		    loopVar = 0;
		    parserState = 10;
		    break;
		}
	    }

	    if(type == SAXTYPE_STOPELEMENT) {
		if(strcmp(string, TAGNAME_COMMANDS) == 0) {
		    parserState = 6;
		    break;
		}
	    }

	    EGOTO(syntaxError);

	case 6: /* *** having seen </matcher>, wait for the end  *** */
	    /*  We should never reach here  */
	    EGOTO(syntaxError);

	case 10: /* ***  Look for patterns + substitutes  *** */
	    if(type == SAXTYPE_STARTELEMENT) {
		if(head == NULL && strcmp(string, TAGNAME_PATTERN) == 0) {
		    parserState = 15;
		    break;
		}

		if(startIndex == (unsigned int)-1 &&
		   strcmp(string, TAGNAME_SUBSTITUTE) == 0) {
		    attrs = saxparsegetattrs(sapa);
		    nattrs = saxparsegetnattrs(sapa);

		    for(i = 0 ; i < nattrs ; i += 2)
			if(strcmp(attrs[i], ATTRNAME_TRIM) == 0) {
			    unsigned int value = atoi(attrs[i + 1]);

			    if(value == 0)
				sapa->SAPA_Mode &= ~SAXTYPE_FTRIMSPACES;

			} else {
			    JSSETERROR2(_ERR_TOPACREATE, 6, attrs[i]);
			    EGOTO(error);
			}

		    startIndex = topa->TOPA_StateQueue.TKFF_Index;
		    parserState = 20;
		    break;
		}
	    }

	    if(type == SAXTYPE_STOPELEMENT) {
		if(strcmp(string, TAGNAME_MATCHER) == 0) {
		    if(head != NULL) {
			/*  OK, there is another non-empty pattern, so add it
			 *   startIndex and endIndex may be -1
			 */
			if(addDefinition(topa, matcherName, head, startIndex, endIndex) == -1)
			    EGOTO(error);

			/*  Check for maximum  */
			if(aux.NVarNames > topa->TOPA_NVariables)
			    topa->TOPA_NVariables = aux.NVarNames;

			if(aux.NCounterNames > topa->TOPA_NCounters)
			    topa->TOPA_NCounters = aux.NCounterNames;

			head = NULL;
			startIndex = endIndex = -1;
		    }

		    parserState = 5;
		    break;
		}
	    }

	    if(type == SAXTYPE_CHARACTER) /*  ignore any characters here  */
		/*  FIXME: Maybe better bark about useless characters?  */
		break;

	    EGOTO(syntaxError);

	case 15: /* ***  Parsing content of <pattern> tag  *** */
	    if(type == SAXTYPE_STARTELEMENT) {
		if(strcmp(string, TAGNAME_CHOICE) == 0) {
		    if((current = getStateDescription(topa, TOMASD_TYPE_CHOICE)) == NULL)
			EGOTO(error);

		    /*  In the TOMASD_Sub field, we collect the 'lNext' chain for
		     *   all branches
		     */
		    current->TOMASD_Sub = NULL; /*  no branch yet  */

		    flag = FLAG_PUSHHDLN | FLAG_NO_LNEXT;

		    parserState = 16;
		    break;
		}

		if(strcmp(string, TAGNAME_REPEAT) == 0) {
		    if((current = getRepeatState(&aux, TOMASD_TYPE_REPEAT,
						 loopVar, sapa)) == NULL)
			EGOTO(error);

		    current->TOMASD_Sub = NULL; /*  no content yet  */

		    /*  A loop appears as a normal element, so also add it to lNext  */
		    flag = FLAG_PUSHHDLN;

		    if(++loopVar > topa->TOPA_MaxLoopCounter)
			topa->TOPA_MaxLoopCounter = loopVar;

		    break;
		}

		if(strcmp(string, TAGNAME_OPTIONAL) == 0) {
		    if(saxparsegetnattrs(sapa) > 0)
			/*  We do not accept attributes for <optional> tags  */
			EGOTO(syntaxError);

		    if((current = getRepeatState(&aux, TOMASD_TYPE_REPEAT,
						 loopVar, NULL)) == NULL)
			EGOTO(error);
		    current->TOMASD_MaxLoop = 1;

		    current->TOMASD_Sub = NULL; /*  no content yet  */

		    /*  A loop appears as a normal element, so also add it to lNext  */
		    flag = FLAG_PUSHHDLN | FLAG_USER;

		    if(++loopVar > topa->TOPA_MaxLoopCounter)
			topa->TOPA_MaxLoopCounter = loopVar;

		    break;
		}

		if(strcmp(string, TAGNAME_EXPRESSION) == 0) {
		    if((current = getExpressionState(&aux, TOMASD_TYPE_EXPRSTART,
						     sapa)) == NULL)
			EGOTO(error);

		    flag = FLAG_PUSH;

		    break;
		}

		if(strcmp(string, TAGNAME_STATEMENT) == 0) {
		    if((current = getExpressionState(&aux, TOMASD_TYPE_STMTSTART,
						     sapa)) == NULL)
			EGOTO(error);

		    flag = FLAG_PUSH;

		    break;
		}
	    }

	    if(type == SAXTYPE_STOPELEMENT) {
		if(strcmp(string, TAGNAME_PATTERN) == 0) {
		    if(!stackisempty()) {
			JSSETERROR2(_ERR_TOPACREATE, 3, string);
			EGOTO(error);
		    }

		    /*  Connect them to the "final" state (== NULL)  */
		    connectNext(lNext, NULL);

		    parserState = 10;
		    break;
		}

		if(strcmp(string, TAGNAME_BRANCH) == 0) {
		    if(stackisempty() ||
		       tos()->State->TOMASD_Type != TOMASD_TYPE_CHOICE) {
			JSSETERROR2(_ERR_TOPACREATE, 3, string);
			EGOTO(error);
		    }

		    /*  Append the lNext chain of this </or> to the next levels  */
		    tos()->LNext = linkTail(tos()->LNext, lNext);
		    lNext = NULL;

		    /*  Hang element in Sub chain of branch  */
		    if((current = tos()->State->TOMASD_Sub) == NULL) {
			tos()->State->TOMASD_Sub = head;

		    } else {
			while(current->TOMASD_Connect)
			    current = current->TOMASD_Connect;

			current->TOMASD_Connect = head;
		    }
		    head->TOMASD_Connect = NULL;

		    current = NULL;
		    head = NULL;

		    parserState = 16;
		    break;
		}

		if(strcmp(string, TAGNAME_REPEAT) == 0) {
		    if(stackisempty() ||
		       tos()->State->TOMASD_Type != TOMASD_TYPE_REPEAT ||
		       (tos()->Flags & FLAG_USER)) {
			JSSETERROR2(_ERR_TOPACREATE, 3, string);
			EGOTO(error);
		    }

		    if(head == NULL) {
			JSSETERROR(_ERR_TOPACREATE, 9);
			EGOTO(error);
		    }

		    current = tos()->State;
		    current->TOMASD_Sub = head;
		    head->TOMASD_Connect = NULL;

		    /*  Produce a loop by connecting the tail of the loop
		     *   content with the loop element itself
		     */
		    connectNext(lNext, current);

		    --loopVar;
		    current = NULL;
		    flag = FLAG_POP;

		    break;
		}

		if(strcmp(string, TAGNAME_OPTIONAL) == 0) {
		    if(stackisempty() ||
		       tos()->State->TOMASD_Type != TOMASD_TYPE_REPEAT ||
		       !(tos()->Flags & FLAG_USER)) {
			JSSETERROR2(_ERR_TOPACREATE, 3, string);
			EGOTO(error);
		    }

		    if(head == NULL) {
			JSSETERROR(_ERR_TOPACREATE, 9);
			EGOTO(error);
		    }

		    current = tos()->State;
		    current->TOMASD_Sub = head;
		    head->TOMASD_Connect = NULL;

		    /*  Produce a loop by connecting the tail of the loop
		     *   content with the loop element itself
		     */
		    connectNext(lNext, current);

		    --loopVar;
		    current = NULL;
		    flag = FLAG_POP;

		    break;
		}

		if(strcmp(string, TAGNAME_EXPRESSION) == 0) {
		    if(stackisempty() ||
		       tos()->State->TOMASD_Type != TOMASD_TYPE_EXPRSTART) {
			JSSETERROR2(_ERR_TOPACREATE, 3, string);
			EGOTO(error);
		    }

		    if((current = getStateDescription(topa, TOMASD_TYPE_EXPREND)) == NULL)
			EGOTO(error);

		    current->TOMASD_ExprIndex = tos()->State->TOMASD_ExprIndex;
		    flag = FLAG_POP;

		    break;
		}

		if(strcmp(string, TAGNAME_STATEMENT) == 0) {
		    if(stackisempty() ||
		       tos()->State->TOMASD_Type != TOMASD_TYPE_STMTSTART) {
			JSSETERROR2(_ERR_TOPACREATE, 3, string);
			EGOTO(error);
		    }

		    if((current = getStateDescription(topa, TOMASD_TYPE_STMTEND)) == NULL)
			EGOTO(error);

		    current->TOMASD_ExprIndex = tos()->State->TOMASD_ExprIndex;
		    flag = FLAG_POP;

		    break;
		}
	    }

	    if(type == SAXTYPE_ELEMENT) {
		if(strcmp(string, TAGNAME_EXPRESSION) == 0) {
		    if((current = getExpressionState(&aux, TOMASD_TYPE_EXPR,
						     sapa)) == NULL)
			EGOTO(error);

		    break;
		}

		if(strcmp(string, TAGNAME_IDENTIFIER) == 0) {
		    if((current = getExpressionState(&aux, TOMASD_TYPE_IDENT,
						     sapa)) == NULL)
			EGOTO(error);

		    break;
		}

		if(strcmp(string, TAGNAME_NUMBER) == 0) {
		    if((current = getExpressionState(&aux, TOMASD_TYPE_NUMBER,
						     sapa)) == NULL)
			EGOTO(error);

		    break;
		}

		if(strcmp(string, TAGNAME_STRING) == 0) {
		    if((current = getExpressionState(&aux, TOMASD_TYPE_STRING,
						     sapa)) == NULL)
			EGOTO(error);

		    break;
		}

		if(strcmp(string, TAGNAME_STATEMENT) == 0) {
		    if((current = getExpressionState(&aux, TOMASD_TYPE_STMT,
						     sapa)) == NULL)
			EGOTO(error);

		    break;
		}
	    }

	    if(type == SAXTYPE_CHARACTER) {
		if((current = getTokenState(&aux, sapa)) == NULL)
		    EGOTO(error);

		break;
	    }

	    EGOTO(syntaxError);

	case 16: /* ***  Parsing <pattern>, within <choice>  *** */
	    if(type == SAXTYPE_STARTELEMENT) {
		if(strcmp(string, TAGNAME_BRANCH) == 0) {
		    head = NULL;
		    parserState = 15;
		    break;
		}
	    }

	    if(type == SAXTYPE_STOPELEMENT) {
		if(strcmp(string, TAGNAME_CHOICE) == 0) {
		    if(stackisempty() ||
		       tos()->State->TOMASD_Type != TOMASD_TYPE_CHOICE) {
			JSSETERROR2(_ERR_TOPACREATE, 3, string);
			EGOTO(error);
		    }

		    /*  A branch node has no next field  */
		    tos()->State->TOMASD_Next = NULL;

		    flag = FLAG_POP;

		    parserState = 15;
		    break;
		}
	    }

	    EGOTO(syntaxError);

	case 20: /* ***  scanning substitute  **** */
	    if(type == SAXTYPE_STARTELEMENT) {
		JSSETERROR2(_ERR_TOPACREATE, 2, string);
		EGOTO(error);
	    }

	    if(type == SAXTYPE_ELEMENT) {
		const char *varName;
		jsctoken_t token;
		unsigned int ind, fType;

		fType = strcmp(string, TAGNAME_EXPRESSION) == 0 ? TOMASD_TYPE_EXPR :
		    strcmp(string, TAGNAME_STATEMENT)      == 0 ? TOMASD_TYPE_STMT :
		    strcmp(string, TAGNAME_IDENTIFIER)     == 0 ? TOMASD_TYPE_IDENT :
		    strcmp(string, TAGNAME_NUMBER)         == 0 ? TOMASD_TYPE_NUMBER :
		    strcmp(string, TAGNAME_STRING)         == 0 ? TOMASD_TYPE_STRING :
		    strcmp(string, TAGNAME_LINENR)         == 0 ? TOMASD_TYPE_LINENR :
		    strcmp(string, TAGNAME_COUNTER)        == 0 ? TOMASD_TYPE_COUNTER :
		    -1;

		if(fType == (unsigned int)-1)
		    EGOTO(syntaxError);

		token.TK_Flags = 0;
		if(fType == TOMASD_TYPE_LINENR) {
		    token.TK_Value = TOPA_TOKEN_LINENR;

		} else {
		    unsigned int flags;

		    if((flags = analyzeVariableTag(&aux, sapa,
						   &varName, 1)) == (unsigned int)-1)
			EGOTO(error);

		    if(varName == NULL) {
			JSSETERROR(_ERR_TOPACREATE, 4);
			EGOTO(error);
		    }

		    if(fType == TOMASD_TYPE_COUNTER) {
			if((ind = getCounterIndex(&aux, varName)) == (unsigned int)-1)
			    EGOTO(error);

			token.TK_Value = TOPA_TOKEN_COUNTER | ind;

		    } else {
			if((ind = findExpressionIndex(&aux, fType, varName)) == (unsigned int)-1) {
			    JSSETERROR2(_ERR_TOPACREATE, 5, varName);
			    EGOTO(error);
			}

			token.TK_Value = ((flags & 1) ? TOPA_TOKEN_TRIMEXPRESSION : TOPA_TOKEN_EXPRESSION) |
			    ind;
		    }
		}

		token.TK_String = NULL;

		if(jctokinsert(&(topa->TOPA_StateQueue), token, 0) == -1)
		    EGOTO(error);

		break;
	    }

	    if(type == SAXTYPE_CHARACTER) {
		if(addCTokens(&aux, string) == -1)
		    EGOTO(error);

		break;
	    }

	    if(type == SAXTYPE_STOPELEMENT) {
		if(strcmp(string, TAGNAME_SUBSTITUTE) != 0) {
		    JSSETERROR2(_ERR_TOPACREATE, 2, string);
		    EGOTO(error);
		}

		/*  Restore mode  */
		sapa->SAPA_Mode |= sapamode & SAXTYPE_FTRIMSPACES;

		endIndex = topa->TOPA_StateQueue.TKFF_Index;

		parserState = 10;
		break;
	    }

	    EGOTO(syntaxError);
	}

	/* ----  */
	if(current != NULL) {
	    if(head == NULL)
		/*  First element, memorize it as the head  */
		head = current;

	    if(lNext != NULL) {
		/*  Connect the previous elements Next field with this element  */
		connectNext(lNext, current);
		lNext = NULL;
	    }

	    assert(lNext == NULL);

	    /*  Memorize this element for next round  */
	    if(!(flag & FLAG_NO_LNEXT)) {
		current->TOMASD_Next = NULL;
		lNext = current;
	    }

	    if(flag & (FLAG_PUSH | FLAG_PUSHHDLN)) {
		unsigned int type1, type2;

		/*  Push 'current' on stack  */
		if(aux.NStack >= aux.MStack) {
		    aux.MStack += 10;
		    if((aux.Stack = (topaparserstack_t *)
			realloc(aux.Stack, aux.MStack * sizeof(*(aux.Stack)))) == NULL)
			EGOTO(error);
		}

		/*  Keep all flags FLAG_USER and higher  */
		aux.Stack[aux.NStack].Flags = (flag & ~(FLAG_USER - 1));
		aux.Stack[aux.NStack].State = current;

		if(current->TOMASD_Type == TOMASD_TYPE_EXPR ||
		   current->TOMASD_Type == TOMASD_TYPE_EXPRSTART) {
		    type1 = TOMASD_TYPE_EXPR;
		    type2 = TOMASD_TYPE_EXPRSTART;
		} else if(current->TOMASD_Type == TOMASD_TYPE_STMT ||
			  current->TOMASD_Type == TOMASD_TYPE_STMTSTART) {
		    type1 = TOMASD_TYPE_STMT;
		    type2 = TOMASD_TYPE_STMTSTART;

		} else
		    type1 = type2 = -1;

		if(type1 != (unsigned int)-1) {
		    /*  Check if the same variable appears already in the stack  */
		    for(i = 0 ; i < aux.NStack ; ++i)
			if((aux.Stack[i].State->TOMASD_Type == type1 ||
			    aux.Stack[i].State->TOMASD_Type == type2) &&
			   aux.Stack[i].State->TOMASD_ExprIndex ==
			   current->TOMASD_ExprIndex) {
			    JSSETERROR(_ERR_TOPACREATE, 8);
			    EGOTO(error);
			}
		}

		if(flag & FLAG_PUSHHDLN) {
		    aux.Stack[aux.NStack].Flags |= 3;
		    aux.Stack[aux.NStack].Head  = head;     head = NULL;
		    aux.Stack[aux.NStack].LNext = lNext;    lNext = NULL;
		}

		aux.NStack++;
	    }
	}

	if(flag & FLAG_POP) {
	    topaparserstack_t *top = tos();

	    aux.NStack--;

	    if(top->Flags & 1) lNext = top->LNext;
	    if(top->Flags & 2) head = top->Head;
	}
    }

    if(parserState != 6) {
	string = "EOF";
	EGOTO(syntaxError);
    }

#undef FLAG_NO_LNEXT
#undef FLAG_PUSH
#undef FLAG_PUSHHDLN
#undef FLAG_POP

#undef stackisempty
#undef tos

    saxparsefree(sapa);

    cleanaux(&aux, 1);

#ifdef TRACETYPE_TOKPATMAT
    if(jstracelevel(TRACETYPE_TOKPATMAT) > 4)
	showStates(topa);
#endif

    if(fclose(fpXml) == EOF) {
	topadelete(topa);
	EGOTO(error);
    }

    /*  Allocate counters  */
    if(topa->TOPA_NCounters > 0) {
	if((topa->TOPA_Counters = (unsigned long *)
	    malloc(topa->TOPA_NCounters * sizeof(*(topa->TOPA_Counters)))) == NULL)
	    EGOTO(error);
    }

    return topa;
syntaxError:
    JSSETERROR2(_ERR_TOPACREATE, 0, string);
error:
    fclose(fpXml);
    JSSETERRFILE2(fileName, linePos[0], linePos[1]);
error0:
    return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static int analyzeVariableTag(topacraux_t *paux, saxparse_t *sapa,
			      const char **pName, int mode)

/************************************************************************/
{
    const char **attrs;
    unsigned int i, nattrs, flags;

    USE(paux);

    *pName = NULL;
    flags = 0;

    attrs = saxparsegetattrs(sapa);
    nattrs = saxparsegetnattrs(sapa);
    for(i = 0 ; i < nattrs ; i += 2)
	if(strcmp(attrs[i], ATTRNAME_NAME) == 0) {
	    *pName = attrs[i + 1];

	} else if((mode & 1) && strcmp(attrs[i], ATTRNAME_TRIM) == 0) {
	    flags |= 1;

	} else {
	    JSSETERROR2(_ERR_TOPACREATE, 6, attrs[i]);
	    EGOTO(error);
	}

    return flags;
error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static tomastatedescr_t *getExpressionState(topacraux_t *paux,
					    unsigned int type, saxparse_t *sapa)

/************************************************************************/
{
    tomastatedescr_t *stateDes;
    const char *variableName;
    unsigned int varIndex, fType;

    fType = type == TOMASD_TYPE_STMTSTART ? TOMASD_TYPE_STMT :
	type == TOMASD_TYPE_EXPRSTART ? TOMASD_TYPE_EXPR :
	type;

    if(analyzeVariableTag(paux, sapa, &variableName, 0) == -1)
	EGOTO(error);

    if(variableName == NULL)
	varIndex = -1;
    else if((varIndex = findExpressionIndex(paux, fType,
					    variableName)) == (unsigned int)-1) {
	varIndex = paux->NVarNames;

	if(varIndex >= paux->MVarNames) {
	    paux->MVarNames += 10;
	    if((paux->VarNames = (topaexprname_t *)
		realloc(paux->VarNames, paux->MVarNames * sizeof(*(paux->VarNames))))
	       == NULL)
		EGOTO(error);
	}

	paux->VarNames[varIndex].Type = fType;
	if((paux->VarNames[varIndex].Name = strdup(variableName)) == NULL)
	    EGOTO(error);

	paux->NVarNames++;
    }

    if((stateDes = getStateDescription(paux->Topa, type)) == NULL)
	EGOTO(error);

    stateDes->TOMASD_ExprIndex = varIndex;

    return stateDes;
error:
    return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static int getCounterIndex(topacraux_t *paux, const char *name)

/************************************************************************/
{
    unsigned int i;

    for(i = 0 ; i < paux->NCounterNames ; ++i)
	if(strcmp(paux->CounterNames[i], name) == 0)
	    break;

    if(i == paux->NCounterNames) {
	if(paux->NCounterNames >= paux->MCounterNames) {
	    paux->MCounterNames += 10;
	    if((paux->CounterNames = (const char **)
		realloc((void *)paux->CounterNames,
			paux->MCounterNames * sizeof(*(paux->CounterNames)))) == NULL)
		EGOTO(error);
	}

	if((paux->CounterNames[i] = strdup(name)) == NULL)
	    EGOTO(error);

	paux->NCounterNames++;
    }

    return i;
error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int findExpressionIndex(topacraux_t *paux, unsigned int type,
			       const char *name)

/************************************************************************/
{
    unsigned int i;

    for(i = 0 ; i < paux->NVarNames ; ++i)
	if(paux->VarNames[i].Type == type &&
	   strcmp(paux->VarNames[i].Name, name) == 0)
	    return i;

    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static tomastatedescr_t *getRepeatState(topacraux_t *paux,
					unsigned int type,
					unsigned int loopVar,
					saxparse_t *sapa)

/************************************************************************/
{
    tomastatedescr_t *stateDes;

    if((stateDes = getStateDescription(paux->Topa, type)) == NULL)
	EGOTO(error);

    stateDes->TOMASD_ILoopVar = loopVar;
    stateDes->TOMASD_MinLoop = -1;
    stateDes->TOMASD_MaxLoop = -1;

    if(sapa != NULL) {
	const char **attrs, *end;
	unsigned int i, nattrs, aflags;

	attrs = saxparsegetattrs(sapa);
	nattrs = saxparsegetnattrs(sapa);
	aflags = 0;
	for(i = 0 ; i < nattrs ; i += 2) {
	    unsigned int value;

	    value = strtoul(attrs[i + 1], (char **)&end, 10);
	    if(*end != '\0') {
		JSSETERROR2(_ERR_TOPACREATE, 7, attrs[i + 1]);
		EGOTO(error);
	    }

	    if(strcmp(attrs[i], ATTRNAME_MIN) == 0) {
		stateDes->TOMASD_MinLoop = value;
		aflags |= 1;

	    } else if(strcmp(attrs[i], ATTRNAME_MAX) == 0) {
		stateDes->TOMASD_MaxLoop = value;
		aflags |= 2;

	    } else if(strcmp(attrs[i], ATTRNAME_COUNT) == 0) {
		stateDes->TOMASD_MinLoop = stateDes->TOMASD_MaxLoop = value;
		aflags |= 4;

	    } else {
		JSSETERROR2(_ERR_TOPACREATE, 6, attrs[i]);
		EGOTO(error);
	    }
	}

	if((aflags & 3) && (aflags & 4)) {
	    JSSETERROR(_ERR_TOPACREATE, 11);
	    EGOTO(error);
	}
    }

    return stateDes;
error:
    return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static tomastatedescr_t *getTokenState(topacraux_t *paux, saxparse_t *sapa)

/************************************************************************/
{
    tomastatedescr_t *stateDes;

    if((stateDes = getStateDescription(paux->Topa, TOMASD_TYPE_TOKENS)) == NULL)
	EGOTO(error);

    stateDes->TOMASD_TokStart = paux->Topa->TOPA_StateQueue.TKFF_Index;

    if((stateDes->TOMASD_TokEnd =
	addCTokens(paux, saxparsegetname(sapa))) == (unsigned int)-1)
	EGOTO(error);

    return stateDes;
error:
    return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static int addCTokens(topacraux_t *paux, const char *string)

/************************************************************************/
{
    extfile_t *efp;
    jsctoken_t token;
    jctokfifo_t *queue;

    queue = &(paux->Topa->TOPA_StateQueue);

    if((efp = efopen(string, NULL)) == NULL)
	EGOTO(error);

    while((token = GETCTOKEN(efp)).TK_Value != CTK_EOF) {
	if(token.TK_Value == CTK_ERROR)
	    EGOTO(error);

	if(jctokinsert(queue, token, 0) == -1)
	    EGOTO(error);
    }

    if(efclose(efp) == -1)
	EGOTO(error);

    return queue->TKFF_Index;
 error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static void connectNext(tomastatedescr_t *list, tomastatedescr_t *next)

/************************************************************************/
{
    while(list != NULL) {
	tomastatedescr_t *lNext;

	lNext = list->TOMASD_Next;
	list->TOMASD_Next = next;
	list = lNext;
    }
}

/*JS*********************************************************************
*   Links the node at the end of the list. Returns the new head of the
*    list
*************************************************************************/

static tomastatedescr_t *linkTail(tomastatedescr_t *list,
				  tomastatedescr_t *node)

/************************************************************************/
{
    tomastatedescr_t *curr;

    if(list == NULL)
	return node;

    curr = list;
    while(curr->TOMASD_Next) curr = curr->TOMASD_Next;

    curr->TOMASD_Next = node;

    return list;
}

/*JS*********************************************************************
*
*************************************************************************/

static int addDefinition(tokpattmatch_t *topa, const char *name,
			 tomastatedescr_t *startState,
			 unsigned int start, unsigned int end)

/************************************************************************/
{
    tokpattdef_t *def;

    if((topa->TOPA_Definitions = (tokpattdef_t *)
	realloc(topa->TOPA_Definitions, (topa->TOPA_NDefinitions + 1) *
		sizeof(*(topa->TOPA_Definitions)))) == NULL)
	EGOTO(error);

    def = &(topa->TOPA_Definitions[topa->TOPA_NDefinitions]);

    def->TOPD_Name = name;
    def->TOPD_StartState = startState;
    def->TOPD_Start = start;
    def->TOPD_End = end;

    topa->TOPA_NDefinitions++;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static void cleanaux(topacraux_t *paux, int flag)

/************************************************************************/
{
    unsigned int i;

    for(i = 0 ; i < paux->NVarNames ; ++i)
	free((void *)paux->VarNames[i].Name);

    for(i = 0 ; i < paux->NCounterNames ; ++i)
	free((void *)paux->CounterNames[i]);

    if(flag) {
	free(paux->VarNames);
	free((void *)paux->CounterNames);
    } else {
	paux->NVarNames = 0;
	paux->NCounterNames = 0;
	paux->NStack = 0;
    }
}

/*JS*********************************************************************
*
*************************************************************************/

void topadelete(tokpattmatch_t *topa)

/************************************************************************/
{
    topamemory_t *stmem;
    matcherstate_t *state, *next;

    jctokfree(&(topa->TOPA_StateQueue), 0);
    jctokfree(&(topa->TOPA_InputQueue), 0);

    for(state = topa->TOPA_MatcherStates ; state ; state = next) {
	next = state->MAST_Next;
	free(state);
    }

    stmem = topa->TOPA_StateMemory;
    while(stmem) {
	topamemory_t *nextMem;

	nextMem = stmem->TOPM_Next;
	free(stmem);
	stmem = nextMem;
    }

    free(topa->TOPA_Definitions);

    if(topa->TOPA_Counters != NULL)
	free(topa->TOPA_Counters);

    free(topa);
}

/*JS*********************************************************************
*   mode bit 0: Preserve comment tokens in the output. Print it after
*               a match.
*   mode bit 1: Preserve comment tokens in the output. Print it before
*               a match.
*   mode bit 2: Only convert matched patterns, discard all other output.
*   mode bit 3: Precede matched patterns with line number
*   mode bit 4: Look for matches with longest token sequence.
*   mode bit 5: Convert newlines to spaces in output.
*************************************************************************/

int topaprocess(tokpattmatch_t *topa, extfile_t *efp, const char *fileName,
		FILE *fpOut, const char **matchers, int mode)

/************************************************************************/
{
    jctokfifo_t *iQueue;
    unsigned int i, offset, count, linenr;

    iQueue = &(topa->TOPA_InputQueue);
    count = 0;

    /*  Ensure empty queue  */
    jctokclear(iQueue);

    /*  Initialize counters  */
    for(i = 0 ; i < topa->TOPA_NCounters ; ++i)
	topa->TOPA_Counters[i] = 0;

    topa->TOPA_Input = efp;

    for(linenr = 1, offset = 0 ; ; ) {
	const jsctoken_t *inputToken;
	matcherstate_t *masta;
	unsigned int ndel;

	/*  Check if EOF is reached. This also ensures we *read*
	 *   at least one token which we can shift afterwards.
	 */
	if((inputToken = indexedToken(topa, iQueue, offset)) == NULL)
	    EGOTO(error);

	if(inputToken->TK_Value == CTK_EOF)
	    break;

	/*  Every token might contain newlines  */
	if(inputToken->TK_String != NULL)
	    linenr += countnlCToken(inputToken->TK_String);

	if(CTK_BLANK_START <= inputToken->TK_Value &&
	   inputToken->TK_Value <= CTK_BLANK_END) {
	    /*  Print blank tokens immediately  */
	    if(!(mode & 4) && printCToken(fpOut, iQueue->TKFF_Token[offset]) == -1)
		EGOTO(error);

	    ++offset;
	    continue;
	}

	masta = NULL;
	for(i = 0 ; i < topa->TOPA_NDefinitions ; ++i) {
	    const tokpattdef_t *pdef;
	    tomastatedescr_t *startState;
	    unsigned int j;

	    pdef = &(topa->TOPA_Definitions[i]);

	    if(matchers != NULL) {
		const char *matName;

		if((matName = pdef->TOPD_Name) == NULL)
		    /*  If patterns are given, do not process unnamed matchers  */
		    continue;

		for(j = 0 ; matchers[j] != NULL ; ++j)
		    if(strcmp(matchers[j], matName) == 0)
			break;

		if(matchers[j] == NULL)
		    continue;
	    }

	    if((startState = pdef->TOPD_StartState) == NULL)
		continue;

	    if(findMatch(topa, iQueue, startState, offset, &masta,
			 (mode & 16) ? 1 : 0) == -1)
		EGOTO(error);

	    if(masta != NULL)
		break;
	}

	if(masta != NULL) {
	    unsigned int k, end;

#ifdef TRACETYPE_TOKPATMAT
	    if(jstracelevel(TRACETYPE_TOKPATMAT) > 4) {
		const matcherstate_t *cmasta;

		for(cmasta = masta ; cmasta ; cmasta = cmasta->MAST_Next) {
		    char buffer[1000];

		    jstrace(TRACETYPE_TOKPATMAT, 4,
			    "#%d: MATCHED %d tokens: '", i,
			    cmasta->MAST_TokenIndex - offset);

		    for(k = offset ; k < cmasta->MAST_TokenIndex ; ++k) {

			sprintCToken(buffer, iQueue->TKFF_Token[k]);
			jstrace(TRACETYPE_TOKPATMAT, 4,
				"%s", buffer);
		    }
		    jstrace(TRACETYPE_TOKPATMAT, 4, "'\n");

		    for(k = 0 ; k < topa->TOPA_NVariables ; ++k) {
			unsigned int s, e;

			s = cmasta->MAST_VariableMatch[k].MASTV_Start;
			e = cmasta->MAST_VariableMatch[k].MASTV_End;
			jstrace(TRACETYPE_TOKPATMAT, 4,
				"  V%d [%d,%d] '", k, s, e);
			for( ; s < e ; ++s) {
			    sprintCToken(buffer, iQueue->TKFF_Token[s]);
			    jstrace(TRACETYPE_TOKPATMAT, 4,
				    "%s", buffer);
			}
			jstrace(TRACETYPE_TOKPATMAT, 4, "'\n");
		    }
		}
	    }
#endif

	    if((mode & 6) == 2) { /*  Print comments beforehand  */
		for(k = offset ; k < masta->MAST_TokenIndex ; ++k) {
		    jsctoken_t token;

		    token = iQueue->TKFF_Token[k];

		    if(token.TK_Value == CTK_COMMENT || token.TK_Value == CTK_CPPCOMMENT)
			printCToken(fpOut, token);
		}
	    }

	    if(mode & 8) {
		if(fileName != NULL)
		    fprintf(fpOut, "l. %d \"%s\": ", linenr, fileName);
		else
		    fprintf(fpOut, "l. %d: ", linenr);
	    }

	    /* ***  Print the substitution  *** */
	    end = topa->TOPA_Definitions[i].TOPD_End;
	    for(k = topa->TOPA_Definitions[i].TOPD_Start ; k < end ; ++k) {
		jsctoken_t token;
		unsigned int ttype;

		token = topa->TOPA_StateQueue.TKFF_Token[k];

		ttype = token.TK_Value & TOPA_TOKEN_TMASK;
		if(ttype == TOPA_TOKEN_EXPRESSION || ttype == TOPA_TOKEN_TRIMEXPRESSION) {
		    topavarmatch_t *vama;
		    unsigned int l, tokVal;

		    vama = &(masta->MAST_VariableMatch[token.TK_Value & TOPA_TOKEN_FMASK]);

		    /*  Print what the variable was found to match  */
		    for(l = vama->MASTV_Start ; l < vama->MASTV_End ;
			l += jctokgetstep(iQueue)) {

			if(ttype == TOPA_TOKEN_TRIMEXPRESSION) {
			    tokVal = iQueue->TKFF_Token[l].TK_Value;
			    if(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END) {
				if(putc(' ', fpOut) == EOF)
				    EGOTO(error);

				continue;
			    }
			}

			tokVal = iQueue->TKFF_Token[l].TK_Value;
			if((mode & 4) && CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END) {
			    if(putc(' ', fpOut) == EOF)
				EGOTO(error);
			} else {
			    if(printCToken(fpOut, iQueue->TKFF_Token[l]) == -1)
				EGOTO(error);
			}
		    }

		} else if(ttype == TOPA_TOKEN_LINENR) {
		    fprintf(fpOut, "%d", linenr);

		} else if(ttype == TOPA_TOKEN_COUNTER) {
		    fprintf(fpOut, "%ld",
			    1 + topa->TOPA_Counters[token.TK_Value & TOPA_TOKEN_FMASK]++);

		} else if(printCToken(fpOut, token) == -1)
		    EGOTO(error);
	    }

	    if((mode & 4) && fputc('\n', fpOut) == EOF)
		EGOTO(error);

	    if((mode & 5) == 1) { /*  Print comments afterwards  */
		for(k = offset ; k < masta->MAST_TokenIndex ; ++k) {
		    jsctoken_t token;

		    token = iQueue->TKFF_Token[k];

		    if(token.TK_Value == CTK_COMMENT || token.TK_Value == CTK_CPPCOMMENT)
			printCToken(fpOut, token);
		}
	    }

	    ++count;

	    /*  Make sure we advance at all. This protects us against an infinite
	     *   loop due to patterns that match nothing. In that case, each
	     *   "nothing" is only matched once
	     */
	    ndel = MAX(1, masta->MAST_TokenIndex - offset);

	    freeMatcherState(topa, masta);

	} else {
	    /*  Nothing matched  */
	    if(!(mode & 4) && printCToken(fpOut, iQueue->TKFF_Token[offset]) == -1)
		EGOTO(error);

	    ndel = 1; /*  start matching from next token  */
	}

	offset += ndel;

	if(offset > GC_INTERVAL) {
	    /*  Shorten queue  */
	    if(jctokdelete(iQueue, 0, offset, 1) == -1)
		EGOTO(error);

	    offset = 0;
	}
    }

    return count;
 error:
    return -1;
}

/*JS*********************************************************************
*   mode bit 0: If set, look for match with longest token sequence.
*        bit 1: Return all possible matches.
*************************************************************************/

static int findMatch(tokpattmatch_t *topa, jctokfifo_t *iQueue,
		     tomastatedescr_t *start,
		     unsigned int startIndex, matcherstate_t **pState,
		     int mode)

/************************************************************************/
{
    matcherstate_t *head, *current, *result;

    /*  Initialize states  */
    if((current = getMatcherState(topa, start, startIndex, NULL)) == NULL)
	EGOTO(error);

    result = NULL;

    head = NULL;
    putState(topa, &head, current);

    /*  MAIN LOOP  */
    while(head != NULL) {
	tomastatedescr_t *currDescr;
	unsigned int newPos;

#ifdef TRACETYPE_TOKPATMAT
	if(jstracelevel(TRACETYPE_TOKPATMAT) > 4)
	    showMatcherStates(topa, head, iQueue);
#endif

	/*  POP  */
	current = head;

	if((currDescr = current->MAST_StateDescription) == NULL) {
	    /*  A complete match! Remove from work list and add to result  */
	    head = head->MAST_Next;

	    if(mode & 2) {
		/*  Collect all results  */
		current->MAST_Next = result;
		result = current;

	    } else if(mode & 1) {
		if(result == NULL) {
		    result = current;
		} else if(current->MAST_TokenIndex > result->MAST_TokenIndex) {
		    /*  current is better  */
		    (void)freeMatcherState(topa, result);
		    result = current;
		} else {
		    freeMatcherState(topa, current);
		}

	    } else {
		/*  One result is enough, so stop loop  */
		result = current;

		/*  Clean up remaining stuff first  */
		while(head)
		    head = freeMatcherState(topa, head);

		break;
	    }

	    /*  Continue looking for other matches  */
	    continue;
	}

	/*  Advance this state to the next one to avoid endless loops  */
	current->MAST_StateDescription = currDescr->TOMASD_Next;

	head = head->MAST_Next; /*  complete the "pop" operation  */

	switch(currDescr->TOMASD_Type) {
	case TOMASD_TYPE_TOKENS:
	    if((newPos = checkTokens(topa, iQueue, current->MAST_TokenIndex,
				     currDescr->TOMASD_TokStart,
				     currDescr->TOMASD_TokEnd, 0)) == (unsigned int)-1)
		EGOTO(error);

	    if(newPos) {
		current->MAST_TokenIndex = newPos;

		putState(topa, &head, current);
		current = NULL;
	    }

	    break;

	case TOMASD_TYPE_REPEAT:
	    {
		unsigned int nrRepeat;

		nrRepeat = current->MAST_LoopCounter[currDescr->TOMASD_ILoopVar];

		/*  Check if we can still make another round  */
		if(currDescr->TOMASD_MaxLoop == (unsigned int)-1 ||
		   nrRepeat < currDescr->TOMASD_MaxLoop) {
		    matcherstate_t *newState;

		    if((newState = getMatcherState(topa, currDescr->TOMASD_Sub,
						   current->MAST_TokenIndex, current)) == NULL)
			EGOTO(error);

		    /*  Loops cannot be empty  */
		    assert(newState->MAST_StateDescription != NULL);

		    newState->MAST_LoopCounter[currDescr->TOMASD_ILoopVar]++;

		    /*  If an upper limit is given, add in reversed sequence so longer matches
		     *   are preferred.
		     */
		    putState(topa, &head, newState);
		}

		if(currDescr->TOMASD_MinLoop == (unsigned int)-1 ||
		   currDescr->TOMASD_MinLoop <= nrRepeat) {
		    /*  We have reached our minimum number of loops, so we might
		     *   advance to the next state
		     */

		    /*  Reset counter for next time usage  */
		    current->MAST_LoopCounter[currDescr->TOMASD_ILoopVar] = 0;

		    putState(topa, &head, current);
		    current = NULL;
		}
	    }

	    break;

	case TOMASD_TYPE_CHOICE:
	    {
		tomastatedescr_t *sub;

		for(sub = currDescr->TOMASD_Sub ; sub ; sub = sub->TOMASD_Connect) {
		    matcherstate_t *newState;

		    if((newState = getMatcherState(topa, sub, current->MAST_TokenIndex,
						   current)) == NULL)
			EGOTO(error);

		    putState(topa, &head, newState);
		}
	    }
	    break;

	case TOMASD_TYPE_EXPR:
	case TOMASD_TYPE_STMT:
	case TOMASD_TYPE_IDENT:
	case TOMASD_TYPE_NUMBER:
	case TOMASD_TYPE_STRING:
	    {
		topavarmatch_t *varmat;

		if(currDescr->TOMASD_ExprIndex != (unsigned int)-1)
		    varmat = &(current->MAST_VariableMatch[currDescr->TOMASD_ExprIndex]);
		else
		    varmat = NULL;

		if(varmat == NULL || varmat->MASTV_Start == (unsigned int)-1) {
		    if(currDescr->TOMASD_Type == TOMASD_TYPE_EXPR || currDescr->TOMASD_Type == TOMASD_TYPE_STMT) {
			if(getExprMatcher(topa, iQueue, current, currDescr->TOMASD_ExprIndex, &head,
					  currDescr->TOMASD_Type == TOMASD_TYPE_STMT ? 1 : 0) == -1)
			    EGOTO(error);

		    } else {
			if(getTokenMatcher(topa, iQueue, current, currDescr->TOMASD_ExprIndex, &head, 0) == -1)
			    EGOTO(error);
		    }

		} else {
		    /*  This variable was already seen, check if token
		     *   sequence matches
		     */
		    if((newPos = checkTokens(topa, iQueue, current->MAST_TokenIndex,
					     varmat->MASTV_Start, varmat->MASTV_End,
					     1)) == (unsigned int)-1)
			EGOTO(error);

		    if(newPos) {
			current->MAST_TokenIndex = newPos;

			putState(topa, &head, current);
			current = NULL;
		    }
		}
	    }
	    break;

	case TOMASD_TYPE_EXPRSTART:
	case TOMASD_TYPE_STMTSTART:
	    /*  It is useless to have an unnamed expression bracketing, but ignore it  */
	    if(currDescr->TOMASD_ExprIndex != (unsigned int)-1) {
		topavarmatch_t *varmat;

		varmat = &(current->MAST_VariableMatch[currDescr->TOMASD_ExprIndex]);

		if(varmat->MASTV_Start == (unsigned int)-1) {
		    /*  Memorize the starting point  */
		    varmat->MASTV_Start = current->MAST_TokenIndex;
		    putState(topa, &head, current);
		    current = NULL;

		} else {
		    /*  the state machine generator ensures that expressions are not nested within themselves  */
		    assert(varmat->MASTV_End != (unsigned int)-1);

		    /*  That is a bit strange: We have already seen and identified this expression, but
		     *   still it appears again wrapping a pattern.
		     *   In such a case, we simply check if this pattern appears again and if so, set
		     *   the determined end index which we later check when encountering TOMASD_TYPE_EXPREND.
		     */
		    if((newPos = checkTokens(topa, iQueue, current->MAST_TokenIndex,
					     varmat->MASTV_Start, varmat->MASTV_End, 1)) == (unsigned int)-1)
			EGOTO(error);

		    if(newPos) {
			/*  We basically move the interval to the new position  */
			varmat->MASTV_Start = current->MAST_TokenIndex;
			varmat->MASTV_End = newPos;

			/*  We *leave* current->MAST_TokenIndex because the pattern is still
			 *   supposed to be matched  */
			putState(topa, &head, current);
			current = NULL;
		    }
		}
	    }
	    break;

	case TOMASD_TYPE_EXPREND:
	case TOMASD_TYPE_STMTEND:
	    /*  It is useless to have an unnamed expression bracketing, but ignore it  */
	    if(currDescr->TOMASD_ExprIndex != (unsigned int)-1) {
		topavarmatch_t *varmat;

		varmat = &(current->MAST_VariableMatch[currDescr->TOMASD_ExprIndex]);

		if(varmat->MASTV_End == (unsigned int)-1) {
		    varmat->MASTV_End = current->MAST_TokenIndex;
		    putState(topa, &head, current);
		    current = NULL;

		} else {
		    /*  [see lengthy comment above at TOMASD_TYPE_EXPRSTART]  */
		    if(current->MAST_TokenIndex == varmat->MASTV_End) {
			putState(topa, &head, current);
			current = NULL;
		    }
		}
	    }

	    break;
	}

	if(current != NULL)
	    freeMatcherState(topa, current);
    }

    if(result != NULL && !(mode & 2))
	result->MAST_Next = NULL;

    *pState = result;

    return result != NULL;
 error:
    return -1;
}

#ifdef TRACETYPE_TOKPATMAT
/*JS*********************************************************************
*
*************************************************************************/

static void showStates(const tokpattmatch_t *topa)

/************************************************************************/
{
    topamemory_t *stmem;
    unsigned int i, nr;

    jstrace(TRACETYPE_TOKPATMAT, 4,
	    "STATES:\n");
    nr = 0;
    for(stmem = topa->TOPA_StateMemory ; stmem ; stmem = stmem->TOPM_Next) {
	for(i = 0 ; i < stmem->TOPM_Index ; ++i) {
	    tomastatedescr_t *currDescr;

	    currDescr = &(stmem->TOPM_Memory[i]);
	    jstrace(TRACETYPE_TOKPATMAT, 4,
		    "  STATE #%d: NEXT #%d: ", nr,
		    findStateNr(topa, currDescr->TOMASD_Next));
	    ++nr;

	    switch(currDescr->TOMASD_Type) {
	    case TOMASD_TYPE_TOKENS:
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"TOKENS [%d,%d] '", currDescr->TOMASD_TokStart,
			currDescr->TOMASD_TokEnd);
		printTokens(&(topa->TOPA_StateQueue),
			    currDescr->TOMASD_TokStart, currDescr->TOMASD_TokEnd);
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"'");
		break;

	    case TOMASD_TYPE_CHOICE:
		{
		    tomastatedescr_t *sub = currDescr->TOMASD_Sub;

		    jstrace(TRACETYPE_TOKPATMAT, 4,
			    "CHOICE [%d", findStateNr(topa, sub));
		    if(sub) {
			for(sub = sub->TOMASD_Connect ; sub ; sub = sub->TOMASD_Connect)
			    jstrace(TRACETYPE_TOKPATMAT, 4,
				    ",%d", findStateNr(topa, sub));
		    }
		    jstrace(TRACETYPE_TOKPATMAT, 4,
			    "]");
		}
		break;

	    case TOMASD_TYPE_REPEAT:
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"REPEAT LVAR %d [%d,%d] BODY #%d", currDescr->TOMASD_ILoopVar,
			currDescr->TOMASD_MinLoop, currDescr->TOMASD_MaxLoop,
			findStateNr(topa, currDescr->TOMASD_Sub));
		break;

	    case TOMASD_TYPE_EXPR:
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"EXPR #%d", currDescr->TOMASD_ExprIndex);
		break;

	    case TOMASD_TYPE_EXPRSTART:
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"EXPRSTART #%d", currDescr->TOMASD_ExprIndex);
		break;

	    case TOMASD_TYPE_EXPREND:
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"EXPREND #%d", currDescr->TOMASD_ExprIndex);
		break;

	    case TOMASD_TYPE_STMT:
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"STMT #%d", currDescr->TOMASD_ExprIndex);
		break;

	    case TOMASD_TYPE_STMTSTART:
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"STMTSTART #%d", currDescr->TOMASD_ExprIndex);
		break;

	    case TOMASD_TYPE_STMTEND:
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"STMTEND #%d", currDescr->TOMASD_ExprIndex);
		break;
	    }

	    jstrace(TRACETYPE_TOKPATMAT, 4, "\n");
	}
    }

    jstrace(TRACETYPE_TOKPATMAT, 4,
	    "DEFINITIONS\n");
    for(i = 0 ; i < topa->TOPA_NDefinitions ; ++i) {
	int start, end;

	start = topa->TOPA_Definitions[i].TOPD_Start;
	end = topa->TOPA_Definitions[i].TOPD_End;

	jstrace(TRACETYPE_TOKPATMAT, 4,
		"  STARTSTATE %d [%d, %d] '",
	       findStateNr(topa, topa->TOPA_Definitions[i].TOPD_StartState),
	       start, end);
	if(start != -1)
	    printTokens(&(topa->TOPA_StateQueue), start, end);
	jstrace(TRACETYPE_TOKPATMAT, 4,
		"'\n");
    }
}

/*JS*********************************************************************
*
*************************************************************************/

static void printTokens(const jctokfifo_t *queue, int start, int end)

/************************************************************************/
{
    for( ; start < end ; start += jctokgetstep(queue)) {
	if(start >= queue->TKFF_Index) {
	    jstrace(TRACETYPE_TOKPATMAT, 4, "#");

	} else {
	    char buffer[1000];
	    int value, ttype;

	    value = queue->TKFF_Token[start].TK_Value;
	    ttype = value & TOPA_TOKEN_TMASK;
	    if(ttype == TOPA_TOKEN_EXPRESSION)
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"<V#%ld>",
			queue->TKFF_Token[start].TK_Value & TOPA_TOKEN_FMASK);
	    else if(ttype == TOPA_TOKEN_LINENR)
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"<LINE>");
	    else if(ttype == TOPA_TOKEN_COUNTER)
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"<C#%ld>",
			queue->TKFF_Token[start].TK_Value & TOPA_TOKEN_FMASK);
	    else if(value == CTK_BLANKSEOL)
		jstrace(TRACETYPE_TOKPATMAT, 4,
			" ");
	    else {
		sprintCToken(buffer, queue->TKFF_Token[start]);
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"%s", buffer);
	    }
	}
    }
}

/*JS*********************************************************************
*
*************************************************************************/

static int findStateNr(const tokpattmatch_t *topa, const tomastatedescr_t *state)

/************************************************************************/
{
    topamemory_t *stmem;
    unsigned int i, nr;

    nr = 0;
    for(stmem = topa->TOPA_StateMemory ; stmem ; stmem = stmem->TOPM_Next) {
	for(i = 0 ; i < stmem->TOPM_Index ; ++i) {
	    if(state == &(stmem->TOPM_Memory[i]))
		return nr;

	    ++nr;
	}
    }

    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static void showMatcherStates(const tokpattmatch_t *topa, const matcherstate_t *head,
			      const jctokfifo_t *iQueue)

/************************************************************************/
{
    const matcherstate_t *current;

    jstrace(TRACETYPE_TOKPATMAT, 4,
	    "CURRENT MATCHERSTATES:\n");
    for(current = head ; current ; current = current->MAST_Next) {
	unsigned int i;

	jstrace(TRACETYPE_TOKPATMAT, 4,
		"  MATSTATE DESC #%d IND %d",
		findStateNr(topa, current->MAST_StateDescription),
		current->MAST_TokenIndex);

	if(topa->TOPA_MaxLoopCounter > 0) {
	    jstrace(TRACETYPE_TOKPATMAT, 4,
		    " LOOP");
	    for(i = 0 ; i < topa->TOPA_MaxLoopCounter ; ++i)
		jstrace(TRACETYPE_TOKPATMAT, 4,
			"%c%d", i == 0 ? '[' : ',', current->MAST_LoopCounter[i]);
	    jstrace(TRACETYPE_TOKPATMAT, 4,
		    "]");
	}

	if(topa->TOPA_NVariables > 0) {
	    jstrace(TRACETYPE_TOKPATMAT, 4,
		    " VARS");
	    for(i = 0 ; i < topa->TOPA_NVariables ; ++i) {
		jstrace(TRACETYPE_TOKPATMAT, 4,
			i == 0 ? ":" : ",");
		if(current->MAST_VariableMatch[i].MASTV_Start != (unsigned int)-1) {
		    jstrace(TRACETYPE_TOKPATMAT, 4,
			    " V%d [%d,%d] '", i,
			    current->MAST_VariableMatch[i].MASTV_Start,
			    current->MAST_VariableMatch[i].MASTV_End);
		    printTokens(iQueue, current->MAST_VariableMatch[i].MASTV_Start,
				current->MAST_VariableMatch[i].MASTV_End);
		    jstrace(TRACETYPE_TOKPATMAT, 4,
			    "'");
		} else
		    jstrace(TRACETYPE_TOKPATMAT, 4,
			    " V%d", i);
	    }
	}
	jstrace(TRACETYPE_TOKPATMAT, 4,
		"\n");
    }
}
#endif

/*JS*********************************************************************
*
*************************************************************************/

static int putState(tokpattmatch_t *topa, matcherstate_t **pHead,
		    matcherstate_t *state)

/************************************************************************/
{
    matcherstate_t *prev, *curr;
    int nVars;

    nVars = topa->TOPA_NVariables;
    for(prev = NULL, curr = *pHead ; curr ;
	prev = curr, curr = curr->MAST_Next) {
	/*  Compare them  */
	if(curr->MAST_TokenIndex == state->MAST_TokenIndex &&
	   curr->MAST_StateDescription == state->MAST_StateDescription) {
	    topavarmatch_t *cvar, *svar;
	    int i;

	    cvar = curr->MAST_VariableMatch;
	    svar = state->MAST_VariableMatch;
	    for(i = 0 ; i < nVars ; ++i)
		if(cvar[i].MASTV_Start != svar[i].MASTV_Start ||
		   cvar[i].MASTV_End != svar[i].MASTV_End)
		    break;

	    if(i == nVars)
		break;
	}
    }

    if(curr != NULL) {
	/*  State already appears  */
	(void)freeMatcherState(topa, state);
	return 0;
    }

    /*  Add at tail  */
    if(prev == NULL)
	*pHead = state;
    else
	prev->MAST_Next = state;

    state->MAST_Next = NULL;

    return 1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int checkTokens(tokpattmatch_t *topa, jctokfifo_t *iQueue,
		       unsigned int position,
		       unsigned int fifoStart,
		       unsigned int fifoEnd, int mode)

/************************************************************************/
{
    jctokfifo_t *queue;

    queue = (mode & 1) ? iQueue : &(topa->TOPA_StateQueue);

#ifdef TRACETYPE_TOKPATMAT
    jstrace(TRACETYPE_TOKPATMAT, 4,
	    "    CHECKTOKENS:\n");
#endif

    while(fifoStart < fifoEnd) {
	const jsctoken_t *inputToken, *token;

	if((inputToken = indexedToken(topa, iQueue, position)) == NULL)
	    EGOTO(error);

	/*  Skip blank tokens in the input  */
	if(CTK_BLANK_START <= inputToken->TK_Value &&
	   inputToken->TK_Value <= CTK_BLANK_END) {
	    position += jctokgetstep(iQueue);
	    continue;
	}

	token = &(queue->TKFF_Token[fifoStart]);

	if(CTK_BLANK_START <= token->TK_Value &&
	   token->TK_Value <= CTK_BLANK_END) {
	    fifoStart += jctokgetstep(queue);
	    continue;
	}

#ifdef TRACETYPE_TOKPATMAT
	if(jstracelevel(TRACETYPE_TOKPATMAT) > 4) {
	    char ibuffer[1000], tbuffer[1000];

	    sprintCToken(ibuffer, *inputToken);
	    sprintCToken(tbuffer, *token);
	    jstrace(TRACETYPE_TOKPATMAT, 4,
		    "      INPUT %d '%s' <=> PATTERN %d '%s'\n",
		    position, ibuffer, fifoStart, tbuffer);
	}
#endif

	if(inputToken->TK_Value != token->TK_Value ||
	   (inputToken->TK_String != NULL &&
	    tkStrcmp(inputToken->TK_String, token->TK_String))) {
	    /*  no match  */
#ifdef TRACETYPE_TOKPATMAT
	    jstrace(TRACETYPE_TOKPATMAT, 4,
		    "      -> not matched\n");
#endif
	    return 0;
	}

	position += jctokgetstep(iQueue);
	fifoStart += jctokgetstep(queue);
    }

#ifdef TRACETYPE_TOKPATMAT
    jstrace(TRACETYPE_TOKPATMAT, 4,
	    "      -> matching up to position %d\n", position);
#endif

    return position;
 error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int getExprMatcher(tokpattmatch_t *topa, jctokfifo_t *iQueue,
			  matcherstate_t *current, unsigned int varIndex,
			  matcherstate_t **pHead, int mode)

/************************************************************************/
{
    unsigned int braCount[3];
    const jsctoken_t *token;
    tomastatedescr_t *descr;
    unsigned int start, end, count, flags;

    descr = current->MAST_StateDescription;

    braCount[0] = braCount[1] = braCount[2] = 0;
    count = 0;

    flags = 0; /*  bit 0: seen a non blank; bit 1: first token is curly bracket  */
    start = current->MAST_TokenIndex;
    for(end = start ;; ++end) {
	if((token = indexedToken(topa, iQueue, end)) == NULL)
	    EGOTO(error);

	if(token->TK_Value == CTK_EOF)
	    break;

	if(CTK_BLANK_START <= token->TK_Value && token->TK_Value <= CTK_BLANK_END)
	    /*  No need to have separate states that only differ by blanks  */
	    continue;

	if(token->TK_Value == '(') {
	    braCount[0]++;

	} else if(token->TK_Value == ')') {
	    if(braCount[0]-- == 0)
		break;

	} else if(token->TK_Value == '[') {
	    braCount[1]++;

	} else if(token->TK_Value == ']') {
	    if(braCount[1]-- == 0)
		break;

	} else if(token->TK_Value == '{') {
	    braCount[2]++;

	    if(!(flags & 1)) flags |= 2;

	} else if(token->TK_Value == '}') {
	    if(braCount[2]-- == 0)
		break;
	}

	flags |= 1;

	/*  Only complete expressions can make up a variable  */
	if(braCount[0] == 0 && braCount[1] == 0 && braCount[2] == 0) {
	    matcherstate_t *newState;

	    /*  Expressions can be separated by commas as well  */
	    if(!(mode & 1) && token->TK_Value == ',')
		break;

	    if((newState = getMatcherState(topa, descr, end + 1, current)) == NULL)
		EGOTO(error);

	    if(varIndex != (unsigned int)-1) {
		newState->MAST_VariableMatch[varIndex].MASTV_Start = start;
		newState->MAST_VariableMatch[varIndex].MASTV_End = end + jctokgetstep(iQueue);
	    }

	    /*  push it  */
	    putState(topa, pHead, newState);
	    ++count;

	    /*  Heuristic method to recognize end of one expression/statement  */
	    if(token->TK_Value == ';' || (flags & 2))
		/*  A fully wrapped compound statement  */
		break;
	}
    }

    return count;
error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int getTokenMatcher(tokpattmatch_t *topa, jctokfifo_t *iQueue,
			   matcherstate_t *current, int varIndex,
			   matcherstate_t **pHead, int mode)

/************************************************************************/
{
    matcherstate_t *newState;
    tomastatedescr_t *descr;
    const jsctoken_t *token;
    unsigned int end;

    descr = current->MAST_StateDescription;

    for(end = current->MAST_TokenIndex ; ; ++end) {
	if((token = indexedToken(topa, iQueue, end)) == NULL)
	    EGOTO(error);

	if(token->TK_Value < CTK_BLANK_START || CTK_BLANK_END < token->TK_Value ||
	   token->TK_Value == CTK_EOF)
	    break;
    }

    switch(descr->TOMASD_Type) {
    case TOMASD_TYPE_IDENT:
	if(token->TK_Value != CTK_IDENT)
	    return 0;

	break;

    case TOMASD_TYPE_NUMBER:
	if(token->TK_Value != CTK_NUMBER && token->TK_Value != CTK_INUMBER &&
	   token->TK_Value != CTK_DNUMBER)
	    return 0;

	break;

    case TOMASD_TYPE_STRING:
	if(token->TK_Value != CTK_STRING && token->TK_Value != CTK_WSTRING &&
	   token->TK_Value != CTK_USTRING &&
	   token->TK_Value != CTK_CHARCONST && token->TK_Value != CTK_UCHARCONST)
	    return 0;

	break;
    }

    if((newState = getMatcherState(topa, descr, end + 1, current)) == NULL)
	EGOTO(error);

    if(varIndex != -1) {
	newState->MAST_VariableMatch[varIndex].MASTV_Start = current->MAST_TokenIndex;
	newState->MAST_VariableMatch[varIndex].MASTV_End = end + jctokgetstep(iQueue);
    }

    putState(topa, pHead, newState);

    return 1;
 error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static tomastatedescr_t *getStateDescription(tokpattmatch_t *topa, unsigned int type)

/************************************************************************/
{
    tomastatedescr_t *descr;
    topamemory_t *stmem;

    for(stmem = topa->TOPA_StateMemory ; stmem ; stmem = stmem->TOPM_Next) {
	if(stmem->TOPM_Index <
	   sizeof(stmem->TOPM_Memory) / sizeof(stmem->TOPM_Memory[0])) {
	    break;
	}
    }

    if(stmem == NULL) {
	if((stmem = (topamemory_t *)malloc(sizeof(*stmem))) == NULL)
	    EGOTO(error);
	stmem->TOPM_Index = 0;
	stmem->TOPM_Next = topa->TOPA_StateMemory;
	topa->TOPA_StateMemory = stmem;
    }

    descr = &(stmem->TOPM_Memory[stmem->TOPM_Index++]);

    descr->TOMASD_Type = type;
    descr->TOMASD_Next = NULL;
    descr->TOMASD_Connect = NULL;

    return descr;
error:
    return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static matcherstate_t *getMatcherState(tokpattmatch_t *topa,
				       tomastatedescr_t *stateDescr,
				       int tokenIndex, matcherstate_t *old)

/************************************************************************/
{
    matcherstate_t *matState;
    unsigned int i, nVars, nLoop;

    nVars = topa->TOPA_NVariables;
    nLoop = topa->TOPA_MaxLoopCounter;

    if((matState = topa->TOPA_MatcherStates) != NULL)
	topa->TOPA_MatcherStates = matState->MAST_Next;
    else if((matState = (matcherstate_t *)
	     malloc(sizeof(*matState) + (nVars > 0 ? nVars - 1 : 0) *
		    sizeof(matState->MAST_VariableMatch[0]) +
		    nLoop * sizeof(*(matState->MAST_LoopCounter)))) == NULL)
	EGOTO(error);

    /*  Initialize  */
    matState->MAST_LoopCounter = (unsigned int *) &(matState->MAST_VariableMatch[nVars]);
    matState->MAST_TokenIndex = tokenIndex;
    matState->MAST_Next = NULL;
    matState->MAST_StateDescription = stateDescr;
    if(old != NULL) {
	for(i = 0 ; i < nVars ; ++i) {
	    matState->MAST_VariableMatch[i].MASTV_Start =
		old->MAST_VariableMatch[i].MASTV_Start;
	    matState->MAST_VariableMatch[i].MASTV_End =
		old->MAST_VariableMatch[i].MASTV_End;
	}

	for(i = 0 ; i < nLoop ; ++i)
	    matState->MAST_LoopCounter[i] = old->MAST_LoopCounter[i];

    } else {
	/*  Properly set up variables to match nothing which is also the
	 *   right result for e. g. <optional> tags
	 */
	for(i = 0 ; i < nVars ; ++i) {
	    matState->MAST_VariableMatch[i].MASTV_Start = -1;
	    matState->MAST_VariableMatch[i].MASTV_End   = -1;
	}

	for(i = 0 ; i < nLoop ; ++i)
	    matState->MAST_LoopCounter[i] = 0;
    }

    return matState;
error:
    return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static matcherstate_t *freeMatcherState(tokpattmatch_t *topa,
					matcherstate_t *state)

/************************************************************************/
{
    matcherstate_t *next;

    next = state->MAST_Next;

    state->MAST_Next =  topa->TOPA_MatcherStates;
    topa->TOPA_MatcherStates = state;

    return next;
}

/*JS*********************************************************************
*
*************************************************************************/

static const jsctoken_t *indexedToken(tokpattmatch_t *topa,
				      jctokfifo_t *iQueue, int tokIndex)

/************************************************************************/
{
    jsctoken_t token;
    static jsctoken_t EOFToken = CTOKEN_GENSTATIC(CTK_EOF, NULL);

    while(iQueue->TKFF_Index <= tokIndex) {
	token = GETCTOKEN(topa->TOPA_Input);
	if(token.TK_Value == CTK_ERROR)
	    EGOTO(error);

	if(jctokinsert(iQueue, token, 0) == -1)
	    EGOTO(error);
    }

    if(tokIndex < iQueue->TKFF_Index)
	return &(iQueue->TKFF_Token[tokIndex]);

    return &EOFToken;
error:
    return NULL;
}
