/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application for preprocessor library.
*
*    Part    : Check for unused preprocessor definitions
*
*              Copyright (C) 2003-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include "jscpp.h"

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSCPP_MODE_PPUNUSED     "ppunused"
/*  END-DEFINITIONS  */

typedef struct {
  char *Ident;
  int Used;
} idused_t;

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doPPUnUsed(jsppenviron_t *ppenv, const char *names[],
				    int nnames, FILE *fpOut, int mode);
static int               findIdent(idused_t *table, int n, const char *ident);

/*********     GLOBAL VARIABLES                                 *********/

/*JS*********************************************************************
*   mode bit 0: Include also definitions from user header.
*        bit 1: Include also definitions from system header.
*************************************************************************/

int doPPUnUsed(jsppenviron_t *ppenv, const char *names[], int nnames,
	       FILE *fpOut, int mode)

/************************************************************************/
{
  idused_t *table;
  int ntab, mtab, i;

  table = NULL;
  ntab = mtab = 0;

  for(i = 0 ; i < nnames ; ++i) {
    jsppstream_t *jps;
    jsctoken_t token;
    int j;

    /*printf("SCAN '%s'\n",names[i]);*/

    if((jps = getPreProcStream(ppenv, names[i], 0)) == NULL)
      EGOTO(error);

    /*  Read and discard tokens  */
    do {
      token = jsppgettoken(jps, NULL, 0);

      if(token.TK_Value == CTK_ERROR)
	EGOTO(error);
    } while(token.TK_Value != CTK_EOF);

    for(j = 0 ; j < JSPP_MACTABSIZE ; ++j) {
      _jsppmacentry_t *entry;

      for(entry = jps->JSPP_Macros[j] ; entry ; entry = entry->ME_Next) {
	char *ident;
	int pos;

	/*  Skip internal definitions, including "__FILE__" etc.  */
	if(entry->ME_Flags & MAFLAG_INTERNAL)
	  continue;

	/*  Skip definitions from system headers  */
	if(!(mode & 2) && (entry->ME_Flags & MAFLAG_SYSTEMHEADER))
	  continue;

	/*  Skip definitions from user headers  */
	if(!(mode & 1) && (entry->ME_Flags & MAFLAG_USERHEADER))
	  continue;

	ident = entry->ME_Ident;
	pos = findIdent(table, ntab, ident);

	/*  Add to list if not already there  */
	if(pos == ntab || strcmp(table[pos].Ident, ident) != 0) {
	  if(ntab >= mtab) {
	    REALLOC_RESIZE(mtab, 100);
	    if((table = (idused_t *)realloc(table, mtab * sizeof(*table)))
	       == NULL)
	      EGOTO(error);
	  }

	  if(pos < ntab)
	    memmove(&table[pos + 1], &table[pos], (ntab - pos) *
		    sizeof(*table));
	  ++ntab;

	  /*  Initialize  */
	  if((table[pos].Ident = strdup(entry->ME_Ident)) == NULL)
	    EGOTO(error);

	  table[pos].Used = 0;
	}

	/*  IDENTIFIER USED  */
	if(entry->ME_Flags & MAFLAG_USED)
	  table[pos].Used = 1;
      }
    }

    if(closePreProcStream(jps, 0) == -1)
      EGOTO(error);
  }

  if(table) {
    for(i = 0 ; i < ntab ; ++i) {
      if(!table[i].Used)
	fprintf(fpOut, "%s\n", table[i].Ident);

      free(table[i].Ident);
    }

    free(table);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int findIdent(idused_t *table, int n, const char *ident)

/************************************************************************/
{
  int lower, upper;

  /*  Binary search  */
  lower = 0;   upper = n;
  while(lower < upper) {
    int m = (lower + upper) >> 1;
    if(strcmp(table[m].Ident, ident) < 0)
      lower = m + 1;
    else
      upper = m;
  }

  return lower;
}
