/*JS*********************************************************************
 *
 *    Program : SHOWSTMTSTEP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Converts compressed statement tracing format of jscpp
 *              into more readable format.
 *
 *              Copyright (C) 2003-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

/*********     DEFINES                                          *********/
#define MAXLINE     1000

typedef struct {
    char File[MAXLINE];
    int LineNr;
} stepstatus_t;

/*********     PROTOTYPES                                       *********/
static int updateStatus(stepstatus_t *status, const char *line);

/*JS*********************************************************************
 *
 *************************************************************************/

int main(int argc, char *argv[])

/************************************************************************/
{
    stepstatus_t status;
    char line[MAXLINE];
    FILE *fp;
    int linenr;

    if(argc != 2) {
	fprintf(stderr, "Usage: showstmtstep <tracefile>\n");
	return 5;
    }

    if((fp = fopen(argv[1], "r")) == NULL)
	goto syserror;

    status.File[0] = '\0';
    status.LineNr = 0;

    for(linenr = 1 ; fgets(line, sizeof(line) - 1, fp) != NULL ; ++linenr) {
	int l;

	/*  Delete newline  */
	if((l = strlen(line)) > 0)
	    line[l - 1] = '\0';

	if((l = updateStatus(&status, line)) == -1) {
	    printf("WARNING: Invalid line '%s'\n", line);
	    continue;
	}

	if(l == 0)
	    printf("%s\n", line);
	else if(status.File[0] != '\0')
	    printf("%-15s %5d\n", status.File, status.LineNr);
    }

    if(fclose(fp) == EOF)
	goto syserror;

    return 0;
 syserror:
    perror("showstmtstep");
    return 30;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int updateStatus(stepstatus_t *status, const char *line)

/************************************************************************/
{
    const char *s;

    if(*line == '#')
	return 0;

    if(isdigit(*line)) {
	/*  Skip initial number + colon + spaces  */
	while(*line && isdigit(*line)) ++line;
	if(*line == ':') ++line;
    }
    while(*line && isspace(*line)) ++line;

    if(*line == '@') {
	status->LineNr = strtol(line + 1, (char **)&s, 10);
	if(s == (line + 1))
	    goto error;

	if(*s == ' ') ++s;
	if(*s != '\0')
	    strcpy(status->File, s);
#if 0
    } else if(*line == '&') {
	strcpy(status->Routine, line + 1);
#endif
    } else if(*line == '+' || *line == '-') {
	int diff;

	diff = strtol(line + 1, (char **)&s, 10);
	if(s == (line + 1))
	    goto error;

	if(*line == '+')
	    status->LineNr += diff;
	else
	    status->LineNr -= diff;
    } else
	return 0;

    return 1;
 error:
    return -1;
}
