/*JS*********************************************************************
*
*    Program : SUPPWRAPMATH
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Wrapper function for standard library math routines.
*
*              Copyright (C) 2005-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <math.h>

/*********     DEFINES                                          *********/
#define WRAPFUNCTION(name)   _jswrap_##name

/*********     PROTOTYPES                                       *********/
extern void _jstracefunc(const char *,const char *, ...);

/*JS*********************************************************************
*   Functions from <math.h>
*************************************************************************/

double WRAPFUNCTION(fabs)(double x)
{
  double retvalue;

  _jstracefunc("+fabs","d",x);
  retvalue =  fabs(x);
  _jstracefunc("-fabs","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(sqrt)(double x)
{
  double retvalue;

  _jstracefunc("+sqrt","d",x);
  retvalue =  sqrt(x);
  _jstracefunc("-sqrt","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(exp)(double x)
{
  double retvalue;

  _jstracefunc("+exp","d",x);
  retvalue =  exp(x);
  _jstracefunc("-exp","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(log)(double x)
{
  double retvalue;

  _jstracefunc("+log","d",x);
  retvalue =  log(x);
  _jstracefunc("-log","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(log10)(double x)
{
  double retvalue;

  _jstracefunc("+log10","d",x);
  retvalue =  log10(x);
  _jstracefunc("-log10","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(frexp)(double x, int *ex)
{
  double retvalue;

  _jstracefunc("+frexp","dp",x,ex);
  retvalue =  frexp(x, ex);
  _jstracefunc("-frexp","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(sin)(double x)
{
  double retvalue;

  _jstracefunc("+sin","d",x);
  retvalue =  sin(x);
  _jstracefunc("-sin","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(cos)(double x)
{
  double retvalue;

  _jstracefunc("+cos","d",x);
  retvalue =  cos(x);
  _jstracefunc("-cos","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(tan)(double x)
{
  double retvalue;

  _jstracefunc("+tan","d",x);
  retvalue =  tan(x);
  _jstracefunc("-tan","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(asin)(double x)
{
  double retvalue;

  _jstracefunc("+asin","d",x);
  retvalue =  asin(x);
  _jstracefunc("-asin","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(acos)(double x)
{
  double retvalue;

  _jstracefunc("+acos","d",x);
  retvalue =  acos(x);
  _jstracefunc("-acos","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(atan)(double x)
{
  double retvalue;

  _jstracefunc("+atan","d",x);
  retvalue =  atan(x);
  _jstracefunc("-atan","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(sinh)(double x)
{
  double retvalue;

  _jstracefunc("+sinh","d",x);
  retvalue =  sinh(x);
  _jstracefunc("-sinh","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(cosh)(double x)
{
  double retvalue;

  _jstracefunc("+cosh","d",x);
  retvalue =  cosh(x);
  _jstracefunc("-cosh","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(tanh)(double x)
{
  double retvalue;

  _jstracefunc("+tanh","d",x);
  retvalue =  tanh(x);
  _jstracefunc("-tanh","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(atan2)(double x, double y)
{
  double retvalue;

  _jstracefunc("+atan2","dd",x,y);
  retvalue =  atan2(x, y);
  _jstracefunc("-atan2","d",retvalue);
  return retvalue;
}
