/*JS*********************************************************************
*
*    Program : SUPPWRAPSTD1
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Wrapper function for standard library routines
*
*              Copyright (C) 2005-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdarg.h>

#include <stdlib.h>
#include <string.h>

#include <ctype.h>
#include <wctype.h>

/*********     DEFINES                                          *********/
#define WRAPFUNCTION(name)   _jswrap_##name

/*********     PROTOTYPES                                       *********/
extern void _jstracefunc(const char *,const char *, ...);

/*JS*********************************************************************
*   Functions from <stdio.h>
*************************************************************************/

FILE *WRAPFUNCTION(fopen)(const char *name, const char *mode)
{
  FILE *retvalue;

  _jstracefunc("+fopen","ss",name,mode);
  retvalue = fopen(name, mode);
  _jstracefunc("-fopen","p",retvalue);
  return retvalue;
}

FILE *WRAPFUNCTION(freopen)(const char *path, const char *mode, FILE *stream)
{
  FILE *retvalue;

  _jstracefunc("+freopen","ssp",path,mode,stream);
  retvalue =  freopen(path, mode, stream);
  _jstracefunc("-freopen","p",retvalue);
  return retvalue;
}

int WRAPFUNCTION(fflush)(FILE *stream)
{
  int retvalue;

  _jstracefunc("+fflush","p",stream);
  retvalue =  fflush(stream);
  _jstracefunc("-fflush","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(fclose)(FILE *fp)
{
  int retvalue;

  _jstracefunc("+fclose","p",fp);
  retvalue =  fclose(fp);
  _jstracefunc("-fclose","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(remove)(const char *pathname)
{
  int retvalue;

  _jstracefunc("+remove","s",pathname);
  retvalue =  remove(pathname);
  _jstracefunc("-remove","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(rename)(const char *oldpath, const char *newpath)
{
  int retvalue;

  _jstracefunc("+rename","ss",oldpath,newpath);
  retvalue =  rename(oldpath, newpath);
  _jstracefunc("-rename","i",retvalue);
  return retvalue;
}

FILE *WRAPFUNCTION(tmpfile) (void)
{
  FILE *retvalue;

  _jstracefunc("+tmpfile","");
  retvalue =  tmpfile();
  _jstracefunc("-tmpfile","p",retvalue);
  return retvalue;
}

void WRAPFUNCTION(setbuf)(FILE *stream, char *buf)
{
  _jstracefunc("+setbuf","ps",stream,buf);
  setbuf(stream, buf);
  _jstracefunc("-setbuf","");
}

void WRAPFUNCTION(setbuffer)(FILE *stream, char *buf, size_t size)
{
  _jstracefunc("+setbuffer","psl",stream,buf,size);
  setbuffer(stream, buf, size);
  _jstracefunc("-setbuffer","");
}

void WRAPFUNCTION(setlinebuf)(FILE *stream)
{
  _jstracefunc("+setlinebuf","p",stream);
  setlinebuf(stream);
  _jstracefunc("-setlinebuf","");
}

int WRAPFUNCTION(setvbuf)(FILE *stream, char *buf, int mode, size_t size)
{
  int retvalue;

  _jstracefunc("+setvbuf","psil",stream,buf,mode,size);
  retvalue =  setvbuf(stream, buf, mode, size);
  _jstracefunc("-setvbuf","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(fgetc)(FILE *stream)
{
  int retvalue;

  _jstracefunc("+fgetc","p",stream);
  retvalue =  fgetc(stream);
  _jstracefunc("-fgetc","i",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(fgets)(char *s, int size, FILE *stream)
{
  char *retvalue;

  _jstracefunc("+fgets","sip",s,size,stream);
  retvalue =  fgets(s, size, stream);
  _jstracefunc("-fgets","s",retvalue);
  return retvalue;
}

int WRAPFUNCTION(getc)(FILE *stream)
{
  int retvalue;

  _jstracefunc("+getc","p",stream);
  retvalue =  getc(stream);
  _jstracefunc("-getc","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(getchar)(void)
{
  int retvalue;

  _jstracefunc("+getchar","");
  retvalue =  getchar();
  _jstracefunc("-getchar","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(ungetc)(int c, FILE *stream)
{
  int retvalue;

  _jstracefunc("+ungetc","ip",c,stream);
  retvalue =  ungetc(c, stream);
  _jstracefunc("-ungetc","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(printf)(const char *format, ...)
{
  va_list ap;
  int ret;

  _jstracefunc("+printf","s",format);
  va_start(ap, format);
  ret = vprintf(format, ap);
  va_end(ap);

  _jstracefunc("-printf","i",ret);
  return ret;
}

int WRAPFUNCTION(fprintf)(FILE *fp, const char *format, ...)
{
  va_list ap;
  int ret;

  _jstracefunc("+fprintf","ps",fp,format);
  va_start(ap, format);
  ret = vfprintf(fp, format, ap);
  va_end(ap);

  _jstracefunc("-fprintf","i",ret);
  return ret;
}

int WRAPFUNCTION(sprintf)(char *dest, const char *format, ...)
{
  va_list ap;
  int ret;

  _jstracefunc("+sprintf","ss",dest,format);
  va_start(ap, format);
  ret = vsprintf(dest, format, ap);
  va_end(ap);

  _jstracefunc("-sprintf","i",ret);
  return ret;
}

int WRAPFUNCTION(vprintf)(const char *format, va_list ap)
{
  int retvalue;

  _jstracefunc("+vprintf","s",format);
  retvalue =  vprintf(format, ap);
  _jstracefunc("-vprintf","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(vfprintf)(FILE *stream, const char *format, va_list ap)
{
  int retvalue;

  _jstracefunc("+vfprintf","ps",stream,format);
  retvalue =  vfprintf(stream, format, ap);
  _jstracefunc("-vfprintf","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(vsprintf)(char *str, const char *format, va_list ap)
{
  int retvalue;

  _jstracefunc("+vsprintf","ss",str,format);
  retvalue =  vsprintf(str, format, ap);
  _jstracefunc("-vsprintf","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(fputc)(int c, FILE *stream)
{
  int retvalue;

  _jstracefunc("+fputc","ip",c,stream);
  retvalue =  fputc(c, stream);
  _jstracefunc("-fputc","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(fputs)(const char *s, FILE *stream)
{
  int retvalue;

  _jstracefunc("+fputs","sp",s,stream);
  retvalue =  fputs(s, stream);
  _jstracefunc("-fputs","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(putc)(int c, FILE *stream)
{
  int retvalue;

  _jstracefunc("+putc","ip",c,stream);
  retvalue =  putc(c, stream);
  _jstracefunc("-putc","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(putchar)(int c)
{
  int retvalue;

  _jstracefunc("+putchar","i",c);
  retvalue =  putchar(c);
  _jstracefunc("-putchar","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(puts)(const char *s)
{
  int retvalue;

  _jstracefunc("+puts","s",s);
  retvalue =  puts(s);
  _jstracefunc("-puts","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(scanf)(const char *format, ...)
{
  va_list ap;
  int ret;

  _jstracefunc("+scanf","s",format);
  va_start(ap, format);
  ret = vscanf(format, ap);
  va_end(ap);

  _jstracefunc("-scanf","i",ret);
  return ret;
}

int WRAPFUNCTION(fscanf)(FILE *fp, const char *format, ...)
{
  va_list ap;
  int ret;

  _jstracefunc("+fscanf","ps",fp,format);
  va_start(ap, format);
  ret = vfscanf(fp, format, ap);
  va_end(ap);

  _jstracefunc("-fscanf","i",ret);
  return ret;
}

int WRAPFUNCTION(sscanf)(const char *str, const char *format, ...)
{
  va_list ap;
  int ret;

  _jstracefunc("+sscanf","ss",str,format);
  va_start(ap, format);
  ret = vsscanf(str, format, ap);
  va_end(ap);

  _jstracefunc("-sscanf","i",ret);
  return ret;
}

int WRAPFUNCTION(vscanf)(const char *format, va_list ap)
{
  int retvalue;

  _jstracefunc("+vscanf","s",format);
  retvalue =  vscanf(format, ap);
  _jstracefunc("-vscanf","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(vsscanf)(const char *str, const char *format, va_list ap)
{
  int retvalue;

  _jstracefunc("+vsscanf","ss",str,format);
  retvalue =  vsscanf(str, format, ap);
  _jstracefunc("-vsscanf","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(vfscanf)(FILE *stream, const char *format, va_list ap)
{
  int retvalue;

  _jstracefunc("+vfscanf","ps",stream,format);
  retvalue =  vfscanf(stream, format, ap);
  _jstracefunc("-vfscanf","i",retvalue);
  return retvalue;
}

size_t WRAPFUNCTION(fread)(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
  size_t retvalue;

  _jstracefunc("+fread","pllp",ptr,size,nmemb,stream);
  retvalue =  fread(ptr, size, nmemb, stream);
  _jstracefunc("-fread","l",retvalue);
  return retvalue;
}

size_t WRAPFUNCTION(fwrite)(const void *ptr, size_t size, size_t nmemb, FILE *stream)
{
  size_t retvalue;

  _jstracefunc("+fwrite","pllp",ptr,size,nmemb,stream);
  retvalue =  fwrite(ptr, size, nmemb, stream);
  _jstracefunc("-fwrite","l",retvalue);
  return retvalue;
}

int WRAPFUNCTION(fseek)(FILE *stream, long offset, int whence)
{
  int retvalue;

  _jstracefunc("+fseek","pli",stream,offset,whence);
  retvalue =  fseek(stream, offset, whence);
  _jstracefunc("-fseek","i",retvalue);
   return retvalue;
}

long WRAPFUNCTION(ftell)(FILE *stream)
{
  long retvalue;

  _jstracefunc("+ftell","p",stream);
  retvalue =  ftell(stream);
  _jstracefunc("-ftell","l",retvalue);
  return retvalue;
}

void WRAPFUNCTION(rewind)(FILE *stream)
{
  _jstracefunc("+rewind","p",stream);
  rewind(stream);
  _jstracefunc("-rewind","");
}

int WRAPFUNCTION(fgetpos)(FILE *stream, fpos_t *pos)
{
  int retvalue;

  _jstracefunc("+fgetpos","pp",stream,pos);
  retvalue =  fgetpos(stream, pos);
  _jstracefunc("-fgetpos","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(fsetpos)(FILE *stream, fpos_t *pos)
{
  int retvalue;

  _jstracefunc("+fsetpos","pp",stream,pos);
  retvalue =  fsetpos(stream, pos);
  _jstracefunc("-fsetpos","i",retvalue);
  return retvalue;
}

void WRAPFUNCTION(clearerr)(FILE *stream)
{
  _jstracefunc("+clearerr","p",stream);
  clearerr(stream);
  _jstracefunc("-clearerr","");
}

int WRAPFUNCTION(feof)(FILE *stream)
{
  int retvalue;

  _jstracefunc("+feof","p",stream);
  retvalue =  feof(stream);
  _jstracefunc("-feof","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(ferror)(FILE *stream)
{
  int retvalue;

  _jstracefunc("+ferror","p",stream);
  retvalue =  ferror(stream);
  _jstracefunc("-ferror","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(fileno)(FILE *stream)
{
  int retvalue;

  _jstracefunc("+fileno","p",stream);
  retvalue =  fileno(stream);
  _jstracefunc("-fileno","i",retvalue);
  return retvalue;
}

/*JS*********************************************************************
*   Functions from <stdlib.h>
*************************************************************************/

int WRAPFUNCTION(putenv)(char *string)
{
  int retvalue;

  _jstracefunc("+putenv","s",string);
  retvalue =  putenv(string);
  _jstracefunc("-putenv","i",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(getenv)(const char *name)
{
  char *retvalue;

  _jstracefunc("+getenv","s",name);
  retvalue =  getenv(name);
  _jstracefunc("-getenv","s",retvalue);
  return retvalue;
}

void *WRAPFUNCTION(calloc)(size_t nmemb, size_t size)
{
  void *retvalue;

  _jstracefunc("+calloc","ll",nmemb,size);
  retvalue =  calloc(nmemb, size);
  _jstracefunc("-calloc","p",retvalue);
  return retvalue;
}

void *WRAPFUNCTION(malloc)(size_t size)
{
  void *retvalue;

  _jstracefunc("+malloc","l",size);
  retvalue =  malloc(size);
  _jstracefunc("-malloc","p",retvalue);
  return retvalue;
}

void WRAPFUNCTION(free)(void *ptr)
{
  _jstracefunc("+free","p",ptr);
  free(ptr);
  _jstracefunc("-free","");
}

void *WRAPFUNCTION(realloc)(void *ptr, size_t size)
{
  void *retvalue;

  _jstracefunc("+realloc","pl",ptr,size);
  retvalue = realloc(ptr, size);
  _jstracefunc("-realloc","p",retvalue);
  return retvalue;
}

unsigned long int WRAPFUNCTION(strtol)(const char *nptr, char **endptr, int base)
{
  unsigned long int retvalue;

  _jstracefunc("+strtol","spi",nptr,endptr,base);
  retvalue =  strtol(nptr, endptr, base);
  _jstracefunc("-strtol","l",retvalue);
  return retvalue;
}

unsigned long int WRAPFUNCTION(strtoul)(const char *nptr, char **endptr, int base)
{
  unsigned long int retvalue;

  _jstracefunc("+strtoul","spi",nptr,endptr,base);
  retvalue =  strtoul(nptr, endptr, base);
  _jstracefunc("-strtoul","l",retvalue);
  return retvalue;
}

int WRAPFUNCTION(atoi)(const char *a)
{
  int retvalue;

  _jstracefunc("+atoi","s",a);
  retvalue =  atoi(a);
  _jstracefunc("-atoi","i",retvalue);
  return retvalue;
}

double WRAPFUNCTION(atof)(const char *a)
{
  double retvalue;

  _jstracefunc("+atof","s",a);
  retvalue =  atof(a);
  _jstracefunc("-atof","d",retvalue);
  return retvalue;
}

double WRAPFUNCTION(strtod)(const char *a, char **b)
{
  double retvalue;

  _jstracefunc("+strtod","sp",a,b);
  retvalue =  strtod(a, b);
  _jstracefunc("-strtod","d",retvalue);
  return retvalue;
}

/*JS*********************************************************************
*   Functions from <ctype.h>
*************************************************************************/

int WRAPFUNCTION(isalnum)(int c)
{
  int retvalue;

  _jstracefunc("+isalnum","i",c);
  retvalue =  isalnum(c);
  _jstracefunc("-isalnum","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(isalpha)(int c)
{
  int retvalue;

  _jstracefunc("+isalpha","i",c);
  retvalue =  isalpha(c);
  _jstracefunc("-isalpha","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iscntrl)(int c)
{
  int retvalue;

  _jstracefunc("+iscntrl","i",c);
  retvalue =  iscntrl(c);
  _jstracefunc("-iscntrl","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(isdigit)(int c)
{
  int retvalue;

  _jstracefunc("+isdigit","i",c);
  retvalue =  isdigit(c);
  _jstracefunc("-isdigit","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(isgraph)(int c)
{
  int retvalue;

  _jstracefunc("+isgraph","i",c);
  retvalue =  isgraph(c);
  _jstracefunc("-isgraph","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(islower)(int c)
{
  int retvalue;

  _jstracefunc("+islower","i",c);
  retvalue =  islower(c);
  _jstracefunc("-islower","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(isprint)(int c)
{
  int retvalue;

  _jstracefunc("+isprint","i",c);
  retvalue =  isprint(c);
  _jstracefunc("-isprint","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(ispunct)(int c)
{
  int retvalue;

  _jstracefunc("+ispunct","i",c);
  retvalue =  ispunct(c);
  _jstracefunc("-ispunct","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(isspace)(int c)
{
  int retvalue;

  _jstracefunc("+isspace","i",c);
  retvalue =  isspace(c);
  _jstracefunc("-isspace","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(isupper)(int c)
{
  int retvalue;

  _jstracefunc("+isupper","i",c);
  retvalue =  isupper(c);
  _jstracefunc("-isupper","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(isxdigit)(int c)
{
  int retvalue;

  _jstracefunc("+isxdigit","i",c);
  retvalue =  isxdigit(c);
  _jstracefunc("-isxdigit","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(tolower)(int c)
{
  int retvalue;

  _jstracefunc("+tolower","i",c);
  retvalue =  tolower(c);
  _jstracefunc("-tolower","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(toupper)(int c)
{
  int retvalue;

  _jstracefunc("+toupper","i",c);
  retvalue =  toupper(c);
  _jstracefunc("-toupper","i",retvalue);
  return retvalue;
}

/*JS*********************************************************************
*   Functions from <wctype.h>
*************************************************************************/

int WRAPFUNCTION(mblen)(const char *s, size_t n)
{
  int retvalue;

  _jstracefunc("+mblen","sl",s,n);
  retvalue =  mblen(s, n);
  _jstracefunc("-mblen","i",retvalue);
  return retvalue;
}

size_t WRAPFUNCTION(mbstowcs)(wchar_t *dest, const char *src, size_t n)
{
  size_t retvalue;

  _jstracefunc("+mbstowcs","wsl",dest,src,n);
  retvalue =  mbstowcs(dest, src, n);
  _jstracefunc("-mbstowcs","l",retvalue);
  return retvalue;
}

int WRAPFUNCTION(mbtowc)(wchar_t *pwc, const char *s, size_t n)
{
  int retvalue;

  _jstracefunc("+mbtowc","wsl",pwc,s,n);
  retvalue =  mbtowc(pwc, s, n);
  _jstracefunc("-mbtowc","i",retvalue);
  return retvalue;
}

size_t WRAPFUNCTION(wcstombs)(char *dest, const wchar_t *src, size_t n)
{
  size_t retvalue;

  _jstracefunc("+wcstombs","swl",dest,src,n);
  retvalue =  wcstombs(dest, src, n);
  _jstracefunc("-wcstombs","l",retvalue);
  return retvalue;
}

int WRAPFUNCTION(wctomb)(char *s, wchar_t wc)
{
  int retvalue;

  _jstracefunc("+wctomb","si",s,wc);
  retvalue =  wctomb(s, wc);
  _jstracefunc("-wctomb","i",retvalue);
  return retvalue;
}

/*JS*********************************************************************
*   Functions from <string.h>
*************************************************************************/

void *WRAPFUNCTION(memchr)(const void *s, int c, size_t n)
{
  void *retvalue;

  _jstracefunc("+memchr","pil",s,c,n);
  retvalue =  memchr(s, c, n);
  _jstracefunc("-memchr","p",retvalue);
  return retvalue;
}

#if 0
void *WRAPFUNCTION(memrchr)(const void *s, int c, size_t n)
{
  void *retvalue;

  _jstracefunc("+memrchr","pil",s,c,n);
  retvalue =  memrchr(s, c, n);
  _jstracefunc("-memrchr","p",retvalue);
  return retvalue;
}
#endif

int WRAPFUNCTION(memcmp)(const void *s1, const void *s2, size_t n)
{
  int retvalue;

  _jstracefunc("+memcmp","ppl",s1,s2,n);
  retvalue =  memcmp(s1, s2, n);
  _jstracefunc("-memcmp","i",retvalue);
  return retvalue;
}

void *WRAPFUNCTION(memcpy)(void *dest, const void *src, size_t n)
{
  void *retvalue;

  _jstracefunc("+memcpy","ppl",dest,src,n);
  retvalue =  memcpy(dest, src, n);
  _jstracefunc("-memcpy","p",retvalue);
  return retvalue;
}

void *WRAPFUNCTION(memmove)(void *dest, const void *src, size_t n)
{
  void *retvalue;

  _jstracefunc("+memmove","ppl",dest,src,n);
  retvalue =  memmove(dest, src, n);
  _jstracefunc("-memmove","p",retvalue);
  return retvalue;
}

void *WRAPFUNCTION(memset)(void *s, int c, size_t n)
{
  void *retvalue;

  _jstracefunc("+memset","pil",s,c,n);
  retvalue =  memset(s, c, n);
  _jstracefunc("-memset","p",retvalue);
  return retvalue;
}

int WRAPFUNCTION(strcmp)(const char *s1, const char *s2)
{
  int retvalue;

  _jstracefunc("+strcmp","ss",s1,s2);
  retvalue =  strcmp(s1, s2);
  _jstracefunc("-strcmp","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(strncmp)(const char *s1, const char *s2, size_t n)
{
  int retvalue;

  _jstracefunc("+strncmp","ssl",s1,s2,n);
  retvalue =  strncmp(s1, s2, n);
  _jstracefunc("-strncmp","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(strcoll)(const char *s1, const char *s2)
{
  int retvalue;

  _jstracefunc("+strcoll","ss",s1,s2);
  retvalue =  strcoll(s1, s2);
  _jstracefunc("-strcoll","i",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strcpy)(char *dest, const char *src)
{
  char *retvalue;

  _jstracefunc("+strcpy","ss",dest,src);
  retvalue =  strcpy(dest, src);
  _jstracefunc("-strcpy","s",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strncpy)(char *dest, const char *src, size_t n)
{
  char *retvalue;

  _jstracefunc("+strncpy","ssl",dest,src,n);
  retvalue =  strncpy(dest, src, n);
  _jstracefunc("-strncpy","s",retvalue);
  return retvalue;
}

size_t WRAPFUNCTION(strspn)(const char *s, const char *accept)
{
  size_t retvalue;

  _jstracefunc("+strspn","ss",s,accept);
  retvalue =  strspn(s, accept);
  _jstracefunc("-strspn","l",retvalue);
  return retvalue;
}

size_t WRAPFUNCTION(strcspn)(const char *s, const char *reject)
{
  size_t retvalue;

  _jstracefunc("+strcspn","ss",s,reject);
  retvalue =  strcspn(s, reject);
  _jstracefunc("-strcspn","l",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strerror)(int errnum)
{
  char *retvalue;

  _jstracefunc("+strerror","i",errnum);
  retvalue =  strerror(errnum);
  _jstracefunc("-strerror","s",retvalue);
  return retvalue;
}

size_t WRAPFUNCTION(strlen)(const char *s)
{
  size_t retvalue;

  _jstracefunc("+strlen","s",s);
  retvalue =  strlen(s);
  _jstracefunc("-strlen","l",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strcat)(char *dest, const char *src)
{
  char *retvalue;

  _jstracefunc("+strcat","ss",dest,src);
  retvalue =  strcat(dest, src);
  _jstracefunc("-strcat","s",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strncat)(char *dest, const char *src, size_t n)
{
  char *retvalue;

  _jstracefunc("+strncat","ssl",dest,src,n);
  retvalue =  strncat(dest, src, n);
  _jstracefunc("-strncat","s",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strpbrk)(const char *s, const char *accept)
{
  char *retvalue;

  _jstracefunc("+strpbrk","ss",s,accept);
  retvalue =  strpbrk(s, accept);
  _jstracefunc("-strpbrk","s",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strchr)(const char *s, int c)
{
  char *retvalue;

  _jstracefunc("+strchr","si",s,c);
  retvalue =  strchr(s, c);
  _jstracefunc("-strchr","s",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strrchr)(const char *s, int c)
{
  char *retvalue;

  _jstracefunc("+strrchr","si",s,c);
  retvalue =  strrchr(s, c);
  _jstracefunc("-strrchr","s",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strstr)(const char *haystack, const char *needle)
{
  char *retvalue;

  _jstracefunc("+strstr","ss",haystack,needle);
  retvalue =  strstr(haystack, needle);
  _jstracefunc("-strstr","s",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strtok)(char *s, const char *delim)
{
  char *retvalue;

  _jstracefunc("+strtok","ss",s,delim);
  retvalue =  strtok(s, delim);
  _jstracefunc("-strtok","s",retvalue);
  return retvalue;
}

size_t WRAPFUNCTION(strxfrm)(char *dest, const char *src, size_t n)
{
  size_t retvalue;

  _jstracefunc("+strxfrm","ssl",dest,src,n);
  retvalue =  strxfrm(dest, src, n);
  _jstracefunc("-strxfrm","l",retvalue);
  return retvalue;
}

char *WRAPFUNCTION(strdup)(const char *a)
{
  char *retvalue;

  _jstracefunc("+strdup","s",a);
  retvalue =  strdup(a);
  _jstracefunc("-strdup","s",retvalue);
  return retvalue;
}

/*JS*********************************************************************
*   Functions from <ctype.h>
*************************************************************************/

int WRAPFUNCTION(iswalnum)(int c)
{
  int retvalue;

  _jstracefunc("+iswalnum","i",c);
  retvalue =  iswalnum(c);
  _jstracefunc("-iswalnum","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswalpha)(int c)
{
  int retvalue;

  _jstracefunc("+iswalpha","i",c);
  retvalue =  iswalpha(c);
  _jstracefunc("-iswalpha","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswcntrl)(int c)
{
  int retvalue;

  _jstracefunc("+iswcntrl","i",c);
  retvalue =  iswcntrl(c);
  _jstracefunc("-iswcntrl","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswdigit)(int c)
{
  int retvalue;

  _jstracefunc("+iswdigit","i",c);
  retvalue =  iswdigit(c);
  _jstracefunc("-iswdigit","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswgraph)(int c)
{
  int retvalue;

  _jstracefunc("+iswgraph","i",c);
  retvalue =  iswgraph(c);
  _jstracefunc("-iswgraph","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswlower)(int c)
{
  int retvalue;

  _jstracefunc("+iswlower","i",c);
  retvalue =  iswlower(c);
  _jstracefunc("-iswlower","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswprint)(int c)
{
  int retvalue;

  _jstracefunc("+iswprint","i",c);
  retvalue =  iswprint(c);
  _jstracefunc("-iswprint","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswpunct)(int c)
{
  int retvalue;

  _jstracefunc("+iswpunct","i",c);
  retvalue =  iswpunct(c);
  _jstracefunc("-iswpunct","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswspace)(int c)
{
  int retvalue;

  _jstracefunc("+iswspace","i",c);
  retvalue =  iswspace(c);
  _jstracefunc("-iswspace","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswupper)(int c)
{
  int retvalue;

  _jstracefunc("+iswupper","i",c);
  retvalue =  iswupper(c);
  _jstracefunc("-iswupper","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(iswxdigit)(int c)
{
  int retvalue;

  _jstracefunc("+iswxdigit","i",c);
  retvalue =  iswxdigit(c);
  _jstracefunc("-iswxdigit","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(towlower)(int c)
{
  int retvalue;

  _jstracefunc("+towlower","i",c);
  retvalue =  towlower(c);
  _jstracefunc("-towlower","i",retvalue);
  return retvalue;
}

int WRAPFUNCTION(towupper)(int c)
{
  int retvalue;

  _jstracefunc("+towupper","i",c);
  retvalue =  towupper(c);
  _jstracefunc("-towupper","i",retvalue);
  return retvalue;
}
