# /*
#
#   General test file for preprocessor. Define one of the symbols
#    TEST1 ... TESTn and run through jscpp.
#
#   This file may be used as a shell script as well (for SH).
#   This file contains solely tests that apply to any ANSI preprocessor.
#

DIR=pptest
if [ ! -d $DIR ] ; then
  mkdir $DIR
fi

#  Do all tests
max=16
i=1
while [ $i -le $max ] ; do
  echo -n " $i.."
  tfile=$DIR/test$i.i
  jscpp -DTEST$i PPTest.c $* -o /tmp/tmp1$$
  sed '1d' </tmp/tmp1$$ >/tmp/tmp2$$

  if [ ! -r $tfile ] ; then
    echo "  --> Creating test file"
    cp /tmp/tmp2$$ $tfile
  elif ! cmp -s $tfile /tmp/tmp2$$ ; then
    echo ""
    diff $tfile /tmp/tmp2$$ || true
  fi
  i=`expr $i + 1`
done
echo ""
rm /tmp/tmp[12]$$
exit
# */

#if TEST1  /* --------------------------------------------------------- */
  /*  Check if empty preprocessor lines are allowed  */
#
#   /* empty  */

  /*  Test simple and complicated macro expansion  */
#  define mac1     This is a simple macro
  mac1

#  define mac2(a,b)  /* Param */  Par a and b
  mac2((1,2,3),(++,--))

#  define mac3(idi)  #idi
  mac3(  Hello world +  )
  mac3(dudel)

#  define mac4(idi)  ONE##idi
#  define ONETWO      ONEONE
  mac4(TWO) mac4(44)

#  define mac5(x)    mac6(x)
#  define mac6(x)    I like x
  mac5(Silke)

/*  Macro definition with empty argument  */
#  define mac7()     empty arguments are *no* problem!
This occurence of mac7() will be expanded,
whereas this one  mac7  not!

#elif TEST2  /* --------------------------------------------------------- */
/*   Check conditional compilation  */
#  define var1 1
#  define var2 0

#  if defined(var1) && !var2 || var2
#   define ANSWER   right
#  else
#   define ANSWER   wrong
#  endif

  The answer is: ANSWER
#elif TEST3  /* --------------------------------------------------------- */
/*   Check pragma and line counting after including  */

/*  No macro expansion should occur in pragma line */
#define unknown wonderful
#  pragma unknown pragma in this "line" /* he he  */
  this is good

/*  Test include command  */
# undef TEST3
# include "PPTest.c"
# define FILENAME "PPTest.c"
# include FILENAME

  Here we are!
#elif TEST4  /* --------------------------------------------------------- */
/*  Check expression evaluation in preprocessor lines  */

#  if 1 - 1 ? 4*5 : 6*7
    Hello
#  else
    This is bad
#  endif
#elif TEST5  /* --------------------------------------------------------- */
/*  Check predefined symbols  */
  Compiler: "__STDC__" ist __STDC__

#  ifndef __DATE__
#   error "__DATE__" is not defined!
#  endif
#  ifndef __TIME__
#   error "__TIME__" is not defined!
#  endif
  File: "__FILE__" is __FILE__
  Line: "__LINE__" is __LINE__
#elif TEST6  /* --------------------------------------------------------- */
/*  Check ANSI required special concatenation behaviour  */
#  define concat(a)    a ## ball
#  define Base         B
#  define Baseball     Sport
  Play concat(Base) a lot!

#  define cat(x,y)  x ## y
#  define xcat(x,y) cat(x,y)

  xcat(xcat(1,2),3)

/*  Macros in sub arguments should be expanded  */
#define macmac  super
  xcat(xcat(macmac,2),3)	xcat(cat(macmac,2),3)
  xcat(macmac,2)		cat(macmac,2)

#elif TEST7  /* --------------------------------------------------------- */
/*  Check generation new symbols via macro expansion  */
#  define concat(a,b) a ## b

  a = concat(+,+) b;
  a concat(<<,=) b;
#elif TEST8  /* --------------------------------------------------------- */
/*  Check correct expansion after non-expansion of macro with parameters */
#  define two(x,y)   Mr x loves y!
  Look at this: two two(Joerg,Silke)
#elif TEST9 /* --------------------------------------------------------- */
/*  Check generation of macro call with parameters through a macro
       expansion  */

  /*  The following property seems a little bit strange, but some other
      preprocessors I have tested seem to behave similar  */
#  define statement(x,y) The x is y today
#  define first_part   statement(sun,very hot
  first_part)

#  define double(x)  (2*(x))
#  define call_with_1(x)  x(1)
  call_with_1(double)

#elif TEST10 /* --------------------------------------------------------- */
/*  Check that old concatenation property doesn't work  */
#define OLDCONCAT(a,b)   a/**/b
OLDCONCAT(hello,world)
#elif TEST11 /* --------------------------------------------------------- */
/* Check that line number remains correct during macro expansion  */
#define drop2(a,b,c) we \
 drop (a) the \
 second (c) \
 argument! (c)
drop2(one,two,three)
drop2(
one
  is
  quite
  long,
  two
  also
is
very
long, three
is
rather
 short)
finally

#if __LINE__ != 181
# error This should be line 182, but it isn't!
#endif

#elif TEST12 /* --------------------------------------------------------- */
/*  Check '#' operator on complicated arguments  */
#define STR(arg)   #arg
#define XSTR(arg)  STR(arg)

#define ONE    1

STR(\n "\n" '\n')

STR("hello")
XSTR("hello")

STR(\n
bb)
XSTR(\n
bb)

STR(#)

#elif TEST13 /* --------------------------------------------------------- */
/*  Something fishy with macro expansion  */
/* Test should give "start (here is the sun ) end"  */
#define first(x)  start second(x) end
#define second(x) (here is x)
#define sun     the sun
first(sun)

#elif TEST14 /* --------------------------------------------------------- */

/*  Check that arithmetric works without overflow  */
#define MAX_LONG     0x7fffffff
#define MAX_ULONG    (2 * MAX_LONG + 1)

#if MAX_ULONG == 0xffffffff
This is good
#else
This is bad
#endif

#elif TEST15 /* --------------------------------------------------------- */
/*  Some special thing with recursion  */
#define RECUR(x)  <RECUR(x)>
#define WANDA(x) here x end

WANDA(RECUR(x))

#elif TEST16 /* --------------------------------------------------------- */

/*  Using character constants in preprocessor conditions  */
#if '1' < '2'
This is right
#else
This is wrong
#endif

#if '8' < '2'
This is wrong
#else
This is right
#endif

#endif
