# /*
#
#   General test file for function dependency. Run through "jscpp -m fdeps".
#
#   This file may be used as a shell script as well (for SH).
#

DIR=test_fdeps
if [ ! -d $DIR ] ; then
  mkdir $DIR
fi

#  Do all tests
max=5
i=1
while [ $i -le $max ] ; do
  echo -n " $i.."
  tfile=$DIR/test$i.i

  jscpp -m fdeps -pps -DTEST$i TestFDeps.c -o /tmp/tmp$$

  if [ ! -r $tfile ] ; then
    echo "  --> Creating test file"
    cp /tmp/tmp$$ $tfile
  elif ! cmp -s $tfile /tmp/tmp$$ ; then
    echo ""
    diff $tfile /tmp/tmp$$ || true
  fi
  i=`expr $i + 1`
done
echo ""
rm /tmp/tmp$$
exit
# */

#ifdef TEST1  /* ********************************************** */
typedef void *FILE;
extern FILE *stdout;
void *malloc(int size);
void free(void *);
int fputs(const char *s, FILE *stream);
char *getenv(const char *name);

struct dummy {
  int malloc;
  int free;
};

void dummy1(void)
{
  int (*routine1)(const char *,FILE *fp);
  char *(*routine2)(const char *);
  struct dummy dudu;

  routine1 = fputs;
  routine2 = getenv;

  printf("hello, world!\n");
  (*routine1)("hello, world!\n",stdout);
  printf("HOME=%s\n",routine2("HOME"));

  dudu->malloc = dudu->free = 0;
}

#elif TEST2 /* ********************************************** */

extern void fussy(const char *fmt, ...);
extern void fussybear(const char *fmt, ...);

void (*FFunc1)(const char *, ...) = fussy,
 	(*FFunc2)(const char *, ...) = fussybear;

void (*FFunc3)(const char *, ...) = fussy;

struct {
  void (*Fct1)(const char *, ...);
  void (*Fct2)(const char *, ...);
} FFuncArr[] = {
  { fussy, NULL      },
  { NULL,  fussybear }
};

void dummy2(void)
{
  void (*func)(const char *, ...) = fussy;
}

#elif TEST3 /* ********************************************** */

int GI, GJ = GI;

void dummy3(void)
{
  int k = GI;
  k = GJ;
}

struct {
  int I,J;
} IValues[] = {
  { GI, 9 },
  { 4, GJ }
};

void dummy4(void)
{
  int i,j;
  struct {
    int I,J;
  } ivalues[] = {
    { GI, 9 },
    { 4, GJ },
    { i, j }
  };
}

#elif TEST4 /* ********************************************** */

extern int printf(char *,...);
void *malloc(int);

int GI;
static int SI;

void dummy5(void)
{
   printf(SI);
}

static void dummy6(void)
{
  malloc();
}

int (*fptr1)(void);
static int (*fptr2)(void);

void dummy7(void)
{
  printf(GI);
  fptr1();
  fptr2();
  dummy2();
}

#elif TEST5 /* ********************************************** */

/*  A "static" prototype makes a function static even though
 *   it itself doesn't have the "static" keyword
 */
static void dummy9(void);

void dummy8(void)
{
  dummy9();
}

void dummy9(void)
{
  printf("hello, world!\n");
}
#endif
