# /*
#
#   General test file for function tracer. Run through "jscpp -m ftrace".
#
#   This file may be used as a shell script as well (for SH).
#

DIR=test_trace
if [ ! -d $DIR ] ; then
  mkdir $DIR
fi

#  Do all tests
max=37
i=1
while [ $i -le $max ] ; do
  echo -n " $i.."
  tfile=$DIR/test$i.i

  jscpp -x0 -CCP -DTEST$i TestFTrace.c -o /tmp/tmp0$$
  sed -e '/^ *$/d' -e 's/^@/#/g' </tmp/tmp0$$ >/tmp/tmp1$$
  jscpp -m ftrace /tmp/tmp1$$ -o /tmp/tmp2$$

  if [ ! -r $tfile ] ; then
    echo "  --> Creating test file"
    cp /tmp/tmp2$$ $tfile
  elif ! cmp -s $tfile /tmp/tmp2$$ ; then
    echo ""
    diff $tfile /tmp/tmp2$$ || true
  fi
  i=`expr $i + 1`
done
echo ""
rm /tmp/tmp[012]$$
exit
# */

#if TEST1

void dummy1()
{
  return;
}

int dummy2(void)
{
  int i;
  i = 1;

  dummy1();

  if(i == 2)
    return -1;
  else if(i == 3) {
    return i;
  }

  return i + 1;
}

void dummy3(void)
{
  int i;

  i = 1;
error1:
error2:
error3:
  return;
}

#elif TEST2

static void dummy4(void)
{
  int i;

  i = 1;
  /*  missing return  */
}

void dummy5(void)
{
  int i = 1;
}

static void dummy6(void *x,int *bad1,int bad2[],int iii,double **bad3[],
                   double ddd,float fff,long ll,const char *sss,char ccc)
{
  int i;
  i = 1;
}

#elif TEST3

static struct bongo dummy7(void)
{
static struct bongo x;

  return x;
}

typedef struct bongo {
  int dummy;
} bongo_t;

static bongo_t dummy8(void)
{
static bongo_t x;

  return x ;
}

static void dummy9(int depth)
{
  if(--depth > 0)
    dummy9(depth);
  return ;
}

#elif TEST4

/*  Tricky case when a macro expansion is attempted, but later abandoned  */
@define max(a,b)    ((a) > (b) ? (a) : (b))

int dummy10(void)
{
  int max = 1;

  return (max);
  return (max) ;
}

#elif TEST5

/*  Some other tricky cases  */
@define mystrcmp(a,b)       strcmp(a,b)

@define NULL             0L

int dummy11(char *a,char *b)
{
  if(a != NULL && b != NULL)
    return mystrcmp(a,b);

  if(a != NULL)
    return mystrcmp     (a,"wanda")        ;

  if(b != NULL)
    return mystrcmp /*  a stupid comment */
    (a,"wanda")  /*  another one  */      ;

  return mystrcmp
    (
     "one",
     "two"
     );
}

#elif TEST6

typedef char CHAR;

int dummy12(CHAR *s)
{
  return 0;
}

int dummy13(const CHAR *s)
{
  return 0;
}

#elif TEST7

enum Dummy {
  Unknown = -1, One = 1, Two, Three
};

enum Dummy dummy14(int arg)
{
  if(arg == 1)
    return One;
  else if(arg == 2)
    return Two;
  else if(arg == 3)
    return Three;
  else
    return Unknown;
}

#elif TEST8

/*  That is bad style, but we have to do it properly  */
@define NEWINT     (int *)malloc(sizeof(int));

extern void *malloc(int);

int *dummy15()
{
  int *ptr = NEWINT;

  return ptr;
}

#elif TEST9

/*  Ugly K&R style  */
dummy16(a,b,c)
double a;
int b;
void *c;
{
  return b;
}

/*  Even worse - leave out some parameter declarations  */
dummy17(p, m)
int *p;
{
  if(p == 0L) return 0;
  while(*p)
  {
    if(*p++ == m)
      return 1;
  }
  return 0;
}

#elif TEST10

/*  Have return statement within preprocessor definition
 *  - cannot be patched
 */
@define RET(x)   if(x) { return -1; } else { return x - 1; }
@define RET2     return 2

int dummy18(int arg)
{
  RET(arg);

  RET2;
}

#elif TEST11

@pragma __JSCPP__ traceformatcall   TRACE((__FILE__,__LINE__,"%m%n","%f"%a))
@pragma __JSCPP__ traceformatreturn TRACE((__FILE__,__LINE__,"%m%P%n","%f"%p%a))

int dummy19(int arg)
{
  return 1 + 2;
}

#elif TEST12

@pragma __JSCPP__ trace dummy20, dummy21 off
@pragma __JSCPP__ trace "dummy*1", "dummy*2"

void dummy20()
{
}

void dummy21()
{
}

void dummy22()
{
}

@pragma __JSCPP__ trace "dummy*3" call
@pragma __JSCPP__ trace "dummy*4" return

int dummy23(int arg)
{

}

int dummy24(int arg)
{
  if(arg < 0) return 0;
  else return 1;
}

@pragma __JSCPP__ trace dummy25 call() return()

int dummy25(int arg)
{
  return 1 + arg;
}

#elif TEST13

dummy26()
{
  return 1 + 1;
}

int dummy27()
{
  return 1 + 1;
}

long int dummy28()
{
  return 1 + 1;
}

static double dummy29 (int a,int b)
{
  return 1 + 1;
}

static int *dummy30()
{
  return 1 + 1;
}

int (*dummy31())(int a,int b)
{
  return 1 + 1;
}

#elif TEST14

@pragma __JSCPP__ trace dummy32 call(b,d)

int dummy32(int a, int b, int c, int d)
{
  return 0;
}

@pragma __JSCPP__ trace dummy33 \
	call((a + b) / (c + d), c, "D" d) \
	return("" a,"" b, @ - a,"" @ + b)

int dummy33(int a, int b, int c, int d)
{
  if(a == b)
    return a + c;
  else if(c == d)
    return b * c;

  return a + b;
}

#elif TEST15

@pragma __JSCPP__ trace dummy34 \
	call(a, b->IVal, *(b->String), c->DVal) \
	return(@->IVal,"" @->DVal,"Murks" (int)(@->DVal))

struct bingo {
  int IVal;
  double DVal;
  char *String;
};

typedef struct bingo *BINGO;

struct bingo *dummy34(int a, struct bingo *b, BINGO c)
{
  if(a == 1)
    return (struct bingo *)c;

  return b;
}

#elif TEST16

/*  Test variable tracing  */
@pragma __JSCPP__ vartrace "*" read write

void dummy35(void)
{
  int i,j,k[2],l = 3;
  int *p;

  k[i = l + i] = i = j = 1;

  i++;
  ++j;
  --l;
  l--;

  p = &i;
}

#elif TEST17

@pragma __JSCPP__ vartrace "*" read write

typedef char CHARARRAY[100];

CHARARRAY dummy36(void)
{
  CHARARRAY a = "Wanda";

  return a;
}

typedef struct {
  int Number;
  char *String;
} STRUCT;

STRUCT dummy37(void)
{
  int i;
  STRUCT s = { i, "one" };
}

#elif TEST18

@pragma __JSCPP__ vartrace "*" read write

void dummy38(void)
{
static int X = 4;
  const int CON = 1234;
}

#elif TEST19

@pragma __JSCPP__ trace "*" off
@pragma __JSCPP__ vartrace "*" read write

int dummy39(int arg)
{
  int i,j;

  i = arg * 2;
  j = arg + 3;

  printf("hello, world!\n",i);

  return --j;
}

#elif TEST20

/*  Test a special problem of the parse to produce
 *   correct original tokens
 */
@define good         append(too)
@define append(name) name##bad
@define toobad       long
good value;

#elif TEST21
typedef char MYCHAR, *MYSTRING;
typedef MYCHAR *MYCHARSTRING;

@pragma __JSCPP__ typeformat MYCHAR *     : "r"
@pragma __JSCPP__ typeformat char *       : "t"
@pragma __JSCPP__ typeformat MYCHARSTRING : "u"   
@pragma __JSCPP__ typeformat MYSTRING     : "w"   
@pragma __JSCPP__ typeformat MYCHAR       : "x"   

MYSTRING dummy40(char c, MYCHARSTRING s1, char *s2)
{
  return s2;
}

@pragma __JSCPP__ typeformat char : "y"

int dummy41(MYCHAR c)
{
  return c;
}

#elif TEST22

typedef struct {
  int Value;
  const char *String;
} mystruct1_t;

@pragma __JSCPP__ typemap mystruct1_t * : "Value" @->Value, "Strong" @->String
int dummy42(int a, mystruct1_t *b, int c, mystruct1_t *d)
{
  return a + c;
}

#elif TEST23

typedef struct {
  int Value;
  const char *String;
} mystruct2_t;

@pragma __JSCPP__ typemap mystruct2_t * : @->Value, @->String

mystruct2_t *dummy43(int a, mystruct2_t *b, int c, mystruct2_t *d)
{
  return b;
}

mystruct2_t dummy44(mystruct2_t *a)
{
  return *a;
}

mystruct2_t *dummy45(mystruct2_t str)
{
  return &str;
}

#elif TEST24

typedef struct {
  int Value;
  const char *String;
} mystruct3_t;

@pragma __JSCPP__ typemap mystruct3_t : @.Value, @.String

mystruct3_t *dummy46(mystruct3_t *a, double x)
{
  return a;
}

mystruct3_t dummy47(mystruct3_t argument)
{
  return argument;
}

mystruct3_t dummy48(mystruct3_t arg)
{
  return *(mystruct3_t *)&arg;
}

mystruct3_t dummy49(mystruct3_t *ptr)
{
  return *ptr;
}

#elif TEST25

@pragma __JSCPP__ typeformat int    : " %d"
@pragma __JSCPP__ typeformat double : " %g"
@pragma __JSCPP__ typeformat float  : " %g"
@pragma __JSCPP__ typeformat char   : " %c"
@pragma __JSCPP__ typeformat char * : " %s"
@pragma __JSCPP__ typeformat __POINTER__  : " %p"

@pragma __JSCPP__ traceformatcall   printf("Enter %n:%f"%a)
@pragma __JSCPP__ traceformatreturn printf("Leave %n:%f"%a)

int dummy50(int ia, double darg, float fff, char cc, char *ss, int *iarr)
{
  return ia + cc;
}

#elif TEST26

struct bingo1 {
  struct bingo2 *Bingo2;
  int Value1;
};

struct bingo2 {
  struct bingo1 *Bingo1;
  int Value2;
};

@pragma __JSCPP__ options indent = 2 recurlevel = 3

@if 0 /*  Let's not produce overkill  */
@pragma __JSCPP__ typemap struct bingo1 * : \
	(@) ? (@)->Bingo2 : 0L, (@) ? (@)->Value1 : -1
@pragma __JSCPP__ typemap struct bingo2 * : \
	(@) ? (@)->Bingo1 : 0L, (@) ? (@)->Value2 : -1
@else
@pragma __JSCPP__ typemap struct bingo1 * : (@)->Bingo2, (@)->Value1
@pragma __JSCPP__ typemap struct bingo2 * : (@)->Bingo1, (@)->Value2
@endif

@pragma __JSCPP__ typeformat struct bingo1 * : "X"
@pragma __JSCPP__ typeformat struct bingo2 * : "Y"

int dummy51(struct bingo1 *ptr1)
{
  return 0;
}

int dummy52(struct bingo2 *ptr2)
{
  return 0;
}

int dummy53(struct bingo1 *ptr1, struct bingo2 *ptr2)
{
  return 0;
}

#elif TEST27

/*  Is the pointer nature properly mainained across a typedef?  */
@pragma __JSCPP__ typeformat char        : " %c"
@pragma __JSCPP__ typeformat char *      : " %s"

typedef char MYCHAR;

int dummy54(MYCHAR *s)
{
  return 0;
}

#elif TEST28

@pragma __JSCPP__ labelformat   tolower("%s")

@pragma __JSCPP__ trace "*" call return label

int dummy55(int i, double d, char c, char *s)
{
  return 0;
}

#elif TEST29

@pragma __JSCPP__ traceformatcall   printf("Level %l: Enter %n")
@pragma __JSCPP__ traceformatreturn printf("Level %l: Leave %n")

@pragma __JSCPP__ tracelevel dummy56, dummy57   7

int dummy56(int a)
{
  return a + 1;
}

double dummy57(double x)
{
  return x * 1.234;
}

char *dummy58(char *s)
{
  return s + 1;
}

#elif TEST30

@pragma __JSCPP__ typemap struct { int A;int *B;} * : (@)->A, (@)->B

int dummy59(struct { int A; int *B; } *ptr)
{
  return 0;
}

int dummy60(struct { int A; int *b; } *ptr)
{
  return 0;
}

#elif TEST31

/*  This makes the "_jstracefname[]" variable unique  */
@line 1 "TestFTrace.c"

@pragma __JSCPP__ trace dummy61 statement

int dummy61(int arg)
{
  int i, j = arg + 1;
  int sum, sum2;

  if(arg < 0)
    return -1;

  if(arg == 0) {
    printf("GOT 0\n");
  }

  for(i = 1 ; i < 10 ; ++i) {
    sum += i;
    sum2 += i * i;
  }

  for( ; j < 10 ; ) {
    sum2 -= 2 * j--;
  }

  return sum;
}

#elif TEST32

/*  This makes the "_jstracefname[]" variable unique  */
@line 1 "TestFTrace.c"

@pragma __JSCPP__ trace dummy62 call return statement

int dummy62(int arg)
{
  int i, j = arg + 1;
  int sum, sum2;

  if(arg < 0)
    return -1;

  if(arg == 0) {
    printf("GOT 0\n");
    printf("GOT 0\n");
  } else
    printf("Where is my zero?\n");

  for(i = 1 ; i < 10 ; ++i) {
    sum += i;
    sum2 += i * i;
  }

  for( ; j < 10 ; )
    sum2 -= 2 * j--;

  if(sum2 & 1) {
    return sum;
  } else
    return sum2;
}

#elif TEST33

@pragma __JSCPP__ vartrace "*" read write

typedef char TEXT[100];
void dummy63(void)
{
  static int bad = 1;
  const int x = 222;
  TEXT a = "123";
  char b[] = "abadf";
  int i = 1;
  printf("Hello\n");
}

#elif TEST34

@pragma __JSCPP__ trace "*" wrap

typedef struct { int Dummy; } FILE;
extern int fputs(const char *, FILE *);
extern int strlen(const char *);

int dummy64(FILE *fp, int arg)
{
  fputs("here", fp);
  if(arg > 0)
    return strlen("hello");

  return strlen("hello");
}

#elif TEST35

extern char *strcpy(char *dest, const char *src);
extern char *strstr(const char *haystack, const char *needle);

@pragma __JSCPP__ trace "*" wrap call

int dummy65(void)
{
  char buffer[100];

  strcpy(buffer, "hello");
  return 1;
}

@pragma __JSCPP__ trace "*" wrap return

char *dummy66(void)
{
  return strstr("Hello world", "world");
}

#elif TEST36

/*  Test the internal generation of integer constants
 *   The result is irrelevant, but it shouldn't abort
 */
struct dummy77 {
  int a,b,c,d;
  double x;
};

static double *Dummy78[7][9];
static double Dummy79[45];

enum {
  AAA = sizeof(int),
  BBB = sizeof(Dummy78),
  CCC = sizeof(Dummy79),
  DDD = sizeof(struct dummy77 *),
  EEE = sizeof(struct dummy77),
} dummy80_e;

int main(void)
{
  printf("%d %d %d %d %d\n", AAA,BBB,CCC,DDD, EEE);
}

#elif TEST37

/*  This makes the "_jstracefname[]" variable unique  */
@line 1 "TestFTrace.c"

/*  Switch off any internal mappings and substitute by 'printf' style  */
@pragma __JSFTRACE__ typeformat off

@pragma __JSFTRACE__ typeformat char        : " %c"   " %s=%c"
@pragma __JSFTRACE__ typeformat short       : " %hd"  " %s=%hd"
@pragma __JSFTRACE__ typeformat int         : " %d"   " %s=%d"
@pragma __JSFTRACE__ typeformat long        : " %ld"  " %s=%ld"

@pragma __JSFTRACE__ typeformat float       : " %g"   " %s=%g"
@pragma __JSFTRACE__ typeformat double      : " %g"   " %s=%g"

@pragma __JSFTRACE__ typeformat char *      : " %s"   " %s=%s"
@pragma __JSFTRACE__ typeformat __POINTER__ : " %p"   " %s=%p"

@pragma __JSFTRACE__ traceformatcall       printf("Calling %n(%f)\n"%a)
@pragma __JSFTRACE__ traceformatreturn     printf("Leaving %n(%f)\n"%a)
@pragma __JSFTRACE__ traceformatstatement  printf("Statement at line %%d, file '%%s'\n", %L, %F)

/*  Enable tracing of statements, function entry and exit. Add labels */
@pragma __JSFTRACE__ trace "*" call return statement label

int dummy78(int n)
{
  int i, isquare;

  for(i = 1 ; i < n ; ++i) {
    isquare = i * i;
    printf("i=%d, i^2=%d\n", i, isquare);
  }

  return 0;
}

#endif
