# /*
#
#   General test file for function separator. Run through "jscpp -m fctsep"
#    and define one of TEST* symbols below
#
#   This file may be used as a shell script as well (for SH).
#

DIR=test_fctsep
if [ ! -d $DIR ] ; then
  mkdir $DIR
fi

#  Do all tests
max=6
i=1
while [ $i -le $max ] ; do
  echo -n " $i.."
  tfile=$DIR/test$i.i

  jscpp -x0 -CCP -DTEST$i TestFctSep.c -o /tmp/tmp0$$
  sed -e '/^ *$/d' -e 's/^@/#/g' </tmp/tmp0$$ >/tmp/tmp1$$
  jscpp -m fctsep /tmp/tmp1$$ -o /tmp/tmp2$$

  if [ ! -r $tfile ] ; then
    echo "  --> Creating test file"
    cp /tmp/tmp2$$ $tfile
  elif ! cmp -s $tfile /tmp/tmp2$$ ; then
    echo ""
    diff $tfile /tmp/tmp2$$ || true
  fi
  i=`expr $i + 1`
done
echo ""
rm /tmp/tmp[012]$$
exit
# */

#ifdef TEST1
extern int printf(const char *format, ...);

int main(int argc,char *argv[])
{
  int i;

  for(i = 0 ; i < 10 ; ++i)
@pragma separate helloworld  /*  simple block, no var usage  */
  {
    printf("hello, world\n");
  }
  return 0;
}
#elif TEST2
extern int printf(const char *format, ...);

int main(int argc,char *argv[])
{
  int i;

  for(i = 0 ; i < 10 ; ++i)
@pragma separate print_i  /*  simple block, use variable  */
  {
    printf("i = %d\n", i);
  }
  return 0;
}
#elif TEST3
extern int printf(const char *format, ...);

int main(int argc,char *argv[])
{
  int i;

  for(i = 0 ; i < 10 ;)
@pragma separate print_i_and_increment  /*  simple block, change variable  */
  {
    printf("i = %d\n", i);
    ++i;
  }
}
#elif TEST4
extern int printf(const char *format, ...);

int main(int argc,char *argv[])
{
  int i;

  for(i = 0 ; i < 10 ; ++i)
@pragma separate submain /*  leave block somewhere else  */
  {
    if(i == 7)
      break;

    if(i == 8)
      continue;

    printf("i = %d\n", i);

    if(i == 11)
      return 10;

    if(i == 12)
      return -i;

    if(i == 13)
      return -1;
  }
  return 0;
}
#elif TEST5
extern int printf(const char *format, ...);

int main(int argc,char *argv[])
{
  long i,j,k,l,n;

  if(argc == 0) goto label1; /*  Here we go!  */

  n = 100;
  for(i = 1 ; i <= n ; ++i)
@pragma separate submain
  {
    long *ptr;
    int m;

    printf("i %d\n", i);
    l = m = 4711;
label1:
    if(i == 10) continue;  /*  That's enough  */
    if(i == 11) return -33; /*  That's bad  */
    if(i == 12) return 555; /*  That's good  */
    ptr = &n;
    printf("i^2 %d\n", i * i);

label2:
    if(i == 15) break; /*  Bye, bye, I am leaving  */

    if(i > 1000) goto error; /*  That's too much!  */

    k = i;
    printf("i^3 %d\n", i * i * i);

    if(i == 12) break;
    printf("m is %d\n", m);
  }

  if(k < 2) goto label1; /*  That's ugly  */
  if(n == 999) goto label2; /*  That's ugly as well  */

  return 0;
error:
  return -1;
}
#elif TEST6
extern int printf(const char *format, ...);

int main(int argc,char *argv[])
{
  int i, j, k;

  j = 10; k = 100;
  for(i = 0 ; i < 10 ; ++i)
@pragma __JSCPP__ separate print_i local(i) reference(j JJ, k) /*  just a test */
  {
    printf("i = %d, j = %d, k = %d\n", i, j, k);
  }
  return 0;
}
#endif
