# /*
#   General test file for parser. Define one of the symbols
#    TEST1 ... TESTn and run through jscpp.
#
#   This file may be used as a shell script as well (for SH).
#   Call with any -D arguments that are needed to mimick the systems
#    compiler (Linux "-D__i386__") or with additional options for
#    jscpp (like "-w").
#

DIR=test_parser
if [ ! -d $DIR ] ; then
  mkdir $DIR
fi

#  Do all tests
max=11
i=1
while [ $i -le $max ] ; do
  echo -n " $i.."
  tfile=$DIR/test$i.i

  jscpp -m fdeps -DTEST$i -Xprtree -o /dev/null TestParser.c >/tmp/tmp$$

  if [ ! -r $tfile ] ; then
    echo "  --> Creating test file"
    cp /tmp/tmp$$ $tfile
  elif ! cmp -s $tfile /tmp/tmp$$ ; then
    echo ""
    diff $tfile /tmp/tmp$$ || true
  fi
  i=`expr $i + 1`
done
echo ""
rm /tmp/tmp$$
exit
# */

/* ---------------------------------------------------------------------- */
#ifdef TEST1
/* ***  Recognize typedef  *** */
typedef int FILE;
FILE *fp;
/* ---------------------------------------------------------------------- */
#elif defined(TEST2)
/* ***  Pointers and complex declarations  *** */
typedef void *FILE;
char *point,**xxx,c,x;
double (*Function)(short,float);
int fprintf(FILE*,const char *,...);
double xsin(double **argu);
struct test {
  int i,j;
  double (*func)(short a,long *b);
} *sPoint,str;

/* ---------------------------------------------------------------------- */
#elif defined(TEST3)
/* ***  Complicated structures  *** */
struct hans {
  int i,j;
  union {
    float f;
    double d;
  } c;
  struct bongo bo;
} *p,**p2;

/* ---------------------------------------------------------------------- */
#elif defined(TEST4)
/* ***  A program  *** */
int main(int argc,char *argv[])
{
  double x;
  double y,z;
  printf("hello, world!\n");
  return(0);
}
dummy()
{
  return(0);
}

k_and_r_routine(a,b,c)
int a,b;
void *c;
{
  printf("Once upon a time in %d and %d\n",a+b,*c);
}

lazyStyle(first_int,second_int)
{
  return(0);
}

/* ---------------------------------------------------------------------- */
#elif defined(TEST5)
/* ***  Test all valid declaration specifiers  *** */
void *a;

char b;
unsigned char b1;
signed char b2;

short c;
unsigned short c1;
signed short c2;
unsigned short int c3;
signed short int c4;

long d;
unsigned long d1;
signed long d2;
unsigned long int d3;
signed long int d4;

int e;
unsigned int e1;
signed int e2;
unsigned e3;
signed e4;

float f;
long float f1;

double g;
float extern h;

int typedef Garbled;

typedef double Double;
typedef int    Integer;

struct h;
struct h1 {
  Double  D;
  Integer I;
};
typedef struct h HType;
typedef struct h2 {
  void *empty;
  char *string;
} H2Type;

union i;
union i1 {
  int First;
  double Second;
  void * Third;
};
typedef union i IType;
typedef union i2 {
  int  Get;
  long United;
} I2Type;

enum j;
enum j1 {
  One = 1, Two, Three
};
typedef enum j JType;
typedef enum j2 {
  Four = 4, Five, Six
} J2Type;

/* ---------------------------------------------------------------------- */
#elif defined(TEST6)
/*  Test name space handling:
 *   1. typedef redefinition in inner block to identifier or enumeration
 *      constant.
 *   2. goto labels identical to typedef
 */
typedef char   One;
typedef int    Two;
typedef long   Three;
typedef float  Four;
typedef double Five;
typedef unsigned char Six;

/*  Test if typedef is visible at the end of the declarator  */
typedef float Tarzan,Jane(Tarzan);

int test(void)
{
  typedef unsigned char One;
  typedef int (*Two)(One arg);
  struct Bongo {
    int A,B,Three;
    enum Bingo {
      /*  These redefinitions of typedefs must be visible in function block */
      Three, Four
    } XX;
  } BB;
  unsigned int Five;

One:
  /*  Test if redefined names are accepted  */
  BB.Three = 1;
  Five = Three;
  Five = Four;

  if(Five != BB.Three) goto One;
  else goto Six;

  return(0);

Six:
  return(-1);
}
/*  Test if typedefs retain their meaning outside function block  */
Three X,Y;
Four K,L;
/* ---------------------------------------------------------------------- */
#elif defined(TEST7)
/* ***  Check if "if .. else" conflict is resolved correctly  *** */
int main(void)
{
  if(1 == 2)
    if(1 == 4)
      test1();
    else
      test2();
  else if(2 == 3)
    test3();
}

/* ---------------------------------------------------------------------- */
#elif defined(TEST8)
/* ***  Check if reusage of variable definitions happens correctly  *** */
int func(int i,int j);
int func2(int i,int j);

main()
{
  int i,j;
}

/* ---------------------------------------------------------------------- */
#elif defined(TEST9)
/* ***  Check "__typeof__" and "__ignore__" keyword  *** */
int *a;
#ifdef CHECK_TYPEOF
__typeof__(*a) b; /*  b has now type "int"  */
#endif

/*  These constructions appear with "gcc" in system header files  */
#define __attribute__(arg)  __ignore__(__attribute__(arg))

int dummy() __attribute__((whatsoever))
{
  printf("hello, world!\n");
}

/* ---------------------------------------------------------------------- */
#elif defined(TEST10)
typedef float T;
int dodo      (const (T[5]));
int dodo_equiv(const id1(T id2[5]));
int dudu      (const (T(int)));
int dudu_equiv(const id1(T id2(int id3)));

/* ---------------------------------------------------------------------- */
#elif defined(TEST11)
/*  Test precedence of all operators  */
main()
{
  int i;

  i = 1 || 2 && 3 | 4 ^ 5 & 6 == 7 != 8 < 9 > 10 <= 11 >= 12 << 13 >> 14
    + 15 * 16 / 17 % 18;
}

#endif
