/*  Show some properties of CURSES library  Joerg Schoen 1998  */
#include <ctype.h>
#include <stdlib.h>

#include <curses.h>

#if defined(A_COLOR)
#define HAS_COLORS
#endif

#define GET_COLORBG(color)    ((((color) - 1) / 8) % 8)
#define GET_COLORFG(color)    (((color) - 1) % 8)

int main(void)
{
  const char *string;
  chtype c;
  int count;

  /*  Initialize screen  */
  initscr();
  cbreak();
  nonl(); noecho();
  keypad(stdscr, TRUE);

  printw("Screen has %d lines and %d columns.\n", LINES, COLS);

  /* ********************************************************************** */
  printw("Relevant environment variables:\n");
#define SHOWENV(var) \
  if((string = getenv(var)) != NULL) \
    printw("\t%s=%s\n", var, string)

  SHOWENV("TERM");
  SHOWENV("LINES");
  SHOWENV("COLS");
  SHOWENV("COLORTERM");
  printw("\n");

  /* ********************************************************************** */
#ifdef HAS_COLORS
  start_color();
  if(has_colors()) {
    int i, j;

    /*  Initialize color pairs with all possible combinations  */
    for(i = 1 ; i < COLOR_PAIRS ; i++) {
      short f, b;
      f = (i - 1) % COLORS;
      b = ((i - 1) / COLORS) % 8;
      init_pair(i, f, b);
    }

    printw("We have %d colors and %d color pairs. Colors can%s be changed.\n",
	   COLORS, COLOR_PAIRS, can_change_color() ? "" : "not");
    for(c = 0 ; c < 2 ; c++) {
      for(i = 1 ; i < COLOR_PAIRS ; i++) {
	char buffer[40];
        short f, b;

        b = GET_COLORBG(i); if(b < 0) b = 0;
        f = GET_COLORFG(i); if(f < 0) f = 0;
	sprintf(buffer, "Col(%1d,%1d)", b, f);
	for(j = 0 ; buffer[j] ; j++)
	  addch((c ? A_BOLD : 0) | COLOR_PAIR(i) | buffer[j]);
	if((i + 1) % 9 == 1) addch('\n');
      }
      printw("\n");
    }
  } else {
    printw("CURSES library has colors, this terminal not.\n");
  }
#else
  printw("CURSES library has no colors.\n");
#endif

  /* ********************************************************************** */
  printw("Attributes:");
#define SHOWATTR(attr, name)	printw(" "); \
				attron(attr); \
				printw(#name); \
				attroff(attr)

#ifdef A_STANDOUT
  SHOWATTR(A_STANDOUT, standout);
#endif
#ifdef A_UNDERLINE
  SHOWATTR(A_UNDERLINE, underline);
#endif
#ifdef A_REVERSE
  SHOWATTR(A_REVERSE, reverse);
#endif
#ifdef A_BLINK
  SHOWATTR(A_BLINK, blink);
#endif
#ifdef A_DIM
  SHOWATTR(A_DIM, dim);
#endif
#ifdef A_BOLD
  SHOWATTR(A_BOLD, bold);
#endif
#ifdef A_PROTECT
  SHOWATTR(A_PROTECT, protect);
#endif
#ifdef A_INVIS
  SHOWATTR(A_INVIS, invisible);
#endif

#undef SHOWATTR
  printw("\n");

  /* ********************************************************************** */
  /*  Check for additional characters  */
  printw("Additional characters:\n\t'");
#define SPACE    printw("' '")
#define SPACE2   printw("'\n\t'")
#ifdef ACS_ULCORNER
  addch(ACS_ULCORNER); SPACE;
#endif
#ifdef ACS_LLCORNER
  addch(ACS_LLCORNER); SPACE;
#endif
#ifdef ACS_URCORNER
  addch(ACS_URCORNER); SPACE;
#endif
#ifdef ACS_LRCORNER
  addch(ACS_LRCORNER); SPACE;
#endif
#ifdef ACS_LTEE
  addch(ACS_LTEE); SPACE;
#endif
#ifdef ACS_RTEE
  addch(ACS_RTEE); SPACE;
#endif
#ifdef ACS_BTEE
  addch(ACS_BTEE); SPACE;
#endif
#ifdef ACS_TTEE
  addch(ACS_TTEE); SPACE;
#endif
#ifdef ACS_HLINE
  addch(ACS_HLINE); SPACE;
#endif
#ifdef ACS_VLINE
  addch(ACS_VLINE); SPACE;
#endif
#ifdef ACS_PLUS
  addch(ACS_PLUS); SPACE;
#endif
#ifdef ACS_S1
  addch(ACS_S1); SPACE;
#endif
#ifdef ACS_S9
  addch(ACS_S9); SPACE;
#endif
#ifdef ACS_DIAMOND
  addch(ACS_DIAMOND); SPACE;
#endif
#ifdef ACS_CKBOARD
  addch(ACS_CKBOARD); SPACE;
#endif
#ifdef ACS_DEGREE
  addch(ACS_DEGREE); SPACE2;
#endif
#ifdef ACS_PLMINUS
  addch(ACS_PLMINUS); SPACE;
#endif
#ifdef ACS_BULLET
  addch(ACS_BULLET); SPACE;
#endif
#ifdef ACS_LARROW
  addch(ACS_LARROW); SPACE;
#endif
#ifdef ACS_RARROW
  addch(ACS_RARROW); SPACE;
#endif
#ifdef ACS_DARROW
  addch(ACS_DARROW); SPACE;
#endif
#ifdef ACS_UARROW
  addch(ACS_UARROW); SPACE;
#endif
#ifdef ACS_BOARD
  addch(ACS_BOARD); SPACE;
#endif
#ifdef ACS_LANTERN
  addch(ACS_LANTERN); SPACE;
#endif
#ifdef ACS_BLOCK
  addch(ACS_BLOCK); SPACE;
#endif
#ifdef ACS_S3
  addch(ACS_S3); SPACE;
#endif
#ifdef ACS_S7
  addch(ACS_S7); SPACE;
#endif
#ifdef ACS_LEQUAL
  addch(ACS_LEQUAL); SPACE;
#endif
#ifdef ACS_GEQUAL
  addch(ACS_GEQUAL); SPACE;
#endif
#ifdef ACS_PI
  addch(ACS_PI); SPACE;
#endif
#ifdef ACS_NEQUAL
  addch(ACS_NEQUAL); SPACE;
#endif
#ifdef ACS_STERLING
  addch(ACS_STERLING);
#endif
  printw("'\n\n");

  /* ********************************************************************** */
#ifdef NCURSES_MOUSE_VERSION
  c = mousemask(ALL_MOUSE_EVENTS|REPORT_MOUSE_POSITION, NULL);

  printw("NCURSES: We can use mouse, ver %d, mask %#x\n",
	 NCURSES_MOUSE_VERSION, c);
#endif

  /* ********************************************************************** */
  printw("Press any key to show key value, 'q' to quit!\n");

  count = 0;
  while(1) {
    refresh();
    c = getch();

#ifdef NCURSES_MOUSE_VERSION
    if(c == KEY_MOUSE) {
      MEVENT mev;
      if(getmouse(&mev) == ERR) goto error;
      printw("\r%5d: MOUSE id %d state %x: x=%d y=%d", count++,
	     mev.id, mev.bstate, mev.x, mev.y);
    } else
#endif
    {
      if(c == 'q') break;

      printw("\rKEY %4ld %#4o '%c'", c, c, (c < 256 && isprint(c)) ? c : '.');
    }

    clrtoeol();
  }

  endwin();

  return 0;
error:
  endwin();
  perror("testcurses");
  return 30;
}
