/*JS*********************************************************************
*
*    Program : TESTEDLIB
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Test editing library.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <jsconfig.h>

#ifndef CONFIG_NO_POSIX
#include <unistd.h>
#endif

#include <jssubs.h>

#include <jedlib.h>

/*********     DEFINES                                          *********/
#define MULTI  3


/*********     PROTOTYPES                                       *********/
static int getLine(void *s, long nr, char *dest, int m);

/*********     STATIC VARIABLES                                 *********/
static long Toggle;

/* **  An example menu definition  ** */
#define MAXDEPTH   4
static struct {
  const char *Names[MAXDEPTH];
  int NNames;
  long Type, Value, *Ptr;
} Menues[] = {
  { { "File"      }, 1, MENU_SUBMENU },
  { { "Edit"      }, 1, MENU_SUBMENU },
  { { "Form"      }, 1, 0, 3 },
  { { "Table"     }, 1, 0, 4 },
  { { "Directory" }, 1, 0, 5 },

  { { "File", "New"     }, 2, 0, 11 },
  { { "File", "Open"    }, 2, 0, 12 },
  { { "File", "Save as" }, 2, 0, 13 },
  { { "File", "Save"    }, 2, 0, 14 },
  { { "File", "Type"    }, 2, MENU_SUBMENU },
  { { "File", "Quit"    }, 2, 0, 16 },

  { { "Edit", "Undo"    }, 2, 0, 21 },
  { { "Edit", "Cut"     }, 2, 0, 22 },
  { { "Edit", "Copy"    }, 2, 0, 23 },
  { { "Edit", "Paste"   }, 2, 0, 24 },
  { { "Edit", "Save"    }, 2, 0, 25 },
  { { "Edit", "Toggle0" }, 2, MENU_VARIABLE + 0, 0, &Toggle },
  { { "Edit", "Toggle1" }, 2, MENU_VARIABLE + 1, 0, &Toggle },
  { { "Edit", "Toggle2" }, 2, MENU_VARIABLE + 2, 0, &Toggle },

  { { "File", "Type", "ASCII"   }, 3, 0, 151 },
  { { "File", "Type", "Binary"  }, 3, 0, 152 },
  { { "File", "Type", "Archive" }, 3, 0, 153 },
  { { "File", "Type", "User"    }, 3, MENU_SUBMENU },

  { { "File", "Type", "User", "One"   }, 4, 0, 1551 },
  { { "File", "Type", "User", "Two"   }, 4, 0, 1552 },
  { { "File", "Type", "User", "Three" }, 4, 0, 1553 },
  { { "File", "Type", "User", "Four"  }, 4, 0, 1554 },
  { { "File", "Type", "User", "Five"  }, 4, 0, 1555 },
};

/* ***  A sample form  *** */
#define LEN  3000
#define LEN2 25
#define N    18
static char Lines[N][LEN] = {
  "Surname",     "Schoen",
  "Firstname",   "Joerg",
  "Age",         "40",
  "Size",        "1.84",
  "Property",    "",
  "Password",    "",
  "Bemerkung",   "This is the one who is responsible\nfor this programming stuff!\n"
};

const char *const Props[] = {
  "One", "Two", "Three", "Four", "Five", "Six", "Seven", "Eight", "Nine",
  "Ten", "Eleven", "Twelve", "Thirteen", "Fourteen", "Fifteen", "Sixteen",
  "Seventeen", "Eighteen", "Nineteen", "Twenty", "Twenty-one", "Twenty-two",
  "Twenty-three", "Twenty-four", "Twenty-five"
};

static formfield_t Forms[N] = {
 { FOFI_FNONSEL | FOFI_FREFRESH | FOFI_FEMPHASIZE | FOFI_TASCII,
   0,   0, 1, -1, Lines[3], -1, -1, NULL, NULL },
 { FOFI_FNONSEL | FOFI_FREFRESH | FOFI_FEMPHASIZE | FOFI_TASCII,
   0,  10, 1, -1, Lines[1], -1, -1, NULL, NULL },

 { FOFI_FNONSEL | FOFI_TASCII, 2,  0, 1,  -1, Lines[0],  -1, -1, NULL, NULL  },
 { FOFI_TASCII | FOFI_FSHOWSIGNS, 2, 10,  1, LEN2, Lines[1],  -1, LEN, Lines[0],
   NULL },

 { FOFI_FNONSEL | FOFI_TASCII, 3,  0, 1,  -1, Lines[2],  -1, -1, NULL, NULL  },
 { FOFI_TASCII,                3, 10, 1, LEN2, Lines[3],  -1, LEN, Lines[2],
   NULL },

 { FOFI_FNONSEL | FOFI_TASCII, 4,  0, 1,  -1, Lines[4],  -1, -1, NULL, NULL  },
 { FOFI_TINUM,                 4, 10, 1, LEN2, Lines[5],  -1, LEN, Lines[4],
   NULL },

 { FOFI_FNONSEL | FOFI_TASCII, 5,  0, 1,  -1, Lines[6],  -1, -1, NULL, NULL  },
 { FOFI_TFNUM,                 5, 10, 1, LEN2, Lines[7],  -1, LEN, Lines[6],
   NULL },

 { FOFI_FNONSEL | FOFI_TASCII, 6,  0, 1,  -1, Lines[8],  -1, -1, NULL  },
 { FOFI_TENUM,                 6, 13, 1, LEN2, (char *)Props, 0,
   sizeof(Props)/sizeof(Props[0]), Lines[8], NULL },

 { FOFI_FNONSEL | FOFI_TASCII, 7,  0, 1,  -1, Lines[10],  -1, -1, NULL  },
 { FOFI_TASCII | FOFI_FHIDDEN, 7, 10,  3,  40, Lines[11],  -1, LEN, Lines[10],
   NULL },

 { FOFI_FNONSEL | FOFI_TASCII, 8,  0, 1,  -1, Lines[12],  -1, -1, NULL  },
 { FOFI_TTEXT | FOFI_FSHOWSIGNS, 8, 10,  3,  40, Lines[13],  -1, LEN, Lines[12],
   NULL },

 { FOFI_TBUTTON | 0,          13, 12, 2,  -1, "<OK>", -1, -1, NULL  },
 { FOFI_TBUTTON | 1,          13, 22, 2,  -1, "<CANCEL>", -1, -1, NULL  }
};

/*JS*********************************************************************
*
*************************************************************************/

int main(int argc, char *argv[])

/************************************************************************/
{
  int mode, y, ret, edinit;
  int nLines, height;
  jedmenufld_t *menu;

  edinit = 1;
  if(argc > 1 && strcmp(argv[1], "-C") == 0) {
    --argc; ++argv;
    edinit |= 2 | 4 | 8 | 16 | 32;
  }

  if(jedinitialize(edinit) == -1)
    EGOTO(error);

  nLines = 150 * MULTI;
  height = LINES - 3;

  mode = 0;
  ret = -1;
  if(argc == 2) {
    if(strcmp(argv[1], "form") == 0)
      ret = 3;
    else if(strcmp(argv[1], "form2") == 0)
      ret = 6;
    else if(strcmp(argv[1], "table") == 0)
      ret = 4;
    else if(strcmp(argv[1], "directory") == 0)
      ret = 5;
  }

  menu = NULL;
  for(y = 0 ; y < sizeof(Menues) / sizeof(Menues[0]) ; ++y) {
    jedmenufld_t *ent;

    if((ent = jedmenufldadd(&menu, Menues[y].Names, Menues[y].NNames,
		      Menues[y].Type)) == NULL)
      goto error;

    if(Menues[y].Ptr) ent->MF_Variable = Menues[y].Ptr;
    else ent->MF_Value = Menues[y].Value;
  }

  if(jedmenudokey(0, 0, menu, -1, mode | MENUM_PRINTONLY)) goto error;
  mvaddstr(3, 1, "Press 'm' to enter menu, 'q' to quit.\n");

  y = 4;
  if(ret != -1) jedwstacktouch(stdscr);
  for(;; ret = -1) {
    int c;

    /*refreshWinStack(7);*/
    if(ret == -1) {
      c = jedwstackgetch(stdscr);
      if(c == 'q' || c == 'Q') break;

      if(c == 'f')
	ret = 3;
      else if(c == 't')
	ret = 4;
      else if(c == 'F')
	ret = 6;
      else {
	if(c != 'm'
#ifdef NCURSES_MOUSE_VERSION
	   && c != KEY_MOUSE
#endif
	   ) {
	  beep();
	  continue;
	}

	ret = jedmenudokey(0, 0, menu,
#ifdef NCURSES_MOUSE_VERSION
		       c == KEY_MOUSE ? c :
#endif
		       -1, MENUM_NEWWIN | mode);
      }
    }

    if(ret == 3 || ret == 6) {
      ret = jedformdisplay(-1, -1, ret == 6 ? 7 : 14, 60, Forms, N, FOFI_MFRAME);
    } else if(ret == 4) {
      ret = jedtabdisplay("Testtable", 2, 2, height, 60, MULTI,
			  nLines, getLine, NULL, NULL, NULL,
			  FOFI_MFRAME | FOFI_NUMTABLE);
#ifndef CONFIG_NO_POSIX
    } else if(ret == 5) {
      char cwd[JS_FILENAME_MAX];
      const char *s;

      if(getcwd(cwd, sizeof(cwd)) == NULL) goto error;
      s = jedshowdirectory(cwd, -1, -1, LINES - 2, COLS - 2, &ret,
			/*FOFI_MPOINTFILES | */FOFI_MFRAME);
      mvprintw(y, 1, "  File '%s'\n", s);
      y++;
#endif
    }

    mvprintw(y, 1, "Returned %d; Toggle %d\n", ret, Toggle);
    y++;

    if(ret == 16) break;
  }

  if(jeddeinitialize() == -1)
    EGOTO(error2);

  return 0;
error:
  jeddeinitialize();
error2:
  jsperror("testedit");
  return 30;
}

/*JS*********************************************************************
*
*************************************************************************/

static int getLine(void *s, long nr, char *dest, int m)

/************************************************************************/
{
  int i;

  i = nr % MULTI;
  nr /= MULTI;
  if(i == 0)
    sprintf(dest, "Wanda %ld hello!!", nr);
  else
    sprintf(dest, "   %ld continued...", nr);

  return 0;
}
