/*JS*********************************************************************
*
*    Program : TESTJSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Sample program to show usage of jscpp library
*
*              Copyright (C) 2006 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>

#include <jscpplib.h>

/*********     DEFINES                                          *********/
#define OPTSTRING         JSCPP_OPTSTRING "h"

/*********     PROTOTYPES                                       *********/
static int dopreprocess(jsppenviron_t *jppenv, const char *name,
			FILE *fpOut);

/*JS*********************************************************************
*
*************************************************************************/

int main(int argc, char *argv[])

/************************************************************************/
{
  const char *string;
  jsppenviron_t ppenv;
  FILE *fp;
  int c;

  if(jsppinit(0) == -1)
    EGOTO(error);

  jsppenvclear(&ppenv);

  /*  Parse options  */
  while((c = jsGetOpt(&argc, argv, OPTSTRING, (const char **)&string))
	!= EOF) {
    switch(c) {
    default:
      /*  Generic option handling for preprocessor  */
      if((c = jsppenvhandle(&ppenv, c, string)) == -1)
	EGOTO(error);

      if(c != -2)
	/*  OK, option handled  */
	break;
      /*FALLTHROUGH*/

    case 'h':
      goto help;
    }
  }

  if(argc < 2) goto help;

  if(argc == 2)
    fp = stdout;
  else if((fp = fopen(argv[2], "w")) == NULL)
    EGOTO(error);

  if(dopreprocess(&ppenv, argv[1], fp) == -1)
    EGOTO(error);

  jsppenvfree(&ppenv);

  return 0;

help:
  fprintf(stderr, "Usage: testjscpp [<opts>] <infile> [<outfile>]\n\
  Sample application for JSCPP library that preprocesses the input file\n\
  to the output file or stdout.\n\
  Standard options:\n%s",
	  JSPPStdOptsHelp);
  return 5;

error:
  jsperror("testjscpp");
  return 40;
}

/*JS*********************************************************************
*
*************************************************************************/

static int dopreprocess(jsppenviron_t *jppenv, const char *name,
			FILE *fpOut)

/************************************************************************/
{
  jsppstream_t *jps;
  jsctoken_t token;

  if((jps = jsppopen2(NULL, jppenv, name, NULL, NULL, 0, 0)) == NULL)
    EGOTO(error);

  while((token = jsppgettoken(jps, NULL, 0)).TK_Value != CTK_EOF) {
    if(PTK_FIRSTINTERN <= token.TK_Value && token.TK_Value <= PTK_LASTINTERN)
      continue;

    if(token.TK_Value == CTK_ERROR)
      EGOTO(error);

    /*  Ignore empty tokens  */
    if(token.TK_Value == PTK_EMPTY) continue;

    /*  Write token to output  */
    if(jspptkprint(jps, fpOut, token) == -1)
      EGOTO(error);
  }

  if(!(jps->JSPP_Mode & JPPM_PRESSPACE))
    /*  Print newline at EOF. The internal logic deferred output
     *   of spaces, so we have to write it here. This deletes any
     *   unnecessary trailing spaces and fixes files that end
     *   without a newline.
     */
    putc('\n', fpOut);

  jsppclose(jps, 0);

  return 0;
error:
  return -1;
}
